Package: tidyILD
Title: Tidy Intensive Longitudinal Data Analysis
Version: 0.4.0
Author: Alex Litovchenko [aut, cre]
Maintainer: Alex Litovchenko <al4877@columbia.edu>
Authors@R: person(given = "Alex", family = "Litovchenko", role = c("aut", "cre"), email = "al4877@columbia.edu")
Description: A reproducible, tidyverse-style framework for intensive longitudinal
  data analysis in R, with built-in methodological safeguards, provenance
  tracking, and reporting tools. Encodes time structure, enforces
  within-between decomposition, provides spacing-aware lags, and integrates
  diagnostics and visualization. Use ild_prepare(), ild_center(), ild_lag(),
  and related functions for a unified pipeline from raw EMA/diary data to
  interpretable models.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: tibble, dplyr, lubridate, rlang, lme4, nlme, ggplot2, mgcv
Suggests: BH, testthat (>= 3.0.0), roxygen2, knitr, rmarkdown, rstan,
        RcppEigen, broom.mixed, clubSandwich, jsonlite, yaml, tsibble,
        brms, KFAS, ctsem
VignetteBuilder: knitr
Collate: 'package.R' 'ild-class.R' 'ild_provenance.R' 'ild_methods.R'
        'ild_compare_pipelines.R' 'ild_compare_fits.R' 'utils.R'
        'ild_schema_tidy_augment.R' 'ild_tidy_helpers.R'
        'ild_augment_helpers.R' 'ild_tsibble.R' 'ild_prepare.R'
        'ild_summary.R' 'ild_center.R' 'ild_decomposition.R'
        'ild_lag.R' 'ild_spacing_class.R' 'ild_spacing.R'
        'ild_design_check.R' 'ild_missing_extensions.R'
        'ild_missing_pattern.R' 'ild_missing_bias.R'
        'ild_missing_model.R' 'ild_ipw_helpers.R' 'ild_ipw_weights.R'
        'ild_iptw_weights.R' 'ild_iptw_msm_weights.R'
        'ild_ipcw_weights.R' 'ild_joint_msm_weights.R'
        'ild_ipw_refit.R' 'ild_msm_bootstrap.R' 'ild_msm_balance.R'
        'ild_tvem.R' 'ild_tvem_plot.R' 'ild_check_lags.R'
        'ild_panel_lag_prepare.R' 'ild_msm_history.R' 'ild_crosslag.R'
        'ild_acf.R' 'ild_align.R' 'ild_lme.R' 'ild_person_model.R'
        'ild_heterogeneity.R' 'ild_robust_se.R' 'ild_model_tidiers.R'
        'ild_guardrail_registry.R' 'ild_missingness_report.R'
        'ild_diagnostics_bundle.R' 'ild_diagnostics_utilities.R'
        'ild_diagnostics.R' 'ild_manifest.R' 'ild_plot.R'
        'ild_generics.R' 'ild_brms.R' 'ild_brms_s3.R'
        'ild_brms_dynamics_formula.R' 'ild_ctsem.R' 'ild_ctsem_s3.R'
        'ild_kfas_spec.R' 'ild_kfas_helpers.R' 'ild_kfas.R'
        'ild_kfas_guardrails.R' 'ild_kfas_tidiers.R'
        'ild_kfas_augment.R' 'ild_kfas_diagnose.R' 'ild_kfas_plot.R'
        'ild_kfas_autoplot.R' 'ild_diagnose_fillers.R'
        'ild_diagnose_methods.R' 'ild_autoplot_bundle.R'
        'ild_circadian.R' 'ild_simulate.R' 'ild_msm_estimand.R'
        'ild_msm_fit.R' 'ild_msm_contrast.R' 'ild_msm_simulation.R'
        'ild_power.R' 'ild_recovery_metrics.R' 'data.R' 'broom.R'
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-18 02:16:17 UTC; alexanderlitovchenko
Repository: CRAN
Date/Publication: 2026-04-18 03:00:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(ild_augment,brmsfit)
S3method(ild_augment,default)
S3method(ild_augment,ild_fit_ctsem)
S3method(ild_augment,ild_fit_kfas)
S3method(ild_augment,lme)
S3method(ild_augment,lmerMod)
S3method(ild_augment_states,ild_fit_kfas)
S3method(ild_autoplot,brmsfit)
S3method(ild_autoplot,default)
S3method(ild_autoplot,ild_diagnostics)
S3method(ild_autoplot,ild_diagnostics_bundle)
S3method(ild_autoplot,ild_fit_ctsem)
S3method(ild_autoplot,ild_fit_kfas)
S3method(ild_autoplot,ild_heterogeneity)
S3method(ild_autoplot,lme)
S3method(ild_autoplot,lmerMod)
S3method(ild_diagnose,brmsfit)
S3method(ild_diagnose,default)
S3method(ild_diagnose,ild_fit_ctsem)
S3method(ild_diagnose,ild_fit_kfas)
S3method(ild_diagnose,lme)
S3method(ild_diagnose,lmerMod)
S3method(ild_heterogeneity,brmsfit)
S3method(ild_heterogeneity,default)
S3method(ild_heterogeneity,lme)
S3method(ild_heterogeneity,lmerMod)
S3method(ild_tidy,brmsfit)
S3method(ild_tidy,default)
S3method(ild_tidy,ild_fit_ctsem)
S3method(ild_tidy,ild_fit_kfas)
S3method(ild_tidy,ild_heterogeneity)
S3method(ild_tidy,lme)
S3method(ild_tidy,lmerMod)
S3method(ild_tidy_states,ild_fit_kfas)
S3method(print,ild_brms_diagnostics)
S3method(print,ild_compare_pipelines)
S3method(print,ild_design_check)
S3method(print,ild_diagnostics)
S3method(print,ild_diagnostics_bundle)
S3method(print,ild_fit_ctsem)
S3method(print,ild_heterogeneity)
S3method(print,ild_msm_bootstrap)
S3method(print,ild_msm_estimand)
S3method(print,ild_msm_fit)
S3method(print,ild_msm_history_spec)
S3method(print,ild_msm_recovery)
S3method(print,tidyild_model)
export(as_ild)
export(augment.ild_lme)
export(augment_ild_model)
export(guardrail_registry)
export(ild_acf)
export(ild_align)
export(ild_as_tsibble)
export(ild_augment)
export(ild_augment_schema)
export(ild_augment_states)
export(ild_autoplot)
export(ild_brms)
export(ild_brms_dynamics_formula)
export(ild_build_msm_history)
export(ild_bundle)
export(ild_center)
export(ild_center_plot)
export(ild_check_lags)
export(ild_circadian)
export(ild_compare_fits)
export(ild_compare_pipelines)
export(ild_crosslag)
export(ild_ctsem)
export(ild_decomposition)
export(ild_design_check)
export(ild_diagnose)
export(ild_diagnostics)
export(ild_diagnostics_bundle)
export(ild_export_provenance)
export(ild_fit)
export(ild_heatmap)
export(ild_heterogeneity)
export(ild_heterogeneity_stratified)
export(ild_history)
export(ild_ipcw_weights)
export(ild_iptw_msm_weights)
export(ild_iptw_weights)
export(ild_ipw_ess)
export(ild_ipw_refit)
export(ild_ipw_weights)
export(ild_joint_msm_weights)
export(ild_kfas)
export(ild_lag)
export(ild_lme)
export(ild_manifest)
export(ild_meta)
export(ild_methods)
export(ild_missing_bias)
export(ild_missing_cohort)
export(ild_missing_compliance)
export(ild_missing_hazard_first)
export(ild_missing_model)
export(ild_missing_pattern)
export(ild_missingness_report)
export(ild_msm_balance)
export(ild_msm_bootstrap)
export(ild_msm_contrast_over_time)
export(ild_msm_diagnose)
export(ild_msm_estimand)
export(ild_msm_fit)
export(ild_msm_history_spec)
export(ild_msm_overlap_plot)
export(ild_msm_recovery)
export(ild_msm_simulate_scenario)
export(ild_panel_lag_prepare)
export(ild_person_distribution)
export(ild_person_model)
export(ild_plot)
export(ild_plot_filtered_vs_smoothed)
export(ild_plot_forecast)
export(ild_plot_predicted_trajectory)
export(ild_plot_states)
export(ild_power)
export(ild_prepare)
export(ild_prior_ild)
export(ild_provenance)
export(ild_recovery_metrics)
export(ild_report)
export(ild_robust_se)
export(ild_simulate)
export(ild_spacing)
export(ild_spacing_class)
export(ild_spaghetti)
export(ild_summary)
export(ild_tidy)
export(ild_tidy_schema)
export(ild_tidy_states)
export(ild_tsibble_meta)
export(ild_tvem)
export(ild_tvem_plot)
export(is_ild)
export(is_ild_diagnostics_bundle)
export(new_ild_diagnostics_bundle)
export(plot_ild_diagnostics)
export(tidy.ild_lme)
export(tidy_ild_model)
export(tidy_ild_msm_bootstrap)
export(validate_ild)
importFrom(dplyr,across)
importFrom(dplyr,all_of)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,group_by)
importFrom(dplyr,lag)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,n)
importFrom(dplyr,row_number)
importFrom(dplyr,select)
importFrom(dplyr,summarise)
importFrom(dplyr,ungroup)
importFrom(ggplot2,aes)
importFrom(ggplot2,facet_wrap)
importFrom(ggplot2,geom_abline)
importFrom(ggplot2,geom_col)
importFrom(ggplot2,geom_hline)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_tile)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,stat_qq)
importFrom(ggplot2,stat_qq_line)
importFrom(ggplot2,theme_minimal)
importFrom(ggplot2,vars)
importFrom(lme4,lmer)
importFrom(nlme,corAR1)
importFrom(nlme,corCAR1)
importFrom(nlme,lme)
importFrom(rlang,.data)
importFrom(rlang,inject)
importFrom(rlang,sym)
importFrom(stats,coef)
importFrom(stats,fitted)
importFrom(stats,formula)
importFrom(stats,predict)
importFrom(stats,rnorm)
importFrom(stats,rstandard)
importFrom(stats,runif)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)

# tidyILD

A **reproducible, tidyverse-style framework** for intensive longitudinal data (ILD) analysis in R, with built-in methodological safeguards, **provenance tracking**, and reporting tools.  
Author: **Alex Litovchenko**.

## Install

```r
remotes::install_github("alitovchenko/tidyILD")
```

## Backend validation benchmarks (developers / CI)

From a **source checkout** of the package (directory containing `DESCRIPTION`), optional cross-backend simulation benchmarks can be run locally:

```bash
Rscript scripts/run-backend-validation-benchmarks.R --tier smoke --out-dir /tmp/bench
Rscript scripts/check-backend-validation-thresholds.R \
  --summary /tmp/bench/benchmark_summary.csv \
  --thresholds inst/benchmarks/thresholds-smoke.json \
  --out /tmp/bench/benchmark_checks.csv
```

This uses `pkgload::load_all()` when available. **Tiers:** `smoke` (fast), `nightly`, `full` — see `inst/dev/BACKEND_VALIDATION_BENCHMARK_CONTRACT.md`. GitHub Actions workflow: `.github/workflows/backend-validation-benchmarks.yml` (scheduled + manual dispatch; artifacts uploaded).

## Quick start

```r
library(tidyILD)

# Prepare: validate time structure, add .ild_* columns and metadata
d <- data.frame(
  id = rep(1:3, each = 5),
  time = rep(as.POSIXct(0:4 * 3600, origin = "1970-01-01"), 3),
  mood = rnorm(15)
)
x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)

# Inspect (summary tibble + list)
ild_summary(x)

# Within-between decomposition
x <- ild_center(x, mood)

# Spacing-aware lags (max_gap from metadata if omitted)
x <- ild_lag(x, mood, mode = "gap_aware", max_gap = 7200)

# Missingness (summary tibble + plot + by_id)
ild_missing_pattern(x, vars = "mood")

# Fit and report: tidy fixed effects, fitted vs observed, residual ACF + QQ
fit <- ild_lme(mood ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
tidy_ild_model(fit)
ild_plot(fit, type = "fitted")
ild_plot_predicted_trajectory(fit, time_var = ".ild_seq") # observed + fitted vs time
diag <- ild_diagnostics(fit); diag; plot_ild_diagnostics(diag)
bundle <- ild_diagnose(fit) # diagnostics bundle; use ild_autoplot(bundle, ...)
```

## Pipeline

- `ild_prepare()` — encode longitudinal structure, spacing, gaps
- `ild_summary()` — one-shot summary
- `ild_center()` — person-mean centering (WP/BP)
- `ild_lag()` — index, gap-aware, or time-window lags (supports `lubridate::hours(2)` etc.)
- `ild_decomposition()` — WP/BP variance and ratio; optional WP vs BP density plot
- `ild_check_lags()` — lag validity (valid/invalid, pct_invalid, lag order)
- `ild_panel_lag_prepare()` — several `ild_lag()` columns + one lag audit table
- `ild_crosslag()` — one-call cross-lag: lag predictor, check lags, fit outcome ~ lag
- `ild_compare_fits()` — AIC/BIC/nobs table for a list of models (optional guardrail counts)
- `ild_brms_dynamics_formula()` — template `brms` formula for random lag slopes (does not fit)
- `ild_acf()` — ACF on a variable or on residuals (pre-model check for AR1)
- `ild_spacing_class()` — regular-ish vs irregular-ish
- `ild_spacing()` — spacing diagnostics (median/IQR, large gaps %, CV) and AR1/CAR1 recommendation
- `ild_design_check()` — aggregate spacing, WP/BP, missingness, and recommendations
- `ild_center_plot()` — standalone WP vs BP density plot
- `ild_missing_pattern()` — missingness by person/variable
- `ild_missing_bias()` — test if missingness is associated with a predictor (informative missingness)
- `ild_align()` — align secondary stream (e.g. wearables) to primary ILD within a time window
- `ild_lme()` — mixed-effects model (lmer or nlme with AR1/CAR1)
- `ild_robust_se()` — cluster-robust variance (clubSandwich); `tidy_ild_model(fit, se = "robust")` for robust SE/CI/p
- `ild_missing_model()` — model missingness from covariates; `ild_ipw_weights()` and `ild_ipw_refit()` for IPW sensitivity
- `ild_iptw_weights()`, `ild_iptw_msm_weights()`, `ild_ipcw_weights()`, `ild_joint_msm_weights()` — treatment/censoring/joint MSM weight pipelines
- `ild_msm_estimand()` + `ild_msm_fit()` — estimand-first MSM runner with explicit inference capability status and `strict_inference`
- `ild_msm_bootstrap()` / `tidy_ild_msm_bootstrap()` — cluster bootstrap inference for weighted `lmer` (fixed or re-estimated weights)
- `ild_msm_balance()`, `ild_ipw_ess()`, `ild_msm_overlap_plot()` — weighted balance, effective sample size, and overlap diagnostics
- `ild_msm_diagnose()` + `ild_msm_contrast_over_time()` — one-call diagnostics bridge and time-indexed post-fit contrasts
- `ild_msm_simulate_scenario()` + `ild_msm_recovery()` — causal simulation and recovery harness with scenario-grid validation
- `ild_ctsem()` — continuous-time latent-dynamics backend with `ild_tidy()`, `ild_augment()`, `ild_diagnose()`, and `ild_autoplot()`
- `ild_tvem()` — time-varying effects (GAM); `ild_tvem_plot()` for the coefficient curve
- **Temporal model choice:** `vignette("temporal-dynamics-model-choice", package = "tidyILD")` (lags vs AR vs TVEM vs state-space); `vignette("brms-dynamics-recipes", package = "tidyILD")` for Bayesian templates
- `ild_person_model()` — fit model per person (N-of-1); `ild_person_distribution()` — plot distribution of estimates
- `ild_diagnostics()` — residual ACF, residuals vs fitted/time (use `print()` for summary)
- `plot_ild_diagnostics()` — build diagnostic plots from an `ild_diagnostics` object
- `ild_plot()` — trajectory, heatmap, gaps, missingness, fitted vs observed, **predicted trajectory** (observed + fitted vs time), residual ACF; optional **`facet_by`** for panels (e.g. cluster)
- `ild_plot_predicted_trajectory()` — convenience wrapper for predicted trajectory vs `time_var`
- `ild_heatmap()`, `ild_spaghetti()` — aliases; pass **`facet_by`** through to `ggplot2::facet_wrap()`
- `ild_diagnose()` — assemble **`ild_diagnostics_bundle`** (data, design, fit, residual, …); **`ild_autoplot(bundle, ...)`** for sectioned plots
- `ild_circadian()` — variable by hour of day (when time is POSIXct)
- `augment_ild_model()` / `ild_augment()` — tibble per `ild_augment_schema()`: `.outcome`, `.fitted`, `.resid`, `.resid_std` (Pearson when available), `engine`, `model_class`, etc.
- `tidy_ild_model()` — fixed-effect table (estimate, SE, CI, p); use `se = "robust"` for cluster-robust inference
- `ild_simulate()` — simulated ILD (n_id, n_time/n_obs_per, ar1, wp_effect, bp_effect, irregular)
- `ild_power()` — simulation-based power for a fixed effect (ild_simulate → ild_lme → effect recovery)
- `ema_example` — built-in dataset (`data(ema_example)`)
- **broom**: Load `broom.mixed` for `tidy(fit)` and `augment(fit)` on `ild_lme` fits.

## Vignettes

- **Specialist backends** (`vignette("ild-specialist-backends", package = "tidyILD")`): when to move beyond `lme4` / default tidyILD fits; handoffs to dynamite, PGEE, DSEM, and multivariate workflows; export patterns after `ild_prepare()` / `ild_center()` / `ild_lag()`.
- **Visualization in tidyILD**: index of plots, bundle sections, `facet_by`, predicted trajectories, and partial-effects templates (`marginaleffects` / `ggeffects` on `_wp` / `_bp`).
- **From raw data to model**: full pipeline with `ild_prepare()` through `ild_lme()` and `ild_plot()`.
- **MSM identification and recovery**: assumptions, estimand-first MSM workflow, strict/degraded inference behavior, and scenario-grid recovery checks.
- **Continuous-time dynamics with ctsem**: `ild_ctsem()` workflow, diagnostics, and guardrails for ctsem fits.
- **Short analysis report**: fit, tidy fixed effects, fitted vs observed, residual ACF and Q-Q.
- **Within-between decomposition and irregular spacing**: centering and gap-aware lags.
- **Glossary and quick-start**: function glossary and checklist.

## pkgdown site

Documentation and vignettes are built with [pkgdown](https://pkgdown.r-lib.org/). From the package root: `pkgdown::build_site()`. Config: `_pkgdown.yml`.

## Version tags

Release tags follow `vMAJOR.MINOR.PATCH` and are listed in version order with:

```sh
git tag -l 'v*' --sort=version:refname
```

Current tags: **v0.0.1**, **v0.2.0**, **v0.3.0** (each annotated message is `tidyILD <version>`).

## License

MIT.

# Helpers for KFAS backend: data prep, KFS extraction, warnings

#' Prepare data for KFAS: single-subject ILD, sorted by time
#'
#' Validates ILD, enforces one distinct \code{.ild_id}, orders rows by time sequence.
#'
#' @param data An object that passes [validate_ild()].
#' @param outcome Character; name of outcome column.
#' @keywords internal
#' @noRd
ild_kfas_prepare_ts <- function(data, outcome) {
  validate_ild(data)
  if (!outcome %in% names(data)) {
    stop("Outcome column '", outcome, "' not found in data.", call. = FALSE)
  }
  n_id <- dplyr::n_distinct(data[[".ild_id"]])
  if (n_id > 1L) {
    stop(
      "ild_kfas() v1 fits one subject at a time. Subset `data` to a single `.ild_id`.",
      call. = FALSE
    )
  }
  ord <- order(data[[".ild_seq"]], data[[".ild_time"]], seq_len(nrow(data)))
  d <- data[ord, , drop = FALSE]
  y <- d[[outcome]]
  if (!is.numeric(y)) {
    y <- suppressWarnings(as.numeric(y))
  }
  list(
    data = d,
    y = y,
    outcome = outcome,
    id_label = as.character(d[[".ild_id"]][1L])
  )
}

#' Irregular spacing warning for local models that assume equal intervals
#' @keywords internal
#' @noRd
ild_kfas_warn_irregular <- function(data) {
  sc <- tryCatch(ild_spacing_class(data), error = function(e) NA_character_)
  if (!is.na(sc) && sc == "irregular-ish") {
    warning(
      "Time spacing is irregular-ish; local level on a regular grid is most interpretable. ",
      "Consider alignment or continuous-time extensions (future backends).",
      call. = FALSE
    )
  }
  invisible(sc)
}

#' Extract smoothed / filtered signal and standardized innovations from KFS output
#' @keywords internal
#' @noRd
ild_kfas_extract_augment <- function(kfs, y, use_smoother = TRUE) {
  # Smoothed observation expectation (signal) when available.
  # KFS often omits muhat for diffuse init; for identity observation y_t = Z*alpha_t,
  # the first smoothed state column matches the signal when Z = (1).
  muhat <- kfs$muhat
  ny <- length(as.numeric(y))
  if (!is.null(muhat) && is.matrix(muhat)) {
    fit <- as.numeric(muhat[, 1L])
  } else {
    ah <- kfs$alphahat
    if (!is.null(ah) && is.matrix(ah) && nrow(ah) == ny) {
      fit <- as.numeric(ah[, 1L])
    } else {
      m <- kfs$m
      if (!is.null(m) && is.matrix(m) && nrow(m) == ny) {
        fit <- as.numeric(m[, 1L])
      } else {
        fit <- rep(NA_real_, ny)
      }
    }
  }
  resid <- as.numeric(y) - fit
  rs <- rep(NA_real_, length(y))
  list(fitted = fit, resid = resid, resid_std = rs)
}

#' Scalar extract from KFAS Q / H arrays (time-invariant)
#' @keywords internal
#' @noRd
ild_kfas_qh_scalar <- function(x) {
  d <- dim(x)
  if (is.null(d)) return(NA_real_)
  if (length(d) >= 3L) {
    return(tryCatch(as.numeric(x[1L, 1L, 1L]), error = function(e) NA_real_))
  }
  tryCatch(as.numeric(x[1L, 1L]), error = function(e) NA_real_)
}

#' First latent state (smoothed) for local level — column 1 of alphahat
#' @keywords internal
#' @noRd
ild_kfas_extract_level_state <- function(kfs) {
  ah <- kfs$alphahat
  if (!is.null(ah) && is.matrix(ah)) {
    return(list(mean = as.numeric(ah[, 1L])))
  }
  list(mean = NULL)
}

#' Rich residual / innovation diagnostics for KFS output (Gaussian)
#' @keywords internal
#' @noRd
ild_kfas_residual_diagnostics <- function(kfs, y) {
  rlang::check_installed("KFAS", reason = "KFAS residual diagnostics")
  innov <- tryCatch(as.numeric(stats::rstandard(kfs)), error = function(e) NULL)
  if (is.null(innov) || length(innov) == 0L) {
    # Fallback: scale one-step prediction errors by their variances (Gaussian)
    v <- tryCatch(as.numeric(kfs$v), error = function(e) NULL)
    Fv <- tryCatch(as.numeric(kfs$F), error = function(e) NULL)
    if (!is.null(v) && !is.null(Fv) && length(v) == length(Fv)) {
      innov <- ifelse(is.finite(Fv) & Fv > 0, v / sqrt(Fv), NA_real_)
    }
  }
  ny <- length(as.numeric(y))
  if (is.null(innov) || length(innov) == 0L) {
    return(list(
      standardized_innovations_summary = NULL,
      acf_innovations = NULL,
      qq = NULL,
      outlier_flags = NULL,
      ljung_box = NULL
    ))
  }
  if (length(innov) != ny) {
    m <- min(length(innov), ny)
    innov <- innov[seq_len(m)]
  }
  n <- length(innov)
  ok <- is.finite(innov)
  iv <- innov[ok]
  summ <- list(
    n = length(iv),
    mean = mean(iv, na.rm = TRUE),
    sd = stats::sd(iv, na.rm = TRUE),
    min = suppressWarnings(min(iv, na.rm = TRUE)),
    max = suppressWarnings(max(iv, na.rm = TRUE)),
    mae = mean(abs(iv - mean(iv, na.rm = TRUE)), na.rm = TRUE)
  )
  acf_obj <- tryCatch(
    stats::acf(iv, plot = FALSE, lag.max = min(20L, max(1L, length(iv) - 1L))),
    error = function(e) NULL
  )
  acf_list <- if (!is.null(acf_obj)) {
    la <- acf_obj$lag
    ac <- acf_obj$acf
    if (is.matrix(la)) la <- as.numeric(la[, 1L]) else as.numeric(la)
    if (is.matrix(ac)) ac <- as.numeric(ac[, 1L]) else as.numeric(ac)
    list(lag = la, acf = ac)
  } else {
    NULL
  }
  qq_cor <- tryCatch({
    qn <- stats::qqnorm(iv, plot = FALSE)
    stats::cor(qn$x, qn$y)
  }, error = function(e) NA_real_)
  thr <- 3.5
  outlier_idx <- which(abs(innov) > thr & is.finite(innov))
  lb <- tryCatch(
    stats::Box.test(iv, type = "Ljung-Box", lag = min(5L, max(1L, length(iv) - 1L))),
    error = function(e) NULL
  )
  lb_list <- if (!is.null(lb)) {
    list(statistic = unname(lb$statistic), p_value = unname(lb$p.value), df = lb$parameter)
  } else {
    NULL
  }
  list(
    standardized_innovations_summary = summ,
    acf_innovations = acf_list,
    qq = list(correlation_normal_scores = qq_cor),
    outlier_flags = list(threshold_abs = thr, indices = outlier_idx, n = length(outlier_idx)),
    ljung_box = lb_list
  )
}

#' Count distinct runs of NA in a logical vector
#' @keywords internal
#' @noRd
ild_kfas_na_run_count <- function(is_na) {
  if (length(is_na) == 0L) return(0L)
  r <- rle(is_na)
  as.integer(sum(r$values))
}

# Section-first autoplot for ild_diagnostics_bundle (see ?ild_autoplot)

#' @rdname ild_autoplot
#' @method ild_autoplot ild_diagnostics_bundle
#' @export
ild_autoplot.ild_diagnostics_bundle <- function(x, section = "residual", type = NULL, ...) {
  if (!inherits(x, "ild_diagnostics_bundle")) {
    stop("Expected an ild_diagnostics_bundle.", call. = FALSE)
  }
  section <- match.arg(section, c("residual", "fit", "predictive", "data", "design", "causal"))
  eng <- x$meta$engine
  if (is.null(eng) && !is.null(x$fit)) {
    eng <- x$fit$engine
  }
  is_kfas <- identical(eng, "KFAS")
  is_ctsem <- identical(eng, "ctsem")

  if (identical(section, "residual")) {
    leg <- x$residual$legacy_ild_diagnostics
    if (is.null(type)) {
      if (!is_kfas && !is_ctsem && !is.null(leg)) {
        return(plot_ild_diagnostics(leg, ...))
      }
      type <- "acf"
    }
    type <- match.arg(type, c("acf", "qq", "fitted"))
    if (is_kfas) {
      switch(type,
        acf = plot_bundle_kfas_residual_acf(x),
        qq = plot_bundle_kfas_residual_qq(x),
        fitted = plot_bundle_kfas_residual_fitted(x)
      )
    } else if (is_ctsem) {
      switch(type,
        acf = plot_bundle_ctsem_residual_acf(x),
        qq = plot_bundle_ctsem_residual_qq(x),
        fitted = plot_bundle_ctsem_residual_fitted(x)
      )
    } else {
      switch(type,
        acf = plot_bundle_residual_acf(x),
        qq = plot_bundle_residual_qq(x),
        fitted = plot_bundle_residual_fitted(x)
      )
    }
  } else {
    choices <- switch(section,
      fit = c("convergence", "heterogeneity"),
      predictive = if (is_kfas) c("forecast", "errors") else if (is_ctsem) c("errors") else c("ppc"),
      data = c("missingness"),
      design = c("coverage"),
      causal = c("weights", "overlap")
    )
    if (is.null(type)) {
      type <- choices[1L]
    }
    type <- match.arg(type, choices)
    switch(section,
      fit = if (identical(type, "heterogeneity")) {
        plot_bundle_fit_heterogeneity(x, ...)
      } else if (is_kfas) {
        plot_bundle_kfas_fit_convergence(x)
      } else {
        plot_bundle_fit_convergence(x)
      },
      predictive = if (is_kfas) {
        switch(type,
          forecast = plot_bundle_kfas_predictive_forecast(x),
          errors = plot_bundle_kfas_predictive_errors(x)
        )
      } else if (is_ctsem) {
        plot_bundle_ctsem_predictive_errors(x)
      } else {
        plot_bundle_predictive_ppc(x, ...)
      },
      data = plot_bundle_data_missingness(x),
      design = plot_bundle_design_coverage(x),
      causal = switch(type,
        weights = plot_bundle_causal_weights(x),
        overlap = plot_bundle_causal_overlap(x, ...)
      )
    )
  }
}

#' @keywords internal
#' @noRd
plot_bundle_residual_acf <- function(x) {
  leg <- x$residual$legacy_ild_diagnostics
  if (!is.null(leg)) {
    pl <- plot_ild_diagnostics(leg, type = "residual_acf")
    return(pl$residual_acf)
  }
  fit <- attr(x, "ild_fit", exact = TRUE)
  dat <- attr(x, "ild_data", exact = TRUE)
  if (is.null(fit) || is.null(dat)) {
    stop(
      "Residual ACF requires residual$legacy_ild_diagnostics or a bundle from ",
      "ild_diagnose() (with ild_fit / ild_data attributes). Re-run ild_diagnose().",
      call. = FALSE
    )
  }
  diag <- ild_diagnostics(fit, data = dat, type = "residual_acf", by_id = TRUE)
  plot_ild_diagnostics(diag, type = "residual_acf")$residual_acf
}

#' @keywords internal
#' @noRd
plot_bundle_residual_qq <- function(x) {
  leg <- x$residual$legacy_ild_diagnostics
  if (!is.null(leg)) {
    pl <- plot_ild_diagnostics(leg, type = "qq")
    return(pl$qq)
  }
  fit <- attr(x, "ild_fit", exact = TRUE)
  dat <- attr(x, "ild_data", exact = TRUE)
  if (is.null(fit) || is.null(dat)) {
    stop(
      "Residual Q-Q requires residual$legacy_ild_diagnostics or ild_diagnose() ",
      "with ild_fit / ild_data attributes.",
      call. = FALSE
    )
  }
  diag <- ild_diagnostics(fit, data = dat, type = "qq", by_id = TRUE)
  plot_ild_diagnostics(diag, type = "qq")$qq
}

#' @keywords internal
#' @noRd
plot_bundle_residual_fitted <- function(x) {
  fit <- attr(x, "ild_fit", exact = TRUE)
  dat <- attr(x, "ild_data", exact = TRUE)
  if (is.null(fit) || is.null(dat)) {
    stop(
      "Fitted vs observed requires attr(bundle, \"ild_fit\") and attr(bundle, \"ild_data\"). ",
      "Re-run ild_diagnose() on the model.",
      call. = FALSE
    )
  }
  ild_plot_fitted(fit, dat)
}

#' @keywords internal
#' @noRd
plot_bundle_fit_convergence <- function(x) {
  fd <- x$fit
  eng <- if (!is.null(fd$engine)) fd$engine else x$meta$engine

  if (identical(eng, "brms")) {
    ct <- fd$convergence$convergence_table %||% fd$convergence_table
    if (is.null(ct) || nrow(ct) == 0L) {
      return(
        ggplot2::ggplot() +
          ggplot2::theme_minimal() +
          ggplot2::labs(title = "Convergence (no fixed-effects summary table)")
      )
    }
    ggplot2::ggplot(ct, ggplot2::aes(
      x = stats::reorder(.data$term, .data$rhat),
      y = .data$rhat
    )) +
      ggplot2::geom_point() +
      ggplot2::geom_hline(yintercept = 1.05, linetype = 2, alpha = 0.7) +
      ggplot2::coord_flip() +
      ggplot2::labs(x = NULL, y = "R-hat", title = "Convergence (fixed effects)") +
      ggplot2::theme_minimal()
  }
  if (identical(eng, "ctsem")) {
    conv <- fd$convergence
    txt <- paste(
      c(
        "Engine: ctsem",
        sprintf("Converged: %s", conv$converged),
        sprintf("Convergence code: %s", conv$convergence_code),
        sprintf("State dimension: %s", fd$state_dimension),
        if (is.finite(fd$log_likelihood)) sprintf("LogLik: %.2f", fd$log_likelihood) else "LogLik: NA"
      ),
      collapse = "\n"
    )
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5, label = txt, hjust = 0.5, vjust = 0.5) +
        ggplot2::theme_void() +
        ggplot2::labs(title = "Convergence (ctsem)") +
        ggplot2::xlim(0, 1) +
        ggplot2::ylim(0, 1)
    )
  }

  fd_conv <- fd$convergence
  sing <- if (!is.null(fd_conv)) fd_conv$singular else fd$singular
  conv <- if (!is.null(fd_conv)) fd_conv$converged else fd$converged
  om <- if (!is.null(fd_conv)) fd_conv$optimizer_messages else fd$optimizer_messages
  txt <- paste(
    c(
      sprintf("Engine: %s", if (!is.null(fd$engine)) fd$engine else eng),
      sprintf("Singular: %s", sing),
      sprintf("Converged: %s", conv),
      if (length(om) > 0L) {
        paste("Optimizer:", paste(om, collapse = "; "))
      } else {
        "Optimizer: (none)"
      }
    ),
    collapse = "\n"
  )
  ggplot2::ggplot() +
    ggplot2::annotate("text", x = 0.5, y = 0.5, label = txt, hjust = 0.5, vjust = 0.5) +
    ggplot2::theme_void() +
    ggplot2::labs(title = "Convergence (frequentist)") +
    ggplot2::xlim(0, 1) +
    ggplot2::ylim(0, 1)
}

#' @keywords internal
#' @noRd
plot_bundle_predictive_ppc <- function(x, ...) {
  eng <- x$meta$engine
  if (!identical(eng, "brms")) {
    stop(
      "Posterior predictive checks (section = \"predictive\", type = \"ppc\") ",
      "require a brmsfit from ild_brms().",
      call. = FALSE
    )
  }
  fit <- attr(x, "ild_fit", exact = TRUE)
  if (is.null(fit)) {
    stop("PPC requires attr(bundle, \"ild_fit\"). Re-run ild_diagnose() on the brmsfit.", call. = FALSE)
  }
  rlang::check_installed("brms")
  brms::pp_check(fit, ...)
}

#' @keywords internal
#' @noRd
plot_bundle_data_missingness <- function(x) {
  dat <- attr(x, "ild_data", exact = TRUE)
  if (is.null(dat)) {
    stop(
      "Missingness plot requires attr(bundle, \"ild_data\"). Re-run ild_diagnose().",
      call. = FALSE
    )
  }
  mp <- ild_missing_pattern(dat, vars = NULL)
  mp$plot
}

#' @keywords internal
#' @noRd
plot_bundle_design_coverage <- function(x) {
  dc <- x$design$ild_design_check
  dat <- attr(x, "ild_data", exact = TRUE)

  if (!is.null(dc$wp_bp) && nrow(dc$wp_bp) > 0L) {
    wp <- dc$wp_bp
    tot <- wp$wp_var + wp$bp_var
    pct_wp <- ifelse(tot > 0, 100 * wp$wp_var / tot, NA_real_)
    pct_bp <- ifelse(tot > 0, 100 * wp$bp_var / tot, NA_real_)
    n <- nrow(wp)
    df <- data.frame(
      variable = rep(as.character(wp$variable), 2L),
      component = rep(c("within-person", "between-person"), each = n),
      pct = c(pct_wp, pct_bp),
      stringsAsFactors = FALSE
    )
    return(
      ggplot2::ggplot(df, ggplot2::aes(x = .data$variable, y = .data$pct, fill = .data$component)) +
        ggplot2::geom_col(position = "stack") +
        ggplot2::labs(
          x = NULL,
          y = "Percent of total variance",
          title = "Design coverage (within vs between)",
          fill = NULL
        ) +
        ggplot2::theme_minimal() +
        ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))
    )
  }

  if (!is.null(dat) && ".ild_dt" %in% names(dat)) {
    dt <- dat[[".ild_dt"]]
    dt <- dt[!is.na(dt)]
    if (length(dt) == 0L) {
      stop("No non-missing .ild_dt values for interval coverage plot.", call. = FALSE)
    }
    df <- data.frame(dt = dt)
    return(
      ggplot2::ggplot(df, ggplot2::aes(x = .data$dt)) +
        ggplot2::geom_histogram(bins = min(30L, max(10L, length(unique(dt)))), fill = "steelblue", color = "white") +
        ggplot2::labs(
          x = expression(Delta * "t (observation intervals)"),
          y = "Count",
          title = "Observation interval coverage",
          subtitle = "WP/BP decomposition unavailable (no model vars in design check)"
        ) +
        ggplot2::theme_minimal()
    )
  }

  stop(
    "Design coverage needs design$ild_design_check$wp_bp or ILD data with `.ild_dt`.",
    call. = FALSE
  )
}

#' @keywords internal
#' @noRd
plot_bundle_causal_weights <- function(x) {
  dat <- attr(x, "ild_data", exact = TRUE)
  if (is.null(dat)) {
    stop("Weights plot requires attr(bundle, \"ild_data\"). Re-run ild_diagnose().", call. = FALSE)
  }
  parts <- list()
  if (".ipw" %in% names(dat)) parts[["joint (.ipw)"]] <- dat[[".ipw"]]
  if (".ipw_treat" %in% names(dat)) parts[["IPTW (.ipw_treat)"]] <- dat[[".ipw_treat"]]
  if (".ipw_censor" %in% names(dat)) parts[["IPCW (.ipw_censor)"]] <- dat[[".ipw_censor"]]
  if (length(parts) == 0L) {
    stop(
      "No IPW columns in ild_data. Use ild_ipw_weights(), ild_iptw_weights() + ild_ipcw_weights() + ild_joint_msm_weights(), or add weight columns first.",
      call. = FALSE
    )
  }
  rows <- list()
  for (nm in names(parts)) {
    w <- parts[[nm]]
    w <- w[is.finite(w)]
    if (length(w) > 0L) {
      rows[[length(rows) + 1L]] <- data.frame(source = nm, w = w)
    }
  }
  if (length(rows) == 0L) {
    stop("No finite weight values to plot.", call. = FALSE)
  }
  df <- do.call(rbind, rows)
  ggplot2::ggplot(df, ggplot2::aes(x = .data$w)) +
    ggplot2::geom_histogram(bins = min(30L, max(10L, length(unique(df$w)))), fill = "gray40", color = "white") +
    ggplot2::facet_wrap(~ source, scales = "free_y", ncol = 1) +
    ggplot2::labs(x = "Weight", y = "Count", title = "Causal / IPW weights") +
    ggplot2::theme_minimal()
}

#' @keywords internal
#' @noRd
plot_bundle_causal_overlap <- function(x, treatment, source = "auto", ...) {
  if (missing(treatment) || is.null(treatment) || !nzchar(as.character(treatment)[1L])) {
    stop(
      "Overlap plot requires treatment = <column name> (e.g. treatment = \"trt\").",
      call. = FALSE
    )
  }
  dat <- attr(x, "ild_data", exact = TRUE)
  if (is.null(dat)) {
    stop("Overlap plot requires attr(bundle, \"ild_data\"). Re-run ild_diagnose().", call. = FALSE)
  }
  ild_msm_overlap_plot(dat, treatment = treatment, source = source)
}

#' @keywords internal
#' @noRd
plot_bundle_ctsem_residual_acf <- function(x) {
  acf_tbl <- x$residual$acf
  if (is.null(acf_tbl) || nrow(acf_tbl) == 0L) {
    stop("ctsem residual ACF diagnostics are unavailable in this bundle.", call. = FALSE)
  }
  ggplot2::ggplot(acf_tbl, ggplot2::aes(x = .data$lag, y = .data$acf)) +
    ggplot2::geom_col() +
    ggplot2::geom_hline(yintercept = 0, linetype = 2) +
    ggplot2::labs(title = "ctsem residual ACF", x = "Lag", y = "ACF") +
    ggplot2::theme_minimal()
}

#' @keywords internal
#' @noRd
plot_bundle_ctsem_residual_qq <- function(x) {
  fit <- attr(x, "ild_fit", exact = TRUE)
  if (is.null(fit)) {
    stop("ctsem residual QQ plot requires attr(bundle, \"ild_fit\").", call. = FALSE)
  }
  ag <- ild_augment(fit)
  ggplot2::ggplot(ag, ggplot2::aes(sample = .data$.resid_std)) +
    ggplot2::stat_qq() +
    ggplot2::stat_qq_line() +
    ggplot2::labs(title = "ctsem standardized residual QQ plot", x = "Theoretical", y = "Sample") +
    ggplot2::theme_minimal()
}

#' @keywords internal
#' @noRd
plot_bundle_ctsem_residual_fitted <- function(x) {
  fit <- attr(x, "ild_fit", exact = TRUE)
  if (is.null(fit)) {
    stop("ctsem fitted diagnostics require attr(bundle, \"ild_fit\").", call. = FALSE)
  }
  ag <- ild_augment(fit)
  ggplot2::ggplot(ag, ggplot2::aes(x = .data$.fitted, y = .data$.resid)) +
    ggplot2::geom_point(alpha = 0.5) +
    ggplot2::geom_hline(yintercept = 0, linetype = 2) +
    ggplot2::labs(title = "ctsem residuals vs fitted", x = "Fitted", y = "Residual") +
    ggplot2::theme_minimal()
}

#' @keywords internal
#' @noRd
plot_bundle_ctsem_predictive_errors <- function(x) {
  om <- x$predictive$obs_metrics
  if (is.null(om)) {
    return(
      ggplot2::ggplot() +
        ggplot2::theme_void() +
        ggplot2::labs(title = "ctsem predictive errors unavailable")
    )
  }
  txt <- paste(
    c(
      sprintf("Engine: ctsem"),
      sprintf("n: %s", om$n),
      sprintf("MAE: %.3f", om$mean_abs_error),
      sprintf("RMSE: %.3f", om$rmse),
      sprintf("Mean residual: %.3f", om$mean_residual)
    ),
    collapse = "\n"
  )
  ggplot2::ggplot() +
    ggplot2::annotate("text", x = 0.5, y = 0.5, label = txt, hjust = 0.5, vjust = 0.5) +
    ggplot2::theme_void() +
    ggplot2::labs(title = "ctsem predictive error summary") +
    ggplot2::xlim(0, 1) +
    ggplot2::ylim(0, 1)
}

#' @keywords internal
#' @noRd
plot_bundle_fit_heterogeneity <- function(x, ...) {
  het <- x$fit$heterogeneity
  if (is.null(het) || !isTRUE(het$available) || is.null(het$object)) {
    stop(
      "Heterogeneity plot requires random effects and a successful ild_heterogeneity() extraction. ",
      "Re-run ild_diagnose() on a model with (e.g.) (predictor | id).",
      call. = FALSE
    )
  }
  dots <- list(...)
  term <- dots$term
  htype <- if (!is.null(dots$heterogeneity_type)) {
    dots$heterogeneity_type
  } else {
    "caterpillar"
  }
  ild_autoplot(het$object, type = htype, term = term)
}

#' Fit a model for missingness (diagnostic / sensitivity)
#'
#' Fits a logistic model predicting whether the outcome is missing from
#' covariates. Use as a diagnostic; then [ild_ipw_weights()] to compute
#' inverse-probability weights and [ild_ipw_refit()] for a sensitivity analysis.
#' This is not a full MNAR solution—treat as diagnostic and sensitivity tooling.
#'
#' **Not** run automatically by [ild_diagnose()]; it is the usual prerequisite for
#' [ild_ipw_weights()] when you want IPW summaries in the bundle (see [ild_diagnostics_utilities]).
#'
#' @section Bundle integration:
#' There is no separate \dQuote{missingness model} slot on the bundle. This function
#' supports the IPW workflow; after [ild_ipw_weights()], [ild_diagnose()] can summarize
#' weights in \code{causal} on \code{\link{ild_diagnostics_bundle}}.
#'
#' @family ild_diagnostics_utilities
#' @param x An ILD object (see [is_ild()]).
#' @param outcome Character. Name of the variable with missingness (e.g. \code{"mood"}).
#' @param predictors Character vector. Names of covariates to predict missingness.
#' @param random Logical. If \code{TRUE}, fit \code{glmer(is_missing ~ ... + (1|id))};
#'   if \code{FALSE} (default), fit \code{glm(is_missing ~ ., family = binomial)}.
#' @param family Passed to \code{glm} / \code{glmer} (default \code{binomial()}).
#' @param ... Passed to \code{glm()} or \code{lme4::glmer()}.
#' @return A list with \code{fit} (the fitted model or \code{NULL} if no/all missing),
#'   \code{tidy} (tibble: term, estimate, std_error, p_value), \code{outcome},
#'   \code{predictors}, and \code{message}. If \code{fit} is not \code{NULL},
#'   \code{p_missing} is a numeric vector of predicted P(missing) per row (aligned to \code{x}).
#' @seealso [ild_diagnose()], [ild_diagnostics_bundle()], [ild_ipw_weights()]
#' @export
#' @examples
#' set.seed(1)
#' d <- ild_simulate(n_id = 15, n_obs_per = 8, seed = 1)
#' d$stress <- rnorm(nrow(d))
#' d$mood <- d$y
#' d$mood[sample(nrow(d), 20)] <- NA
#' x <- ild_prepare(d, id = "id", time = "time")
#' mm <- ild_missing_model(x, "mood", "stress")
#' mm$tidy
ild_missing_model <- function(x, outcome, predictors, random = FALSE,
                              family = stats::binomial(), ...) {
  validate_ild(x)
  nms <- names(x)
  if (!outcome %in% nms) stop("outcome '", outcome, "' not found in data.", call. = FALSE)
  for (p in predictors) {
    if (!p %in% nms) stop("predictor '", p, "' not found in data.", call. = FALSE)
  }
  is_missing <- as.integer(is.na(x[[outcome]]))
  n_miss <- sum(is_missing)
  n_obs <- sum(!is_missing)
  miss_args <- list(outcome = outcome, predictors = predictors, random = random)
  miss_outputs <- list(n_missing = n_miss, n_obs = n_obs)
  if (n_miss == 0L) {
    out <- list(
      fit = NULL,
      tidy = tibble::tibble(term = character(), estimate = double(), std_error = double(), p_value = double()),
      outcome = outcome,
      predictors = predictors,
      message = "No missingness in outcome; nothing to model."
    )
    attr(out, "ild_provenance") <- ild_new_analysis_provenance(x, "ild_missing_model", miss_args, miss_outputs)
    return(out)
  }
  if (n_obs == 0L) {
    out <- list(
      fit = NULL,
      tidy = tibble::tibble(term = character(), estimate = double(), std_error = double(), p_value = double()),
      outcome = outcome,
      predictors = predictors,
      message = "All outcome values are missing; cannot fit model."
    )
    attr(out, "ild_provenance") <- ild_new_analysis_provenance(x, "ild_missing_model", miss_args, miss_outputs)
    return(out)
  }
  id_col <- ".ild_id"
  df <- as.data.frame(x[, c(id_col, predictors), drop = FALSE])
  df$is_missing <- is_missing
  df$id <- df[[id_col]]
  form <- stats::as.formula(paste("is_missing ~", paste(predictors, collapse = " + ")))
  fit <- if (random && requireNamespace("lme4", quietly = TRUE)) {
    tryCatch(
      lme4::glmer(stats::update(form, . ~ . + (1 | id)), data = df, family = family, ...),
      error = function(e) stats::glm(form, data = df, family = family, ...)
    )
  } else {
    stats::glm(form, data = df, family = family, ...)
  }
  co <- summary(fit)$coefficients
  if (is.null(co) || nrow(co) == 0L) {
    tidy_tbl <- tibble::tibble(term = character(), estimate = double(), std_error = double(), p_value = double())
  } else {
    tidy_tbl <- tibble::tibble(
      term = rownames(co),
      estimate = as.vector(co[, "Estimate"]),
      std_error = as.vector(co[, "Std. Error"]),
      p_value = if ("Pr(>|z|)" %in% colnames(co)) as.vector(co[, "Pr(>|z|)"]) else rep(NA_real_, nrow(co))
    )
  }
  p_missing <- stats::predict(fit, type = "response")
  if (length(p_missing) != nrow(x)) {
    p_missing <- rep(NA_real_, nrow(x))
  }
  out <- list(
    fit = fit,
    tidy = tidy_tbl,
    outcome = outcome,
    predictors = predictors,
    p_missing = p_missing,
    message = "Missingness model fit; use ild_ipw_weights() for IPW and ild_ipw_refit() for sensitivity."
  )
  attr(out, "ild_provenance") <- ild_new_analysis_provenance(x, "ild_missing_model", list(
    outcome = outcome,
    predictors = predictors,
    random = random
  ), list(n_missing = n_miss, n_obs = n_obs))
  out
}

test_that("ild_missing_model returns fit and tidy when there is missingness", {
  set.seed(1)
  d <- ild_simulate(n_id = 10, n_obs_per = 8, seed = 1)
  d$stress <- rnorm(nrow(d))
  d$mood <- d$y
  d$mood[sample(nrow(d), 15)] <- NA
  x <- ild_prepare(d, id = "id", time = "time")
  mm <- ild_missing_model(x, "mood", "stress")
  expect_true(is.list(mm))
  expect_named(mm, c("fit", "tidy", "outcome", "predictors", "p_missing", "message"))
  expect_s3_class(mm$fit, "glm")
  expect_s3_class(mm$tidy, "tbl_df")
  expect_equal(mm$outcome, "mood")
  expect_equal(mm$predictors, "stress")
  expect_equal(length(mm$p_missing), nrow(x))
})

test_that("ild_missing_model returns early when no missingness", {
  set.seed(1)
  d <- ild_simulate(n_id = 5, n_obs_per = 4, seed = 1)
  d$stress <- rnorm(nrow(d))
  x <- ild_prepare(d, id = "id", time = "time")
  mm <- ild_missing_model(x, "y", "stress")
  expect_null(mm$fit)
  expect_true(grepl("No missingness", mm$message))
})

test_that("ild_ipw_weights returns data with .ipw and preserves ILD attrs", {
  set.seed(1)
  d <- ild_simulate(n_id = 10, n_obs_per = 6, seed = 1)
  d$stress <- rnorm(nrow(d))
  d$mood <- d$y
  d$mood[sample(nrow(d), 10)] <- NA
  x <- ild_prepare(d, id = "id", time = "time")
  mm <- ild_missing_model(x, "mood", "stress")
  xw <- ild_ipw_weights(x, mm, stabilize = TRUE, trim = c(0.01, 0.99))
  expect_true(".ipw" %in% names(xw))
  expect_true(all(is.finite(xw$.ipw)))
  expect_equal(attr(xw, "ild_id"), attr(x, "ild_id"))
})

test_that("ild_ipw_refit runs and returns lmerMod with ild_data", {
  set.seed(1)
  d <- ild_simulate(n_id = 8, n_obs_per = 6, seed = 1)
  d$stress <- rnorm(nrow(d))
  d$mood <- d$y
  d$mood[sample(nrow(d), 8)] <- NA
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, mood)
  mm <- ild_missing_model(x, "mood", "stress")
  xw <- ild_ipw_weights(x, mm, stabilize = TRUE)
  fit0 <- ild_lme(mood ~ mood_bp + mood_wp + stress + (1 | id), data = xw,
    ar1 = FALSE, warn_no_ar1 = FALSE)
  fitw <- ild_ipw_refit(fit0, data = xw)
  expect_true(inherits(fitw, "lmerMod"))
  expect_false(is.null(attr(fitw, "ild_data")))
  tbl <- tidy_ild_model(fitw)
  expect_s3_class(tbl, "tbl_df")
  expect_true(all(ild_tidy_schema()$required %in% names(tbl)))
})

test_that("ild_ipw_refit errors when ar1 = TRUE", {
  expect_error(ild_ipw_refit(structure(list(), class = "lmerMod"), data = mtcars, ar1 = TRUE), "ar1 = TRUE")
})

# Compare fitted models on information criteria (and optional guardrail counts)

#' Compare fitted models (AIC, BIC, nobs)
#'
#' Builds a compact comparison table for a named list of fitted objects.
#' Uses [stats::AIC()], [stats::BIC()], and [stats::nobs()] where those methods
#' exist. This is **not** a likelihood-ratio test and does not establish that
#' models are nested; interpret only when model comparison is scientifically
#' appropriate for your estimand.
#'
#' @param fits A named (or unnamed) list of fitted model objects.
#' @param guardrail_bundles Optional named list of [ild_diagnostics_bundle] objects
#'   aligned with \code{fits} (same length and order). When provided, adds
#'   \code{n_guardrails} per row.
#' @return A tibble with columns \code{model}, \code{engine} (from [ild_methods()]
#'   truncated to one line when long), \code{aic}, \code{bic}, \code{n_obs},
#'   \code{converged} (best-effort), \code{n_guardrails} (if bundles given),
#'   and \code{notes} (fit-specific caveats).
#' @seealso \code{vignette("temporal-dynamics-model-choice", package = "tidyILD")}.
#' @export
ild_compare_fits <- function(fits, guardrail_bundles = NULL) {
  if (!is.list(fits) || length(fits) == 0L) {
    stop("fits must be a non-empty list.", call. = FALSE)
  }
  nm <- names(fits)
  if (is.null(nm)) {
    nm <- paste0("model", seq_along(fits))
  }
  if (any(!nzchar(nm))) {
    nm[!nzchar(nm)] <- paste0("model", which(!nzchar(nm)))
  }
  if (!is.null(guardrail_bundles) && length(guardrail_bundles) != length(fits)) {
    stop("guardrail_bundles must be NULL or have the same length as fits.", call. = FALSE)
  }
  rows <- vector("list", length(fits))
  for (i in seq_along(fits)) {
    fit <- fits[[i]]
    lab <- nm[i]
    eng <- tryCatch(as.character(ild_methods(fit))[1L], error = function(e) NA_character_)
    if (is.character(eng) && nchar(eng) > 120L) {
      eng <- paste0(substr(eng, 1L, 117L), "...")
    }
    aic <- tryCatch(stats::AIC(fit), error = function(e) NA_real_)
    bic <- tryCatch(stats::BIC(fit), error = function(e) NA_real_)
    nobs <- tryCatch(as.integer(stats::nobs(fit)), error = function(e) NA_integer_)
    conv <- ild_compare_fits_converged(fit)
    notes <- character(1L)
    if (inherits(fit, "lmerMod") && requireNamespace("lme4", quietly = TRUE)) {
      if (isTRUE(lme4::isSingular(fit))) {
        notes <- "lmer singular fit"
      }
    }
    n_gr <- NA_integer_
    if (!is.null(guardrail_bundles)) {
      b <- guardrail_bundles[[i]]
      if (!is.null(b) && inherits(b, "ild_diagnostics_bundle") && tibble::is_tibble(b$guardrails)) {
        n_gr <- as.integer(nrow(b$guardrails))
      }
    }
    rows[[i]] <- tibble::tibble(
      model = lab,
      engine = eng,
      aic = as.numeric(aic)[1L],
      bic = as.numeric(bic)[1L],
      n_obs = nobs,
      converged = conv,
      n_guardrails = n_gr,
      notes = notes
    )
  }
  dplyr::bind_rows(rows)
}

#' @keywords internal
#' @noRd
ild_compare_fits_converged <- function(fit) {
  if (inherits(fit, "lmerMod")) {
    conv <- fit@optinfo$conv$lme4
    if (!is.null(conv$code) && is.numeric(conv$code)) {
      return(as.integer(conv$code) == 0L)
    }
    return(NA)
  }
  if (inherits(fit, "lme")) {
    return(TRUE)
  }
  if (inherits(fit, "brmsfit") && requireNamespace("brms", quietly = TRUE)) {
    rh <- tryCatch(max(brms::rhat(fit), na.rm = TRUE), error = function(e) NA_real_)
    return(is.finite(rh) && rh < 1.09)
  }
  if (inherits(fit, "gam")) {
    return(TRUE)
  }
  NA
}

test_that("ild_compare_fits returns AIC/BIC for two lmer models", {
  d <- ild_simulate(n_id = 8, n_obs_per = 10, seed = 904L)
  d$x <- rnorm(nrow(d))
  x <- ild_prepare(d, id = "id", time = "time")
  f1 <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
  f2 <- ild_lme(y ~ x + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
  cmp <- ild_compare_fits(list(simple = f1, with_x = f2))
  expect_true(all(c("model", "aic", "bic", "n_obs", "converged") %in% names(cmp)))
  expect_equal(nrow(cmp), 2L)
  expect_true(is.finite(cmp$aic[1]) && is.finite(cmp$aic[2]))
})

test_that("ild_compare_fits passes guardrail_bundles through", {
  d <- ild_simulate(n_id = 5, n_obs_per = 8, seed = 905L)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
  b <- ild_diagnose(fit, data = x, type = "qq")
  cmp <- ild_compare_fits(list(a = fit), guardrail_bundles = list(b))
  expect_true("n_guardrails" %in% names(cmp))
  expect_equal(cmp$n_guardrails[1], nrow(b$guardrails))
})

test_that("ild_brms_dynamics_formula returns formula with lag and notes", {
  tpl <- ild_brms_dynamics_formula("y", "x_lag1", id_var = "id")
  expect_s3_class(tpl$formula, "formula")
  expect_true(nzchar(tpl$notes))
})

# Internal helpers shared by ild_ipw_weights, ild_iptw_weights, ild_ipcw_weights, ild_joint_msm_weights

#' Trim numeric weights at quantiles
#' @param w Numeric vector of weights.
#' @param trim Numeric of length 2, quantiles (e.g. \code{c(0.01, 0.99)}).
#' @return \code{w} with values clipped to quantile bounds.
#' @noRd
.ild_ipw_trim <- function(w, trim = c(0.01, 0.99)) {
  if (length(trim) < 2L || !is.numeric(trim)) return(w)
  qq <- stats::quantile(w, probs = trim, na.rm = TRUE)
  pmin(pmax(w, qq[1L]), qq[2L])
}

#' Restore ILD column attributes after adding a column (in-place style)
#' @param x ILD object.
#' @param attrs Result of \code{attributes(x)} before modification.
#' @return \code{x} with attributes restored.
#' @noRd
.ild_ipw_restore_attrs <- function(x, attrs) {
  for (a in c("ild_id", "ild_time", "ild_gap_threshold", "ild_n_units", "ild_n_obs", "ild_spacing", "class", "tidyILD")) {
    if (!is.null(attrs[[a]])) attr(x, a) <- attrs[[a]]
  }
  if (!is.null(attrs$names)) attr(x, "names") <- names(x)
  x
}

#' Refit an ILD model with inverse-probability weights (sensitivity analysis)
#'
#' Takes a fit from [ild_lme()] (or a formula and data) and refits the model
#' using observation weights from [ild_ipw_weights()]. Only the lme4 (lmer)
#' path is supported; nlme (ar1 = TRUE) is not supported for weighted refits.
#' This is a sensitivity tool, not a full MNAR solution.
#'
#' @param fit_or_formula Either a fitted model from [ild_lme()] (lmerMod), or a
#'   formula. If a formula, \code{data} and \code{weights} are required.
#' @param data An ILD object (see [is_ild()]) containing a weight column (default \code{.ipw})
#'   from [ild_ipw_weights()] or joint MSM weights from [ild_joint_msm_weights()], when refitting
#'   from a fit (or the data to fit when \code{fit_or_formula} is a formula).
#' @param weights Character. Name of the weight column in \code{data} (default \code{".ipw"}).
#' @param ar1 Logical. Must be \code{FALSE} for IPW refit (nlme weighted refit not supported).
#' @param ... Passed to [ild_lme()] or \code{lme4::lmer()}.
#' @return A fitted model (lmerMod) with \code{attr(..., "ild_data")} set so that
#'   [tidy_ild_model()] and [ild_diagnostics()] work. For inference under estimated
#'   weights, see [ild_msm_bootstrap()] and [ild_msm_inference].
#' @export
#' @examples
#' set.seed(1)
#' d <- ild_simulate(n_id = 12, n_obs_per = 10, seed = 1)
#' d$stress <- rnorm(nrow(d))
#' d$mood <- d$y
#' d$mood[sample(nrow(d), 25)] <- NA
#' x <- ild_prepare(d, id = "id", time = "time")
#' x <- ild_center(x, mood)
#' mm <- ild_missing_model(x, "mood", "stress")
#' xw <- ild_ipw_weights(x, mm, stabilize = TRUE)
#' fit0 <- ild_lme(mood ~ mood_bp + mood_wp + stress + (1 | id), data = xw,
#'   ar1 = FALSE, warn_no_ar1 = FALSE)
#' fitw <- ild_ipw_refit(fit0, data = xw)
#' tidy_ild_model(fitw)
ild_ipw_refit <- function(fit_or_formula, data, weights = ".ipw", ar1 = FALSE, ...) {
  if (ar1) {
    stop("ild_ipw_refit() does not support ar1 = TRUE (nlme weighted refit not supported). Use ar1 = FALSE.", call. = FALSE)
  }
  validate_ild(data)
  if (!weights %in% names(data)) {
    stop("weights column '", weights, "' not found in data. Run ild_ipw_weights() first.", call. = FALSE)
  }
  wvec <- data[[weights]]
  if (inherits(fit_or_formula, "lmerMod")) {
    formula <- stats::formula(fit_or_formula)
    args <- list(formula = formula, data = data, weights = wvec)
    args <- c(args, list(...))
    fit <- do.call(lme4::lmer, args)
    attr(fit, "ild_data") <- data
    attr(fit, "ild_ar1") <- FALSE
    attr(fit, "ild_provenance") <- ild_new_analysis_provenance(
      fit_or_formula,
      "ild_ipw_refit",
      list(weights = weights),
      list()
    )
    return(fit)
  }
  if (inherits(fit_or_formula, "lme")) {
    stop("ild_ipw_refit() does not support nlme (lme) fits. Refit with ild_lme(..., ar1 = FALSE) then ild_ipw_refit.", call. = FALSE)
  }
  if (inherits(fit_or_formula, "formula")) {
    args <- c(list(formula = fit_or_formula, data = data, weights = wvec), list(...))
    fit <- do.call(lme4::lmer, args)
    attr(fit, "ild_data") <- data
    attr(fit, "ild_ar1") <- FALSE
    attr(fit, "ild_provenance") <- ild_new_analysis_provenance(
      data,
      "ild_ipw_refit",
      list(weights = weights),
      list()
    )
    return(fit)
  }
  stop("fit_or_formula must be an ild_lme() fit (lmerMod) or a formula.", call. = FALSE)
}

# Sequential MSM IPTW (ild_iptw_msm_weights)

test_that("ild_iptw_msm_weights cumulative weight matches manual cumprod", {
  set.seed(1201)
  d <- ild_simulate(n_id = 8, n_obs_per = 5, seed = 1201)
  d$stress <- rnorm(nrow(d))
  d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
  d <- ild_prepare(d, id = "id", time = "time")
  x <- ild_iptw_msm_weights(d, treatment = "trt", history = ~ stress,
    stabilize = "marginal", trim = numeric(0))
  fits <- attr(x, "ild_iptw_msm_fits")
  expect_true(length(fits) >= 1L)
  n <- nrow(x)
  r <- rep(NA_real_, n)
  seqv <- x[[".ild_seq"]]
  A <- x[["trt"]]
  if (is.factor(A)) A <- as.integer(A) - 1L
  A <- as.numeric(A)
  for (nm in names(fits)) {
    t <- as.numeric(nm)
    idx <- which(as.numeric(seqv) == t & !is.na(A))
    fit <- fits[[nm]]
    p_hat <- stats::predict(fit, newdata = as.data.frame(x[idx, , drop = FALSE]), type = "response")
    p_hat <- pmax(pmin(p_hat, 1 - 1e-6), 1e-6)
    Ai <- A[idx]
    p1 <- mean(Ai, na.rm = TRUE)
    p1 <- pmax(pmin(p1, 1 - 1e-6), 1e-6)
    r[idx] <- ifelse(Ai >= 0.5, p1 / p_hat, (1 - p1) / (1 - p_hat))
  }
  id <- x[[".ild_id"]]
  ord <- order(id, seqv, seq_len(n))
  ido <- id[ord]
  r_ord <- r[ord]
  r_ord[is.na(r_ord)] <- 1
  w_ord <- rep(NA_real_, n)
  i <- 1L
  while (i <= n) {
    j <- i
    while (j < n && ido[j + 1L] == ido[i]) j <- j + 1L
    idx <- i:j
    w_ord[idx] <- cumprod(r_ord[idx])
    i <- j + 1L
  }
  w <- rep(NA_real_, n)
  w[ord] <- w_ord
  expect_equal(x$.ipw_treat, w, tolerance = 1e-5)
})

test_that("ild_iptw_msm_weights + joint + refit + diagnose (smoke)", {
  set.seed(1202)
  d <- ild_simulate(n_id = 15, n_obs_per = 6, seed = 1202)
  d$stress <- rnorm(nrow(d))
  d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
  d <- ild_prepare(d, id = "id", time = "time")
  d <- ild_center(d, y)
  d <- ild_lag(d, stress, mode = "gap_aware", max_gap = Inf)
  d <- ild_lag(d, trt, mode = "gap_aware", max_gap = Inf)
  x <- ild_iptw_msm_weights(d, "trt", history = ~ stress_lag1 + trt_lag1)
  x <- ild_ipcw_weights(x, predictors = "stress")
  x <- ild_joint_msm_weights(x)
  fit <- ild_lme(y ~ y_bp + y_wp + stress + (1 | id), data = x,
    ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
  fitw <- ild_ipw_refit(fit, data = x)
  expect_true(inherits(fitw, "lmerMod"))
  b <- ild_diagnose(fitw, data = x, type = "qq", causal_detail = TRUE)
  expect_true(!is.null(b$causal$msm_treat_weight_by_occasion))
})

# IPTW / IPCW / joint MSM weights

test_that("ild_iptw_weights + ild_ipcw_weights + ild_joint_msm_weights pipeline", {
  set.seed(901)
  d <- ild_simulate(n_id = 25, n_obs_per = 8, seed = 901)
  d$stress <- rnorm(nrow(d))
  d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_iptw_weights(x, treatment = "trt", predictors = "stress")
  x <- ild_ipcw_weights(x, predictors = "stress")
  expect_true(all(is.finite(x$.ipw_treat)))
  expect_true(all(is.finite(x$.ipw_censor)))
  expect_true(all(x$.ipw_treat > 0))
  expect_true(all(x$.ipw_censor > 0))
  xj <- ild_joint_msm_weights(x, stabilize = "none", trim = numeric(0))
  expect_equal(xj$.ipw, xj$.ipw_treat * xj$.ipw_censor)
  xm <- ild_joint_msm_weights(x, stabilize = "mean1", trim = numeric(0))
  expect_equal(mean(xm$.ipw, na.rm = TRUE), 1, tolerance = 1e-4)
  wprod <- xm$.ipw_treat * xm$.ipw_censor
  expect_equal(xm$.ipw, wprod / mean(wprod, na.rm = TRUE), tolerance = 1e-3)
})

test_that("ild_ipw_refit and ild_diagnose work with joint MSM .ipw", {
  set.seed(902)
  d <- ild_simulate(n_id = 20, n_obs_per = 7, seed = 902)
  d$stress <- rnorm(nrow(d))
  d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  x <- ild_iptw_weights(x, "trt", "stress")
  x <- ild_ipcw_weights(x, "stress")
  x <- ild_joint_msm_weights(x)
  fit <- ild_lme(y ~ y_bp + y_wp + stress + (1 | id), data = x,
    ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
  fitw <- ild_ipw_refit(fit, data = x)
  expect_true(inherits(fitw, "lmerMod"))
  b <- ild_diagnose(fitw, data = x, type = "qq")
  expect_true(!is.null(b$causal))
  expect_true(!is.null(b$causal$weight_summary))
  expect_true(!is.null(b$causal$weight_summary_treat))
  expect_true(!is.null(b$causal$weight_summary_censor))
})

test_that("ild_iptw_weights rejects non-binary treatment", {
  set.seed(903)
  d <- ild_simulate(n_id = 10, n_obs_per = 5, seed = 903)
  d$stress <- rnorm(nrow(d))
  d$trt3 <- rep(1:3, length.out = nrow(d))
  x <- ild_prepare(d, id = "id", time = "time")
  expect_error(
    ild_iptw_weights(x, "trt3", "stress"),
    "binary treatment"
  )
})

#' Align a secondary stream to primary ILD within a time window
#'
#' For each row in the primary ILD, finds observations in the secondary
#' data set (same id, time within \code{window} before the primary time)
#' and attaches an aggregated value (e.g. mean, median, or closest).
#' Use when combining self-report with wearables or other streams that
#' have different timestamps.
#'
#' @param primary An ILD object (see [is_ild()]); the stream to keep as rows.
#' @param secondary A data frame with id and time columns and the value variable(s) to align.
#' @param value_var Character. Name of the column in \code{secondary} to align (e.g. \code{"heart_rate"}).
#' @param window Numeric or lubridate duration. Time window (same units as \code{.ild_time_num},
#'   e.g. seconds for POSIXct). Only secondary observations with time in
#'   \code{(primary_time - window, primary_time]} are used.
#' @param time_secondary Character. Name of the time column in \code{secondary} (default \code{"time"}).
#' @param fun Character. Aggregation for values in window: \code{"mean"}, \code{"median"}, or \code{"closest"} (most recent in window).
#' @return The primary data with a new column \code{<value_var>_aligned} (numeric; NA where no secondary obs in window).
#' @export
#' @examples
#' prim <- ild_prepare(
#'   data.frame(
#'     id = rep(1:2, each = 3),
#'     time = as.POSIXct(rep(c(0, 3600, 7200), 2), origin = "1970-01-01"),
#'     y = rnorm(6)
#'   ),
#'   id = "id", time = "time"
#' )
#' sec <- data.frame(
#'   id = rep(1:2, each = 4),
#'   time = as.POSIXct(rep(c(0, 1800, 3600, 5400), 2), origin = "1970-01-01"),
#'   heart_rate = 60 + rnorm(8, 0, 5)
#' )
#' ild_align(prim, sec, "heart_rate", window = 3600, fun = "mean")
ild_align <- function(primary, secondary, value_var, window,
                     time_secondary = "time", fun = c("mean", "median", "closest")) {
  validate_ild(primary)
  fun <- match.arg(fun)
  meta <- ild_meta(primary)
  id_name <- meta$ild_id
  if (!id_name %in% names(secondary)) stop("secondary must have id column '", id_name, "'.", call. = FALSE)
  if (!time_secondary %in% names(secondary)) stop("secondary must have time column '", time_secondary, "'.", call. = FALSE)
  if (!value_var %in% names(secondary)) stop("secondary must have value column '", value_var, "'.", call. = FALSE)
  window_num <- if (is.numeric(window)) as.numeric(window)[1] else {
    if (inherits(window, "Period") || inherits(window, "Duration")) as.numeric(lubridate::as.duration(window)) else as.numeric(window)[1]
  }
  t_sec_num <- ild_time_to_num(secondary[[time_secondary]])
  primary_id <- primary[[".ild_id"]]
  primary_t <- primary[[".ild_time_num"]]
  n <- nrow(primary)
  aligned <- rep(NA_real_, n)
  for (i in seq_len(n)) {
    pid <- primary_id[i]
    pt <- primary_t[i]
    if (is.na(pt)) next
    lo <- pt - window_num
    idx <- which(secondary[[id_name]] == pid & t_sec_num > lo & t_sec_num <= pt)
    if (length(idx) == 0) next
    if (fun == "closest") {
      imax <- idx[which.max(t_sec_num[idx])]
      aligned[i] <- secondary[[value_var]][imax]
    } else {
      vals <- secondary[[value_var]][idx]
      vals <- vals[!is.na(vals)]
      if (length(vals) == 0) next
      if (fun == "mean") aligned[i] <- mean(vals) else aligned[i] <- stats::median(vals)
    }
  }
  out <- tibble::as_tibble(primary)
  out[[paste0(value_var, "_aligned")]] <- aligned
  out <- restore_ild_attrs(primary, out)
  out <- ild_add_step(out, "ild_align",
    list(value_var = value_var, window = window, time_secondary = time_secondary, fun = fun),
    list(created = paste0(value_var, "_aligned"))
  )
  out
}

test_that("ild_align attaches aligned column with mean fun", {
  prim <- ild_prepare(
    data.frame(
      id = rep(1:2, each = 3),
      time = as.POSIXct(rep(c(0, 3600, 7200), 2), origin = "1970-01-01"),
      y = rnorm(6)
    ),
    id = "id", time = "time"
  )
  sec <- data.frame(
    id = rep(1:2, each = 4),
    time = as.POSIXct(rep(c(0, 1800, 3600, 5400), 2), origin = "1970-01-01"),
    hr = 70 + rnorm(8, 0, 2)
  )
  out <- ild_align(prim, sec, "hr", window = 3600, fun = "mean")
  expect_true("hr_aligned" %in% names(out))
  expect_equal(nrow(out), 6)
  expect_true(is.numeric(out$hr_aligned))
})

test_that("ild_align closest returns value from latest time in window", {
  prim <- ild_prepare(
    data.frame(id = 1, time = as.POSIXct(3600, origin = "1970-01-01"), y = 1),
    id = "id", time = "time"
  )
  sec <- data.frame(id = 1, time = as.POSIXct(c(0, 1800, 3600), origin = "1970-01-01"), v = c(10, 20, 30))
  out <- ild_align(prim, sec, "v", window = 3600, fun = "closest")
  expect_equal(out$v_aligned, 30)
})

test_that("ild_align errors when secondary missing required columns", {
  prim <- ild_prepare(
    data.frame(id = 1, time = as.POSIXct(0, origin = "1970-01-01"), y = 1),
    id = "id", time = "time"
  )
  sec <- data.frame(other_id = 1, time = as.POSIXct(0, origin = "1970-01-01"), hr = 60)
  expect_error(ild_align(prim, sec, "hr", window = 3600), "id column")
})

# Guardrail registry and evaluation for ild_diagnostics_bundle$guardrails
# Schema: rule_id, section, severity, triggered, message, recommendation

#' @importFrom tibble tibble
NULL

ILD_GUARDRAIL_REGISTRY <- tibble::tibble(
  rule_id = c(
    "GR_SINGULAR_RANDOM_EFFECTS",
    "GR_POOR_POSTERIOR_MIXING",
    "GR_LOW_ESS",
    "GR_MIXED_PREDICTOR_NOT_DECOMPOSED",
    "GR_IRREGULAR_SPACING_NO_RESID_COR",
    "GR_HIGH_TIMING_GAP_RATE",
    "GR_DROPOUT_LATE_CONCENTRATION",
    "GR_IPW_WEIGHTS_UNSTABLE",
    "GR_MSM_COMPONENT_WEIGHTS_UNSTABLE",
    "GR_MSM_BALANCE_SMD_HIGH",
    "GR_MSM_ESS_LOW",
    "GR_KFAS_HIGH_IRREGULARITY_FOR_DISCRETE_TIME",
    "GR_KFAS_SHORT_SERIES_FOR_STATE_SPACE",
    "GR_KFAS_STATE_DIMENSION_HIGH_FOR_N",
    "GR_KFAS_DEGENERATE_VARIANCE_ESTIMATE",
    "GR_KFAS_NONCONVERGENCE",
    "GR_KFAS_MANY_MISSING_OUTCOME_SEGMENTS",
    "GR_KFAS_UNMODELED_BETWEEN_PERSON_HETEROGENEITY",
    "GR_CTSEM_NONCONVERGENCE",
    "GR_CTSEM_UNSTABLE_DYNAMICS",
    "GR_CTSEM_SHORT_SERIES_FOR_COMPLEX_DYNAMICS",
    "GR_RE_SLOPE_VARIANCE_VERSUS_RESIDUAL_LOW",
    "GR_PERSON_SPECIFIC_SLOPES_EMPIRICALLY_TIGHT",
    "GR_LAG_MEAN_STRONG_RESIDUAL_ACF_NO_AR",
    "GR_INDEX_LAG_IRREGULAR_SPACING"
  ),
  section = c(
    "fit", "fit", "fit",
    "design", "data", "data",
    "missingness", "causal", "causal", "causal", "causal",
    "design", "data", "fit", "fit", "fit", "data", "fit",
    "fit", "fit", "data",
    "fit", "fit",
    "residual", "data"
  ),
  severity = c(
    "warning", "warning", "warning",
    "warning", "warning", "info",
    "warning", "warning", "warning", "info", "info",
    "warning", "warning", "warning", "warning", "warning", "info", "warning",
    "warning", "warning", "warning",
    "info", "info",
    "warning", "info"
  ),
  default_message = c(
    "Estimated random-effect covariance is singular or near-singular.",
    "Max R-hat exceeds a conventional mixing threshold.",
    "Bulk effective sample size is low for at least one fixed-effect parameter.",
    "Predictor(s) vary within and between persons but are not WP/BP decomposed.",
    "Irregular spacing is present but residual temporal correlation (AR1/CAR1) is not modeled.",
    "A large fraction of time intervals are flagged as gaps.",
    "Missingness on the outcome is more concentrated in the later part of the study timeline.",
    "Inverse probability weights show extreme variability (possible instability).",
    "IPTW or IPCW component weights show extreme variability (possible instability).",
    "At least one covariate shows large weighted standardized mean difference after weighting.",
    "Effective sample size from the analysis weights is low relative to the number of rows.",
    "Interval timing is highly irregular relative to the median step; discrete-time KFAS models assume a consistent time index.",
    "The observed series is short for the requested state-space complexity.",
    "Latent state dimension is large relative to series length.",
    "Estimated state or observation variance is effectively on the boundary (near zero or degenerate).",
    "Numerical optimization for the state-space model did not report clean convergence.",
    "The outcome has many separate missing segments in time order.",
    "Independent single-subject state-space fits are not a substitute for pooled multilevel latent-dynamics models (e.g. ctsem).",
    "ctsem optimization did not report clean convergence.",
    "Estimated continuous-time dynamics appear unstable or implausibly large.",
    "The observed series may be too short for the requested continuous-time latent dynamics complexity.",
    "Random slope standard deviation is small relative to residual standard deviation.",
    "Person-specific partial-pooling slopes show very little empirical spread versus residual noise.",
    "Pooled residual ACF at lag 1 is large while the mean includes lag columns and residual AR1/CAR1 is not modeled.",
    "Index-based ild_lag() was used on irregular-ish spacing; row-order lags may not align with elapsed time between observations."
  ),
  default_recommendation = c(
    "Consider simpler random structure, stronger priors (Bayesian), or rescaling predictors.",
    "Run longer chains, reparameterize, or tighten priors.",
    "Increase iterations or use stronger priors; interpret posterior means with caution.",
    "Use ild_center() to create _wp and _bp components to avoid conflation bias.",
    "Consider ild_lme(..., ar1 = TRUE) or interpret residual autocorrelation diagnostics cautiously.",
    "Review gap_threshold and missingness; lags and spacing-based assumptions may be strained.",
    "Investigate selective dropout (MNAR) and sensitivity analyses; consider IPW or models for missingness.",
    "Review the missingness model, trim bounds, or stabilize weights before interpreting results.",
    "Review treatment and censoring models, trim bounds, or stabilize weights before interpreting MSM results.",
    "Review confounder selection, functional forms, overlap, or weight stabilization.",
    "Consider trimming, stabilizing, or respecifying weight models; interpret weighted estimates cautiously.",
    "Align or aggregate to a regular grid, or use continuous-time modeling for irregular spacing.",
    "Collect more time points or simplify the state specification (e.g. fewer latent components).",
    "Simplify the model (fewer states) or extend the series before interpreting complex latent structure.",
    "Rescale the outcome, re-check identification, or consider stronger priors / constraints in KFAS.",
    "Try different starting values, optimization method, or check the model specification.",
    "Consider explicit missingness modeling or sensitivity analyses for intermittent observations.",
    "Do not treat stacked single-ID KFAS fits as a pooled latent model; use hierarchical latent frameworks if that is the estimand.",
    "Try alternative starting values, simplify the ctsem model, or inspect time scaling and identification.",
    "Rescale time/outcome or constrain unstable drift parameters before interpretation.",
    "Collect more observations per person or simplify latent dynamics before strong inferences.",
    "Use ild_heterogeneity() and VarCorr(); consider a simpler random structure if slopes are not substantively heterogeneous.",
    "Compare to ild_person_model() and population fixed effects; confirm whether random slopes merit interpretation.",
    "Consider ild_lme(..., ar1 = TRUE) or a dynamics model that matches spacing; see vignette(\"temporal-dynamics-model-choice\", package = \"tidyILD\").",
    "Prefer ild_lag(..., mode = \"gap_aware\" or \"time_window\"), align to a grid, or use continuous-time models when appropriate."
  )
)

#' Guardrail rule registry (analysis safety layer)
#'
#' Returns the canonical catalog of \strong{guardrail} rules tidyILD may use. Each rule has a
#' stable \code{rule_id}, \code{section} (bundle slot), \code{severity}, and default
#' \code{default_message} / \code{default_recommendation} text. When you call
#' \code{\link{ild_diagnose}}, \strong{triggered} rules appear as rows in
#' \code{\link{ild_diagnostics_bundle}$guardrails} with \code{message} and
#' \code{recommendation} possibly customized for the run.
#'
#' @return A tibble with columns \code{rule_id}, \code{section}, \code{severity},
#'   \code{default_message}, \code{default_recommendation}.
#' @seealso \code{\link{ild_diagnose}}, \code{\link{ild_diagnostics_bundle}}
#' @export
guardrail_registry <- function() {
  ILD_GUARDRAIL_REGISTRY
}

#' @keywords internal
#' @noRd
guardrail_registry_lookup <- function(rule_id) {
  r <- ILD_GUARDRAIL_REGISTRY
  m <- match(rule_id, r$rule_id)
  if (is.na(m)) {
    return(list(
      section = "unknown",
      severity = "info",
      default_message = "",
      default_recommendation = ""
    ))
  }
  list(
    section = r$section[m],
    severity = r$severity[m],
    default_message = r$default_message[m],
    default_recommendation = r$default_recommendation[m]
  )
}

#' Numeric rank for guardrail severity (higher = stronger)
#'
#' \code{info} < \code{warning}; unknown levels rank lowest (0).
#' @param severity Character vector of severity labels.
#' @return Integer vector of same length as \code{severity}.
#' @keywords internal
#' @noRd
guardrail_severity_rank <- function(severity) {
  s <- tolower(as.character(severity))
  ifelse(s == "warning", 2L, ifelse(s == "info", 1L, 0L))
}

#' Highest severity among triggered guardrails
#'
#' @param severities Character vector (e.g. from \code{guardrails$severity}).
#' @return Single severity string (\code{"warning"} or \code{"info"}) or
#'   \code{NA_character_} if empty.
#' @keywords internal
#' @noRd
guardrail_max_severity <- function(severities) {
  if (length(severities) == 0L) {
    return(NA_character_)
  }
  sev <- as.character(severities)
  r <- guardrail_severity_rank(sev)
  sev[which.max(r)[1]]
}

#' Structured summary and short narrative for guardrails tibble
#'
#' Used by \code{\link{print.ild_diagnostics_bundle}}, \code{\link{ild_report}},
#' and \code{\link{ild_methods}}. Does not include full \code{message} text.
#'
#' @param guardrails A guardrails tibble (e.g. \code{bundle$guardrails}), or an
#'   \code{\link{ild_diagnostics_bundle}} (uses \code{$guardrails}).
#' @param max_rule_ids Maximum unique \code{rule_id} values to list in the narrative (default 5).
#' @return A list: \code{n} (integer), \code{max_severity} (character or \code{NA}),
#'   \code{rule_ids} (character, unique, order preserved), \code{narrative} (single string, may be empty).
#' @keywords internal
#' @noRd
ild_guardrails_summary <- function(guardrails, max_rule_ids = 5L) {
  if (inherits(guardrails, "ild_diagnostics_bundle")) {
    guardrails <- guardrails$guardrails
  }
  if (is.null(guardrails) || !tibble::is_tibble(guardrails) || nrow(guardrails) == 0L) {
    return(list(
      n = 0L,
      max_severity = NA_character_,
      rule_ids = character(0),
      narrative = ""
    ))
  }
  n <- nrow(guardrails)
  max_sev <- guardrail_max_severity(guardrails$severity)
  rids <- unique(as.character(guardrails$rule_id))
  show_ids <- rids[seq_len(min(length(rids), as.integer(max_rule_ids)))]
  id_part <- if (length(show_ids) > 0L) {
    paste(show_ids, collapse = ", ")
  } else {
    ""
  }
  sev_part <- if (!is.na(max_sev) && nzchar(max_sev)) {
    sprintf("max severity %s", max_sev)
  } else {
    ""
  }
  narrative <- if (n > 0L) {
    paste0(
      sprintf("%d methodological guardrail%s triggered", n, if (n == 1L) "" else "s"),
      if (nzchar(sev_part)) paste0(" (", sev_part, ")") else "",
      if (nzchar(id_part)) paste0(": ", id_part) else "",
      "."
    )
  } else {
    ""
  }
  list(
    n = as.integer(n),
    max_severity = max_sev,
    rule_ids = rids,
    narrative = narrative
  )
}

#' Empty guardrails tibble (canonical schema)
#' @keywords internal
#' @noRd
guardrails_empty_tibble <- function() {
  tibble::tibble(
    rule_id = character(),
    section = character(),
    severity = character(),
    triggered = logical(),
    message = character(),
    recommendation = character()
  )
}

#' Assemble guardrails from evaluated rows (rule_id + trigger + optional overrides)
#' @keywords internal
#' @noRd
guardrail_finalize_rows <- function(rows) {
  if (length(rows) == 0L) {
    return(guardrails_empty_tibble())
  }
  out_rule <- character()
  out_sec <- character()
  out_sev <- character()
  out_trig <- logical()
  out_msg <- character()
  out_rec <- character()
  for (z in rows) {
    if (!isTRUE(z$triggered)) next
    rid <- as.character(z$rule_id)[1]
    def <- guardrail_registry_lookup(rid)
    msg <- z$message
    if (is.null(msg) || (length(msg) == 1L && !nzchar(msg))) {
      msg <- def$default_message
    } else {
      msg <- as.character(msg)[1]
    }
    rec <- z$recommendation
    if (is.null(rec) || (length(rec) == 1L && !nzchar(rec))) {
      rec <- def$default_recommendation
    } else {
      rec <- as.character(rec)[1]
    }
    out_rule <- c(out_rule, rid)
    out_sec <- c(out_sec, def$section)
    out_sev <- c(out_sev, def$severity)
    out_trig <- c(out_trig, TRUE)
    out_msg <- c(out_msg, msg)
    out_rec <- c(out_rec, rec)
  }
  tibble::tibble(
    rule_id = out_rule,
    section = out_sec,
    severity = out_sev,
    triggered = out_trig,
    message = out_msg,
    recommendation = out_rec
  )
}

#' @keywords internal
#' @noRd
ild_formula_includes_lag_terms <- function(f) {
  if (is.null(f)) {
    return(FALSE)
  }
  rhs <- tryCatch(stats::terms(f), error = function(e) NULL)
  if (is.null(rhs)) {
    return(FALSE)
  }
  labels <- attr(rhs, "term.labels")
  any(grepl("_lag[0-9]+$|_lag_window$", labels))
}

#' @keywords internal
#' @noRd
ild_bundle_residual_acf_lag1 <- function(bundle) {
  ac <- tryCatch(bundle$residual$stats$acf$pooled, error = function(e) NULL)
  if (is.null(ac) || nrow(ac) == 0L) {
    return(NA_real_)
  }
  hit <- ac[ac$lag %in% c(1, 1L), , drop = FALSE]
  if (nrow(hit) == 0L) {
    return(NA_real_)
  }
  as.numeric(hit$acf[1L])
}

#' @keywords internal
#' @noRd
ild_provenance_has_index_lag <- function(data) {
  p <- ild_get_history(data)
  if (is.null(p) || is.null(p$steps)) {
    return(FALSE)
  }
  for (s in p$steps) {
    if (identical(as.character(s$step), "ild_lag")) {
      am <- s$args$mode
      if (identical(am, "index")) {
        return(TRUE)
      }
    }
  }
  FALSE
}

#' Evaluate design / data / causal guardrails from bundle + model
#' @keywords internal
#' @noRd
evaluate_guardrails_contextual <- function(object, data, bundle, engine = c("lmer", "lme", "brms", "ctsem")) {
  engine <- match.arg(engine)
  rows <- list()
  if (!is.null(data) && is_ild(data)) {
    f <- tryCatch(stats::formula(object), error = function(e) NULL)
    if (!is.null(f)) {
      unc <- ild_detect_uncentered_predictors(data, f)
      if (length(unc) > 0L) {
        rows[[length(rows) + 1L]] <- list(
          rule_id = "GR_MIXED_PREDICTOR_NOT_DECOMPOSED",
          triggered = TRUE,
          message = sprintf(
            "Predictor(s) %s vary within and between persons but are not WP/BP decomposed.",
            paste(paste0("'", unc, "'"), collapse = ", ")
          ),
          recommendation = NULL
        )
      }
    }
    if (engine %in% c("lmer", "lme")) {
      ar1_dyn <- tryCatch(isTRUE(attr(object, "ild_ar1", exact = TRUE)), error = function(e) FALSE)
      if (!ar1_dyn && !is.null(f) && ild_formula_includes_lag_terms(f)) {
        acf1 <- ild_bundle_residual_acf_lag1(bundle)
        if (is.finite(acf1) && acf1 > 0.2) {
          rows[[length(rows) + 1L]] <- list(
            rule_id = "GR_LAG_MEAN_STRONG_RESIDUAL_ACF_NO_AR",
            triggered = TRUE,
            message = sprintf(
              "Pooled residual ACF at lag 1 is %.2f with lag term(s) in the mean and residual AR1/CAR1 not used.",
              acf1
            ),
            recommendation = NULL
          )
        }
      }
      sc2 <- tryCatch(ild_spacing_class(data), error = function(e) NA_character_)
      if (!is.na(sc2) && sc2 == "irregular-ish" && ild_provenance_has_index_lag(data)) {
        rows[[length(rows) + 1L]] <- list(
          rule_id = "GR_INDEX_LAG_IRREGULAR_SPACING",
          triggered = TRUE,
          message = NULL,
          recommendation = NULL
        )
      }
    }
    ar1 <- tryCatch(isTRUE(attr(object, "ild_ar1", exact = TRUE)), error = function(e) FALSE)
    sc <- tryCatch(ild_spacing_class(data), error = function(e) NA_character_)
    if (!is.na(sc) && sc == "irregular-ish" && !ar1) {
      rows[[length(rows) + 1L]] <- list(
        rule_id = "GR_IRREGULAR_SPACING_NO_RESID_COR",
        triggered = TRUE,
        message = NULL,
        recommendation = NULL
      )
    }
    pg <- if (!is.null(bundle$data)) bundle$data$pct_gap else NA_real_
    if (is.na(pg)) {
      pg <- tryCatch(ild_summary(data)$pct_gap, error = function(e) NA_real_)
    }
    if (is.finite(pg) && pg > 25) {
      rows[[length(rows) + 1L]] <- list(
        rule_id = "GR_HIGH_TIMING_GAP_RATE",
        triggered = TRUE,
        message = sprintf(
          "About %.1f%% of intervals are gaps (threshold for rule > 25%%).",
          pg
        ),
        recommendation = NULL
      )
    }
    rn <- tryCatch(ild_response_name(object), error = function(e) NULL)
    if (!is.null(rn) && rn %in% names(data)) {
      if (guardrail_dropout_late_heuristic(data, rn)) {
        rows[[length(rows) + 1L]] <- list(
          rule_id = "GR_DROPOUT_LATE_CONCENTRATION",
          triggered = TRUE,
          message = sprintf(
            "Outcome '%s' is missing more often in the later half of the pooled timeline than in the earlier half.",
            rn
          ),
          recommendation = NULL
        )
      }
    }
  }
  if (!is.null(bundle$causal) && is.list(bundle$causal$weight_summary)) {
    ws <- bundle$causal$weight_summary
    mx <- ws$max
    mn <- ws$min
    if (is.finite(mx) && is.finite(mn) && mn > 0 && (mx / mn) > 50) {
      rows[[length(rows) + 1L]] <- list(
        rule_id = "GR_IPW_WEIGHTS_UNSTABLE",
        triggered = TRUE,
        message = sprintf("IPW weight range ratio max/min = %.1f (threshold > 50).", mx / mn),
        recommendation = NULL
      )
    }
  }
  if (!is.null(data) && is_ild(data)) {
    .wm_ratio <- function(w) {
      wf <- w[is.finite(w) & w > 0]
      if (length(wf) < 2L) return(NA_real_)
      max(wf) / min(wf)
    }
    for (cn in c(".ipw_treat", ".ipw_censor")) {
      if (cn %in% names(data)) {
        r <- .wm_ratio(data[[cn]])
        if (is.finite(r) && r > 50) {
          rows[[length(rows) + 1L]] <- list(
            rule_id = "GR_MSM_COMPONENT_WEIGHTS_UNSTABLE",
            triggered = TRUE,
            message = sprintf(
              "Component %s weight range ratio max/min = %.1f (threshold > 50).",
              cn, r
            ),
            recommendation = NULL
          )
        }
      }
    }
  }
  if (!is.null(bundle$causal) && is.list(bundle$causal$balance) && !is.null(bundle$causal$balance$table)) {
    tab <- bundle$causal$balance$table
    if (inherits(tab, "tbl_df") && nrow(tab) > 0L) {
      mx <- suppressWarnings(max(abs(tab$smd), na.rm = TRUE))
      if (is.finite(mx) && mx > 0.25) {
        rows[[length(rows) + 1L]] <- list(
          rule_id = "GR_MSM_BALANCE_SMD_HIGH",
          triggered = TRUE,
          message = sprintf("Maximum |SMD| across balance table = %.3f (threshold > 0.25).", mx),
          recommendation = NULL
        )
      }
    }
    ess <- bundle$causal$balance$ess
    ess_min <- if (inherits(ess, "tbl_df")) {
      suppressWarnings(min(ess$ess, na.rm = TRUE))
    } else {
      as.numeric(ess)[1L]
    }
    n <- if (!is.null(data) && is_ild(data)) nrow(data) else NA_integer_
    if (is.finite(ess_min) && is.finite(n) && ess_min < max(30, 0.05 * n)) {
      rows[[length(rows) + 1L]] <- list(
        rule_id = "GR_MSM_ESS_LOW",
        triggered = TRUE,
        message = sprintf(
          "Minimum ESS = %.1f (threshold: max(30, 0.05 * N) with N = %d).",
          ess_min, n
        ),
        recommendation = NULL
      )
    }
  }
  if (identical(engine, "ctsem") && !is.null(data) && is_ild(data)) {
    st_dim <- tryCatch(as.numeric(bundle$fit$state_dimension)[1L], error = function(e) NA_real_)
    n <- nrow(data)
    min_n <- if (is.finite(st_dim) && st_dim > 1) 40 else 25
    if (is.finite(n) && n < min_n) {
      rows[[length(rows) + 1L]] <- list(
        rule_id = "GR_CTSEM_SHORT_SERIES_FOR_COMPLEX_DYNAMICS",
        triggered = TRUE,
        message = sprintf(
          "n = %d observations; recommend at least %d for this ctsem complexity.",
          n, min_n
        ),
        recommendation = NULL
      )
    }
  }
  guardrail_finalize_rows(rows)
}

#' @keywords internal
#' @noRd
guardrail_dropout_late_heuristic <- function(data, outcome) {
  if (!outcome %in% names(data)) return(FALSE)
  ord <- order(data[[".ild_seq"]], seq_len(nrow(data)))
  miss <- is.na(data[[outcome]])
  miss <- miss[ord]
  n <- length(miss)
  if (n < 20L) return(FALSE)
  mid <- floor(n / 2)
  r1 <- mean(miss[seq_len(mid)])
  r2 <- mean(miss[(mid + 1L):n])
  r2 > pmax(1.5 * r1, r1 + 0.05, na.rm = TRUE)
}

#' Evaluate fit-level guardrails (frequentist + Bayesian)
#' @keywords internal
#' @noRd
evaluate_guardrails_fit <- function(object, fit_diag, engine = c("lmer", "lme", "brms", "ctsem")) {
  engine <- match.arg(engine)
  fd <- fit_bundle_flat_for_guardrails(fit_diag, engine = engine)
  rows <- list()
  if (engine %in% c("lmer", "lme")) {
    sing <- fd$singular
    if (!is.na(sing) && isTRUE(sing)) {
      rows[[length(rows) + 1L]] <- list(
        rule_id = "GR_SINGULAR_RANDOM_EFFECTS",
        triggered = TRUE,
        message = NULL,
        recommendation = NULL
      )
    }
  }
  if (engine %in% c("lmer", "lme", "brms")) {
    het <- fit_diag$heterogeneity
    if (!is.null(het) && isTRUE(het$available) && !is.null(het$object)) {
      sig <- tryCatch(stats::sigma(object), error = function(e) NA_real_)
      sm <- het$object$summary
      # isTRUE guards nrow/sig comparisons that can be NA; skip heuristics when sigma is
      # numerically degenerate (e.g. failed lmer) to avoid false guardrails / if(NA).
      sig_ok <- is.finite(sig) && isTRUE(sig > 1e-8)
      if (!is.null(sm) && isTRUE(nrow(sm) > 0L) && sig_ok) {
        hit_var <- FALSE
        for (k in seq_len(nrow(sm))) {
          termk <- as.character(sm$term[k])
          if (termk %in% c("(Intercept)", "Intercept")) {
            next
          }
          vcs <- sm$varcorr_sdcor[k]
          if (is.finite(vcs) && vcs < 0.01 * sig) {
            hit_var <- TRUE
            break
          }
        }
        if (hit_var) {
          rows[[length(rows) + 1L]] <- list(
            rule_id = "GR_RE_SLOPE_VARIANCE_VERSUS_RESIDUAL_LOW",
            triggered = TRUE,
            message = NULL,
            recommendation = NULL
          )
        }
        sdt_max <- suppressWarnings(max(sm$sd_total, na.rm = TRUE))
        if (is.finite(sdt_max) && sdt_max < 0.05 * sig) {
          rows[[length(rows) + 1L]] <- list(
            rule_id = "GR_PERSON_SPECIFIC_SLOPES_EMPIRICALLY_TIGHT",
            triggered = TRUE,
            message = NULL,
            recommendation = NULL
          )
        }
      }
    }
  }
  if (engine == "brms") {
    mx <- fd$max_rhat
    if (!is.na(mx) && mx > 1.05) {
      rows[[length(rows) + 1L]] <- list(
        rule_id = "GR_POOR_POSTERIOR_MIXING",
        triggered = TRUE,
        message = sprintf("Max R-hat = %.3f (threshold 1.05).", mx),
        recommendation = NULL
      )
    }
    ct <- fd$convergence_table
    if (!is.null(ct) && nrow(ct) > 0L && "ess_bulk" %in% names(ct)) {
      mness <- suppressWarnings(min(ct$ess_bulk, na.rm = TRUE))
      if (is.finite(mness) && mness < 100) {
        rows[[length(rows) + 1L]] <- list(
          rule_id = "GR_LOW_ESS",
          triggered = TRUE,
          message = sprintf("Minimum bulk ESS = %.0f (threshold 100).", mness),
          recommendation = NULL
        )
      }
    }
  }
  if (engine == "ctsem") {
    conv <- tryCatch(fit_diag$convergence$converged, error = function(e) NA)
    if (!is.na(conv) && !isTRUE(conv)) {
      rows[[length(rows) + 1L]] <- list(
        rule_id = "GR_CTSEM_NONCONVERGENCE",
        triggered = TRUE,
        message = NULL,
        recommendation = NULL
      )
    }
    # Heuristic instability signal: very large absolute drift entries.
    drift_max <- tryCatch(as.numeric(fit_diag$drift_abs_max)[1L], error = function(e) NA_real_)
    if (is.finite(drift_max) && drift_max > 5) {
      rows[[length(rows) + 1L]] <- list(
        rule_id = "GR_CTSEM_UNSTABLE_DYNAMICS",
        triggered = TRUE,
        message = sprintf("Maximum absolute drift estimate = %.2f (heuristic threshold > 5).", drift_max),
        recommendation = NULL
      )
    }
  }
  guardrail_finalize_rows(rows)
}

#' @keywords internal
#' @noRd
guardrails_bind <- function(...) {
  parts <- list(...)
  nonempty <- parts[vapply(parts, function(z) nrow(z) > 0L, logical(1))]
  if (length(nonempty) == 0L) {
    return(guardrails_empty_tibble())
  }
  dplyr::bind_rows(nonempty)
}

# Canonical guardrail rule_id triggers — deterministic fixtures + fit-level unit tests

test_that("evaluate_guardrails_fit triggers GR_SINGULAR_RANDOM_EFFECTS (lmer)", {
  gr <- tidyILD:::evaluate_guardrails_fit(
    structure(list(), class = "lmerMod"),
    list(singular = TRUE),
    engine = "lmer"
  )
  expect_true(any(gr$rule_id == "GR_SINGULAR_RANDOM_EFFECTS"))
})

test_that("evaluate_guardrails_fit triggers GR_POOR_POSTERIOR_MIXING and GR_LOW_ESS (brms)", {
  gr_mix <- tidyILD:::evaluate_guardrails_fit(
    structure(list(), class = "brmsfit"),
    list(max_rhat = 1.08, convergence_table = tibble::tibble(ess_bulk = 200)),
    engine = "brms"
  )
  expect_true(any(gr_mix$rule_id == "GR_POOR_POSTERIOR_MIXING"))

  gr_ess <- tidyILD:::evaluate_guardrails_fit(
    structure(list(), class = "brmsfit"),
    list(
      max_rhat = 1.01,
      convergence_table = tibble::tibble(ess_bulk = c(150, 50))
    ),
    engine = "brms"
  )
  expect_true(any(gr_ess$rule_id == "GR_LOW_ESS"))
})

test_that("evaluate_guardrails_fit/contextual trigger ctsem guardrails", {
  gr_fit <- tidyILD:::evaluate_guardrails_fit(
    structure(list(), class = "ild_fit_ctsem"),
    list(
      convergence = list(converged = FALSE),
      drift_abs_max = 6
    ),
    engine = "ctsem"
  )
  expect_true(any(gr_fit$rule_id == "GR_CTSEM_NONCONVERGENCE"))
  expect_true(any(gr_fit$rule_id == "GR_CTSEM_UNSTABLE_DYNAMICS"))

  d <- ild_simulate(n_id = 1, n_obs_per = 12, seed = 7110)
  x <- ild_prepare(d, id = "id", time = "time")
  b <- ild_diagnostics_bundle(
    fit = list(state_dimension = 2),
    data = list(pct_gap = NA_real_),
    design = list(),
    residual = list(),
    predictive = list(),
    missingness = list(),
    causal = list(),
    meta = list(),
    warnings = tibble::tibble(),
    guardrails = tidyILD:::guardrails_empty_tibble(),
    summary_text = character()
  )
  gr_ctx <- tidyILD:::evaluate_guardrails_contextual(
    structure(list(), class = "ild_fit_ctsem"),
    x,
    b,
    engine = "ctsem"
  )
  expect_true(any(gr_ctx$rule_id == "GR_CTSEM_SHORT_SERIES_FOR_COMPLEX_DYNAMICS"))
})

test_that("ild_diagnose triggers GR_MIXED_PREDICTOR_NOT_DECOMPOSED", {
  d <- ild_simulate(n_id = 8, n_obs_per = 5, seed = 7001)
  d$stress <- as.numeric(scale(rnorm(nrow(d)))) +
    rep(as.numeric(scale(rnorm(8))), each = 5L) * 0.3
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ stress + y_bp + y_wp + (1 | id), data = x, ar1 = FALSE,
    warn_no_ar1 = FALSE, warn_uncentered = FALSE)
  b <- ild_diagnose(fit, data = x, type = "qq")
  expect_true(any(b$guardrails$rule_id == "GR_MIXED_PREDICTOR_NOT_DECOMPOSED"))
})

test_that("ild_diagnose triggers GR_IRREGULAR_SPACING_NO_RESID_COR and GR_HIGH_TIMING_GAP_RATE", {
  d <- ild_simulate(n_id = 5, n_obs_per = 8, seed = 7002)
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 0)
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x, type = "qq")
  expect_true(any(b$guardrails$rule_id == "GR_IRREGULAR_SPACING_NO_RESID_COR"))
  expect_true(any(b$guardrails$rule_id == "GR_HIGH_TIMING_GAP_RATE"))
})

test_that("ild_diagnose triggers GR_IPW_WEIGHTS_UNSTABLE", {
  d <- ild_simulate(n_id = 6, n_obs_per = 5, seed = 7003)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  w <- rep(c(1, 60), length.out = nrow(x))
  x$.ipw <- w
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x, type = "qq")
  expect_true(any(b$guardrails$rule_id == "GR_IPW_WEIGHTS_UNSTABLE"))
})

test_that("evaluate_guardrails_contextual triggers GR_MSM_COMPONENT_WEIGHTS_UNSTABLE", {
  d <- ild_simulate(n_id = 6, n_obs_per = 5, seed = 7003)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  w <- rep(c(1, 60), length.out = nrow(x))
  x$.ipw_treat <- w
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  gr <- tidyILD:::evaluate_guardrails_contextual(fit, x, ild_diagnostics_bundle(), engine = "lmer")
  expect_true(any(gr$rule_id == "GR_MSM_COMPONENT_WEIGHTS_UNSTABLE"))
})

test_that("evaluate_guardrails_contextual triggers GR_DROPOUT_LATE_CONCENTRATION", {
  d <- ild_simulate(n_id = 4, n_obs_per = 6, seed = 7004)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  x$y[x$.ild_seq > 3] <- NA
  xc <- x[stats::complete.cases(x$y), , drop = FALSE]
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = xc, ar1 = FALSE, warn_no_ar1 = FALSE)
  expect_true(tidyILD:::guardrail_dropout_late_heuristic(x, "y"))
  gr <- tidyILD:::evaluate_guardrails_contextual(fit, x, ild_diagnostics_bundle(), engine = "lmer")
  expect_true(any(gr$rule_id == "GR_DROPOUT_LATE_CONCENTRATION"))
})

test_that("ild_guardrails_summary and print bundle include guardrail summary", {
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 7005)
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 0)
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x, type = "qq")
  gs <- tidyILD:::ild_guardrails_summary(b)
  expect_true(gs$n >= 1L)
  expect_true(nzchar(gs$narrative))
  pr <- utils::capture.output(print(b))
  expect_true(any(grepl("Summary:", pr, fixed = TRUE)))
  expect_true(any(grepl("guardrails:", pr, fixed = TRUE)))
  expect_true(any(grepl("key rule_id", pr, fixed = TRUE)))
})

test_that("ild_methods appends guardrails when bundle has rows", {
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 7006)
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 0)
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x, type = "qq")
  m <- ild_methods(fit, bundle = b)
  expect_true(grepl("Methodological cautions (tidyILD guardrails)", m, fixed = TRUE))
})

test_that("ild_report includes diagnostics_summary guardrails fields when triggered", {
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 7007)
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 0)
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  r <- ild_report(fit)
  expect_true(!is.null(r$diagnostics_summary$guardrails_narrative))
  expect_true(!is.null(r$diagnostics_summary$guardrails))
  expect_true(!is.null(r$methods_with_guardrails))
  expect_true(nchar(r$diagnostics_summary$guardrails_narrative) > 0)
})

# Canonical contract-regression fixtures — deterministic seeds, semantic expectations.
# Loaded automatically by testthat.

#' @keywords internal
contract_tidy_required <- function() {
  ild_tidy_schema()$required
}

#' @keywords internal
contract_augment_required <- function() {
  ild_augment_schema()$required
}

#' Assert ild_tidy / ild_augment column contracts for a frequentist ild_lme fit
#' @keywords internal
contract_assert_tidy_augment_lmer <- function(fit) {
  td <- tidy_ild_model(fit, object = FALSE)
  req <- contract_tidy_required()
  expect_true(all(req %in% names(td)))
  ag <- augment_ild_model(fit)
  expect_true(all(contract_augment_required() %in% names(ag)))
}

#' Core bundle sections populated by ild_diagnose(lmer)
#' @keywords internal
contract_assert_bundle_dense_lmer <- function(b) {
  expect_true(!is.null(b$meta))
  expect_true(!is.null(b$data))
  expect_true(!is.null(b$data$cohort))
  expect_true(!is.null(b$data$obs_per_id))
  expect_true(!is.null(b$design))
  expect_true(!is.null(b$design$flags))
  expect_true(!is.null(b$fit))
  expect_true(!is.null(b$residual))
  expect_true(!is.null(b$predictive))
  expect_true(!is.null(b$missingness))
  expect_true(!is.null(attr(b, "ild_fit", exact = TRUE)))
  expect_true(!is.null(attr(b, "ild_data", exact = TRUE)))
}

#' Frequentist autoplot entry points return ggplot
#' @keywords internal
contract_assert_autoplot_frequentist_core <- function(b) {
  expect_s3_class(ild_autoplot(b, section = "residual", type = "acf"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "residual", type = "qq"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "residual", type = "fitted"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "fit", type = "convergence"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "data", type = "missingness"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "design", type = "coverage"), "ggplot")
}

## --- Fixtures (fixed seeds) -------------------------------------------------

#' Regular-ish spacing, WP/BP decomposition, no intentional guardrails
fixture_contract_regular_decomposed <- function() {
  d <- ild_simulate(n_id = 10, n_obs_per = 8, seed = 9101L)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x, type = "qq")
  list(
    name = "regular_decomposed",
    data = x,
    fit = fit,
    bundle = b,
    spacing_class = ild_spacing_class(x),
    expect_rule_ids = character(0),
    expect_methods_guardrail = FALSE
  )
}

#' Irregular spacing, no AR1 — irregular + high gap guardrails
fixture_contract_irregular_no_ar1 <- function() {
  d <- ild_simulate(n_id = 5, n_obs_per = 8, seed = 9102L)
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 0)
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x, type = "qq")
  list(
    name = "irregular_no_ar1",
    data = x,
    fit = fit,
    bundle = b,
    spacing_class = ild_spacing_class(x),
    expect_rule_ids = c("GR_IRREGULAR_SPACING_NO_RESID_COR", "GR_HIGH_TIMING_GAP_RATE"),
    expect_methods_guardrail = TRUE
  )
}

#' Merge singular-fit guardrail row into a real diagnosed bundle (contract for GR_SINGULAR)
fixture_contract_singular_guardrail <- function() {
  f <- fixture_contract_regular_decomposed()
  gr <- tidyILD:::evaluate_guardrails_fit(
    structure(list(), class = "lmerMod"),
    list(singular = TRUE),
    engine = "lmer"
  )
  f$bundle$guardrails <- dplyr::bind_rows(f$bundle$guardrails, gr)
  f$name <- "singular_lmer_guardrail"
  f$expect_rule_ids <- "GR_SINGULAR_RANDOM_EFFECTS"
  f$expect_methods_guardrail <- TRUE
  f
}

#' Unstable IPW weights
fixture_contract_ipw_unstable <- function() {
  d <- ild_simulate(n_id = 6, n_obs_per = 5, seed = 9103L)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  x$.ipw <- rep(c(1, 60), length.out = nrow(x))
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x, type = "qq")
  list(
    name = "ipw_unstable",
    data = x,
    fit = fit,
    bundle = b,
    spacing_class = ild_spacing_class(x),
    expect_rule_ids = "GR_IPW_WEIGHTS_UNSTABLE",
    expect_methods_guardrail = TRUE
  )
}

#' Late dropout: fit on complete cases; full data has NA pattern; merge contextual dropout rows
fixture_contract_dropout_late <- function() {
  d <- ild_simulate(n_id = 4, n_obs_per = 6, seed = 9104L)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  x$y[x$.ild_seq > 3] <- NA
  xc <- x[stats::complete.cases(x$y), , drop = FALSE]
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = xc, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = xc, type = "qq")
  gr <- tidyILD:::evaluate_guardrails_contextual(fit, x, b, engine = "lmer")
  gr <- gr[gr$rule_id == "GR_DROPOUT_LATE_CONCENTRATION", , drop = FALSE]
  b$guardrails <- dplyr::bind_rows(b$guardrails, gr)
  list(
    name = "dropout_late",
    data_full = x,
    data_fit = xc,
    fit = fit,
    bundle = b,
    expect_rule_ids = "GR_DROPOUT_LATE_CONCENTRATION",
    expect_methods_guardrail = TRUE
  )
}

#' Brms: diagnosed bundle + merged poor-mixing / low-ESS guardrails (no MCMC dependency)
fixture_contract_brms_poor_mixing_merged <- function() {
  skip_if_not_installed("brms")
  d <- ild_simulate(n_id = 5, n_obs_per = 4, seed = 9105L)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_brms(
    y ~ y_bp + y_wp + (1 | id),
    data = x,
    chains = 1,
    iter = 80,
    warmup = 40,
    refresh = 0,
    seed = 9105L,
    silent = 2,
    backend = "rstan",
    warn_uncentered = FALSE
  )
  b <- ild_diagnose(fit, type = "all")
  gr_mix <- tidyILD:::evaluate_guardrails_fit(
    fit,
    list(max_rhat = 1.09, convergence_table = tibble::tibble(ess_bulk = 200, term = "b_Intercept")),
    engine = "brms"
  )
  gr_ess <- tidyILD:::evaluate_guardrails_fit(
    fit,
    list(
      max_rhat = 1.01,
      convergence_table = tibble::tibble(ess_bulk = c(40, 30), term = c("a", "b"))
    ),
    engine = "brms"
  )
  b$guardrails <- dplyr::bind_rows(b$guardrails, gr_mix, gr_ess)
  list(
    name = "brms_poor_mixing_merged",
    data = x,
    fit = fit,
    bundle = b,
    expect_rule_ids = c("GR_POOR_POSTERIOR_MIXING", "GR_LOW_ESS"),
    expect_methods_guardrail = TRUE
  )
}

test_that("GR_LAG_MEAN_STRONG_RESIDUAL_ACF_NO_AR triggers with synthetic ACF", {
  d <- ild_simulate(n_id = 10, n_obs_per = 14, seed = 906L)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_lag(x, y, n = 1L, mode = "gap_aware")
  x <- dplyr::filter(x, !is.na(.data$y_lag1))
  fit <- ild_lme(y ~ y_lag1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
  b <- ild_diagnose(fit, data = x, type = c("residual_acf", "qq"))
  b$residual$stats$acf$pooled <- tibble::tibble(lag = c(0, 1), acf = c(1, 0.35))
  gr <- tidyILD:::evaluate_guardrails_contextual(fit, x, b, engine = "lmer")
  expect_true(any(gr$rule_id == "GR_LAG_MEAN_STRONG_RESIDUAL_ACF_NO_AR"))
})

test_that("GR_INDEX_LAG_IRREGULAR_SPACING triggers with index lag on irregular ILD", {
  d <- ild_simulate(n_id = 6, n_obs_per = 10, irregular = TRUE, seed = 907L)
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 0)
  x <- ild_lag(x, y, n = 1L, mode = "index")
  x <- dplyr::filter(x, !is.na(.data$y_lag1))
  fit <- ild_lme(y ~ y_lag1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
  b <- ild_diagnose(fit, data = x, type = "qq")
  gr <- tidyILD:::evaluate_guardrails_contextual(fit, x, b, engine = "lmer")
  expect_true(any(gr$rule_id == "GR_INDEX_LAG_IRREGULAR_SPACING"))
})

# tsibble interoperability: capture provenance on input, optional round-trip via ild_as_tsibble()

#' Capture tsibble metadata before coercion to plain tibble (internal)
#' @keywords internal
#' @noRd
ild_tsibble_capture_meta <- function(data) {
  if (!requireNamespace("tsibble", quietly = TRUE)) {
    return(NULL)
  }
  if (!inherits(data, "tbl_ts")) {
    return(NULL)
  }
  kv <- tsibble::key_vars(data)
  iv <- tsibble::index_var(data)
  intv <- tryCatch(tsibble::interval(data), error = function(e) NULL)
  int_txt <- tryCatch({
    if (!is.null(intv)) format(intv) else NA_character_
  }, error = function(e) NA_character_)
  reg <- tryCatch(tsibble::is_regular(data), error = function(e) NA)
  idx_vec <- tryCatch(tsibble::index(data), error = function(e) NULL)
  idx_class <- if (!is.null(idx_vec)) class(idx_vec)[[1L]] else NA_character_
  list(
    source_class = "tbl_ts",
    key_vars = kv,
    index_var = iv,
    interval_format = int_txt,
    is_regular = reg,
    tsibble_version = tryCatch(
      as.character(utils::packageVersion("tsibble")),
      error = function(e) NA_character_
    ),
    index_type = idx_class
  )
}

#' Merge tsibble declaration with empirical ILD spacing (internal)
#' @keywords internal
#' @noRd
ild_spacing_tsibble_bridge <- function(spacing, ts_meta) {
  if (is.null(ts_meta)) {
    return(spacing)
  }
  spacing$tsibble <- list(
    declared_interval = ts_meta$interval_format,
    key_vars = ts_meta$key_vars,
    index_var = ts_meta$index_var,
    tsibble_was_regular = ts_meta$is_regular,
    median_dt_sec_empirical = spacing$median_dt,
    iqr_dt_sec_empirical = spacing$iqr_dt,
    note = paste0(
      "Declared tsibble interval/regularity reflects the input tbl_ts before ild_prepare(). ",
      "Empirical median_dt / iqr_dt are from .ild_dt after sorting and duplicate handling; ",
      "they may differ if rows were dropped or timestamps changed."
    )
  )
  spacing
}

#' Tsibble provenance from an ILD object
#'
#' If [ild_prepare()] was called on a \code{tbl_ts}, metadata from the source
#' tsibble (key, index, interval summary, regularity) is stored in
#' \code{attr(x, "tidyILD")$tsibble}. Returns \code{NULL} if the object was
#' not prepared from a tsibble.
#'
#' @param x An object that passes [validate_ild()].
#' @return A named list or \code{NULL}.
#' @seealso [ild_prepare()], [ild_as_tsibble()],
#'   \href{../doc/tsibble-interoperability.html}{\code{vignette("tsibble-interoperability", package = "tidyILD")}}
#' @export
ild_tsibble_meta <- function(x) {
  validate_ild(x)
  b <- attr(x, "tidyILD", exact = TRUE)
  if (is.null(b) || is.null(b$tsibble)) {
    return(NULL)
  }
  b$tsibble
}

#' Convert an ILD object to a tsibble
#'
#' Wraps [tsibble::as_tsibble()] using the subject and time column names from
#' [ild_meta()] as key and index. When [ild_prepare()] recorded tsibble
#' provenance (\code{ild_tsibble_meta()} is non-\code{NULL}), the \code{regular}
#' argument is set from the stored \code{is_regular} flag so the reconstructed
#' \code{interval} often matches the original for unchanged data. Otherwise
#' defaults to \code{regular = TRUE}.
#'
#' @param x An object that passes [validate_ild()].
#' @param ... Optional arguments passed to [tsibble::as_tsibble()] (e.g.
#'   override \code{regular}).
#' @return A \code{tbl_ts} object (see \pkg{tsibble}).
#' @examples
#' \donttest{
#' d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 1)
#' x <- ild_prepare(d, id = "id", time = "time")
#' if (requireNamespace("tsibble", quietly = TRUE)) {
#'   ts <- ild_as_tsibble(x)
#'   class(ts)
#' }
#' }
#' @seealso \href{../doc/tsibble-interoperability.html}{\code{vignette("tsibble-interoperability", package = "tidyILD")}}
#' @importFrom rlang inject sym
#' @export
ild_as_tsibble <- function(x, ...) {
  validate_ild(x)
  if (!requireNamespace("tsibble", quietly = TRUE)) {
    stop("Package 'tsibble' is required for ild_as_tsibble(). Install it with install.packages(\"tsibble\").", call. = FALSE)
  }
  m <- ild_meta(x)
  id_col <- m$ild_id
  time_col <- m$ild_time
  dots <- list(...)
  regular_arg <- dots$regular
  dots$regular <- NULL
  tsm <- ild_tsibble_meta(x)
  if (!is.null(regular_arg)) {
    regular_use <- regular_arg
  } else if (!is.null(tsm) && is.logical(tsm$is_regular) && length(tsm$is_regular) == 1L && !is.na(tsm$is_regular)) {
    regular_use <- tsm$is_regular
  } else {
    regular_use <- TRUE
  }
  out <- rlang::inject(tsibble::as_tsibble(
    tibble::as_tibble(x),
    key = !!rlang::sym(id_col),
    index = !!rlang::sym(time_col),
    regular = regular_use,
    !!!dots
  ))
  if (!is.null(tsm) && isTRUE(tsm$is_regular) && !isTRUE(tsibble::is_regular(out))) {
    warning(
      "ild_as_tsibble(): reconstructed tsibble is not regular; empirical spacing may differ from the original tbl_ts (sorting, duplicates, or row drops in ild_prepare).",
      call. = FALSE
    )
  }
  out
}

test_that("ild_as_tsibble round-trips key and index", {
  skip_if_not_installed("tsibble")
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 11)
  x <- ild_prepare(d, id = "id", time = "time")
  ts <- ild_as_tsibble(x)
  expect_true(inherits(ts, "tbl_ts"))
  expect_equal(tsibble::key_vars(ts), "id")
  expect_equal(tsibble::index_var(ts), "time")
})

test_that("ild_prepare infers id and time from tbl_ts", {
  skip_if_not_installed("tsibble")
  d <- ild_simulate(n_id = 3, n_obs_per = 4, seed = 12)
  x0 <- ild_prepare(d, id = "id", time = "time")
  ts <- ild_as_tsibble(x0)
  x1 <- ild_prepare(ts)
  expect_true(is_ild(x1))
  expect_equal(attr(x1, "ild_id", exact = TRUE), "id")
  expect_equal(attr(x1, "ild_time", exact = TRUE), "time")
  expect_equal(nrow(x0), nrow(x1))
})

test_that("ild_prepare omit id/time captures tsibble meta and spacing bridge", {
  skip_if_not_installed("tsibble")
  suppressPackageStartupMessages(library(tsibble))
  t <- yearmonth(c(201901, 201902, 201903, 201901, 201902, 201903))
  d <- tsibble(
    id = rep(c("a", "b"), each = 3),
    t = t,
    y = rnorm(6),
    key = id,
    index = t
  )
  x <- ild_prepare(d)
  tm <- ild_tsibble_meta(x)
  expect_type(tm, "list")
  expect_identical(tm$source_class, "tbl_ts")
  expect_identical(tm$key_vars, "id")
  expect_identical(tm$index_var, "t")
  expect_true(is.logical(tm$is_regular) && !is.na(tm$is_regular))
  sp <- attr(x, "ild_spacing", exact = TRUE)
  expect_true(!is.null(sp$tsibble))
  expect_identical(sp$tsibble$index_var, "t")
})

test_that("ild_prepare explicit id/time on tbl_ts still captures meta", {
  skip_if_not_installed("tsibble")
  suppressPackageStartupMessages(library(tsibble))
  t <- yearmonth(c(201901, 201902, 201903))
  d <- tsibble(id = rep(1, 3), t = t, y = 1:3, key = id, index = t)
  x <- ild_prepare(d, id = "id", time = "t")
  tm <- ild_tsibble_meta(x)
  expect_identical(tm$source_class, "tbl_ts")
})

test_that("ild_as_tsibble round-trip preserves interval for unchanged regular series", {
  skip_if_not_installed("tsibble")
  suppressPackageStartupMessages(library(tsibble))
  t <- yearmonth(c(201901, 201902, 201903))
  d <- tsibble(id = rep(1, 3), t = t, y = 1:3, key = id, index = t)
  iv0 <- tsibble::interval(d)
  x <- ild_prepare(d)
  d2 <- ild_as_tsibble(x)
  expect_s3_class(d2, "tbl_ts")
  expect_true(identical(tsibble::interval(d2), iv0))
})

test_that("non-tsibble ild_prepare has NULL ild_tsibble_meta", {
  d <- data.frame(id = c(1, 1), time = 1:2, y = 1:2)
  x <- ild_prepare(d, id = "id", time = "time")
  expect_null(ild_tsibble_meta(x))
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
has_kfas <- requireNamespace("KFAS", quietly = TRUE)

## ----example, eval = has_kfas-------------------------------------------------
library(tidyILD)
set.seed(1)
d <- ild_simulate(n_id = 1, n_obs_per = 60, seed = 42)
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_center(x, y)
fit <- suppressWarnings(
  ild_kfas(x, outcome = "y", state_spec = "local_level", time_units = "sim_steps")
)
b <- ild_diagnose(fit)
class(b)
ild_autoplot(b, section = "residual", type = "acf")

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----eval = FALSE-------------------------------------------------------------
# library(tidyILD)
# 
# d <- ild_simulate(n_id = 1, n_obs_per = 60, seed = 501)
# x <- ild_prepare(d, id = "id", time = "time")
# x <- ild_center(x, y)
# 
# fit_ct <- ild_ctsem(
#   data = x,
#   outcome = "y",
#   model_type = "stanct",
#   chains = 1,
#   iter = 400
# )
# 
# fit_ct
# td <- ild_tidy(fit_ct)
# ag <- ild_augment(fit_ct)
# dg <- ild_diagnose(fit_ct)

## ----eval = FALSE-------------------------------------------------------------
# ild_autoplot(fit_ct, type = "fitted_vs_actual")
# ild_autoplot(fit_ct, type = "residual_time")
# ild_autoplot(dg, section = "fit", type = "convergence")
# ild_autoplot(dg, section = "residual", type = "acf")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----pattern------------------------------------------------------------------
library(tidyILD)
set.seed(11)
d <- ild_simulate(n_id = 25, n_obs_per = 12, seed = 11)
d$stress <- rnorm(nrow(d))
d$mood <- d$y
miss_i <- sample(nrow(d), 45)
d$mood[miss_i] <- NA
x <- ild_prepare(d, id = "id", time = "time")
mp <- ild_missing_pattern(x, vars = c("mood", "stress"), outcome = "mood")
mp$summary
head(mp$by_id, 3)

## ----compliance---------------------------------------------------------------
cm <- ild_missing_compliance(x, outcome = "mood", expected_occasions = 12L)
summary(cm$pct_nonmissing_outcome)

## ----cohort-hazard------------------------------------------------------------
coh <- ild_missing_cohort(x, outcome = "mood", plot = FALSE)
head(coh$by_occasion)
head(ild_missing_hazard_first(x, outcome = "mood"))

## ----report-------------------------------------------------------------------
rpt <- ild_missingness_report(
  x,
  outcome = "mood",
  predictors = "stress",
  fit_missing_model = TRUE,
  random = FALSE,
  cohort_plot = FALSE
)
names(rpt)
rpt$snippets["overview"]

## ----ipw-template, eval = FALSE-----------------------------------------------
# mm <- ild_missing_model(x, outcome = "mood", predictors = c("stress"), random = TRUE)
# x_w <- ild_ipw_weights(x, mm, stabilize = TRUE)
# fit_w <- ild_ipw_refit(mood ~ stress + (1 | id), data = x_w, weights = ".ipw")

## ----cc-vs-full, eval = FALSE-------------------------------------------------
# x_cc <- dplyr::filter(x, !is.na(mood))
# fit_full <- ild_lme(mood ~ stress + (1 | id), data = x, warn_uncentered = FALSE)
# fit_cc <- ild_lme(mood ~ stress + (1 | id), data = x_cc, warn_uncentered = FALSE)

# Methods narrative from provenance
# Uses ild_provenance(), ild_flatten_provenance() from ild_provenance.R
# Step order is preserved exactly (data steps then analysis steps).

# Format a "created" vector for methods text (internal)
.format_created <- function(o) {
  if (is.null(o) || length(o) == 0) return("")
  cre <- o$created
  if (is.null(cre)) return("")
  if (length(cre) == 1) return(paste0(", creating ", cre, "."))
  paste0(", creating ", paste(cre, collapse = ", "), ".")
}

#' Turn one step record into a methods sentence (internal)
#' @param s Step record (list with step, args, outputs).
#' @param context Optional. The object x passed to ild_methods(); used to read ild_robust_se when step is ild_lme.
#' @param robust_se Optional. Passed from ild_methods(robust_se = ); used when step is ild_lme to mention cluster-robust SEs.
#' @noRd
ild_step_to_sentence <- function(s, context = NULL, robust_se = NULL) {
  step <- s$step
  a <- s$args
  o <- s$outputs
  if (step == "ild_prepare") {
    if (is.numeric(a$gap_threshold) && is.finite(a$gap_threshold)) {
      gt_phrase <- paste0("a gap threshold of ", a$gap_threshold, " units")
    } else {
      gt_phrase <- "no gap threshold"
    }
    out <- paste0(
      "Data were prepared using ild_prepare() with participant ID ", .def(a$id, "id"),
      ", time variable ", .def(a$time, "time"),
      ", and ", gt_phrase
    )
    if (!is.null(o$spacing_class) && length(o$spacing_class) == 1 && nchar(o$spacing_class) > 0) {
      out <- paste0(out, " (spacing class: ", o$spacing_class, ")")
    }
    if (!is.null(o$n_id) && !is.null(o$n_obs)) {
      out <- paste0(out, " (N = ", o$n_id, " persons, n = ", o$n_obs, " observations)")
    }
    return(paste0(out, "."))
  }
  if (step == "ild_center") {
    vars <- a$vars
    if (length(vars) > 1) vars <- paste(vars, collapse = ", ")
    type <- .def(a$type, "person_mean")
    out <- paste0(
      "Predictor(s) ", vars, " were ",
      if (type == "grand_mean") "grand-mean" else "person-mean",
      " centered using ild_center()"
    )
    out <- paste0(out, .format_created(o))
    if (!endsWith(out, ".")) out <- paste0(out, ".")
    return(out)
  }
  if (step == "ild_lag") {
    vars <- a$vars
    if (length(vars) > 1) vars <- paste(vars, collapse = ", ")
    n <- .def(a$n, 1)
    mode <- .def(a$mode, "index")
    max_gap <- a$max_gap
    if (is.null(max_gap) || (is.numeric(max_gap) && !is.finite(max_gap))) max_gap <- "none"
    out <- paste0(
      "A ", mode, " lag of ", vars, " was computed using ild_lag() with lag ", n,
      " and max gap ", max_gap
    )
    if (identical(mode, "time_window") && (!is.null(a$window) || !is.null(a$resolution))) {
      out <- paste0(out, ", time window ", .def(a$window, "as specified"),
        if (!is.null(a$resolution)) paste0(" and resolution ", a$resolution) else "")
    }
    out <- paste0(out, .format_created(o))
    if (!endsWith(out, ".")) out <- paste0(out, ".")
    return(out)
  }
  if (step == "ild_align") {
    out <- paste0(
      "Variable ", .def(a$value_var, "value"), " was aligned from secondary data using ild_align() with window ", .def(a$window, "window")
    )
    out <- paste0(out, .format_created(o))
    if (!endsWith(out, ".")) out <- paste0(out, ".")
    return(out)
  }
  if (step == "ild_ipw_weights") {
    stab <- if (isTRUE(a$stabilize)) "stabilized" else "unstabilized"
    out <- paste0("Inverse probability weights (", stab, ") were added using ild_ipw_weights()")
    if (!is.null(a$trim) && length(a$trim) >= 2) {
      out <- paste0(out, " with trim quantiles ", paste(a$trim, collapse = " and "))
    }
    out <- paste0(out, .format_created(o))
    if (!endsWith(out, ".")) out <- paste0(out, ".")
    return(out)
  }
  if (step == "ild_lme") {
    engine <- .def(a$method, "lmer")
    out <- paste0(
      "A mixed-effects model was fit using ild_lme() (", engine, ") with formula ", .def(a$formula, "as specified")
    )
    if (isTRUE(a$ar1) && !is.null(a$correlation_class) && !is.na(a$correlation_class) && nchar(a$correlation_class) > 0) {
      out <- paste0(out, " and ", a$correlation_class, " correlation structure for residuals")
    } else {
      out <- paste0(out, " with AR1 disabled")
    }
    if (!is.null(o$n_obs) && !is.null(o$n_id)) {
      out <- paste0(out, " (n = ", o$n_obs, " observations, N = ", o$n_id, " persons)")
    }
    out <- paste0(out, ".")
    rs <- robust_se
    if (is.null(rs) && !is.null(context)) {
      rs <- attr(context, "ild_robust_se", exact = TRUE)
      if (is.null(rs) && is.list(context) && !is.null(context$fit)) {
        rs <- attr(context$fit, "ild_robust_se", exact = TRUE)
      }
    }
    if (length(rs) == 1 && is.character(rs) && nchar(rs) > 0) {
      out <- paste0(out, " Fixed effects were reported with cluster-robust standard errors (", rs, ").")
    }
    return(out)
  }
  if (step == "ild_brms") {
    pr <- a$prior_summary_text
    if (is.null(pr) || !nzchar(as.character(pr)[1])) pr <- "see prior_summary(fit)"
    pr_short <- substr(gsub("\n", " ", as.character(pr)), 1L, 500L)
    if (nchar(as.character(pr)) > 500) pr_short <- paste0(pr_short, " [...]")
    out <- paste0(
      "A Bayesian mixed-effects model was fit using ild_brms() with formula ", .def(a$formula, "as specified"),
      ". Sampling used ", .def(a$chains, "?"), " chains, ", .def(a$iter, "?"), " iterations per chain (warmup ",
      .def(a$warmup, "?"), "). ",
      "Sampler: adapt_delta = ", .def(a$adapt_delta, "default"),
      ", max_treedepth = ", .def(a$max_treedepth, "default"),
      if (!is.null(a$seed) && !is.na(a$seed)) paste0(", seed = ", a$seed) else "",
      ". Priors (brms::prior_summary): ", pr_short
    )
    if (!is.null(o$n_obs) && !is.null(o$n_id)) {
      out <- paste0(out, " (n = ", o$n_obs, ", N = ", o$n_id, " persons)")
    }
    out <- paste0(out, ".")
    if (!is.null(o$max_rhat) && is.finite(o$max_rhat)) {
      out <- paste0(out, " Max R-hat: ", round(o$max_rhat, 3), ".")
    }
    return(out)
  }
  if (step == "ild_kfas") {
    st <- .def(a$state_spec, "local_level")
    fam <- .def(a$observation_family, "gaussian")
    tu <- .def(a$time_units, "time units")
    fc <- .def(a$fit_context, "single_series")
    sm <- if (isTRUE(a$smoother)) {
      "smoothed and filtered state estimates (KFS smoothing)"
    } else {
      "filtered state estimates only (smoothing off)"
    }
    struct <- if (identical(st, "local_level")) {
      "local level (random walk plus observation noise)"
    } else {
      st
    }
    dots_note <- ""
    if (!is.null(a$fit_ssm_dots) && length(a$fit_ssm_dots) > 0L) {
      dn <- paste(names(a$fit_ssm_dots), a$fit_ssm_dots, sep = " = ", collapse = "; ")
      if (nzchar(dn)) dots_note <- paste0(" Additional fitSSM controls: ", dn, ".")
    }
    fh <- a$forecast_horizon
    fh_note <- if (!is.null(fh) && is.finite(fh) && fh >= 1L) {
      paste0(" Forecast horizon was set to ", fh, " step(s).")
    } else {
      ""
    }
    out <- paste0(
      "A Gaussian state-space model was fit using ild_kfas() with outcome ", .def(a$outcome, "outcome"),
      ", latent structure ", struct, " (", fam, " observations), and time indexing in ", tu, ". ",
      "KFS used ", sm, ". ",
      "Fit context: ", fc, " (not pooled across persons unless you explicitly stacked multiple series).",
      if (isTRUE(a$irregular_time)) {
        " Irregular observation spacing was acknowledged (irregular_time = TRUE)."
      } else {
        " Equal spacing was assumed for the time index after preparation."
      },
      fh_note,
      dots_note
    )
    if (!is.null(o$n_obs)) {
      out <- paste0(out, " (n = ", o$n_obs, " time points")
      if (!is.null(o$n_id)) out <- paste0(out, ", N = ", o$n_id, " person(s)")
      out <- paste0(out, ").")
    } else {
      out <- paste0(out, ".")
    }
    if (!is.null(o$logLik) && is.finite(o$logLik)) {
      out <- paste0(out, " Log-likelihood: ", round(o$logLik, 2), ".")
    }
    return(out)
  }
  if (step == "ild_ctsem") {
    out <- paste0(
      "A continuous-time latent-dynamics model was fit using ild_ctsem() with outcome ",
      .def(a$outcome, "outcome"),
      ", fit type ", .def(a$model_type, "ctsem default"),
      ", id column ", .def(a$id_col, ".ild_id"),
      ", and time column ", .def(a$time_col, ".ild_time_num"),
      "."
    )
    if (!is.null(o$n_obs) && !is.null(o$n_id)) {
      out <- paste0(out, " (n = ", o$n_obs, ", N = ", o$n_id, " persons).")
    }
    if (!is.null(o$converged) && !is.na(o$converged)) {
      out <- paste0(out, " Converged: ", o$converged, ".")
    }
    return(out)
  }
  if (step == "ild_diagnostics") {
    types <- a$type
    if (is.null(types) || length(types) == 0) {
      return("Residual diagnostics were examined.")
    }
    labels <- character(0)
    if ("residual_acf" %in% types) labels <- c(labels, "autocorrelation")
    if ("residual_time" %in% types) labels <- c(labels, "residuals vs time and fitted")
    if ("qq" %in% types) labels <- c(labels, "Q-Q")
    if (length(labels) == 0) labels <- "requested diagnostics"
    return(paste0("Residual diagnostics were examined using ", paste(labels, collapse = " and "), " diagnostics."))
  }
  if (step == "ild_tvem") {
    return(paste0(
      "A time-varying effects model was fit using ild_tvem() for outcome ", .def(a$outcome, "outcome"),
      " and predictor ", .def(a$predictor, "predictor"), "."
    ))
  }
  if (step == "ild_power") {
    return(paste0(
      "Simulation-based power analysis was run with n_sim=", .def(a$n_sim, "?"), ", effect_size=", .def(a$effect_size, "?"), ", test_term=", .def(a$test_term, "?"), "."
    ))
  }
  if (step == "ild_missing_model") {
    preds <- a$predictors
    if (length(preds) > 1) preds <- paste(preds, collapse = ", ")
    return(paste0(
      "A missingness model was fit for outcome ", .def(a$outcome, "outcome"),
      " and predictors ", preds, "."
    ))
  }
  if (step == "ild_crosslag") {
    out <- paste0(
      "A cross-lag model was fit using ild_crosslag() with outcome ", .def(a$outcome, "outcome"),
      ", predictor ", .def(a$predictor, "predictor"), ", lag ", .def(a$lag, 1)
    )
    if (isTRUE(a$ar1)) out <- paste0(out, " and AR1 correlation") else out <- paste0(out, " and AR1 disabled")
    return(paste0(out, "."))
  }
  if (step == "ild_ipw_refit") {
    wcol <- .def(a$weights, ".ipw")
    return(paste0(
      "Weighted estimation was used: the model was refit with inverse-probability weights using ild_ipw_refit() (weight column: ", wcol, ")."
    ))
  }
  paste0("Step ", step, " was applied.")
}

# Default value helper (internal)
.def <- function(x, y) if (is.null(x)) y else x

#' Generate methods-style narrative from provenance
#'
#' Takes an ILD data object, a model fit, or a diagnostics object and produces
#' a concise methods-style paragraph based on the recorded provenance (data
#' preparation, centering, lagging, modeling, etc.).
#'
#' @param x An ILD object (see [is_ild()]), a model from [ild_lme()] or [ild_tvem()],
#'   a diagnostics object from [ild_diagnostics()], or another object with
#'   [ild_provenance()] (e.g. [ild_power()] result, [ild_missing_model()] result).
#' @param robust_se Optional. If you reported fixed effects with cluster-robust SEs
#'   via [tidy_ild_model()] with \code{se = "robust"}, pass the type here (e.g. \code{"CR2"})
#'   so the methods text can mention it.
#' @param bundle Optional. An object from [ild_diagnose()] (class
#'   \code{\link{ild_diagnostics_bundle}}). When provided and
#'   \code{nrow(bundle$guardrails) > 0}, a short methodological-cautions sentence
#'   is appended after the provenance paragraph.
#' @param ... Unused.
#' @return A single character string (one or more sentences) suitable for a
#'   methods section. Use \code{cat()} or \code{print()} to display.
#' @details
#' After fitting, call \code{diag <- ild_diagnose(fit)} and pass
#' \code{bundle = diag} to surface triggered guardrails in the methods narrative
#' without duplicating [ild_report()]'s structured \code{diagnostics_summary}.
#' @export
#' @examples
#' set.seed(1)
#' d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 1)
#' x <- ild_prepare(d, id = "id", time = "time")
#' x <- ild_center(x, y)
#' ild_methods(x)
ild_methods <- function(x, robust_se = NULL, bundle = NULL, ...) {
  prov <- ild_provenance(x)
  if (is.null(prov)) {
    base <- "No provenance recorded for this object."
  } else {
    steps <- ild_flatten_provenance(prov)
    if (length(steps) == 0) {
      base <- "Provenance has no steps recorded."
    } else {
      sentences <- vapply(steps, function(s) ild_step_to_sentence(s, context = x, robust_se = robust_se), character(1L))
      base <- paste(sentences, collapse = " ")
    }
  }
  if (inherits(bundle, "ild_diagnostics_bundle") && nrow(bundle$guardrails) > 0L) {
    gs <- ild_guardrails_summary(bundle)
    if (nzchar(gs$narrative)) {
      base <- paste0(
        base,
        " Methodological cautions (tidyILD guardrails): ",
        gs$narrative
      )
    }
  }
  base
}

#' Assemble a light report from a model fit
#'
#' Builds a structured list with methods narrative (from [ild_methods()]),
#' fixed-effects table (from [tidy_ild_model()]), a short diagnostics summary,
#' and the raw provenance. Optionally exports provenance to a file.
#'
#' @param fit A fitted model from [ild_lme()] (or a list with a \code{fit} component, e.g. from [ild_crosslag()]).
#' @param export_provenance_path Optional. If provided, [ild_export_provenance()] is called to write provenance to this path; the path is included in the returned list.
#' @param robust_se Optional. Passed to [ild_methods()] when building the methods text (e.g. \code{"CR2"} if you used [tidy_ild_model()] with \code{se = "robust"}).
#' @param ... Unused.
#' @return A list with a stable schema: \code{meta} (list with \code{n_obs}, \code{n_id}, \code{engine} when available), \code{methods}, \code{model_table}, \code{diagnostics_summary} (includes \code{guardrails_narrative} and \code{guardrails} when [ild_diagnose()] succeeds), \code{provenance}, \code{provenance_export_path} (character or NULL). When guardrails are triggered, \code{methods_with_guardrails} repeats the methods paragraph with guardrails appended (same pattern as \code{\link{ild_methods}(fit, bundle = d)}).
#' @export
#' @examples
#' set.seed(1)
#' x <- ild_prepare(ild_simulate(n_id = 5, n_obs_per = 6, seed = 1), id = "id", time = "time")
#' fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
#' r <- ild_report(fit)
#' r$methods
#' r$model_table
ild_report <- function(fit, export_provenance_path = NULL, robust_se = NULL, ...) {
  if (is.list(fit) && !is.null(fit$fit)) fit <- fit$fit
  meta <- list(n_obs = NULL, n_id = NULL, engine = NULL)
  ild_data <- attr(fit, "ild_data", exact = TRUE)
  if (!is.null(ild_data) && nrow(ild_data) > 0) {
    meta$n_obs <- nrow(ild_data)
    id_col <- ild_meta(ild_data)$ild_id
    if (!is.null(id_col) && id_col %in% names(ild_data)) {
      meta$n_id <- length(unique(ild_data[[id_col]]))
    }
  }
  meta$engine <- if (inherits(fit, "ild_fit_kfas")) {
    "KFAS"
  } else if (inherits(fit, "ild_fit_ctsem")) {
    "ctsem"
  } else if (inherits(fit, "lme")) {
    "lme"
  } else if (inherits(fit, "lmerMod")) {
    "lmer"
  } else if (inherits(fit, "brmsfit")) {
    "brms"
  } else {
    NULL
  }
  diag_bundle <- tryCatch(
    ild_diagnose(
      fit,
      type = if (inherits(fit, "brmsfit")) {
        "all"
      } else if (inherits(fit, "ild_fit_kfas")) {
        NULL
      } else if (inherits(fit, "ild_fit_ctsem")) {
        NULL
      } else {
        c("residual_acf", "qq")
      }
    ),
    error = function(e) NULL
  )
  methods_text <- ild_methods(fit, robust_se = robust_se)
  methods_with_guardrails <- if (!is.null(diag_bundle)) {
    ild_methods(fit, robust_se = robust_se, bundle = diag_bundle)
  } else {
    NULL
  }
  model_table <- tryCatch(
    if (inherits(fit, "brmsfit") || inherits(fit, "ild_fit_kfas")) {
      ild_tidy(fit)
    } else {
      tidy_ild_model(fit, object = FALSE)
    },
    error = function(e) NULL
  )
  diagnostics_summary <- if (is.null(diag_bundle)) {
    list(meta = NULL, summary_text = "Diagnostics could not be computed.")
  } else {
    gs <- ild_guardrails_summary(diag_bundle$guardrails)
    out <- list(
      meta = diag_bundle$meta,
      summary_text = paste(diag_bundle$summary_text, collapse = " "),
      guardrails_narrative = gs$narrative,
      guardrails = list(
        n = gs$n,
        max_severity = gs$max_severity,
        rule_ids = gs$rule_ids
      )
    )
    if (gs$n == 0L) {
      out$guardrails_narrative <- ""
      out$guardrails <- list(n = 0L, max_severity = NA_character_, rule_ids = character(0))
    }
    out
  }
  provenance <- ild_provenance(fit)
  provenance_export_path <- NULL
  if (!is.null(export_provenance_path) && length(export_provenance_path) == 1L && is.character(export_provenance_path) && nchar(export_provenance_path) > 0 && !is.null(provenance)) {
    ild_export_provenance(fit, export_provenance_path)
    provenance_export_path <- export_provenance_path
  }
  out <- list(
    meta = meta,
    methods = methods_text,
    model_table = model_table,
    diagnostics_summary = diagnostics_summary,
    provenance = provenance,
    provenance_export_path = provenance_export_path
  )
  if (!is.null(methods_with_guardrails) && !identical(methods_with_guardrails, methods_text)) {
    out$methods_with_guardrails <- methods_with_guardrails
  }
  out
}

test_that("ild_methods returns non-empty string for data object", {
  set.seed(1)
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  out <- ild_methods(x)
  expect_type(out, "character")
  expect_true(nchar(out) > 0)
  expect_true(grepl("ild_prepare", out))
  expect_true(grepl("ild_center", out))
})

test_that("ild_methods returns non-empty string for model object", {
  set.seed(1)
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  out <- ild_methods(fit)
  expect_type(out, "character")
  expect_true(nchar(out) > 0)
  expect_true(grepl("ild_lme", out))
})

test_that("ild_methods returns message when no provenance", {
  out <- ild_methods(data.frame(a = 1))
  expect_equal(out, "No provenance recorded for this object.")
})

test_that("ild_compare_pipelines returns structure and detects match", {
  set.seed(1)
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 1)
  x1 <- ild_prepare(d, id = "id", time = "time")
  x1 <- ild_center(x1, y)
  x2 <- ild_prepare(d, id = "id", time = "time")
  x2 <- ild_center(x2, y)
  cmp <- ild_compare_pipelines(x1, x2)
  expect_s3_class(cmp, "ild_compare_pipelines")
  expect_true(is.character(cmp$summary))
  expect_true(length(cmp$only_in_first) >= 0)
  expect_true(length(cmp$only_in_second) >= 0)
})

test_that("ild_compare_pipelines detects different gap_threshold", {
  set.seed(1)
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 1)
  x1 <- ild_prepare(d, id = "id", time = "time", gap_threshold = 10)
  x2 <- ild_prepare(d, id = "id", time = "time", gap_threshold = 20)
  cmp <- ild_compare_pipelines(x1, x2)
  expect_true(length(cmp$differing) > 0 || any(grepl("gap", cmp$summary, ignore.case = TRUE)))
})

test_that("ild_export_provenance writes JSON file when jsonlite available", {
  skip_if_not_installed("jsonlite")
  set.seed(1)
  d <- ild_simulate(n_id = 3, n_obs_per = 4, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  path <- tempfile(fileext = ".json")
  ild_export_provenance(x, path)
  expect_true(file.exists(path))
  txt <- readLines(path)
  expect_true(any(grepl("ild_prepare", txt)))
  expect_true(any(grepl("object_type", txt)))
})

test_that("ild_export_provenance errors when no provenance", {
  expect_error(ild_export_provenance(data.frame(a = 1), tempfile(fileext = ".json")), "no provenance")
})

test_that("ild_export_provenance errors when path missing", {
  set.seed(1)
  x <- ild_prepare(ild_simulate(n_id = 2, n_obs_per = 3, seed = 1), id = "id", time = "time")
  expect_error(ild_export_provenance(x), "path")
})

#' Fit a linear mixed-effects model to ILD
#'
#' When `ar1 = FALSE`, fits with [lme4::lmer()] (no residual correlation).
#' When `ar1 = TRUE`, fits with [nlme::lme()] using a residual correlation
#' structure: CAR1 (continuous-time) by default for irregular spacing,
#' or AR1 when spacing is regular-ish. Use [ild_spacing_class()] to inform
#' the choice; override with `correlation_class`.
#'
#' @param formula Fixed-effects formula. For `ar1 = TRUE`, must be fixed-only
#'   (e.g. `y ~ x`); random structure is set to `~ 1 | .ild_id` internally.
#'   For `ar1 = FALSE`, formula may include random effects (e.g. `y ~ x + (1|id)`).
#' @param data An ILD object (see [is_ild()]).
#' @param ar1 Logical. If `TRUE`, fit with nlme and residual AR1/CAR1
#'   correlation; if `FALSE`, fit with lme4 (no residual correlation).
#' @param correlation_class Character. `"auto"` (default) uses [ild_spacing_class()]
#'   to choose CAR1 (irregular-ish) or AR1 (regular-ish). Use `"CAR1"` or
#'   `"AR1"` to override.
#' @param random For `ar1 = TRUE`, the random effects formula (default
#'   `~ 1 | .ild_id`). Must use `.ild_id` as grouping for correlation to match.
#' @param warn_no_ar1 If `TRUE` (default), warn when `ar1 = FALSE` that
#'   temporal autocorrelation is not modeled.
#' @param warn_uncentered If `TRUE` (default), warn when a predictor in the
#'   formula varies both within and between persons but is not decomposed
#'   (no \code{_wp}/\code{_bp}); suggests using [ild_center()].
#' @param ... Passed to [lme4::lmer()] or [nlme::lme()].
#' @return A fitted model object (class `lmerMod` or `lme`) with attribute
#'   `ild_data` (the ILD data) and `ild_ar1` (logical). When `ar1 = TRUE`,
#'   the returned object has class `ild_lme` prepended and attribute
#'   `ild_random_resolved` (the formula actually passed to nlme, e.g. `~ 1 | M2ID`).
#'   See [ild_diagnostics()] and [ild_plot()].
#' @examples
#' # lme4 path: formula includes random effects
#' set.seed(1)
#' dat <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 1)
#' dat <- ild_prepare(dat, id = "id", time = "time")
#' dat <- ild_center(dat, y)
#' fit_lmer <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = dat,
#'                     ar1 = FALSE, warn_no_ar1 = FALSE)
#' # nlme path (may not converge on all platforms; see ?nlme::lme)
#' \dontrun{
#' fit_lme <- ild_lme(y ~ y_bp + y_wp, data = dat,
#'                    random = ~ 1 | id, ar1 = TRUE)
#' }
#'
#' @importFrom lme4 lmer
#' @importFrom nlme corAR1 corCAR1 lme
#' @export
ild_lme <- function(formula,
                    data,
                    ar1 = FALSE,
                    correlation_class = c("auto", "AR1", "CAR1"),
                    random = ~ 1 | .ild_id,
                    warn_no_ar1 = TRUE,
                    warn_uncentered = TRUE,
                    ...) {
  validate_ild(data)
  if (warn_uncentered) {
    uncentered <- ild_detect_uncentered_predictors(data, formula)
    if (length(uncentered) > 0) {
      warning("Predictor(s) '", paste(uncentered, collapse = "', '"),
              "' vary both within and between persons. Consider using ild_center(",
              paste(uncentered, collapse = ", "), ") to separate within- and ",
              "between-person effects and avoid conflation bias.", call. = FALSE)
    }
  }
  correlation_class <- match.arg(correlation_class)
  if (ar1) {
    if (ild_formula_has_random(formula)) {
      stop("When ar1 = TRUE, use a fixed-effects-only formula and pass random ",
           "effects via the random argument. Formula must not contain (1|id) or similar. ",
           "In model formulas, | is interpreted as an lme4-style random-effects operator.",
           call. = FALSE)
    }
    if (!requireNamespace("nlme", quietly = TRUE)) {
      stop("Package 'nlme' is required for ar1 = TRUE.", call. = FALSE)
    }
    meta <- ild_meta(data)
    id_name <- meta$ild_id
    cor_class <- if (correlation_class == "auto") {
      if (ild_spacing_class(data) == "regular-ish") "AR1" else "CAR1"
    } else {
      correlation_class
    }
    cor_form <- if (cor_class == "CAR1") {
      nlme::corCAR1(form = stats::as.formula(paste0("~ .ild_time_num | ", id_name)))
    } else {
      nlme::corAR1(form = stats::as.formula(paste0("~ .ild_seq | ", id_name)))
    }
    random_form <- if (identical(random, ~ 1 | .ild_id)) {
      stats::as.formula(paste0("~ 1 | ", id_name))
    } else {
      random
    }
    fit <- nlme::lme(
      fixed = formula,
      data = data,
      random = random_form,
      correlation = cor_form,
      ...
    )
    attr(fit, "ild_data") <- data  # both engines set ild_data for augment_ild_model/diagnostics
    attr(fit, "ild_ar1") <- TRUE
    attr(fit, "ild_correlation_class") <- cor_class
    attr(fit, "ild_random_resolved") <- random_form
    attr(fit, "ild_provenance") <- ild_new_analysis_provenance(data, "ild_lme", list(
      formula = deparse(formula),
      random = deparse(random_form),
      ar1 = TRUE,
      correlation_class = cor_class,
      method = "lme",
      backend = "nlme"
    ), list(
      n_obs = nrow(data),
      n_id = length(unique(data[[".ild_id"]])),
      fit_engine = "lme",
      backend_version = ild_backend_pkg_version("nlme")
    ))
    class(fit) <- c(class(fit), "ild_lme")
    return(fit)
  }
  if (warn_no_ar1) {
    message("Temporal autocorrelation is not modeled (ar1 = FALSE). Consider ar1 = TRUE for ILD.")
  }
  if (!requireNamespace("lme4", quietly = TRUE)) {
    stop("Package 'lme4' is required for ar1 = FALSE.", call. = FALSE)
  }
  fit <- lme4::lmer(formula, data = data, ...)
  attr(fit, "ild_data") <- data  # both engines set ild_data for augment_ild_model/diagnostics
  attr(fit, "ild_ar1") <- FALSE
  attr(fit, "ild_provenance") <- ild_new_analysis_provenance(data, "ild_lme", list(
    formula = deparse(formula),
    ar1 = FALSE,
    correlation_class = NA_character_,
    method = "lmer",
    backend = "lme4"
  ), list(
    n_obs = nrow(data),
    n_id = length(unique(data[[".ild_id"]])),
    fit_engine = "lmer",
    backend_version = ild_backend_pkg_version("lme4")
  ))
  # Do not add ild_lme to class for S4 lmerMod (breaks residuals/fitted dispatch)
  fit
}

#' Detect if formula contains random effects (e.g. (1|id))
#' @param f Formula or call to recurse.
#' @return Logical.
#' @noRd
ild_formula_has_random <- function(f) {
  if (is.call(f)) {
    if (identical(f[[1]], as.name("|"))) return(TRUE)
    for (i in seq_along(f)) {
      if (i == 1) next
      if (ild_formula_has_random(f[[i]])) return(TRUE)
    }
  }
  FALSE
}

#' Predictors that vary both WP and BP but are not _wp/_bp (uncentered).
#' @param data ILD object
#' @param formula Model formula (fixed or mixed)
#' @return Character vector of variable names to warn about (may be empty).
#' @noRd
ild_detect_uncentered_predictors <- function(data, formula) {
  all_v <- all.vars(formula)
  if (length(all_v) <= 1) return(character())
  preds <- all_v[-1]
  nms <- names(data)
  id_col <- ".ild_id"
  out <- character()
  for (v in preds) {
    if (!v %in% nms) next
    if (!is.numeric(data[[v]])) next
    if (grepl("_wp$|_bp$", v)) next
    means_i <- tapply(data[[v]], data[[id_col]], mean, na.rm = TRUE)
    bp_var <- stats::var(as.vector(means_i))
    if (is.na(bp_var) || bp_var < 1e-10) next
    wp_var_by_id <- tapply(data[[v]], data[[id_col]], function(z) stats::var(z, na.rm = TRUE))
    wp_var_by_id[is.na(wp_var_by_id)] <- 0
    wp_var <- mean(wp_var_by_id, na.rm = TRUE)
    if (is.na(wp_var) || wp_var < 1e-10) next
    out <- c(out, v)
  }
  unique(out)
}

test_that("Phase 1 pipeline: prepare -> summary -> center -> lag -> missing_pattern", {
  set.seed(42)
  d <- data.frame(
    id = rep(1:3, each = 5),
    time = rep(as.POSIXct(0:4 * 3600, origin = "1970-01-01"), 3),
    mood = rnorm(15),
    stress = rnorm(15)
  )
  # Introduce one missing
  d$mood[3] <- NA
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)
  expect_true(is_ild(x))
  s <- ild_summary(x)
  expect_equal(s$n_units, 3)
  expect_equal(s$n_obs, 15)
  x <- ild_center(x, mood, stress)
  expect_true("mood_bp" %in% names(x))
  expect_true("mood_wp" %in% names(x))
  x <- ild_lag(x, mood, mode = "gap_aware", max_gap = 7200)
  expect_true("mood_lag1" %in% names(x))
  mp <- ild_missing_pattern(x, vars = "mood")
  expect_equal(mp$overall$mood$n_na, 1)
  expect_equal(ild_spacing_class(x), "regular-ish")
})

test_that("Phase 2 pipeline: prepare -> lme -> diagnostics -> plot", {
  d <- ild_simulate(n_id = 4, n_obs_per = 6, seed = 10)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  expect_true(inherits(fit, "lmerMod") && !is.null(attr(fit, "ild_data")))
  diag <- ild_diagnostics(fit, data = x)
  expect_s3_class(diag, "ild_diagnostics")
  p1 <- ild_plot(x, type = "trajectory", var = "y")
  p2 <- ild_plot(fit, type = "fitted")
  expect_s3_class(p1, "ggplot")
  expect_s3_class(p2, "ggplot")
})

test_that("ild_lme ar1 = FALSE fits with lmer", {
  d <- ild_simulate(n_id = 4, n_obs_per = 8, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  expect_true(inherits(fit, "lmerMod"))
  expect_false(attr(fit, "ild_ar1"))
  expect_true(!is.null(attr(fit, "ild_data")))
  expect_equal(length(stats::residuals(fit)), nrow(x))
})

test_that("ild_lme ar1 = TRUE fits with nlme and CAR1", {
  d <- ild_simulate(n_id = 3, n_obs_per = 6, irregular = TRUE, seed = 2)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1, data = x, ar1 = TRUE, correlation_class = "CAR1")
  expect_s3_class(fit, "lme")
  expect_true(inherits(fit, "ild_lme"))
  expect_true(attr(fit, "ild_ar1"))
  expect_equal(attr(fit, "ild_correlation_class"), "CAR1")
})

test_that("ild_lme correlation_class override", {
  d <- ild_simulate(n_id = 3, n_obs_per = 6, seed = 3)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1, data = x, ar1 = TRUE, correlation_class = "AR1")
  expect_equal(attr(fit, "ild_correlation_class"), "AR1")
})

test_that("ild_lme engine selection: ar1 FALSE returns lmerMod, ar1 TRUE returns lme", {
  d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 10)
  x <- ild_prepare(d, id = "id", time = "time")
  fit_lmer <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  fit_lme <- ild_lme(y ~ 1, data = x, ar1 = TRUE)
  expect_true(inherits(fit_lmer, "lmerMod"))
  expect_s3_class(fit_lme, "lme")
})

test_that("ild_lme default random mapping works when id column is not .ild_id", {
  dat <- data.frame(
    M2ID = rep(1:3, each = 4),
    time = rep(as.POSIXct(0:3, origin = "1970-01-01"), 3),
    y = rnorm(12)
  )
  x <- ild_prepare(dat, id = "M2ID", time = "time")
  fit <- ild_lme(y ~ 1, data = x, ar1 = TRUE)
  expect_equal(ild_meta(attr(fit, "ild_data"))$ild_id, "M2ID")
  expect_s3_class(fit, "lme")
  resolved <- attr(fit, "ild_random_resolved")
  expect_false(is.null(resolved))
  dep <- paste(deparse(resolved), collapse = " ")
  expect_true(grepl("M2ID", dep, fixed = TRUE))
  expect_false(grepl(".ild_id", dep, fixed = TRUE))
})

test_that("ild_lme ar1 TRUE rejects lme4 random syntax in formula", {
  d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 20)
  x <- ild_prepare(d, id = "id", time = "time")
  expect_error(
    ild_lme(y ~ 1 + (1 | id), data = x, ar1 = TRUE),
    "fixed-effects-only"
  )
})

test_that("ild_lme correlation default uses CAR1 for irregular data", {
  d <- ild_simulate(n_id = 3, n_obs_per = 6, irregular = TRUE, seed = 30)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1, data = x, ar1 = TRUE)
  expect_equal(attr(fit, "ild_correlation_class"), "CAR1")
  expect_true(inherits(fit$modelStruct$corStruct, "corCAR1"))
})

test_that("ild_lme warns when predictor is uncentered (WP and BP variance)", {
  d <- ild_simulate(n_id = 4, n_obs_per = 6, seed = 1)
  d$stress <- d$y + rnorm(nrow(d), 0, 0.5)
  x <- ild_prepare(d, id = "id", time = "time")
  expect_warning(
    ild_lme(y ~ stress + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE),
    "ild_center"
  )
})

test_that("ild_lme does not warn about centering when predictors are _wp/_bp", {
  d <- ild_simulate(n_id = 4, n_obs_per = 6, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- suppressWarnings(
    ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  )
  expect_true(inherits(fit, "lmerMod"))
})

test_that("ild_lme warn_uncentered = FALSE suppresses uncentered warning", {
  d <- ild_simulate(n_id = 4, n_obs_per = 6, seed = 1)
  d$stress <- d$y + rnorm(nrow(d), 0, 0.5)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ stress + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE,
                 warn_uncentered = FALSE)
  expect_true(inherits(fit, "lmerMod"))
})

test_that("ild_diagnose lmer bundle has denser data, design, fit, predictive sections", {
  d <- ild_simulate(n_id = 6, n_obs_per = 5, seed = 400)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x)
  expect_true(!is.null(b$data$cohort))
  expect_true(!is.null(b$data$obs_per_id))
  expect_true(!is.null(b$data$outcome_summaries))
  expect_true(!is.null(b$design$flags))
  expect_true(!is.null(b$design$time_coverage))
  expect_equal(b$design$occasion_imbalance, b$data$obs_per_id)
  expect_true(!is.null(b$fit$optimizer$reml))
  expect_true(!is.null(b$fit$random_effects_structure$theta))
  expect_true(!is.null(b$fit$residual_correlation))
  expect_true(!is.null(b$fit$rank$fixed_effects_X))
  expect_true(!is.null(b$predictive$obs_metrics$n))
  expect_true(!is.null(b$predictive$n))
  expect_true(!is.null(b$predictive$mean_residual))
  expect_true(!is.null(b$predictive$max_abs_error))
  expect_true(!is.null(b$missingness$summary) || !is.null(b$missingness$note))
})

test_that("ild_diagnose causal_detail adds weight_detail when .ipw present", {
  d <- ild_simulate(n_id = 5, n_obs_per = 4, seed = 401)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  x$.ipw <- rep(1, nrow(x))
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x, causal_detail = TRUE)
  expect_true(!is.null(b$causal$weight_detail))
  expect_true(!is.null(b$causal$weight_detail$quantiles))
})

test_that("ild_acf on data returns acf tibble", {
  d <- ild_simulate(n_id = 5, n_obs_per = 10, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_acf(x, "y")
  expect_named(out, "acf")
  expect_s3_class(out$acf, "tbl_df")
  expect_true(all(c("lag", "acf") %in% names(out$acf)))
  expect_true(nrow(out$acf) >= 1)
})

test_that("ild_acf on fit returns residual ACF", {
  d <- ild_simulate(n_id = 4, n_obs_per = 8, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  out <- ild_acf(fit)
  expect_named(out, "acf")
  expect_true(all(c("lag", "acf") %in% names(out$acf)))
})

test_that("ild_acf with by_id adds acf_by_id", {
  d <- ild_simulate(n_id = 3, n_obs_per = 6, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_acf(x, "y", by_id = TRUE)
  expect_true("acf_by_id" %in% names(out))
  expect_true(is.list(out$acf_by_id))
  expect_equal(length(out$acf_by_id), 3)
})

test_that("ild_acf on fit without ild_data errors", {
  fit <- lme4::lmer(y ~ 1 + (1 | id), data = data.frame(id = rep(1:2, each = 5), y = rnorm(10)))
  expect_error(ild_acf(fit), "refit using ild_lme")
})

test_that("ild_robust_se requires ild_data", {
  dat <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 1)
  dat <- ild_prepare(dat, id = "id", time = "time")
  dat <- ild_center(dat, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = dat, ar1 = FALSE, warn_no_ar1 = FALSE)
  attr(fit, "ild_data") <- NULL
  expect_error(ild_robust_se(fit), "refit using ild_lme")
})

test_that("ild_robust_se returns list with vcov, type, cluster_name, engine", {
  skip_if_not_installed("clubSandwich")
  set.seed(1)
  dat <- ild_simulate(n_id = 8, n_obs_per = 6, seed = 1)
  dat <- ild_prepare(dat, id = "id", time = "time")
  dat <- ild_center(dat, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = dat, ar1 = FALSE, warn_no_ar1 = FALSE)
  rv <- ild_robust_se(fit, type = "CR2")
  expect_type(rv, "list")
  expect_named(rv, c("vcov", "type", "cluster_name", "engine"))
  expect_equal(rv$type, "CR2")
  expect_equal(rv$engine, "lmer")
  expect_true(is.matrix(rv$vcov))
  expect_equal(nrow(rv$vcov), 3L)  # (Intercept), y_bp, y_wp
  expect_equal(ncol(rv$vcov), 3L)
})

test_that("tidy_ild_model with se = robust returns numeric SEs and CIs", {
  skip_if_not_installed("clubSandwich")
  set.seed(1)
  dat <- ild_simulate(n_id = 8, n_obs_per = 6, seed = 1)
  dat <- ild_prepare(dat, id = "id", time = "time")
  dat <- ild_center(dat, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = dat, ar1 = FALSE, warn_no_ar1 = FALSE)
  tbl <- tidy_ild_model(fit, se = "robust", robust_type = "CR2")
  expect_s3_class(tbl, "tbl_df")
  expect_true(all(ild_tidy_schema()$required %in% names(tbl)))
  expect_true(all(is.finite(tbl$std_error)))
  expect_true(all(is.finite(tbl$conf_low)))
  expect_true(all(is.finite(tbl$conf_high)))
  expect_true(all(is.finite(tbl$p_value)))
})

test_that("tidy_ild_model se = model unchanged when se = model", {
  set.seed(1)
  dat <- ild_simulate(n_id = 6, n_obs_per = 5, seed = 1)
  dat <- ild_prepare(dat, id = "id", time = "time")
  dat <- ild_center(dat, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = dat, ar1 = FALSE, warn_no_ar1 = FALSE)
  t_model <- tidy_ild_model(fit, se = "model")
  t_default <- tidy_ild_model(fit)
  expect_equal(t_model, t_default)
})

# Provenance helpers: attr(x, "tidyILD")$provenance (data) and attr(obj, "ild_provenance") (analysis)
# Uses .TIDYILD_ATTR from ild-class.R (Collate: ild_provenance after ild-class)

# Attribute name for analysis provenance on model/diagnostics/power/etc. objects
.ILD_PROVENANCE_ATTR <- "ild_provenance"

#' Append a step to the ILD provenance (internal)
#'
#' Ensures \code{attr(x, "tidyILD")$provenance} exists, then appends one step
#' record. If the tidyILD bundle is missing, builds a minimal one from
#' \code{ild_*} attributes. Modifies \code{x} in place by setting the attribute.
#'
#' @param x An ILD object (see [is_ild()]).
#' @param step Character; step name (e.g. \code{"ild_prepare"}, \code{"ild_center"}).
#' @param args Named list of user-facing arguments (serializable).
#' @param outputs Optional named list of step outputs (e.g. \code{created}, \code{n_id}).
#' @return \code{x} (invisibly), with \code{attr(x, "tidyILD")$provenance$steps} updated.
#' @noRd
ild_add_step <- function(x, step, args, outputs = NULL) {
  bundle <- attr(x, .TIDYILD_ATTR, exact = TRUE)
  if (is.null(bundle)) {
    bundle <- list(
      id_col = attr(x, "ild_id", exact = TRUE),
      time_col = attr(x, "ild_time", exact = TRUE),
      tz = "UTC",
      gap_threshold = attr(x, "ild_gap_threshold", exact = TRUE),
      created = NA,
      spacing = attr(x, "ild_spacing", exact = TRUE)
    )
  }
  if (is.null(bundle$provenance)) {
    pv <- tryCatch(
      as.character(utils::packageVersion("tidyILD")),
      error = function(e) "0.0.0"
    )
    bundle$provenance <- list(version = pv, schema_version = "1", object_type = "ild_data", steps = list())
  }
  step_id <- as.character(length(bundle$provenance$steps) + 1L)
  step_record <- list(
    step_id = step_id,
    step = step,
    timestamp = format(Sys.time(), "%Y-%m-%dT%H:%M:%S"),
    args = args,
    outputs = outputs
  )
  bundle$provenance$steps <- c(bundle$provenance$steps, list(step_record))
  attr(x, .TIDYILD_ATTR) <- bundle
  invisible(x)
}

#' Flatten provenance into an ordered list of steps (internal)
#' Data steps first (from source_data_provenance or prov$steps), then analysis_steps.
#' @param prov Return value of ild_provenance(x).
#' @return List of step records (each with step, args, outputs; step_id if present).
#' @noRd
ild_flatten_provenance <- function(prov) {
  if (is.null(prov)) return(list())
  out <- list()
  if (!is.null(prov$source_data_provenance) && length(prov$source_data_provenance$steps) > 0) {
    out <- c(out, prov$source_data_provenance$steps)
  }
  if (!is.null(prov$steps) && length(prov$steps) > 0) {
    out <- c(out, prov$steps)
  }
  if (!is.null(prov$analysis_steps) && length(prov$analysis_steps) > 0) {
    out <- c(out, prov$analysis_steps)
  }
  out
}

#' Get raw provenance object (internal)
#' @param x An ILD object.
#' @return \code{attr(x, "tidyILD")$provenance} or \code{NULL}.
#' @noRd
ild_get_history <- function(x) {
  bundle <- attr(x, .TIDYILD_ATTR, exact = TRUE)
  if (is.null(bundle)) return(NULL)
  bundle$provenance
}

#' Map analysis step name to object_type (internal)
#' @noRd
ild_provenance_object_type <- function(step) {
  switch(
    step,
    ild_lme = "ild_model",
    ild_brms = "ild_model",
    ild_tvem = "ild_model",
    ild_crosslag = "ild_model",
    ild_panel_lag_prepare = "ild_data",
    ild_ipw_refit = "ild_model",
    ild_msm_bootstrap = "ild_msm_bootstrap",
    ild_msm_fit = "ild_msm_fit",
    ild_msm_recovery = "ild_msm_recovery",
    ild_ctsem = "ild_model",
    ild_kfas = "ild_model",
    ild_diagnostics = "ild_diagnostics",
    ild_power = "ild_power",
    ild_missing_model = "ild_missingness",
    "ild_other"
  )
}

#' Build analysis provenance for a non-data object (internal)
#'
#' Used by ild_lme(), ild_diagnostics(), ild_tvem(), ild_power(), ild_missing_model()
#' to attach a provenance record to the returned fit/list. Structure:
#' \code{list(version, object_type, source_data_provenance, analysis_steps)}.
#'
#' @param source ILD data (for data provenance snapshot), or an object with
#'   \code{attr(., "ild_provenance")} (e.g. fit from ild_lme), or \code{NULL}.
#' @param step Character; analysis step name (e.g. \code{"ild_lme"}, \code{"ild_diagnostics"}).
#' @param args Named list of serializable arguments.
#' @param outputs Optional named list of step outputs.
#' @return List with \code{version}, \code{object_type}, \code{source_data_provenance}, \code{analysis_steps}.
#' @noRd
ild_new_analysis_provenance <- function(source = NULL, step, args, outputs = NULL) {
  pv <- tryCatch(
    as.character(utils::packageVersion("tidyILD")),
    error = function(e) "0.0.0"
  )
  object_type <- ild_provenance_object_type(step)
  step_id <- "1"
  if (!is.null(source)) {
    existing <- attr(source, .ILD_PROVENANCE_ATTR, exact = TRUE)
    if (!is.null(existing) && is.list(existing) && length(existing$analysis_steps) > 0) {
      step_id <- as.character(length(existing$analysis_steps) + 1L)
    }
  }
  step_record <- list(
    step_id = step_id,
    step = step,
    timestamp = format(Sys.time(), "%Y-%m-%dT%H:%M:%S"),
    args = args,
    outputs = outputs
  )
  schema_version <- "1"
  if (is.null(source)) {
    return(list(
      version = pv,
      schema_version = schema_version,
      object_type = object_type,
      source_data_provenance = NULL,
      analysis_steps = list(step_record)
    ))
  }
  existing <- attr(source, .ILD_PROVENANCE_ATTR, exact = TRUE)
  if (!is.null(existing) && is.list(existing)) {
    return(list(
      version = pv,
      schema_version = schema_version,
      object_type = object_type,
      source_data_provenance = existing$source_data_provenance,
      analysis_steps = c(existing$analysis_steps, list(step_record))
    ))
  }
  data_prov <- ild_get_history(source)
  list(
    version = pv,
    schema_version = schema_version,
    object_type = object_type,
    source_data_provenance = data_prov,
    analysis_steps = list(step_record)
  )
}

#' Return the raw provenance object
#'
#' For ILD data: returns \code{attr(x, "tidyILD")$provenance} (version + steps from
#' preprocessing). For analysis objects (e.g. fits from [ild_lme()], [ild_diagnostics()],
#' [ild_tvem()], [ild_power()], [ild_missing_model()]): returns \code{attr(x, "ild_provenance")},
#' which has \code{source_data_provenance} (snapshot of data provenance) and
#' \code{analysis_steps} (list of analysis step records).
#'
#' @param x An ILD object (see [is_ild()]) or an analysis object with \code{ild_provenance} attribute.
#' @return For data: list with \code{version} and \code{steps}. For analysis: list with
#'   \code{version}, \code{source_data_provenance}, \code{analysis_steps}. \code{NULL} if none.
#' @export
ild_provenance <- function(x) {
  ap <- attr(x, .ILD_PROVENANCE_ATTR, exact = TRUE)
  if (!is.null(ap)) return(ap)
  ild_get_history(x)
}

#' Print a human-readable log of preprocessing or analysis steps
#'
#' For ILD data, displays data provenance steps. For analysis objects (e.g. from
#' [ild_lme()], [ild_diagnostics()]), displays source data provenance (if any)
#' and analysis steps. Use [ild_provenance()] to get the raw structured object.
#'
#' @param x An ILD object (see [is_ild()]) or an analysis object with \code{ild_provenance}.
#' @return The provenance object (from [ild_provenance()]) invisibly, or a message if none.
#' @export
ild_history <- function(x) {
  prov <- ild_provenance(x)
  if (is.null(prov)) {
    message("No provenance recorded for this object.")
    return(invisible(NULL))
  }
  if (!is.null(prov$analysis_steps)) {
    cat("ILD analysis provenance (tidyILD ", prov$version, ")\n", sep = "")
    if (!is.null(prov$source_data_provenance) && length(prov$source_data_provenance$steps) > 0) {
      cat("  [Source data steps: ", length(prov$source_data_provenance$steps), "]\n", sep = "")
    }
    for (i in seq_along(prov$analysis_steps)) {
      s <- prov$analysis_steps[[i]]
      cat("  ", i, ". ", s$step, " @ ", s$timestamp, "\n", sep = "")
      if (length(s$args) > 0) {
        cat("      args: ", paste(names(s$args), collapse = ", "), "\n", sep = "")
      }
      if (length(s$outputs) > 0 && !is.null(s$outputs)) {
        cat("      outputs: ", paste(names(s$outputs), collapse = ", "), "\n", sep = "")
      }
    }
    return(invisible(prov))
  }
  cat("ILD provenance (tidyILD ", prov$version, ")\n", sep = "")
  for (i in seq_along(prov$steps)) {
    s <- prov$steps[[i]]
    cat("  ", i, ". ", s$step, " @ ", s$timestamp, "\n", sep = "")
    if (length(s$args) > 0) {
      cat("      args: ", paste(names(s$args), collapse = ", "), "\n", sep = "")
    }
    if (length(s$outputs) > 0 && !is.null(s$outputs)) {
      cat("      outputs: ", paste(names(s$outputs), collapse = ", "), "\n", sep = "")
    }
  }
  invisible(prov)
}

#' Export provenance to a JSON or YAML file
#'
#' Writes the full provenance structure (data or analysis) to a file for
#' reproducibility supplements, preregistration appendices, or lab archiving.
#' Requires the \pkg{jsonlite} package for JSON or \pkg{yaml} for YAML.
#'
#' @param x An ILD object or an analysis object with [ild_provenance()].
#' @param path Character. File path to write (e.g. \code{"analysis_provenance.json"}).
#' @param format Character. \code{"auto"} (default) infers from file extension
#'   (\code{.json} -> JSON, \code{.yaml} or \code{.yml} -> YAML). Use \code{"json"}
#'   or \code{"yaml"} to force a format.
#' @return The \code{path} invisibly, after writing the file.
#' @export
ild_export_provenance <- function(x, path, format = c("auto", "json", "yaml")) {
  format <- match.arg(format)
  if (missing(path) || is.null(path) || !is.character(path) || length(path) != 1L || is.na(path) || nchar(path) == 0) {
    stop("path must be a non-empty character string (file path).", call. = FALSE)
  }
  prov <- ild_provenance(x)
  if (is.null(prov)) {
    stop("Object has no provenance to export.", call. = FALSE)
  }
  if (format == "auto") {
    ext <- tolower(sub("^.*\\.([^.]+)$", "\\1", path))
    if (ext %in% c("yml", "yaml")) format <- "yaml" else format <- "json"
  }
  if (format == "json") {
    if (!requireNamespace("jsonlite", quietly = TRUE)) {
      stop("Package 'jsonlite' is required for JSON export. Install with install.packages(\"jsonlite\").", call. = FALSE)
    }
    txt <- jsonlite::toJSON(prov, auto_unbox = TRUE, pretty = TRUE, null = "null")
    writeLines(txt, path, useBytes = TRUE)
  } else {
    if (!requireNamespace("yaml", quietly = TRUE)) {
      stop("Package 'yaml' is required for YAML export. Install with install.packages(\"yaml\").", call. = FALSE)
    }
    txt <- yaml::as.yaml(prov)
    writeLines(txt, path, useBytes = TRUE)
  }
  invisible(path)
}

test_that("ild_prepare then ild_center yields two provenance steps", {
  d <- data.frame(
    id = c(1, 1, 2, 2),
    time = as.POSIXct(c(0, 10, 0, 10), origin = "1970-01-01"),
    x = c(10, 12, 20, 22)
  )
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, x)
  prov <- ild_provenance(x)
  expect_false(is.null(prov))
  expect_equal(length(prov$steps), 2L)
  expect_equal(prov$steps[[1]]$step, "ild_prepare")
  expect_equal(prov$steps[[2]]$step, "ild_center")
})

test_that("ild_provenance(x) has version and steps with step/timestamp/args/outputs", {
  d <- data.frame(id = c(1, 1, 2, 2), time = as.POSIXct(0:3, origin = "1970-01-01"), x = 1:4)
  x <- ild_prepare(d, id = "id", time = "time")
  prov <- ild_provenance(x)
  expect_false(is.null(prov))
  expect_true(is.character(prov$version))
  expect_true(nchar(prov$version) >= 1L)
  expect_true(is.list(prov$steps))
  expect_equal(length(prov$steps), 1L)
  s1 <- prov$steps[[1]]
  expect_equal(s1$step, "ild_prepare")
  expect_true("timestamp" %in% names(s1))
  expect_true(grepl("^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}$", s1$timestamp))
  expect_true(is.list(s1$args))
  expect_true("id" %in% names(s1$args))
  expect_true("time" %in% names(s1$args))
  expect_true(is.list(s1$outputs))
  expect_true("n_id" %in% names(s1$outputs))
  expect_true("n_obs" %in% names(s1$outputs))
})

test_that("ild_history(x) runs without error and prints step names", {
  d <- data.frame(id = c(1, 1, 2, 2), time = as.POSIXct(0:3, origin = "1970-01-01"), x = 1:4)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, x)
  prov <- ild_history(x)
  expect_false(is.null(prov))
  expect_equal(length(prov$steps), 2L)
  expect_output(ild_history(x), "ild_prepare")
  expect_output(ild_history(x), "ild_center")
})

test_that("ild_history(x) returns message when no provenance", {
  d <- data.frame(id = c(1, 1), time = as.POSIXct(0:1, origin = "1970-01-01"), x = 1:2)
  x <- ild_prepare(d, id = "id", time = "time")
  # Object has provenance from ild_prepare; test ild_provenance on NULL-like is via as_ild
  # For "no provenance" we'd need a legacy object without provenance - ild_normalize adds empty steps now.
  # So instead: ild_provenance returns NULL for non-ILD
  expect_null(ild_provenance(data.frame(a = 1)))
  expect_message(ild_history(data.frame(a = 1)), "No provenance")
})

test_that("analysis objects have ild_provenance and ild_provenance() returns it", {
  set.seed(1)
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  ap <- ild_provenance(fit)
  expect_false(is.null(ap))
  expect_true(is.list(ap$analysis_steps))
  expect_equal(length(ap$analysis_steps), 1L)
  expect_equal(ap$analysis_steps[[1]]$step, "ild_lme")
  expect_false(is.null(ap$source_data_provenance))
  expect_equal(length(ap$source_data_provenance$steps), 2L)
  expect_output(ild_history(fit), "ild_lme")
})

test_that("ild_missing_model return has analysis provenance, data provenance unchanged", {
  set.seed(1)
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 1)
  d$stress <- rnorm(nrow(d))
  d$mood <- d$y
  d$mood[sample(nrow(d), 5)] <- NA
  x <- ild_prepare(d, id = "id", time = "time")
  data_steps_before <- length(ild_provenance(x)$steps)
  mm <- ild_missing_model(x, "mood", "stress")
  expect_false(is.null(attr(mm, "ild_provenance")))
  expect_equal(attr(mm, "ild_provenance")$analysis_steps[[1]]$step, "ild_missing_model")
  expect_equal(length(ild_provenance(x)$steps), data_steps_before)
})

test_that("data provenance has object_type and steps have step_id", {
  d <- data.frame(id = c(1, 1, 2, 2), time = as.POSIXct(0:3, origin = "1970-01-01"), x = 1:4)
  x <- ild_prepare(d, id = "id", time = "time")
  prov <- ild_provenance(x)
  expect_equal(prov$object_type, "ild_data")
  expect_equal(prov$steps[[1]]$step_id, "1")
  x <- ild_center(x, x)
  prov <- ild_provenance(x)
  expect_equal(prov$steps[[2]]$step_id, "2")
})

test_that("analysis provenance has object_type and analysis_steps have step_id", {
  set.seed(1)
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  ap <- ild_provenance(fit)
  expect_equal(ap$object_type, "ild_model")
  expect_equal(ap$analysis_steps[[1]]$step_id, "1")
})

#' Joint MSM weights from IPTW and IPCW components
#'
#' Multiplies \code{.ipw_treat} and \code{.ipw_censor} into a single analysis
#' weight column (default \code{.ipw}) for use with [ild_ipw_refit()] and
#' [ild_diagnose()]. Optional scaling sets the mean joint weight to 1.
#'
#' @section Bundle integration:
#' Sets \code{.ipw} (or \code{joint_name}) so existing causal diagnostics and
#' guardrails apply to the **joint** weight.
#'
#' @family ild_diagnostics_utilities
#' @param x An ILD object containing \code{.ipw_treat} and \code{.ipw_censor}
#'   (from [ild_iptw_weights()] or [ild_iptw_msm_weights()], and [ild_ipcw_weights()]).
#' @param stabilize Character. \code{"none"}: \code{w_treat * w_censor} only.
#'   \code{"mean1"} (default): divide joint weights by their mean so the mean is 1
#'   (helps comparability with stabilized components).
#' @param trim Numeric of length 2. Quantiles for trimming the **joint** vector
#'   (default \code{c(0.01, 0.99)}).
#' @param joint_name Character. Name of the combined weight column (default \code{".ipw"}).
#' @return \code{x} with \code{joint_name} set; component columns unchanged.
#' @seealso [ild_iptw_weights()], [ild_iptw_msm_weights()], [ild_ipcw_weights()], [ild_ipw_refit()]
#' @export
#' @examples
#' set.seed(4)
#' d <- ild_simulate(n_id = 15, n_obs_per = 8, seed = 4)
#' d$stress <- rnorm(nrow(d))
#' d$trt <- as.integer(d$stress > 0)
#' x <- ild_prepare(d, id = "id", time = "time")
#' x <- ild_iptw_weights(x, "trt", "stress")
#' x <- ild_ipcw_weights(x, "stress")
#' x <- ild_joint_msm_weights(x)
#' summary(x$.ipw)
ild_joint_msm_weights <- function(x,
                                  stabilize = c("mean1", "none"),
                                  trim = c(0.01, 0.99),
                                  joint_name = ".ipw") {
  validate_ild(x)
  stabilize <- match.arg(stabilize)
  if (!".ipw_treat" %in% names(x)) {
    stop("Column '.ipw_treat' not found. Run ild_iptw_weights() first.", call. = FALSE)
  }
  if (!".ipw_censor" %in% names(x)) {
    stop("Column '.ipw_censor' not found. Run ild_ipcw_weights() first.", call. = FALSE)
  }
  wt <- x$.ipw_treat
  wc <- x$.ipw_censor
  if (length(wt) != nrow(x) || length(wc) != nrow(x)) {
    stop("Component weight length must match nrow(x).", call. = FALSE)
  }
  w <- wt * wc
  if (stabilize == "mean1") {
    m <- mean(w, na.rm = TRUE)
    if (is.finite(m) && m > 0) w <- w / m
  }
  if (length(trim) >= 2L && is.numeric(trim)) {
    w <- .ild_ipw_trim(w, trim)
  }
  attrs <- attributes(x)
  x[[joint_name]] <- w
  x <- .ild_ipw_restore_attrs(x, attrs)
  ild_add_step(x, "ild_joint_msm_weights",
    list(stabilize = stabilize, trim = trim, joint_name = joint_name),
    list(created = joint_name)
  )
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)

## ----prepare------------------------------------------------------------------
library(tidyILD)
# Simulate simple ILD
d <- ild_simulate(n_id = 10, n_obs_per = 12, irregular = TRUE, seed = 42)
# Prepare: encode time structure and add .ild_* columns
x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)

## ----summary------------------------------------------------------------------
ild_summary(x)
ild_spacing_class(x)

## ----center_lag---------------------------------------------------------------
x <- ild_center(x, y)
x <- ild_lag(x, y, mode = "gap_aware", max_gap = 7200)

## ----lme_no_ar1---------------------------------------------------------------
fit0 <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)

## ----lme_ar1------------------------------------------------------------------
fit1 <- ild_lme(y ~ 1, data = x, ar1 = TRUE, correlation_class = "CAR1")

## ----diagnostics--------------------------------------------------------------
diag <- ild_diagnostics(fit1, data = x)
names(diag)  # meta, data, stats
names(plot_ild_diagnostics(diag))  # plot names for requested types
# Pooled residual ACF (tibble)
head(diag$stats$acf$pooled)
# By-id ACF when by_id = TRUE: one tibble per person
names(diag$stats$acf$by_id)
head(diag$stats$acf$by_id[[1]])

## ----plot_trajectory, fig.alt = "Trajectory plot"-----------------------------
ild_plot(x, type = "trajectory", var = "y", max_ids = 5)

## ----plot_fitted, fig.alt = "Fitted vs observed"------------------------------
ild_plot(fit1, type = "fitted")

## ----msm_weights, eval = FALSE------------------------------------------------
# # Example skeleton (not run in the vignette build)
# x2 <- ild_simulate(n_id = 12, n_obs_per = 10, seed = 1)
# x2$stress <- rnorm(nrow(x2))
# x2$trt <- rbinom(nrow(x2), 1L, 0.45)
# x2 <- ild_prepare(x2, id = "id", time = "time")
# x2 <- ild_center(x2, y)
# x2 <- ild_iptw_weights(x2, treatment = "trt", predictors = "stress")
# # Sequential A_t: x2 <- ild_lag(x2, stress); x2 <- ild_lag(x2, trt); ...
# # x2 <- ild_iptw_msm_weights(x2, treatment = "trt", history = ~ stress_lag1 + trt_lag1)
# x2 <- ild_ipcw_weights(x2, predictors = "stress")
# x2 <- ild_joint_msm_weights(x2)
# fit_msm <- ild_lme(y ~ y_bp + y_wp + stress + (1 | id), data = x2,
#   ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
# fit_msm_w <- ild_ipw_refit(fit_msm, data = x2)

## ----msm_bootstrap, eval = requireNamespace("lme4", quietly = TRUE)-----------
set.seed(3)
xb <- ild_simulate(n_id = 10, n_obs_per = 5, seed = 3)
xb$stress <- rnorm(nrow(xb))
xb <- ild_prepare(xb, id = "id", time = "time")
xb <- ild_center(xb, y)
xb$.ipw <- runif(nrow(xb), 0.85, 1.15)
fb <- ild_lme(y ~ y_bp + y_wp + stress + (1 | id), data = xb,
  ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
fwb <- ild_ipw_refit(fb, data = xb, weights = ".ipw")
bs_fixed <- ild_msm_bootstrap(fwb, n_boot = 20L, weight_policy = "fixed_weights", seed = 3)
tidy_ild_msm_bootstrap(bs_fixed)
# reestimate_weights: weights_fn must return ILD with the weight column, e.g. re-run IPTW pipeline:
bs_re <- ild_msm_bootstrap(fwb, n_boot = 12L, weight_policy = "reestimate_weights",
  seed = 4, weights_fn = function(d) { d$.ipw <- runif(nrow(d), 0.85, 1.15); d })

## ----reproducibility----------------------------------------------------------
# Optional: build a manifest with scenario and seed, then bundle the fit for saving
manifest <- ild_manifest(seed = 42, scenario = ild_summary(x), include_session = FALSE)
bundle <- ild_bundle(fit1, manifest = manifest, label = "model_ar1")
# saveRDS(bundle, "run.rds")  # one file with result + manifest + label
names(bundle)

# Cross-backend validation harness — loaded by testthat and sourced by CI scripts.
# Depends on tidyILD being loaded (library / load_all). See inst/dev/BACKEND_VALIDATION_BENCHMARK_CONTRACT.md.

#' @keywords internal
harness_tier_rank <- function(tier) {
  tier <- match.arg(tier, c("smoke", "nightly", "full"))
  switch(tier, smoke = 1L, nightly = 2L, full = 3L)
}

#' @keywords internal
harness_scenario_in_tier <- function(scenario_tier_min, requested_tier) {
  harness_tier_rank(requested_tier) >= harness_tier_rank(scenario_tier_min)
}

#' Scenario manifest (v1) — columns used by runner
#' @keywords internal
harness_scenario_manifest <- function() {
  tibble::tibble(
    scenario_id = c("mlm_power_ri", "kfas_local_level", "ctsem_univariate", "mlm_lag_x"),
    scenario_type = c("mlm_power", "kfas", "ctsem", "mlm_lag"),
    tier_min = c("smoke", "smoke", "nightly", "smoke"),
    mlm_n_id = c(18L, NA_integer_, NA_integer_, NA_integer_),
    mlm_n_obs_per = c(10L, NA_integer_, NA_integer_, NA_integer_),
    mlm_effect_size = c(0.35, NA_real_, NA_real_, NA_real_),
    mlm_test_term = c("x", NA_character_, NA_character_, NA_character_),
    kfas_n_id = c(NA_integer_, 1L, NA_integer_, NA_integer_),
    kfas_n_obs_per = c(NA_integer_, 36L, NA_integer_, NA_integer_),
    ctsem_n_obs_per = c(NA_integer_, NA_integer_, 42L, NA_integer_),
    mlm_lag_n_id = c(NA_integer_, NA_integer_, NA_integer_, 20L),
    mlm_lag_n_obs_per = c(NA_integer_, NA_integer_, NA_integer_, 10L),
    mlm_lag_beta = c(NA_real_, NA_real_, NA_real_, 0.4),
    stringsAsFactors = FALSE
  )
}

#' @keywords internal
harness_default_backends_for_tier <- function(tier) {
  tier <- match.arg(tier, c("smoke", "nightly", "full"))
  if (tier == "smoke") {
    return(c("lme4", "kfas"))
  }
  if (tier == "nightly") {
    return(c("lme4", "nlme", "brms", "kfas", "ctsem"))
  }
  c("lme4", "nlme", "brms", "kfas", "ctsem")
}

#' @keywords internal
harness_backend_installed <- function(backend) {
  backend <- match.arg(backend, c("lme4", "nlme", "brms", "kfas", "ctsem"), several.ok = TRUE)
  vapply(backend, function(b) {
    switch(b,
      lme4 = requireNamespace("lme4", quietly = TRUE),
      nlme = requireNamespace("nlme", quietly = TRUE),
      brms = requireNamespace("brms", quietly = TRUE) &&
        (requireNamespace("rstan", quietly = TRUE) || requireNamespace("cmdstanr", quietly = TRUE)),
      kfas = requireNamespace("KFAS", quietly = TRUE),
      ctsem = requireNamespace("ctsem", quietly = TRUE) && requireNamespace("rstan", quietly = TRUE),
      FALSE
    )
  }, logical(1L))
}

#' @keywords internal
harness_expand_runs <- function(manifest, tier, backends_requested) {
  rows <- list()
  for (i in seq_len(nrow(manifest))) {
    sid <- manifest$scenario_id[i]
    stype <- manifest$scenario_type[i]
    tmin <- manifest$tier_min[i]
    if (!harness_scenario_in_tier(tmin, tier)) {
      next
    }
    bes <- switch(stype,
      mlm_power = intersect(backends_requested, c("lme4", "nlme", "brms")),
      mlm_lag = intersect(backends_requested, "lme4"),
      kfas = if ("kfas" %in% backends_requested) "kfas" else character(),
      ctsem = if ("ctsem" %in% backends_requested) "ctsem" else character(),
      character()
    )
    for (b in bes) {
      rows[[length(rows) + 1L]] <- tibble::tibble(
        scenario_id = sid,
        scenario_type = stype,
        backend = b
      )
    }
  }
  if (length(rows) == 0L) {
    return(tibble::tibble(
      scenario_id = character(),
      scenario_type = character(),
      backend = character()
    ))
  }
  dplyr::bind_rows(rows)
}

#' @keywords internal
harness_generate_mlm_power_data <- function(seed,
                                            n_id,
                                            n_obs_per,
                                            effect_size,
                                            test_term = "x") {
  set.seed(as.integer(seed))
  d <- ild_simulate(
    n_id = n_id,
    n_obs_per = n_obs_per,
    seed = as.integer(seed)
  )
  tt <- as.character(test_term)[1L]
  d[[tt]] <- stats::rnorm(nrow(d))
  d$y <- d$y + as.numeric(effect_size) * d[[tt]]
  ild_prepare(d, id = "id", time = "time")
}

#' @keywords internal
harness_generate_kfas_data <- function(seed, n_id, n_obs_per) {
  d <- ild_simulate(
    n_id = as.integer(n_id),
    n_obs_per = as.integer(n_obs_per),
    seed = as.integer(seed)
  )
  ild_prepare(d, id = "id", time = "time")
}

#' @keywords internal
harness_generate_mlm_lag_data <- function(seed, n_id, n_obs_per, beta_lag) {
  set.seed(as.integer(seed))
  d <- ild_simulate(
    n_id = as.integer(n_id),
    n_obs_per = as.integer(n_obs_per),
    seed = as.integer(seed)
  )
  d$x <- stats::rnorm(nrow(d))
  prep <- ild_prepare(d, id = "id", time = "time")
  prep2 <- ild_lag(prep, x, n = 1L, mode = "gap_aware")
  b <- as.numeric(beta_lag)[1L]
  prep2$y <- b * prep2$x_lag1 + stats::rnorm(nrow(prep2), sd = 0.75)
  prep2
}

#' @keywords internal
harness_generate_ctsem_data <- function(seed, n_obs_per) {
  d <- ild_simulate(
    n_id = 1L,
    n_obs_per = as.integer(n_obs_per),
    irregular = TRUE,
    seed = as.integer(seed)
  )
  x <- ild_prepare(d, id = "id", time = "time")
  ild_center(x, y)
}

#' @keywords internal
harness_resolve_tidy_row <- function(td, target_term) {
  if (is.null(td) || nrow(td) == 0L) {
    return(NULL)
  }
  if (target_term %in% td$term) {
    return(td[td$term == target_term, , drop = FALSE][1L, ])
  }
  pat <- paste0("^", gsub("([][{}()+*^$|\\\\?.])", "\\\\\\1", target_term), "$")
  hit <- td$term[grepl(pat, td$term)]
  if (length(hit) >= 1L) {
    return(td[td$term == hit[1L], , drop = FALSE][1L, ])
  }
  NULL
}

#' @keywords internal
harness_wald_interval <- function(est, se, level = 0.95) {
  if (!is.finite(est) || !is.finite(se) || se <= 0) {
    return(c(NA_real_, NA_real_))
  }
  z <- stats::qnorm(1 - (1 - level) / 2)
  c(est - z * se, est + z * se)
}

#' @keywords internal
harness_converged_heuristic <- function(fit, backend) {
  if (is.null(fit)) {
    return(FALSE)
  }
  if (backend == "lme4") {
    if (inherits(fit, "lmerMod")) {
      conv <- fit@optinfo$conv$lme4
      if (!is.null(conv$code) && is.numeric(conv$code)) {
        return(as.integer(conv$code) == 0L)
      }
    }
    return(TRUE)
  }
  if (backend == "nlme") {
    return(TRUE)
  }
  if (backend == "brms") {
    if (!requireNamespace("brms", quietly = TRUE)) {
      return(FALSE)
    }
    rh <- tryCatch(max(brms::rhat(fit), na.rm = TRUE), error = function(e) NA_real_)
    return(is.finite(rh) && rh < 1.08)
  }
  if (backend == "kfas") {
    return(inherits(fit, "ild_fit_kfas"))
  }
  if (backend == "ctsem") {
    return(inherits(fit, "ild_fit_ctsem"))
  }
  TRUE
}

#' @keywords internal
harness_singular_heuristic <- function(fit, backend) {
  if (backend != "lme4" || !inherits(fit, "lmerMod")) {
    return(NA)
  }
  if (!requireNamespace("lme4", quietly = TRUE)) {
    return(NA)
  }
  isTRUE(lme4::isSingular(fit))
}

#' Collect guardrail summary from diagnostics bundle
#' @keywords internal
collect_diagnostics_signals <- function(bundle) {
  if (is.null(bundle) || !inherits(bundle, "ild_diagnostics_bundle")) {
    return(list(
      n_guardrails = 0L,
      guardrail_rule_ids = ""
    ))
  }
  gr <- bundle$guardrails
  if (is.null(gr) || !tibble::is_tibble(gr) || nrow(gr) == 0L) {
    return(list(n_guardrails = 0L, guardrail_rule_ids = ""))
  }
  ids <- gr$rule_id
  ids <- ids[!is.na(ids) & nzchar(as.character(ids))]
  list(
    n_guardrails = nrow(gr),
    guardrail_rule_ids = paste(unique(as.character(ids)), collapse = ";")
  )
}

#' Fit one case; returns list(fit, error_message, elapsed_sec)
#' @keywords internal
harness_fit_one <- function(backend,
                            scenario_id,
                            scenario_type,
                            data,
                            manifest_row,
                            brms_iter = 200L,
                            brms_chains = 1L,
                            brms_warmup = 100L,
                            ctsem_iter = 150L,
                            ctsem_chains = 1L) {
  fit <- NULL
  err <- NA_character_
  t0 <- proc.time()[[3L]]
  if (backend == "lme4" && scenario_type == "mlm_power") {
    tt <- manifest_row$mlm_test_term
    ff <- stats::as.formula(paste("y ~", tt, "+ (1 | id)"))
    fit <- tryCatch(
      suppressWarnings(ild_lme(
        ff,
        data = data,
        ar1 = FALSE,
        warn_no_ar1 = FALSE,
        warn_uncentered = FALSE
      )),
      error = function(e) {
        err <<- conditionMessage(e)
        NULL
      }
    )
  } else if (backend == "lme4" && scenario_type == "mlm_lag") {
    fit <- tryCatch(
      suppressWarnings(ild_lme(
        y ~ x_lag1 + (1 | id),
        data = data,
        ar1 = FALSE,
        warn_no_ar1 = FALSE,
        warn_uncentered = FALSE
      )),
      error = function(e) {
        err <<- conditionMessage(e)
        NULL
      }
    )
  } else if (backend == "nlme" && scenario_type == "mlm_power") {
    tt <- manifest_row$mlm_test_term
    ff <- stats::as.formula(paste("y ~", tt))
    fit <- tryCatch(
      suppressWarnings(ild_lme(
        ff,
        data = data,
        ar1 = TRUE,
        warn_no_ar1 = FALSE,
        warn_uncentered = FALSE
      )),
      error = function(e) {
        err <<- conditionMessage(e)
        NULL
      }
    )
  } else if (backend == "brms" && scenario_type == "mlm_power") {
    tt <- manifest_row$mlm_test_term
    ff <- stats::as.formula(paste("y ~", tt, "+ (1 | id)"))
    fit <- tryCatch(
      suppressWarnings(ild_brms(
        ff,
        data = data,
        warn_uncentered = FALSE,
        iter = brms_iter,
        chains = brms_chains,
        warmup = brms_warmup,
        refresh = 0,
        seed = 1L
      )),
      error = function(e) {
        err <<- conditionMessage(e)
        NULL
      }
    )
  } else if (backend == "kfas" && scenario_type == "kfas") {
    fit <- tryCatch(
      suppressWarnings(ild_kfas(
        data = data,
        outcome = "y",
        time_units = "hours",
        irregular_time = TRUE
      )),
      error = function(e) {
        err <<- conditionMessage(e)
        NULL
      }
    )
  } else if (backend == "ctsem" && scenario_type == "ctsem") {
    fit <- tryCatch(
      suppressWarnings(ild_ctsem(
        data = data,
        outcome = "y",
        model_type = "stanct",
        iter = ctsem_iter,
        chains = ctsem_chains
      )),
      error = function(e) {
        err <<- conditionMessage(e)
        NULL
      }
    )
  }
  elapsed <- proc.time()[[3L]] - t0
  list(fit = fit, error_message = err, elapsed_sec = as.numeric(elapsed))
}

#' One raw benchmark row
#' @keywords internal
harness_extract_metrics_row <- function(fit,
                                        backend,
                                        scenario_type,
                                        truth,
                                        test_term,
                                        nominal_level = 0.95,
                                        elapsed_sec = NA_real_,
                                        fit_error = NA_character_) {
  skipped_reason <- ""
  fe_ok <- length(fit_error) > 0L && !is.na(fit_error[1L]) && nzchar(as.character(fit_error[1L]))
  if (fe_ok) {
    return(tibble::tibble(
      truth = truth,
      estimate = NA_real_,
      bias = NA_real_,
      squared_error = NA_real_,
      std_error = NA_real_,
      ci_low = NA_real_,
      ci_high = NA_real_,
      covered = NA,
      nominal_level = nominal_level,
      elapsed_sec = elapsed_sec,
      converged = FALSE,
      singular = NA,
      fit_error = as.character(fit_error[1L]),
      n_guardrails = 0L,
      guardrail_rule_ids = "",
      skipped_reason = ""
    ))
  }
  if (is.null(fit)) {
    return(tibble::tibble(
      truth = truth,
      estimate = NA_real_,
      bias = NA_real_,
      squared_error = NA_real_,
      std_error = NA_real_,
      ci_low = NA_real_,
      ci_high = NA_real_,
      covered = NA,
      nominal_level = nominal_level,
      elapsed_sec = elapsed_sec,
      converged = FALSE,
      singular = harness_singular_heuristic(fit, backend),
      fit_error = if (is.na(fit_error)) "null fit" else fit_error,
      n_guardrails = 0L,
      guardrail_rule_ids = "",
      skipped_reason = ""
    ))
  }

  data <- attr(fit, "ild_data", exact = TRUE)
  bundle <- tryCatch(
    suppressWarnings(ild_diagnose(fit, data = data, type = "qq")),
    error = function(e) NULL
  )
  sig <- collect_diagnostics_signals(bundle)

  td <- tryCatch(ild_tidy(fit), error = function(e) NULL)
  row <- NULL
  if (scenario_type %in% c("mlm_power", "mlm_lag")) {
    row <- harness_resolve_tidy_row(td, test_term)
  } else if (scenario_type == "kfas") {
    row <- harness_resolve_tidy_row(td, test_term %||% "observation_variance")
  } else if (scenario_type == "ctsem") {
    row <- if (!is.null(td) && nrow(td) > 0L) td[1L, , drop = FALSE] else NULL
  }

  est <- NA_real_
  se <- NA_real_
  lo <- NA_real_
  hi <- NA_real_
  if (!is.null(row)) {
    est <- as.numeric(row$estimate)[1L]
    se <- as.numeric(row$std_error)[1L]
    lo <- as.numeric(row$conf_low)[1L]
    hi <- as.numeric(row$conf_high)[1L]
  }
  if (scenario_type %in% c("mlm_power", "mlm_lag") && is.finite(se) && (!is.finite(lo) || !is.finite(hi))) {
    w <- harness_wald_interval(est, se, nominal_level)
    lo <- w[1L]
    hi <- w[2L]
  }

  conv <- harness_converged_heuristic(fit, backend)
  sing <- harness_singular_heuristic(fit, backend)
  covered <- NA
  bias <- NA_real_
  sqe <- NA_real_
  if (is.finite(truth) && is.finite(est)) {
    bias <- est - truth
    sqe <- bias^2
    if (is.finite(lo) && is.finite(hi)) {
      covered <- truth >= lo & truth <= hi
    }
  }

  tibble::tibble(
    truth = truth,
    estimate = est,
    bias = bias,
    squared_error = sqe,
    std_error = se,
    ci_low = lo,
    ci_high = hi,
    covered = covered,
    nominal_level = nominal_level,
    elapsed_sec = elapsed_sec,
    converged = conv,
    singular = sing,
    fit_error = NA_character_,
    n_guardrails = sig$n_guardrails,
    guardrail_rule_ids = sig$guardrail_rule_ids,
    skipped_reason = skipped_reason
  )
}

#' @keywords internal
`%||%` <- function(x, y) {
  if (is.null(x)) y else x
}

#' Summarize raw benchmark tibble
#' @keywords internal
summarize_validation_metrics <- function(raw_df) {
  if (nrow(raw_df) == 0L) {
    return(raw_df)
  }
  raw_df <- tibble::as_tibble(raw_df)
  raw_df$has_skip <- !is.na(raw_df$skipped_reason) & nzchar(as.character(raw_df$skipped_reason))
  raw_df$install_skipped <- raw_df$skipped_reason == "backend_not_installed"
  dplyr::group_by(raw_df, .data$tier, .data$scenario_id, .data$backend) %>%
    dplyr::summarise(
      n_replicates = dplyr::n(),
      n_skipped = sum(.data$has_skip, na.rm = TRUE),
      n_install_skipped = sum(.data$install_skipped, na.rm = TRUE),
      n_ran = sum(!.data$install_skipped, na.rm = TRUE),
      n_converged = sum(.data$converged %in% TRUE & !.data$install_skipped, na.rm = TRUE),
      convergence_rate = ifelse(
        sum(!.data$install_skipped) > 0L,
        mean(.data$converged[!.data$install_skipped] %in% TRUE, na.rm = TRUE),
        NA_real_
      ),
      mean_bias = mean(.data$bias[!.data$install_skipped], na.rm = TRUE),
      rmse = sqrt(mean(.data$squared_error[!.data$install_skipped], na.rm = TRUE)),
      mean_coverage = ifelse(
        sum(!is.na(.data$covered[!.data$install_skipped])) > 0L,
        mean(.data$covered[!.data$install_skipped] %in% TRUE, na.rm = TRUE),
        NA_real_
      ),
      n_coverage_defined = sum(!is.na(.data$covered[!.data$install_skipped])),
      calibration_gap = ifelse(
        sum(!is.na(.data$covered[!.data$install_skipped])) > 0L,
        dplyr::first(.data$nominal_level) -
          mean(.data$covered[!.data$install_skipped] %in% TRUE, na.rm = TRUE),
        NA_real_
      ),
      mean_elapsed_sec = mean(.data$elapsed_sec[!.data$install_skipped], na.rm = TRUE),
      mean_n_guardrails = mean(.data$n_guardrails[!.data$install_skipped], na.rm = TRUE),
      .groups = "drop"
    )
}

#' Parse --key value / --key=value CLI args
#' @keywords internal
harness_parse_cli_args <- function(args = NULL) {
  if (is.null(args)) {
    args <- commandArgs(trailingOnly = TRUE)
  }
  out <- list(
    tier = "smoke",
    backends = NULL,
    n_sim = 3L,
    seed = 20260406L,
    out_dir = file.path(getwd(), "benchmark_artifacts"),
    run_id = format(Sys.time(), "bench-%Y%m%d-%H%M%S"),
    git_sha = NA_character_,
    brms_iter = 200L,
    brms_chains = 1L,
    ctsem_iter = 150L,
    ctsem_chains = 1L
  )
  if (length(args) == 0L) {
    return(out)
  }
  i <- 1L
  while (i <= length(args)) {
    a <- args[i]
    if (grepl("^--tier=", a)) {
      out$tier <- sub("^--tier=", "", a)
    } else if (a == "--tier" && i < length(args)) {
      out$tier <- args[i + 1L]
      i <- i + 1L
    } else if (grepl("^--backends=", a)) {
      out$backends <- strsplit(sub("^--backends=", "", a), ",", fixed = TRUE)[[1L]]
    } else if (a == "--backends" && i < length(args)) {
      out$backends <- strsplit(args[i + 1L], ",", fixed = TRUE)[[1L]]
      i <- i + 1L
    } else if (grepl("^--n-sim=", a)) {
      out$n_sim <- as.integer(sub("^--n-sim=", "", a))
    } else if (a == "--n-sim" && i < length(args)) {
      out$n_sim <- as.integer(args[i + 1L])
      i <- i + 1L
    } else if (grepl("^--seed=", a)) {
      out$seed <- as.integer(sub("^--seed=", "", a))
    } else if (a == "--seed" && i < length(args)) {
      out$seed <- as.integer(args[i + 1L])
      i <- i + 1L
    } else if (grepl("^--out-dir=", a)) {
      out$out_dir <- sub("^--out-dir=", "", a)
    } else if (a == "--out-dir" && i < length(args)) {
      out$out_dir <- args[i + 1L]
      i <- i + 1L
    } else if (grepl("^--run-id=", a)) {
      out$run_id <- sub("^--run-id=", "", a)
    } else if (a == "--run-id" && i < length(args)) {
      out$run_id <- args[i + 1L]
      i <- i + 1L
    } else if (grepl("^--git-sha=", a)) {
      out$git_sha <- sub("^--git-sha=", "", a)
    } else if (a == "--git-sha" && i < length(args)) {
      out$git_sha <- args[i + 1L]
      i <- i + 1L
    }
    i <- i + 1L
  }
  if (is.null(out$backends) || length(out$backends) == 0L) {
    out$backends <- harness_default_backends_for_tier(out$tier)
  }
  out
}

#' Run full benchmark matrix; returns list(raw, summary, metadata, skipped_backends)
#' @keywords internal
harness_run_benchmark <- function(tier = c("smoke", "nightly", "full"),
                                    backends = NULL,
                                    n_sim = 3L,
                                    seed = 20260406L,
                                    run_id = NULL,
                                    git_sha = NA_character_,
                                    brms_iter = 200L,
                                    brms_chains = 1L,
                                    ctsem_iter = 150L,
                                    ctsem_chains = 1L) {
  tier <- match.arg(tier)
  if (is.null(backends)) {
    backends <- harness_default_backends_for_tier(tier)
  }
  if (is.null(run_id)) {
    run_id <- format(Sys.time(), "bench-%Y%m%d-%H%M%S")
  }
  manifest <- harness_scenario_manifest()
  grid <- harness_expand_runs(manifest, tier, backends)
  skipped_backends <- backends[!harness_backend_installed(backends)]
  rows <- list()
  if (nrow(grid) == 0L) {
    raw <- tibble::tibble()
  } else {
    rep_idx <- 0L
    for (r in seq_len(nrow(grid))) {
      sid <- grid$scenario_id[r]
      st <- grid$scenario_type[r]
      be <- grid$backend[r]
      mrow <- manifest[manifest$scenario_id == sid, , drop = FALSE][1L, ]
      if (!harness_backend_installed(be)) {
        for (k in seq_len(n_sim)) {
          rep_idx <- rep_idx + 1L
          rows[[length(rows) + 1L]] <- tibble::tibble(
            run_id = run_id,
            git_sha = git_sha %||% NA_character_,
            tier = tier,
            scenario_id = sid,
            backend = be,
            replicate_id = k,
            truth = NA_real_,
            estimate = NA_real_,
            bias = NA_real_,
            squared_error = NA_real_,
            std_error = NA_real_,
            ci_low = NA_real_,
            ci_high = NA_real_,
            covered = NA,
            nominal_level = 0.95,
            elapsed_sec = NA_real_,
            converged = FALSE,
            singular = NA,
            fit_error = NA_character_,
            n_guardrails = 0L,
            guardrail_rule_ids = "",
            skipped_reason = "backend_not_installed"
          )
        }
        next
      }
      for (k in seq_len(n_sim)) {
        rep_idx <- rep_idx + 1L
        rep_seed <- as.integer(seed) + rep_idx
        data <- NULL
        truth <- NA_real_
        tt <- "x"
        if (st == "mlm_power") {
          data <- harness_generate_mlm_power_data(
            rep_seed,
            mrow$mlm_n_id,
            mrow$mlm_n_obs_per,
            mrow$mlm_effect_size,
            mrow$mlm_test_term
          )
          truth <- as.numeric(mrow$mlm_effect_size)
          tt <- as.character(mrow$mlm_test_term)
        } else if (st == "kfas") {
          data <- harness_generate_kfas_data(
            rep_seed,
            mrow$kfas_n_id,
            mrow$kfas_n_obs_per
          )
          truth <- NA_real_
          tt <- "observation_variance"
        } else if (st == "ctsem") {
          data <- harness_generate_ctsem_data(rep_seed, mrow$ctsem_n_obs_per)
          truth <- NA_real_
          tt <- NA_character_
        } else if (st == "mlm_lag") {
          data <- harness_generate_mlm_lag_data(
            rep_seed,
            mrow$mlm_lag_n_id,
            mrow$mlm_lag_n_obs_per,
            mrow$mlm_lag_beta
          )
          truth <- as.numeric(mrow$mlm_lag_beta)
          tt <- "x_lag1"
        }
        fo <- harness_fit_one(
          be, sid, st, data, mrow,
          brms_iter = brms_iter,
          brms_chains = brms_chains,
          brms_warmup = max(50L, brms_iter %/% 2L),
          ctsem_iter = ctsem_iter,
          ctsem_chains = ctsem_chains
        )
        err <- fo$error_message
        if (!is.na(err) && nzchar(err)) {
          mr <- harness_extract_metrics_row(
            NULL, be, st, truth, tt,
            elapsed_sec = fo$elapsed_sec,
            fit_error = err
          )
        } else {
          mr <- harness_extract_metrics_row(
            fo$fit, be, st, truth, tt,
            elapsed_sec = fo$elapsed_sec
          )
        }
        rows[[length(rows) + 1L]] <- tibble::tibble(
          run_id = run_id,
          git_sha = git_sha %||% NA_character_,
          tier = tier,
          scenario_id = sid,
          backend = be,
          replicate_id = k
        ) %>%
          dplyr::bind_cols(mr)
      }
    }
    raw <- dplyr::bind_rows(rows)
  }
  summ <- if (nrow(raw) > 0L) summarize_validation_metrics(raw) else tibble::tibble()
  meta <- list(
    package_version = as.character(utils::packageVersion("tidyILD")),
    r_version = paste(R.version$major, R.version$minor, sep = "."),
    platform = R.version$platform,
    git_sha = git_sha,
    seed_base = seed,
    tier = tier,
    backends_requested = backends,
    backends_skipped_install = skipped_backends,
    scenario_manifest_version = 2L,
    timestamp_utc = format(Sys.time(), tz = "UTC", usetz = TRUE)
  )
  list(raw = raw, summary = summ, metadata = meta, skipped_backends = skipped_backends)
}

#' Read thresholds JSON (requires jsonlite)
#' @keywords internal
harness_read_thresholds_json <- function(path) {
  if (!requireNamespace("jsonlite", quietly = TRUE)) {
    stop("jsonlite required to read thresholds (Suggests).", call. = FALSE)
  }
  jsonlite::read_json(path, simplifyVector = FALSE)
}

#' Evaluate summary against thresholds; returns tibble of check rows
#' @keywords internal
harness_evaluate_thresholds <- function(summary_df,
                                        thresholds_list,
                                        warn_only_backends = c("brms", "ctsem")) {
  if (nrow(summary_df) == 0L) {
    return(tibble::tibble(
      scenario_id = character(),
      backend = character(),
      check = character(),
      status = character(),
      detail = character()
    ))
  }
  defs <- thresholds_list$defaults %||% list()
  by_be <- thresholds_list$by_backend %||% list()
  by_sc <- thresholds_list$by_scenario %||% list()
  min_cov <- as.numeric(defs$min_coverage %||% 0.65)
  max_fail <- as.numeric(defs$max_convergence_failure_rate %||% 0.5)
  min_conv_n <- as.integer(defs$min_n_converged %||% 1L)

  out <- list()
  for (i in seq_len(nrow(summary_df))) {
    sid <- summary_df$scenario_id[i]
    be <- summary_df$backend[i]
    sc_row <- summary_df[i, , drop = FALSE]
    be_cfg <- by_be[[be]] %||% list()
    sc_cfg <- by_sc[[sid]] %||% list()
    warn_only <- isTRUE(be_cfg$warn_only) || isTRUE(sc_cfg$warn_only) ||
      be %in% warn_only_backends
    min_cov_i <- as.numeric(sc_cfg$min_coverage %||% be_cfg$min_coverage %||% min_cov)
    max_fail_i <- as.numeric(
      sc_cfg$max_convergence_failure_rate %||%
        be_cfg$max_convergence_failure_rate %||% max_fail
    )
    min_conv_i <- as.integer(sc_cfg$min_n_converged %||% be_cfg$min_n_converged %||% min_conv_n)

    n_ran <- sc_row$n_ran
    if (!is.finite(n_ran) || n_ran <= 0L) {
      out[[length(out) + 1L]] <- tibble::tibble(
        scenario_id = sid,
        backend = be,
        check = "convergence",
        status = "pass",
        detail = "no runs (backend not installed or filtered)"
      )
      next
    }

    fail_rate <- 1 - as.numeric(sc_row$convergence_rate)
    if (!is.finite(fail_rate)) {
      fail_rate <- 1
    }

    st_conv <- if (sc_row$n_converged < min_conv_i) {
      "fail"
    } else if (fail_rate > max_fail_i) {
      "fail"
    } else {
      "pass"
    }
    detail_conv <- sprintf(
      "n_converged=%g, rate=%.3f, max_fail=%.3f",
      sc_row$n_converged, sc_row$convergence_rate, max_fail_i
    )
    out[[length(out) + 1L]] <- tibble::tibble(
      scenario_id = sid,
      backend = be,
      check = "convergence",
      status = if (warn_only && st_conv == "fail") "warn" else st_conv,
      detail = detail_conv
    )

    n_cov <- sc_row$n_coverage_defined
    if (is.finite(n_cov) && n_cov > 0L && is.finite(sc_row$mean_coverage)) {
      st_cov <- if (sc_row$mean_coverage < min_cov_i) "fail" else "pass"
      out[[length(out) + 1L]] <- tibble::tibble(
        scenario_id = sid,
        backend = be,
        check = "coverage",
        status = if (warn_only && st_cov == "fail") "warn" else st_cov,
        detail = sprintf("mean_cov=%.3f min=%.3f", sc_row$mean_coverage, min_cov_i)
      )
    }
  }
  dplyr::bind_rows(out)
}

#' Exit status helper: 0 pass, 1 fail
#' @keywords internal
harness_threshold_exit_status <- function(checks_df) {
  if (nrow(checks_df) == 0L) {
    return(0L)
  }
  if (any(checks_df$status == "fail")) {
    return(1L)
  }
  0L
}

test_that("harness scenario manifest is stable", {
  m <- harness_scenario_manifest()
  expect_true(all(c("scenario_id", "scenario_type", "tier_min") %in% names(m)))
  expect_true("mlm_power_ri" %in% m$scenario_id)
  expect_true("kfas_local_level" %in% m$scenario_id)
})

test_that("harness_expand_runs respects tier and backends", {
  m <- harness_scenario_manifest()
  g <- harness_expand_runs(m, "smoke", c("lme4", "kfas"))
  expect_true(all(g$backend %in% c("lme4", "kfas")))
  expect_true(any(g$scenario_id == "mlm_power_ri" & g$backend == "lme4"))
  expect_true(any(g$scenario_id == "mlm_lag_x" & g$backend == "lme4"))
  expect_true(any(g$scenario_id == "kfas_local_level" & g$backend == "kfas"))
  g2 <- harness_expand_runs(m, "smoke", "lme4")
  expect_false(any(g2$backend == "kfas"))
  g3 <- harness_expand_runs(m, "nightly", c("ctsem"))
  expect_true(any(g3$scenario_id == "ctsem_univariate"))
})

test_that("harness_parse_cli_args parses flags", {
  a <- harness_parse_cli_args(c("--tier", "nightly", "--n-sim", "5", "--seed", "1"))
  expect_equal(a$tier, "nightly")
  expect_equal(a$n_sim, 5L)
  expect_equal(a$seed, 1L)
})

test_that("summarize_validation_metrics and thresholds are deterministic", {
  est <- c(0.3, 0.4, 0.36)
  truth <- 0.35
  bias <- est - truth
  raw <- tibble::tibble(
    run_id = "r1",
    git_sha = "abc",
    tier = "smoke",
    scenario_id = "mlm_power_ri",
    backend = "lme4",
    replicate_id = 1:3,
    truth = truth,
    estimate = est,
    bias = bias,
    squared_error = bias^2,
    std_error = c(0.1, 0.1, 0.1),
    ci_low = est - 0.2,
    ci_high = est + 0.2,
    covered = c(TRUE, TRUE, TRUE),
    nominal_level = 0.95,
    elapsed_sec = c(1, 1, 1),
    converged = TRUE,
    singular = FALSE,
    fit_error = NA_character_,
    n_guardrails = 0L,
    guardrail_rule_ids = "",
    skipped_reason = ""
  )
  s <- summarize_validation_metrics(raw)
  expect_equal(nrow(s), 1L)
  expect_true(is.finite(s$convergence_rate))
  expect_true(s$mean_coverage <= 1)

  thr <- list(
    defaults = list(min_coverage = 0.5, max_convergence_failure_rate = 0.5, min_n_converged = 1L),
    by_backend = list(),
    by_scenario = list()
  )
  chk <- harness_evaluate_thresholds(s, thr)
  expect_true(all(chk$status %in% c("pass", "warn", "fail")))
  expect_equal(harness_threshold_exit_status(chk), 0L)
})

test_that("harness_evaluate_thresholds fails on bad convergence", {
  s <- tibble::tibble(
    tier = "smoke",
    scenario_id = "mlm_power_ri",
    backend = "lme4",
    n_replicates = 5L,
    n_skipped = 0L,
    n_install_skipped = 0L,
    n_ran = 5L,
    n_converged = 0L,
    convergence_rate = 0,
    mean_bias = NA_real_,
    rmse = NA_real_,
    mean_coverage = NA_real_,
    n_coverage_defined = 0L,
    calibration_gap = NA_real_,
    mean_elapsed_sec = 1,
    mean_n_guardrails = 0
  )
  thr <- list(
    defaults = list(min_coverage = 0.5, max_convergence_failure_rate = 0.1, min_n_converged = 1L),
    by_backend = list(),
    by_scenario = list()
  )
  chk <- harness_evaluate_thresholds(s, thr)
  expect_true(any(chk$status == "fail"))
  expect_equal(harness_threshold_exit_status(chk), 1L)
})

test_that("harness smoke benchmark runs for lme4 when available", {
  res <- harness_run_benchmark(
    tier = "smoke",
    backends = "lme4",
    n_sim = 2L,
    seed = 4242L,
    run_id = "test-smoke",
    git_sha = "testsha"
  )
  expect_true(nrow(res$raw) >= 2L)
  expect_true(all(c("estimate", "converged", "scenario_id", "backend") %in% names(res$raw)))
  expect_true(all(res$raw$backend == "lme4"))
})

test_that("harness_read_thresholds_json reads package thresholds", {
  skip_if_not_installed("jsonlite")
  p <- system.file("benchmarks", "thresholds-smoke.json", package = "tidyILD")
  skip_if(!nzchar(p), "installed package has no inst/benchmarks (use load_all)")
  thr <- harness_read_thresholds_json(p)
  expect_type(thr, "list")
  expect_true(!is.null(thr$defaults))
})

#' Simulate simple ILD for examples, tests, and power analysis
#'
#' Generates a tibble with id, time, and outcome \code{y}. Optionally uses
#' AR(1) within-person correlation and configurable WP/BP variance. Use
#' [ild_prepare()] after to get a proper ILD object.
#'
#' @param n_id Integer. Number of persons (default 5).
#' @param n_obs_per Integer. Observations per person (default 10).
#' @param n_time Integer. Alias for \code{n_obs_per} (observations per person). If provided, overrides \code{n_obs_per}.
#' @param irregular Logical. If \code{TRUE}, add random jitter to time (default \code{FALSE}).
#' @param ar1 Numeric or \code{NULL}. If numeric, within-person AR(1) correlation (e.g. 0.4). If \code{NULL} or 0, no AR (default \code{NULL}).
#' @param wp_effect Numeric. Scale (SD) of within-person innovation (default 0.5).
#' @param bp_effect Numeric. Scale (SD) of between-person random intercept (default 1).
#' @param seed Integer. Random seed for reproducibility (default 42).
#' @return A data frame with columns \code{id}, \code{time} (POSIXct), and \code{y}.
#' @seealso \href{../doc/benchmark-simulation-recovery.html}{\code{vignette("benchmark-simulation-recovery", package = "tidyILD")}}
#'   (simulation benchmarks and DGP description).
#' @importFrom stats runif rnorm
#' @examples
#' d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 1)
#' x <- ild_prepare(d, id = "id", time = "time")
#' d2 <- ild_simulate(n_id = 100, n_time = 50, ar1 = 0.4, wp_effect = 0.6,
#'   bp_effect = 0.3, irregular = TRUE, seed = 1)
#' @export
ild_simulate <- function(n_id = 5L,
                        n_obs_per = 10L,
                        n_time = NULL,
                        irregular = FALSE,
                        ar1 = NULL,
                        wp_effect = 0.5,
                        bp_effect = 1,
                        seed = 42L) {
  set.seed(seed)
  n_id <- as.integer(n_id)[1]
  if (!is.null(n_time)) n_obs_per <- as.integer(n_time)[1]
  else n_obs_per <- as.integer(n_obs_per)[1]
  phi <- if (is.null(ar1) || (is.numeric(ar1) && ar1 == 0)) 0 else as.numeric(ar1)[1]
  id <- rep(seq_len(n_id), each = n_obs_per)
  base_time <- rep(seq(0, by = 3600, length.out = n_obs_per), n_id)
  if (irregular) {
    jitter <- runif(length(base_time), -600, 600)
    base_time <- base_time + jitter
    base_time <- pmax(0, base_time)
  }
  time <- as.POSIXct(base_time, origin = "1970-01-01")
  person_eff <- rep(rnorm(n_id, 0, bp_effect), each = n_obs_per)
  time_eff <- base_time / 3600 * 0.1
  n_tot <- n_id * n_obs_per
  if (phi != 0) {
    innov <- rnorm(n_tot, 0, wp_effect)
    y <- numeric(n_tot)
    for (i in seq_len(n_id)) {
      idx <- (i - 1) * n_obs_per + seq_len(n_obs_per)
      y[idx][1] <- person_eff[idx][1] + time_eff[idx][1] + innov[idx][1]
      for (t in 2:n_obs_per) {
        y[idx][t] <- person_eff[idx][t] + time_eff[idx][t] +
          phi * (y[idx][t - 1] - person_eff[idx][t - 1] - time_eff[idx][t - 1]) + innov[idx][t]
      }
    }
  } else {
    y <- person_eff + time_eff + rnorm(n_tot, 0, wp_effect)
  }
  data.frame(id = id, time = time, y = y)
}

test_that("ild_check_lags returns expected columns", {
  d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_lag(x, y, mode = "gap_aware", max_gap = 7200)
  out <- ild_check_lags(x, lag_vars = "y_lag1", max_gap = 7200)
  expect_named(out, c("var", "lag", "n_valid", "n_invalid", "n_first", "n_total", "pct_valid", "pct_invalid"))
  expect_equal(out$var, "y_lag1")
  expect_equal(out$lag, "1")
  expect_equal(out$n_first, 3)
  expect_equal(out$pct_invalid, 100 - out$pct_valid)
})

test_that("ild_check_lags with no lag vars returns empty tibble", {
  x <- ild_prepare(ild_simulate(n_id = 2, n_obs_per = 3, seed = 1), id = "id", time = "time")
  out <- ild_check_lags(x, lag_vars = character(0))
  expect_equal(nrow(out), 0)
})

test_that("ild_check_lags parses lag and reports pct_invalid", {
  d <- ild_simulate(n_id = 2, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_lag(x, y, n = 2, mode = "gap_aware", max_gap = 7200)
  out <- ild_check_lags(x)
  expect_true("lag" %in% names(out))
  expect_true("pct_invalid" %in% names(out))
  expect_equal(out$lag, "2")
})

test_that("ild_crosslag returns fit, lag_term, lag_check, data", {
  d <- ild_simulate(n_id = 8, n_obs_per = 6, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_crosslag(x, y, y, lag = 1, ar1 = FALSE, warn_no_ar1 = FALSE)
  expect_named(out, c("fit", "lag_term", "lag_check", "data"))
  expect_true(inherits(out$fit, "lmerMod"))
  expect_s3_class(out$lag_term, "tbl_df")
  expect_true("y_lag1" %in% out$lag_term$term)
  expect_s3_class(out$lag_check, "tbl_df")
  expect_equal(out$lag_check$var, "y_lag1")
  expect_true(is_ild(out$data))
  expect_true("y_lag1" %in% names(out$data))
})

test_that("ild_crosslag with include_diagnostics adds diagnostics slot", {
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 2)
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_crosslag(x, "y", "y", lag = 1, mode = "index", include_diagnostics = TRUE,
                      ar1 = FALSE)
  expect_true("diagnostics" %in% names(out))
  if (!is.null(out$diagnostics)) expect_s3_class(out$diagnostics, "ild_diagnostics")
})

test_that("ild_crosslag errors on missing outcome or predictor", {
  x <- ild_prepare(ild_simulate(n_id = 3, n_obs_per = 4, seed = 1), id = "id", time = "time")
  expect_error(ild_crosslag(x, "nonexistent", "y", lag = 1, warn_no_ar1 = FALSE), "not found")
  expect_error(ild_crosslag(x, "y", "nonexistent", lag = 1, warn_no_ar1 = FALSE), "not found")
})

test_that("ild_circadian returns ggplot when .ild_time is POSIXct", {
  d <- ild_simulate(n_id = 5, n_obs_per = 12, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  p <- ild_circadian(x, y)
  expect_s3_class(p, "gg")
  p2 <- ild_circadian(x, y, type = "line")
  expect_s3_class(p2, "gg")
})

test_that("ild_circadian errors when .ild_time is not POSIXct", {
  d <- data.frame(id = rep(1:2, each = 5), time = 1:5, y = rnorm(10))
  x <- ild_prepare(d, id = "id", time = "time")
  expect_error(ild_circadian(x, y), "POSIXct")
})

test_that("ild_circadian errors on missing variable", {
  x <- ild_prepare(ild_simulate(n_id = 2, n_obs_per = 3, seed = 1), id = "id", time = "time")
  expect_error(ild_circadian(x, nonexistent), "not found")
})

test_that("ild_tvem fits on simulated ILD and has expected class and meta", {
  set.seed(1)
  d <- ild_simulate(n_id = 8, n_obs_per = 12, seed = 1)
  d$x <- rnorm(nrow(d))
  x <- ild_prepare(d, id = "id", time = "time")
  tv <- ild_tvem(x, "y", "x", k = 5L, re_id = TRUE)
  expect_true(inherits(tv, "tidyild_tvem"))
  expect_true(inherits(tv, "gam"))
  meta <- attr(tv, "ild_tvem_meta")
  expect_named(meta, c("outcome", "predictor", "time_var", "k", "re_id"))
  expect_equal(meta$outcome, "y")
  expect_equal(meta$predictor, "x")
  expect_equal(meta$time_var, ".ild_time_num")
  expect_equal(meta$k, 5L)
})

test_that("ild_tvem_plot returns a ggplot", {
  set.seed(1)
  d <- ild_simulate(n_id = 6, n_obs_per = 10, seed = 1)
  d$x <- rnorm(nrow(d))
  x <- ild_prepare(d, id = "id", time = "time")
  tv <- ild_tvem(x, "y", "x", k = 5L, re_id = TRUE)
  p <- ild_tvem_plot(tv, n_grid = 50L)
  expect_s3_class(p, "ggplot")
})

test_that("ild_tvem runs in under 2 seconds on small data", {
  set.seed(1)
  d <- ild_simulate(n_id = 5, n_obs_per = 8, seed = 1)
  d$x <- rnorm(nrow(d))
  x <- ild_prepare(d, id = "id", time = "time")
  elapsed <- system.time(ild_tvem(x, "y", "x", k = 5L))["elapsed"]
  expect_lt(elapsed, 2)
})

test_that("ild_simulate returns data frame with id, time, y", {
  d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 1)
  expect_s3_class(d, "data.frame")
  expect_true(all(c("id", "time", "y") %in% names(d)))
  expect_equal(nrow(d), 3 * 5)
})

test_that("ild_simulate n_time overrides n_obs_per", {
  d <- ild_simulate(n_id = 2, n_time = 7, seed = 1)
  expect_equal(nrow(d), 2 * 7)
})

test_that("ild_simulate with ar1 produces data", {
  d <- ild_simulate(n_id = 5, n_obs_per = 10, ar1 = 0.4, seed = 1)
  expect_equal(nrow(d), 50)
  expect_true(is.numeric(d$y))
})

test_that("ild_simulate wp_effect and bp_effect scale output", {
  d1 <- ild_simulate(n_id = 10, n_obs_per = 20, wp_effect = 0.1, bp_effect = 2, seed = 1)
  d2 <- ild_simulate(n_id = 10, n_obs_per = 20, wp_effect = 1, bp_effect = 0.1, seed = 1)
  expect_true(stats::var(d1$y) > stats::var(d2$y))
})

test_that("ild_missing_bias returns list with estimate, p_value, message", {
  set.seed(1)
  d <- ild_simulate(n_id = 15, n_obs_per = 8, seed = 1)
  d$stress <- rnorm(nrow(d))
  d$mood <- d$y
  d$mood[sample(nrow(d), 20)] <- NA
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_missing_bias(x, "mood", "stress")
  expect_named(out, c("predictor", "estimate", "std_error", "p_value", "message"))
  expect_equal(out$predictor, "stress")
  expect_true(is.numeric(out$estimate))
  expect_true(is.numeric(out$p_value))
  expect_true(is.character(out$message))
})

test_that("ild_missing_bias with no missingness returns NA and message", {
  d <- ild_simulate(n_id = 5, n_obs_per = 4, seed = 1)
  d$stress <- rnorm(nrow(d))
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_missing_bias(x, "y", "stress")
  expect_true(is.na(out$estimate))
  expect_true(grepl("No missingness", out$message))
})

test_that("ild_missing_bias errors on missing variable names", {
  d <- ild_simulate(n_id = 5, n_obs_per = 4, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  expect_error(ild_missing_bias(x, "nonexistent", "y"), "not found")
  expect_error(ild_missing_bias(x, "y", "nonexistent"), "not found")
})

test_that("ild_design_check returns expected structure", {
  d <- ild_simulate(n_id = 8, n_obs_per = 6, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_design_check(x)
  expect_s3_class(out, "ild_design_check")
  expect_named(out, c("summary", "spacing_class", "spacing", "recommendation", "wp_bp", "missingness"))
  expect_s3_class(out$summary, "tbl_df")
  expect_equal(out$spacing_class, out$spacing$spacing_class)
  expect_true(grepl("AR1|CAR1", out$recommendation))
  expect_null(out$wp_bp)
  expect_null(out$missingness)
})

test_that("ild_design_check with vars includes wp_bp and missingness", {
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 2)
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_design_check(x, vars = "y")
  expect_s3_class(out$wp_bp, "tbl_df")
  expect_true(all(c("variable", "wp_var", "bp_var", "ratio") %in% names(out$wp_bp)))
  expect_true(is.list(out$missingness))
  expect_named(out$missingness, c("summary", "pct_na"))
})

test_that("print.ild_design_check runs without error", {
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 3)
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_design_check(x, vars = "y")
  expect_output(print(out), "Spacing class")
  expect_output(print(out), "Recommendation")
})

test_that("ild_panel_lag_prepare creates lag columns and check table", {
  d <- ild_simulate(n_id = 6, n_obs_per = 8, seed = 902L)
  d$x <- rnorm(nrow(d))
  d$z <- rnorm(nrow(d))
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_panel_lag_prepare(x, c("x", "z"), n = c(1L, 2L), mode = "gap_aware")
  expect_true(all(c("x_lag1", "z_lag2") %in% names(out$data)))
  expect_equal(out$lag_vars, c("x_lag1", "z_lag2"))
  expect_true(nrow(out$check) == 2L)
  expect_true(all(out$check$var %in% out$lag_vars))
})

test_that("ild_panel_lag_prepare recycles n", {
  d <- ild_simulate(n_id = 4, n_obs_per = 6, seed = 903L)
  d$x <- rnorm(nrow(d))
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_panel_lag_prepare(x, c("x", "y"), n = 1L, mode = "index")
  expect_true(all(c("x_lag1", "y_lag1") %in% names(out$data)))
})

test_that("ild_spacing returns expected elements and recommendation", {
  d <- ild_simulate(n_id = 5, n_obs_per = 10, irregular = TRUE, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)
  out <- ild_spacing(x)
  expect_named(out, c("median_interval", "iqr", "large_gaps_pct",
    "coefficient_of_variation", "recommendation", "spacing_class"))
  expect_true(is.numeric(out$median_interval))
  expect_true(is.character(out$recommendation))
  expect_true(out$spacing_class %in% c("regular-ish", "irregular-ish"))
  expect_true(grepl("CAR1|AR1", out$recommendation))
})

test_that("ild_spacing with regular-ish data recommends AR1", {
  d <- ild_simulate(n_id = 5, n_obs_per = 10, irregular = FALSE, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = Inf)
  out <- ild_spacing(x)
  expect_equal(out$spacing_class, "regular-ish")
  expect_true(grepl("AR1", out$recommendation))
})

test_that("ild_lag time_window mode with closest_prior", {
  d <- data.frame(
    id = c(1, 1, 1, 1),
    time = as.POSIXct(c(0, 5, 10, 20), origin = "1970-01-01"),
    x = c(10, 20, 30, 40)
  )
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_lag(x, x, mode = "time_window", window = 15, resolution = "closest_prior")
  # Row 4: time=20, window (5,20); obs at 5,10 in window; closest_prior = 30 (at 10)
  expect_true("x_lag_window" %in% names(out))
  expect_equal(out$x_lag_window[4], 30)
})

test_that("ild_lag time_window mean_in_window", {
  d <- data.frame(
    id = c(1, 1, 1),
    time = as.POSIXct(c(0, 5, 10), origin = "1970-01-01"),
    x = c(10, 20, 30)
  )
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_lag(x, x, mode = "time_window", window = 15, resolution = "mean_in_window")
  # Row 3: time=10, window (-5,10); obs at 0,5 in window; mean = 15
  expect_equal(out$x_lag_window[3], 15)
})

test_that("ild_lag time_window requires window", {
  x <- ild_prepare(ild_simulate(n_id = 2, n_obs_per = 3, seed = 1), id = "id", time = "time")
  expect_error(ild_lag(x, y, mode = "time_window"), "window")
})

test_that("ild_lag time_window accepts lubridate duration", {
  # time in seconds: 0, 3600, 7200. hours(2) = 7200 sec; at 7200, window (0, 7200) has obs at 0, 3600; closest_prior = 20
  d <- data.frame(
    id = 1,
    time = as.POSIXct(c(0, 3600, 7200), origin = "1970-01-01"),
    x = c(10, 20, 30)
  )
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_lag(x, x, mode = "time_window", window = lubridate::hours(2))
  expect_equal(out$x_lag_window[3], 20)
})

test_that("ild_heterogeneity brmsfit matches coef summary structure", {
  skip_if_not_installed("brms")
  skip_on_cran()
  d <- ild_simulate(n_id = 6, n_obs_per = 5, seed = 201)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_brms(
    y ~ y_bp + y_wp + (1 | id),
    data = x,
    chains = 1, iter = 80, warmup = 40,
    refresh = 0, seed = 201, silent = 2,
    backend = "rstan",
    warn_uncentered = FALSE
  )
  h <- ild_heterogeneity(fit)
  expect_s3_class(h, "ild_heterogeneity")
  expect_identical(h$meta$engine, "brms")
  expect_true(nrow(h$random_effects) >= 6L)
  expect_true(all(is.finite(h$random_effects$estimate_total)))
  p <- ild_autoplot(h, type = "caterpillar", term = "Intercept")
  expect_s3_class(p, "ggplot")
})

test_that("ild_fit backend brms matches ild_brms attributes", {
  skip_if_not_installed("brms")
  skip_on_cran()
  d <- ild_simulate(n_id = 5, n_obs_per = 4, seed = 201)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- suppressWarnings(ild_fit(
    y ~ y_bp + y_wp + (1 | id),
    data = x,
    backend = "brms",
    chains = 1,
    iter = 50,
    warmup = 25,
    refresh = 0,
    seed = 201,
    silent = 2,
    warn_uncentered = FALSE
  ))
  expect_s3_class(fit, "brmsfit")
  expect_true(!is.null(attr(fit, "ild_data", exact = TRUE)))
  expect_equal(attr(fit, "ild_posterior", exact = TRUE)$engine, "brms")
})

test_that("ild_fit errors when prior is set for lme4 backend", {
  d <- ild_simulate(n_id = 4, n_obs_per = 3, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  expect_error(
    ild_fit(y ~ (1 | id), data = x, backend = "lme4", prior = "not_null"),
    "prior is only used when backend"
  )
})

test_that("ild_fit errors when prior_template non-default for lme4", {
  d <- ild_simulate(n_id = 4, n_obs_per = 3, seed = 2)
  x <- ild_prepare(d, id = "id", time = "time")
  expect_error(
    ild_fit(y ~ (1 | id), data = x, backend = "lme4", prior_template = "weakly_informative"),
    "prior_template is only used when backend"
  )
})

test_that("ild_fit errors when correlation_class set for brms", {
  skip_if_not_installed("brms")
  d <- ild_simulate(n_id = 4, n_obs_per = 3, seed = 3)
  x <- ild_prepare(d, id = "id", time = "time")
  expect_error(
    ild_fit(y ~ (1 | id), data = x, backend = "brms", correlation_class = "CAR1"),
    "correlation_class is only used when backend"
  )
})

#' Cluster-robust variance-covariance matrix for ILD model fits
#'
#' Computes cluster-robust (sandwich) variance estimators with small-sample
#' corrections via the \pkg{clubSandwich} package. Use with [tidy_ild_model()]
#' via \code{se = "robust"} for fixed-effect inference. Requires
#' \code{attr(fit, "ild_data")}; refit with [ild_lme()] if missing.
#'
#' For fits from [ild_ipw_refit()] with **estimated** IPW/IPCW weights, these SEs
#' apply to the **weighted outcome (mixed) model** only; they do **not** incorporate
#' first-stage uncertainty from the propensity/censoring models unless you treat
#' weights as fixed. For MSM-style inference, prefer [ild_msm_bootstrap()] with an
#' appropriate \code{weight_policy}; see [ild_msm_inference].
#'
#' @param fit A fitted model from [ild_lme()] (lmerMod or lme).
#' @param type Character. Correction type: \code{"CR2"} (recommended), \code{"CR3"}, or \code{"CR0"}.
#' @param cluster Either \code{"id"} (default; cluster by the ILD id column from
#'   \code{ild_data}) or \code{"data"} to use a user-supplied vector via \code{cluster_vec}.
#' @param cluster_vec When \code{cluster = "data"}, a vector of cluster IDs aligned to
#'   the model rows (same length and order as \code{attr(fit, "ild_data")}).
#' @param ... Passed to \code{clubSandwich::vcovCR()}.
#' @return A list with \code{vcov} (matrix), \code{type}, \code{cluster_name},
#'   \code{engine} (\code{"lmer"} or \code{"lme"}), and optionally \code{message}
#'   if a fallback was used (e.g. lme not fully supported on this build).
#' @seealso [tidy_ild_model()] with \code{se = "robust"}, \code{clubSandwich::vcovCR},
#'   [ild_msm_bootstrap()], [ild_msm_inference].
#' @examples
#' if (requireNamespace("clubSandwich", quietly = TRUE)) {
#'   set.seed(1)
#'   dat <- ild_simulate(n_id = 8, n_obs_per = 6, seed = 1)
#'   dat <- ild_prepare(dat, id = "id", time = "time")
#'   dat <- ild_center(dat, y)
#'   fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = dat, ar1 = FALSE, warn_no_ar1 = FALSE)
#'   rv <- ild_robust_se(fit, type = "CR2")
#'   rv$engine
#'   dim(rv$vcov)
#' }
#' @export
ild_robust_se <- function(fit, type = c("CR2", "CR3", "CR0"), cluster = c("id", "data"), cluster_vec = NULL, ...) {
  type <- match.arg(type)
  cluster <- match.arg(cluster)
  rlang::check_installed("clubSandwich", reason = "to use cluster-robust standard errors")
  data <- attr(fit, "ild_data", exact = TRUE)
  if (is.null(data)) {
    stop("Fit was not produced by tidyILD; refit using ild_lme() so the fit carries ild_data.",
         call. = FALSE)
  }
  validate_ild(data)
  engine <- if (inherits(fit, "lme")) "lme" else "lmer"
  cluster_vec <- if (cluster == "id") {
    meta <- ild_meta(data)
    id_name <- meta$ild_id
    if (!id_name %in% names(data)) stop("ild_data must contain id column '", id_name, "'.", call. = FALSE)
    data[[id_name]]
  } else {
    if (is.null(cluster_vec)) stop("When cluster = \"data\", supply cluster_vec.", call. = FALSE)
    if (length(cluster_vec) != nrow(data)) {
      stop("cluster_vec length (", length(cluster_vec), ") must match ild_data rows (", nrow(data), ").", call. = FALSE)
    }
    cluster_vec
  }
  cluster_name <- if (cluster == "id") ild_meta(data)$ild_id else "user"
  out <- tryCatch({
    vc <- clubSandwich::vcovCR(fit, cluster = cluster_vec, type = type, ...)
    list(vcov = vc, type = type, cluster_name = cluster_name, engine = engine)
  }, error = function(e) {
    list(
      vcov = stats::vcov(fit),
      type = type,
      cluster_name = cluster_name,
      engine = engine,
      message = paste0("clubSandwich::vcovCR failed (", conditionMessage(e), "); returning model vcov.")
    )
  })
  if (!is.null(out$message)) {
    warning(out$message, call. = FALSE)
  }
  out
}

#' Time-of-day pattern plot for ILD (circadian-style)
#'
#' Plots a variable by hour of day (or time-of-day) when \code{.ild_time} is
#' POSIXct. Useful for EMA (e.g. mood or activity by hour). Does not add
#' columns to the ILD object; hour is derived internally for plotting.
#'
#' @param x An ILD object (see [is_ild()]) with \code{.ild_time} as POSIXct.
#' @param var Character or symbol. Variable to plot (e.g. mood, activity).
#' @param type Character. \code{"boxplot"} (default) or \code{"line"} (mean per hour with optional SE).
#' @return A ggplot object (variable by hour of day).
#' @export
#' @examples
#' d <- ild_simulate(n_id = 5, n_obs_per = 12, seed = 1)
#' x <- ild_prepare(d, id = "id", time = "time")
#' ild_circadian(x, y)
ild_circadian <- function(x, var, type = c("boxplot", "line")) {
  validate_ild(x)
  type <- match.arg(type)
  var <- as.character(rlang::ensym(var))
  if (!var %in% names(x)) stop("Variable '", var, "' not found in data.", call. = FALSE)
  tcol <- x[[".ild_time"]]
  if (!lubridate::is.POSIXt(tcol)) {
    stop(".ild_time must be POSIXct/POSIXlt for circadian plot. Use ild_plot(type = 'trajectory') for other time axes.",
         call. = FALSE)
  }
  hour <- as.integer(lubridate::hour(tcol))
  df <- data.frame(
    hour = hour,
    value = x[[var]],
    id = x[[".ild_id"]]
  )
  df <- df[!is.na(df$value), ]
  if (nrow(df) == 0) stop("No non-NA values for '", var, "'.", call. = FALSE)
  if (type == "boxplot") {
    p <- ggplot2::ggplot(df, ggplot2::aes(x = factor(.data$hour), y = .data$value)) +
      ggplot2::geom_boxplot(fill = "gray80", outlier.alpha = 0.5) +
      ggplot2::labs(x = "Hour of day", y = var, title = paste0(var, " by hour of day")) +
      ggplot2::theme_minimal()
  } else {
    agg <- dplyr::summarise(
      dplyr::group_by(df, .data$hour),
      mean = mean(.data$value, na.rm = TRUE),
      se = stats::sd(.data$value, na.rm = TRUE) / sqrt(dplyr::n()),
      .groups = "drop"
    )
    p <- ggplot2::ggplot(agg, ggplot2::aes(x = .data$hour, y = .data$mean)) +
      ggplot2::geom_line(linewidth = 1) +
      ggplot2::geom_point() +
      ggplot2::geom_errorbar(ggplot2::aes(ymin = .data$mean - .data$se, ymax = .data$mean + .data$se), width = 0.2) +
      ggplot2::labs(x = "Hour of day", y = paste0("Mean ", var), title = paste0(var, " by hour of day")) +
      ggplot2::theme_minimal()
  }
  p
}

#' Missingness workflow report (orchestration)
#'
#' Bundles person-level compliance, pattern summaries, cohort and hazard tables,
#' optional [ild_missing_model()], heuristic flags, and short text snippets for
#' methods sections. This is **diagnostic and reporting** tooling, not a substitute
#' for formal sensitivity analysis or MNAR models.
#'
#' @family ild_diagnostics_utilities
#' @param x An ILD object (see [is_ild()]).
#' @param outcome Character; outcome column with possible `NA`.
#' @param predictors Optional character vector passed to [ild_missing_model()]
#'   when `fit_missing_model` is `TRUE`. If empty, no missingness model is fit.
#' @param fit_missing_model Logical; fit [ild_missing_model()] when predictors
#'   are non-empty.
#' @param random Logical; passed to [ild_missing_model()] (`glmer` vs `glm`).
#' @param expected_occasions Optional integer for [ild_missing_compliance()].
#' @param max_ids,seed Passed to [ild_missing_pattern()] for large cohorts.
#' @param cohort_plot Logical; include ggplot for [ild_missing_cohort()].
#' @return A list (class `ild_missingness_report`) with:
#'   \describe{
#'     \item{compliance}{Tibble from [ild_missing_compliance()].}
#'     \item{pattern}{Output of [ild_missing_pattern()] (includes compliance
#'       columns on `by_id` if `outcome` was passed through).}
#'     \item{cohort}{List from [ild_missing_cohort()].}
#'     \item{hazard}{Tibble from [ild_missing_hazard_first()].}
#'     \item{flags}{Named list: `dropout_late_pooled` (logical from same
#'       heuristic as guardrail `GR_DROPOUT_LATE_CONCENTRATION`).}
#'     \item{missing_model}{Result of [ild_missing_model()] or `NULL`.}
#'     \item{snippets}{Character vector of short paragraphs for copy-paste.}
#'   }
#' @seealso [ild_missing_pattern()], [ild_missing_bias()], [ild_ipw_weights()],
#'   `vignette("ild-missingness-workflow", package = "tidyILD")`
#' @export
ild_missingness_report <- function(x,
                                   outcome,
                                   predictors = NULL,
                                   fit_missing_model = TRUE,
                                   random = FALSE,
                                   expected_occasions = NULL,
                                   max_ids = NULL,
                                   seed = NULL,
                                   cohort_plot = TRUE) {
  validate_ild(x)
  if (!outcome %in% names(x)) {
    stop("outcome '", outcome, "' not found in data.", call. = FALSE)
  }
  predictors <- if (is.null(predictors)) character(0) else unique(as.character(predictors))
  predictors <- predictors[nzchar(predictors)]
  vars <- unique(c(outcome, predictors))
  compliance <- ild_missing_compliance(x, outcome, expected_occasions = expected_occasions)
  pattern <- ild_missing_pattern(
    x,
    vars = vars,
    outcome = outcome,
    expected_occasions = expected_occasions,
    max_ids = max_ids,
    seed = seed
  )
  cohort <- ild_missing_cohort(x, outcome, plot = isTRUE(cohort_plot))
  hazard <- ild_missing_hazard_first(x, outcome)
  flags <- list(dropout_late_pooled = guardrail_dropout_late_heuristic(x, outcome))
  mm <- NULL
  if (isTRUE(fit_missing_model) && length(predictors) > 0L) {
    mm <- tryCatch(
      ild_missing_model(x, outcome, predictors, random = random),
      error = function(e) NULL
    )
  }
  snippets <- .ild_missingness_snippets(
    outcome = outcome,
    predictors = predictors,
    flags = flags,
    missing_model = mm
  )
  structure(
    list(
      compliance = compliance,
      pattern = pattern,
      cohort = cohort,
      hazard = hazard,
      flags = flags,
      missing_model = mm,
      snippets = snippets
    ),
    class = "ild_missingness_report"
  )
}

#' @keywords internal
#' @noRd
.ild_missingness_snippets <- function(outcome, predictors, flags, missing_model) {
  sn <- c(
    overview = paste0(
      "Outcome ", outcome, " was summarized with tidyILD person-level compliance, ",
      "cohort observed fractions by occasion (.ild_seq), and a discrete-time hazard ",
      "of first missing row (ordinal schedule; see ?ild_missing_hazard_first)."
    ),
    mar_cc = paste0(
      "Under MAR on the outcome, mixed models fitted to all available rows use the ",
      "observed-data likelihood for the outcome conditional on random effects; ",
      "this does not address informative missingness (MNAR). Compare complete-case ",
      "subsets to full-data fits as a coarse sensitivity check."
    ),
    mnar = paste0(
      "If missingness may depend on unobserved outcomes or latent states, treat ",
      "logistic missingness models and IPW as sensitivity analyses, not proof of MAR."
    )
  )
  if (isTRUE(flags$dropout_late_pooled)) {
    sn <- c(
      sn,
      dropout_late = paste0(
        "Pooled outcome missingness is higher in the later half of the (sorted) ",
        "timeline than the earlier half (heuristic). Investigate burden, fatigue, ",
        "or selective dropout."
      )
    )
  }
  if (length(predictors) > 0L) {
    sn <- c(
      sn,
      ipw = paste0(
        "After fitting ild_missing_model(), consider ild_ipw_weights() and ",
        "ild_ipw_refit() for inverse-probability weighting as one sensitivity route ",
        "(see vignette and ?ild_ipw_weights)."
      )
    )
  }
  if (!is.null(missing_model) && !is.null(missing_model$message)) {
    sn <- c(sn, missing_model_note = missing_model$message)
  }
  sn
}

#' Inverse probability of treatment weights (IPTW)
#'
#' Fits a **single pooled** treatment propensity model (logistic regression)
#' over all person-occasions and computes stabilized or unstabilized IPTW weights
#' (one factor per row). Use this for a **time-invariant** treatment or descriptive
#' pooling. For **time-varying** \eqn{A_t} with a sequential MSM estimand, use
#' [ild_iptw_msm_weights()] instead. Outcome missingness IPW is
#' [ild_missing_model()] + [ild_ipw_weights()].
#'
#' **Assumptions:** Positivity (\eqn{0 < P(A|X) < 1}), correct specification of
#' the treatment model. This is sensitivity / MSM-style tooling, not a substitute
#' for careful causal design.
#'
#' @section Bundle integration:
#' Adds \code{.ipw_treat}. Use [ild_joint_msm_weights()] to combine with
#' \code{.ipw_censor} into \code{.ipw} for [ild_ipw_refit()] and [ild_diagnose()].
#'
#' @family ild_diagnostics_utilities
#' @param x An ILD object (see [is_ild()]).
#' @param treatment Character. Name of a **binary** treatment column (numeric
#'   \code{0}/\code{1} or two-level factor).
#' @param predictors Character vector of covariate names in \code{x} (main terms only).
#' @param stabilize Logical. If \code{TRUE} (default), stabilized weights
#'   \eqn{P(A)/P(A|X)} for the observed \eqn{A}; otherwise \eqn{1/P(A|X)}.
#' @param trim Numeric of length 2. Quantiles for trimming (default \code{c(0.01, 0.99)}).
#' @param ... Passed to \code{stats::glm()}.
#' @return \code{x} with added column \code{.ipw_treat}. Attributes include
#'   \code{ild_iptw_fit} (the fitted \code{glm} object).
#' @seealso [ild_iptw_msm_weights()], [ild_ipcw_weights()], [ild_joint_msm_weights()],
#'   [ild_ipw_refit()]
#' @export
#' @examples
#' set.seed(2)
#' d <- ild_simulate(n_id = 20, n_obs_per = 6, seed = 2)
#' d$stress <- rnorm(nrow(d))
#' d$trt <- as.integer(d$stress > 0)
#' x <- ild_prepare(d, id = "id", time = "time")
#' x <- ild_iptw_weights(x, treatment = "trt", predictors = "stress")
#' summary(x$.ipw_treat)
ild_iptw_weights <- function(x, treatment, predictors, stabilize = TRUE,
                             trim = c(0.01, 0.99), ...) {
  validate_ild(x)
  nms <- names(x)
  if (!treatment %in% nms) {
    stop("treatment '", treatment, "' not found in data.", call. = FALSE)
  }
  for (p in predictors) {
    if (!p %in% nms) stop("predictor '", p, "' not found in data.", call. = FALSE)
  }
  Araw <- x[[treatment]]
  if (is.factor(Araw)) {
    if (nlevels(Araw) != 2L) {
      stop("ild_iptw_weights() currently supports binary treatment only (two-level factor).", call. = FALSE)
    }
    A <- as.integer(Araw) - 1L
  } else {
    A <- as.numeric(Araw)
    u <- unique(stats::na.omit(A))
    if (length(u) > 2L || !all(u %in% c(0, 1))) {
      stop("ild_iptw_weights() currently supports binary treatment (0/1 or two-level factor).", call. = FALSE)
    }
  }
  id_col <- ".ild_id"
  pred_cols <- if (length(predictors) > 0L) c(id_col, predictors) else id_col
  df <- as.data.frame(x[, pred_cols, drop = FALSE])
  df[[treatment]] <- A
  rhs <- if (length(predictors) > 0L) paste(predictors, collapse = " + ") else "1"
  form <- stats::as.formula(paste(treatment, "~", rhs))
  fit <- stats::glm(form, data = df, family = stats::binomial(), ...)
  p_hat <- stats::predict(fit, newdata = df, type = "response")
  p_hat <- pmax(pmin(p_hat, 1 - 1e-6), 1e-6)
  p1 <- mean(A, na.rm = TRUE)
  p1 <- pmax(pmin(p1, 1 - 1e-6), 1e-6)
  if (stabilize) {
    w <- ifelse(A >= 0.5, p1 / p_hat, (1 - p1) / (1 - p_hat))
  } else {
    w <- ifelse(A >= 0.5, 1 / p_hat, 1 / (1 - p_hat))
  }
  if (length(trim) >= 2L && is.numeric(trim)) {
    w <- .ild_ipw_trim(w, trim)
  }
  attrs <- attributes(x)
  x$.ipw_treat <- w
  x <- .ild_ipw_restore_attrs(x, attrs)
  attr(x, "ild_iptw_fit") <- fit
  ild_add_step(x, "ild_iptw_weights",
    list(treatment = treatment, predictors = predictors, stabilize = stabilize, trim = trim),
    list(created = ".ipw_treat")
  )
}

# MSM history-builder utilities (on top of ild_lag)

#' Build a declarative MSM history specification
#'
#' Creates a lightweight spec object used by [ild_build_msm_history()] to generate
#' lagged confounder/treatment history columns with deterministic names (e.g.
#' \code{stress_lag1}, \code{trt_lag2}).
#'
#' @param vars Character vector of variable names to lag.
#' @param lags Integer vector of lag orders. Recycled across \code{vars}. Defaults to \code{1L}.
#' @param mode Passed to [ild_lag()] (default \code{"gap_aware"}).
#' @param max_gap Passed to [ild_lag()] for \code{mode = "gap_aware"}.
#' @return Object of class \code{ild_msm_history_spec}.
#' @export
#' @examples
#' s <- ild_msm_history_spec(vars = c("stress", "trt"), lags = 1:2)
#' s
ild_msm_history_spec <- function(vars,
                                 lags = 1L,
                                 mode = c("gap_aware", "index", "time_window"),
                                 max_gap = Inf) {
  mode <- match.arg(mode)
  if (missing(vars) || length(vars) == 0L) {
    stop("vars must include at least one variable name.", call. = FALSE)
  }
  vars <- unique(as.character(vars))
  lags <- as.integer(lags)
  lags <- unique(lags[is.finite(lags) & lags >= 1L])
  if (length(lags) == 0L) {
    stop("lags must contain at least one positive integer.", call. = FALSE)
  }
  out <- list(
    vars = vars,
    lags = sort(lags),
    mode = mode,
    max_gap = max_gap
  )
  class(out) <- c("ild_msm_history_spec", "list")
  out
}

#' @export
print.ild_msm_history_spec <- function(x, ...) {
  cat("MSM history spec\n")
  cat("  vars: ", paste(x$vars, collapse = ", "), "\n", sep = "")
  cat("  lags: ", paste(x$lags, collapse = ", "), "\n", sep = "")
  cat("  mode: ", x$mode, "\n", sep = "")
  if (!is.null(x$max_gap)) cat("  max_gap: ", x$max_gap, "\n", sep = "")
  invisible(x)
}

#' Build MSM lagged history columns from a spec
#'
#' Applies [ild_lag()] repeatedly according to an [ild_msm_history_spec()] and
#' records a manifest under \code{attr(x, "ild_msm_history_manifest")}.
#'
#' @param data An ILD object.
#' @param spec Object from [ild_msm_history_spec()].
#' @param mode Optional override for \code{spec$mode}.
#' @param max_gap Optional override for \code{spec$max_gap}.
#' @return ILD object with lagged columns added; attributes include
#'   \code{ild_msm_history_manifest} (tibble with \code{variable}, \code{lag},
#'   \code{column}).
#' @export
#' @examples
#' d <- ild_simulate(n_id = 6, n_obs_per = 5, seed = 1)
#' d$stress <- rnorm(nrow(d))
#' d$trt <- rbinom(nrow(d), 1, 0.4)
#' d <- ild_prepare(d, id = "id", time = "time")
#' hs <- ild_msm_history_spec(c("stress", "trt"), lags = 1:2)
#' d2 <- ild_build_msm_history(d, hs)
#' attr(d2, "ild_msm_history_manifest")
ild_build_msm_history <- function(data, spec, mode = NULL, max_gap = NULL) {
  validate_ild(data)
  if (!inherits(spec, "ild_msm_history_spec")) {
    stop("spec must be from ild_msm_history_spec().", call. = FALSE)
  }
  mode_use <- if (is.null(mode)) spec$mode else match.arg(mode, c("gap_aware", "index", "time_window"))
  gap_use <- if (is.null(max_gap)) spec$max_gap else max_gap

  out <- data
  rows <- vector("list", length(spec$vars) * length(spec$lags))
  k <- 0L
  for (v in spec$vars) {
    if (!v %in% names(out)) stop("Variable '", v, "' not found in data.", call. = FALSE)
    for (lg in spec$lags) {
      k <- k + 1L
      # Evaluate ild_lag with a symbol to preserve current naming conventions.
      out <- eval(
        bquote(
          ild_lag(.(out), .(as.name(v)), n = .(lg), mode = .(mode_use), max_gap = .(gap_use))
        )
      )
      rows[[k]] <- tibble::tibble(
        variable = v,
        lag = as.integer(lg),
        column = paste0(v, "_lag", as.integer(lg))
      )
    }
  }
  manifest <- dplyr::bind_rows(rows)
  attr(out, "ild_msm_history_manifest") <- manifest
  out
}

test_that("ild_msm_history_spec validates inputs", {
  expect_error(ild_msm_history_spec(vars = character()), "vars")
  s <- ild_msm_history_spec(vars = c("a", "b"), lags = c(1, 2))
  expect_s3_class(s, "ild_msm_history_spec")
  expect_equal(s$lags, c(1L, 2L))
})

test_that("ild_build_msm_history creates deterministic lag columns + manifest", {
  set.seed(2101)
  d <- ild_simulate(n_id = 6, n_obs_per = 5, seed = 2101)
  d$stress <- rnorm(nrow(d))
  d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.4))
  d <- ild_prepare(d, id = "id", time = "time")
  hs <- ild_msm_history_spec(vars = c("stress", "trt"), lags = 1:2)
  x <- ild_build_msm_history(d, hs)
  expect_true(all(c("stress_lag1", "stress_lag2", "trt_lag1", "trt_lag2") %in% names(x)))
  mf <- attr(x, "ild_msm_history_manifest", exact = TRUE)
  expect_s3_class(mf, "tbl_df")
  expect_equal(nrow(mf), 4L)
})

#' Spacing diagnostics and correlation-structure recommendation
#'
#' Reports observation intervals in human-friendly units (e.g. hours) and
#' recommends AR1 vs CAR1 for use in [ild_lme()]. Surfaces the same logic
#' that \code{ild_lme(..., ar1 = TRUE)} uses internally so users can see
#' why a correlation structure was chosen.
#'
#' @param x An ILD object (see [is_ild()]).
#' @param gap_large_hours Numeric. Intervals (in hours) above which to count
#'   as "large gaps" for \code{large_gaps_pct} (default 12). Ignored if time
#'   is not in seconds (e.g. numeric day indices).
#' @return A list with \code{median_interval} (hours), \code{iqr} (hours),
#'   \code{large_gaps_pct} (percent of intervals \code{> gap_large_hours}),
#'   \code{coefficient_of_variation}, \code{recommendation} (character: use CAR1 or AR1),
#'   and \code{spacing_class} (\code{regular-ish} or \code{irregular-ish}).
#' @export
#' @examples
#' d <- ild_simulate(n_id = 5, n_obs_per = 10, irregular = TRUE, seed = 1)
#' x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)
#' ild_spacing(x)
ild_spacing <- function(x, gap_large_hours = 12) {
  validate_ild(x)
  meta <- ild_meta(x)
  spacing <- meta$ild_spacing
  dt <- x[[".ild_dt"]]
  dt <- dt[!is.na(dt)]
  if (length(dt) == 0) {
    return(list(
      median_interval = NA_real_,
      iqr = NA_real_,
      large_gaps_pct = NA_real_,
      coefficient_of_variation = NA_real_,
      recommendation = "Insufficient intervals; cannot recommend.",
      spacing_class = NA_character_
    ))
  }
  median_dt_sec <- if (!is.null(spacing$median_dt)) spacing$median_dt else stats::median(dt)
  iqr_dt_sec <- if (!is.null(spacing$iqr_dt)) spacing$iqr_dt else stats::IQR(dt)
  # Assume .ild_time_num is in seconds (POSIXct); convert to hours for display
  sec_per_hour <- 3600
  median_interval_h <- median_dt_sec / sec_per_hour
  iqr_h <- iqr_dt_sec / sec_per_hour
  # CV: use IQR/median as robust proxy (consistent with ild_spacing_class)
  cv <- if (median_dt_sec <= 0 || is.na(median_dt_sec)) NA_real_ else {
    if (is.na(iqr_dt_sec) || iqr_dt_sec == 0) 0 else iqr_dt_sec / median_dt_sec
  }
  gap_large_sec <- gap_large_hours * sec_per_hour
  large_gaps_pct <- mean(dt > gap_large_sec) * 100
  spacing_class <- ild_spacing_class(x)
  recommendation <- if (spacing_class == "irregular-ish") {
    "Use CAR1 correlation structure."
  } else {
    "Use AR1 correlation structure."
  }
  list(
    median_interval = median_interval_h,
    iqr = iqr_h,
    large_gaps_pct = large_gaps_pct,
    coefficient_of_variation = cv,
    recommendation = recommendation,
    spacing_class = spacing_class
  )
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----eval = FALSE-------------------------------------------------------------
# library(tidyILD)
# 
# d <- ild_msm_simulate_scenario(n_id = 100, n_obs_per = 12, true_ate = 0.5, seed = 101)
# d <- ild_center(d, y)
# 
# hist_spec <- ild_msm_history_spec(vars = c("stress", "trt"), lags = 1:2)
# d <- ild_build_msm_history(d, hist_spec)
# 
# estimand <- ild_msm_estimand(type = "ate", regime = "static", treatment = "trt")
# 
# fit_obj <- ild_msm_fit(
#   estimand = estimand,
#   data = d,
#   outcome_formula = y ~ y_bp + y_wp + stress + trt + (1 | id),
#   history = ~ stress_lag1 + trt_lag1,
#   predictors_censor = "stress",
#   inference = "bootstrap",
#   n_boot = 200,
#   strict_inference = FALSE
# )
# 
# fit_obj
# fit_obj$inference$status
# fit_obj$inference$reason

## ----eval = FALSE-------------------------------------------------------------
# rec <- ild_msm_recovery(
#   n_sim = 100,
#   n_id = 120,
#   n_obs_per = 12,
#   true_ate = 0.5,
#   n_boot = 200,
#   inference = "bootstrap",
#   seed = 1001,
#   censoring = TRUE
# )
# 
# rec$summary
# rec$summary_by_scenario

## ----eval = FALSE-------------------------------------------------------------
# grid <- tibble::tibble(
#   scenario_id = c("baseline", "positivity_stress", "misspecified_treatment"),
#   positivity_stress = c(1, 1.8, 1),
#   misspec_treatment_model = c(FALSE, FALSE, TRUE)
# )
# 
# rec_grid <- ild_msm_recovery(
#   n_sim = 50,
#   n_id = 120,
#   n_obs_per = 12,
#   true_ate = 0.5,
#   n_boot = 200,
#   inference = "bootstrap",
#   scenario_grid = grid,
#   seed = 1101
# )
# 
# rec_grid$summary_by_scenario

#' Tidy and augment ild_lme fits with broom.mixed
#'
#' These S3 methods delegate to [broom.mixed::tidy()] and [broom.mixed::augment()]
#' on the underlying model object so that ild_lme fits work in tidy workflows.
#' Package \pkg{broom.mixed} must be attached (e.g. \code{library(broom.mixed)}).
#'
#' @param x A fitted model from [ild_lme()].
#' @param ... Passed to \code{broom.mixed::tidy()} or \code{broom.mixed::augment()}.
#' @return Same as the corresponding broom.mixed method.
#' @name broom_ild_lme
NULL

#' @rdname broom_ild_lme
#' @export
tidy.ild_lme <- function(x, ...) {
  if (!requireNamespace("broom.mixed", quietly = TRUE)) {
    stop("Package 'broom.mixed' is required for tidy.ild_lme(). Install it and run library(broom.mixed).", call. = FALSE)
  }
  class(x) <- setdiff(class(x), "ild_lme")
  broom.mixed::tidy(x, ...)
}

#' @rdname broom_ild_lme
#' @export
augment.ild_lme <- function(x, ...) {
  if (!requireNamespace("broom.mixed", quietly = TRUE)) {
    stop("Package 'broom.mixed' is required for augment.ild_lme(). Install it and run library(broom.mixed).", call. = FALSE)
  }
  class(x) <- setdiff(class(x), "ild_lme")
  broom.mixed::augment(x, ...)
}

# Autoplot for ild_fit_kfas

#' @rdname ild_autoplot
#' @param type Character: \code{"states"} (smoothed level vs time), \code{"innovations"}
#'   (standardized), or \code{"signal"} (observation expectation vs outcome).
#' @method ild_autoplot ild_fit_kfas
#' @export
ild_autoplot.ild_fit_kfas <- function(x, type = c("states", "innovations", "signal"), ...) {
  rlang::check_installed("KFAS", reason = "to plot ild_kfas fits")
  type <- match.arg(type)
  dat <- attr(x, "ild_data", exact = TRUE)
  if (is.null(dat)) {
    stop("ild_autoplot() for ild_fit_kfas requires attr(fit, \"ild_data\").", call. = FALSE)
  }
  resp <- attr(x, "ild_response", exact = TRUE)
  y <- if (!is.null(resp)) dat[[resp]] else NA
  kfs <- x$kfs
  n <- nrow(dat)
  tidx <- seq_len(n)

  if (type == "states") {
    return(ild_plot_states(x, state_index = 1L, ...))
  }

  if (type == "innovations") {
    r <- tryCatch(stats::rstandard(kfs), error = function(e) NULL)
    if (is.null(r)) {
      stop("Could not extract standardized innovations from KFS output.", call. = FALSE)
    }
    df <- data.frame(time = tidx, r = as.numeric(r)[seq_len(n)], stringsAsFactors = FALSE)
    return(
      ggplot2::ggplot(df, ggplot2::aes(x = .data$time, y = .data$r)) +
        ggplot2::geom_line() +
        ggplot2::geom_hline(yintercept = 0, linetype = 2) +
        ggplot2::labs(
          title = "Standardized one-step-ahead innovations",
          x = "Time index",
          y = "rstandard(KFS)"
        ) +
        ggplot2::theme_minimal()
    )
  }

  # signal
  ex <- ild_kfas_extract_augment(kfs, y, TRUE)
  df <- data.frame(
    time = tidx,
    y = as.numeric(y),
    fitted = ex$fitted,
    stringsAsFactors = FALSE
  )
  ggplot2::ggplot(df, ggplot2::aes(x = .data$time)) +
    ggplot2::geom_line(ggplot2::aes(y = .data$y, colour = "y")) +
    ggplot2::geom_line(ggplot2::aes(y = .data$fitted, colour = "fitted")) +
    ggplot2::labs(
      title = "Outcome vs signal expectation",
      x = "Time index",
      y = NULL,
      colour = NULL
    ) +
    ggplot2::theme_minimal()
}

# S3 methods for ild_brms() / brmsfit objects with attr(fit, "ild_data")

#' @rdname ild_tidy
#' @param intervals Logical. For \code{brmsfit} from [ild_brms()], include
#'   posterior \code{rhat}, \code{ess_bulk}, \code{ess_tail} (default \code{TRUE}).
#' @param object Logical. Passed through for non-brms fits; ignored for \code{brmsfit}.
#' @method ild_tidy brmsfit
#' @export
ild_tidy.brmsfit <- function(x, intervals = TRUE, object = FALSE, ...) {
  if (is.null(attr(x, "ild_data", exact = TRUE))) {
    stop(
      "ild_tidy() for brmsfit requires a model fitted with ild_brms() (ild_data attribute missing).",
      call. = FALSE
    )
  }
  rlang::check_installed("brms")
  sf <- summary(x)$fixed
  if (is.null(sf) || nrow(sf) == 0L) {
    stop("No fixed effects in brms summary.", call. = FALSE)
  }
  term <- rownames(sf)
  n <- length(term)
  estimate <- as.numeric(sf[["Estimate"]])
  std_error <- as.numeric(sf[["Est.Error"]])
  conf_low <- as.numeric(sf[["l-95% CI"]])
  conf_high <- as.numeric(sf[["u-95% CI"]])
  p_val <- rep(NA_real_, n)
  stat <- ifelse(is.finite(std_error) & std_error > 0, estimate / std_error, NA_real_)
  int_type <- rep("quantile", n)
  optional <- list()
  if (isTRUE(intervals)) {
    optional$rhat <- as.numeric(sf[["Rhat"]])
    optional$ess_bulk <- as.numeric(sf[["Bulk_ESS"]])
    optional$ess_tail <- as.numeric(sf[["Tail_ESS"]])
  }
  out <- ild_tidy_assemble(
    term = term,
    estimate = estimate,
    std_error = std_error,
    conf_low = conf_low,
    conf_high = conf_high,
    p_value = p_val,
    statistic = stat,
    interval_type = int_type,
    engine = "brms",
    model_class = "brmsfit",
    component = rep("fixed", n),
    effect_level = NULL,
    optional = optional
  )
  if (isTRUE(object)) {
    return(structure(list(meta = list(engine = "brms"), table = out), class = "tidyild_model"))
  }
  out
}

#' @rdname ild_augment
#' @param summary Logical. For \code{brmsfit}, use posterior predictive summary
#'   for fitted values (default \code{TRUE}).
#' @method ild_augment brmsfit
#' @export
ild_augment.brmsfit <- function(x, summary = TRUE, ...) {
  data <- attr(x, "ild_data", exact = TRUE)
  if (is.null(data)) {
    stop(
      "ild_augment() for brmsfit requires ild_brms() (ild_data attribute missing).",
      call. = FALSE
    )
  }
  rlang::check_installed("brms")
  validate_ild(data)
  mf <- tryCatch(stats::model.frame(x), error = function(e) stats::model.frame(stats::formula(x), data = data))
  y <- stats::model.response(mf)
  n <- nrow(data)
  if (isTRUE(summary)) {
    fi <- stats::fitted(x, summary = TRUE, ...)
    est <- fi[, "Estimate", drop = TRUE]
    lo <- fi[, "Q2.5", drop = TRUE]
    hi <- fi[, "Q97.5", drop = TRUE]
  } else {
    fi <- stats::fitted(x, summary = FALSE, ...)
    est <- apply(fi, 2L, mean)
    lo <- apply(fi, 2L, stats::quantile, probs = 0.025)
    hi <- apply(fi, 2L, stats::quantile, probs = 0.975)
  }
  y_num <- as.numeric(y)
  if (length(y_num) != n) y_num <- rep(NA_real_, n)
  est <- as.numeric(est)
  lo <- as.numeric(lo)
  hi <- as.numeric(hi)
  resid <- y_num - est
  r_std <- ild_augment_pearson_residuals(x, n)
  optional <- list(
    .fitted_lower = lo,
    .fitted_upper = hi
  )
  ild_augment_assemble(
    .ild_id = data[[".ild_id"]],
    .ild_time = data[[".ild_time"]],
    .outcome = y_num,
    .fitted = est,
    .resid = resid,
    .resid_std = r_std,
    engine = "brms",
    model_class = ild_tidy_model_class_string(x),
    optional = optional
  )
}

#' @noRd
ild_brms_ppc_summary <- function(fit, ndraws = 500L) {
  dat <- attr(fit, "ild_data", exact = TRUE)
  mf <- tryCatch(stats::model.frame(fit), error = function(e) NULL)
  if (is.null(mf)) mf <- stats::model.frame(stats::formula(fit), data = dat)
  y <- stats::model.response(mf)
  y <- as.numeric(y)
  ndraws <- max(50L, min(as.integer(ndraws), 2000L))
  yrep <- tryCatch(
    brms::posterior_predict(fit, ndraws = ndraws),
    error = function(e) NULL
  )
  if (is.null(yrep)) {
    return(list(error = "posterior_predict failed"))
  }
  mean_obs <- mean(y, na.rm = TRUE)
  sd_obs <- stats::sd(y, na.rm = TRUE)
  mean_rep <- mean(rowMeans(yrep))
  sd_rep <- mean(apply(yrep, 1L, function(z) stats::sd(as.numeric(z), na.rm = TRUE)))
  list(
    n_draws = ndraws,
    mean_obs = mean_obs,
    sd_obs = sd_obs,
    mean_rep = mean_rep,
    sd_rep = sd_rep,
    mean_abs_diff = abs(mean_obs - mean_rep),
    sd_abs_diff = abs(sd_obs - sd_rep)
  )
}

#' @noRd
ild_brms_diagnose_text <- function(conv, samp, ppc) {
  parts <- character()
  if (!is.null(conv) && nrow(conv) > 0L) {
    mx <- suppressWarnings(max(conv$rhat, na.rm = TRUE))
    mn <- suppressWarnings(min(c(conv$ess_bulk, conv$ess_tail), na.rm = TRUE))
    parts <- c(parts, sprintf(
      "Convergence: max R-hat = %.3f; min ESS (bulk/tail) = %.0f.",
      mx, mn
    ))
  }
  if (!is.null(samp)) {
    parts <- c(parts, sprintf(
      "Sampler: %s chains, iter = %s, warmup = %s, adapt_delta = %s, divergent transitions = %s.",
      samp$chains, samp$iter, samp$warmup,
      if (is.na(samp$adapt_delta)) "NA" else samp$adapt_delta,
      if (is.na(samp$n_divergent)) "NA" else samp$n_divergent
    ))
  }
  if (!is.null(ppc) && is.null(ppc$error)) {
    parts <- c(parts, sprintf(
      "PPC (posterior predictive): |mean_obs - mean_rep| = %.4f, |sd_obs - sd_rep| = %.4f.",
      ppc$mean_abs_diff, ppc$sd_abs_diff
    ))
  }
  paste(parts, collapse = " ")
}

#' @export
print.ild_brms_diagnostics <- function(x, ...) {
  cat("ILD Bayesian diagnostics (legacy brms object; use ild_diagnose() for ild_diagnostics_bundle)\n")
  cat(" ", x$summary_text, "\n", sep = "")
  invisible(x)
}

#' @rdname ild_autoplot
#' @method ild_autoplot brmsfit
#' @export
ild_autoplot.brmsfit <- function(x, type = "pp_check", ...) {
  if (is.null(attr(x, "ild_data", exact = TRUE))) {
    stop("ild_autoplot() for brmsfit requires ild_brms().", call. = FALSE)
  }
  rlang::check_installed("brms")
  if (type == "pp_check") {
    return(brms::pp_check(x, ...))
  }
  if (type %in% c("fitted_vs_actual", "fitted")) {
    return(ild_plot(x, type = "fitted_vs_actual", ...))
  }
  stop("Unknown type for ild_autoplot.brmsfit: use 'pp_check' or 'fitted_vs_actual'.", call. = FALSE)
}

# KFAS-specific guardrails for ild_diagnose.ild_fit_kfas

#' Evaluate KFAS methodological guardrails
#' @keywords internal
#' @noRd
evaluate_guardrails_kfas <- function(object, data, bundle = NULL) {
  rows <- list()
  if (!inherits(object, "ild_fit_kfas")) {
    return(guardrail_finalize_rows(rows))
  }

  spec <- object$spec
  n <- nrow(data)
  sm <- tryCatch(ild_summary(data), error = function(e) NULL)
  sc <- tryCatch(ild_spacing_class(data), error = function(e) NA_character_)

  # GR_KFAS_HIGH_IRREGULARITY_FOR_DISCRETE_TIME
  if (!is.null(sm) && !is.na(sc) && sc == "irregular-ish") {
    med <- sm$summary$median_dt_sec[1]
    iqr <- sm$summary$iqr_dt_sec[1]
    cv <- if (is.finite(med) && med > 0 && is.finite(iqr)) iqr / med else NA_real_
    if (is.finite(cv) && cv > 0.75) {
      rows[[length(rows) + 1L]] <- list(
        rule_id = "GR_KFAS_HIGH_IRREGULARITY_FOR_DISCRETE_TIME",
        triggered = TRUE,
        message = sprintf(
          "IQR/median interval ratio = %.2f on discrete-time indexing; KFAS treats steps as uniform.",
          cv
        ),
        recommendation = NULL
      )
    }
  }

  # GR_KFAS_SHORT_SERIES_FOR_STATE_SPACE — local level needs sufficient n
  state_spec <- spec$state_spec
  min_n <- if (identical(state_spec, "local_level")) 15L else 25L
  if (n < min_n) {
    rows[[length(rows) + 1L]] <- list(
      rule_id = "GR_KFAS_SHORT_SERIES_FOR_STATE_SPACE",
      triggered = TRUE,
      message = sprintf("n = %d time points; recommend at least ~%d for stable %s estimation.", n, min_n, state_spec),
      recommendation = NULL
    )
  }

  # GR_KFAS_STATE_DIMENSION_HIGH_FOR_N
  m_state <- tryCatch(ncol(object$kfs$alphahat), error = function(e) NA_integer_)
  if (is.finite(m_state) && m_state >= 2L && n < 20L * m_state) {
    rows[[length(rows) + 1L]] <- list(
      rule_id = "GR_KFAS_STATE_DIMENSION_HIGH_FOR_N",
      triggered = TRUE,
      message = sprintf(
        "Smoothed state dimension = %d with n = %d; complexity may exceed information in the series.",
        m_state, n
      ),
      recommendation = NULL
    )
  }

  # GR_KFAS_DEGENERATE_VARIANCE_ESTIMATE
  fm <- object$kfas_model
  qv <- ild_kfas_qh_scalar(fm$Q)
  hv <- ild_kfas_qh_scalar(fm$H)
  eps <- 1e-8
  if ((is.finite(qv) && qv < eps) || (is.finite(hv) && hv < eps)) {
    rows[[length(rows) + 1L]] <- list(
      rule_id = "GR_KFAS_DEGENERATE_VARIANCE_ESTIMATE",
      triggered = TRUE,
      message = sprintf(
        "Estimated variances Q = %g, H = %g may be on the boundary.",
        qv, hv
      ),
      recommendation = NULL
    )
  }

  # GR_KFAS_NONCONVERGENCE
  opt <- object$fit_ssm$optim.out
  if (!is.null(opt)) {
    conv <- opt$convergence
    if (!is.null(conv) && !isTRUE(conv == 0L)) {
      rows[[length(rows) + 1L]] <- list(
        rule_id = "GR_KFAS_NONCONVERGENCE",
        triggered = TRUE,
        message = sprintf("Optimizer convergence code %s (see fit$fit_ssm$optim.out).", conv),
        recommendation = NULL
      )
    }
  }

  # GR_KFAS_MANY_MISSING_OUTCOME_SEGMENTS
  pre <- object$preprocessing
  if (!is.null(pre) && is.list(pre)) {
    nseg <- pre$n_na_segments
    pct <- pre$pct_outcome_na
    if (is.finite(nseg) && nseg >= 3L && is.finite(pct) && pct > 0.1) {
      rows[[length(rows) + 1L]] <- list(
        rule_id = "GR_KFAS_MANY_MISSING_OUTCOME_SEGMENTS",
        triggered = TRUE,
        message = sprintf(
          "%d separate NA segments in the outcome (%.0f%% NA before omitting rows).",
          nseg, 100 * pct
        ),
        recommendation = NULL
      )
    }
  }

  # GR_KFAS_UNMODELED_BETWEEN_PERSON_HETEROGENEITY
  fc <- spec$fit_context
  if (identical(fc, "independent_series_per_id")) {
    rows[[length(rows) + 1L]] <- list(
      rule_id = "GR_KFAS_UNMODELED_BETWEEN_PERSON_HETEROGENEITY",
      triggered = TRUE,
      message = NULL,
      recommendation = NULL
    )
  }

  guardrail_finalize_rows(rows)
}

#' Simulation-based power analysis for a fixed effect in ILD models
#'
#' Estimates empirical power by repeatedly simulating data with a known effect
#' (via [ild_simulate()] plus one added predictor), fitting with [ild_lme()],
#' and counting the proportion of runs where the target term is significant
#' (Wald p < alpha). The workflow (simulate, fit, reject/retain) mirrors
#' simulation-based power in packages like mixpower; \code{ild_power()} is
#' focused on ILD and \code{ild_lme()}. For multi-parameter grids, LRT, or
#' general LMMs, consider mixpower.
#'
#' The data-generating process adds one predictor (name from \code{test_term})
#' as standard normal and adds \code{effect_size * predictor} to the outcome
#' on top of the base [ild_simulate()] DGP (id, time, y). No change to
#' \code{ild_simulate()} is required.
#'
#' For \code{ar1 = FALSE} (lmer), the lme4 backend does not report p-values;
#' inference for the test term uses a Wald z-approximation (estimate / SE)
#' so that power is still computed. For \code{ar1 = TRUE} (nlme), p-values
#' come from the model summary.
#'
#' @param formula Fixed-effects formula including the predictor to power for
#'   and random effects, e.g. \code{y ~ x + (1 | id)}. For \code{ar1 = TRUE},
#'   use a fixed-only formula (random passed internally).
#' @param n_sim Integer. Number of simulation replications (default 500).
#' @param n_id Integer. Number of persons per replication.
#' @param n_obs_per Integer. Observations per person per replication.
#' @param effect_size Numeric. True coefficient for \code{test_term} in the DGP.
#' @param test_term Character or \code{NULL}. Which fixed-effect term to test.
#'   If \code{NULL}, taken as the first non-intercept fixed-effect term from
#'   the model (inferred from the formula).
#' @param alpha Numeric. Significance level for rejection (default 0.05).
#' @param ar1 Logical. If \code{TRUE}, fit with nlme and residual AR1/CAR1
#'   (default \code{FALSE}).
#' @param seed Integer. Base random seed; replication \code{i} uses \code{seed + i}.
#' @param return_sims Logical. If \code{TRUE}, include a tibble of per-run
#'   estimate, std_error, p_value, rejected in the result (default \code{FALSE}).
#' @param verbose Logical. If \code{TRUE}, message progress (default \code{TRUE}).
#' @param ... Passed to [ild_simulate()] (e.g. \code{irregular}, \code{wp_effect},
#'   \code{bp_effect}) and to [ild_lme()].
#' @return A list: \code{power} (proportion of converged runs with p < alpha),
#'   \code{n_sim}, \code{n_reject}, \code{n_converged}, \code{n_failed},
#'   \code{alpha}, \code{test_term}. If \code{return_sims = TRUE}, also
#'   \code{sim_results} (tibble of per-run results).
#' @seealso [ild_recovery_metrics()] to summarize bias, RMSE, and coverage from
#'   \code{return_sims}. \href{../doc/benchmark-simulation-recovery.html}{\code{vignette("benchmark-simulation-recovery", package = "tidyILD")}}
#'   (worked example).
#' @examples
#' set.seed(42)
#' res <- ild_power(
#'   formula = y ~ x + (1 | id),
#'   n_sim = 25L,
#'   n_id = 15L,
#'   n_obs_per = 10L,
#'   effect_size = 0.3,
#'   seed = 42L,
#'   verbose = FALSE
#' )
#' res$power
#' res$n_reject
#' @importFrom stats rnorm
#' @export
ild_power <- function(formula,
                     n_sim = 500L,
                     n_id,
                     n_obs_per,
                     effect_size,
                     test_term = NULL,
                     alpha = 0.05,
                     ar1 = FALSE,
                     seed = 42L,
                     return_sims = FALSE,
                     verbose = TRUE,
                     ...) {
  formula <- stats::as.formula(formula)
  n_sim <- as.integer(n_sim)[1]
  if (missing(n_id)) stop("n_id is required.", call. = FALSE)
  if (missing(n_obs_per)) stop("n_obs_per is required.", call. = FALSE)
  n_id <- as.integer(n_id)[1]
  n_obs_per <- as.integer(n_obs_per)[1]
  if (missing(effect_size)) stop("effect_size is required.", call. = FALSE)
  effect_size <- as.numeric(effect_size)[1]
  alpha <- as.numeric(alpha)[1]
  seed <- as.integer(seed)[1]

  # Resolve test_term: first non-intercept fixed-effect variable from formula
  all_v <- all.vars(formula)
  if (length(all_v) < 2L) stop("formula must include at least one predictor.", call. = FALSE)
  rhs_vars <- all_v[-1L]
  # Exclude grouping factor (id) so that y ~ 1 + (1|id) has no fixed predictor
  fixed_vars <- setdiff(rhs_vars, c("id"))
  if (length(fixed_vars) == 0L) stop("formula must include at least one predictor.", call. = FALSE)
  if (is.null(test_term)) {
    test_term <- fixed_vars[1L]
  }
  test_term <- as.character(test_term)[1L]

  # ... is passed only to ild_simulate (irregular, wp_effect, bp_effect, etc.)
  dots <- list(...)
  sim_args <- c(list(n_id = n_id, n_obs_per = n_obs_per), dots)

  n_reject <- 0L
  n_failed <- 0L
  sim_results <- if (return_sims) {
    list(estimate = numeric(n_sim), std_error = numeric(n_sim),
         p_value = numeric(n_sim), rejected = logical(n_sim), converged = logical(n_sim))
  } else {
    NULL
  }

  for (i in seq_len(n_sim)) {
    set.seed(seed + i)
    # Generate base data (ild_simulate does not add the predictor)
    sim_args$seed <- seed + i
    d <- do.call(ild_simulate, sim_args)
    # Add predictor and true effect
    d[[test_term]] <- stats::rnorm(nrow(d))
    d$y <- d$y + effect_size * d[[test_term]]
    # Prepare and fit
    prep <- tryCatch(
      ild_prepare(d, id = "id", time = "time"),
      error = function(e) NULL
    )
    if (is.null(prep)) {
      n_failed <- n_failed + 1L
      if (return_sims) {
        sim_results$estimate[i] <- NA_real_
        sim_results$std_error[i] <- NA_real_
        sim_results$p_value[i] <- NA_real_
        sim_results$rejected[i] <- FALSE
        sim_results$converged[i] <- FALSE
      }
      next
    }
    fit <- tryCatch(
      suppressWarnings(ild_lme(
        formula, prep,
        ar1 = ar1,
        warn_no_ar1 = FALSE,
        warn_uncentered = FALSE
      )),
      error = function(e) NULL
    )
    if (is.null(fit)) {
      n_failed <- n_failed + 1L
      if (return_sims) {
        sim_results$estimate[i] <- NA_real_
        sim_results$std_error[i] <- NA_real_
        sim_results$p_value[i] <- NA_real_
        sim_results$rejected[i] <- FALSE
        sim_results$converged[i] <- FALSE
      }
      next
    }
    tbl <- tryCatch(tidy_ild_model(fit, object = FALSE), error = function(e) NULL)
    if (is.null(tbl)) {
      n_failed <- n_failed + 1L
      if (return_sims) {
        sim_results$estimate[i] <- NA_real_
        sim_results$std_error[i] <- NA_real_
        sim_results$p_value[i] <- NA_real_
        sim_results$rejected[i] <- FALSE
        sim_results$converged[i] <- FALSE
      }
      next
    }
    row <- tbl[tbl$term == test_term, ]
    if (nrow(row) != 1L) {
      n_failed <- n_failed + 1L
      if (return_sims) {
        sim_results$estimate[i] <- NA_real_
        sim_results$std_error[i] <- NA_real_
        sim_results$p_value[i] <- NA_real_
        sim_results$rejected[i] <- FALSE
        sim_results$converged[i] <- FALSE
      }
      next
    }
    pval <- row$p_value[1L]
    if (is.na(pval)) {
      # lme4 does not report p-values; use Wald z-approximation for power
      est <- row$estimate[1L]
      se <- row$std_error[1L]
      if (!is.na(se) && se > 0) pval <- 2 * (1 - stats::pnorm(abs(est / se)))
    }
    rej <- !is.na(pval) && pval < alpha
    if (rej) n_reject <- n_reject + 1L
    if (return_sims) {
      sim_results$estimate[i] <- row$estimate[1L]
      sim_results$std_error[i] <- row$std_error[1L]
      sim_results$p_value[i] <- pval
      sim_results$rejected[i] <- rej
      sim_results$converged[i] <- TRUE
    }
    if (verbose && (i %% 100L == 0L)) message("ild_power: ", i, "/", n_sim, " replications.")
  }

  n_converged <- n_sim - n_failed
  power <- if (n_converged > 0L) n_reject / n_converged else NA_real_

  out <- list(
    power = power,
    n_sim = n_sim,
    n_reject = n_reject,
    n_converged = n_converged,
    n_failed = n_failed,
    alpha = alpha,
    test_term = test_term
  )
  if (return_sims && !is.null(sim_results)) {
    out$sim_results <- tibble::tibble(
      sim = seq_len(n_sim),
      estimate = sim_results$estimate,
      std_error = sim_results$std_error,
      p_value = sim_results$p_value,
      rejected = sim_results$rejected,
      converged = sim_results$converged
    )
  }
  attr(out, "ild_provenance") <- ild_new_analysis_provenance(NULL, "ild_power", list(
    formula = deparse(formula),
    n_sim = n_sim,
    n_id = n_id,
    n_obs_per = n_obs_per,
    effect_size = effect_size,
    test_term = test_term,
    alpha = alpha,
    ar1 = ar1
  ), list(n_reject = n_reject, n_converged = n_converged, n_failed = n_failed, power = power))
  out
}

test_that("ild_power returns expected structure", {
  skip_on_cran()
  res <- ild_power(
    formula = y ~ x + (1 | id),
    n_sim = 20L,
    n_id = 15L,
    n_obs_per = 10L,
    effect_size = 0.3,
    seed = 1L,
    verbose = FALSE
  )
  expect_named(res, c("power", "n_sim", "n_reject", "n_converged", "n_failed", "alpha", "test_term"))
  expect_equal(res$n_sim, 20L)
  expect_equal(res$test_term, "x")
  expect_true(is.numeric(res$power))
  expect_true(res$power >= 0 && res$power <= 1)
  expect_true(res$n_reject >= 0L && res$n_reject <= res$n_sim)
  expect_equal(res$n_converged + res$n_failed, res$n_sim)
  expect_equal(res$alpha, 0.05)
})

test_that("ild_power with effect_size 0 gives type I rate near alpha", {
  skip_on_cran()
  res <- ild_power(
    formula = y ~ x + (1 | id),
    n_sim = 100L,
    n_id = 20L,
    n_obs_per = 12L,
    effect_size = 0,
    alpha = 0.05,
    seed = 42L,
    verbose = FALSE
  )
  # Type I rate should be roughly alpha; allow wide margin for simulation noise
  expect_true(res$power >= 0 && res$power <= 0.25)
})

test_that("ild_power with large effect_size gives high power", {
  skip_on_cran()
  res <- ild_power(
    formula = y ~ x + (1 | id),
    n_sim = 50L,
    n_id = 40L,
    n_obs_per = 20L,
    effect_size = 1.5,
    seed = 123L,
    verbose = FALSE
  )
  expect_true(res$power >= 0.7)
})

test_that("ild_power with test_term explicitly set", {
  skip_on_cran()
  res <- ild_power(
    formula = y ~ x + (1 | id),
    n_sim = 15L,
    n_id = 10L,
    n_obs_per = 8L,
    effect_size = 0.4,
    test_term = "x",
    seed = 2L,
    verbose = FALSE
  )
  expect_equal(res$test_term, "x")
})

test_that("ild_power errors on missing required args", {
  expect_error(
    ild_power(formula = y ~ x + (1 | id), n_sim = 10L, n_id = 5L, n_obs_per = 5L),
    "effect_size"
  )
  expect_error(
    ild_power(formula = y ~ x + (1 | id), n_sim = 10L, effect_size = 0.5, n_obs_per = 5L),
    "n_id"
  )
  expect_error(
    ild_power(formula = y ~ x + (1 | id), n_sim = 10L, n_id = 5L, effect_size = 0.5),
    "n_obs_per"
  )
})

test_that("ild_power errors on formula without predictor", {
  expect_error(
    ild_power(formula = y ~ 1 + (1 | id), n_sim = 5L, n_id = 5L, n_obs_per = 5L, effect_size = 0.5),
    "at least one predictor"
  )
})

test_that("ild_power with return_sims includes sim_results", {
  skip_on_cran()
  res <- ild_power(
    formula = y ~ x + (1 | id),
    n_sim = 10L,
    n_id = 8L,
    n_obs_per = 6L,
    effect_size = 0.2,
    seed = 99L,
    return_sims = TRUE,
    verbose = FALSE
  )
  expect_true("sim_results" %in% names(res))
  expect_s3_class(res$sim_results, "tbl_df")
  expect_equal(nrow(res$sim_results), 10L)
  expect_true(all(c("sim", "estimate", "std_error", "p_value", "rejected", "converged") %in% names(res$sim_results)))
})

test_that("ild_power with ar1 = TRUE runs nlme path", {
  skip_on_cran()
  res <- ild_power(
    formula = y ~ x,
    n_sim = 8L,
    n_id = 10L,
    n_obs_per = 6L,
    effect_size = 0.5,
    ar1 = TRUE,
    seed = 77L,
    verbose = FALSE
  )
  expect_named(res, c("power", "n_sim", "n_reject", "n_converged", "n_failed", "alpha", "test_term"))
  expect_equal(res$test_term, "x")
  expect_true(is.numeric(res$power))
})

# MSM estimand-first runner (v1.1 scaffold)

#' Fit an MSM-style weighted mixed model from an estimand spec
#'
#' High-level runner that takes an [ild_msm_estimand()] plus ILD data, builds
#' treatment (and optional censoring) weights, refits a weighted \code{lmer},
#' and optionally computes inference via robust SEs or cluster bootstrap.
#'
#' v1.1 keeps the v1 workflow and adds explicit capability signaling for
#' inference/regime paths that are degraded or unsupported.
#'
#' @param estimand Object from [ild_msm_estimand()].
#' @param data ILD data.
#' @param outcome_formula Outcome model formula for [ild_lme()] (typically includes treatment term).
#' @param history Predictors for treatment process weights:
#'   one-sided formula (e.g. \code{~ stress_lag1 + trt_lag1}) or character vector.
#' @param history_spec Optional [ild_msm_history_spec()] to build lagged history before weighting.
#' @param treatment_engine \code{"sequential_msm"} (default) or \code{"pooled"}.
#' @param predictors_censor Optional character vector for IPCW model. If supplied, joint weights
#'   are built with [ild_joint_msm_weights()] and stored in \code{weights_col}.
#' @param weights_col Column used by [ild_ipw_refit()] (default \code{".ipw"}).
#' @param stabilize_treat Treatment-weight stabilization mode. For sequential MSM:
#'   \code{"marginal"}, \code{"prior_treatment"}, \code{"none"}. For pooled IPTW:
#'   mapped to logical \code{stabilize}.
#' @param prior_treatment Required when \code{stabilize_treat = "prior_treatment"} in sequential MSM.
#' @param stabilize_joint Joint-weight stabilization for [ild_joint_msm_weights()].
#' @param trim Quantile trimming applied during weight construction.
#' @param inference \code{"none"} (default), \code{"robust"}, or \code{"bootstrap"}.
#' @param robust_type Passed to [ild_robust_se()] and [tidy_ild_model()] when \code{inference = "robust"}.
#' @param n_boot Passed to [ild_msm_bootstrap()] when \code{inference = "bootstrap"}.
#' @param weight_policy Passed to [ild_msm_bootstrap()] when \code{inference = "bootstrap"}.
#' @param weights_fn Passed to [ild_msm_bootstrap()] when \code{inference = "bootstrap"} and
#'   \code{weight_policy = "reestimate_weights"}.
#' @param seed Optional seed for bootstrap inference.
#' @param strict_inference Logical. If \code{TRUE}, stop when requested inference or
#'   regime dispatch cannot be delivered as requested. If \code{FALSE} (default),
#'   degrade with warning and record machine-readable status.
#' @param ... Passed to [ild_lme()].
#' @return Object of class \code{ild_msm_fit} with elements:
#'   \code{estimand}, \code{history_spec}, \code{weights_data}, \code{fit}, \code{inference},
#'   \code{treatment_engine}, \code{weights_col}. Analysis provenance is attached.
#' @export
#' @examples
#' set.seed(31)
#' d <- ild_simulate(n_id = 12, n_obs_per = 6, seed = 31)
#' d$stress <- rnorm(nrow(d))
#' d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
#' d <- ild_prepare(d, id = "id", time = "time")
#' d <- ild_center(d, y)
#' d <- ild_lag(d, stress, max_gap = Inf)
#' d <- ild_lag(d, trt, max_gap = Inf)
#' est <- ild_msm_estimand(treatment = "trt")
#' res <- ild_msm_fit(
#'   estimand = est,
#'   data = d,
#'   outcome_formula = y ~ y_bp + y_wp + stress + trt + (1 | id),
#'   history = ~ stress_lag1 + trt_lag1,
#'   predictors_censor = "stress",
#'   inference = "none",
#'   warn_no_ar1 = FALSE,
#'   warn_uncentered = FALSE
#' )
#' res$fit
ild_msm_fit <- function(estimand,
                        data,
                        outcome_formula,
                        history,
                        history_spec = NULL,
                        treatment_engine = c("sequential_msm", "pooled"),
                        predictors_censor = NULL,
                        weights_col = ".ipw",
                        stabilize_treat = c("marginal", "prior_treatment", "none"),
                        prior_treatment = NULL,
                        stabilize_joint = c("mean1", "none"),
                        trim = c(0.01, 0.99),
                        inference = c("none", "robust", "bootstrap"),
                        robust_type = c("CR2", "CR3", "CR0"),
                        n_boot = 200L,
                        weight_policy = c("fixed_weights", "reestimate_weights"),
                        weights_fn = NULL,
                        seed = NULL,
                        strict_inference = FALSE,
                        ...) {
  validate_ild_msm_estimand(estimand)
  validate_ild(data)
  treatment_engine <- match.arg(treatment_engine)
  stabilize_treat <- match.arg(stabilize_treat)
  stabilize_joint <- match.arg(stabilize_joint)
  inference <- match.arg(inference)
  robust_type <- match.arg(robust_type)
  weight_policy <- match.arg(weight_policy)
  strict_inference <- isTRUE(strict_inference)

  if (!estimand$treatment %in% names(data)) {
    stop("Treatment column '", estimand$treatment, "' not found in data.", call. = FALSE)
  }
  if (!estimand$time_var %in% names(data)) {
    stop("time_var '", estimand$time_var, "' not found in data.", call. = FALSE)
  }
  if (missing(outcome_formula)) {
    stop("outcome_formula is required.", call. = FALSE)
  }
  if (missing(history)) {
    stop("history is required (formula or character vector).", call. = FALSE)
  }

  x <- data
  regime_status <- list(
    status = "ok",
    reason = "none",
    message = "Regime dispatch completed."
  )
  if (identical(estimand$regime, "dynamic")) {
    x$.msm_regime_assignment <- .ild_msm_eval_dynamic_rule(x, estimand)
    regime_status <- list(
      status = "degraded",
      reason = "dynamic_regime_weighting_not_implemented",
      message = "Dynamic regime assignment was evaluated, but weighting still follows observed treatment in v1.1."
    )
    if (strict_inference) {
      stop(regime_status$message, call. = FALSE)
    }
    warning(regime_status$message, call. = FALSE)
  }
  if (!identical(estimand$type, "ate")) {
    msg_type <- "Estimand type is accepted by schema but not fully implemented in runner; proceeding with ATE-compatible pathway."
    if (strict_inference) stop(msg_type, call. = FALSE)
    warning(msg_type, call. = FALSE)
  }
  if (!is.null(history_spec)) {
    x <- ild_build_msm_history(x, history_spec)
  }

  if (treatment_engine == "sequential_msm") {
    x <- ild_iptw_msm_weights(
      x,
      treatment = estimand$treatment,
      history = history,
      time_var = estimand$time_var,
      stabilize = stabilize_treat,
      prior_treatment = prior_treatment,
      trim = trim
    )
  } else {
    pred <- .ild_history_to_predictors(history)
    if (length(pred) == 0L) {
      stop("Pooled IPTW requires at least one predictor in history.", call. = FALSE)
    }
    x <- ild_iptw_weights(
      x,
      treatment = estimand$treatment,
      predictors = pred,
      stabilize = !identical(stabilize_treat, "none"),
      trim = trim
    )
  }

  if (!is.null(predictors_censor) && length(predictors_censor) > 0L) {
    x <- ild_ipcw_weights(x, predictors = predictors_censor, trim = trim)
    x <- ild_joint_msm_weights(x, stabilize = stabilize_joint, trim = trim, joint_name = weights_col)
  } else {
    if (!".ipw_treat" %in% names(x)) stop("Expected .ipw_treat after treatment weighting.", call. = FALSE)
    attrs <- attributes(x)
    x[[weights_col]] <- x$.ipw_treat
    x <- .ild_ipw_restore_attrs(x, attrs)
  }

  base_fit <- ild_lme(outcome_formula, data = x, ar1 = FALSE, ...)
  fitw <- ild_ipw_refit(base_fit, data = x, weights = weights_col)

  inf <- .ild_msm_inference_shell(
    requested = inference,
    strict_inference = strict_inference
  )
  if (inference == "robust") {
    rb_try <- tryCatch(
      {
        vc <- ild_robust_se(fitw, type = robust_type)
        td <- tidy_ild_model(fitw, se = "robust", robust_type = robust_type)
        list(ok = TRUE, vcov = vc, tidy = td, error = NULL)
      },
      error = function(e) list(ok = FALSE, vcov = NULL, tidy = NULL, error = e)
    )
    if (isTRUE(rb_try$ok)) {
      inf$status <- "ok"
      inf$reason <- "none"
      inf$message <- "Robust cluster-robust inference delivered."
      inf$method <- "robust"
      inf$raw$robust <- list(vcov = rb_try$vcov, tidy = rb_try$tidy)
      inf$summary <- .ild_msm_inference_summary(rb_try$tidy, method = "robust")
    } else {
      msg <- paste0(
        "Robust inference unavailable for this weighted lmer path: ",
        conditionMessage(rb_try$error),
        ". Falling back to model-based Wald intervals."
      )
      if (strict_inference) stop(msg, call. = FALSE)
      warning(msg, call. = FALSE)
      td <- tidy_ild_model(fitw, se = "model")
      inf$status <- "degraded"
      inf$reason <- "robust_weighted_lmer_not_supported"
      inf$message <- msg
      inf$method <- "model_fallback"
      inf$raw$robust <- list(vcov = NULL, tidy = NULL, error = conditionMessage(rb_try$error))
      inf$summary <- .ild_msm_inference_summary(td, method = "model_fallback")
    }
  } else if (inference == "bootstrap") {
    bres <- ild_msm_bootstrap(
      fit = fitw,
      n_boot = n_boot,
      weight_policy = weight_policy,
      weights_fn = weights_fn,
      seed = seed
    )
    inf$method <- "bootstrap"
    inf$raw$bootstrap <- list(result = bres)
    inf$summary <- .ild_msm_inference_summary(tidy_ild_msm_bootstrap(bres), method = "bootstrap")
    if (is.null(bres$n_success) || bres$n_success <= 0L) {
      msg <- "Bootstrap inference did not produce any successful replicates."
      if (strict_inference) stop(msg, call. = FALSE)
      warning(msg, call. = FALSE)
      inf$status <- "unsupported"
      inf$reason <- "bootstrap_no_successful_replicates"
      inf$message <- msg
    } else if (bres$n_success < bres$n_boot) {
      msg <- paste0(
        "Bootstrap inference delivered with partial failures (",
        bres$n_success, "/", bres$n_boot, " successful replicates)."
      )
      if (strict_inference) stop(msg, call. = FALSE)
      warning(msg, call. = FALSE)
      inf$status <- "degraded"
      inf$reason <- "bootstrap_partial_failure"
      inf$message <- msg
    } else {
      inf$status <- "ok"
      inf$reason <- "none"
      inf$message <- "Bootstrap inference delivered."
    }
  } else {
    inf$status <- "ok"
    inf$reason <- "none"
    inf$message <- "Inference not requested."
    inf$method <- "none"
    inf$summary <- tibble::tibble(
      term = character(),
      terms = character(),
      estimate = numeric(),
      std_error = numeric(),
      conf_low = numeric(),
      conf_high = numeric(),
      interval_type = character(),
      method = character()
    )
  }

  out <- list(
    estimand = estimand,
    history_spec = history_spec,
    weights_data = x,
    fit = fitw,
    inference = inf,
    regime_status = regime_status,
    treatment_engine = treatment_engine,
    weights_col = weights_col
  )
  class(out) <- c("ild_msm_fit", "list")
  attr(out, "ild_provenance") <- ild_new_analysis_provenance(
    fitw,
    "ild_msm_fit",
    list(
      estimand = estimand,
      treatment_engine = treatment_engine,
      weights_col = weights_col,
      inference = inference,
      inference_status = inf$status,
      inference_reason = inf$reason,
      regime_status = regime_status$status,
      regime_reason = regime_status$reason,
      strict_inference = strict_inference
    ),
    list(n_obs = nrow(x), n_id = length(unique(x[[".ild_id"]])))
  )
  out
}

#' @export
print.ild_msm_fit <- function(x, ...) {
  cat("tidyILD MSM estimand fit\n")
  cat("  estimand: ", x$estimand$type, " (", x$estimand$regime, ")\n", sep = "")
  cat("  treatment: ", x$estimand$treatment, "\n", sep = "")
  cat("  treatment_engine: ", x$treatment_engine, "\n", sep = "")
  cat("  weights_col: ", x$weights_col, "\n", sep = "")
  if (!is.null(x$inference)) {
    cat("  inference: ", x$inference$method, " [", x$inference$status, "]\n", sep = "")
  }
  if (!is.null(x$regime_status)) {
    cat("  regime_status: ", x$regime_status$status, "\n", sep = "")
  }
  invisible(x)
}

#' @keywords internal
#' @noRd
.ild_history_to_predictors <- function(history) {
  if (inherits(history, "formula")) {
    return(tryCatch(attr(stats::terms(history), "term.labels"), error = function(e) character(0)))
  }
  as.character(history)
}

#' Diagnose an \code{ild_msm_fit} result in one call
#'
#' Convenience bridge to [ild_diagnose()] using \code{object$fit} and
#' \code{object$weights_data}.
#'
#' @param object Output from [ild_msm_fit()].
#' @param ... Passed to [ild_diagnose()].
#' @return An \code{\link{ild_diagnostics_bundle}}.
#' @export
ild_msm_diagnose <- function(object, ...) {
  if (!inherits(object, "ild_msm_fit")) {
    stop("object must be from ild_msm_fit().", call. = FALSE)
  }
  ild_diagnose(object$fit, data = object$weights_data, ...)
}

#' @keywords internal
#' @noRd
.ild_msm_inference_shell <- function(requested, strict_inference) {
  list(
    requested = requested,
    strict_inference = strict_inference,
    status = "unsupported",
    reason = "pending",
    message = "Inference path not evaluated.",
    method = requested,
    summary = NULL,
    raw = list()
  )
}

#' @keywords internal
#' @noRd
.ild_msm_inference_summary <- function(tidy_tbl, method) {
  req <- c("term", "estimate", "std_error", "conf_low", "conf_high", "interval_type")
  miss <- setdiff(req, names(tidy_tbl))
  if (length(miss) > 0L) {
    stop("Inference tidy table missing required columns: ", paste(miss, collapse = ", "), call. = FALSE)
  }
  out <- tidy_tbl[, req, drop = FALSE]
  out$terms <- out$term
  out$method <- method
  tibble::as_tibble(out)[, c("term", "terms", "estimate", "std_error", "conf_low", "conf_high", "interval_type", "method")]
}

#' @keywords internal
#' @noRd
.ild_msm_eval_dynamic_rule <- function(data, estimand) {
  rs <- estimand$regime_spec
  if (is.null(rs) || !identical(rs$class, "dynamic")) {
    stop("estimand$regime_spec must describe a dynamic regime.", call. = FALSE)
  }
  rule <- rs$rule
  trt <- estimand$treatment
  out <- NULL
  if (is.character(rule)) {
    if (identical(rule, "always_treat")) out <- rep(1L, nrow(data))
    if (identical(rule, "never_treat")) out <- rep(0L, nrow(data))
    if (identical(rule, "as_observed")) out <- as.integer(data[[trt]])
  } else if (is.function(rule)) {
    out <- rule(data)
  }
  if (is.null(out) || length(out) != nrow(data)) {
    stop("Dynamic regime rule must return length nrow(data).", call. = FALSE)
  }
  out <- as.integer(out)
  bad <- !is.finite(out) | !out %in% c(0L, 1L)
  if (any(bad)) {
    stop("Dynamic regime assignment must be binary 0/1 for all rows.", call. = FALSE)
  }
  out
}

test_that("ild_msm_estimand builds backward-compatible static ATE object", {
  e <- ild_msm_estimand(treatment = "trt")
  expect_s3_class(e, "ild_msm_estimand")
  expect_equal(e$type, "ate")
  expect_equal(e$regime, "static")
  expect_true(is.list(e$contrast))
  expect_identical(e$target_time, "all")
})

test_that("ild_msm_estimand accepts structured regime specification", {
  e <- ild_msm_estimand(
    type = "att",
    treatment = "trt",
    regime = list(class = "dynamic", rule = "always_treat"),
    target_time = c(3, 5),
    contrast = list(label = "treated_vs_control", treated = 1, control = 0)
  )
  expect_equal(e$type, "att")
  expect_equal(e$regime, "dynamic")
  expect_true(is.list(e$regime_spec))
  expect_equal(e$regime_spec$class, "dynamic")
  expect_equal(e$target_time, c(3, 5))
})

test_that("ild_msm_fit returns structured object with provenance", {
  d <- msm_fixture_data(seed = 2201, n_id = 12, n_obs_per = 6)
  e <- ild_msm_estimand(treatment = "trt")
  fit_obj <- ild_msm_fit(
    estimand = e,
    data = d,
    outcome_formula = y ~ y_bp + y_wp + stress + trt + (1 | id),
    history = ~ stress_lag1 + trt_lag1,
    predictors_censor = "stress",
    inference = "none",
    warn_no_ar1 = FALSE,
    warn_uncentered = FALSE
  )
  expect_s3_class(fit_obj, "ild_msm_fit")
  expect_true(inherits(fit_obj$fit, "lmerMod"))
  expect_true(is_ild(fit_obj$weights_data))
  expect_true(".ipw" %in% names(fit_obj$weights_data))
  expect_true(is.list(fit_obj$inference))
  expect_equal(fit_obj$inference$status, "ok")
  pv <- attr(fit_obj, "ild_provenance", exact = TRUE)
  expect_true(is.list(pv))
  expect_equal(pv$object_type, "ild_msm_fit")
})

test_that("ild_msm_fit supports robust and bootstrap inference modes", {
  d <- msm_fixture_data(seed = 2202, n_id = 10, n_obs_per = 5)
  e <- ild_msm_estimand(treatment = "trt")

  r1 <- ild_msm_fit(
    estimand = e,
    data = d,
    outcome_formula = y ~ y_bp + y_wp + stress + trt + (1 | id),
    history = ~ stress_lag1 + trt_lag1,
    inference = "robust",
    robust_type = "CR2",
    warn_no_ar1 = FALSE,
    warn_uncentered = FALSE
  )
  expect_true(r1$inference$method %in% c("robust", "model_fallback"))
  expect_true(r1$inference$status %in% c("ok", "degraded"))
  expect_s3_class(r1$inference$summary, "tbl_df")
  expect_true(all(c("term", "terms", "estimate", "std_error", "conf_low", "conf_high", "interval_type", "method") %in% names(r1$inference$summary)))

  r2 <- ild_msm_fit(
    estimand = e,
    data = d,
    outcome_formula = y ~ y_bp + y_wp + stress + trt + (1 | id),
    history = ~ stress_lag1 + trt_lag1,
    inference = "bootstrap",
    n_boot = 6L,
    seed = 1,
    warn_no_ar1 = FALSE,
    warn_uncentered = FALSE
  )
  expect_equal(r2$inference$method, "bootstrap")
  expect_equal(r2$inference$status, "ok")
  expect_s3_class(r2$inference$raw$bootstrap$result, "ild_msm_bootstrap")
})

test_that("dynamic regime is scaffolded with explicit degraded status", {
  d <- msm_fixture_data(seed = 2203, n_id = 10, n_obs_per = 5)
  e_dyn <- ild_msm_estimand(
    treatment = "trt",
    regime = "dynamic",
    dynamic_rule = "always_treat",
    target_time = "final",
    contrast = list(label = "always_vs_never")
  )
  expect_warning(
    fit_dyn <- ild_msm_fit(
      estimand = e_dyn,
      data = d,
      outcome_formula = y ~ y_bp + y_wp + stress + trt + (1 | id),
      history = ~ stress_lag1 + trt_lag1,
      inference = "none",
      strict_inference = FALSE,
      warn_no_ar1 = FALSE,
      warn_uncentered = FALSE
    ),
    "Dynamic regime assignment"
  )
  expect_equal(fit_dyn$regime_status$status, "degraded")
  expect_true(".msm_regime_assignment" %in% names(fit_dyn$weights_data))
  expect_error(
    ild_msm_fit(
      estimand = e_dyn,
      data = d,
      outcome_formula = y ~ y_bp + y_wp + stress + trt + (1 | id),
      history = ~ stress_lag1 + trt_lag1,
      inference = "none",
      strict_inference = TRUE,
      warn_no_ar1 = FALSE,
      warn_uncentered = FALSE
    )
  )
})

test_that("ild_msm_contrast_over_time returns time-indexed contrasts", {
  d <- msm_fixture_data(seed = 2204, n_id = 10, n_obs_per = 5)
  e <- ild_msm_estimand(treatment = "trt")
  fit_obj <- ild_msm_fit(
    estimand = e,
    data = d,
    outcome_formula = y ~ y_bp + y_wp + stress + trt + (1 | id),
    history = ~ stress_lag1 + trt_lag1,
    inference = "none",
    warn_no_ar1 = FALSE,
    warn_uncentered = FALSE
  )
  ct <- ild_msm_contrast_over_time(fit_obj, time_var = ".ild_seq")
  expect_s3_class(ct, "tbl_df")
  expect_true(all(c("term", "estimate", "std_error", "conf_low", "conf_high", "interval_type", "method") %in% names(ct)))
  expect_true(nrow(ct) >= 1L)
  dg <- ild_msm_diagnose(fit_obj, balance = FALSE)
  expect_s3_class(dg, "ild_diagnostics_bundle")
})

# Internal column names (reserved; never required from user)
.ILD_COLS <- c(
  ".ild_id",
  ".ild_time",
  ".ild_time_num",
  ".ild_seq",
  ".ild_dt",
  ".ild_gap"
)

# Attribute names set by ild_prepare(), read via ild_meta()
.ILD_ATTRS <- c(
  "ild_id",
  "ild_time",
  "ild_gap_threshold",
  "ild_n_units",
  "ild_n_obs",
  "ild_spacing"
)

# Optional primary metadata bundle; when present, ild_meta() can read from it
.TIDYILD_ATTR <- "tidyILD"

#' Check if an object is a valid ILD tibble
#'
#' Returns TRUE if the object has all required `.ild_*` columns and
#' `ild_*` metadata attributes (as set by [ild_prepare()]).
#'
#' @param x Any object.
#' @return Logical.
#' @export
is_ild <- function(x) {
  if (!inherits(x, "data.frame")) return(FALSE)
  nms <- names(x)
  has_cols <- all(.ILD_COLS %in% nms)
  if (!has_cols) return(FALSE)
  attrs <- .ILD_ATTRS
  for (a in attrs) {
    if (is.null(attr(x, a, exact = TRUE))) return(FALSE)
  }
  TRUE
}

#' Upgrade legacy ILD object to tidyILD schema (internal)
#'
#' If \code{attr(x, "tidyILD")} is missing but \code{ild_*} attributes exist,
#' builds the tidyILD bundle and sets it; adds \code{tidyild_df} to class.
#' Emits a one-time warning. Never modifies user columns.
#' @param x An object that may be a legacy ILD tibble.
#' @return \code{x} (possibly with \code{attr(x, "tidyILD")} and class updated).
#' @noRd
ild_normalize_internal <- function(x) {
  if (!inherits(x, "data.frame")) return(x)
  if (!is.null(attr(x, .TIDYILD_ATTR, exact = TRUE))) return(x)
  ild_id <- attr(x, "ild_id", exact = TRUE)
  ild_time <- attr(x, "ild_time", exact = TRUE)
  ild_gap_threshold <- attr(x, "ild_gap_threshold", exact = TRUE)
  ild_spacing <- attr(x, "ild_spacing", exact = TRUE)
  if (is.null(ild_id) || is.null(ild_time) || is.null(ild_gap_threshold)) return(x)
  pv <- tryCatch(as.character(utils::packageVersion("tidyILD")), error = function(e) "0.0.0")
  attr(x, .TIDYILD_ATTR) <- list(
    id_col = ild_id,
    time_col = ild_time,
    tz = "UTC",
    gap_threshold = ild_gap_threshold,
    created = NA,
    spacing = ild_spacing,
    provenance = list(version = pv, schema_version = "1", object_type = "ild_data", steps = list())
  )
  if (!inherits(x, "tidyild_df")) {
    class(x) <- c("tidyild_df", "ild_tbl", setdiff(class(x), c("tidyild_df", "ild_tbl")))
  }
  if (is.null(getOption("tidyILD.schema_upgraded", NULL))) {
    warning(
      "ILD internal schema upgraded: attr(x, \"tidyILD\") and class \"tidyild_df\" added. ",
      "This warning is shown once per session.",
      call. = FALSE
    )
    options(tidyILD.schema_upgraded = TRUE)
  }
  x
}

#' Validate an ILD object and error if invalid
#'
#' Checks presence and types of `.ild_*` columns and `ild_*` attributes.
#' Errors with a clear message if anything is missing or invalid.
#' Calls \code{ild_normalize_internal()} so legacy objects get \code{attr(x, "tidyILD")} and class \code{tidyild_df}.
#'
#' @param x Object to validate (expected to be an ILD tibble).
#' @return Invisibly returns \code{x} if valid.
#' @export
validate_ild <- function(x) {
  ild_normalize_internal(x)
  if (!inherits(x, "data.frame")) {
    stop("ILD object must be a data frame or tibble.", call. = FALSE)
  }
  nms <- names(x)
  missing_cols <- setdiff(.ILD_COLS, nms)
  if (length(missing_cols) > 0) {
    stop(
      "ILD object is missing required columns: ",
      paste(missing_cols, collapse = ", "),
      ". Run ild_prepare() first.",
      call. = FALSE
    )
  }
  for (a in .ILD_ATTRS) {
    if (is.null(attr(x, a, exact = TRUE))) {
      stop("ILD object is missing attribute: ", a, ". Run ild_prepare() first.", call. = FALSE)
    }
  }
  # Basic type checks
  if (!is.numeric(x[[".ild_time_num"]])) {
    stop(".ild_time_num must be numeric.", call. = FALSE)
  }
  seq_vals <- x[[".ild_seq"]]
  if (!is.numeric(seq_vals) || any(is.na(seq_vals)) || any(seq_vals != as.integer(seq_vals))) {
    stop(".ild_seq must be integer-valued.", call. = FALSE)
  }
  if (!is.logical(x[[".ild_gap"]]) && !all(is.na(x[[".ild_gap"]]))) {
    stop(".ild_gap must be logical (or NA for first row per person).", call. = FALSE)
  }
  invisible(x)
}

#' Get ILD metadata attributes
#'
#' Returns the metadata attributes set by [ild_prepare()]: user-facing
#' id/time column names, gap threshold, n_units, n_obs, and spacing
#' (descriptive stats only).
#'
#' @param x An ILD object (see [is_ild()]).
#' @return A named list of metadata (ild_id, ild_time, ild_gap_threshold,
#'   ild_n_units, ild_n_obs, ild_spacing). \code{ild_spacing} includes overall
#'   stats and may contain \code{by_id}, a tibble of per-person spacing stats.
#' @export
ild_meta <- function(x) {
  validate_ild(x)
  out <- list(
    ild_id            = attr(x, "ild_id", exact = TRUE),
    ild_time          = attr(x, "ild_time", exact = TRUE),
    ild_gap_threshold = attr(x, "ild_gap_threshold", exact = TRUE),
    ild_n_units       = attr(x, "ild_n_units", exact = TRUE),
    ild_n_obs         = attr(x, "ild_n_obs", exact = TRUE),
    ild_spacing       = attr(x, "ild_spacing", exact = TRUE)
  )
  out
}

#' Internal constructor for ILD tibbles
#'
#' Builds the tibble with all `.ild_*` columns and sets all `ild_*`
#' attributes plus \code{attr(x, "tidyILD")}. Only used by ild_prepare(); not exported.
#'
#' @param data A data frame or tibble (already sorted, with user id/time and
#'   value columns plus the computed .ild_* columns).
#' @param ild_id Character; user-facing id column name.
#' @param ild_time Character; user-facing time column name.
#' @param gap_threshold Numeric; threshold used for .ild_gap.
#' @param spacing List; descriptive spacing stats (median, IQR, etc.).
#' @param tz Character; time zone for the bundle (default \code{"UTC"}).
#' @param tsibble_meta Optional list from \code{ild_tsibble_capture_meta()} when
#'   \code{data} was a \code{tbl_ts}; stored under \code{attr(x, "tidyILD")$tsibble}.
#' @return A tibble with class \code{tidyild_df}, \code{ild_tbl}, and ILD attributes.
#' @noRd
new_ild_df <- function(data, ild_id, ild_time, gap_threshold, spacing, tz = "UTC", tsibble_meta = NULL) {
  stopifnot(
    is.character(ild_id), length(ild_id) == 1,
    is.character(ild_time), length(ild_time) == 1,
    is.numeric(gap_threshold), length(gap_threshold) == 1,
    is.list(spacing)
  )
  n_units <- length(unique(data[[".ild_id"]]))
  n_obs   <- nrow(data)
  x <- tibble::as_tibble(data)
  attr(x, "ild_id")            <- ild_id
  attr(x, "ild_time")          <- ild_time
  attr(x, "ild_gap_threshold") <- gap_threshold
  attr(x, "ild_n_units")       <- n_units
  attr(x, "ild_n_obs")         <- n_obs
  attr(x, "ild_spacing")       <- spacing
  pv <- tryCatch(as.character(utils::packageVersion("tidyILD")), error = function(e) "0.0.0")
  tidy_bundle <- list(
    id_col = ild_id,
    time_col = ild_time,
    tz = tz,
    gap_threshold = gap_threshold,
    created = Sys.time(),
    spacing = spacing,
    provenance = list(version = pv, schema_version = "1", object_type = "ild_data", steps = list())
  )
  if (!is.null(tsibble_meta)) {
    tidy_bundle$tsibble <- tsibble_meta
  }
  attr(x, .TIDYILD_ATTR) <- tidy_bundle
  class(x) <- c("tidyild_df", "ild_tbl", class(x))
  x
}

#' Coerce to ILD object
#'
#' If the object already has the required `.ild_*` columns and
#' attributes, validates and returns it (with \code{tidyild_df} and \code{ild_tbl} class if missing).
#' Otherwise errors.
#'
#' @param x A data frame or tibble that may already be ILD-shaped.
#' @return An ILD tibble with class \code{tidyild_df} and \code{ild_tbl}.
#' @export
as_ild <- function(x) {
  validate_ild(x)
  if (!inherits(x, "tidyild_df")) {
    class(x) <- c("tidyild_df", "ild_tbl", setdiff(class(x), c("tidyild_df", "ild_tbl")))
  }
  if (!inherits(x, "ild_tbl")) {
    class(x) <- c("ild_tbl", class(x))
  }
  x
}

#' Copy ILD attributes and class from one object to another (internal)
#' @noRd
restore_ild_attrs <- function(from, to) {
  for (a in .ILD_ATTRS) attr(to, a) <- attr(from, a, exact = TRUE)
  tidyild <- attr(from, .TIDYILD_ATTR, exact = TRUE)
  if (!is.null(tidyild)) attr(to, .TIDYILD_ATTR) <- tidyild
  class(to) <- class(from)
  to
}

# Satisfy R CMD check (tidy eval and rlang)
if (getRversion() >= "2.15.1") {
  utils::globalVariables(c(".data", ".ild_time_num", ":=", "y"))
}

test_that("ild_center adds _bp and _wp columns", {
  d <- data.frame(
    id = c(1, 1, 1, 2, 2, 2),
    time = as.POSIXct(rep(0:2, 2), origin = "1970-01-01"),
    x = c(10, 12, 14, 20, 22, 24)
  )
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_center(x, x)
  expect_true("x_bp" %in% names(out))
  expect_true("x_wp" %in% names(out))
  # Person 1 mean = 12, person 2 mean = 22
  expect_equal(unique(out$x_bp[out$.ild_id == 1]), 12)
  expect_equal(unique(out$x_bp[out$.ild_id == 2]), 22)
  expect_equal(out$x_wp, c(-2, 0, 2, -2, 0, 2))
})

test_that("ild_center preserves ILD attributes", {
  d <- data.frame(id = c(1, 1, 2, 2), time = as.POSIXct(0:3, origin = "1970-01-01"), x = 1:4)
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_center(x, x)
  expect_true(is_ild(out))
  expect_equal(attr(out, "ild_n_units"), 2)
})

test_that("ild_center is idempotent for same variable", {
  d <- data.frame(id = c(1, 1, 2, 2), time = as.POSIXct(0:3, origin = "1970-01-01"), x = c(1, 2, 3, 4))
  x <- ild_prepare(d, id = "id", time = "time")
  out1 <- ild_center(x, x)
  out2 <- ild_center(out1, x)
  expect_equal(out1$x_bp, out2$x_bp)
  expect_equal(out1$x_wp, out2$x_wp)
})

test_that("ild_center grand_mean adds _gm and _wp_gm", {
  d <- data.frame(id = c(1, 1, 2, 2), time = as.POSIXct(0:3, origin = "1970-01-01"), x = c(1, 2, 3, 4))
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_center(x, x, type = "grand_mean")
  expect_true("x_gm" %in% names(out))
  expect_true("x_wp_gm" %in% names(out))
  expect_equal(unique(out$x_gm), 2.5)
})

test_that("ild_center fails on non-numeric variable", {
  d <- data.frame(id = c(1, 1, 2, 2), time = as.POSIXct(0:3, origin = "1970-01-01"), x = c("a", "b", "c", "d"))
  x <- ild_prepare(d, id = "id", time = "time")
  expect_error(ild_center(x, x), "not numeric")
})

test_that("ild_center naming prefix produces BP_var and WP_var", {
  d <- data.frame(id = c(1, 1, 2, 2), time = as.POSIXct(0:3, origin = "1970-01-01"), x = c(10, 12, 20, 22))
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_center(x, x, naming = "prefix")
  expect_true("BP_x" %in% names(out))
  expect_true("WP_x" %in% names(out))
  expect_equal(unique(out$BP_x[out$.ild_id == 1]), 11)
})

test_that("is_ild returns FALSE for non-data frames", {
  expect_false(is_ild(1))
  expect_false(is_ild("a"))
  expect_false(is_ild(NULL))
})

test_that("is_ild returns FALSE when required columns missing", {
  df <- data.frame(.ild_id = 1, .ild_time = Sys.time(), .ild_time_num = 1)
  expect_false(is_ild(df))
})

test_that("validate_ild errors on non-data frame", {
  expect_error(validate_ild(1), "data frame or tibble")
})

test_that("validate_ild errors when required columns missing", {
  df <- data.frame(a = 1)
  expect_error(validate_ild(df), "missing required columns")
})

test_that("ild_meta returns list with expected names", {
  # Use ild_prepare to get a valid ILD object
  d <- data.frame(
    id = c(1, 1, 2, 2),
    time = as.POSIXct(c(0, 1, 0, 1), origin = "1970-01-01"),
    x = 1:4
  )
  x <- ild_prepare(d, id = "id", time = "time")
  meta <- ild_meta(x)
  expect_type(meta, "list")
  expect_named(meta, c("ild_id", "ild_time", "ild_gap_threshold", "ild_n_units", "ild_n_obs", "ild_spacing"))
  expect_equal(meta$ild_id, "id")
  expect_equal(meta$ild_time, "time")
  expect_equal(meta$ild_n_units, 2)
  expect_equal(meta$ild_n_obs, 4)
})

test_that("as_ild validates and adds tidyild_df and ild_tbl class", {
  d <- data.frame(
    id = c(1, 1, 2, 2),
    time = as.POSIXct(c(0, 1, 0, 1), origin = "1970-01-01"),
    x = 1:4
  )
  x <- ild_prepare(d, id = "id", time = "time")
  y <- as_ild(x)
  expect_s3_class(y, "tidyild_df")
  expect_s3_class(y, "ild_tbl")
})

test_that("ild_prepare creates all .ild_* columns", {
  d <- data.frame(
    id = c(1, 1, 2, 2),
    time = as.POSIXct(c(0, 10, 0, 10), origin = "1970-01-01"),
    x = 1:4
  )
  x <- ild_prepare(d, id = "id", time = "time")
  expect_true(is_ild(x))
  expect_named(x, c("id", "time", "x", ".ild_id", ".ild_time", ".ild_time_num", ".ild_seq", ".ild_dt", ".ild_gap"))
  expect_equal(x[[".ild_seq"]], c(1, 2, 1, 2))
  expect_equal(x[[".ild_dt"]], c(NA, 10, NA, 10))
})

test_that("ild_prepare sets attr tidyILD and class tidyild_df", {
  d <- data.frame(id = c(1, 1, 2, 2), time = as.POSIXct(c(0, 10, 0, 10), origin = "1970-01-01"), x = 1:4)
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 5)
  expect_true(inherits(x, "tidyild_df"))
  expect_true(inherits(x, "ild_tbl"))
  tidyild <- attr(x, "tidyILD", exact = TRUE)
  expect_false(is.null(tidyild))
  expect_equal(tidyild$id_col, "id")
  expect_equal(tidyild$time_col, "time")
  expect_equal(tidyild$gap_threshold, 5)
  expect_true(!is.null(tidyild$created))
  expect_true(!is.null(tidyild$spacing))
})

test_that("ild_prepare .ild_seq is 1:n_i per id and .ild_dt equals diff of .ild_time_num", {
  d <- data.frame(id = c(1, 1, 1, 2, 2), time = as.POSIXct(c(0, 5, 10, 0, 5), origin = "1970-01-01"), x = 1:5)
  x <- ild_prepare(d, id = "id", time = "time")
  expect_equal(x[[".ild_seq"]], c(1, 2, 3, 1, 2))
  dt_expected <- c(NA, 5, 5, NA, 5)
  expect_equal(x[[".ild_dt"]], dt_expected)
})

test_that("ild_prepare sorts by id then time", {
  d <- data.frame(
    id = c(2, 1, 1, 2),
    time = as.POSIXct(c(5, 0, 10, 0), origin = "1970-01-01"),
    x = 1:4
  )
  x <- ild_prepare(d, id = "id", time = "time")
  expect_equal(x[[".ild_id"]], c(1, 1, 2, 2))
  expect_equal(x[[".ild_time_num"]], c(0, 10, 0, 5))
})

test_that("ild_prepare gap_threshold sets .ild_gap", {
  d <- data.frame(
    id = c(1, 1, 1),
    time = as.POSIXct(c(0, 5, 100), origin = "1970-01-01"),
    x = 1:3
  )
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 10)
  expect_equal(x[[".ild_gap"]], c(NA, FALSE, TRUE))
})

test_that("ild_prepare duplicate_handling first keeps first row", {
  d <- data.frame(
    id = c(1, 1, 1),
    time = as.POSIXct(c(0, 0, 10), origin = "1970-01-01"),
    x = c(10, 20, 30)
  )
  x <- ild_prepare(d, id = "id", time = "time", duplicate_handling = "first")
  expect_equal(nrow(x), 2)
  expect_equal(x$x, c(10, 30))
})

test_that("ild_prepare duplicate_handling last keeps last row", {
  d <- data.frame(
    id = c(1, 1, 1),
    time = as.POSIXct(c(0, 0, 10), origin = "1970-01-01"),
    x = c(10, 20, 30)
  )
  x <- ild_prepare(d, id = "id", time = "time", duplicate_handling = "last")
  expect_equal(nrow(x), 2)
  expect_equal(x$x, c(20, 30))
})

test_that("ild_prepare duplicate_handling error stops on duplicates", {
  d <- data.frame(
    id = c(1, 1),
    time = as.POSIXct(c(0, 0), origin = "1970-01-01"),
    x = 1:2
  )
  expect_error(ild_prepare(d, id = "id", time = "time", duplicate_handling = "error"), "Duplicate")
})

test_that("ild_prepare with one observation per person", {
  d <- data.frame(id = 1:3, time = as.POSIXct(0:2, origin = "1970-01-01"), x = 1:3)
  x <- ild_prepare(d, id = "id", time = "time")
  expect_equal(x[[".ild_seq"]], c(1, 1, 1))
  expect_true(all(is.na(x[[".ild_dt"]])))
  expect_true(all(is.na(x[[".ild_gap"]])))
})

test_that("ild_prepare spacing is descriptive list", {
  d <- data.frame(
    id = c(1, 1, 2, 2),
    time = as.POSIXct(c(0, 10, 0, 10), origin = "1970-01-01"),
    x = 1:4
  )
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 5)
  spacing <- attr(x, "ild_spacing")
  expect_type(spacing, "list")
  expect_true("median_dt" %in% names(spacing))
  expect_true("iqr_dt" %in% names(spacing))
  expect_true("pct_gap" %in% names(spacing))
  expect_true("by_id" %in% names(spacing))
  expect_s3_class(spacing$by_id, "tbl_df")
  expect_named(spacing$by_id, c("id", "median_dt", "iqr_dt", "n_intervals", "pct_gap"))
})

test_that("ild_prepare spacing by_id has one row per person and correct stats", {
  # id 1: times 0, 10, 20 -> intervals 10, 10; median=10, iqr=0. id 2: times 0, 5, 15 -> intervals 5, 10; median=7.5.
  # gap_threshold=8: id1 intervals 10,10 both >8 so pct_gap=100; id2 intervals 5,10 so one >8, pct_gap=50.
  d <- data.frame(
    id = c(1, 1, 1, 2, 2, 2),
    time = as.POSIXct(c(0, 10, 20, 0, 5, 15), origin = "1970-01-01"),
    x = 1:6
  )
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 8)
  by_id <- attr(x, "ild_spacing")$by_id
  expect_equal(nrow(by_id), 2)
  by_id <- by_id[order(by_id$id), ]
  expect_equal(by_id$id, c(1, 2))
  expect_equal(by_id$median_dt, c(10, 7.5))
  expect_equal(by_id$n_intervals, c(2L, 2L))
  expect_equal(by_id$pct_gap, c(100, 50))
})

test_that("ild_prepare requires both id and time unless data is tbl_ts", {
  d <- data.frame(id = 1, time = 1, x = 1)
  expect_error(ild_prepare(d), regexp = "tsibble|Provide both")
  expect_error(ild_prepare(d, id = "id"), "both id and time")
})

test_that("ild_prepare duplicate_handling collapse aggregates with collapse_fn", {
  d <- data.frame(
    id = c(1, 1, 1),
    time = as.POSIXct(c(0, 0, 10), origin = "1970-01-01"),
    x = c(10, 20, 30)
  )
  x <- ild_prepare(d, id = "id", time = "time", duplicate_handling = "collapse", collapse_fn = list(x = mean))
  expect_equal(nrow(x), 2)
  expect_equal(x$x[x$.ild_time_num == 0], 15)
  expect_equal(x$x[x$.ild_time_num == 10], 30)
})

test_that("ild_lag index mode equals previous row within person", {
  d <- data.frame(
    id = c(1, 1, 1, 2, 2),
    time = as.POSIXct(c(0, 10, 20, 0, 10), origin = "1970-01-01"),
    x = c(10, 20, 30, 40, 50)
  )
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_lag(x, x, n = 1, mode = "index")
  expect_equal(out$x_lag1, c(NA, 10, 20, NA, 40))
})

test_that("ild_lag no cross-person contamination", {
  d <- data.frame(
    id = c(1, 1, 2, 2),
    time = as.POSIXct(c(0, 10, 0, 10), origin = "1970-01-01"),
    x = c(100, 200, 300, 400)
  )
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_lag(x, x, mode = "index")
  expect_true(is.na(out$x_lag1[1]))
  expect_equal(out$x_lag1[2], 100)
  expect_true(is.na(out$x_lag1[3]))
  expect_equal(out$x_lag1[4], 300)
})

test_that("ild_lag gap_aware sets NA when interval exceeds max_gap", {
  d <- data.frame(
    id = c(1, 1, 1, 1),
    time = as.POSIXct(c(0, 5, 10, 100), origin = "1970-01-01"),
    x = c(1, 2, 3, 4)
  )
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_lag(x, x, n = 1, mode = "gap_aware", max_gap = 20)
  # .ild_dt: NA, 5, 5, 90. So lag at row 4 would be 3, but dt[4]=90 > 20 -> NA
  expect_equal(out$x_lag1, c(NA, 1, 2, NA))
})

test_that("ild_lag gap_aware keeps value when interval within max_gap", {
  d <- data.frame(
    id = c(1, 1, 1),
    time = as.POSIXct(c(0, 5, 10), origin = "1970-01-01"),
    x = c(1, 2, 3)
  )
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_lag(x, x, n = 1, mode = "gap_aware", max_gap = 20)
  expect_equal(out$x_lag1, c(NA, 1, 2))
})

test_that("ild_lag preserves ILD attributes", {
  d <- data.frame(id = c(1, 1, 2, 2), time = as.POSIXct(0:3, origin = "1970-01-01"), x = 1:4)
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_lag(x, x, mode = "index")
  expect_true(is_ild(out))
})

test_that("ild_lag gap_aware uses max_gap from metadata when max_gap is NULL", {
  d <- data.frame(
    id = c(1, 1, 1, 1),
    time = as.POSIXct(c(0, 5, 10, 100), origin = "1970-01-01"),
    x = c(1, 2, 3, 4)
  )
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 20)
  out <- ild_lag(x, x, n = 1, mode = "gap_aware", max_gap = NULL)
  expect_equal(out$x_lag1, c(NA, 1, 2, NA))
})

test_that("ild_lag n=2 lags by two rows", {
  d <- data.frame(
    id = c(1, 1, 1, 1),
    time = as.POSIXct(0:3, origin = "1970-01-01"),
    x = c(10, 20, 30, 40)
  )
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_lag(x, x, n = 2, mode = "index")
  expect_equal(out$x_lag2, c(NA, NA, 10, 20))
})

#' Prepare a data frame as an ILD (intensive longitudinal data) object
#'
#' Validates and encodes longitudinal structure: parses time, sorts by id and
#' time, handles duplicate timestamps, and adds internal columns (`.ild_*`)
#' and metadata. All downstream functions assume the result of `ild_prepare()`.
#'
#' @param data A data frame or tibble with at least an id and a time column.
#'   If \code{id} and \code{time} are both omitted, \code{data} must be a
#'   \code{tbl_ts} from \pkg{tsibble} (requires a single key column and an index);
#'   key and index names are inferred (see [ild_as_tsibble()] for the reverse).
#'
#'   \strong{Tsibble interoperability:} Accepting a \code{tbl_ts} is an input
#'   convenience; the result is always a plain ILD tibble (the \code{tbl_ts} class
#'   is dropped). Metadata from the source tsibble (key, index, interval summary,
#'   regularity) is stored in \code{attr(x, "tidyILD")$tsibble} when available;
#'   use [ild_tsibble_meta()] to read it. Use [ild_as_tsibble()] for a
#'   best-effort round-trip. \strong{Conceptual choice (keys):} ILD uses one
#'   subject identifier per row; \code{tbl_ts} must have exactly one key
#'   column. Compound keys are not supported—combine levels into a single id
#'   column before calling \code{ild_prepare()}.
#' @param id Character. Name of the subject/unit identifier column. Omit both
#'   \code{id} and \code{time} when \code{data} is a tsibble (see above).
#' @param time Character. Name of the time column (Date, POSIXct, or numeric).
#' @param gap_threshold Numeric. Time distance above which an interval is
#'   flagged as a gap (`.ild_gap` TRUE). Same units as the numeric time
#'   (e.g. seconds if time is POSIXct). Use `Inf` to disable gap flagging.
#' @param duplicate_handling Character. How to handle duplicate timestamps
#'   within the same id: `"first"` (keep first), `"last"` (keep last),
#'   `"error"` (stop with an error), `"collapse"` (aggregate with \code{collapse_fn}).
#' @param collapse_fn Named list of functions, one per variable to collapse.
#'   Used only when \code{duplicate_handling = "collapse"}. E.g. \code{list(x = mean, y = function(z) z[1])}.
#'   Variables not in \code{collapse_fn} keep their first value within the duplicate group.
#' @return An ILD tibble with `.ild_*` columns and metadata attributes.
#'   Spacing metadata (see [ild_meta()]) includes overall stats and a
#'   \code{by_id} tibble of per-person spacing stats (median_dt, iqr_dt,
#'   n_intervals, pct_gap). Use [ild_summary()] to inspect and check gap
#'   flags before modeling.
#' @seealso \href{../doc/tsibble-interoperability.html}{\code{vignette("tsibble-interoperability", package = "tidyILD")}}
#'   (tsibble ingestion, provenance, \code{ild_as_tsibble()}).
#' @importFrom dplyr arrange group_by mutate n row_number summarise ungroup
#' @importFrom tibble as_tibble
#' @export
ild_prepare <- function(data,
                        id = NULL,
                        time = NULL,
                        gap_threshold = Inf,
                        duplicate_handling = c("first", "last", "error", "collapse"),
                        collapse_fn = NULL) {
  duplicate_handling <- match.arg(duplicate_handling)
  if (duplicate_handling == "collapse" && (is.null(collapse_fn) || !is.list(collapse_fn))) {
    stop("When duplicate_handling = 'collapse', provide a named list of functions in collapse_fn (e.g. list(x = mean)).", call. = FALSE)
  }
  if (!is.data.frame(data)) stop("'data' must be a data frame or tibble.", call. = FALSE)
  ts_meta <- NULL
  if (xor(is.null(id), is.null(time))) {
    stop("Provide both id and time, or omit both and pass a tsibble (tbl_ts) with one key and an index.", call. = FALSE)
  }
  if (is.null(id) && is.null(time)) {
    if (!requireNamespace("tsibble", quietly = TRUE)) {
      stop("Package 'tsibble' is required when id and time are omitted. Install it or supply id and time explicitly.", call. = FALSE)
    }
    if (!inherits(data, "tbl_ts")) {
      stop("Provide both id and time, or omit both and pass a tsibble (tbl_ts) with one key and an index.", call. = FALSE)
    }
    kv <- tsibble::key_vars(data)
    iv <- tsibble::index_var(data)
    if (length(kv) != 1L) {
      stop("tsibble must have exactly one key column for ild_prepare (subject id); found ", length(kv), ".", call. = FALSE)
    }
    id <- kv[1L]
    time <- iv
    ts_meta <- ild_tsibble_capture_meta(data)
    data <- tibble::as_tibble(data)
  }
  id <- as.character(id)[1]
  time <- as.character(time)[1]
  if (!id %in% names(data)) stop("Column '", id, "' not found in data.", call. = FALSE)
  if (!time %in% names(data)) stop("Column '", time, "' not found in data.", call. = FALSE)

  if (is.null(ts_meta) && requireNamespace("tsibble", quietly = TRUE) && inherits(data, "tbl_ts")) {
    ts_meta <- ild_tsibble_capture_meta(data)
  }

  # Avoid masking user columns: work on a copy and add .ild_* columns
  out <- tibble::as_tibble(data)
  out[[".ild_id"]]   <- out[[id]]
  out[[".ild_time"]] <- out[[time]]
  out[[".ild_time_num"]] <- ild_time_to_num(out[[time]])

  # Sort by id then time
  out <- dplyr::arrange(out, .data[[".ild_id"]], .data[[".ild_time_num"]])

  # Duplicate handling: within each id, keep first/last per timestamp, error, or collapse
  out <- ild_handle_duplicates(out, duplicate_handling, collapse_fn)

  # Within-person sequence and intervals
  out <- dplyr::group_by(out, .data[[".ild_id"]])
  out <- dplyr::mutate(out,
    .ild_seq = seq_len(dplyr::n()),
    .ild_dt = .ild_time_num - dplyr::lag(.ild_time_num, default = NA_real_),
    .ild_gap = dplyr::if_else(
      is.na(.data[[".ild_dt"]]),
      NA,
      .data[[".ild_dt"]] > gap_threshold
    )
  )
  out <- dplyr::ungroup(out)

  # Descriptive spacing stats (no hard classifier)
  dt_vals <- out[[".ild_dt"]]
  spacing <- ild_spacing_stats(dt_vals, gap_threshold = if (is.finite(gap_threshold)) gap_threshold else NULL)
  spacing$by_id <- ild_spacing_by_id(out, ".ild_id", ".ild_dt",
    gap_threshold = if (is.finite(gap_threshold)) gap_threshold else NULL)
  if (!is.null(ts_meta)) {
    spacing <- ild_spacing_tsibble_bridge(spacing, ts_meta)
  }

  out <- new_ild_df(
    data = out,
    ild_id = id,
    ild_time = time,
    gap_threshold = gap_threshold,
    spacing = spacing,
    tsibble_meta = ts_meta
  )
  out <- ild_add_step(out, "ild_prepare",
    list(id = id, time = time, gap_threshold = gap_threshold, duplicate_handling = duplicate_handling),
    list(
      n_id = attr(out, "ild_n_units", exact = TRUE),
      n_obs = attr(out, "ild_n_obs", exact = TRUE),
      spacing_class = ild_spacing_class(out),
      source_was_tsibble = !is.null(ts_meta),
      tsibble_interval_declared = if (!is.null(ts_meta)) ts_meta$interval_format else NULL
    )
  )
  out
}

#' Within each .ild_id, resolve duplicates by keeping first/last row per .ild_time_num, error, or collapse
#' @noRd
ild_handle_duplicates <- function(data, handling, collapse_fn = NULL) {
  id_col <- ".ild_id"
  time_col <- ".ild_time_num"
  grp <- dplyr::group_by(data, .data[[id_col]], .data[[time_col]])
  ns <- dplyr::summarise(grp, .n = dplyr::n(), .groups = "drop")
  if (handling == "error" && any(ns$.n > 1)) {
    stop("Duplicate timestamps within the same id found. Set duplicate_handling to 'first', 'last', or 'collapse'.", call. = FALSE)
  }
  if (handling == "error") return(dplyr::ungroup(dplyr::group_by(data, .data[[id_col]], .data[[time_col]])))
  if (handling == "collapse") {
    fn_list <- collapse_fn
    grp <- dplyr::group_by(data, .data[[id_col]], .data[[time_col]])
    out <- dplyr::summarise(grp, dplyr::across(
      dplyr::everything(),
      function(col) {
        nm <- dplyr::cur_column()
        if (nm %in% names(fn_list)) fn_list[[nm]](col) else col[1]
      }
    ), .groups = "drop")
    return(out)
  }
  grp <- dplyr::group_by(data, .data[[id_col]], .data[[time_col]])
  grp <- dplyr::mutate(grp,
    .ild_dup_rn = dplyr::row_number(),
    .ild_dup_n = dplyr::n()
  )
  grp <- dplyr::ungroup(grp)
  keep <- if (handling == "first") {
    grp[[".ild_dup_rn"]] == 1L
  } else {
    grp[[".ild_dup_rn"]] == grp[[".ild_dup_n"]]
  }
  out <- grp[keep, ]
  out[[".ild_dup_rn"]] <- NULL
  out[[".ild_dup_n"]] <- NULL
  out
}

test_that("ild_missing_pattern returns by_id and overall", {
  d <- data.frame(
    id = c(1, 1, 2, 2),
    time = as.POSIXct(c(0, 10, 0, 10), origin = "1970-01-01"),
    x = c(1, NA, 3, 4)
  )
  x <- ild_prepare(d, id = "id", time = "time")
  mp <- ild_missing_pattern(x, vars = "x")
  expect_named(mp, c("summary", "plot", "by_id", "overall", "n_complete", "vars"))
  expect_s3_class(mp$summary, "tbl_df")
  expect_true(inherits(mp$plot, "gg"))
  expect_equal(nrow(mp$by_id), 2)
  expect_equal(mp$overall$x$n_na, 1)
  expect_equal(mp$n_complete, 3)
})

test_that("ild_missing_pattern with no missing", {
  d <- data.frame(id = c(1, 1, 2, 2), time = as.POSIXct(0:3, origin = "1970-01-01"), x = 1:4)
  x <- ild_prepare(d, id = "id", time = "time")
  mp <- ild_missing_pattern(x, vars = "x")
  expect_equal(mp$overall$x$n_na, 0)
  expect_equal(mp$n_complete, 4)
})

test_that("ild_missing_pattern with max_ids and seed subsets persons", {
  d <- data.frame(
    id = rep(1:5, each = 3),
    time = as.POSIXct(rep(0:2, 5) + rep((0:4) * 100, each = 3), origin = "1970-01-01"),
    x = 1:15
  )
  x <- ild_prepare(d, id = "id", time = "time")
  mp <- ild_missing_pattern(x, vars = "x", max_ids = 2, seed = 42)
  expect_equal(nrow(mp$by_id), 2)
  expect_equal(nrow(mp$summary), 1)
  expect_true(inherits(mp$plot, "gg"))
})

# Stable KFAS test fixtures: DGPs, named seeds, shared expectations
# Loaded automatically by testthat.

# Chosen so the Gaussian local-level DGP does not land on a pathological ML optimum.
KFAS_TEST_SEED_RECOVERY <- 42L
KFAS_TEST_SEED_CONTRACT <- 9001L

#' Simulate Gaussian local level + noise; return [ild_prepare()]d data and true level.
#' Use moderate \code{Q}, \code{H} (e.g. 1) so ML variance recovery is numerically stable.
#' @keywords internal
kfas_dgp_local_level <- function(n, Q = 1, H = 1, seed) {
  set.seed(seed)
  Q <- as.numeric(Q)[1L]
  H <- as.numeric(H)[1L]
  alpha <- numeric(n)
  alpha[1L] <- stats::rnorm(1L, 0, sqrt(Q))
  if (n >= 2L) {
    for (t in 2L:n) {
      alpha[t] <- alpha[t - 1L] + stats::rnorm(1L, 0, sqrt(Q))
    }
  }
  y <- alpha + stats::rnorm(n, 0, sqrt(H))
  raw <- data.frame(
    id = 1L,
    time = seq_len(n),
    y = y,
    stringsAsFactors = FALSE
  )
  list(
    data = ild_prepare(raw, id = "id", time = "time"),
    true_level = alpha,
    Q = Q,
    H = H
  )
}

#' Expect core KFAS bundle slots used by autoplot and reports.
#' @keywords internal
expect_kfas_bundle_core <- function(b) {
  expect_s3_class(b, "ild_diagnostics_bundle")
  expect_no_error(validate_ild_diagnostics_bundle(b))
  expect_identical(b$meta$engine, "KFAS")
  expect_true(is.list(b$fit))
  expect_true(!is.null(b$fit$engine))
  expect_true("log_likelihood" %in% names(b$fit))
  expect_true(is.list(b$fit$optimizer))
  # Residual ACF list shape (when innovations computed)
  ac <- b$residual$acf
  if (!is.null(ac) && is.list(ac)) {
    expect_true(!is.null(ac$lag))
    expect_true(!is.null(ac$acf))
    expect_true(length(ac$lag) == length(ac$acf))
  }
}

#' Single-subject ILD with highly variable observation intervals (high IQR/median on .ild_dt).
#' @keywords internal
kfas_fixture_irregular_intervals <- function(n = 55L, seed = 91001L) {
  set.seed(seed)
  gaps <- rep(c(1, 10), length.out = max(1L, n - 1L))
  tvec <- cumsum(c(0, gaps))
  y <- stats::rnorm(length(tvec))
  raw <- data.frame(id = 1L, time = tvec, y = y, stringsAsFactors = FALSE)
  ild_prepare(raw, id = "id", time = "time")
}

#' Outcome with several NA runs and enough NA share to trigger missing-segment guardrail.
#' @keywords internal
kfas_fixture_segmented_na_outcome <- function(seed = 91002L) {
  set.seed(seed)
  n <- 50L
  y <- stats::rnorm(n)
  # Four separate NA runs; 8 missing -> 16% > 10%
  y[c(3L, 12L, 13L, 22L, 23L, 24L, 38L, 45L)] <- NA
  raw <- data.frame(id = 1L, time = seq_len(n), y = y, stringsAsFactors = FALSE)
  ild_prepare(raw, id = "id", time = "time")
}

# Contract invariant tests (ILD object)

test_that("contract: within each .ild_id, .ild_time_num is non-decreasing", {
  d <- data.frame(
    id = c(1, 1, 1, 2, 2),
    time = as.POSIXct(c(0, 5, 10, 0, 5), origin = "1970-01-01"),
    x = 1:5
  )
  x <- ild_prepare(d, id = "id", time = "time")
  for (uid in unique(x[[".ild_id"]])) {
    tnum <- x[[".ild_time_num"]][x[[".ild_id"]] == uid]
    expect_true(all(diff(tnum) >= 0, na.rm = TRUE))
  }
})

test_that("contract: .ild_seq is exactly 1:n_i per id", {
  d <- data.frame(id = c(1, 1, 1, 2, 2), time = as.POSIXct(0:4, origin = "1970-01-01"), x = 1:5)
  x <- ild_prepare(d, id = "id", time = "time")
  expect_equal(x[[".ild_seq"]], c(1, 2, 3, 1, 2))
})

test_that("contract: .ild_dt equals .ild_time_num - lag(.ild_time_num) within id", {
  d <- data.frame(id = c(1, 1, 1, 2, 2), time = as.POSIXct(c(0, 3, 7, 0, 4), origin = "1970-01-01"), x = 1:5)
  x <- ild_prepare(d, id = "id", time = "time")
  dt_expected <- c(NA, 3, 4, NA, 4)
  expect_equal(x[[".ild_dt"]], dt_expected)
})

test_that("contract: .ild_gap is TRUE where .ild_dt > gap_threshold, NA for first row per id", {
  d <- data.frame(id = c(1, 1, 1), time = as.POSIXct(c(0, 5, 100), origin = "1970-01-01"), x = 1:3)
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 10)
  expect_equal(x[[".ild_gap"]], c(NA, FALSE, TRUE))
})

test_that("contract: .ild_time_num consistent with .ild_time for POSIXct", {
  d <- data.frame(id = 1, time = as.POSIXct(100, origin = "1970-01-01"), x = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  expect_equal(x[[".ild_time_num"]], 100)
})

test_that("ild_spacing_class returns regular-ish for equal intervals", {
  d <- data.frame(
    id = c(1, 1, 1, 2, 2, 2),
    time = as.POSIXct(rep(c(0, 10, 20), 2), origin = "1970-01-01"),
    x = 1:6
  )
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 100)
  expect_equal(ild_spacing_class(x), "regular-ish")
})

test_that("ild_spacing_class returns irregular-ish when cv high", {
  d <- data.frame(
    id = c(1, 1, 1),
    time = as.POSIXct(c(0, 5, 100), origin = "1970-01-01"),
    x = 1:3
  )
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 10)
  expect_equal(ild_spacing_class(x), "irregular-ish")
})

test_that("ild_spacing_class is overridable via cv_threshold and pct_gap_threshold", {
  d <- data.frame(
    id = c(1, 1, 1),
    time = as.POSIXct(c(0, 5, 100), origin = "1970-01-01"),
    x = 1:3
  )
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 10)
  # Relax both so classification becomes regular-ish
  expect_equal(ild_spacing_class(x, cv_threshold = 1, pct_gap_threshold = 100), "regular-ish")
})

test_that("ild_summary returns expected structure", {
  d <- data.frame(
    id = c(1, 1, 2, 2),
    time = as.POSIXct(c(0, 10, 0, 10), origin = "1970-01-01"),
    x = 1:4
  )
  # gap_threshold 50 so intervals (10, 10) are not gaps
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 50)
  s <- ild_summary(x)
  expect_named(s, c("summary", "n_units", "n_obs", "time_range", "spacing", "n_gaps", "pct_gap"))
  expect_s3_class(s$summary, "tbl_df")
  expect_equal(names(s$summary), c("n_id", "n_obs", "time_min", "time_max", "prop_gap", "median_dt_sec", "iqr_dt_sec"))
  expect_equal(s$n_units, 2)
  expect_equal(s$n_obs, 4)
  expect_equal(s$time_range, c(0, 10))
  expect_equal(s$n_gaps, 0)
})

test_that("ild_summary counts gaps", {
  d <- data.frame(
    id = c(1, 1, 1),
    time = as.POSIXct(c(0, 5, 100), origin = "1970-01-01"),
    x = 1:3
  )
  x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 10)
  s <- ild_summary(x)
  expect_equal(s$n_gaps, 1)
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
has_tsibble <- requireNamespace("tsibble", quietly = TRUE)

## ----ingest, eval = has_tsibble-----------------------------------------------
suppressPackageStartupMessages({
  library(tidyILD)
  library(tsibble)
})
t <- seq.POSIXt(as.POSIXct("2020-01-01", tz = "UTC"), by = "hour", length.out = 4)
d <- tsibble(
  id = rep(1L, 4),
  t = t,
  y = 1:4,
  key = id,
  index = t
)
x <- ild_prepare(d)
class(x)

## ----meta, eval = has_tsibble-------------------------------------------------
str(ild_tsibble_meta(x), max.level = 1)

## ----spacing-bridge, eval = has_tsibble---------------------------------------
sp <- ild_spacing(x)
names(sp$tsibble)

## ----summary-ts, eval = has_tsibble-------------------------------------------
names(ild_summary(x))
ild_summary(x)$tsibble[c("key_vars", "index_var", "interval_format", "is_regular")]

## ----roundtrip, eval = has_tsibble--------------------------------------------
d2 <- ild_as_tsibble(x)
tsibble::interval(d2)
identical(tsibble::interval(d), tsibble::interval(d2))

# Recovery summaries from ild_power(..., return_sims = TRUE)

#' Recovery metrics from simulation replications
#'
#' Summarizes **bias**, **RMSE**, and **nominal Wald interval coverage** for a known
#' coefficient using the per-replication table from [ild_power()] when
#' \code{return_sims = TRUE}. Purely descriptive; does not run new simulations.
#'
#' @param sim_results A data frame or tibble with columns \code{estimate},
#'   \code{std_error}, and \code{converged} (as in \code{ild_power(..., return_sims = TRUE)$sim_results}).
#' @param truth Known value of the estimand (e.g. \code{effect_size} in [ild_power()]).
#' @param level Nominal coverage level for Wald intervals (default \code{0.95}).
#' @param converged_only If \code{TRUE} (default), only rows with \code{converged == TRUE}
#'   are used.
#' @return A one-row \code{\link[tibble]{tibble}} with \code{truth}, \code{n} (rows used),
#'   \code{n_total} (rows in \code{sim_results} if \code{converged_only}), \code{mean_estimate},
#'   \code{bias}, \code{rmse}, \code{coverage}, \code{level}.
#' @seealso [ild_power()]
#' @export
ild_recovery_metrics <- function(sim_results,
                                 truth,
                                 level = 0.95,
                                 converged_only = TRUE) {
  req <- c("estimate", "std_error", "converged")
  nm <- names(sim_results)
  if (!all(req %in% nm)) {
    stop(
      "sim_results must contain columns: ", paste(req, collapse = ", "),
      call. = FALSE
    )
  }
  truth <- as.numeric(truth)[1L]
  level <- as.numeric(level)[1L]
  if (level <= 0 || level >= 1) {
    stop("level must be in (0, 1).", call. = FALSE)
  }
  n_total <- nrow(sim_results)
  sim <- if (isTRUE(converged_only)) {
    sim_results[sim_results$converged %in% TRUE, , drop = FALSE]
  } else {
    sim_results
  }
  n <- nrow(sim)
  if (n == 0L) {
    return(tibble::tibble(
      truth = truth,
      n = 0L,
      n_total = n_total,
      mean_estimate = NA_real_,
      bias = NA_real_,
      rmse = NA_real_,
      coverage = NA_real_,
      level = level
    ))
  }
  z <- stats::qnorm(1 - (1 - level) / 2)
  est <- sim$estimate
  se <- sim$std_error
  lo <- est - z * se
  hi <- est + z * se
  covers <- !is.na(se) & se > 0 & truth >= lo & truth <= hi
  tibble::tibble(
    truth = truth,
    n = n,
    n_total = n_total,
    mean_estimate = mean(est, na.rm = TRUE),
    bias = mean(est - truth, na.rm = TRUE),
    rmse = sqrt(mean((est - truth)^2, na.rm = TRUE)),
    coverage = mean(covers, na.rm = TRUE),
    level = level
  )
}

test_that("ild_recovery_metrics matches hand-computed bias rmse coverage", {
  sim <- tibble::tibble(
    estimate = c(1.0, 2.0, 10.0),
    std_error = c(1, 1, 1),
    converged = c(TRUE, TRUE, TRUE)
  )
  truth <- 2.0
  m <- ild_recovery_metrics(sim, truth = truth, level = 0.95)
  expect_equal(m$bias, mean(sim$estimate - truth))
  expect_equal(m$rmse, sqrt(mean((sim$estimate - truth)^2)))
  expect_equal(m$n, 3L)
  expect_equal(m$n_total, 3L)
  # truth=2 in Wald interval for est=1,2; not for est=10 (interval ~ [8.04, 11.96])
  expect_equal(m$coverage, 2 / 3)
})

test_that("ild_recovery_metrics filters converged_only", {
  sim <- tibble::tibble(
    estimate = c(1, 2),
    std_error = c(1, 1),
    converged = c(TRUE, FALSE)
  )
  m <- ild_recovery_metrics(sim, truth = 1, converged_only = TRUE)
  expect_equal(m$n, 1L)
  m2 <- ild_recovery_metrics(sim, truth = 1, converged_only = FALSE)
  expect_equal(m2$n, 2L)
})

test_that("ild_recovery_metrics errors on missing columns", {
  expect_error(ild_recovery_metrics(data.frame(a = 1), truth = 1), "sim_results must contain")
})

test_that("ild_recovery_metrics empty converged returns NA summaries", {
  sim <- tibble::tibble(
    estimate = 1,
    std_error = 1,
    converged = FALSE
  )
  m <- ild_recovery_metrics(sim, truth = 0, converged_only = TRUE)
  expect_equal(m$n, 0L)
  expect_true(is.na(m$bias))
})

#' Check lag variable validity (gap-aware)
#'
#' Given an ILD object and lag variable names, reports how many lagged
#' values are valid vs invalid (NA because the time distance to the
#' lagged row exceeded a threshold). Useful to audit lag columns before
#' modeling without re-specifying \code{max_gap}.
#'
#' @param x An ILD object (see [is_ild()]) that contains lag columns
#'   (e.g. from [ild_lag()] with \code{mode = "gap_aware"}).
#' @param lag_vars Character vector of lag column names (e.g. \code{"y_lag1"}).
#'   If \code{NULL}, attempts to detect columns ending in \code{_lag{n}} or \code{_lag_window}.
#' @param max_gap Numeric. Threshold used to define invalid (same units as
#'   \code{.ild_time_num}). If \code{NULL}, uses \code{ild_meta(x)$ild_gap_threshold}.
#' @return A data frame with one row per lag variable: \code{var}, \code{lag} (parsed lag order or \code{"window"}),
#'   \code{n_valid}, \code{n_invalid}, \code{n_first}, \code{n_total}, \code{pct_valid}, \code{pct_invalid}.
#' @importFrom tibble tibble as_tibble
#' @export
ild_check_lags <- function(x, lag_vars = NULL, max_gap = NULL) {
  validate_ild(x)
  if (is.null(lag_vars)) {
    nms <- names(x)
    lag_vars <- nms[grepl("_lag[0-9]+$|_lag_window$", nms)]
  }
  if (length(lag_vars) == 0) {
    return(tibble::tibble(
      var = character(), lag = character(), n_valid = integer(), n_invalid = integer(),
      n_first = integer(), n_total = integer(), pct_valid = double(), pct_invalid = double()
    ))
  }
  if (is.null(max_gap)) {
    meta <- ild_meta(x)
    max_gap <- meta$ild_gap_threshold
    if (is.null(max_gap) || !is.finite(max_gap)) max_gap <- Inf
  }
  id_col <- ".ild_id"
  seq_col <- ".ild_seq"
  dt_col <- ".ild_dt"
  n_first <- sum(x[[seq_col]] == 1L, na.rm = TRUE)
  out <- vector("list", length(lag_vars))
  j <- 0L
  for (i in seq_along(lag_vars)) {
    v <- lag_vars[i]
    if (!v %in% names(x)) {
      warning("Lag variable '", v, "' not found in data.", call. = FALSE)
      next
    }
    j <- j + 1L
    could_have_lag <- x[[seq_col]] > 1L
    n_valid <- sum(!is.na(x[[v]]) & could_have_lag, na.rm = TRUE)
    n_invalid <- sum(is.na(x[[v]]) & could_have_lag, na.rm = TRUE)
    n_total <- sum(could_have_lag, na.rm = TRUE)
    pct_valid <- if (n_total > 0) 100 * n_valid / n_total else NA_real_
    pct_invalid <- if (n_total > 0) 100 * n_invalid / n_total else NA_real_
    lag_parsed <- if (grepl("_lag_window$", v)) "window" else {
      z <- sub(".*_lag([0-9]+)$", "\\1", v)
      if (z == v) NA_character_ else z
    }
    out[[j]] <- list(
      var = v,
      lag = lag_parsed,
      n_valid = as.integer(n_valid),
      n_invalid = as.integer(n_invalid),
      n_first = as.integer(n_first),
      n_total = as.integer(n_total),
      pct_valid = pct_valid,
      pct_invalid = pct_invalid
    )
  }
  if (j == 0) {
    return(tibble::tibble(
      var = character(), lag = character(), n_valid = integer(), n_invalid = integer(),
      n_first = integer(), n_total = integer(), pct_valid = double(), pct_invalid = double()
    ))
  }
  tibble::as_tibble(do.call(rbind, lapply(out[seq_len(j)], as.data.frame, stringsAsFactors = FALSE)))
}

# MSM simulation and recovery harness (causal validation)

#' Simulate a simple longitudinal MSM scenario
#'
#' Generates ILD with time-varying treatment, confounding, and optional monotone
#' censoring/dropout under a known treatment effect (\code{true_ate}).
#'
#' @param n_id Number of persons.
#' @param n_obs_per Planned observations per person.
#' @param true_ate True additive treatment effect on outcome.
#' @param censoring If \code{TRUE}, apply monotone dropout and return observed rows only.
#' @param positivity_stress Multiplier for treatment/censoring logits
#'   (>1 worsens overlap; <1 improves overlap).
#' @param treatment_strength Base strength of treatment dependence on confounders/history.
#' @param censoring_strength Base strength of dropout dependence on stress/treatment.
#' @param seed Integer seed.
#' @return ILD object with columns \code{stress} and binary \code{trt}; attributes include
#'   \code{true_ate}.
#' @export
#' @examples
#' d <- ild_msm_simulate_scenario(n_id = 20, n_obs_per = 8, true_ate = 0.5, seed = 12)
#' ild_summary(d)$summary
ild_msm_simulate_scenario <- function(n_id = 60L,
                                      n_obs_per = 10L,
                                      true_ate = 0.5,
                                      censoring = TRUE,
                                      positivity_stress = 1,
                                      treatment_strength = 0.8,
                                      censoring_strength = 0.6,
                                      seed = 42L) {
  set.seed(as.integer(seed)[1L])
  d <- ild_simulate(n_id = n_id, n_obs_per = n_obs_per, seed = seed)
  d$stress <- as.numeric(scale(rnorm(nrow(d))))
  d <- ild_prepare(d, id = "id", time = "time")

  # Sequential treatment process A_t ~ stress_t + A_{t-1}
  ord <- order(d$.ild_id, d$.ild_seq, seq_len(nrow(d)))
  ido <- d$.ild_id[ord]
  stress_o <- d$stress[ord]
  trt_o <- rep(NA_integer_, nrow(d))
  i <- 1L
  while (i <= nrow(d)) {
    j <- i
    while (j < nrow(d) && ido[j + 1L] == ido[i]) j <- j + 1L
    prev <- 0
    for (k in i:j) {
      p <- stats::plogis(
        -0.2 + positivity_stress * (treatment_strength * stress_o[k] + 1.0 * prev)
      )
      trt_o[k] <- stats::rbinom(1L, 1L, p)
      prev <- trt_o[k]
    }
    i <- j + 1L
  }
  trt <- rep(NA_integer_, nrow(d))
  trt[ord] <- trt_o
  d$trt <- trt

  # Outcome with true causal treatment effect.
  d$y <- d$y + as.numeric(true_ate) * d$trt + 0.4 * d$stress

  if (isTRUE(censoring)) {
    keep <- rep(TRUE, nrow(d))
    for (pid in unique(d$.ild_id)) {
      idx <- which(d$.ild_id == pid)
      idx <- idx[order(d$.ild_seq[idx])]
      dropped <- FALSE
      for (k in seq_along(idx)) {
        ii <- idx[k]
        if (!dropped) {
          p_drop <- stats::plogis(
            -3.5 + positivity_stress * (censoring_strength * d$stress[ii] + 0.5 * d$trt[ii])
          )
          dr <- stats::rbinom(1L, 1L, p_drop)
          if (dr == 1L && k < length(idx)) {
            dropped <- TRUE
          }
        } else {
          keep[ii] <- FALSE
        }
      }
    }
    raw <- as.data.frame(d[keep, c("id", "time", "y", "stress", "trt"), drop = FALSE])
    d <- ild_prepare(raw, id = "id", time = "time")
  }

  attr(d, "true_ate") <- as.numeric(true_ate)
  d
}

#' MSM recovery simulation harness
#'
#' Repeats an MSM analysis over simulated datasets with known \code{true_ate} and
#' summarizes estimation bias, RMSE, CI coverage, and positivity stress summaries.
#'
#' @param n_sim Number of simulations.
#' @param n_id Number of persons per simulation.
#' @param n_obs_per Planned observations per person.
#' @param true_ate True additive treatment effect.
#' @param n_boot Bootstrap replicates when \code{inference = "bootstrap"}.
#' @param inference \code{"bootstrap"} (default), \code{"robust"}, or \code{"none"}.
#' @param seed Base seed.
#' @param censoring If \code{TRUE}, include monotone dropout in DGP and fit IPCW+joint pipeline.
#' @param scenario_grid Optional data frame/list of scenario settings. Supported columns:
#'   \code{scenario_id, n_id, n_obs_per, true_ate, censoring, positivity_stress,
#'   treatment_strength, censoring_strength, misspec_treatment_model}.
#' @return List with \code{summary}, \code{summary_by_scenario}, and \code{sim_results}.
#' @export
#' @examples
#' \dontrun{
#' out <- ild_msm_recovery(n_sim = 10, n_id = 40, n_obs_per = 10, true_ate = 0.5, n_boot = 50)
#' out$summary
#' }
ild_msm_recovery <- function(n_sim = 100L,
                             n_id = 80L,
                             n_obs_per = 10L,
                             true_ate = 0.5,
                             n_boot = 200L,
                             inference = c("bootstrap", "robust", "none"),
                             seed = 1001L,
                             censoring = TRUE,
                             scenario_grid = NULL) {
  inference <- match.arg(inference)
  n_sim <- as.integer(n_sim)[1L]
  if (!is.finite(n_sim) || n_sim < 1L) stop("n_sim must be >= 1.", call. = FALSE)
  scen <- .ild_msm_build_scenario_grid(
    scenario_grid = scenario_grid,
    n_id = n_id,
    n_obs_per = n_obs_per,
    true_ate = true_ate,
    censoring = censoring
  )
  rows <- vector("list", nrow(scen) * n_sim)
  at <- 1L
  for (j in seq_len(nrow(scen))) {
    sj <- scen[j, , drop = FALSE]
    for (i in seq_len(n_sim)) {
      s <- as.integer(seed) + i + (j - 1L) * 10000L
      dat <- ild_msm_simulate_scenario(
        n_id = sj$n_id[[1L]],
        n_obs_per = sj$n_obs_per[[1L]],
        true_ate = sj$true_ate[[1L]],
        censoring = sj$censoring[[1L]],
        positivity_stress = sj$positivity_stress[[1L]],
        treatment_strength = sj$treatment_strength[[1L]],
        censoring_strength = sj$censoring_strength[[1L]],
        seed = s
      )
      dat <- ild_center(dat, y)
      hs <- ild_msm_history_spec(vars = c("stress", "trt"), lags = 1L)
      dat <- ild_build_msm_history(dat, hs)
      est <- ild_msm_estimand(type = "ate", regime = "static", treatment = "trt")
      hist_formula <- if (isTRUE(sj$misspec_treatment_model[[1L]])) {
        ~ stress_lag1
      } else {
        ~ stress_lag1 + trt_lag1
      }

      fit_obj <- tryCatch(
        ild_msm_fit(
          estimand = est,
          data = dat,
          outcome_formula = y ~ y_bp + y_wp + stress + trt + (1 | id),
          history = hist_formula,
          treatment_engine = "sequential_msm",
          predictors_censor = if (isTRUE(sj$censoring[[1L]])) "stress" else NULL,
          weights_col = ".ipw",
          inference = inference,
          n_boot = n_boot,
          seed = s,
          warn_no_ar1 = FALSE,
          warn_uncentered = FALSE
        ),
        error = function(e) NULL
      )
      if (is.null(fit_obj)) {
        rows[[at]] <- tibble::tibble(
          scenario_id = sj$scenario_id[[1L]],
          true_ate = sj$true_ate[[1L]],
          sim = i,
          estimate = NA_real_,
          std_error = NA_real_,
          conf_low = NA_real_,
          conf_high = NA_real_,
          covered = NA,
          ess = NA_real_,
          weight_ratio = NA_real_,
          propensity_extreme_rate = NA_real_,
          inference_status = "unsupported",
          inference_method = inference,
          converged = FALSE
        )
        at <- at + 1L
        next
      }

      coef_est <- as.numeric(lme4::fixef(fit_obj$fit)[["trt"]])
      rr <- fit_obj$inference$summary
      rr <- rr[rr$term == "trt", , drop = FALSE]
      se <- if (nrow(rr) == 1L) rr$std_error else NA_real_
      lo <- if (nrow(rr) == 1L) rr$conf_low else NA_real_
      hi <- if (nrow(rr) == 1L) rr$conf_high else NA_real_

      ess <- ild_ipw_ess(fit_obj$weights_data, weights_col = fit_obj$weights_col)
      wf <- fit_obj$weights_data[[fit_obj$weights_col]]
      wf <- wf[is.finite(wf) & wf > 0]
      wr <- if (length(wf) >= 2L) max(wf) / min(wf) else NA_real_
      overlap <- .ild_msm_propensity_overlap_summary(fit_obj$weights_data)

      rows[[at]] <- tibble::tibble(
        scenario_id = sj$scenario_id[[1L]],
        true_ate = sj$true_ate[[1L]],
        sim = i,
        estimate = coef_est,
        std_error = se,
        conf_low = lo,
        conf_high = hi,
        covered = is.finite(lo) && is.finite(hi) && lo <= sj$true_ate[[1L]] && sj$true_ate[[1L]] <= hi,
        ess = ess,
        weight_ratio = wr,
        propensity_extreme_rate = overlap$extreme_rate,
        inference_status = fit_obj$inference$status,
        inference_method = fit_obj$inference$method,
        converged = TRUE
      )
      at <- at + 1L
    }
  }

  sim_results <- dplyr::bind_rows(rows)
  summary_by_scenario <- .ild_msm_recovery_summary_by_scenario(sim_results)
  summary <- .ild_msm_recovery_summary_overall(summary_by_scenario, n_sim = n_sim)

  out <- list(
    summary = summary,
    summary_by_scenario = summary_by_scenario,
    sim_results = sim_results,
    scenarios = scen
  )
  class(out) <- c("ild_msm_recovery", "list")
  attr(out, "ild_provenance") <- ild_new_analysis_provenance(
    NULL,
    "ild_msm_recovery",
    list(
      n_sim = n_sim,
      n_id = n_id,
      n_obs_per = n_obs_per,
      true_ate = true_ate,
      inference = inference,
      n_scenarios = nrow(scen)
    ),
    list(n_success = sum(sim_results$converged, na.rm = TRUE))
  )
  out
}

#' @export
print.ild_msm_recovery <- function(x, ...) {
  cat("MSM recovery summary\n")
  print(x$summary)
  invisible(x)
}

#' @keywords internal
#' @noRd
.ild_msm_build_scenario_grid <- function(scenario_grid, n_id, n_obs_per, true_ate, censoring) {
  if (is.null(scenario_grid)) {
    return(tibble::tibble(
      scenario_id = "baseline",
      n_id = as.integer(n_id),
      n_obs_per = as.integer(n_obs_per),
      true_ate = as.numeric(true_ate),
      censoring = isTRUE(censoring),
      positivity_stress = 1,
      treatment_strength = 0.8,
      censoring_strength = 0.6,
      misspec_treatment_model = FALSE
    ))
  }
  scen <- tibble::as_tibble(scenario_grid)
  if (!"scenario_id" %in% names(scen)) {
    scen$scenario_id <- paste0("scenario_", seq_len(nrow(scen)))
  }
  defaults <- list(
    n_id = as.integer(n_id),
    n_obs_per = as.integer(n_obs_per),
    true_ate = as.numeric(true_ate),
    censoring = isTRUE(censoring),
    positivity_stress = 1,
    treatment_strength = 0.8,
    censoring_strength = 0.6,
    misspec_treatment_model = FALSE
  )
  for (nm in names(defaults)) {
    if (!nm %in% names(scen)) scen[[nm]] <- defaults[[nm]]
  }
  scen
}

#' @keywords internal
#' @noRd
.ild_msm_propensity_overlap_summary <- function(data) {
  fits <- attr(data, "ild_iptw_msm_fits", exact = TRUE)
  if (is.null(fits) || !length(fits)) {
    return(list(extreme_rate = NA_real_))
  }
  pv <- unlist(lapply(fits, function(f) tryCatch(stats::fitted(f), error = function(e) numeric())), use.names = FALSE)
  pv <- pv[is.finite(pv)]
  if (!length(pv)) return(list(extreme_rate = NA_real_))
  list(extreme_rate = mean(pv < 0.05 | pv > 0.95, na.rm = TRUE))
}

#' @keywords internal
#' @noRd
.ild_msm_recovery_summary_by_scenario <- function(sim_results) {
  grouped <- dplyr::group_by(sim_results, .data$scenario_id, .data$true_ate)
  dplyr::summarise(
    grouped,
    n_sim = dplyr::n(),
    n_success = sum(.data$converged & is.finite(.data$estimate), na.rm = TRUE),
    failure_rate = mean(!.data$converged, na.rm = TRUE),
    bias = mean(.data$estimate - .data$true_ate, na.rm = TRUE),
    rmse = sqrt(mean((.data$estimate - .data$true_ate)^2, na.rm = TRUE)),
    coverage = mean(.data$covered, na.rm = TRUE),
    ess_mean = mean(.data$ess, na.rm = TRUE),
    ess_min = min(.data$ess, na.rm = TRUE),
    ess_p10 = stats::quantile(.data$ess, probs = 0.10, na.rm = TRUE, type = 7),
    ess_p50 = stats::quantile(.data$ess, probs = 0.50, na.rm = TRUE, type = 7),
    ess_p90 = stats::quantile(.data$ess, probs = 0.90, na.rm = TRUE, type = 7),
    weight_ratio_median = stats::median(.data$weight_ratio, na.rm = TRUE),
    propensity_extreme_rate = mean(.data$propensity_extreme_rate, na.rm = TRUE),
    inference_degraded_rate = mean(.data$inference_status != "ok", na.rm = TRUE),
    inference_unsupported_rate = mean(.data$inference_status == "unsupported", na.rm = TRUE),
    inference_method_modal = names(sort(table(.data$inference_method), decreasing = TRUE))[1L],
    mean_ci_width = mean(.data$conf_high - .data$conf_low, na.rm = TRUE),
    coverage_calibration_gap = mean(.data$covered, na.rm = TRUE) - 0.95,
    .groups = "drop"
  )
}

#' @keywords internal
#' @noRd
.ild_msm_recovery_summary_overall <- function(summary_by_scenario, n_sim) {
  tibble::tibble(
    n_sim = n_sim,
    n_scenarios = nrow(summary_by_scenario),
    n_success = sum(summary_by_scenario$n_success, na.rm = TRUE),
    bias = mean(summary_by_scenario$bias, na.rm = TRUE),
    rmse = mean(summary_by_scenario$rmse, na.rm = TRUE),
    coverage = mean(summary_by_scenario$coverage, na.rm = TRUE),
    ess_mean = mean(summary_by_scenario$ess_mean, na.rm = TRUE),
    ess_min = min(summary_by_scenario$ess_min, na.rm = TRUE),
    ess_p10 = mean(summary_by_scenario$ess_p10, na.rm = TRUE),
    ess_p50 = mean(summary_by_scenario$ess_p50, na.rm = TRUE),
    ess_p90 = mean(summary_by_scenario$ess_p90, na.rm = TRUE),
    weight_ratio_median = stats::median(summary_by_scenario$weight_ratio_median, na.rm = TRUE),
    inference_degraded_rate = mean(summary_by_scenario$inference_degraded_rate, na.rm = TRUE),
    inference_unsupported_rate = mean(summary_by_scenario$inference_unsupported_rate, na.rm = TRUE),
    coverage_calibration_gap = mean(summary_by_scenario$coverage_calibration_gap, na.rm = TRUE)
  )
}

test_that("ild_msm_simulate_scenario returns ILD with treatment and truth", {
  d <- ild_msm_simulate_scenario(
    n_id = 12,
    n_obs_per = 6,
    true_ate = 0.4,
    positivity_stress = 1.2,
    seed = 2301
  )
  expect_true(is_ild(d))
  expect_true(all(c("stress", "trt") %in% names(d)))
  expect_equal(attr(d, "true_ate", exact = TRUE), 0.4)
})

test_that("ild_msm_recovery fast tier returns summary and per-sim metrics", {
  set.seed(2302)
  out <- ild_msm_recovery(
    n_sim = 3L,
    n_id = 20L,
    n_obs_per = 6L,
    true_ate = 0.5,
    n_boot = 8L,
    inference = "bootstrap",
    seed = 99L,
    censoring = TRUE
  )
  expect_s3_class(out, "ild_msm_recovery")
  expect_s3_class(out$summary, "tbl_df")
  expect_s3_class(out$summary_by_scenario, "tbl_df")
  expect_s3_class(out$sim_results, "tbl_df")
  expect_true(all(c("bias", "rmse", "coverage", "ess_mean", "ess_p50") %in% names(out$summary)))
  expect_true(all(c("failure_rate", "inference_degraded_rate", "inference_unsupported_rate") %in% names(out$summary_by_scenario)))
  expect_true(nrow(out$sim_results) == 3L)
  pv <- attr(out, "ild_provenance", exact = TRUE)
  expect_true(is.list(pv))
  expect_equal(pv$object_type, "ild_msm_recovery")
})

test_that("ild_msm_recovery scenario grid works (extended tier)", {
  skip_on_cran()
  grid <- msm_recovery_scenario_grid_fixture()
  out <- ild_msm_recovery(
    n_sim = 2L,
    n_id = 20L,
    n_obs_per = 6L,
    true_ate = 0.5,
    inference = "none",
    scenario_grid = grid,
    seed = 2310
  )
  expect_equal(nrow(out$summary_by_scenario), nrow(grid))
  expect_true(all(c("scenario_id", "propensity_extreme_rate", "coverage_calibration_gap", "inference_method_modal") %in% names(out$summary_by_scenario)))
  expect_equal(nrow(out$sim_results), 2L * nrow(grid))
})

# Internal helpers for tidyILD (spacing stats, duplicate handling, etc.)
# Not exported.

#' Compute descriptive spacing stats from .ild_dt (excluding NA)
#' Returns list: median_dt, iqr_dt, n_intervals, pct_gap (if gap_threshold given), etc.
#' @noRd
ild_spacing_stats <- function(dt, gap_threshold = NULL) {
  dt <- dt[!is.na(dt)]
  n <- length(dt)
  if (n == 0) {
    return(list(
      median_dt = NA_real_,
      iqr_dt = NA_real_,
      n_intervals = 0L,
      pct_gap = NA_real_
    ))
  }
  median_dt <- stats::median(dt)
  iqr_dt <- stats::IQR(dt)
  pct_gap <- if (!is.null(gap_threshold) && is.finite(gap_threshold)) {
    mean(dt > gap_threshold) * 100
  } else {
    NA_real_
  }
  list(
    median_dt = median_dt,
    iqr_dt = iqr_dt,
    n_intervals = n,
    pct_gap = pct_gap
  )
}

#' Per-person spacing stats: one row per id (median_dt, iqr_dt, n_intervals, pct_gap).
#' Returns tibble with columns id, median_dt, iqr_dt, n_intervals, pct_gap.
#' @noRd
ild_spacing_by_id <- function(data, id_col, dt_col, gap_threshold = NULL) {
  if (!nrow(data)) {
    return(tibble::tibble(
      id = data[[id_col]],
      median_dt = double(nrow(data)),
      iqr_dt = double(nrow(data)),
      n_intervals = integer(nrow(data)),
      pct_gap = double(nrow(data))
    ))
  }
  g <- dplyr::group_by(data, .data[[id_col]])
  out <- dplyr::summarise(g,
    id = .data[[id_col]][1],
    median_dt = stats::median(.data[[dt_col]], na.rm = TRUE),
    iqr_dt = stats::IQR(.data[[dt_col]], na.rm = TRUE),
    n_intervals = as.integer(sum(!is.na(.data[[dt_col]]))),
    pct_gap = if (!is.null(gap_threshold) && is.finite(gap_threshold)) {
      d <- .data[[dt_col]]
      d <- d[!is.na(d)]
      if (length(d) == 0) NA_real_ else mean(d > gap_threshold) * 100
    } else NA_real_,
    .groups = "drop"
  )
  out[[id_col]] <- NULL
  out$median_dt[is.nan(out$median_dt)] <- NA_real_
  out$iqr_dt[is.nan(out$iqr_dt)] <- NA_real_
  tibble::as_tibble(out)
}

#' Convert time column to numeric (seconds from epoch) for .ild_time_num
#' Handles Date, POSIXct, POSIXlt, and numeric (passed through).
#' @noRd
ild_time_to_num <- function(t) {
  if (is.numeric(t)) return(as.numeric(t))
  if (lubridate::is.Date(t)) {
    return(as.numeric(as.POSIXct(t)))
  }
  if (lubridate::is.POSIXt(t)) {
    return(as.numeric(t))
  }
  # try parsing as character to POSIXct
  t <- tryCatch(lubridate::as_datetime(t), error = function(e) NULL)
  if (is.null(t)) stop("Time column could not be converted to numeric (Date/POSIXct/numeric).", call. = FALSE)
  as.numeric(t)
}

#' Installed package version for provenance (internal)
#' @param pkg Character package name (e.g. \code{"lme4"}).
#' @return Character version or \code{NA_character_} if not installed.
#' @noRd
ild_backend_pkg_version <- function(pkg) {
  tryCatch(as.character(utils::packageVersion(pkg)), error = function(e) NA_character_)
}

#' Within-person and between-person decomposition (centering)
#'
#' For each selected variable, computes the person mean (between-person
#' component) and the within-person deviation (variable minus person mean).
#' Use `*_wp` at level-1 and `*_bp` at level-2 or in cross-level interactions
#' to avoid ecological fallacy and conflation bias. Selected variables must be numeric.
#'
#' @param x An ILD object (see [is_ild()]).
#' @param ... Variables to center (tidy-select). Unquoted names or a single
#'   character vector of column names. Must be numeric.
#' @param type Character. `"person_mean"` (default) for person-mean centering
#'   (x_bp, x_wp); `"grand_mean"` for grand-mean centering (x_gm, x_wp_gm).
#' @param naming Character. `"suffix"` (default): new columns \code{var_bp}, \code{var_wp};
#'   \code{"prefix"}: new columns \code{BP_var}, \code{WP_var}.
#' @return The same ILD tibble with additional columns. ILD attributes are preserved.
#' @importFrom dplyr group_by mutate select ungroup
#' @importFrom rlang .data
#' @export
ild_center <- function(x, ..., type = c("person_mean", "grand_mean"),
                      naming = c("suffix", "prefix")) {
  validate_ild(x)
  type <- match.arg(type)
  naming <- match.arg(naming)
  vars <- names(dplyr::select(x, ...))
  if (length(vars) == 0) stop("No variables selected for centering.", call. = FALSE)
  id_col <- ".ild_id"
  for (v in vars) {
    if (!v %in% names(x)) stop("Variable '", v, "' not found in data.", call. = FALSE)
    if (!is.numeric(x[[v]])) {
      stop("Variable '", v, "' is not numeric. Centering requires numeric variables.", call. = FALSE)
    }
  }
  out <- x
  created_nms <- character(0)
  for (v in vars) {
    out <- dplyr::group_by(out, .data[[id_col]])
    if (naming == "suffix") {
      bp_nm <- paste0(v, "_bp")
      wp_nm <- paste0(v, "_wp")
    } else {
      bp_nm <- paste0("BP_", v)
      wp_nm <- paste0("WP_", v)
    }
    out <- dplyr::mutate(out,
      !!bp_nm := mean(.data[[v]], na.rm = TRUE),
      !!wp_nm := .data[[v]] - mean(.data[[v]], na.rm = TRUE)
    )
    out <- dplyr::ungroup(out)
    created_nms <- c(created_nms, bp_nm, wp_nm)
    if (type == "grand_mean") {
      gm <- mean(out[[v]], na.rm = TRUE)
      out[[paste0(v, "_gm")]] <- gm
      out[[paste0(v, "_wp_gm")]] <- out[[v]] - gm
      created_nms <- c(created_nms, paste0(v, "_gm"), paste0(v, "_wp_gm"))
    }
  }
  out <- restore_ild_attrs(x, out)
  out <- ild_add_step(out, "ild_center",
    list(vars = vars, type = type, naming = naming),
    list(created = created_nms)
  )
  out
}

#' Sequential MSM inverse probability of treatment weights (IPTW for A_t)
#'
#' For **time-varying binary treatment** \eqn{A_t}, fits a separate treatment
#' propensity model at each occasion \eqn{t} (pooling persons at that time):
#' \eqn{P(A_t \mid \bar{L}_t)} where \eqn{\bar{L}_t} is supplied by the user
#' (e.g. lags from [ild_lag()], prior treatment, time-varying confounders). The
#' stabilized factor at each person-occasion is multiplied **within person**
#' over time to yield a **cumulative** MSM IPTW in \code{.ipw_treat}.
#'
#' This differs from [ild_iptw_weights()], which fits **one pooled** logistic
#' over all rows (appropriate for a single treatment decision or descriptive
#' pooling, not the standard sequential MSM estimand for \eqn{\bar{A}_K}).
#'
#' **Assumptions:** Sequential exchangeability given \eqn{\bar{L}_t}, positivity
#' at each \eqn{t}, and correct models. Variance (e.g. bootstrap by person) is
#' not computed here.
#'
#' @section Stabilization:
#' \describe{
#'   \item{\code{marginal}}{Numerator at each occasion uses the sample marginal
#'     \eqn{P(A_t = 1)} among rows at that time (same pattern as pooled stabilized
#'     IPTW).}
#'   \item{\code{prior_treatment}}{Numerator uses \eqn{P(A_t \mid A_{t-1})} from
#'     a logistic model \code{A_t ~ prior_treatment} fit at each occasion (pass
#'     \code{prior_treatment} column name). Denominator remains \eqn{P(A_t \mid \bar{L}_t)}.}
#'   \item{\code{none}}{Unstabilized \eqn{1/P(A_t \mid \bar{L}_t)} for the observed \eqn{A_t}.}
#' }
#'
#' @section Bundle integration:
#' Sets \code{.ipw_treat} for use with [ild_ipcw_weights()] and
#' [ild_joint_msm_weights()].
#'
#' @family ild_diagnostics_utilities
#' @param x An ILD object (see [is_ild()]).
#' @param treatment Character. Binary \eqn{A_t} column (\code{0}/\code{1} or two-level factor).
#' @param history Predictors \eqn{\bar{L}_t} for the **denominator** model: a
#'   \code{character} vector of column names, or a one-sided \code{formula}
#'   like \code{~ stress + trt_lag1}. Empty history fits an intercept-only model.
#' @param time_var Character. Occasion index column (default \code{".ild_seq"}).
#' @param stabilize Character: \code{"marginal"}, \code{"prior_treatment"}, or \code{"none"}.
#' @param prior_treatment Character. Required when \code{stabilize = "prior_treatment"}:
#'   name of the \eqn{A_{t-1}} (or prior treatment) column for the numerator model.
#' @param trim Numeric of length 2. Quantiles applied to the **final** cumulative
#'   \code{.ipw_treat} (default \code{c(0.01, 0.99)}). Use length-0 numeric to skip.
#' @param ... Passed to \code{stats::glm()} for each occasion (denominator model;
#'   numerator model when \code{prior_treatment} is used also uses \code{...}).
#' @return \code{x} with \code{.ipw_treat} set to the cumulative sequential weight.
#'   Attributes: \code{ild_iptw_msm_fits} (named list of denominator \code{glm} fits
#'   by occasion), \code{ild_iptw_msm_numerator_fits} (same for numerator when
#'   \code{prior_treatment}).
#' @seealso [ild_iptw_weights()], [ild_ipcw_weights()], [ild_joint_msm_weights()]
#' @export
#' @examples
#' set.seed(11)
#' d <- ild_simulate(n_id = 20, n_obs_per = 6, seed = 11)
#' d$stress <- rnorm(nrow(d))
#' d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.4))
#' d <- ild_prepare(d, id = "id", time = "time")
#' d <- ild_lag(d, stress, mode = "gap_aware", max_gap = Inf)
#' d <- ild_lag(d, trt, mode = "gap_aware", max_gap = Inf)
#' x <- ild_iptw_msm_weights(d, treatment = "trt", history = ~ stress_lag1 + trt_lag1)
#' summary(x$.ipw_treat)
ild_iptw_msm_weights <- function(x,
                                 treatment,
                                 history,
                                 time_var = ".ild_seq",
                                 stabilize = c("marginal", "prior_treatment", "none"),
                                 prior_treatment = NULL,
                                 trim = c(0.01, 0.99),
                                 ...) {
  validate_ild(x)
  stabilize <- match.arg(stabilize)
  if (stabilize == "prior_treatment" && (is.null(prior_treatment) || !nzchar(prior_treatment))) {
    stop("prior_treatment must name a column when stabilize = \"prior_treatment\".", call. = FALSE)
  }
  nms <- names(x)
  if (!treatment %in% nms) stop("treatment '", treatment, "' not found in data.", call. = FALSE)
  if (!time_var %in% nms) stop("time_var '", time_var, "' not found in data.", call. = FALSE)
  if (!is.null(prior_treatment) && !prior_treatment %in% nms) {
    stop("prior_treatment '", prior_treatment, "' not found in data.", call. = FALSE)
  }

  pred_names <- if (inherits(history, "formula")) {
    tryCatch(
      attr(stats::terms(history), "term.labels"),
      error = function(e) character(0)
    )
  } else {
    history
  }
  for (p in pred_names) {
    if (!p %in% nms) stop("predictor '", p, "' not found in data.", call. = FALSE)
  }

  id <- x[[".ild_id"]]
  seqv <- x[[time_var]]
  Araw <- x[[treatment]]
  if (is.factor(Araw)) {
    if (nlevels(Araw) != 2L) {
      stop("Treatment must be binary (two-level factor).", call. = FALSE)
    }
    A <- as.integer(Araw) - 1L
  } else {
    A <- as.numeric(Araw)
    u <- unique(stats::na.omit(A))
    if (length(u) > 2L || !all(u %in% c(0, 1))) {
      stop("Treatment must be binary (0/1 or two-level factor).", call. = FALSE)
    }
  }

  n <- nrow(x)
  r <- rep(NA_real_, n)
  fits_den <- list()
  fits_num <- list()

  rhs_den <- if (length(pred_names) > 0L) paste(pred_names, collapse = " + ") else "1"
  form_den <- stats::as.formula(paste(treatment, "~", rhs_den))

  occ <- sort(unique(as.numeric(seqv)))
  for (t in occ) {
    idx <- which(as.numeric(seqv) == t & !is.na(A))
    if (length(idx) < 2L) next # need variation to fit; rows left NA then neutralized in cumprod
    df <- as.data.frame(x[idx, c(treatment, pred_names), drop = FALSE])
    df[[treatment]] <- A[idx]
    fit_den <- tryCatch(
      stats::glm(form_den, data = df, family = stats::binomial(), ...),
      error = function(e) NULL
    )
    if (is.null(fit_den)) next
    fits_den[[as.character(t)]] <- fit_den
    p_hat <- stats::predict(fit_den, newdata = as.data.frame(x[idx, , drop = FALSE]), type = "response")
    p_hat <- pmax(pmin(p_hat, 1 - 1e-6), 1e-6)
    Ai <- A[idx]

    if (stabilize == "none") {
      ri <- ifelse(Ai >= 0.5, 1 / p_hat, 1 / (1 - p_hat))
    } else if (stabilize == "marginal") {
      p1 <- mean(Ai, na.rm = TRUE)
      p1 <- pmax(pmin(p1, 1 - 1e-6), 1e-6)
      ri <- ifelse(Ai >= 0.5, p1 / p_hat, (1 - p1) / (1 - p_hat))
    } else {
      form_num <- stats::as.formula(paste(treatment, "~", prior_treatment))
      df_num <- as.data.frame(x[idx, c(treatment, prior_treatment), drop = FALSE])
      df_num[[treatment]] <- A[idx]
      fit_num <- tryCatch(
        stats::glm(form_num, data = df_num, family = stats::binomial(), ...),
        error = function(e) NULL
      )
      if (is.null(fit_num)) next
      fits_num[[as.character(t)]] <- fit_num
      num_p <- stats::predict(fit_num, newdata = as.data.frame(x[idx, , drop = FALSE]), type = "response")
      num_p <- pmax(pmin(num_p, 1 - 1e-6), 1e-6)
      ri <- ifelse(Ai >= 0.5, num_p / p_hat, (1 - num_p) / (1 - p_hat))
    }
    r[idx] <- ri
  }

  if (length(fits_den) == 0L) {
    stop(
      "No occasion-level treatment models were fitted (need >=2 rows with non-NA treatment per occasion, or glm failed).",
      call. = FALSE
    )
  }

  ord <- order(id, seqv, seq_len(n))
  ido <- id[ord]
  r_ord <- r[ord]
  w_ord <- rep(NA_real_, n)
  i <- 1L
  while (i <= n) {
    g <- ido[i]
    j <- i
    while (j < n && ido[j + 1L] == g) j <- j + 1L
    idx <- i:j
    rr <- r_ord[idx]
    rr[is.na(rr)] <- 1
    w_ord[idx] <- cumprod(rr)
    i <- j + 1L
  }
  w <- rep(NA_real_, n)
  w[ord] <- w_ord
  if (length(trim) >= 2L && is.numeric(trim)) {
    w <- .ild_ipw_trim(w, trim)
  }
  attrs <- attributes(x)
  x$.ipw_treat <- w
  x <- .ild_ipw_restore_attrs(x, attrs)
  attr(x, "ild_iptw_msm_fits") <- fits_den
  attr(x, "ild_iptw_msm_numerator_fits") <- if (length(fits_num) > 0L) fits_num else NULL
  hist_arg <- if (inherits(history, "formula")) {
    paste(deparse(history, width.cutoff = 500L), collapse = " ")
  } else {
    history
  }
  ild_add_step(x, "ild_iptw_msm_weights",
    list(
      treatment = treatment, history = hist_arg, time_var = time_var,
      stabilize = stabilize, prior_treatment = prior_treatment, trim = trim
    ),
    list(created = ".ipw_treat", mode = "sequential_msm")
  )
}

ctsem_fixture_data <- function(seed = 9401L, n_id = 1L, n_obs_per = 60L) {
  d <- ild_simulate(n_id = n_id, n_obs_per = n_obs_per, seed = seed)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  x
}

# Tidiers for ild_kfas objects (KFAS backend)

#' @rdname ild_tidy
#' @method ild_tidy ild_fit_kfas
#' @export
ild_tidy.ild_fit_kfas <- function(x, ...) {
  rlang::check_installed("KFAS", reason = "to use ild_tidy() for ild_kfas fits")
  fm <- x$kfas_model
  rows <- list()
  # Observation and state innovation variances (Gaussian local level)
  qv <- ild_kfas_qh_scalar(fm$Q)
  hv <- ild_kfas_qh_scalar(fm$H)
  if (is.finite(qv)) {
    rows[[length(rows) + 1L]] <- list(
      term = "level_variance",
      component = "variance",
      effect_level = "auxiliary",
      estimate = qv,
      std_error = NA_real_,
      conf_low = NA_real_,
      conf_high = NA_real_,
      p_value = NA_real_,
      statistic = NA_real_,
      interval_type = "none"
    )
  }
  if (is.finite(hv)) {
    rows[[length(rows) + 1L]] <- list(
      term = "observation_variance",
      component = "variance",
      effect_level = "auxiliary",
      estimate = hv,
      std_error = NA_real_,
      conf_low = NA_real_,
      conf_high = NA_real_,
      p_value = NA_real_,
      statistic = NA_real_,
      interval_type = "none"
    )
  }
  if (length(rows) == 0L) {
    return(
      ild_tidy_assemble(
        term = "(no_variance_extracted)",
        estimate = NA_real_,
        std_error = NA_real_,
        conf_low = NA_real_,
        conf_high = NA_real_,
        p_value = NA_real_,
        statistic = NA_real_,
        interval_type = "none",
        engine = "KFAS",
        model_class = "ild_fit_kfas",
        component = "auxiliary",
        effect_level = "auxiliary",
        optional = NULL
      )
    )
  }
  n <- length(rows)
  term <- vapply(rows, function(z) z$term, character(1L))
  comp <- vapply(rows, function(z) z$component, character(1L))
  el <- vapply(rows, function(z) z$effect_level, character(1L))
  est <- vapply(rows, function(z) z$estimate, numeric(1L))
  se <- rep(NA_real_, n)
  cl <- rep(NA_real_, n)
  ch <- rep(NA_real_, n)
  pv <- rep(NA_real_, n)
  st <- rep(NA_real_, n)
  int <- vapply(rows, function(z) z$interval_type, character(1L))
  ild_tidy_assemble(
    term = term,
    estimate = est,
    std_error = se,
    conf_low = cl,
    conf_high = ch,
    p_value = pv,
    statistic = st,
    interval_type = int,
    engine = "KFAS",
    model_class = "ild_fit_kfas",
    component = comp,
    effect_level = el,
    optional = NULL
  )
}

#' Tidy per-time state summaries for ild_kfas
#'
#' Returns a tibble with one row per time index and columns for the smoothed
#' level (and optional variance). Separate from [ild_tidy()] which tidies
#' **parameters**.
#'
#' @param x An object from [ild_kfas()].
#' @param ... Reserved.
#' @return A \code{\link[tibble]{tibble}}.
#' @export
ild_tidy_states <- function(x, ...) {
  UseMethod("ild_tidy_states")
}

#' @export
ild_tidy_states.ild_fit_kfas <- function(x, ...) {
  rlang::check_installed("KFAS", reason = "to use ild_tidy_states()")
  kfs <- x$kfs
  ah <- kfs$alphahat
  V <- kfs$V
  n <- if (!is.null(ah) && is.matrix(ah)) nrow(ah) else 0L
  if (n == 0L) {
    return(tibble::tibble(
      time_index = integer(),
      state = character(),
      estimate = numeric(),
      std_error = numeric()
    ))
  }
  lev <- as.numeric(ah[, 1L])
  se <- rep(NA_real_, n)
  if (!is.null(V) && length(dim(V)) >= 2L) {
    # diagonal of state covariance per time (approximate)
    for (i in seq_len(n)) {
      se[i] <- tryCatch(sqrt(V[1L, 1L, i]), error = function(e) NA_real_)
    }
  }
  tibble::tibble(
    time_index = seq_len(n),
    state = rep("level", n),
    estimate = lev,
    std_error = se
  )
}

#' Fit a model separately per person (N-of-1 / idiographic)
#'
#' Splits the ILD by person and fits the same formula (e.g. \code{lm}) within each.
#' Returns a tibble of person-level estimates for teaching, N-of-1 analysis, or
#' inspecting heterogeneity. Use [ild_person_distribution()] to visualize the
#' distribution of estimates across persons.
#'
#' @param formula A formula (e.g. \code{y ~ x}). Used for each person's \code{lm}.
#' @param data An ILD object (see [is_ild()]).
#' @param method Character. Currently only \code{"lm"} (default).
#' @param min_obs Integer. Minimum observations per person to fit (default 2).
#'   Persons with fewer are omitted or get NA rows.
#' @return A tibble with columns \code{.ild_id} (or the id column name from metadata),
#'   \code{term}, \code{estimate}, \code{std_error}, \code{p_value}, and optionally
#'   \code{sigma}, \code{n_obs}. One row per person per term (long format).
#' @export
#' @examples
#' d <- ild_simulate(n_id = 5, n_obs_per = 8, seed = 1)
#' x <- ild_prepare(d, id = "id", time = "time")
#' pm <- ild_person_model(y ~ 1, x)
#' ild_person_distribution(pm, term = "(Intercept)")
ild_person_model <- function(formula, data, method = c("lm"), min_obs = 2L) {
  validate_ild(data)
  method <- match.arg(method)
  min_obs <- as.integer(min_obs)[1L]
  if (min_obs < 1L) min_obs <- 1L
  id_col <- ".ild_id"
  ids <- unique(data[[id_col]])
  out_rows <- list()
  for (i in seq_along(ids)) {
    id_val <- ids[i]
    sub <- data[data[[id_col]] == id_val, ]
    if (nrow(sub) < min_obs) next
    fit <- tryCatch(stats::lm(formula, data = sub), error = function(e) NULL)
    if (is.null(fit)) next
    co <- summary(fit)$coefficients
    if (is.null(co) || nrow(co) == 0) next
    for (j in seq_len(nrow(co))) {
      out_rows[[length(out_rows) + 1L]] <- list(
        .ild_id = id_val,
        term = rownames(co)[j],
        estimate = co[j, "Estimate"],
        std_error = co[j, "Std. Error"],
        p_value = co[j, "Pr(>|t|)"],
        sigma = stats::sigma(fit),
        n_obs = nrow(sub)
      )
    }
  }
  if (length(out_rows) == 0) {
    return(tibble::tibble(
      .ild_id = data[[id_col]][0],
      term = character(),
      estimate = double(),
      std_error = double(),
      p_value = double(),
      sigma = double(),
      n_obs = integer()
    ))
  }
  tibble::as_tibble(do.call(rbind, lapply(out_rows, as.data.frame, stringsAsFactors = FALSE)))
}

#' Plot distribution of person-level estimates from ild_person_model
#'
#' Draws a histogram or density of the selected term's estimates across persons.
#' Useful to visualize heterogeneity (e.g. distribution of slopes or intercepts).
#'
#' @param person_fit Tibble returned by [ild_person_model()] (columns \code{term}, \code{estimate}, etc.).
#' @param term Character. Which term to plot (e.g. \code{"(Intercept)"} or a covariate name).
#'   If \code{NULL}, the first term in the table is used.
#' @param type Character. \code{"histogram"} (default) or \code{"density"}.
#' @return A ggplot object.
#' @export
ild_person_distribution <- function(person_fit, term = NULL, type = c("histogram", "density")) {
  type <- match.arg(type)
  if (!inherits(person_fit, "data.frame") || !"estimate" %in% names(person_fit)) {
    stop("person_fit must be a tibble from ild_person_model() with an 'estimate' column.", call. = FALSE)
  }
  if (!"term" %in% names(person_fit)) {
    stop("person_fit must have a 'term' column.", call. = FALSE)
  }
  terms_avail <- unique(person_fit$term)
  if (length(terms_avail) == 0) stop("No terms in person_fit.", call. = FALSE)
  if (is.null(term)) term <- terms_avail[1]
  if (!term %in% terms_avail) stop("term '", term, "' not found in person_fit.", call. = FALSE)
  sub <- person_fit[person_fit$term == term, ]
  est <- sub$estimate[!is.na(sub$estimate)]
  if (length(est) == 0) stop("No non-NA estimates for term '", term, "'.", call. = FALSE)
  df <- data.frame(estimate = est)
  if (type == "histogram") {
    p <- ggplot2::ggplot(df, ggplot2::aes(x = .data$estimate)) +
      ggplot2::geom_histogram(bins = min(30L, max(5L, length(est) %/% 2)), fill = "gray70", color = "gray40") +
      ggplot2::labs(x = "Estimate", y = "Count", title = paste0("Distribution of ", term, " (by person)")) +
      ggplot2::theme_minimal()
  } else {
    p <- ggplot2::ggplot(df, ggplot2::aes(x = .data$estimate)) +
      ggplot2::geom_density(fill = "gray70", alpha = 0.6) +
      ggplot2::labs(x = "Estimate", y = "Density", title = paste0("Distribution of ", term, " (by person)")) +
      ggplot2::theme_minimal()
  }
  p
}

test_that("ild_person_model returns tibble with term, estimate, std_error, p_value", {
  d <- ild_simulate(n_id = 6, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  pm <- ild_person_model(y ~ 1, x)
  expect_s3_class(pm, "tbl_df")
  expect_true(all(c(".ild_id", "term", "estimate", "std_error", "p_value") %in% names(pm)))
  expect_equal(unique(pm$term), "(Intercept)")
  expect_equal(nrow(pm), 6)
})

test_that("ild_person_model with predictor returns one row per person per term", {
  d <- ild_simulate(n_id = 4, n_obs_per = 6, seed = 2)
  x <- ild_prepare(d, id = "id", time = "time")
  pm <- ild_person_model(y ~ 1, x)
  expect_gte(nrow(pm), 4)
  expect_true(is.numeric(pm$estimate))
})

test_that("ild_person_model skips persons with too few obs when min_obs > n_obs", {
  d <- ild_simulate(n_id = 3, n_obs_per = 2, seed = 3)
  x <- ild_prepare(d, id = "id", time = "time")
  pm <- ild_person_model(y ~ 1, x, min_obs = 3)
  expect_equal(nrow(pm), 0)
})

test_that("ild_person_distribution returns ggplot", {
  d <- ild_simulate(n_id = 8, n_obs_per = 5, seed = 4)
  x <- ild_prepare(d, id = "id", time = "time")
  pm <- ild_person_model(y ~ 1, x)
  p <- ild_person_distribution(pm, term = "(Intercept)")
  expect_s3_class(p, "gg")
  p2 <- ild_person_distribution(pm, term = "(Intercept)", type = "density")
  expect_s3_class(p2, "gg")
})

# ild_diagnose() S3 methods — all return ild_diagnostics_bundle

#' @keywords internal
#' @noRd
ild_model_predictor_names <- function(object, data) {
  f <- tryCatch(stats::formula(object), error = function(e) NULL)
  if (is.null(f)) return(character())
  intersect(all.vars(f), names(data))
}

#' @keywords internal
#' @noRd
ild_diagnose_outcome_vars <- function(object, data, preds) {
  rn <- tryCatch(ild_response_name(object), error = function(e) NULL)
  ov <- unique(c(stats::na.omit(c(rn, preds))))
  ov[ov %in% names(data)]
}

#' @rdname ild_diagnose
#' @method ild_diagnose lmerMod
#' @export
ild_diagnose.lmerMod <- function(object,
                                 data = NULL,
                                 type = c("residual_acf", "residual_time", "qq"),
                                 by_id = TRUE,
                                 missing_model = FALSE,
                                 missing_model_predictors = NULL,
                                 causal_detail = FALSE,
                                 balance = FALSE,
                                 balance_treatment = NULL,
                                 balance_covariates = NULL,
                                 balance_weights_col = ".ipw_treat",
                                 balance_by_occasion = FALSE,
                                 ...) {
  if (is.null(data)) data <- attr(object, "ild_data", exact = TRUE)
  if (is.null(data)) stop("Provide ILD data or refit with ild_lme() so the model has ild_data.", call. = FALSE)
  validate_ild(data)
  type <- match.arg(type, choices = c("residual_acf", "residual_time", "qq"), several.ok = TRUE)
  preds <- ild_model_predictor_names(object, data)
  outcome_vars <- ild_diagnose_outcome_vars(object, data, preds)
  fd <- fill_diagnostics_fit_lmerMod(object)
  wg <- collect_freq_warnings_guardrails(object, fd)
  miss <- fill_diagnostics_missingness_section(data, vars = preds)
  if (isTRUE(missing_model)) {
    mpreds <- if (is.null(missing_model_predictors)) preds else missing_model_predictors
    rn <- tryCatch(ild_response_name(object), error = function(e) NULL)
    if (length(mpreds) > 0L && !is.null(rn) && rn %in% names(data)) {
      mm <- tryCatch(ild_missing_model(data, outcome = rn, predictors = mpreds), error = function(e) NULL)
      if (!is.null(mm)) {
        miss$missing_model_diagnostic <- list(
          outcome = mm$outcome,
          predictors = mm$predictors,
          tidy = mm$tidy,
          message = mm$message
        )
      }
    } else {
      miss$missing_model_diagnostic <- list(
        message = "missing_model requires a response column in data and at least one predictor (set missing_model_predictors if formula has none)."
      )
    }
  }
  bundle <- ild_diagnostics_bundle(
    meta = list(
      engine = "lmer",
      n_obs = nrow(data),
      n_id = length(unique(data[[".ild_id"]])),
      ar1 = isTRUE(attr(object, "ild_ar1", exact = TRUE))
    ),
    data = fill_diagnostics_data(data, outcome_vars = outcome_vars),
    design = fill_diagnostics_design(data, vars = if (length(preds)) preds else NULL),
    fit = fd,
    residual = fill_diagnostics_residual_frequentist(object, data, type, by_id),
    predictive = fill_diagnostics_predictive_frequentist(object, data),
    missingness = miss,
    causal = fill_diagnostics_causal(
      data,
      causal_detail = causal_detail,
      balance = balance,
      balance_treatment = balance_treatment,
      balance_covariates = balance_covariates,
      balance_weights_col = balance_weights_col,
      balance_by_occasion = balance_by_occasion
    ),
    warnings = wg$warnings,
    guardrails = guardrails_empty_tibble(),
    summary_text = character()
  )
  bundle$design$occasion_imbalance <- bundle$data$obs_per_id
  bundle$guardrails <- guardrails_bind(
    evaluate_guardrails_fit(object, fd, engine = "lmer"),
    evaluate_guardrails_contextual(object, data, bundle, engine = "lmer")
  )
  bundle$summary_text <- build_diagnostics_bundle_summary(bundle)
  bundle <- enrich_bundle_semantic_sections(bundle)
  attr(bundle, "ild_fit") <- object
  attr(bundle, "ild_data") <- data
  bundle
}

#' @rdname ild_diagnose
#' @method ild_diagnose lme
#' @export
ild_diagnose.lme <- function(object,
                             data = NULL,
                             type = c("residual_acf", "residual_time", "qq"),
                             by_id = TRUE,
                             missing_model = FALSE,
                             missing_model_predictors = NULL,
                             causal_detail = FALSE,
                             balance = FALSE,
                             balance_treatment = NULL,
                             balance_covariates = NULL,
                             balance_weights_col = ".ipw_treat",
                             balance_by_occasion = FALSE,
                             ...) {
  if (is.null(data)) data <- attr(object, "ild_data", exact = TRUE)
  if (is.null(data)) stop("Provide ILD data or refit with ild_lme() so the model has ild_data.", call. = FALSE)
  validate_ild(data)
  type <- match.arg(type, choices = c("residual_acf", "residual_time", "qq"), several.ok = TRUE)
  preds <- ild_model_predictor_names(object, data)
  outcome_vars <- ild_diagnose_outcome_vars(object, data, preds)
  fd <- fill_diagnostics_fit_lme(object)
  wg <- collect_freq_warnings_guardrails(object, fd)
  miss <- fill_diagnostics_missingness_section(data, vars = preds)
  if (isTRUE(missing_model)) {
    mpreds <- if (is.null(missing_model_predictors)) preds else missing_model_predictors
    rn <- tryCatch(ild_response_name(object), error = function(e) NULL)
    if (length(mpreds) > 0L && !is.null(rn) && rn %in% names(data)) {
      mm <- tryCatch(ild_missing_model(data, outcome = rn, predictors = mpreds), error = function(e) NULL)
      if (!is.null(mm)) {
        miss$missing_model_diagnostic <- list(
          outcome = mm$outcome,
          predictors = mm$predictors,
          tidy = mm$tidy,
          message = mm$message
        )
      }
    } else {
      miss$missing_model_diagnostic <- list(
        message = "missing_model requires a response column in data and at least one predictor (set missing_model_predictors if formula has none)."
      )
    }
  }
  bundle <- ild_diagnostics_bundle(
    meta = list(
      engine = "lme",
      n_obs = nrow(data),
      n_id = length(unique(data[[".ild_id"]])),
      ar1 = isTRUE(attr(object, "ild_ar1", exact = TRUE))
    ),
    data = fill_diagnostics_data(data, outcome_vars = outcome_vars),
    design = fill_diagnostics_design(data, vars = if (length(preds)) preds else NULL),
    fit = fd,
    residual = fill_diagnostics_residual_frequentist(object, data, type, by_id),
    predictive = fill_diagnostics_predictive_frequentist(object, data),
    missingness = miss,
    causal = fill_diagnostics_causal(
      data,
      causal_detail = causal_detail,
      balance = balance,
      balance_treatment = balance_treatment,
      balance_covariates = balance_covariates,
      balance_weights_col = balance_weights_col,
      balance_by_occasion = balance_by_occasion
    ),
    warnings = wg$warnings,
    guardrails = guardrails_empty_tibble(),
    summary_text = character()
  )
  bundle$design$occasion_imbalance <- bundle$data$obs_per_id
  bundle$guardrails <- guardrails_bind(
    evaluate_guardrails_fit(object, fd, engine = "lme"),
    evaluate_guardrails_contextual(object, data, bundle, engine = "lme")
  )
  bundle$summary_text <- build_diagnostics_bundle_summary(bundle)
  bundle <- enrich_bundle_semantic_sections(bundle)
  attr(bundle, "ild_fit") <- object
  attr(bundle, "ild_data") <- data
  bundle
}

#' @rdname ild_diagnose
#' @method ild_diagnose brmsfit
#' @export
ild_diagnose.brmsfit <- function(object,
                                 data = NULL,
                                 type = c("all", "convergence", "sampler", "ppc"),
                                 by_id = TRUE,
                                 ppc_ndraws = 500L,
                                 missing_model = FALSE,
                                 missing_model_predictors = NULL,
                                 causal_detail = FALSE,
                                 balance = FALSE,
                                 balance_treatment = NULL,
                                 balance_covariates = NULL,
                                 balance_weights_col = ".ipw_treat",
                                 balance_by_occasion = FALSE,
                                 ...) {
  if (is.null(attr(object, "ild_data", exact = TRUE))) {
    stop("ild_diagnose() for brmsfit requires ild_brms() (ild_data attribute missing).", call. = FALSE)
  }
  rlang::check_installed("brms")
  data <- attr(object, "ild_data", exact = TRUE)
  validate_ild(data)
  if (length(type) == 1L) {
    type <- match.arg(type, choices = c("all", "convergence", "sampler", "ppc"))
  }
  types <- if (identical(type, "all")) {
    c("convergence", "sampler", "ppc")
  } else {
    match.arg(type, c("convergence", "sampler", "ppc"), several.ok = TRUE)
  }
  preds <- ild_model_predictor_names(object, data)
  outcome_vars <- ild_diagnose_outcome_vars(object, data, preds)
  ffit <- fill_diagnostics_fit_brms(object)
  pred_sec <- if ("ppc" %in% types) {
    fill_diagnostics_predictive_brms(object, ppc_ndraws = ppc_ndraws)
  } else {
    fill_diagnostics_predictive_brms_augment_only(object)
  }
  wg <- collect_brms_warnings_guardrails(object, ffit)
  miss <- fill_diagnostics_missingness_section(data, vars = preds)
  if (isTRUE(missing_model)) {
    mpreds <- if (is.null(missing_model_predictors)) preds else missing_model_predictors
    rn <- tryCatch(ild_response_name(object), error = function(e) NULL)
    if (length(mpreds) > 0L && !is.null(rn) && rn %in% names(data)) {
      mm <- tryCatch(ild_missing_model(data, outcome = rn, predictors = mpreds), error = function(e) NULL)
      if (!is.null(mm)) {
        miss$missing_model_diagnostic <- list(
          outcome = mm$outcome,
          predictors = mm$predictors,
          tidy = mm$tidy,
          message = mm$message
        )
      }
    } else {
      miss$missing_model_diagnostic <- list(
        message = "missing_model requires a response column in data and at least one predictor (set missing_model_predictors if formula has none)."
      )
    }
  }
  bundle <- ild_diagnostics_bundle(
    meta = list(engine = "brms", brms_types = types, n_obs = nrow(data), n_id = length(unique(data[[".ild_id"]]))),
    data = fill_diagnostics_data(data, outcome_vars = outcome_vars),
    design = fill_diagnostics_design(data, vars = if (length(preds)) preds else NULL),
    fit = ffit,
    residual = fill_diagnostics_residual_brms(object),
    predictive = pred_sec,
    missingness = miss,
    causal = fill_diagnostics_causal(
      data,
      causal_detail = causal_detail,
      balance = balance,
      balance_treatment = balance_treatment,
      balance_covariates = balance_covariates,
      balance_weights_col = balance_weights_col,
      balance_by_occasion = balance_by_occasion
    ),
    warnings = wg$warnings,
    guardrails = guardrails_empty_tibble(),
    summary_text = character()
  )
  bundle$design$occasion_imbalance <- bundle$data$obs_per_id
  bundle$guardrails <- guardrails_bind(
    evaluate_guardrails_fit(object, ffit, engine = "brms"),
    evaluate_guardrails_contextual(object, data, bundle, engine = "brms")
  )
  bundle$summary_text <- c(
    build_diagnostics_bundle_summary(bundle),
    build_diagnostics_bundle_summary_brms(bundle, types)
  )
  bundle <- enrich_bundle_semantic_sections(bundle)
  attr(bundle, "ild_fit") <- object
  attr(bundle, "ild_data") <- data
  bundle
}

#' @keywords internal
#' @noRd
build_diagnostics_bundle_summary_brms <- function(bundle, types) {
  parts <- character()
  ps <- bundle$fit$posterior_summary$ild_posterior %||% bundle$fit$ild_posterior
  if ("sampler" %in% types && !is.null(ps)) {
    parts <- c(parts, sprintf(
      "Sampler: %s chains, iter = %s, warmup = %s, divergent = %s.",
      ps$chains, ps$iter, ps$warmup,
      if (is.na(ps$n_divergent)) "NA" else ps$n_divergent
    ))
  }
  ct <- bundle$fit$convergence$convergence_table %||% bundle$fit$convergence_table
  if ("convergence" %in% types && !is.null(ct)) {
    if (nrow(ct) > 0L) {
      mx <- suppressWarnings(max(ct$rhat, na.rm = TRUE))
      parts <- c(parts, sprintf("Max R-hat (fixed): %.3f.", mx))
    }
  }
  parts
}

#' @rdname ild_diagnose
#' @method ild_diagnose ild_fit_ctsem
#' @export
ild_diagnose.ild_fit_ctsem <- function(object,
                                       data = NULL,
                                       type = NULL,
                                       by_id = NULL,
                                       missing_model = FALSE,
                                       missing_model_predictors = NULL,
                                       causal_detail = FALSE,
                                       balance = FALSE,
                                       balance_treatment = NULL,
                                       balance_covariates = NULL,
                                       balance_weights_col = ".ipw_treat",
                                       balance_by_occasion = FALSE,
                                       ...) {
  rlang::check_installed("ctsem", reason = "to diagnose ild_ctsem fits")
  data <- if (is.null(data)) attr(object, "ild_data", exact = TRUE) else data
  if (is.null(data)) {
    stop("ild_diagnose() for ild_fit_ctsem requires data or attr(fit, \"ild_data\").", call. = FALSE)
  }
  validate_ild(data)
  rn <- attr(object, "ild_response", exact = TRUE)
  outcome_vars <- if (!is.null(rn) && rn %in% names(data)) rn else character()
  fd <- fill_diagnostics_fit_ctsem(object)
  miss <- fill_diagnostics_missingness_section(data, vars = outcome_vars)
  bundle <- ild_diagnostics_bundle(
    meta = list(
      engine = "ctsem",
      n_obs = nrow(data),
      n_id = length(unique(data[[".ild_id"]]))
    ),
    data = fill_diagnostics_data(data, outcome_vars = outcome_vars),
    design = fill_diagnostics_design(data, vars = NULL),
    fit = fd,
    residual = fill_diagnostics_residual_ctsem(object, data),
    predictive = fill_diagnostics_predictive_ctsem(object),
    missingness = miss,
    causal = fill_diagnostics_causal(
      data,
      causal_detail = causal_detail,
      balance = balance,
      balance_treatment = balance_treatment,
      balance_covariates = balance_covariates,
      balance_weights_col = balance_weights_col,
      balance_by_occasion = balance_by_occasion
    ),
    warnings = tibble::tibble(),
    guardrails = guardrails_empty_tibble(),
    summary_text = character()
  )
  bundle$design$occasion_imbalance <- bundle$data$obs_per_id
  bundle$guardrails <- guardrails_bind(
    evaluate_guardrails_fit(object, fd, engine = "ctsem"),
    evaluate_guardrails_contextual(object, data, bundle, engine = "ctsem")
  )
  bundle$summary_text <- build_diagnostics_bundle_summary(bundle)
  bundle <- enrich_bundle_semantic_sections(bundle)
  attr(bundle, "ild_fit") <- object
  attr(bundle, "ild_data") <- data
  bundle
}

#' @rdname ild_diagnose
#' @method ild_diagnose default
#' @export
ild_diagnose.default <- function(object, ...) {
  stop(
    "No ild_diagnose() method for class ",
    paste(class(object), collapse = ", "),
    ". Use lmerMod, lme, brmsfit, ild_fit_kfas, or ild_fit_ctsem.",
    call. = FALSE
  )
}

#' Cross-lag model: lag predictor then fit outcome ~ lag
#'
#' One-call pipeline: [ild_lag()] the predictor, [ild_check_lags()] to audit,
#' then [ild_lme()] to fit outcome on the lagged predictor. Returns the fit,
#' the lag-term coefficient (estimate, CI, p), and lag validity check.
#'
#' @param data An ILD object (see [is_ild()]).
#' @param outcome Character or symbol. Name of the outcome variable (e.g. \code{"mood"}).
#' @param predictor Character or symbol. Name of the predictor to lag (e.g. \code{"stress"}).
#' @param lag Integer. Lag order (default 1).
#' @param mode Character. Passed to [ild_lag()]: \code{"index"}, \code{"gap_aware"} (default), or \code{"time_window"}.
#' @param max_gap Numeric or NULL. Passed to [ild_lag()] when \code{mode = "gap_aware"}.
#' @param ar1 Logical. If \code{TRUE}, fit with [ild_lme()] using AR1/CAR1 (default \code{FALSE}).
#' @param include_diagnostics Logical. If \code{TRUE}, attach [ild_diagnostics()] to the return (default \code{FALSE}).
#' @param ... Passed to [ild_lme()].
#' @return A list: \code{fit} (fitted model), \code{lag_term} (one-row tibble from [tidy_ild_model()] for the lag variable),
#'   \code{lag_check} (tibble from [ild_check_lags()]), \code{data} (ILD with lag column), and optionally \code{diagnostics}.
#' @export
#' @examples
#' d <- ild_simulate(n_id = 10, n_obs_per = 8, seed = 1)
#' x <- ild_prepare(d, id = "id", time = "time")
#' out <- ild_crosslag(x, "y", "y", lag = 1, ar1 = FALSE, warn_no_ar1 = FALSE)
#' out$lag_term
#' out$lag_check
ild_crosslag <- function(data,
                         outcome,
                         predictor,
                         lag = 1L,
                         mode = c("gap_aware", "index", "time_window"),
                         max_gap = NULL,
                         ar1 = FALSE,
                         include_diagnostics = FALSE,
                         ...) {
  validate_ild(data)
  mode <- match.arg(mode)
  outcome <- as.character(rlang::ensym(outcome))
  predictor <- as.character(rlang::ensym(predictor))
  if (!outcome %in% names(data)) stop("outcome '", outcome, "' not found in data.", call. = FALSE)
  if (!predictor %in% names(data)) stop("predictor '", predictor, "' not found in data.", call. = FALSE)
  lag <- as.integer(lag)[1L]
  if (lag < 1L) stop("'lag' must be >= 1.", call. = FALSE)
  id_name <- ild_meta(data)$ild_id
  lag_var_name <- paste0(predictor, "_lag", lag)
  data_lagged <- ild_lag(data, dplyr::all_of(predictor), n = lag, mode = mode, max_gap = max_gap)
  lag_check <- ild_check_lags(data_lagged, lag_vars = lag_var_name, max_gap = max_gap)
  f <- stats::as.formula(paste0(outcome, " ~ ", lag_var_name, " + (1|", id_name, ")"))
  dots <- list(...)
  dots$warn_no_ar1 <- NULL
  dots$warn_uncentered <- NULL
  fit <- do.call(ild_lme, c(list(formula = f, data = data_lagged, ar1 = ar1, warn_no_ar1 = FALSE, warn_uncentered = FALSE), dots))
  tidy_tbl <- tidy_ild_model(fit, object = FALSE)
  lag_term <- tidy_tbl[tidy_tbl$term == lag_var_name, ]
  if (nrow(lag_term) == 0) lag_term <- tidy_tbl[0, ]
  out <- list(
    fit = fit,
    lag_term = lag_term,
    lag_check = lag_check,
    data = data_lagged
  )
  attr(out, "ild_provenance") <- ild_new_analysis_provenance(
    data_lagged,
    "ild_crosslag",
    list(
      outcome = outcome,
      predictor = predictor,
      lag = lag,
      mode = mode,
      max_gap = max_gap,
      ar1 = ar1
    ),
    list(lag_var = lag_var_name, n_obs = nrow(data_lagged))
  )
  if (include_diagnostics) {
    diag_out <- tryCatch(
      ild_diagnostics(fit, type = c("residual_acf", "qq")),
      error = function(e) {
        warning("Could not compute diagnostics: ", conditionMessage(e), call. = FALSE)
        NULL
      }
    )
    out["diagnostics"] <- list(diag_out)
  }
  out
}

# Canonical diagnostics bundle (Phase 1 contract -- structure only; engines populate later)

#' Top-level slot names for an ILD diagnostics bundle (internal contract)
#' @keywords internal
#' @noRd
ILD_DIAGNOSTICS_BUNDLE_SLOTS <- c(
  "meta",
  "data",
  "design",
  "fit",
  "residual",
  "predictive",
  "missingness",
  "causal",
  "warnings",
  "guardrails",
  "summary_text"
)

#' Canonical ILD diagnostics bundle
#'
#' The bundle has identical top-level \strong{slot names} for every estimation backend.
#' Content differs by engine; unavailable sections are \code{NULL}.
#' \code{warnings} and \code{guardrails} are always tibbles (possibly 0 rows).
#' Guardrails use the canonical schema documented in \code{\link{guardrail_registry}}:
#' \code{rule_id}, \code{section}, \code{severity}, \code{triggered}, \code{message},
#' \code{recommendation} (only triggered rules are stored as rows).
#'
#' Several standalone functions **feed** these slots when you use \code{\link{ild_diagnose}};
#' see \code{\link{ild_diagnostics_utilities}} (\code{\link{ild_design_check}},
#' \code{\link{ild_missing_pattern}}, \code{\link{ild_missing_model}}, \code{\link{ild_ipw_weights}}).
#'
#' Bundles returned by \code{\link{ild_diagnose}} also set \code{attr(bundle, "ild_fit")} and
#' \code{attr(bundle, "ild_data")} for \code{\link{ild_autoplot}}; re-run \code{ild_diagnose} if
#' these are missing (e.g. after loading a saved bundle without the fitted object).
#'
#' The \code{\link{print}} method lists each slot and prints a short \strong{summary} line:
#' number of warning rows, number of guardrail rows, highest guardrail severity
#' (\code{info} < \code{warning}), and up to five triggered \code{rule_id} values.
#'
#' @section Slots:
#' \describe{
#'   \item{\code{meta}}{List or \code{NULL}: run ids, engine, dimensions.}
#'   \item{\code{data}}{List or \code{NULL}: spacing, gaps, compliance, missingness, distributions.}
#'   \item{\code{design}}{List or \code{NULL}: within/between variation, time coverage, occasions, imbalance.}
#'   \item{\code{fit}}{List or \code{NULL}: convergence, singularity, optimizer / MCMC diagnostics.}
#'   \item{\code{residual}}{List or \code{NULL}: ACF, Q-Q, fitted vs observed, heteroskedasticity.}
#'   \item{\code{predictive}}{List or \code{NULL}: CV, PPC, forecast error.}
#'   \item{\code{missingness}}{List or \code{NULL}: IPW / imputation summaries when applicable.}
#'   \item{\code{causal}}{List or \code{NULL}: weights, positivity, causal diagnostics when applicable.}
#'   \item{\code{warnings}}{Tibble: structured warnings.}
#'   \item{\code{guardrails}}{Tibble: methodological guardrails (see \code{\link{guardrail_registry}}).}
#'   \item{\code{summary_text}}{Character: short narrative summary.}
#' }
#'
#' @param meta,data,design,fit,residual,predictive,missingness,causal Optional sections
#'   (list or structured object, or \code{NULL}).
#' @param warnings,guardrails Tibbles (default empty).
#' @param summary_text Character vector (default empty).
#' @return An object of class \code{ild_diagnostics_bundle}.
#' @seealso \code{\link{ild_diagnose}}, \code{\link{guardrail_registry}},
#'   \code{\link{ild_diagnostics_utilities}},
#'   \code{\link{ild_tidy_schema}}, \code{\link{ild_augment_schema}}.
#'   Per-section field semantics: \code{vignette("developer-contracts", package = "tidyILD")}.
#' @export
ild_diagnostics_bundle <- function(meta = NULL,
                                 data = NULL,
                                 design = NULL,
                                 fit = NULL,
                                 residual = NULL,
                                 predictive = NULL,
                                 missingness = NULL,
                                 causal = NULL,
                                 warnings = NULL,
                                 guardrails = NULL,
                                 summary_text = character()) {
  if (is.null(warnings)) warnings <- tibble::tibble()
  if (is.null(guardrails)) guardrails <- guardrails_empty_tibble()
  if (!tibble::is_tibble(warnings)) {
    warnings <- tibble::as_tibble(warnings)
  }
  if (!tibble::is_tibble(guardrails)) {
    guardrails <- tibble::as_tibble(guardrails)
  }
  if (!is.character(summary_text)) summary_text <- as.character(summary_text)
  out <- list(
    meta = meta,
    data = data,
    design = design,
    fit = fit,
    residual = residual,
    predictive = predictive,
    missingness = missingness,
    causal = causal,
    warnings = warnings,
    guardrails = guardrails,
    summary_text = summary_text
  )
  validate_ild_diagnostics_bundle(out)
  class(out) <- c("ild_diagnostics_bundle", "list")
  out
}

#' @rdname ild_diagnostics_bundle
#' @export
new_ild_diagnostics_bundle <- ild_diagnostics_bundle

#' @rdname ild_diagnostics_bundle
#' @return \code{is_ild_diagnostics_bundle}: logical.
#' @param x Object to test.
#' @export
is_ild_diagnostics_bundle <- function(x) {
  inherits(x, "ild_diagnostics_bundle")
}

#' @keywords internal
#' @noRd
validate_ild_diagnostics_bundle <- function(x) {
  if (!is.list(x)) stop("Bundle must be a list.", call. = FALSE)
  nm <- names(x)
  if (!identical(nm, ILD_DIAGNOSTICS_BUNDLE_SLOTS)) {
    stop(
      "Diagnostics bundle must have exactly these names in order: ",
      paste(ILD_DIAGNOSTICS_BUNDLE_SLOTS, collapse = ", "),
      call. = FALSE
    )
  }
  if (!tibble::is_tibble(x$warnings)) stop("Bundle $warnings must be a tibble.", call. = FALSE)
  if (!tibble::is_tibble(x$guardrails)) stop("Bundle $guardrails must be a tibble.", call. = FALSE)
  gr_req <- c("rule_id", "section", "severity", "triggered", "message", "recommendation")
  if (!all(gr_req %in% names(x$guardrails))) {
    stop(
      "Bundle $guardrails must have columns: ",
      paste(gr_req, collapse = ", "),
      call. = FALSE
    )
  }
  if (!is.character(x$summary_text)) stop("Bundle $summary_text must be character.", call. = FALSE)
  invisible(x)
}

#' @export
print.ild_diagnostics_bundle <- function(x, ...) {
  cat("ILD diagnostics bundle\n")
  for (s in ILD_DIAGNOSTICS_BUNDLE_SLOTS) {
    if (s %in% c("warnings", "guardrails")) {
      n <- nrow(x[[s]])
      cat("  ", s, ": tibble [", n, " x ", ncol(x[[s]]), "]\n", sep = "")
      next
    }
    if (s == "summary_text") {
      cat("  summary_text: ", length(x[[s]]), " string(s)\n", sep = "")
      next
    }
    v <- x[[s]]
    if (is.null(v)) {
      cat("  ", s, ": NULL\n", sep = "")
    } else {
      cat("  ", s, ": <non-null>\n", sep = "")
    }
  }
  nw <- nrow(x$warnings)
  ng <- nrow(x$guardrails)
  cat("Summary:\n")
  cat("  warnings: ", nw, " row(s)\n", sep = "")
  cat("  guardrails: ", ng, " row(s)", sep = "")
  if (ng > 0L) {
    max_sev <- guardrail_max_severity(x$guardrails$severity)
    rids <- unique(as.character(x$guardrails$rule_id))
    show <- rids[seq_len(min(length(rids), 5L))]
    cat(
      "; highest severity: ",
      if (!is.na(max_sev) && nzchar(max_sev)) max_sev else "-",
      "\n  key rule_id(s): ",
      if (length(show) > 0L) paste(show, collapse = ", ") else "-",
      "\n",
      sep = ""
    )
  } else {
    cat("\n")
  }
  invisible(x)
}

test_that("ild_diagnostics_bundle has fixed slots and accepts NULL sections", {
  b <- ild_diagnostics_bundle(
    meta = list(engine = "test"),
    fit = list(ok = TRUE)
  )
  expect_s3_class(b, "ild_diagnostics_bundle")
  expect_named(b, names(ild_diagnostics_bundle()))
  expect_null(b$data)
  expect_equal(b$meta$engine, "test")
  expect_true(tibble::is_tibble(b$warnings))
  expect_true(tibble::is_tibble(b$guardrails))
  expect_equal(
    names(b$guardrails),
    c("rule_id", "section", "severity", "triggered", "message", "recommendation")
  )
})

test_that("guardrail_registry lists rules and matches bundle schema", {
  gr <- guardrail_registry()
  expect_s3_class(gr, "tbl_df")
  expect_true(all(c("rule_id", "section", "severity", "default_message", "default_recommendation") %in% names(gr)))
  expect_true(nrow(gr) >= 8L)
})

test_that("validate_ild_diagnostics_bundle rejects malformed list", {
  bad <- list(meta = NULL)
  expect_error(
    tidyILD:::validate_ild_diagnostics_bundle(bad),
    "exactly these names"
  )
  bad2 <- ild_diagnostics_bundle()
  bad2$guardrails <- tibble::tibble(rule = "x")
  expect_error(
    tidyILD:::validate_ild_diagnostics_bundle(bad2),
    "guardrails must have columns"
  )
})

test_that("ild_tidy_schema and ild_augment_schema return required and optional names", {
  ts <- ild_tidy_schema()
  expect_true(all(c("term", "conf_low", "conf_high", "engine") %in% ts$required))
  expect_true(all(c("rhat", "pd") %in% ts$optional))
  as <- ild_augment_schema()
  expect_true(all(c(".ild_id", ".outcome", ".resid_std", "engine") %in% as$required))
  expect_true(all(c(".fitted_lower", ".state") %in% as$optional))
})

# Internal fillers for ild_diagnostics_bundle assembly (engine-agnostic contract).
# Standalone section providers are documented under ?ild_diagnostics_utilities
# (ild_design_check, ild_missing_pattern, ild_missing_model, ild_ipw_weights).

#' Flatten nested \code{fit} slice for guardrails and legacy callers
#' @param engine Optional engine label (\code{"lmer"}, \code{"lme"}, \code{"brms"}, \code{"ctsem"}) when
#'   \code{fit$engine} is missing (e.g. unit-test fixtures).
#' @keywords internal
#' @noRd
fit_bundle_flat_for_guardrails <- function(fit, engine = NULL) {
  if (is.null(fit) || !is.list(fit)) {
    return(list(
      singular = NA,
      max_rhat = NA_real_,
      convergence_table = NULL,
      ild_posterior = NULL
    ))
  }
  eng <- engine %||% fit$engine
  if (identical(eng, "brms")) {
    conv <- fit$convergence
    if (inherits(conv, "data.frame")) {
      conv <- list(convergence_table = conv, max_rhat = fit$max_rhat)
    }
    if (is.null(conv) || length(conv) == 0L) {
      conv <- list(
        convergence_table = fit$convergence_table,
        max_rhat = fit$max_rhat
      )
    }
    ps <- fit$posterior_summary$ild_posterior %||% fit$ild_posterior
    list(
      singular = NA,
      max_rhat = conv$max_rhat %||% fit$max_rhat,
      convergence_table = conv$convergence_table %||% fit$convergence_table,
      ild_posterior = ps
    )
  } else if (identical(eng, "ctsem")) {
    conv <- fit$convergence
    list(
      singular = NA,
      max_rhat = NA_real_,
      convergence_table = NULL,
      ild_posterior = NULL,
      converged = tryCatch(conv$converged, error = function(e) NA),
      drift_abs_max = fit$drift_abs_max %||% NA_real_
    )
  } else {
    conv <- fit$convergence
    if (inherits(conv, "data.frame")) {
      conv <- list(singular = fit$singular, converged = fit$converged, optimizer_messages = fit$optimizer_messages)
    }
    if (is.null(conv) || length(conv) == 0L) {
      conv <- list(singular = fit$singular, converged = fit$converged, optimizer_messages = fit$optimizer_messages)
    }
    list(
      singular = conv$singular %||% fit$singular,
      max_rhat = NA_real_,
      convergence_table = NULL,
      ild_posterior = NULL
    )
  }
}

#' Enrich \code{data} / \code{design} with parallel semantic blocks (cohort, per_id, timing, imbalance).
#' @keywords internal
#' @noRd
enrich_bundle_semantic_sections <- function(bundle) {
  if (!is.null(bundle$data)) {
    d <- bundle$data
    if (is.null(d$per_id) && !is.null(d$obs_per_id)) {
      bundle$data$per_id <- d$obs_per_id
    }
    if (is.null(d$imbalance) && !is.null(d$obs_per_id)) {
      bundle$data$imbalance <- d$obs_per_id
    }
    if (is.null(d$timing)) {
      bundle$data$timing <- list(
        spacing_class = d$spacing_class,
        n_gaps = d$n_gaps,
        pct_gap = d$pct_gap,
        median_dt_sec = d$median_dt_sec,
        iqr_dt_sec = d$iqr_dt_sec,
        time_range = d$time_range
      )
    }
    if (is.null(d$distribution)) {
      bundle$data$distribution <- list(
        outcome_summaries = d$outcome_summaries,
        missingness_rates = d$missingness_rates
      )
    }
  }
  if (!is.null(bundle$design)) {
    g <- bundle$design
    if (!is.null(bundle$data$cohort)) {
      bundle$design$cohort <- bundle$data$cohort
    }
    if (!is.null(bundle$data$obs_per_id)) {
      bundle$design$per_id <- bundle$data$obs_per_id
    }
    if (is.null(g$timing) && !is.null(g$time_coverage)) {
      bundle$design$timing <- list(
        time_coverage = g$time_coverage,
        spacing_class = g$spacing_class
      )
    }
    if (is.null(g$imbalance) && !is.null(g$occasion_imbalance)) {
      bundle$design$imbalance <- g$occasion_imbalance
    }
  }
  bundle
}

#' @keywords internal
#' @noRd
`%||%` <- function(x, y) {
  if (is.null(x)) y else x
}

#' Data-layer diagnostics (ILD object)
#' @param outcome_vars Optional character vector of column names for \code{outcome_summaries} and \code{missingness_rates}.
#' @keywords internal
#' @noRd
fill_diagnostics_data <- function(data, include_missing_pattern = TRUE, outcome_vars = NULL) {
  validate_ild(data)
  sm <- ild_summary(data)
  sc <- ild_spacing_class(data)
  mp <- NULL
  if (isTRUE(include_missing_pattern)) {
    mp <- tryCatch(ild_missing_pattern(data, vars = NULL), error = function(e) NULL)
  }
  n_int <- if (!is.null(sm$spacing$n_intervals)) sm$spacing$n_intervals else {
    sum(!is.na(data[[".ild_dt"]]))
  }
  cohort <- list(
    n_id = as.integer(sm$summary$n_id[1]),
    n_obs = as.integer(sm$summary$n_obs[1]),
    n_intervals = as.integer(n_int)
  )
  tab <- as.integer(table(data[[".ild_id"]]))
  obs_per_id <- list(
    min = as.integer(min(tab)),
    median = stats::median(tab),
    max = as.integer(max(tab)),
    sd = if (length(tab) > 1L) stats::sd(as.numeric(tab)) else NA_real_
  )
  outcome_summaries <- NULL
  missingness_rates <- NULL
  if (!is.null(outcome_vars) && length(outcome_vars) > 0L) {
    outcome_vars <- unique(outcome_vars[outcome_vars %in% names(data)])
    if (length(outcome_vars) > 0L) {
      rows <- lapply(outcome_vars, function(v) {
        x <- data[[v]]
        ntot <- length(x)
        nna <- sum(is.na(x))
        tibble::tibble(
          variable = v,
          mean = if (is.numeric(x)) mean(x, na.rm = TRUE) else NA_real_,
          sd = if (is.numeric(x)) stats::sd(x, na.rm = TRUE) else NA_real_,
          min = if (is.numeric(x)) suppressWarnings(min(x, na.rm = TRUE)) else NA_real_,
          max = if (is.numeric(x)) suppressWarnings(max(x, na.rm = TRUE)) else NA_real_,
          pct_na = if (ntot > 0L) 100 * nna / ntot else NA_real_,
          n = ntot
        )
      })
      outcome_summaries <- do.call(rbind, rows)
      missingness_rates <- tibble::tibble(
        variable = outcome_vars,
        pct_na = vapply(outcome_vars, function(v) {
          x <- data[[v]]
          if (length(x) == 0L) return(NA_real_)
          100 * mean(is.na(x))
        }, double(1L))
      )
    }
  }
  list(
    summary = sm$summary,
    spacing_class = sc,
    n_gaps = sm$n_gaps,
    pct_gap = sm$pct_gap,
    median_dt_sec = sm$summary$median_dt_sec[1],
    iqr_dt_sec = sm$summary$iqr_dt_sec[1],
    time_range = sm$time_range,
    cohort = cohort,
    obs_per_id = obs_per_id,
    outcome_summaries = outcome_summaries,
    missingness_rates = missingness_rates,
    missing_pattern = if (!is.null(mp)) {
      list(summary = mp$summary, overall = mp$overall, n_complete = mp$n_complete)
    } else {
      NULL
    }
  )
}

#' Design-layer diagnostics
#' @keywords internal
#' @noRd
fill_diagnostics_design <- function(data, vars = NULL) {
  validate_ild(data)
  dc <- ild_design_check(data, vars = vars)
  sm <- ild_summary(data)
  tr <- sm$time_range
  span_sec <- if (length(tr) >= 2L && all(is.finite(tr))) (tr[2L] - tr[1L]) else NA_real_
  time_coverage <- list(
    min = if (length(tr) >= 1L) tr[1L] else NA_real_,
    max = if (length(tr) >= 2L) tr[2L] else NA_real_,
    span_sec = span_sec
  )
  has_wp_bp <- !is.null(dc$wp_bp) && (inherits(dc$wp_bp, "data.frame") || inherits(dc$wp_bp, "tbl_df")) &&
    nrow(dc$wp_bp) > 0L
  flags <- list(
    has_wp_bp = has_wp_bp,
    spacing_class = dc$spacing_class,
    irregular = identical(dc$spacing_class, "irregular-ish")
  )
  list(
    ild_design_check = dc,
    spacing_class = dc$spacing_class,
    recommendation = dc$recommendation,
    wp_bp = dc$wp_bp,
    design_missingness = dc$missingness,
    flags = flags,
    time_coverage = time_coverage
  )
}

#' Fit diagnostics -- lmerMod
#' @keywords internal
#' @noRd
fill_diagnostics_fit_lmerMod <- function(fit) {
  rescor_note <- "lme4::lmer has no Gaussian AR1/CAR1 residual correlation; use nlme::lme via ild_lme(ar1 = TRUE)."
  out <- list(
    engine = "lmer",
    convergence = list(
      singular = NA,
      converged = NA,
      optimizer_messages = character()
    ),
    rank = list(fixed_effects_X = NA_integer_),
    optimizer = list(
      reml = NA,
      name = NA_character_,
      optinfo = NULL
    ),
    residual_correlation = list(
      modeled = FALSE,
      structure = "none",
      note = rescor_note
    ),
    random_effects_structure = list(theta = NULL)
  )
  if (!inherits(fit, "lmerMod")) return(out)
  out$convergence$singular <- lme4::isSingular(fit)
  out$optimizer$reml <- tryCatch(lme4::getME(fit, "is_REML"), error = function(e) NA)
  out$optimizer$name <- tryCatch(as.character(fit@optinfo$optimizer)[1L], error = function(e) NA_character_)
  th <- tryCatch(lme4::getME(fit, "theta"), error = function(e) NULL)
  if (!is.null(th) && length(th) > 0L) {
    out$random_effects_structure$theta <- list(
      length = length(th),
      min = min(th, na.rm = TRUE),
      max = max(th, na.rm = TRUE)
    )
  }
  X <- tryCatch(lme4::getME(fit, "X"), error = function(e) NULL)
  if (!is.null(X) && nrow(X) > 0L) {
    out$rank$fixed_effects_X <- tryCatch(qr(X)$rank, error = function(e) NA_integer_)
  }
  ar1 <- isTRUE(attr(fit, "ild_ar1", exact = TRUE))
  out$residual_correlation <- list(
    modeled = ar1,
    structure = if (ar1) "ild_ar1_attr" else "none",
    note = rescor_note
  )
  opt <- fit@optinfo
  out$optimizer$optinfo <- list(
    conv = opt$conv,
    finopt = opt$finopt
  )
  cv <- opt$conv$lme4
  if (!is.null(cv)) {
    if (is.list(cv)) {
      nums <- suppressWarnings(as.numeric(unlist(cv, use.names = FALSE)))
      out$convergence$converged <- length(nums) > 0L && all(nums == 0, na.rm = TRUE)
    } else {
      out$convergence$converged <- isTRUE(as.integer(cv)[1L] == 0L)
    }
  }
  if (length(opt$warnings) > 0L) {
    out$convergence$optimizer_messages <- vapply(
      opt$warnings,
      function(w) tryCatch(conditionMessage(w), error = function(e) "warning"),
      character(1L)
    )
  }
  out$heterogeneity <- fill_diagnostics_heterogeneity(fit)
  out
}

#' Fit diagnostics -- nlme lme
#' @keywords internal
#' @noRd
fill_diagnostics_fit_lme <- function(fit) {
  out <- list(
    engine = "lme",
    convergence = list(
      singular = NA,
      converged = NA,
      apVar_ok = NA,
      optimizer_messages = character()
    ),
    rank = list(fixed_effects_X = NA_integer_),
    optimizer = list(note = "nlme::lme -- see model object for optimization details."),
    residual_correlation = list(modeled = FALSE, class = NA_character_, coef_corStruct = NULL),
    random_effects_structure = list(note = "See VarCorr(model) and random effects in summary().")
  )
  if (!inherits(fit, "lme")) return(out)
  out$convergence$converged <- TRUE
  out$convergence$apVar_ok <- !is.null(fit$apVar) && !any(is.na(fit$apVar))
  if (!is.null(fit$apVar) && any(is.na(fit$apVar))) {
    out$convergence$singular <- TRUE
  } else {
    out$convergence$singular <- FALSE
  }
  corr_class <- attr(fit, "ild_correlation_class", exact = TRUE)
  coef_cs <- tryCatch(
    if (!is.null(fit$modelStruct$corStruct)) coef(fit$modelStruct$corStruct) else NULL,
    error = function(e) NULL
  )
  has_cor <- !is.null(fit$modelStruct$corStruct)
  out$residual_correlation <- list(
    modeled = isTRUE(attr(fit, "ild_ar1", exact = TRUE)) || has_cor,
    class = if (!is.null(corr_class)) as.character(corr_class)[1L] else NA_character_,
    coef_corStruct = coef_cs
  )
  out$heterogeneity <- fill_diagnostics_heterogeneity(fit)
  out
}

#' Fit diagnostics -- brms (uses attr ild_posterior)
#' @keywords internal
#' @noRd
fill_diagnostics_fit_brms <- function(fit) {
  ps <- attr(fit, "ild_posterior", exact = TRUE)
  if (is.null(ps)) ps <- ild_posterior_attr(fit, prior_template = NULL)
  conv_tbl <- NULL
  sf <- tryCatch(summary(fit)$fixed, error = function(e) NULL)
  if (!is.null(sf) && nrow(sf) > 0L) {
    conv_tbl <- tibble::tibble(
      term = rownames(sf),
      rhat = as.numeric(sf[["Rhat"]]),
      ess_bulk = as.numeric(sf[["Bulk_ESS"]]),
      ess_tail = as.numeric(sf[["Tail_ESS"]])
    )
  }
  mx <- ild_brms_max_rhat(fit)
  out <- list(
    engine = "brms",
    convergence = list(
      convergence_table = conv_tbl,
      max_rhat = mx
    ),
    sampler = list(
      n_divergent = ps$n_divergent,
      n_max_treedepth_hits = ps$n_max_treedepth_hits,
      chains = ps$chains,
      iter = ps$iter,
      warmup = ps$warmup
    ),
    residual_correlation = list(
      modeled = NA,
      note = "See brms family/residual structure and posterior predictive checks for residual behavior."
    ),
    posterior_summary = list(ild_posterior = ps)
  )
  out$heterogeneity <- fill_diagnostics_heterogeneity(fit)
  out
}

#' Residual diagnostics -- frequentist (embeds legacy [ild_diagnostics()] for plotting)
#' @keywords internal
#' @noRd
fill_diagnostics_residual_frequentist <- function(object, data, type, by_id) {
  diag <- ild_diagnostics(object, data = data, type = type, by_id = by_id)
  aug <- tryCatch(augment_ild_model(object), error = function(e) NULL)
  res_sd <- NA_real_
  cor_obs_fitted <- NA_real_
  if (!is.null(aug) && nrow(aug) > 0L) {
    res_sd <- stats::sd(aug$.resid, na.rm = TRUE)
    if (all(is.finite(aug$.fitted)) && stats::sd(aug$.fitted, na.rm = TRUE) > 0) {
      cor_obs_fitted <- stats::cor(aug$.outcome, aug$.fitted, use = "pairwise.complete.obs")
    }
  }
  list(
    engine = if (inherits(object, "lme")) "lme" else "lmer",
    legacy_ild_diagnostics = diag,
    residual_sd = res_sd,
    cor_observed_fitted = cor_obs_fitted,
    stats = diag$stats,
    meta = diag$meta
  )
}

#' Residual layer--  brms (lighter; legacy ACF not computed here)
#' @keywords internal
#' @noRd
fill_diagnostics_residual_brms <- function(fit) {
  aug <- tryCatch(ild_augment(fit), error = function(e) NULL)
  out <- list(engine = "brms", legacy_ild_diagnostics = NULL, residual_sd = NA_real_)
  if (!is.null(aug) && ".resid" %in% names(aug)) {
    out$residual_sd <- stats::sd(aug$.resid, na.rm = TRUE)
    out$mean_abs_resid <- mean(abs(aug$.resid), na.rm = TRUE)
  }
  out
}

#' Predictive layer -- frequentist (observed vs fitted summaries)
#' @keywords internal
#' @noRd
fill_diagnostics_predictive_frequentist <- function(object, data) {
  aug <- tryCatch(augment_ild_model(object), error = function(e) NULL)
  if (is.null(aug)) return(NULL)
  om <- fill_diagnostics_predictive_obs_metrics(aug)
  eng <- if (inherits(object, "lme")) "lme" else "lmer"
  c(
    list(
      engine = eng,
      obs_metrics = om,
      simulation_checks = list(ppc = NULL)
    ),
    om
  )
}

#' @keywords internal
#' @noRd
fill_diagnostics_predictive_obs_metrics <- function(aug) {
  if (is.null(aug) || !all(c(".outcome", ".fitted") %in% names(aug))) {
    return(list())
  }
  y <- aug$.outcome
  f <- aug$.fitted
  ok <- is.finite(y) & is.finite(f)
  n <- sum(ok)
  ye <- y[ok]
  fe <- f[ok]
  res <- ye - fe
  list(
    n = as.integer(n),
    mean_abs_error = if (n > 0L) mean(abs(res), na.rm = TRUE) else NA_real_,
    rmse = if (n > 0L) sqrt(mean(res^2, na.rm = TRUE)) else NA_real_,
    mean_residual = if (n > 0L) mean(res, na.rm = TRUE) else NA_real_,
    max_abs_error = if (n > 0L) max(abs(res), na.rm = TRUE) else NA_real_,
    cor_observed_fitted = if (n > 1L) stats::cor(ye, fe, use = "pairwise.complete.obs") else NA_real_
  )
}

#' Predictive layer -- brms augment-only (no PPC)
#' @keywords internal
#' @noRd
fill_diagnostics_predictive_brms_augment_only <- function(fit) {
  aug <- tryCatch(ild_augment(fit), error = function(e) NULL)
  om <- fill_diagnostics_predictive_obs_metrics(aug)
  c(
    list(
      engine = "brms",
      obs_metrics = om,
      simulation_checks = list(ppc = NULL)
    ),
    om
  )
}

#' Predictive layer -- brms (PPC summary)
#' @keywords internal
#' @noRd
fill_diagnostics_predictive_brms <- function(fit, ppc_ndraws) {
  ppc <- ild_brms_ppc_summary(fit, ndraws = ppc_ndraws)
  aug <- tryCatch(ild_augment(fit), error = function(e) NULL)
  om <- fill_diagnostics_predictive_obs_metrics(aug)
  c(
    list(
      engine = "brms",
      obs_metrics = om,
      simulation_checks = list(ppc = ppc),
      ppc = ppc
    ),
    om
  )
}

#' Fit layer -- ctsem
#' @keywords internal
#' @noRd
fill_diagnostics_fit_ctsem <- function(object) {
  fit <- object$ct_fit
  conv <- tryCatch(fit$optimization$convergence, error = function(e) NA_integer_)
  # ctsem/OpenMx may expose empty-length convergence slots on some platforms; guard
  # so `if (!is.finite(conv))` never sees logical(0).
  if (length(conv) != 1L || !is.finite(conv)) {
    conv <- tryCatch(fit$opt$convergence, error = function(e) NA_integer_)
  }
  if (length(conv) != 1L || !is.finite(conv)) {
    conv <- NA_integer_
  }
  ll <- tryCatch(as.numeric(stats::logLik(fit)[1L]), error = function(e) NA_real_)
  drift_abs_max <- tryCatch({
    dr <- suppressWarnings(as.numeric(fit$DRIFT))
    dr <- dr[is.finite(dr)]
    if (length(dr) == 0L) NA_real_ else max(abs(dr))
  }, error = function(e) NA_real_)
  list(
    engine = "ctsem",
    convergence = list(
      converged = if (is.finite(conv)) isTRUE(conv == 0L) else .ild_ctsem_is_converged(object),
      convergence_code = conv,
      note = "Convergence extracted from ctsem optimization slots when available."
    ),
    optimizer = list(
      note = "See ct_fit object for optimizer details."
    ),
    log_likelihood = ll,
    state_dimension = tryCatch(as.integer(ncol(fit$DRIFT)), error = function(e) NA_integer_),
    drift_abs_max = drift_abs_max,
    residual_correlation = list(
      modeled = TRUE,
      structure = "continuous_time_state_dynamics",
      note = "Temporal dependence is represented through continuous-time latent dynamics."
    )
  )
}

#' Residual layer -- ctsem
#' @keywords internal
#' @noRd
fill_diagnostics_residual_ctsem <- function(object, data) {
  aug <- tryCatch(ild_augment(object), error = function(e) NULL)
  if (is.null(aug)) {
    return(list(engine = "ctsem", legacy_ild_diagnostics = NULL))
  }
  acf_obj <- tryCatch(stats::acf(aug$.resid, plot = FALSE, na.action = stats::na.pass), error = function(e) NULL)
  acf_tbl <- if (!is.null(acf_obj)) {
    tibble::tibble(lag = as.numeric(acf_obj$lag), acf = as.numeric(acf_obj$acf))
  } else {
    tibble::tibble(lag = numeric(), acf = numeric())
  }
  qq_cor <- tryCatch(
    stats::cor(
      stats::qnorm(stats::ppoints(sum(is.finite(aug$.resid_std)))),
      sort(aug$.resid_std[is.finite(aug$.resid_std)])
    ),
    error = function(e) NA_real_
  )
  list(
    engine = "ctsem",
    residual_sd = stats::sd(aug$.resid, na.rm = TRUE),
    acf = acf_tbl,
    qq = list(correlation = qq_cor),
    legacy_ild_diagnostics = NULL
  )
}

#' Predictive layer -- ctsem
#' @keywords internal
#' @noRd
fill_diagnostics_predictive_ctsem <- function(object) {
  aug <- tryCatch(ild_augment(object), error = function(e) NULL)
  if (is.null(aug)) {
    return(list(engine = "ctsem", obs_metrics = NULL, simulation_checks = list(ppc = NULL)))
  }
  om <- fill_diagnostics_predictive_obs_metrics(aug)
  c(
    list(
      engine = "ctsem",
      obs_metrics = om,
      simulation_checks = list(ppc = NULL)
    ),
    om
  )
}

#' Missingness-focused slice (formula / model variables when available)
#' @keywords internal
#' @noRd
fill_diagnostics_missingness_section <- function(data, vars = NULL) {
  validate_ild(data)
  if (is.null(vars) || length(vars) == 0L) {
    return(list(
      note = "No model variables supplied; see data$missing_pattern for global missingness.",
      summary = NULL,
      overall = NULL,
      pct_na_by_var = NULL
    ))
  }
  mp <- tryCatch(ild_missing_pattern(data, vars = vars), error = function(e) NULL)
  if (is.null(mp)) {
    return(list(
      note = NULL,
      summary = NULL,
      overall = NULL,
      pct_na_by_var = NULL
    ))
  }
  pct_na_by_var <- NULL
  if (!is.null(mp$summary) && nrow(mp$summary) > 0L && "pct_na" %in% names(mp$summary)) {
    pct_na_by_var <- mp$summary[, intersect(c("variable", "pct_na"), names(mp$summary)), drop = FALSE]
  }
  list(
    note = NULL,
    summary = mp$summary,
    overall = mp$overall,
    pct_na_by_var = pct_na_by_var
  )
}

#' Causal / IPW slice when weight columns present
#' @keywords internal
#' @noRd
fill_diagnostics_causal <- function(data,
                                    causal_detail = FALSE,
                                    balance = FALSE,
                                    balance_treatment = NULL,
                                    balance_covariates = NULL,
                                    balance_weights_col = ".ipw_treat",
                                    balance_by_occasion = FALSE) {
  if (is.null(data) || !is_ild(data)) return(NULL)
  wcols <- grep("^\\.ipw|ipw_weight|\\.w_", names(data), value = TRUE, ignore.case = TRUE)
  if (length(wcols) == 0L && !isTRUE(balance)) return(NULL)
  out <- list(columns_found = wcols)
  .summ_w <- function(w) {
    list(
      min = min(w, na.rm = TRUE),
      max = max(w, na.rm = TRUE),
      mean = mean(w, na.rm = TRUE)
    )
  }
  if (".ipw" %in% names(data)) {
    w <- data[[".ipw"]]
    out$weight_summary <- .summ_w(w)
    if (isTRUE(causal_detail)) {
      wf <- w[is.finite(w)]
      if (length(wf) > 0L) {
        qs <- stats::quantile(wf, probs = c(0.05, 0.25, 0.5, 0.75, 0.95), na.rm = TRUE, names = TRUE)
        out$weight_detail <- list(
          quantiles = qs,
          sum_w = sum(wf, na.rm = TRUE)
        )
      }
    }
  }
  if (".ipw_treat" %in% names(data)) {
    out$weight_summary_treat <- .summ_w(data[[".ipw_treat"]])
    if (isTRUE(causal_detail)) {
      wf <- data[[".ipw_treat"]][is.finite(data[[".ipw_treat"]])]
      if (length(wf) > 0L) {
        out$weight_detail_treat <- list(
          quantiles = stats::quantile(wf, probs = c(0.05, 0.25, 0.5, 0.75, 0.95), na.rm = TRUE, names = TRUE)
        )
      }
    }
  }
  if (".ipw_censor" %in% names(data)) {
    out$weight_summary_censor <- .summ_w(data[[".ipw_censor"]])
    if (isTRUE(causal_detail)) {
      wf <- data[[".ipw_censor"]][is.finite(data[[".ipw_censor"]])]
      if (length(wf) > 0L) {
        out$weight_detail_censor <- list(
          quantiles = stats::quantile(wf, probs = c(0.05, 0.25, 0.5, 0.75, 0.95), na.rm = TRUE, names = TRUE)
        )
      }
    }
  }
  if (isTRUE(causal_detail) && ".ipw_treat" %in% names(data) && ".ild_seq" %in% names(data) &&
      !is.null(attr(data, "ild_iptw_msm_fits", exact = TRUE))) {
    wt <- data[[".ipw_treat"]]
    sq <- data[[".ild_seq"]]
    uq <- sort(unique(as.numeric(sq)))
    occ_summ <- vector("list", length(uq))
    names(occ_summ) <- as.character(uq)
    for (tt in uq) {
      wm <- wt[as.numeric(sq) == tt & is.finite(wt)]
      if (length(wm) > 0L) {
        occ_summ[[as.character(tt)]] <- list(
          min = min(wm, na.rm = TRUE),
          max = max(wm, na.rm = TRUE),
          mean = mean(wm, na.rm = TRUE)
        )
      }
    }
    out$msm_treat_weight_by_occasion <- occ_summ
  }
  if (isTRUE(balance) && length(balance_covariates) > 0L && !is.null(balance_treatment) &&
      balance_treatment %in% names(data) && balance_weights_col %in% names(data)) {
    out$balance <- list(
      table = ild_msm_balance(
        data,
        treatment = balance_treatment,
        covariates = balance_covariates,
        weights_col = balance_weights_col,
        by_occasion = balance_by_occasion
      ),
      ess = ild_ipw_ess(
        data,
        weights_col = balance_weights_col,
        by_occasion = balance_by_occasion
      )
    )
  }
  out
}

#' Build warnings tibble
#' @param rows List of named lists with source, level, message, code
#' @keywords internal
#' @noRd
collect_diagnostics_warnings <- function(rows = list()) {
  if (length(rows) == 0L) {
    return(tibble::tibble(
      source = character(),
      level = character(),
      message = character(),
      code = character()
    ))
  }
  tibble::tibble(
    source = vapply(rows, function(z) .fld(z, "source"), character(1)),
    level = vapply(rows, function(z) .fld(z, "level"), character(1)),
    message = vapply(rows, function(z) .fld(z, "message"), character(1)),
    code = vapply(rows, function(z) .fld(z, "code"), character(1))
  )
}

#' @keywords internal
#' @noRd
.fld <- function(z, nm) {
  v <- z[[nm]]
  if (is.null(v)) NA_character_ else as.character(v)[1]
}

#' Build guardrails tibble (delegates to registry; uses \code{rule_id} schema)
#' @keywords internal
#' @noRd
collect_diagnostics_guardrails <- function(rows = list()) {
  guardrail_finalize_rows(rows)
}

.fld2 <- function(z, nm) {
  v <- z[[nm]]
  if (is.null(v)) NA_character_ else paste(as.character(v), collapse = " ")
}

#' Assemble warnings for frequentist fits (guardrails via \code{\link{evaluate_guardrails_fit}})
#' @keywords internal
#' @noRd
collect_freq_warnings_guardrails <- function(fit, fit_diag) {
  warns <- list()
  if (inherits(fit, "lmerMod") && length(fit@optinfo$warnings) > 0L) {
    for (i in seq_along(fit@optinfo$warnings)) {
      warns[[length(warns) + 1L]] <- list(
        source = "lme4",
        level = "warning",
        message = conditionMessage(fit@optinfo$warnings[[i]]),
        code = "lmer_warning"
      )
    }
  }
  list(
    warnings = collect_diagnostics_warnings(warns),
    guardrails = guardrails_empty_tibble()
  )
}

#' @keywords internal
#' @noRd
collect_brms_warnings_guardrails <- function(fit, fit_diag) {
  warns <- list()
  ps <- fit_diag$posterior_summary$ild_posterior %||% fit_diag$ild_posterior
  if (!is.null(ps) && !is.na(ps$n_divergent) && ps$n_divergent > 0L) {
    warns[[length(warns) + 1L]] <- list(
      source = "stan",
      level = "warning",
      message = sprintf("%s divergent transition(s) after warmup.", ps$n_divergent),
      code = "divergent_transitions"
    )
  }
  list(
    warnings = collect_diagnostics_warnings(warns),
    guardrails = guardrails_empty_tibble()
  )
}

#' Collapse bundle into short narrative strings
#' @keywords internal
#' @noRd
build_diagnostics_bundle_summary <- function(bundle) {
  parts <- character()
  if (!is.null(bundle$meta$engine)) {
    parts <- c(parts, sprintf("Engine: %s.", bundle$meta$engine))
  }
  if (!is.null(bundle$fit)) {
    ff <- bundle$fit
    sing <- ff$convergence$singular %||% ff$singular
    if (!is.null(sing) && !is.na(sing) && isTRUE(sing)) {
      parts <- c(parts, "Singular / problematic fit variance structure reported.")
    }
    xrk <- ff$rank$fixed_effects_X %||% ff$X_rank
    if (!is.null(xrk) && !is.na(xrk)) {
      parts <- c(parts, sprintf("Fixed-effects design rank (X): %s.", xrk))
    }
    mr <- ff$convergence$max_rhat %||% ff$max_rhat
    if (!is.null(mr) && is.finite(mr)) {
      parts <- c(parts, sprintf("Max R-hat: %.3f.", mr))
    }
  }
  if (!is.null(bundle$data$obs_per_id)) {
    opi <- bundle$data$obs_per_id
    if (!is.null(opi$min) && !is.null(opi$max)) {
      parts <- c(parts, sprintf(
        "Observations per person: min = %s, median = %.1f, max = %s.",
        opi$min, opi$median, opi$max
      ))
    }
  }
  if (!is.null(bundle$data$outcome_summaries) && nrow(bundle$data$outcome_summaries) > 0L &&
        "pct_na" %in% names(bundle$data$outcome_summaries)) {
    pm <- suppressWarnings(max(bundle$data$outcome_summaries$pct_na, na.rm = TRUE))
    if (is.finite(pm) && pm > 0) {
      parts <- c(parts, sprintf("Outcome(s) missingness up to %.1f%% (see data$outcome_summaries).", pm))
    }
  }
  if (!is.null(bundle$residual$legacy_ild_diagnostics)) {
    d <- bundle$residual$legacy_ild_diagnostics
    parts <- c(parts, sprintf(
      "Residual diagnostics (%s): n_obs = %s.",
      d$meta$engine, d$meta$n_obs
    ))
  }
  ppc_raw <- bundle$predictive$simulation_checks$ppc %||% bundle$predictive$ppc
  if (!is.null(bundle$predictive) && is.list(ppc_raw) && is.null(ppc_raw$error)) {
    p <- ppc_raw
    parts <- c(parts, sprintf(
      "PPC: |mean_obs - mean_rep| = %.4f; |sd_obs - sd_rep| = %.4f.",
      p$mean_abs_diff, p$sd_abs_diff
    ))
  }
  if (nrow(bundle$warnings) > 0L) {
    parts <- c(parts, sprintf("%d diagnostic warning(s) recorded.", nrow(bundle$warnings)))
  }
  if (nrow(bundle$guardrails) > 0L) {
    parts <- c(parts, sprintf("%d guardrail(s) triggered (see guardrail_registry).", nrow(bundle$guardrails)))
  }
  parts
}

# =============================================================================
# tidyILD — backend adapter template (NOT part of the package build)
# =============================================================================
#
# Copy sections into new R files under R/ when adding an estimation engine.
# Normative checklist: inst/dev/DEVELOPER_CONTRACTS.md §4 ("Backend adapter
# checklist"). This file is a **skeleton** only; replace MYENGINE, MYCLASS,
# ild_myfit(), etc. with real names.
#
# Obligations (avoid fragmentation across backends):
#   1. Tidier       — ild_tidy() method, ild_tidy_schema() columns
#   2. Augmenter    — ild_augment() method, ild_augment_schema() columns
#   3. Diagnose     — ild_diagnose() method → ild_diagnostics_bundle
#   4. Bundle shape — validate_ild_diagnostics_bundle; semantic fit/predictive
#   5. Autoplot     — bundle must carry attr(ild_fit), attr(ild_data)
#   6. Provenance   — attr(fit, "ild_provenance") for ild_methods()
#   7. Guardrails   — evaluate_guardrails_fit(fit_diag, engine = "...")
#   8. Tests        — schema + bundle + optional rule_id regression
#   9. Docs         — roxygen, DEVELOPER_CONTRACTS if slots change, NEWS
#
# Reference implementations in this package:
#   - Frequentist: R/ild_model_tidiers.R (tidy/augment), R/ild_diagnose_methods.R
#   - Bayesian:    R/ild_brms_s3.R, R/ild_brms.R (provenance), R/ild_diagnose_fillers.R
#
# =============================================================================

## ---------------------------------------------------------------------------
## 0. File layout (typical new engine)
## ---------------------------------------------------------------------------
#
# R/ild_myengine.R          — public fitter ild_myfit(); ild_data + ild_provenance
# R/ild_myengine_s3.R       — ild_tidy.MYCLASS, ild_augment.MYCLASS, ild_autoplot.brmsfit-style if needed
# R/ild_diagnose_methods.R  — add ild_diagnose.MYCLASS (or separate file + Collate)
# R/ild_diagnose_fillers.R  — add fill_diagnostics_fit_MYENGINE(), predictive, or source myengine_fillers.R
#
# tests/testthat/test-myengine.R — schema + bundle + smoke
# inst/dev/DEVELOPER_CONTRACTS.md — update §1 if fit/predictive semantics change
#

## ---------------------------------------------------------------------------
## 1. Tidier — conform to ild_tidy_schema()
## ---------------------------------------------------------------------------
#
# Required columns (see ?ild_tidy_schema): term, component, effect_level,
# estimate, std_error, conf_low, conf_high, statistic, p_value, interval_type,
# engine, model_class. Optional: rhat, ess_bulk, ess_tail, pd, rope_low, rope_high.
#
# Register in R/ild_generics.R:
#   @export ild_tidy.MYCLASS
#   ild_tidy.MYCLASS <- function(x, ...) { ... }
#
# Validate in tests:
#   expect_true(all(ild_tidy_schema()$required %in% names(out)))
#

if (FALSE) {
  #' @rdname ild_tidy
  #' @method ild_tidy MYCLASS
  #' @export
  ild_tidy.MYCLASS <- function(x, ...) {
    # Assemble a tibble; set engine = "myengine", model_class = class(x)[1]
    # Map terms to component / effect_level per DEVELOPER_CONTRACTS.md §2.3–2.4
    out <- tibble::tibble(
      term = character(),
      component = character(),
      effect_level = character(),
      estimate = double(),
      std_error = double(),
      conf_low = double(),
      conf_high = double(),
      statistic = double(),
      p_value = double(),
      interval_type = character(),
      engine = "myengine",
      model_class = class(x)[1]
    )
    out
  }
}

## ---------------------------------------------------------------------------
## 2. Augmenter — conform to ild_augment_schema()
## ---------------------------------------------------------------------------
#
# Required: .ild_id, .ild_time, .outcome, .fitted, .resid, .resid_std,
#           engine, model_class
# The fitted object must have attr(x, "ild_data") from your fitter (validate_ild).
#
# .resid_std: use residuals(x, type = "pearson") only when length-matched; else NA.
#

if (FALSE) {
  #' @rdname ild_augment
  #' @method ild_augment MYCLASS
  #' @export
  ild_augment.MYCLASS <- function(x, ...) {
    dat <- attr(x, "ild_data", exact = TRUE)
    if (is.null(dat)) {
      stop("MYCLASS fits require attr(fit, \"ild_data\") from ild_myfit().", call. = FALSE)
    }
    # Build tibble aligned to rows of dat; set engine / model_class columns
    out <- tibble::tibble(
      .ild_id = dat$.ild_id,
      .ild_time = dat$.ild_time,
      .outcome = NA_real_,
      .fitted = NA_real_,
      .resid = NA_real_,
      .resid_std = NA_real_,
      engine = "myengine",
      model_class = class(x)[1]
    )
    out
  }
}

## ---------------------------------------------------------------------------
## 3. ild_diagnose — ild_diagnostics_bundle only
## ---------------------------------------------------------------------------
#
# Pattern: mirror R/ild_diagnose_methods.R (lmerMod / lme / brmsfit).
# Use ild_diagnostics_bundle() constructor; slots in ILD_DIAGNOSTICS_BUNDLE_SLOTS.
# Reuse fill_diagnostics_data(), fill_diagnostics_design(), etc. where possible.
# Add fill_diagnostics_fit_myengine() in ild_diagnose_fillers.R for nested
# fit semantics (convergence, rank, optimizer, …) parallel to other engines.
# End with:
#   bundle$summary_text <- c(build_diagnostics_bundle_summary(bundle), ...)
#   bundle <- enrich_bundle_semantic_sections(bundle)
#   attr(bundle, "ild_fit") <- object
#   attr(bundle, "ild_data") <- data
#
# Guardrails:
#   bundle$guardrails <- guardrails_bind(
#     evaluate_guardrails_fit(object, fit_diag, engine = "myengine"),
#     evaluate_guardrails_contextual(object, data, bundle, engine = "myengine")
#   )
# Extend match.arg(engine) in evaluate_guardrails_fit() and
# evaluate_guardrails_contextual() when adding a new engine id (see
# R/ild_guardrail_registry.R). Until then, use the closest existing engine
# only for prototyping.
#
# fit_diag for brms-like rules: fit_bundle_flat_for_guardrails(bundle$fit, engine = "...")
# reads nested or flat fit slices (see R/ild_diagnose_fillers.R).
#

if (FALSE) {
  #' @rdname ild_diagnose
  #' @method ild_diagnose MYCLASS
  #' @export
  ild_diagnose.MYCLASS <- function(object, data = NULL, ...) {
    data <- if (is.null(data)) attr(object, "ild_data", exact = TRUE) else data
    validate_ild(data)
    preds <- NULL # e.g. all.vars(stats::terms(stats::formula(object)))
    outcome_vars <- ild_diagnose_outcome_vars(object, data, preds)
    fit_diag <- fill_diagnostics_fit_myengine(object)
    b <- ild_diagnostics_bundle(
      meta = list(engine = "myengine", n_obs = nrow(data), n_id = length(unique(data[[".ild_id"]]))),
      data = fill_diagnostics_data(data, outcome_vars = outcome_vars),
      design = fill_diagnostics_design(data, vars = if (length(preds)) preds else NULL),
      fit = fit_diag,
      residual = fill_diagnostics_residual_myengine(object, data),
      predictive = fill_diagnostics_predictive_myengine(object, data),
      missingness = fill_diagnostics_missingness_section(data, object, preds),
      causal = fill_diagnostics_causal(data, causal_detail = FALSE),
      warnings = tibble::tibble(),
      guardrails = guardrails_empty_tibble()
    )
    b$guardrails <- guardrails_bind(
      evaluate_guardrails_fit(object, fit_diag, engine = "myengine"),
      evaluate_guardrails_contextual(object, data, b, engine = "myengine")
    )
    b$summary_text <- build_diagnostics_bundle_summary(b)
    b <- enrich_bundle_semantic_sections(b)
    attr(b, "ild_fit") <- object
    attr(b, "ild_data") <- data
    b
  }
}

## ---------------------------------------------------------------------------
## 4. Autoplot — compatibility through the bundle
## ---------------------------------------------------------------------------
#
# Users call ild_autoplot(bundle, section = ..., type = ...).
# Implementations live in R/ild_autoplot_bundle.R: add plotters that read
# bundle$fit$convergence, bundle$predictive$simulation_checks$ppc, etc.
# Do not require users to pass the fit again if attributes are set (see
# ?ild_diagnostics_bundle).
#
# For a new section-specific plot, add a branch in ild_autoplot.ild_diagnostics_bundle
# and a plot_bundle_*() helper.
#

## ---------------------------------------------------------------------------
## 5. Provenance — ild_methods() and reporting
## ---------------------------------------------------------------------------
#
# After fitting:
#   attr(fit, "ild_data") <- data
#   attr(fit, "ild_provenance") <- ild_new_analysis_provenance(
#     step = "ild_myfit",
#     args = list(formula = formula, ...),
#     outputs = list(engine = "myengine", ...)
#   )
#
# See R/ild_provenance.R and R/ild_methods.R.
#

## ---------------------------------------------------------------------------
## 6. Tests (minimum)
## ---------------------------------------------------------------------------
#
# - ild_tidy: all ild_tidy_schema()$required columns present (or documented waiver).
# - ild_augment: all ild_augment_schema()$required columns present.
# - ild_diagnose: expect_s3_class(b, "ild_diagnostics_bundle");
#                 expect_no_error(validate_ild_diagnostics_bundle(b))
# - Guardrails: if engine participates in fit-level rules, add a row to
#   tests/testthat/test-guardrails-triggers.R (deterministic fit_diag fixture)
#   or extend test-contract-regression.R patterns.
# - Optional: tests/testthat/helper-contract-fixtures.R for merged contract tests.
#

## ---------------------------------------------------------------------------
## 7. Documentation updates
## ---------------------------------------------------------------------------
#
# - @rdname shared with sibling methods where possible.
# - Describe engine-specific bundle slots in ild_diagnostics_bundle Details or
#   a dedicated topic page.
# - NEWS.md: mention new engine and any new guardrail rule_ids.
# - If normative bundle/schema text changes, edit inst/dev/DEVELOPER_CONTRACTS.md
#   (vignette developer-contracts embeds it).
#

## ---------------------------------------------------------------------------
## 8. NAMESPACE / Collate
## ---------------------------------------------------------------------------
#
# - Export new S3 methods via roxygen2.
# - Add new source files to Collate in DESCRIPTION **before** files that call them.
#

NULL

# Multi-variable lag preparation for panel dynamics workflows

#' Prepare several lag columns and audit them
#'
#' Sequentially applies [ild_lag()] to each named variable with the same
#' \code{mode}, \code{max_gap}, \code{window}, and \code{resolution}, then runs
#' [ild_check_lags()] once on all created lag columns. Use this when building
#' multivariate lag specifications (e.g. several predictors or lag orders) with
#' consistent gap or time-window semantics.
#'
#' @param data An ILD object (see [is_ild()]).
#' @param variables Character vector of column names to lag.
#' @param n Integer vector of lag orders, recycled to \code{length(variables)}.
#'   Ignored for \code{mode = "time_window"} (see [ild_lag()]).
#' @param mode,max_gap,window,resolution Passed to each [ild_lag()] call.
#' @return A list with \code{data} (ILD with lag columns), \code{lag_vars}
#'   (names of created columns), \code{check} (tibble from [ild_check_lags()]),
#'   and \code{spec} (list of arguments for provenance).
#' @seealso \code{vignette("temporal-dynamics-model-choice", package = "tidyILD")},
#'   [ild_crosslag()], [ild_lag()].
#' @export
ild_panel_lag_prepare <- function(data,
                                  variables,
                                  n = 1L,
                                  mode = c("gap_aware", "index", "time_window"),
                                  max_gap = NULL,
                                  window = NULL,
                                  resolution = c("closest_prior", "last_in_window", "mean_in_window")) {
  validate_ild(data)
  mode <- match.arg(mode)
  resolution <- match.arg(resolution)
  variables <- unique(as.character(variables))
  if (length(variables) == 0L) {
    stop("variables must be non-empty.", call. = FALSE)
  }
  for (v in variables) {
    if (!v %in% names(data)) {
      stop("Variable '", v, "' not found in data.", call. = FALSE)
    }
  }
  n <- rep(as.integer(n), length.out = length(variables))
  out <- data
  lag_vars <- character(length(variables))
  for (i in seq_along(variables)) {
    v <- variables[i]
    ni <- n[i]
    out <- ild_lag(
      out,
      dplyr::all_of(v),
      n = ni,
      mode = mode,
      max_gap = max_gap,
      window = window,
      resolution = resolution
    )
    lag_vars[i] <- if (mode == "time_window") {
      paste0(v, "_lag_window")
    } else {
      paste0(v, "_lag", ni)
    }
  }
  chk <- ild_check_lags(out, lag_vars = lag_vars, max_gap = max_gap)
  spec <- list(
    variables = variables,
    n = n,
    mode = mode,
    max_gap = max_gap,
    window = window,
    resolution = resolution
  )
  out <- ild_add_step(
    out,
    "ild_panel_lag_prepare",
    spec,
    list(created = lag_vars)
  )
  list(
    data = out,
    lag_vars = lag_vars,
    check = chk,
    spec = spec
  )
}

# Package-level documentation for ?tidyILD
# Do not add code; this file exists only for the roxygen block below.

#' tidyILD: Tidy Intensive Longitudinal Data Analysis
#'
#' @description
#' tidyILD is a reproducible, tidyverse-style framework for intensive longitudinal
#' data (ILD) analysis in R, with built-in methodological safeguards, provenance
#' tracking, and reporting tools. It supports ecological momentary assessment
#' (EMA) and diary studies with a tidy pipeline from raw data to mixed-effects
#' models: explicit time structure, within-between decomposition, spacing-aware
#' lags, and diagnostics. Use it when you have repeated measures per person over
#' time and want consistent handling of time, gaps, centering, and residual
#' correlation (AR1/CAR1).
#'
#' @details
#' All ILD structure (`.ild_*` columns and `ild_*` metadata) is created only
#' by \code{\link{ild_prepare}} (via the internal constructor). Downstream
#' functions expect data prepared with \code{ild_prepare()}. For the full
#' workflow and applications, see the vignettes. Estimands that require
#' \strong{joint multivariate dynamics}, \strong{penalized high-dimensional}
#' longitudinal models, or full \strong{latent-variable DSEM} are better handled
#' by specialist packages after preprocessing; see
#' \code{vignette("ild-specialist-backends", package = "tidyILD")}.
#'
#' @section Getting started:
#' A minimal workflow: simulate or load data, prepare with
#' \code{\link{ild_prepare}}, inspect with \code{\link{ild_summary}}, apply
#' \code{\link{ild_center}} and \code{\link{ild_lag}}, fit with
#' \code{\link{ild_lme}} or \code{\link{ild_fit}} (same engines plus optional
#' \code{backend = "brms"}), then \code{\link{ild_diagnostics}} or
#' \code{\link{ild_plot}}. State-space: \code{\link{ild_kfas}}. See the examples below.
#'
#' @section Function index by topic:
#' \describe{
#'   \item{Contracts (schemas)}{\code{\link{ild_diagnostics_bundle}}, \code{\link{guardrail_registry}},
#'     \code{\link{ild_tidy_schema}}, \code{\link{ild_augment_schema}}}
#'   \item{Setup and validation}{\code{\link{ild_prepare}}, \code{\link{ild_as_tsibble}},
#'     \code{\link{as_ild}},
#'     \code{\link{is_ild}}, \code{\link{validate_ild}}, \code{\link{ild_meta}}}
#'   \item{Summaries and inspection}{\code{\link{ild_summary}},
#'     \code{\link{ild_spacing_class}}, \code{\link{ild_spacing}},
#'     \code{\link{ild_design_check}}, \code{\link{ild_missing_pattern}},
#'     \code{\link{ild_missing_compliance}}, \code{\link{ild_missing_cohort}},
#'     \code{\link{ild_missing_hazard_first}}, \code{\link{ild_missingness_report}},
#'     \code{\link{ild_missing_bias}}, \code{\link{ild_missing_model}},
#'     \code{\link{ild_ipw_weights}}, \code{\link{ild_ipw_refit}},
#'     \code{\link{ild_plot}} (types: trajectory, gaps, missingness, predicted_trajectory).
#'     These overlap with \code{\link{ild_diagnostics_utilities}} (bundle section providers).}
#'   \item{Within-person and lags}{\code{\link{ild_center}}, \code{\link{ild_center_plot}},
#'     \code{\link{ild_decomposition}}, \code{\link{ild_lag}}, \code{\link{ild_check_lags}},
#'     \code{\link{ild_panel_lag_prepare}}, \code{\link{ild_crosslag}}, \code{\link{ild_align}}}
#'   \item{Modeling}{\code{\link{ild_fit}} (unified: lme4, nlme, brms), \code{\link{ild_lme}},
#'     \code{\link{ild_brms}} (Bayesian), \code{\link{ild_kfas}} (KFAS state-space; not via \code{ild_fit}),
#'     \code{\link{ild_person_model}}, \code{\link{ild_heterogeneity}}, \code{\link{ild_heterogeneity_stratified}},
#'     \code{\link{ild_tvem}} (time-varying effects)}
#'   \item{Diagnostics and visualization}{\code{\link{ild_diagnose}} (\code{\link{ild_diagnostics_bundle}}),
#'     \code{\link{ild_diagnostics_utilities}}, \code{\link{ild_acf}}, \code{\link{ild_diagnostics}},
#'     \code{\link{ild_plot}} (types: fitted, predicted_trajectory, residual_acf; optional \code{facet_by}),
#'     \code{\link{ild_plot_predicted_trajectory}}, \code{\link{ild_heatmap}},
#'     \code{\link{ild_spaghetti}}, \code{\link{ild_circadian}}, \code{\link{ild_tvem_plot}}}
#'   \item{Provenance and methods}{\code{\link{ild_provenance}}, \code{\link{ild_history}},
#'     \code{\link{ild_methods}}, \code{\link{ild_report}}, \code{\link{ild_compare_pipelines}},
#'     \code{\link{ild_compare_fits}}, \code{\link{ild_export_provenance}}}
#'   \item{Reproducibility}{\code{\link{ild_manifest}}, \code{\link{ild_bundle}}}
#'   \item{Package standards (developers)}{\code{vignette("developer-contracts", package = "tidyILD")};
#'     normative spec: \code{system.file("dev", "DEVELOPER_CONTRACTS.md", package = "tidyILD")}}
#'   \item{Utilities and data}{\code{\link{ild_simulate}}, \code{\link{ild_power}}, \code{\link{ema_example}}}
#'   \item{Person-level}{\code{\link{ild_person_model}}, \code{\link{ild_person_distribution}}}
#'   \item{Model tidiers}{\code{\link{ild_prior_ild}}, \code{\link{ild_tidy}}, \code{\link{ild_augment}},
#'     \code{\link{ild_diagnose}}, \code{\link{ild_autoplot}},
#'     \code{\link{augment_ild_model}}, \code{\link{tidy_ild_model}}
#'     (model or robust SE via \code{se = "robust"}), \code{\link{ild_robust_se}};
#'     \code{\link{tidy.ild_lme}}, \code{\link{augment.ild_lme}} (broom.mixed, see \code{\link{broom_ild_lme}})}
#' }
#'
#' @section Visualization:
#' \code{\link{ild_plot}}, \code{\link{ild_spaghetti}}, and \code{\link{ild_heatmap}} cover trajectories,
#' heatmaps, gaps, and missingness; optional \code{facet_by} adds panels (e.g. by cluster).
#' For observed and fitted lines over time, use \code{\link{ild_plot_predicted_trajectory}} or
#' \code{ild_plot(..., type = "predicted_trajectory")}. After \code{\link{ild_diagnose}}, use
#' \code{\link{ild_autoplot}} on the bundle. Backend-specific plots include \code{\link{ild_tvem_plot}},
#' \code{\link{ild_plot_states}}, \code{\link{ild_plot_forecast}}, and \code{\link{ild_plot_filtered_vs_smoothed}}
#' (KFAS). See \code{vignette("visualization-in-tidyILD", package = "tidyILD")} for a question-to-function map
#' and recipes (including partial effects via \strong{marginaleffects} / \strong{ggeffects} on \code{_wp} / \code{_bp} columns).
#'
#' @section Vignettes:
#' \code{browseVignettes("tidyILD")} lists all vignettes. Key entries:
#' \itemize{
#'   \item \emph{From raw data to model with tidyILD} --- Full pipeline:
#'     prepare, inspect, center, lag, fit, diagnose.
#'   \item \emph{Visualization in tidyILD} --- Index of plots, bundle sections, \code{facet_by},
#'     predicted trajectories, and partial-effects templates.
#'   \item \emph{Short analysis report} --- Fit, tidy fixed effects,
#'     fitted vs observed, residual ACF and Q-Q.
#'   \item \emph{Within-between decomposition and irregular spacing} ---
#'     Centering (BP/WP), gap-aware lags, spacing classification.
#'   \item \emph{Temporal dynamics: choosing a model for ILD} ---
#'     Maps estimands to lags, residual AR (\code{ild_lme}), time-varying effects (\code{ild_tvem}),
#'     and state-space backends (\code{ild_kfas}, \code{ild_ctsem}).
#'   \item \emph{Specialist backends: when to move beyond the default stack} ---
#'     Handoffs to dynamite, PGEE, DSEM, and multivariate workflows; export patterns after prepare/center/lag.
#'   \item \emph{Reproducible ILD workflows with tidyILD provenance} ---
#'     Inspect history, generate methods text, ild_report(), export and compare pipelines.
#'   \item \emph{Glossary and quick-start checklist} --- Table of main
#'     functions and a short checklist.
#'   \item \emph{Heterogeneity and person-specific effects} --- BLUPs vs \code{ild_person_model},
#'     \code{ild_heterogeneity()}, bundle plots.
#'   \item \emph{Missingness in ILD: diagnostics and sensitivity routes} ---
#'     \code{ild_missingness_report()}, MAR/MNAR context, IPW templates.
#' }
#'
#' @section Key concepts:
#' \itemize{
#'   \item \strong{ILD:} Intensive longitudinal data; many repeated
#'     measurements per person over time (e.g. EMA).
#'   \item \strong{Within-between decomposition:} \code{\link{ild_center}}
#'     adds \code{_bp} (person mean) and \code{_wp} (within-person deviation);
#'     use WP for within-person effects and BP for between-person or
#'     cross-level terms.
#'   \item \strong{Spacing-aware lags:} \code{\link{ild_lag}} supports
#'     \code{index}, \code{gap_aware} (NA when \code{gap > max_gap}), and
#'     \code{time_window}; avoids misalignment from assuming equal spacing.
#'   \item \strong{Residual correlation:} \code{\link{ild_lme}} can fit nlme
#'     with AR1 or CAR1 for residual autocorrelation;
#'     \code{\link{ild_spacing_class}} helps choose \code{regular-ish} vs
#'     \code{irregular-ish} spacing.
#'   \item \strong{Person-level:} \code{\link{ild_person_model}} fits models separately
#'     per participant; \code{\link{ild_person_distribution}} plots the distribution of
#'     estimates across persons (N-of-1 / idiographic). \code{\link{ild_heterogeneity}} summarizes
#'     partial-pooling person-specific effects from mixed models (contrasts with no pooling).
#' }
#'
#' @author Alex Litovchenko \email{al4877@columbia.edu}
#'
#' @seealso
#' \code{\link{browseVignettes}} and \code{vignette(package = "tidyILD")} for
#' vignettes. Core entry points: \code{\link{ild_prepare}}, \code{\link{ild_lme}}.
#' \code{vignette("ild-specialist-backends", package = "tidyILD")} for using
#' tidyILD as a preprocessing layer with external multivariate or high-dimensional estimators.
#' Related packages: \pkg{lme4}, \pkg{nlme} (model backends),
#' \pkg{broom.mixed} (tidiers).
#'
#' @examples
#' library(tidyILD)
#' d <- ild_simulate(n_id = 10, n_obs_per = 12, irregular = TRUE, seed = 42)
#' x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)
#' ild_summary(x)
#' x <- ild_center(x, y)
#' x <- ild_lag(x, y, mode = "gap_aware", max_gap = 7200)
#' fit <- ild_lme(y ~ 1, data = x, ar1 = TRUE, correlation_class = "CAR1")
#' ild_diagnostics(fit, data = x)
#' ild_plot(fit, type = "fitted")
#'
#' @docType package
#' @name tidyILD-package
#' @importFrom stats fitted predict rstandard
NULL

## ----include=FALSE------------------------------------------------------------
knitr::opts_chunk$set(echo = FALSE)
md_path <- if (file.exists("../inst/dev/DEVELOPER_CONTRACTS.md")) {
  "../inst/dev/DEVELOPER_CONTRACTS.md"
} else {
  system.file("dev", "DEVELOPER_CONTRACTS.md", package = "tidyILD")
}

## ----contracts, results='asis'------------------------------------------------
if (!nzchar(md_path) || !file.exists(md_path)) {
  stop("DEVELOPER_CONTRACTS.md not found; expected inst/dev/DEVELOPER_CONTRACTS.md", call. = FALSE)
}
cat(knitr::knit_child(md_path, quiet = TRUE))

# Internal helpers for ild_augment() schema (ILD_AUGMENT_REQUIRED_COLS / optional)

#' Principled Pearson residuals for augment (sparse: NA if unavailable or length mismatch)
#' @keywords internal
#' @noRd
ild_augment_pearson_residuals <- function(fit, n_expected) {
  out <- rep(NA_real_, n_expected)
  r <- tryCatch(
    stats::residuals(fit, type = "pearson"),
    error = function(e) NULL
  )
  if (is.null(r)) return(out)
  r <- as.numeric(unname(as.vector(r)))
  if (length(r) != n_expected) return(out)
  r
}

#' Build augmented tibble: required columns + optional schema columns
#' @keywords internal
#' @noRd
ild_augment_assemble <- function(.ild_id,
                                 .ild_time,
                                 .outcome,
                                 .fitted,
                                 .resid,
                                 .resid_std,
                                 engine,
                                 model_class,
                                 optional = NULL) {
  n <- length(.ild_id)
  out <- tibble::tibble(
    .ild_id = .ild_id,
    .ild_time = .ild_time,
    .outcome = .outcome,
    .fitted = .fitted,
    .resid = .resid,
    .resid_std = .resid_std,
    engine = engine,
    model_class = model_class
  )
  for (nm in ILD_AUGMENT_OPTIONAL_COLS) {
    if (!is.null(optional) && !is.null(optional[[nm]])) {
      out[[nm]] <- optional[[nm]]
    } else {
      out[[nm]] <- rep(NA_real_, n)
    }
  }
  out
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----plot-examples------------------------------------------------------------
library(tidyILD)
set.seed(1)
d <- ild_simulate(n_id = 24, n_obs_per = 10, seed = 1)
d$cluster <- rep(LETTERS[1:3], length.out = nrow(d))
x <- ild_prepare(d, id = "id", time = "time")
ild_spaghetti(x, var = "y", facet_by = "cluster", max_ids = 12L)

fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
ild_plot_predicted_trajectory(fit, time_var = ".ild_seq", max_ids = 8L, facet_by = "cluster")

## ----partial-effects-template, eval = FALSE-----------------------------------
# # install.packages(c("marginaleffects", "ggeffects"))  # if needed
# x <- ild_center(x, x)
# fit <- ild_lme(y ~ x_wp + x_bp + (1 | id), data = x, warn_uncentered = FALSE)
# # Example (syntax may vary by package version):
# # marginaleffects::plot_predictions(fit, condition = "x_wp")
# # marginaleffects::plot_predictions(fit, condition = "x_bp")
# # ggeffects::ggpredict(fit, terms = "x_wp")  # then plot() or ggplot2 layer

# Compare two pipelines (data or analysis) by provenance
# Uses ild_provenance(), ild_flatten_provenance() from ild_provenance.R, ild_methods.R

#' Compare provenance of two objects and report differences
#'
#' Flattens provenance for each object (data steps plus analysis steps), then
#' compares step sequences and key arguments. Useful to compare preprocessing
#' pipelines, model settings, or to see what changed between two analyses.
#'
#' @param x1 First object (ILD data, model fit, or diagnostics with provenance).
#' @param x2 Second object (same types as x1).
#' @return A list of class \code{ild_compare_pipelines} with \code{only_in_first}
#'   (step names only in x1), \code{only_in_second}, \code{differing} (list of
#'   per-step arg differences), and \code{summary} (character vector of
#'   human-readable differences). If either object has no provenance, returns
#'   a list with \code{message} and empty comparison components.
#' @export
#' @examples
#' set.seed(1)
#' d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 1)
#' x1 <- ild_prepare(d, id = "id", time = "time")
#' x1 <- ild_center(x1, y)
#' x2 <- ild_prepare(d, id = "id", time = "time")
#' x2 <- ild_center(x2, y)
#' ild_compare_pipelines(x1, x2)
ild_compare_pipelines <- function(x1, x2) {
  p1 <- ild_provenance(x1)
  p2 <- ild_provenance(x2)
  if (is.null(p1) && is.null(p2)) {
    out <- list(
      only_in_first = character(),
      only_in_second = character(),
      differing = list(),
      summary = character(),
      message = "Neither object has provenance."
    )
    class(out) <- "ild_compare_pipelines"
    return(out)
  }
  if (is.null(p1)) {
    out <- list(
      only_in_first = character(),
      only_in_second = character(),
      differing = list(),
      summary = "First object has no provenance.",
      message = "x1 has no recorded provenance."
    )
    class(out) <- "ild_compare_pipelines"
    return(out)
  }
  if (is.null(p2)) {
    out <- list(
      only_in_first = character(),
      only_in_second = character(),
      differing = list(),
      summary = "Second object has no provenance.",
      message = "x2 has no recorded provenance."
    )
    class(out) <- "ild_compare_pipelines"
    return(out)
  }
  steps1 <- ild_flatten_provenance(p1)
  steps2 <- ild_flatten_provenance(p2)
  names1 <- vapply(steps1, function(s) s$step, character(1L))
  names2 <- vapply(steps2, function(s) s$step, character(1L))
  only_first <- setdiff(names1, names2)
  only_second <- setdiff(names2, names1)
  common <- intersect(names1, names2)
  differing <- list()
  summary_lines <- character()
  for (step_name in common) {
    idx1 <- which(names1 == step_name)[1L]
    idx2 <- which(names2 == step_name)[1L]
    if (is.na(idx1) || is.na(idx2)) next
    a1 <- steps1[[idx1]]$args
    a2 <- steps2[[idx2]]$args
    if (is.null(a1)) a1 <- list()
    if (is.null(a2)) a2 <- list()
    all_keys <- unique(c(names(a1), names(a2)))
    for (k in all_keys) {
      v1 <- a1[[k]]
      v2 <- a2[[k]]
      if (identical(v1, v2)) next
      differing <- c(differing, list(list(step = step_name, arg = k, value1 = v1, value2 = v2)))
      v1_str <- if (length(v1) > 1) paste(v1, collapse = ", ") else as.character(v1)
      v2_str <- if (length(v2) > 1) paste(v2, collapse = ", ") else as.character(v2)
      if (nchar(v1_str) > 40) v1_str <- paste0(substr(v1_str, 1, 37), "...")
      if (nchar(v2_str) > 40) v2_str <- paste0(substr(v2_str, 1, 37), "...")
      summary_lines <- c(summary_lines, paste0("Different ", k, " (", step_name, "): ", v1_str, " vs ", v2_str))
    }
  }
  if (length(only_first) > 0) {
    summary_lines <- c(summary_lines, paste0("Only in first: ", paste(only_first, collapse = ", ")))
  }
  if (length(only_second) > 0) {
    summary_lines <- c(summary_lines, paste0("Only in second: ", paste(only_second, collapse = ", ")))
  }
  if (length(summary_lines) == 0 && length(only_first) == 0 && length(only_second) == 0) {
    summary_lines <- "Pipelines match (same steps and arguments)."
  }
  out <- list(
    only_in_first = only_first,
    only_in_second = only_second,
    differing = differing,
    summary = summary_lines
  )
  class(out) <- "ild_compare_pipelines"
  out
}

#' @export
print.ild_compare_pipelines <- function(x, ...) {
  if (!is.null(x$message)) {
    cat(x$message, "\n")
    return(invisible(x))
  }
  cat("Pipeline comparison\n")
  if (length(x$only_in_first) > 0) {
    cat("  Only in first: ", paste(x$only_in_first, collapse = ", "), "\n", sep = "")
  }
  if (length(x$only_in_second) > 0) {
    cat("  Only in second: ", paste(x$only_in_second, collapse = ", "), "\n", sep = "")
  }
  for (s in x$summary) {
    cat("  ", s, "\n", sep = "")
  }
  invisible(x)
}

# Internal state-space specification vocabulary for ild_kfas() (KFAS backend)

#' @keywords internal
ILD_KFAS_STATE_SPECS <- c(
  "local_level",
  "local_trend",
  "ar1_state",
  "regression_local_level",
  "regression_ar1_state"
)

ILD_KFAS_SCHEMA_VERSION <- "0.1"

#' Validate and normalize state_spec for ild_kfas
#' @keywords internal
#' @noRd
ild_kfas_normalize_state_spec <- function(state_spec, implemented_only = TRUE) {
  ss <- state_spec[1L]
  if (!ss %in% ILD_KFAS_STATE_SPECS) {
    stop(
      "state_spec must be one of: ",
      paste(ILD_KFAS_STATE_SPECS, collapse = ", "),
      call. = FALSE
    )
  }
  if (isTRUE(implemented_only) && !ss %in% c("local_level")) {
    stop(
      "state_spec = '", ss, "' is not implemented yet. Currently supported: local_level.",
      call. = FALSE
    )
  }
  ss
}

#' Build internal spec list for provenance and downstream methods
#' @keywords internal
#' @noRd
ild_kfas_internal_spec <- function(state_spec,
                                   observation_family,
                                   time_units,
                                   irregular_time,
                                   fit_context = "single_series",
                                   pool_mode = "none") {
  list(
    state_spec = state_spec,
    observation_family = observation_family,
    time_units = time_units,
    irregular_time = irregular_time,
    fit_context = fit_context,
    pool_mode = pool_mode,
    schema_version = ILD_KFAS_SCHEMA_VERSION
  )
}

#' Plot the time-varying coefficient from a TVEM fit
#'
#' Builds a line plot of the smooth term for the time-varying effect of the
#' predictor (with optional confidence band). Uses a grid over the time variable
#' and [mgcv::predict.gam()] with \code{type = "terms"}.
#'
#' @param tvem_fit A fitted object from [ild_tvem()] (class \code{tidyild_tvem}).
#' @param n_grid Integer. Number of points over the time range for the curve (default 100).
#' @param level Numeric. Confidence level for the band (default 0.95).
#' @return A \code{ggplot} object (time on x-axis, estimated effect on y-axis).
#' @export
#' @examples
#' set.seed(1)
#' d <- ild_simulate(n_id = 10, n_obs_per = 15, seed = 1)
#' d$x <- rnorm(nrow(d))
#' x <- ild_prepare(d, id = "id", time = "time")
#' tv <- ild_tvem(x, "y", "x", k = 5, re_id = TRUE)
#' ild_tvem_plot(tv)
ild_tvem_plot <- function(tvem_fit, n_grid = 100L, level = 0.95) {
  if (!inherits(tvem_fit, "tidyild_tvem")) {
    stop("tvem_fit must be from ild_tvem().", call. = FALSE)
  }
  meta <- attr(tvem_fit, "ild_tvem_meta", exact = TRUE)
  if (is.null(meta)) stop("tvem_fit missing ild_tvem_meta.", call. = FALSE)
  time_var <- meta$time_var
  predictor <- meta$predictor
  outcome <- meta$outcome
  dat <- tvem_fit$model
  if (is.null(dat) || !time_var %in% names(dat)) {
    stop("Model frame missing or missing time_var.", call. = FALSE)
  }
  t_range <- range(dat[[time_var]], na.rm = TRUE)
  t_grid <- seq(t_range[1], t_range[2], length.out = n_grid)
  newdata <- data.frame(t_grid, 1)
  names(newdata) <- c(time_var, predictor)
  if (".ild_id" %in% names(dat) && meta$re_id) {
    id_levs <- levels(factor(dat[[".ild_id"]]))
    newdata$.ild_id <- factor(id_levs[1L], levels = id_levs)
  }
  pred <- mgcv::predict.gam(tvem_fit, newdata = newdata, type = "terms", se.fit = TRUE)
  term_names <- colnames(pred$fit)
  idx <- which(grepl(predictor, term_names, fixed = TRUE) | grepl("by", term_names, fixed = TRUE))
  if (length(idx) == 0L) idx <- 2L
  if (length(idx) > 1L) idx <- idx[1L]
  fit_vals <- as.vector(pred$fit[, idx])
  se_vals <- as.vector(pred$se.fit[, idx])
  half <- (1 - level) / 2
  z <- stats::qnorm(1 - half)
  df_plot <- data.frame(
    time = t_grid,
    effect = fit_vals,
    se = se_vals,
    ci_low = fit_vals - z * se_vals,
    ci_high = fit_vals + z * se_vals
  )
  ggplot2::ggplot(df_plot, ggplot2::aes(x = .data$time, y = .data$effect)) +
    ggplot2::geom_ribbon(ggplot2::aes(ymin = .data$ci_low, ymax = .data$ci_high), alpha = 0.2, fill = "gray50") +
    ggplot2::geom_line(linewidth = 1) +
    ggplot2::labs(
      x = time_var,
      y = paste0("Time-varying effect (", predictor, ")"),
      title = paste0("TVEM: ", outcome, " ~ s(", time_var, ") + s(", time_var, ", by = ", predictor, ")")
    ) +
    ggplot2::theme_minimal()
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
has_kfas <- requireNamespace("KFAS", quietly = TRUE)

## ----spacing_demo, eval = TRUE------------------------------------------------
library(tidyILD)
set.seed(3)
d <- ild_simulate(n_id = 1, n_obs_per = 50, irregular = TRUE, seed = 11)
x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)
ild_summary(x)$summary
ild_spacing_class(x)

## ----kfas_irregular, eval = has_kfas------------------------------------------
fit <- suppressWarnings(
  ild_kfas(
    x,
    outcome = "y",
    state_spec = "local_level",
    time_units = "seconds",
    irregular_time = TRUE
  )
)
diag <- ild_diagnose(fit)
# When spacing is irregular-ish and IQR/median interval ratio > 0.75, expect
# GR_KFAS_HIGH_IRREGULARITY_FOR_DISCRETE_TIME among triggered guardrails:
diag$guardrails[, c("rule_id", "message")]

#' Within-person and between-person variance decomposition
#'
#' Reports WP and BP variance and their ratio for selected variables. Use as a
#' diagnostic and teaching tool: large ratio suggests within-person variance
#' dominates; small ratio suggests between-person differences dominate. Helps
#' avoid conflating WP and BP effects in modeling.
#'
#' @param x An ILD object (see [is_ild()]).
#' @param ... Variables to decompose (tidy-select). Must be numeric.
#' @param plot Logical. If \code{TRUE}, return a list with \code{table} and \code{plot} (WP vs BP density overlay). Default \code{FALSE}.
#' @return A tibble with columns \code{variable}, \code{wp_var}, \code{bp_var}, \code{ratio} (wp_var / bp_var; \code{Inf} if bp_var is 0). If \code{plot = TRUE}, a list with \code{table} and \code{plot} (ggplot).
#' @export
#' @examples
#' d <- ild_simulate(n_id = 10, n_obs_per = 8, seed = 1)
#' x <- ild_prepare(d, id = "id", time = "time")
#' ild_decomposition(x, y)
#' @importFrom dplyr select
ild_decomposition <- function(x, ..., plot = FALSE) {
  validate_ild(x)
  id_col <- ".ild_id"
  vars <- names(dplyr::select(x, ...))
  if (length(vars) == 0) stop("No variables selected for decomposition.", call. = FALSE)
  for (v in vars) {
    if (!v %in% names(x)) stop("Variable '", v, "' not found in data.", call. = FALSE)
    if (!is.numeric(x[[v]])) stop("Variable '", v, "' is not numeric.", call. = FALSE)
  }
  out_rows <- vector("list", length(vars))
  for (i in seq_along(vars)) {
    v <- vars[i]
    means_i <- tapply(x[[v]], x[[id_col]], mean, na.rm = TRUE)
    bp_var <- stats::var(as.vector(means_i))
    if (is.na(bp_var)) bp_var <- 0
    wp_var_by_id <- tapply(x[[v]], x[[id_col]], function(z) stats::var(z, na.rm = TRUE))
    wp_var_by_id[is.na(wp_var_by_id)] <- 0
    wp_var <- mean(wp_var_by_id, na.rm = TRUE)
    if (is.na(wp_var)) wp_var <- 0
    ratio <- if (bp_var > 0) wp_var / bp_var else Inf
    out_rows[[i]] <- list(variable = v, wp_var = wp_var, bp_var = bp_var, ratio = ratio)
  }
  tbl <- tibble::tibble(
    variable = vapply(out_rows, `[[`, character(1), "variable"),
    wp_var = vapply(out_rows, `[[`, double(1), "wp_var"),
    bp_var = vapply(out_rows, `[[`, double(1), "bp_var"),
    ratio = vapply(out_rows, `[[`, double(1), "ratio")
  )
  if (!plot) return(tbl)
  p <- plot_ild_decomposition(x, vars)
  list(table = tbl, plot = p)
}

#' Plot WP vs BP density for decomposition (internal / called by ild_decomposition when plot=TRUE)
#' @param x ILD object
#' @param vars character vector of variable names
#' @return ggplot object (overlay density of WP and BP components for first variable)
#' @noRd
plot_ild_decomposition <- function(x, vars) {
  id_col <- ".ild_id"
  v <- vars[1]
  means_i <- tapply(x[[v]], x[[id_col]], mean, na.rm = TRUE)
  person_means <- means_i[as.character(x[[id_col]])]
  wp_vals <- x[[v]] - person_means
  bp_vals <- as.vector(means_i)
  df_wp <- data.frame(value = wp_vals, component = "WP (within-person deviation)")
  df_bp <- data.frame(value = bp_vals, component = "BP (person mean)")
  df <- rbind(df_wp, df_bp)
  ggplot2::ggplot(df, ggplot2::aes(x = .data$value, color = .data$component, fill = .data$component)) +
    ggplot2::geom_density(alpha = 0.3) +
    ggplot2::labs(title = paste0("WP vs BP: ", v), x = v, y = "Density") +
    ggplot2::theme_minimal()
}

#' Standalone WP/BP centering plot
#'
#' Shows within-person deviation and between-person (person mean) distribution
#' for selected variable(s). Uses the same plot as [ild_decomposition(..., plot = TRUE)].
#' Useful when you only want the visualization without the variance table.
#'
#' @param x An ILD object (see [is_ild()]).
#' @param ... Variables to plot (tidy-select). Must be numeric. Only the first is plotted.
#' @return A ggplot object (WP vs BP density overlay for the first selected variable).
#' @export
#' @examples
#' d <- ild_simulate(n_id = 10, n_obs_per = 8, seed = 1)
#' x <- ild_prepare(d, id = "id", time = "time")
#' ild_center_plot(x, y)
ild_center_plot <- function(x, ...) {
  validate_ild(x)
  vars <- names(dplyr::select(x, ...))
  if (length(vars) == 0) stop("No variables selected for center plot.", call. = FALSE)
  for (v in vars) {
    if (!v %in% names(x)) stop("Variable '", v, "' not found in data.", call. = FALSE)
    if (!is.numeric(x[[v]])) stop("Variable '", v, "' is not numeric.", call. = FALSE)
  }
  plot_ild_decomposition(x, vars)
}

test_that("ild_decomposition returns table with wp_var, bp_var, ratio", {
  d <- ild_simulate(n_id = 10, n_obs_per = 8, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_decomposition(x, "y")
  expect_s3_class(out, "tbl_df")
  expect_named(out, c("variable", "wp_var", "bp_var", "ratio"))
  expect_equal(out$variable, "y")
  expect_true(is.numeric(out$wp_var) && is.numeric(out$bp_var) && is.numeric(out$ratio))
  expect_true(out$wp_var >= 0 && out$bp_var >= 0)
})

test_that("ild_decomposition with plot returns list with table and plot", {
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  out <- ild_decomposition(x, "y", plot = TRUE)
  expect_true(is.list(out))
  expect_named(out, c("table", "plot"))
  expect_s3_class(out$table, "tbl_df")
  expect_s3_class(out$plot, "gg")
})

test_that("ild_decomposition errors on non-numeric variable", {
  d <- ild_simulate(n_id = 3, n_obs_per = 4, seed = 1)
  d$cat <- rep(c("a", "b"), length.out = nrow(d))
  x <- ild_prepare(d, id = "id", time = "time")
  expect_error(ild_decomposition(x, "cat"), "not numeric")
})

test_that("ild_center_plot returns ggplot matching decomposition plot", {
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  p <- ild_center_plot(x, "y")
  expect_s3_class(p, "gg")
  p2 <- ild_decomposition(x, "y", plot = TRUE)$plot
  expect_s3_class(p2, "gg")
})

#' ILD design diagnostics: spacing, WP/BP, missingness, and recommendations
#'
#' Aggregates [ild_summary()], [ild_spacing()], [ild_spacing_class()], and
#' optionally [ild_decomposition()] and [ild_missing_pattern()] into one
#' design summary. Use before modeling to see spacing class, correlation
#' recommendation, within- vs between-person variance, and missingness.
#'
#' Also a **section provider** for [ild_diagnose()] (see [ild_diagnostics_utilities]).
#'
#' @section Bundle integration:
#' [ild_diagnose()] embeds this object in \code{design$ild_design_check} on the
#' \code{\link{ild_diagnostics_bundle}}.
#'
#' @family ild_diagnostics_utilities
#' @param x An ILD object (see [is_ild()]).
#' @param vars Optional character vector of variable names for decomposition
#'   and missingness. If \code{NULL}, only spacing and summary are computed;
#'   \code{wp_bp} and \code{missingness} will be \code{NULL}.
#' @return A list of class \code{ild_design_check}: \code{summary} (from ild_summary),
#'   \code{spacing_class} (regular-ish / irregular-ish), \code{spacing} (from ild_spacing),
#'   \code{recommendation} (AR1/CAR1 text), \code{wp_bp} (decomposition tibble or NULL),
#'   \code{missingness} (list with \code{summary} tibble and \code{pct_na} overall, or NULL).
#'   Use \code{print()} for a human-readable summary.
#' @seealso [ild_diagnose()], [ild_diagnostics_bundle()]
#' @export
#' @examples
#' d <- ild_simulate(n_id = 10, n_obs_per = 8, irregular = TRUE, seed = 1)
#' x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)
#' ild_design_check(x, vars = "y")
ild_design_check <- function(x, vars = NULL) {
  validate_ild(x)
  sum_out <- ild_summary(x)
  spacing_out <- ild_spacing(x)
  spacing_class <- ild_spacing_class(x)
  wp_bp <- NULL
  missingness <- NULL
  if (!is.null(vars)) {
    vars <- intersect(vars, names(x))
    ild_cols <- c(".ild_id", ".ild_time", ".ild_time_num", ".ild_seq", ".ild_dt", ".ild_gap")
    vars <- setdiff(vars, ild_cols)
    if (length(vars) > 0) {
      wp_bp <- ild_decomposition(x, dplyr::all_of(vars))
      miss_out <- ild_missing_pattern(x, vars = vars)
      pct_na <- if (nrow(miss_out$summary) > 0) mean(miss_out$summary$pct_na) else NA_real_
      missingness <- list(summary = miss_out$summary, pct_na = pct_na)
    }
  }
  out <- list(
    summary = sum_out$summary,
    spacing_class = spacing_class,
    spacing = spacing_out,
    recommendation = spacing_out$recommendation,
    wp_bp = wp_bp,
    missingness = missingness
  )
  class(out) <- c("ild_design_check", "list")
  out
}

#' @export
print.ild_design_check <- function(x, ...) {
  cat("ILD design check\n")
  cat("---------------\n")
  if (!is.null(x$summary) && nrow(x$summary) > 0) {
    cat("N persons:", x$summary$n_id[1], "| N obs:", x$summary$n_obs[1], "\n")
  }
  cat("Spacing class:", x$spacing_class, "\n")
  cat("Recommendation:", x$recommendation, "\n")
  if (!is.null(x$wp_bp) && nrow(x$wp_bp) > 0) {
    tot_var <- x$wp_bp$wp_var + x$wp_bp$bp_var
    pct_wp <- ifelse(tot_var > 0, 100 * x$wp_bp$wp_var / tot_var, NA_real_)
    pct_bp <- ifelse(tot_var > 0, 100 * x$wp_bp$bp_var / tot_var, NA_real_)
    cat("Within-person variance:", round(pct_wp[1], 1), "%\n")
    cat("Between-person variance:", round(pct_bp[1], 1), "%\n")
  }
  if (!is.null(x$missingness)) {
    cat("Missingness (mean pct across vars):", round(x$missingness$pct_na, 1), "%\n")
  }
  invisible(x)
}

#' Create a reproducibility manifest
#'
#' Captures timestamp, optional seed, optional scenario fingerprint,
#' session info, and optional git SHA for use when saving or serializing
#' results (e.g. after [ild_lme()] or [ild_diagnostics()]). The return
#' value is a serializable list suitable for [saveRDS()] or [ild_bundle()].
#'
#' @param seed Optional integer. Seed used for the run (e.g. from
#'   [ild_simulate()] or set before fitting). Not captured automatically;
#'   pass explicitly if you want it in the manifest.
#' @param scenario Optional. Named list or character string describing the
#'   run (e.g. formula, n_obs, n_id, ar1). Build from [ild_summary()] or
#'   a short list when calling after [ild_lme()] / [ild_diagnostics()].
#' @param include_session Logical. If `TRUE` (default), include
#'   [utils::sessionInfo()] in the manifest. Set to `FALSE` to reduce size.
#' @param include_git Logical. If `TRUE`, attempt to record the current
#'   git commit SHA from \code{git_path}. Default `FALSE`.
#' @param git_path Character. Path to the repository root (default
#'   \code{"."}). Used only when \code{include_git = TRUE}.
#' @return A list with elements \code{timestamp} (POSIXct), \code{seed}
#'   (integer or NULL), \code{scenario} (as provided or NULL),
#'   \code{session_info} (list from sessionInfo() or NULL),
#'   \code{git_sha} (length-1 character or NA). All elements are
#'   serializable.
#' @examples
#' m <- ild_manifest()
#' names(m)
#' m <- ild_manifest(seed = 42, scenario = list(n_obs = 100, formula = "y ~ x"))
#' m$seed
#' m$scenario
#' @export
ild_manifest <- function(seed = NULL,
                        scenario = NULL,
                        include_session = TRUE,
                        include_git = FALSE,
                        git_path = ".") {
  timestamp <- Sys.time()
  session_info <- if (include_session) utils::sessionInfo() else NULL
  git_sha <- NA_character_
  if (include_git && is.character(git_path) && length(git_path) == 1) {
    out <- tryCatch(
      system2("git", c("-C", git_path, "rev-parse", "HEAD"), stdout = TRUE, stderr = FALSE),
      error = function(e) character(0)
    )
    if (length(out) >= 1 && nzchar(trimws(out[1]))) {
      git_sha <- trimws(out[1])
    }
  }
  list(
    timestamp = timestamp,
    seed = seed,
    scenario = scenario,
    session_info = session_info,
    git_sha = git_sha
  )
}

#' Bundle a result with a reproducibility manifest
#'
#' Combines a result (e.g. a fit from [ild_lme()] or output from
#' [ild_diagnostics()]) with a manifest and optional label for one-shot
#' saving. Typical use: \code{saveRDS(ild_bundle(fit, label = "model_ar1"), "run.rds")}.
#' You can build a manifest with [ild_manifest()] and pass \code{scenario}
#' (e.g. from [ild_summary()]) and \code{seed} before bundling.
#'
#' @param result Any object (e.g. fitted model, diagnostics list).
#' @param manifest List. Reproducibility manifest from [ild_manifest()].
#'   If \code{NULL}, [ild_manifest()] is called with default arguments.
#' @param label Optional character. Short label for the run (e.g.
#'   \code{"model_ar1"} or \code{"diagnostics"}).
#' @return A list with elements \code{result}, \code{manifest}, \code{label},
#'   suitable for [saveRDS()].
#' @examples
#' dat <- ild_prepare(ild_simulate(seed = 1), "id", "time")
#' fit <- ild_lme(y ~ 1 + (1 | id), dat, ar1 = FALSE, warn_no_ar1 = FALSE)
#' b <- ild_bundle(fit, label = "ar1")
#' names(b)
#' b <- ild_bundle(fit, manifest = ild_manifest(seed = 1, scenario = list(n_obs = 50)), label = "run1")
#' @export
ild_bundle <- function(result, manifest = NULL, label = NULL) {
  if (is.null(manifest)) manifest <- ild_manifest()
  list(result = result, manifest = manifest, label = label)
}

test_that("ild_manifest returns list with expected names", {
  m <- ild_manifest()
  expect_type(m, "list")
  expect_named(m, c("timestamp", "seed", "scenario", "session_info", "git_sha"))
  expect_s3_class(m$timestamp, "POSIXct")
  expect_true(difftime(Sys.time(), m$timestamp, units = "secs") < 60)
  expect_null(m$seed)
  expect_null(m$scenario)
  expect_type(m$session_info, "list")
  expect_length(m$git_sha, 1)
})

test_that("ild_manifest include_session FALSE gives NULL session_info", {
  m <- ild_manifest(include_session = FALSE)
  expect_null(m$session_info)
})

test_that("ild_manifest stores seed and scenario", {
  m <- ild_manifest(seed = 42, scenario = list(a = 1, n_obs = 100))
  expect_equal(m$seed, 42)
  expect_equal(m$scenario, list(a = 1, n_obs = 100))
})

test_that("ild_manifest include_git returns character or NA", {
  m <- ild_manifest(include_git = TRUE)
  expect_length(m$git_sha, 1)
  expect_true(is.character(m$git_sha))
  expect_true(is.na(m$git_sha) || nzchar(m$git_sha))
})

test_that("ild_bundle returns result, manifest, label", {
  b <- ild_bundle(1 + 1, label = "x")
  expect_named(b, c("result", "manifest", "label"))
  expect_equal(b$result, 2)
  expect_equal(b$label, "x")
  expect_named(b$manifest, c("timestamp", "seed", "scenario", "session_info", "git_sha"))
})

test_that("ild_bundle with NULL manifest calls ild_manifest()", {
  b <- ild_bundle("fit", manifest = NULL)
  expect_equal(b$result, "fit")
  expect_null(b$label)
  expect_type(b$manifest, "list")
  expect_true("timestamp" %in% names(b$manifest))
})

test_that("ild_bundle preserves user-supplied manifest and label", {
  m <- ild_manifest(seed = 1, scenario = list(formula = "y ~ x"), include_session = FALSE)
  b <- ild_bundle("result", manifest = m, label = "run1")
  expect_equal(b$manifest$seed, 1)
  expect_equal(b$manifest$scenario$formula, "y ~ x")
  expect_equal(b$label, "run1")
})

test_that("ild_bundle round-trip via saveRDS preserves structure", {
  skip_on_cran()  # file I/O; run locally and in CI
  tf <- tempfile(fileext = ".rds")
  on.exit(unlink(tf, force = TRUE))
  b <- ild_bundle(1 + 1, manifest = ild_manifest(include_session = FALSE), label = "x")
  saveRDS(b, tf)
  b2 <- readRDS(tf)
  expect_named(b2, c("result", "manifest", "label"))
  expect_equal(b2$result, b$result)
  expect_equal(b2$label, b$label)
  expect_named(b2$manifest, names(b$manifest))
})

# Unified S3 API for fitting, tidying, augmentation, diagnostics, and plotting
# Backends: lme4/nlme via ild_lme(); brms via ild_brms().
# State-space / latent dynamics wrappers: use ild_kfas() or ild_ctsem().

#' Fit a mixed model to ILD data (unified entry point)
#'
#' Façade that selects the estimation backend explicitly. Equivalent to calling
#' [ild_lme()] for \code{"lme4"} / \code{"nlme"}, or [ild_brms()] for \code{"brms"}.
#' Named functions remain the primary APIs for full documentation and examples.
#'
#' **State-space / latent-dynamics models** are not mixed-model formulas:
#' use [ild_kfas()] or [ild_ctsem()] directly;
#' see \code{vignette("kfas-choosing-backend", package = "tidyILD")}.
#'
#' To pass \code{backend} to \code{brms::brm()} (e.g. \code{"rstan"} vs \code{"cmdstanr"}), use
#' [ild_brms()] directly: \code{ild_fit()} reserves \code{backend} for the tidyILD engine
#' (\code{"lme4"}, \code{"nlme"}, \code{"brms"}).
#'
#' @param formula Passed to [ild_lme()] or [ild_brms()].
#' @param data An ILD object (see [is_ild()]).
#' @param backend Character. \code{"lme4"} (default, [lme4::lmer()]), \code{"nlme"}
#'   ([nlme::lme()] with AR1/CAR1 residual structure), or \code{"brms"} ([brms::brm()]).
#' @param correlation_class,random,warn_no_ar1 Passed to [ild_lme()] when
#'   \code{backend} is \code{"lme4"} or \code{"nlme"}. Ignored when \code{backend = "brms"},
#'   except \code{correlation_class} must be \code{"auto"} for \code{"brms"} (non-\code{"auto"}
#'   values are for residual correlation under \code{nlme} only and will error).
#' @param warn_uncentered Passed to [ild_lme()] or [ild_brms()].
#' @param prior,prior_template Passed to [ild_brms()] when \code{backend = "brms"}.
#'   For \code{"lme4"} / \code{"nlme"}, \code{prior} must be \code{NULL} and
#'   \code{prior_template} must remain the default (\code{"default"}); otherwise an error
#'   is raised so mistaken cross-backend arguments are not silently ignored.
#' @param ... Passed to [ild_lme()] / [lme4::lmer()] / [nlme::lme()] or [ild_brms()].
#' @return A fitted model: \code{lmerMod} / \code{lme} from [ild_lme()], or \code{brmsfit}
#'   from [ild_brms()], each with the same attributes as those functions document.
#' @seealso [ild_kfas()] and [ild_ctsem()] for state-space / latent-dynamics
#'   backends (not available via \code{ild_fit()}).
#' @export
ild_fit <- function(formula,
                    data,
                    backend = c("lme4", "nlme", "brms"),
                    correlation_class = c("auto", "AR1", "CAR1"),
                    random = ~ 1 | .ild_id,
                    warn_no_ar1 = TRUE,
                    warn_uncentered = TRUE,
                    prior = NULL,
                    prior_template = c("default", "weakly_informative", "minimal_shrinkage"),
                    ...) {
  validate_ild(data)
  backend <- match.arg(backend)
  prior_template_resolved <- match.arg(prior_template)

  if (backend == "brms") {
    cc <- match.arg(correlation_class, c("auto", "AR1", "CAR1"))
    if (cc != "auto") {
      stop(
        "correlation_class is only used when backend = \"nlme\". ",
        "For brms, use ild_fit(..., backend = \"brms\") without AR1/CAR1 options, or use ild_brms() directly.",
        call. = FALSE
      )
    }
    return(ild_brms(
      formula = formula,
      data = data,
      prior = prior,
      prior_template = prior_template_resolved,
      warn_uncentered = warn_uncentered,
      ...
    ))
  }

  if (!is.null(prior)) {
    stop("prior is only used when backend = \"brms\". Use ild_fit(..., backend = \"brms\") or ild_brms().", call. = FALSE)
  }
  if (prior_template_resolved != "default") {
    stop("prior_template is only used when backend = \"brms\". Use ild_fit(..., backend = \"brms\") or ild_brms().", call. = FALSE)
  }

  if (backend == "lme4") {
    return(ild_lme(
      formula, data,
      ar1 = FALSE,
      warn_no_ar1 = warn_no_ar1,
      warn_uncentered = warn_uncentered,
      ...
    ))
  }
  ild_lme(
    formula, data,
    ar1 = TRUE,
    correlation_class = correlation_class,
    random = random,
    warn_no_ar1 = warn_no_ar1,
    warn_uncentered = warn_uncentered,
    ...
  )
}

#' Tidy fixed effects from an ILD model fit (S3 generic)
#'
#' Dispatches to [tidy_ild_model()] for \code{lmerMod} and \code{lme} objects from [ild_lme()],
#' or the \code{brmsfit} method for [ild_brms()]. All methods return tables conforming to
#' \code{\link{ild_tidy_schema}}.
#'
#' @param x A fitted model from [ild_lme()].
#' @param ... Passed to [tidy_ild_model()].
#' @export
ild_tidy <- function(x, ...) {
  UseMethod("ild_tidy")
}

#' @rdname ild_tidy
#' @method ild_tidy lmerMod
#' @export
ild_tidy.lmerMod <- function(x, ...) {
  tidy_ild_model(x, ...)
}

#' @rdname ild_tidy
#' @method ild_tidy lme
#' @export
ild_tidy.lme <- function(x, ...) {
  tidy_ild_model(x, ...)
}

#' @rdname ild_tidy
#' @method ild_tidy default
#' @export
ild_tidy.default <- function(x, ...) {
  stop(
    "No ild_tidy method for class ", paste(class(x), collapse = ", "),
    ". Use tidy_ild_model() for ild_lme() or ild_tidy() for ild_brms() fits.",
    call. = FALSE
  )
}

#' Augment observations with fitted values and residuals (S3 generic)
#'
#' Dispatches to [augment_ild_model()] for \code{lmerMod} and \code{lme} objects from [ild_lme()],
#' or the \code{brmsfit} method for [ild_brms()]. All methods return tables conforming to
#' \code{\link{ild_augment_schema}}.
#'
#' @param x A fitted model from [ild_lme()].
#' @param ... Passed to [augment_ild_model()].
#' @export
ild_augment <- function(x, ...) {
  UseMethod("ild_augment")
}

#' @rdname ild_augment
#' @method ild_augment lmerMod
#' @export
ild_augment.lmerMod <- function(x, ...) {
  augment_ild_model(x, ...)
}

#' @rdname ild_augment
#' @method ild_augment lme
#' @export
ild_augment.lme <- function(x, ...) {
  augment_ild_model(x, ...)
}

#' @rdname ild_augment
#' @method ild_augment default
#' @export
ild_augment.default <- function(x, ...) {
  stop(
    "No ild_augment method for class ", paste(class(x), collapse = ", "),
    ". Use augment_ild_model() for ild_lme() or ild_augment() for ild_brms() fits.",
    call. = FALSE
  )
}

#' Engine-agnostic diagnostics façade
#'
#' Returns a single \code{\link{ild_diagnostics_bundle}} for \code{lmerMod}, \code{lme},
#' \code{brmsfit}, \code{ild_fit_kfas}, and \code{ild_fit_ctsem}.
#' Sections are parallel across engines;
#' see \code{?ild_diagnostics_bundle}. Residual ACF/Q-Q for frequentist models are stored in
#' \code{residual$legacy_ild_diagnostics} for [plot_ild_diagnostics()]. For raw residual-only
#' computation without the bundle, call [ild_diagnostics()] directly.
#'
#' **Building blocks:** the same analyses are available as standalone functions
#' ([ild_design_check()], [ild_missing_pattern()], etc.); see [ild_diagnostics_utilities].
#'
#' @param object A fitted model.
#' @param data Optional ILD data (defaults to \code{attr(object, "ild_data")}).
#' @param type Residual diagnostic types for \code{lmerMod}/\code{lme} (passed to [ild_diagnostics()]).
#'   For \code{brmsfit}: \code{"all"} or any of \code{"convergence"}, \code{"sampler"}, \code{"ppc"}.
#' @param by_id For \code{lmerMod}/\code{lme}: ACF by person.
#' @param ppc_ndraws For \code{brmsfit}: posterior predictive draws.
#' @param missing_model If \code{TRUE}, fit \code{\link{ild_missing_model}} for the response (optional; can be slow). Requires at least one predictor.
#' @param missing_model_predictors Character vector for \code{ild_missing_model}; default \code{NULL} uses formula predictors.
#' @param causal_detail If \code{TRUE}, add extra quantile summaries for \code{.ipw} when present.
#' @param balance If \code{TRUE}, add weighted covariate balance (SMD) and ESS to \code{causal$balance}
#'   when \code{balance_covariates} and \code{balance_treatment} are set.
#' @param balance_treatment Character. Binary treatment column for balance (required if \code{balance = TRUE}).
#' @param balance_covariates Character vector of covariate columns for SMDs (required if \code{balance = TRUE}).
#' @param balance_weights_col Weights column for balance (default \code{".ipw_treat"}; use \code{".ipw"} for joint MSM).
#' @param balance_by_occasion If \code{TRUE}, balance and ESS are stratified by \code{.ild_seq}.
#' @param ... Reserved.
#' @return An \code{\link{ild_diagnostics_bundle}}.
#' @seealso \code{\link{ild_autoplot}}, \code{\link{ild_diagnostics_bundle}},
#'   \code{\link{guardrail_registry}}, \code{\link{ild_diagnostics_utilities}},
#'   \code{\link{ild_diagnostics}}
#' @export
ild_diagnose <- function(object, ...) {
  UseMethod("ild_diagnose")
}

#' Autoplot for ILD model fits or diagnostics (S3 generic)
#'
#' For \code{ild_diagnostics_bundle} from [ild_diagnose()], dispatches by \code{section}
#' and \code{type} (section-first; see \code{?ild_autoplot} method for
#' \code{ild_diagnostics_bundle}). For [ild_diagnostics()], calls [plot_ild_diagnostics()]
#' directly. For \code{lmerMod} or \code{lme} fits from [ild_lme()], calls [ild_plot()]
#' (default \code{type = "fitted_vs_actual"}).
#'
#' @param x A bundle, \code{ild_diagnostics} object, or a fitted model with \code{ild_data}.
#' @param type When \code{x} is \code{lmerMod} or \code{lme}: passed to [ild_plot()]
#'   (default \code{"fitted_vs_actual"}). When \code{x} is \code{brmsfit}: \code{"pp_check"}
#'   or \code{"fitted_vs_actual"}. When \code{x} is \code{ild_fit_ctsem}:
#'   \code{"fitted_vs_actual"}, \code{"residual_time"}, or \code{"qq"}.
#'   When \code{x} is \code{\link{ild_diagnostics_bundle}}:
#'   plot \code{type} within \code{section} — for \code{section = "residual"} one of
#'   \code{"acf"}, \code{"qq"}, \code{"fitted"}; for \code{section = "causal"} one of
#'   \code{"weights"} (histograms of IPW columns) or \code{"overlap"} (propensity densities;
#'   pass \code{treatment = "..."}). Other sections use a single type each
#'   (\code{"convergence"}, \code{"ppc"}, \code{"missingness"}, \code{"coverage"}).
#'   Use \code{type = NULL} with \code{section = "residual"} and
#'   legacy residual diagnostics to obtain the full named list from [plot_ild_diagnostics()];
#'   otherwise \code{NULL} picks a default per section.
#' @param section For \code{\link{ild_diagnostics_bundle}} only: \code{"residual"},
#'   \code{"fit"}, \code{"predictive"}, \code{"data"}, \code{"design"}, or \code{"causal"}.
#' @param ... Passed to [plot_ild_diagnostics()] (\code{ild_diagnostics} objects, or bundle
#'   \code{section = "residual"} with \code{type = NULL} and legacy diagnostics), to
#'   [ild_plot()] (\code{lmerMod}/\code{lme}), or to \code{brms::pp_check()} (bundle
#'   \code{section = "predictive"}, \code{type = "ppc"}).
#' @export
ild_autoplot <- function(x, ...) {
  UseMethod("ild_autoplot")
}

#' @rdname ild_autoplot
#' @method ild_autoplot ild_diagnostics
#' @export
ild_autoplot.ild_diagnostics <- function(x, ...) {
  plot_ild_diagnostics(x, ...)
}

#' @rdname ild_autoplot
#' @method ild_autoplot lmerMod
#' @export
ild_autoplot.lmerMod <- function(x, type = "fitted_vs_actual", ...) {
  ild_plot(x, type = type, ...)
}

#' @rdname ild_autoplot
#' @method ild_autoplot lme
#' @export
ild_autoplot.lme <- function(x, type = "fitted_vs_actual", ...) {
  ild_plot(x, type = type, ...)
}

#' @rdname ild_autoplot
#' @method ild_autoplot default
#' @export
ild_autoplot.default <- function(x, ...) {
  stop(
    "No ild_autoplot method for class ", paste(class(x), collapse = ", "),
    ". Use plot_ild_diagnostics() or ild_plot().",
    call. = FALSE
  )
}

test_that("ild_fit lme4 path matches ild_lme ar1=FALSE", {
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  f <- ild_fit(y ~ y_bp + y_wp + (1 | id), data = x, backend = "lme4", warn_no_ar1 = FALSE)
  g <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  expect_equal(lme4::fixef(f), lme4::fixef(g))
  prov <- attr(f, "ild_provenance", exact = TRUE)
  last_args <- prov$analysis_steps[[length(prov$analysis_steps)]]$args
  expect_equal(last_args$backend, "lme4")
})

test_that("ild_tidy and ild_augment dispatch to tidiers", {
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 2)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  tt <- ild_tidy(fit)
  expect_true(nrow(tt) >= 1L)
  expect_equal(tt, tidy_ild_model(fit))
  aug <- ild_augment(fit)
  expect_true(nrow(aug) == nrow(x))
  expect_equal(aug, augment_ild_model(fit))
})

test_that("ild_diagnose returns bundle with legacy ild_diagnostics in residual", {
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 3)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  a <- ild_diagnose(fit, data = x, type = "qq")
  b <- ild_diagnostics(fit, data = x, type = "qq")
  expect_s3_class(a, "ild_diagnostics_bundle")
  expect_equal(a$residual$legacy_ild_diagnostics$meta$type, b$meta$type)
})

test_that("ild_autoplot dispatches for diagnostics and lmerMod", {
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 4)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  diag <- ild_diagnostics(fit, data = x, type = "qq")
  p1 <- ild_autoplot(diag)
  expect_type(p1, "list")
  expect_true("qq" %in% names(p1))
  p2 <- ild_autoplot(fit, type = "fitted_vs_actual")
  expect_s3_class(p2, "ggplot")
})

test_that("ild_brms attaches ild_data, ild_posterior, ild_provenance", {
  skip_if_not_installed("brms")
  skip_on_cran()
  d <- ild_simulate(n_id = 6, n_obs_per = 5, seed = 99)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_brms(
    y ~ y_bp + y_wp + (1 | id),
    data = x,
    chains = 1, iter = 60, warmup = 30,
    refresh = 0, seed = 99, silent = 2,
    backend = "rstan",
    warn_uncentered = FALSE
  )
  expect_true(!is.null(attr(fit, "ild_data", exact = TRUE)))
  ps <- attr(fit, "ild_posterior", exact = TRUE)
  expect_equal(ps$engine, "brms")
  expect_true(!is.null(ps$prior_summary))
  prov <- attr(fit, "ild_provenance", exact = TRUE)
  last <- prov$analysis_steps[[length(prov$analysis_steps)]]
  expect_equal(last$step, "ild_brms")
})

test_that("ild_tidy and ild_augment work for ild_brms fits", {
  skip_if_not_installed("brms")
  skip_on_cran()
  d <- ild_simulate(n_id = 5, n_obs_per = 4, seed = 100)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_brms(
    y ~ y_bp + y_wp + (1 | id),
    data = x,
    chains = 1, iter = 50, warmup = 25,
    refresh = 0, seed = 100, silent = 2,
    backend = "rstan",
    warn_uncentered = FALSE
  )
  tt <- ild_tidy(fit)
  expect_true(all(ild_tidy_schema()$required %in% names(tt)))
  expect_true(all(c("rhat", "ess_bulk", "ess_tail") %in% names(tt)))
  aug <- ild_augment(fit)
  expect_true(all(ild_augment_schema()$required %in% names(aug)))
  expect_true(all(c(".fitted_lower", ".fitted_upper") %in% names(aug)))
  expect_equal(nrow(aug), nrow(x))
})

test_that("ild_diagnose.brmsfit returns ild_diagnostics_bundle", {
  skip_if_not_installed("brms")
  skip_on_cran()
  d <- ild_simulate(n_id = 5, n_obs_per = 4, seed = 101)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_brms(
    y ~ y_bp + y_wp + (1 | id),
    data = x,
    chains = 1, iter = 50, warmup = 25,
    refresh = 0, seed = 101, silent = 2,
    backend = "rstan",
    warn_uncentered = FALSE
  )
  diag <- ild_diagnose(fit, type = "convergence")
  expect_s3_class(diag, "ild_diagnostics_bundle")
  expect_true(!is.null(diag$fit$convergence$convergence_table))
})

test_that("ild_methods mentions priors and sampler for ild_brms", {
  skip_if_not_installed("brms")
  skip_on_cran()
  d <- ild_simulate(n_id = 4, n_obs_per = 4, seed = 102)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_brms(
    y ~ y_bp + y_wp + (1 | id),
    data = x,
    chains = 1, iter = 40, warmup = 20,
    refresh = 0, seed = 102, silent = 2,
    backend = "rstan",
    warn_uncentered = FALSE
  )
  txt <- ild_methods(fit)
  expect_true(grepl("ild_brms|Bayesian|prior|chains|warmup|adapt_delta", txt, ignore.case = TRUE))
})

test_that("ild_heterogeneity lmer random intercept", {
  d <- ild_simulate(n_id = 12, n_obs_per = 8, seed = 42)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_wp + y_bp + (1 | id), data = x)
  h <- ild_heterogeneity(fit)
  expect_s3_class(h, "ild_heterogeneity")
  expect_true(nrow(h$random_effects) >= 12L)
  expect_true(all(c("level_id", "term", "estimate_total") %in% names(h$random_effects)))
  tt <- ild_tidy(h)
  expect_equal(nrow(tt), nrow(h$random_effects))
  p <- ild_autoplot(h, type = "histogram", term = "(Intercept)")
  expect_s3_class(p, "ggplot")
})

test_that("ild_heterogeneity lmer random slope", {
  d <- ild_simulate(n_id = 15, n_obs_per = 10, seed = 43)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- suppressWarnings(
    ild_lme(y ~ y_wp + (y_wp | id), data = x,
            control = lme4::lmerControl(optimizer = "bobyqa", optCtrl = list(maxfun = 2e5)))
  )
  h <- ild_heterogeneity(fit, term = "y_wp")
  expect_true(any(h$summary$term == "y_wp"))
  expect_true(is.finite(h$summary$prop_gt_zero[h$summary$term == "y_wp"][1]))
  p <- ild_autoplot(h, type = "caterpillar", term = "y_wp")
  expect_s3_class(p, "ggplot")
})

test_that("ild_heterogeneity threshold scale raw", {
  d <- ild_simulate(n_id = 10, n_obs_per = 6, seed = 44)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_wp + (1 | id), data = x)
  h <- ild_heterogeneity(fit, threshold = -1e9, scale = "raw")
  expect_true(all(is.finite(h$summary$prop_gt_threshold) | is.na(h$summary$prop_gt_threshold)))
})

test_that("ild_diagnose attaches heterogeneity for lmer", {
  d <- ild_simulate(n_id = 8, n_obs_per = 5, seed = 45)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_wp + (1 | id), data = x)
  b <- ild_diagnose(fit)
  expect_true(!is.null(b$fit$heterogeneity))
  if (isTRUE(b$fit$heterogeneity$available)) {
    expect_s3_class(b$fit$heterogeneity$object, "ild_heterogeneity")
    p <- ild_autoplot(b, section = "fit", type = "heterogeneity", term = "(Intercept)")
    expect_s3_class(p, "ggplot")
  }
})

test_that("ild_heterogeneity_stratified returns rows or empty tibble", {
  d <- ild_simulate(n_id = 20, n_obs_per = 6, seed = 46)
  d$grp <- rep(c("A", "B"), length.out = nrow(d))
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  out <- ild_heterogeneity_stratified(
    y ~ y_wp + (1 | id),
    data = x,
    subgroup = "grp",
    min_n_id = 5L
  )
  expect_true("subgroup_level" %in% names(out) || nrow(out) == 0L)
})

# KFAS: contract depth, simulation recovery, missingness-as-documented, guardrails,
# and skipped placeholders for future state_spec DGPs.

# Helpers: helper-kfas-fixtures.R

test_that("KFAS contract depth: tidy/augment/diagnose structure", {
  skip_if_not_installed("KFAS")
  dgp <- kfas_dgp_local_level(
    n = 80L,
    Q = 1,
    H = 1,
    seed = KFAS_TEST_SEED_CONTRACT
  )
  fit <- suppressWarnings(
    ild_kfas(dgp$data, outcome = "y", state_spec = "local_level", time_units = "steps")
  )

  tt <- ild_tidy(fit)
  sch <- ild_tidy_schema()
  expect_true(all(sch$required %in% names(tt)))
  expect_true(all(tt$engine == "KFAS"))
  expect_true(all(tt$model_class == "ild_fit_kfas"))
  expect_true(all(tt$component %in% c("variance", "auxiliary")))
  expect_true(all(tt$interval_type == "none"))

  aug <- ild_augment(fit)
  a_sch <- ild_augment_schema()
  expect_true(all(a_sch$required %in% names(aug)))
  expect_equal(nrow(aug), nrow(attr(fit, "ild_data")))
  expect_true(is.numeric(aug$.fitted))
  expect_true(is.numeric(aug$.resid))
  expect_true(abs(stats::cor(aug[[".outcome"]], aug[[".fitted"]], use = "complete.obs")) > 0.2)

  b <- ild_diagnose(fit)
  expect_kfas_bundle_core(b)
  expect_true(is.character(b$summary_text) || is.null(b$summary_text))
})

test_that("KFAS recovery: local level Q/H and smoothed state track simulation", {
  skip_if_not_installed("KFAS")
  n <- 500L
  Q <- 1
  H <- 1
  dgp <- kfas_dgp_local_level(n, Q = Q, H = H, seed = KFAS_TEST_SEED_RECOVERY)
  fit <- suppressWarnings(
    ild_kfas(dgp$data, outcome = "y", state_spec = "local_level", time_units = "steps")
  )

  tt <- ild_tidy(fit)
  q_hat <- tt$estimate[tt$term == "level_variance"]
  h_hat <- tt$estimate[tt$term == "observation_variance"]
  expect_length(q_hat, 1L)
  expect_length(h_hat, 1L)
  expect_true(is.finite(q_hat) && is.finite(h_hat) && q_hat > 0 && h_hat > 0)
  expect_true(max(q_hat / Q, Q / q_hat) < 5)
  expect_true(max(h_hat / H, H / h_hat) < 5)

  ah <- fit$kfs$alphahat
  expect_true(is.matrix(ah))
  sm <- as.numeric(ah[, 1L])
  expect_equal(length(sm), nrow(attr(fit, "ild_data")))
  expect_true(stats::cor(dgp$true_level, sm) > 0.85)

  fh <- 5L
  fit_f <- suppressWarnings(
    ild_kfas(
      dgp$data,
      outcome = "y",
      state_spec = "local_level",
      time_units = "steps",
      forecast_horizon = fh
    )
  )
  # Forecast path is ggplot + optional KFAS ahead mean (API varies by KFAS version)
  expect_s3_class(ild_plot_forecast(fit_f), "ggplot")
})

test_that("KFAS missingness: NA outcomes omitted with warning; preprocessing counts", {
  skip_if_not_installed("KFAS")
  x <- kfas_fixture_segmented_na_outcome()
  expect_true(sum(is.na(x$y)) >= 1L)
  expect_warning(
    fit <- ild_kfas(x, outcome = "y", state_spec = "local_level", time_units = "steps"),
    "Removing rows with NA"
  )
  expect_equal(fit$preprocessing$n_rows_input, nrow(x))
  expect_equal(fit$preprocessing$n_outcome_na, sum(is.na(x$y)))
  expect_equal(fit$preprocessing$n_rows_used, sum(!is.na(x$y)))
  expect_true(fit$preprocessing$n_na_segments >= 3L)
})

test_that("KFAS guardrail: high irregularity for discrete-time indexing", {
  skip_if_not_installed("KFAS")
  x <- kfas_fixture_irregular_intervals(n = 55L)
  sc <- ild_spacing_class(x)
  expect_true(sc == "irregular-ish" || !is.na(sc))
  fit <- suppressWarnings(
    ild_kfas(x, outcome = "y", state_spec = "local_level", time_units = "steps", irregular_time = TRUE)
  )
  b <- ild_diagnose(fit)
  expect_true(any(b$guardrails$rule_id == "GR_KFAS_HIGH_IRREGULARITY_FOR_DISCRETE_TIME"))
})

test_that("KFAS guardrail: many missing outcome segments", {
  skip_if_not_installed("KFAS")
  x <- kfas_fixture_segmented_na_outcome()
  fit <- suppressWarnings(
    ild_kfas(x, outcome = "y", state_spec = "local_level", time_units = "steps")
  )
  b <- ild_diagnose(fit)
  expect_true(any(b$guardrails$rule_id == "GR_KFAS_MANY_MISSING_OUTCOME_SEGMENTS"))
})

# --- Future state_spec DGP recovery (unimplemented wrappers): placeholders ---

test_that("KFAS recovery DGP: local trend (skipped until state_spec implemented)", {
  skip_if_not_installed("KFAS")
  skip("local_trend not implemented in ild_kfas(); see inst/dev/KFAS_V1_BACKEND.md")
  n <- 300L
  Q <- c(0.1, 0.1)
  H <- 0.5
  expect_true(n > 0 && length(Q) == 2L && H > 0)
})

test_that("KFAS recovery DGP: AR(1) latent (skipped until state_spec implemented)", {
  skip_if_not_installed("KFAS")
  skip("ar1_state not implemented in ild_kfas(); see inst/dev/KFAS_V1_BACKEND.md")
  phi <- 0.7
  Q <- 0.2
  H <- 0.4
  expect_true(abs(phi) < 1 && Q > 0 && H > 0)
})

test_that("KFAS recovery DGP: regression + local level (skipped until state_spec implemented)", {
  skip_if_not_installed("KFAS")
  skip("regression_local_level not implemented in ild_kfas(); see inst/dev/KFAS_V1_BACKEND.md")
  n <- 250L
  beta <- 0.3
  expect_true(n > 5L && is.finite(beta))
})

test_that("KFAS guardrail: state dimension high for n (skipped until multi-state local_level or larger state_spec)", {
  skip_if_not_installed("KFAS")
  skip("GR_KFAS_STATE_DIMENSION_HIGH_FOR_N requires m_state >= 2; v1 local_level is scalar state")
  expect_true(TRUE)
})

test_that("ild_ctsem fits and supports tidy/augment/diagnose/autoplot contracts", {
  skip_if_not_installed("ctsem")
  x <- ctsem_fixture_data(seed = 9401L, n_id = 1L, n_obs_per = 50L)
  fit <- suppressWarnings(
    ild_ctsem(
      data = x,
      outcome = "y",
      model_type = "stanct",
      iter = 200,
      chains = 1
    )
  )
  expect_s3_class(fit, "ild_fit_ctsem")
  expect_true(!is.null(attr(fit, "ild_data", exact = TRUE)))
  expect_true(!is.null(attr(fit, "ild_provenance", exact = TRUE)))

  td <- ild_tidy(fit)
  expect_true(all(ild_tidy_schema()$required %in% names(td)))
  expect_true(all(td$engine == "ctsem"))

  ag <- ild_augment(fit)
  expect_true(all(ild_augment_schema()$required %in% names(ag)))
  expect_equal(nrow(ag), nrow(attr(fit, "ild_data", exact = TRUE)))

  b <- ild_diagnose(fit)
  expect_s3_class(b, "ild_diagnostics_bundle")
  expect_no_error(validate_ild_diagnostics_bundle(b))
  expect_equal(b$meta$engine, "ctsem")

  expect_s3_class(ild_autoplot(fit, type = "fitted_vs_actual"), "ggplot")
  expect_s3_class(ild_autoplot(fit, type = "residual_time"), "ggplot")
  expect_s3_class(ild_autoplot(fit, type = "qq"), "ggplot")

  expect_s3_class(ild_autoplot(b, section = "residual", type = "acf"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "residual", type = "qq"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "fit", type = "convergence"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "data", type = "missingness"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "design", type = "coverage"), "ggplot")

  m <- ild_methods(fit)
  expect_true(grepl("ild_ctsem", m, fixed = TRUE) || grepl("continuous-time", m, fixed = TRUE))
})

# Contract regression: semantics (sections, guardrails, tidy/augment, autoplot, methods text)

test_that("contract: regular spacing + decomposed predictors — dense bundle, tidy/augment, plots, no guardrail methods", {
  fx <- fixture_contract_regular_decomposed()
  expect_equal(fx$spacing_class, "regular-ish")
  contract_assert_bundle_dense_lmer(fx$bundle)
  contract_assert_tidy_augment_lmer(fx$fit)
  contract_assert_autoplot_frequentist_core(fx$bundle)
  expect_equal(length(fx$expect_rule_ids), 0L)
  expect_false(fx$expect_methods_guardrail)
  m <- ild_methods(fx$fit, bundle = fx$bundle)
  expect_false(grepl("Methodological cautions (tidyILD guardrails)", m, fixed = TRUE))
})

test_that("contract: irregular spacing + no AR1 — expected guardrails and methods text", {
  fx <- fixture_contract_irregular_no_ar1()
  expect_equal(fx$spacing_class, "irregular-ish")
  contract_assert_bundle_dense_lmer(fx$bundle)
  contract_assert_tidy_augment_lmer(fx$fit)
  contract_assert_autoplot_frequentist_core(fx$bundle)
  ids <- fx$bundle$guardrails$rule_id
  for (r in fx$expect_rule_ids) {
    expect_true(r %in% ids, info = r)
  }
  expect_true(fx$expect_methods_guardrail)
  m <- ild_methods(fx$fit, bundle = fx$bundle)
  expect_true(grepl("Methodological cautions (tidyILD guardrails)", m, fixed = TRUE))
  gs <- tidyILD:::ild_guardrails_summary(fx$bundle)
  expect_true(gs$n >= 1L)
  expect_true(nzchar(gs$narrative))
})

test_that("contract: singular guardrail row merged — rule id and methods narrative", {
  fx <- fixture_contract_singular_guardrail()
  contract_assert_bundle_dense_lmer(fx$bundle)
  contract_assert_tidy_augment_lmer(fx$fit)
  expect_true("GR_SINGULAR_RANDOM_EFFECTS" %in% fx$bundle$guardrails$rule_id)
  m <- ild_methods(fx$fit, bundle = fx$bundle)
  expect_true(grepl("GR_SINGULAR_RANDOM_EFFECTS", m, fixed = TRUE) ||
    grepl("guardrail", m, ignore.case = TRUE))
})

test_that("contract: unstable IPW — causal weights plot + guardrail", {
  fx <- fixture_contract_ipw_unstable()
  contract_assert_bundle_dense_lmer(fx$bundle)
  contract_assert_tidy_augment_lmer(fx$fit)
  expect_true("GR_IPW_WEIGHTS_UNSTABLE" %in% fx$bundle$guardrails$rule_id)
  expect_s3_class(ild_autoplot(fx$bundle, section = "causal", type = "weights"), "ggplot")
  m <- ild_methods(fx$fit, bundle = fx$bundle)
  expect_true(grepl("Methodological cautions (tidyILD guardrails)", m, fixed = TRUE))
})

test_that("contract: late dropout — merged guardrail + methods text", {
  fx <- fixture_contract_dropout_late()
  contract_assert_bundle_dense_lmer(fx$bundle)
  contract_assert_tidy_augment_lmer(fx$fit)
  expect_true("GR_DROPOUT_LATE_CONCENTRATION" %in% fx$bundle$guardrails$rule_id)
  m <- ild_methods(fx$fit, bundle = fx$bundle)
  expect_true(grepl("Methodological cautions (tidyILD guardrails)", m, fixed = TRUE))
})

test_that("contract: brms bundle + merged poor mixing / low ESS guardrails", {
  skip_if_not_installed("brms")
  skip_on_cran()
  fx <- fixture_contract_brms_poor_mixing_merged()
  expect_true(!is.null(fx$bundle$meta))
  expect_true(!is.null(fx$bundle$fit))
  expect_true(!is.null(attr(fx$bundle, "ild_fit", exact = TRUE)))
  td <- ild_tidy(fx$fit)
  expect_true(all(contract_tidy_required() %in% names(td)))
  for (r in fx$expect_rule_ids) {
    expect_true(r %in% fx$bundle$guardrails$rule_id, info = r)
  }
  expect_s3_class(ild_autoplot(fx$bundle, section = "fit", type = "convergence"), "ggplot")
  p <- ild_autoplot(fx$bundle, section = "predictive", type = "ppc", ndraws = 10)
  expect_s3_class(p, "ggplot")
  m <- ild_methods(fx$fit, bundle = fx$bundle)
  expect_true(grepl("Methodological cautions (tidyILD guardrails)", m, fixed = TRUE))
})

test_that("contract: ctsem backend satisfies tidy/augment/bundle contracts", {
  skip_if_not_installed("ctsem")
  x <- ctsem_fixture_data(seed = 9410L, n_id = 1L, n_obs_per = 45L)
  fit <- suppressWarnings(
    ild_ctsem(
      data = x,
      outcome = "y",
      model_type = "stanct",
      iter = 200,
      chains = 1
    )
  )
  td <- ild_tidy(fit)
  expect_true(all(contract_tidy_required() %in% names(td)))
  ag <- ild_augment(fit)
  expect_true(all(contract_augment_required() %in% names(ag)))
  b <- ild_diagnose(fit)
  expect_s3_class(b, "ild_diagnostics_bundle")
  expect_no_error(validate_ild_diagnostics_bundle(b))
  expect_s3_class(ild_autoplot(b, section = "fit", type = "convergence"), "ggplot")
  m <- ild_methods(fit, bundle = b)
  expect_true(grepl("ild_ctsem", m, fixed = TRUE) || grepl("continuous-time", m, fixed = TRUE))
})

test_that("ild_kfas local_level fits and matches tidy/augment/diagnose contracts", {
  skip_if_not_installed("KFAS")
  d <- ild_simulate(n_id = 1, n_obs_per = 40, seed = 9001)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- suppressWarnings(
    ild_kfas(x, outcome = "y", state_spec = "local_level", time_units = "sim_steps")
  )
  expect_s3_class(fit, "ild_fit_kfas")
  expect_true(!is.null(attr(fit, "ild_data")))
  expect_true(!is.null(attr(fit, "ild_provenance")))

  tt <- ild_tidy(fit)
  expect_true(all(ild_tidy_schema()$required %in% names(tt)))
  expect_true(all(tt$engine == "KFAS"))

  aug <- ild_augment(fit)
  expect_true(all(ild_augment_schema()$required %in% names(aug)))
  expect_true(nrow(aug) == nrow(attr(fit, "ild_data")))

  st <- ild_tidy_states(fit)
  expect_equal(nrow(st), nrow(attr(fit, "ild_data")))

  b <- ild_diagnose(fit)
  expect_s3_class(b, "ild_diagnostics_bundle")
  expect_no_error(validate_ild_diagnostics_bundle(b))
  expect_true(!is.null(b$residual$standardized_prediction_errors))
  expect_true(b$meta$engine == "KFAS")
  expect_true("log_likelihood" %in% names(b$fit))

  p <- suppressWarnings(ild_autoplot(fit, type = "states"))
  expect_s3_class(p, "ggplot")

  expect_s3_class(ild_autoplot(b, section = "residual", type = "acf"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "residual", type = "qq"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "residual", type = "fitted"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "fit", type = "convergence"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "predictive", type = "forecast"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "predictive", type = "errors"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "data", type = "missingness"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "design", type = "coverage"), "ggplot")

  expect_s3_class(ild_plot_states(fit), "ggplot")
  expect_s3_class(ild_plot_filtered_vs_smoothed(fit), "ggplot")
  expect_s3_class(ild_plot_forecast(fit), "ggplot")

  mt <- ild_methods(fit)
  expect_true(grepl("ild_kfas", mt, fixed = TRUE) || grepl("state-space", mt, fixed = TRUE))
})

test_that("KFAS-specific guardrails can trigger", {
  skip_if_not_installed("KFAS")
  d <- ild_simulate(n_id = 1, n_obs_per = 10, seed = 9003)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- suppressWarnings(
    ild_kfas(
      x,
      outcome = "y",
      state_spec = "local_level",
      time_units = "steps",
      fit_context = "independent_series_per_id"
    )
  )
  b <- ild_diagnose(fit)
  expect_true(any(b$guardrails$rule_id == "GR_KFAS_SHORT_SERIES_FOR_STATE_SPACE"))
  expect_true(any(b$guardrails$rule_id == "GR_KFAS_UNMODELED_BETWEEN_PERSON_HETEROGENEITY"))
})

test_that("ild_kfas rejects multiple subjects in v1", {
  skip_if_not_installed("KFAS")
  d <- ild_simulate(n_id = 3, n_obs_per = 10, seed = 9002)
  x <- ild_prepare(d, id = "id", time = "time")
  expect_error(
    ild_kfas(x, outcome = "y", time_units = "d"),
    "one subject"
  )
})

test_that("ild_diagnose attaches ild_fit and ild_data on bundle", {
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 201)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x)
  expect_identical(attr(b, "ild_fit", exact = TRUE), fit)
  expect_identical(attr(b, "ild_data", exact = TRUE), x)
})

test_that("ild_autoplot bundle: residual legacy list when type is NULL", {
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 202)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x)
  pl <- ild_autoplot(b)
  expect_type(pl, "list")
  expect_true(length(pl) >= 1L)
})

test_that("ild_autoplot bundle: section residual types return ggplot", {
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 203)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x)
  expect_s3_class(ild_autoplot(b, section = "residual", type = "acf"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "residual", type = "qq"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "residual", type = "fitted"), "ggplot")
})

test_that("ild_autoplot bundle: fit, data, design sections", {
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 204)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x)
  expect_s3_class(ild_autoplot(b, section = "fit", type = "convergence"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "data", type = "missingness"), "ggplot")
  expect_s3_class(ild_autoplot(b, section = "design", type = "coverage"), "ggplot")
})

test_that("ild_autoplot bundle: causal weights errors without .ipw", {
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 205)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x)
  expect_error(ild_autoplot(b, section = "causal", type = "weights"), "IPW")
})

test_that("ild_autoplot bundle: causal weights with .ipw", {
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 206)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  x$.ipw <- rep(1, nrow(x))
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x)
  expect_s3_class(ild_autoplot(b, section = "causal", type = "weights"), "ggplot")
})

test_that("ild_autoplot bundle: brms residual acf, convergence, ppc", {
  skip_if_not_installed("brms")
  skip_on_cran()
  d <- ild_simulate(n_id = 5, n_obs_per = 4, seed = 207)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_brms(
    y ~ y_bp + y_wp + (1 | id),
    data = x,
    chains = 1, iter = 50, warmup = 25,
    refresh = 0, seed = 207, silent = 2,
    backend = "rstan",
    warn_uncentered = FALSE
  )
  b <- ild_diagnose(fit, type = "all")
  expect_s3_class(ild_autoplot(b, section = "fit", type = "convergence"), "ggplot")
  p <- ild_autoplot(b, section = "predictive", type = "ppc", ndraws = 20)
  expect_s3_class(p, "ggplot")
})

test_that("ild_autoplot bundle: frequentist predictive ppc errors", {
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 208)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  b <- ild_diagnose(fit, data = x)
  expect_error(ild_autoplot(b, section = "predictive", type = "ppc"), "brmsfit")
})

# S3 methods for ild_ctsem backend

#' @rdname ild_tidy
#' @method ild_tidy ild_fit_ctsem
#' @export
ild_tidy.ild_fit_ctsem <- function(x, ...) {
  rlang::check_installed("ctsem", reason = "to use ild_tidy() for ild_ctsem fits")
  pars <- .ild_ctsem_extract_parameters(x$ct_fit)
  if (nrow(pars) == 0L) {
    return(
      ild_tidy_assemble(
        term = "(no_parameters_extracted)",
        estimate = NA_real_,
        std_error = NA_real_,
        conf_low = NA_real_,
        conf_high = NA_real_,
        p_value = NA_real_,
        statistic = NA_real_,
        interval_type = "Wald",
        engine = "ctsem",
        model_class = "ild_fit_ctsem",
        component = "auxiliary",
        effect_level = "auxiliary",
        optional = NULL
      )
    )
  }
  z <- stats::qnorm(0.975)
  pars$std_error[!is.finite(pars$std_error)] <- NA_real_
  pars$statistic <- ifelse(
    is.finite(pars$estimate) & is.finite(pars$std_error) & pars$std_error > 0,
    pars$estimate / pars$std_error,
    NA_real_
  )
  pars$p_value <- ifelse(
    is.finite(pars$statistic),
    2 * (1 - stats::pnorm(abs(pars$statistic))),
    NA_real_
  )
  pars$conf_low <- ifelse(
    is.finite(pars$estimate) & is.finite(pars$std_error),
    pars$estimate - z * pars$std_error,
    NA_real_
  )
  pars$conf_high <- ifelse(
    is.finite(pars$estimate) & is.finite(pars$std_error),
    pars$estimate + z * pars$std_error,
    NA_real_
  )
  component <- vapply(pars$term, .ild_ctsem_component_from_term, character(1L))
  effect_level <- ifelse(component == "fixed", "population", "auxiliary")
  ild_tidy_assemble(
    term = pars$term,
    estimate = pars$estimate,
    std_error = pars$std_error,
    conf_low = pars$conf_low,
    conf_high = pars$conf_high,
    p_value = pars$p_value,
    statistic = pars$statistic,
    interval_type = rep("Wald", nrow(pars)),
    engine = "ctsem",
    model_class = "ild_fit_ctsem",
    component = component,
    effect_level = effect_level,
    optional = NULL
  )
}

#' @rdname ild_augment
#' @method ild_augment ild_fit_ctsem
#' @export
ild_augment.ild_fit_ctsem <- function(x, ...) {
  rlang::check_installed("ctsem", reason = "to use ild_augment() for ild_ctsem fits")
  data <- attr(x, "ild_data", exact = TRUE)
  if (is.null(data)) {
    stop("ild_augment() for ild_fit_ctsem requires attr(fit, \"ild_data\").", call. = FALSE)
  }
  validate_ild(data)
  resp <- attr(x, "ild_response", exact = TRUE)
  if (is.null(resp) || !resp %in% names(data)) {
    stop("ild_response attribute missing or not present in data; refit with ild_ctsem().", call. = FALSE)
  }
  n <- nrow(data)
  y <- suppressWarnings(as.numeric(data[[resp]]))
  if (length(y) != n) {
    stop("Outcome length does not match ILD rows after coercion; check missing outcome values.", call. = FALSE)
  }
  fit <- .ild_ctsem_extract_fitted(x$ct_fit, n_expected = n)
  if (length(fit) != n || !is.numeric(fit)) {
    fit <- rep(NA_real_, n)
  }
  resid <- rep(NA_real_, n)
  ok <- is.finite(y) & is.finite(fit)
  resid[ok] <- (y - fit)[ok]
  resid_std <- .ild_ctsem_extract_resid_std(x$ct_fit, n_expected = n)

  ild_augment_assemble(
    .ild_id = data[[".ild_id"]],
    .ild_time = data[[".ild_time"]],
    .outcome = y,
    .fitted = fit,
    .resid = resid,
    .resid_std = resid_std,
    engine = "ctsem",
    model_class = "ild_fit_ctsem",
    optional = NULL
  )
}

#' @rdname ild_autoplot
#' @method ild_autoplot ild_fit_ctsem
#' @param type For \code{ild_fit_ctsem}: \code{"fitted_vs_actual"} (default),
#'   \code{"residual_time"}, or \code{"qq"}.
#' @export
ild_autoplot.ild_fit_ctsem <- function(x, type = c("fitted_vs_actual", "residual_time", "qq"), ...) {
  type <- match.arg(type)
  ag <- ild_augment(x)
  if (type == "fitted_vs_actual") {
    return(
      ggplot2::ggplot(ag, ggplot2::aes(x = .data$.outcome, y = .data$.fitted)) +
        ggplot2::geom_point(alpha = 0.5) +
        ggplot2::geom_abline(intercept = 0, slope = 1, linetype = 2) +
        ggplot2::labs(title = "ctsem fitted vs observed", x = "Observed", y = "Fitted") +
        ggplot2::theme_minimal()
    )
  }
  if (type == "residual_time") {
    return(
      ggplot2::ggplot(ag, ggplot2::aes(x = .data$.ild_time, y = .data$.resid)) +
        ggplot2::geom_point(alpha = 0.5) +
        ggplot2::geom_hline(yintercept = 0, linetype = 2) +
        ggplot2::labs(title = "ctsem residuals over time", x = ".ild_time", y = "Residual") +
        ggplot2::theme_minimal()
    )
  }
  ggplot2::ggplot(ag, ggplot2::aes(sample = .data$.resid_std)) +
    ggplot2::stat_qq() +
    ggplot2::stat_qq_line() +
    ggplot2::labs(title = "ctsem standardized residual QQ plot", x = "Theoretical", y = "Sample") +
    ggplot2::theme_minimal()
}

#' @keywords internal
#' @noRd
.ild_ctsem_component_from_term <- function(term) {
  u <- toupper(as.character(term)[1L])
  if (grepl("DRIFT|CINT|T0MEAN|LAMBDA|BETA", u)) return("fixed")
  if (grepl("DIFF|MANIFESTVAR|T0VAR|COV|VAR", u)) return("auxiliary")
  "auxiliary"
}

#' @keywords internal
#' @noRd
.ild_ctsem_extract_parameters <- function(fit) {
  # Prefer coef() for stable names, then attempt SE extraction from summary.
  est <- tryCatch(stats::coef(fit), error = function(e) NULL)
  if (is.null(est)) {
    est <- tryCatch(fit$coefficients, error = function(e) NULL)
  }
  est <- suppressWarnings(as.numeric(unlist(est)))
  nm <- names(unlist(tryCatch(stats::coef(fit), error = function(e) fit$coefficients)))
  if (is.null(nm) || length(nm) != length(est)) {
    nm <- paste0("par_", seq_along(est))
  }
  if (!length(est)) {
    return(tibble::tibble(term = character(), estimate = numeric(), std_error = numeric()))
  }
  se_map <- .ild_ctsem_extract_se_map(fit)
  se <- unname(se_map[nm])
  se[is.na(se)] <- NA_real_
  tibble::tibble(term = nm, estimate = est, std_error = as.numeric(se))
}

#' @keywords internal
#' @noRd
.ild_ctsem_extract_se_map <- function(fit) {
  out <- numeric()
  sm <- tryCatch(summary(fit), error = function(e) NULL)
  if (is.null(sm)) return(out)
  mats <- list()
  if (is.matrix(sm) || is.data.frame(sm)) mats <- c(mats, list(sm))
  if (is.list(sm)) {
    for (nm in names(sm)) {
      obj <- sm[[nm]]
      if (is.matrix(obj) || is.data.frame(obj)) mats <- c(mats, list(obj))
    }
  }
  for (m in mats) {
    mm <- as.data.frame(m)
    cn <- tolower(names(mm))
    i_est <- match(TRUE, cn %in% c("estimate", "est"), nomatch = 0L)
    i_se <- match(TRUE, cn %in% c("std.error", "std.error.", "se", "std.error.est"), nomatch = 0L)
    if (i_est > 0L && i_se > 0L) {
      rn <- rownames(mm)
      if (!is.null(rn) && length(rn) == nrow(mm)) {
        tmp <- suppressWarnings(as.numeric(mm[[i_se]]))
        names(tmp) <- rn
        out <- tmp
        break
      }
    }
  }
  out
}

#' @keywords internal
#' @noRd
.ild_ctsem_extract_fitted <- function(fit, n_expected) {
  out <- rep(NA_real_, n_expected)
  f1 <- tryCatch(stats::fitted(fit), error = function(e) NULL)
  if (is.matrix(f1) || is.data.frame(f1)) {
    v <- tryCatch(suppressWarnings(as.numeric(f1[, 1L])), error = function(e) NULL)
    if (!is.null(v) && length(v) == n_expected) out <- v
  } else if (is.numeric(f1) && length(f1) == n_expected) {
    out <- suppressWarnings(as.numeric(f1))
  }
  if (all(!is.finite(out))) {
    # ctPredict is not always exported from ctsem; avoid ctsem::ctPredict (R CMD check).
    # Do not use return() inside tryCatch expr: it exits .ild_ctsem_extract_fitted() with NULL
    # instead of the intended length-n_expected vector (breaks ild_augment / tibble).
    pred <- tryCatch(
      {
        pred_raw <- NULL
        if (requireNamespace("ctsem", quietly = TRUE)) {
          fn <- get0("ctPredict", envir = asNamespace("ctsem"), inherits = FALSE, ifnotfound = NULL)
          if (is.function(fn)) {
            pred_raw <- fn(fit)
          }
        }
        pred_raw
      },
      error = function(e) NULL
    )
    if (!is.null(pred)) {
      # ctPredict shape varies by ctsem version; coercion can error (e.g. list-column y).
      v <- tryCatch(
        {
          vv <- NULL
          if (is.data.frame(pred) && "y" %in% names(pred)) {
            vv <- suppressWarnings(as.numeric(unlist(pred$y, recursive = FALSE, use.names = FALSE)))
          } else if (is.list(pred) && !is.null(pred$yhat)) {
            vv <- suppressWarnings(as.numeric(unlist(pred$yhat, recursive = FALSE, use.names = FALSE)))
          }
          if (!is.null(vv) && length(vv) == n_expected) vv else NULL
        },
        error = function(e) NULL
      )
      if (!is.null(v)) out <- v
    }
  }
  out
}

#' @keywords internal
#' @noRd
.ild_ctsem_extract_resid_std <- function(fit, n_expected) {
  out <- rep(NA_real_, n_expected)
  r <- tryCatch(stats::residuals(fit, type = "pearson"), error = function(e) NULL)
  if (is.matrix(r) || is.data.frame(r)) r <- r[, 1L]
  r <- suppressWarnings(as.numeric(r))
  if (is.numeric(r) && length(r) == n_expected) return(r)
  out
}

#' Fit a time-varying effects model (TVEM) for ILD
#'
#' Fits a GAM with a smooth in time and a time-varying coefficient for a
#' predictor using [mgcv::gam()]. Use [ild_tvem_plot()] to plot the
#' time-varying effect. Requires \code{.ild_time_num} (or a numeric time column).
#'
#' @param data An ILD object (see [is_ild()]).
#' @param outcome Character. Name of the outcome variable.
#' @param predictor Character. Name of the predictor with a time-varying effect.
#' @param time_var Character. Name of the time variable (default \code{".ild_time_num"}).
#' @param k Integer. Basis dimension for smooth terms (default 10).
#' @param re_id Logical. If \code{TRUE} (default), include a random intercept by person (\code{s(.ild_id, bs="re")}).
#' @param ... Passed to [mgcv::gam()].
#' @return A fitted \code{gam} object with class \code{c("tidyild_tvem", "gam", ...)}
#'   and attribute \code{ild_tvem_meta} (list with \code{outcome}, \code{predictor},
#'   \code{time_var}, \code{k}, \code{re_id}).
#' @export
#' @examples
#' set.seed(1)
#' d <- ild_simulate(n_id = 10, n_obs_per = 15, seed = 1)
#' d$x <- rnorm(nrow(d))
#' x <- ild_prepare(d, id = "id", time = "time")
#' tv <- ild_tvem(x, "y", "x", k = 5, re_id = TRUE)
#' ild_tvem_plot(tv)
ild_tvem <- function(data, outcome, predictor, time_var = ".ild_time_num",
                     k = 10L, re_id = TRUE, ...) {
  validate_ild(data)
  nms <- names(data)
  if (!outcome %in% nms) stop("outcome '", outcome, "' not found in data.", call. = FALSE)
  if (!predictor %in% nms) stop("predictor '", predictor, "' not found in data.", call. = FALSE)
  if (!time_var %in% nms) stop("time_var '", time_var, "' not found in data.", call. = FALSE)
  if (!is.numeric(data[[time_var]])) stop("time_var must be numeric.", call. = FALSE)
  if (!is.numeric(data[[predictor]])) stop("predictor must be numeric.", call. = FALSE)
  k <- as.integer(k)[1L]
  if (k < 3L) k <- 3L
  dat <- as.data.frame(data)
  if (re_id && ".ild_id" %in% nms) {
    dat$.ild_id <- factor(dat[[".ild_id"]])
    form <- stats::as.formula(paste0(
      outcome, " ~ s(", time_var, ", k = ", k, ") + s(", time_var,
      ", by = ", predictor, ", k = ", k, ") + s(.ild_id, bs = 're')"
    ))
  } else {
    form <- stats::as.formula(paste0(
      outcome, " ~ s(", time_var, ", k = ", k, ") + s(", time_var,
      ", by = ", predictor, ", k = ", k, ")"
    ))
  }
  fit <- mgcv::gam(form, data = dat, ...)
  attr(fit, "ild_tvem_meta") <- list(
    outcome = outcome,
    predictor = predictor,
    time_var = time_var,
    k = k,
    re_id = re_id
  )
  attr(fit, "ild_provenance") <- ild_new_analysis_provenance(data, "ild_tvem", list(
    outcome = outcome,
    predictor = predictor,
    time_var = time_var,
    k = k,
    re_id = re_id
  ), list(n_obs = nrow(data)))
  class(fit) <- c("tidyild_tvem", class(fit))
  fit
}

#' Test whether missingness is associated with a predictor (informative missingness)
#'
#' Fits a logistic model of missingness (binary: is the outcome NA?) on a
#' predictor variable. Use as a diagnostic: if the predictor is significant,
#' missingness may be informative and results could be biased. This function
#' does not correct for missingness; it flags the assumption for sensitivity analyses.
#'
#' @param x An ILD object (see [is_ild()]).
#' @param outcome_var Character. Name of the variable with missingness (e.g. \code{"mood"}).
#' @param predictor_var Character. Name of the suspected predictor of missingness (e.g. \code{"stress"}).
#' @param random Logical. If \code{TRUE}, fit a mixed-effects logistic model
#'   \code{is_missing ~ predictor + (1 | id)} via \code{lme4::glmer}; if \code{FALSE}
#'   (default), fit \code{glm(is_missing ~ predictor, family = binomial)}.
#' @return A list with \code{predictor} (name), \code{estimate}, \code{std_error},
#'   \code{p_value}, and \code{message} (short note about informative missingness).
#' @export
#' @examples
#' set.seed(1)
#' d <- ild_simulate(n_id = 20, n_obs_per = 10, seed = 1)
#' d$stress <- rnorm(nrow(d))
#' d$mood <- d$y
#' d$mood[sample(nrow(d), 30)] <- NA  # some missing
#' x <- ild_prepare(d, id = "id", time = "time")
#' ild_missing_bias(x, "mood", "stress")
ild_missing_bias <- function(x, outcome_var, predictor_var, random = FALSE) {
  validate_ild(x)
  nms <- names(x)
  if (!outcome_var %in% nms) stop("outcome_var '", outcome_var, "' not found in data.", call. = FALSE)
  if (!predictor_var %in% nms) stop("predictor_var '", predictor_var, "' not found in data.", call. = FALSE)
  is_missing <- as.integer(is.na(x[[outcome_var]]))
  pred_vals <- x[[predictor_var]]
  if (!is.numeric(pred_vals)) stop("predictor_var must be numeric.", call. = FALSE)
  n_miss <- sum(is_missing)
  n_obs <- sum(!is_missing)
  if (n_miss == 0) {
    return(list(
      predictor = predictor_var,
      estimate = NA_real_,
      std_error = NA_real_,
      p_value = NA_real_,
      message = "No missingness in outcome; nothing to model."
    ))
  }
  if (n_obs == 0) {
    return(list(
      predictor = predictor_var,
      estimate = NA_real_,
      std_error = NA_real_,
      p_value = NA_real_,
      message = "All outcome values are missing; cannot fit model."
    ))
  }
  id_col <- ".ild_id"
  df <- data.frame(
    is_missing = is_missing,
    pred = pred_vals,
    id = x[[id_col]]
  )
  if (random && requireNamespace("lme4", quietly = TRUE)) {
    fit <- tryCatch(
      lme4::glmer(is_missing ~ pred + (1 | id), data = df, family = stats::binomial),
      error = function(e) NULL
    )
    if (is.null(fit)) {
      fit_glm <- stats::glm(is_missing ~ pred, data = df, family = stats::binomial)
      co <- summary(fit_glm)$coefficients
      est <- co["pred", "Estimate"]
      se <- co["pred", "Std. Error"]
      pval <- co["pred", "Pr(>|z|)"]
    } else {
      co <- summary(fit)$coefficients
      est <- co["pred", "Estimate"]
      se <- co["pred", "Std. Error"]
      pval <- co["pred", "Pr(>|z|)"]
    }
  } else {
    fit_glm <- stats::glm(is_missing ~ pred, data = df, family = stats::binomial)
    co <- summary(fit_glm)$coefficients
    est <- co["pred", "Estimate"]
    se <- co["pred", "Std. Error"]
    pval <- co["pred", "Pr(>|z|)"]
  }
  msg <- if (!is.na(pval) && pval < 0.05) {
    paste0("Missingness may be associated with ", predictor_var,
           "; consider sensitivity analyses (e.g. multiple imputation or selection models).")
  } else {
    paste0("No strong evidence that missingness is associated with ", predictor_var, " (p = ", round(pval, 4), ").")
  }
  list(
    predictor = predictor_var,
    estimate = est,
    std_error = se,
    p_value = pval,
    message = msg
  )
}

#' Summarize missingness pattern in ILD
#'
#' Returns a tabular summary of missingness by person and/or by variable,
#' plus an optional heatmap plot. Complements [ild_summary()] and supports
#' checking data before modeling. When \code{vars = NULL}, all non-internal
#' data columns are used (observation presence across variables).
#'
#' Also a **section provider** for [ild_diagnose()] (see [ild_diagnostics_utilities]).
#'
#' @section Bundle integration:
#' [ild_diagnose()] passes summaries into \code{data$missing_pattern} (global) and
#' \code{missingness} (model variables) on the \code{\link{ild_diagnostics_bundle}}.
#'
#' @family ild_diagnostics_utilities
#' @param x An ILD object (see [is_ild()]).
#' @param vars Optional character vector of variable names to summarize.
#'   If \code{NULL}, all non-.ild_* data columns are used.
#' @param max_ids Optional integer. If set, subset to this many persons (sampled)
#'   before computing \code{by_id}, \code{summary}, and \code{plot} to handle large N.
#' @param seed Optional integer. Seed for sampling when \code{max_ids} is set.
#' @param outcome Optional character; if set, \code{by_id} is left-joined with
#'   [ild_missing_compliance()] for that column (interpretable adherence metrics).
#' @param expected_occasions Passed to [ild_missing_compliance()] when \code{outcome} is set.
#' @return A list with: \code{summary} (tibble: one row per var, columns var, n_obs, n_na, pct_na),
#'   \code{plot} (ggplot2 object for missingness heatmap), \code{by_id}, \code{overall},
#'   \code{n_complete}, \code{vars}. When \code{outcome} is set, \code{by_id} includes
#'   compliance columns (see [ild_missing_compliance()]).
#' @seealso [ild_diagnose()], [ild_diagnostics_bundle()]
#' @importFrom dplyr across all_of group_by left_join summarise
#' @importFrom rlang .data
#' @importFrom tibble tibble
#' @export
ild_missing_pattern <- function(x, vars = NULL, max_ids = NULL, seed = NULL,
                                outcome = NULL, expected_occasions = NULL) {
  validate_ild(x)
  ild_cols <- c(".ild_id", ".ild_time", ".ild_time_num", ".ild_seq", ".ild_dt", ".ild_gap")
  all_nms <- names(x)
  data_cols <- setdiff(all_nms, ild_cols)
  if (length(data_cols) == 0) stop("No data columns to summarize for missingness.", call. = FALSE)
  if (!is.null(vars)) {
    vars <- intersect(vars, all_nms)
    if (length(vars) == 0) stop("No selected variables found in data.", call. = FALSE)
  } else {
    vars <- data_cols
  }
  x_work <- x
  if (!is.null(max_ids) && is.finite(max_ids) && max_ids > 0) {
    if (!is.null(seed)) set.seed(seed)
    ids <- unique(x[[".ild_id"]])
    if (length(ids) > max_ids) {
      ids_use <- sample(ids, size = min(max_ids, length(ids)))
      x_work <- x[x[[".ild_id"]] %in% ids_use, ]
    }
  }
  id_col <- ".ild_id"
  time_var <- ".ild_seq"
  by_id <- dplyr::group_by(x_work, .data[[id_col]])
  by_id <- dplyr::summarise(by_id, dplyr::across(
    dplyr::all_of(vars),
    list(n_obs = ~ sum(!is.na(.x)), n_na = ~ sum(is.na(.x))),
    .names = "{.col}_{.fn}"
  ), .groups = "drop")
  overall <- list()
  summary_rows <- list()
  for (v in vars) {
    n_na <- sum(is.na(x_work[[v]]))
    n_obs <- nrow(x_work)
    overall[[v]] <- list(n_na = n_na, n_obs = n_obs, pct_na = 100 * n_na / n_obs)
    summary_rows[[v]] <- tibble::tibble(var = v, n_obs = n_obs, n_na = n_na, pct_na = 100 * n_na / n_obs)
  }
  summary_tbl <- do.call(rbind, summary_rows)
  if (length(summary_rows) > 0) summary_tbl <- tibble::as_tibble(summary_tbl)
  n_complete <- sum(Reduce(`&`, lapply(vars, function(v) !is.na(x_work[[v]]))))
  parts <- lapply(vars, function(v) {
    data.frame(
      id = x_work[[id_col]],
      time = x_work[[time_var]],
      variable = v,
      missing = is.na(x_work[[v]]),
      stringsAsFactors = FALSE
    )
  })
  long <- do.call(rbind, parts)
  p <- ggplot2::ggplot(long, ggplot2::aes(x = .data$time, y = factor(.data$id), fill = .data$missing)) +
    ggplot2::geom_tile() +
    ggplot2::scale_fill_manual(
      values = c("FALSE" = "gray90", "TRUE" = "darkred"),
      labels = c("observed", "missing")
    ) +
    ggplot2::labs(x = "Time (sequence)", y = "Person", fill = NULL, title = "Missingness") +
    ggplot2::theme_minimal()
  if (length(vars) > 1) p <- p + ggplot2::facet_wrap(ggplot2::vars(.data$variable), ncol = 1)
  if (!is.null(outcome)) {
    if (!outcome %in% names(x_work)) {
      stop("outcome '", outcome, "' not found in data.", call. = FALSE)
    }
    comp <- ild_missing_compliance(x_work, outcome, expected_occasions = expected_occasions)
    by_id <- dplyr::left_join(by_id, comp, by = ".ild_id")
  }
  list(
    summary = summary_tbl,
    plot = p,
    by_id = by_id,
    overall = overall,
    n_complete = n_complete,
    vars = vars
  )
}

# MSM / IPW covariate balance, effective sample size, and propensity overlap plots

#' @title MSM and IPW balance diagnostics
#' @name ild_msm_balance_topic
#' @description
#' After building weights ([ild_iptw_weights()], [ild_iptw_msm_weights()], [ild_joint_msm_weights()]),
#' assess **covariate balance** (weighted SMD), **effective sample size** (ESS), and **overlap**
#' of fitted propensities. These complement weight summaries in [fill_diagnostics_causal()] and
#' guardrails on weight ranges.
#'
#' **Weights:** Use \code{.ipw_treat} to check balance for the **treatment** mechanism;
#' use \code{.ipw} for balance in the **joint** MSM pseudo-population used in the weighted outcome
#' model—they need not coincide.
#'
#' @seealso [ild_msm_balance()], [ild_ipw_ess()], [ild_msm_overlap_plot()],
#'   [ild_iptw_weights()], [ild_iptw_msm_weights()], [ild_joint_msm_weights()],
#'   [ild_msm_estimand()], [ild_msm_fit()], [ild_msm_recovery()]
#' @keywords internal
NULL

#' Effective sample size from weights
#'
#' For nonnegative weights \eqn{w}, \eqn{\mathrm{ESS} = (\sum w)^2 / \sum w^2}.
#'
#' @param data An ILD object.
#' @param weights_col Name of the nonnegative weight column (e.g. \code{".ipw"} or \code{".ipw_treat"}).
#' @param by_occasion If \code{TRUE}, one ESS per level of \code{time_var}.
#' @param time_var Column for occasion index (default \code{".ild_seq"}).
#' @return A single numeric (pooled) or a \code{tibble} with \code{occasion} and \code{ess} when
#'   \code{by_occasion = TRUE}.
#' @export
#' @examples
#' set.seed(1)
#' d <- ild_simulate(n_id = 10, n_obs_per = 5, seed = 1)
#' x <- ild_prepare(d, id = "id", time = "time")
#' x$.ipw <- runif(nrow(x), 0.5, 1.5)
#' ild_ipw_ess(x, ".ipw")
#' ild_ipw_ess(x, ".ipw", by_occasion = TRUE)
ild_ipw_ess <- function(data, weights_col = ".ipw", by_occasion = FALSE,
                        time_var = ".ild_seq") {
  validate_ild(data)
  if (!weights_col %in% names(data)) {
    stop("weights_col '", weights_col, "' not found in data.", call. = FALSE)
  }
  w <- as.numeric(data[[weights_col]])
  w[!is.finite(w) | w < 0] <- NA_real_
  if (!by_occasion) {
    return(.ild_ess_from_w(w))
  }
  if (!time_var %in% names(data)) {
    stop("time_var '", time_var, "' not found in data.", call. = FALSE)
  }
  tv <- data[[time_var]]
  u <- sort(unique(as.numeric(tv)))
  rows <- vector("list", length(u))
  for (i in seq_along(u)) {
    t <- u[i]
    idx <- as.numeric(tv) == t
    rows[[i]] <- tibble::tibble(
      occasion = t,
      ess = .ild_ess_from_w(w[idx])
    )
  }
  dplyr::bind_rows(rows)
}

#' @keywords internal
#' @noRd
.ild_ess_from_w <- function(w) {
  w <- w[is.finite(w) & w >= 0]
  if (length(w) == 0L) return(NA_real_)
  sw <- sum(w)
  sw2 <- sum(w^2)
  if (sw2 <= 0) return(NA_real_)
  (sw^2) / sw2
}

#' Standardized mean difference (binary treatment) with IPW-style weights
#'
#' Pooled SD: \eqn{\sqrt{(v_1 + v_0)/2}} where \eqn{v_g} is the weighted variance of \eqn{X}
#' in the weighted treated (\eqn{A=1}) and control (\eqn{A=0}) pseudo-samples.
#'
#' @keywords internal
#' @noRd
.ild_smd_weighted <- function(x, A, w) {
  A <- as.numeric(A)
  if (is.factor(A)) A <- as.integer(A) - 1L
  ok <- is.finite(x) & is.finite(A) & is.finite(w) & w > 0 & A %in% c(0, 1)
  if (!any(ok)) return(NA_real_)
  x <- x[ok]
  A <- A[ok]
  w <- w[ok]
  wt1 <- w * A
  wt0 <- w * (1 - A)
  sw1 <- sum(wt1)
  sw0 <- sum(wt0)
  if (sw1 <= 0 || sw0 <= 0) return(NA_real_)
  m1 <- sum(wt1 * x) / sw1
  m0 <- sum(wt0 * x) / sw0
  v1 <- sum(wt1 * (x - m1)^2) / sw1
  v0 <- sum(wt0 * (x - m0)^2) / sw0
  sp <- sqrt((v1 + v0) / 2)
  if (!is.finite(sp) || sp <= 0) return(NA_real_)
  (m1 - m0) / sp
}

#' Covariate balance (weighted SMD) for MSM / IPW
#'
#' Computes **standardized mean differences** between weighted treated and control groups for
#' each named covariate. When \code{by_occasion = TRUE}, balance is computed **within** each
#' occasion (stratum); when \code{FALSE}, all rows are pooled into one pseudo-population.
#'
#' @param data An ILD object with \code{treatment}, \code{covariates}, and \code{weights_col}.
#' @param treatment Character. Binary treatment column (\code{0}/\code{1} or two-level factor).
#' @param covariates Character vector of numeric or binary covariate columns.
#' @param weights_col Analysis weights (e.g. \code{".ipw_treat"} or \code{".ipw"}).
#' @param by_occasion If \code{TRUE}, stratify by \code{time_var}.
#' @param time_var Occasion index column (default \code{".ild_seq"}).
#' @return A \code{tibble} with columns \code{stratum}, \code{covariate}, \code{smd},
#'   \code{mean_treated}, \code{mean_control}, \code{ess_stratum}, \code{weight_col}.
#' @export
#' @examples
#' set.seed(2)
#' d <- ild_simulate(n_id = 12, n_obs_per = 6, seed = 2)
#' d$stress <- rnorm(nrow(d))
#' d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
#' x <- ild_prepare(d, id = "id", time = "time")
#' x$.ipw_treat <- runif(nrow(x), 0.8, 1.2)
#' ild_msm_balance(x, treatment = "trt", covariates = "stress", weights_col = ".ipw_treat")
ild_msm_balance <- function(data,
                            treatment,
                            covariates,
                            weights_col = ".ipw_treat",
                            by_occasion = FALSE,
                            time_var = ".ild_seq") {
  validate_ild(data)
  if (!treatment %in% names(data)) stop("treatment column not found.", call. = FALSE)
  if (!weights_col %in% names(data)) stop("weights_col not found.", call. = FALSE)
  for (cv in covariates) {
    if (!cv %in% names(data)) stop("covariate '", cv, "' not found.", call. = FALSE)
  }
  Araw <- data[[treatment]]
  if (is.factor(Araw)) {
    if (nlevels(Araw) != 2L) stop("treatment must be binary.", call. = FALSE)
    A <- as.integer(Araw) - 1L
  } else {
    A <- as.numeric(Araw)
  }
  w <- as.numeric(data[[weights_col]])
  if (!by_occasion) {
    out <- vector("list", length(covariates))
    ess <- .ild_ess_from_w(w)
    for (i in seq_along(covariates)) {
      cv <- covariates[i]
      x <- as.numeric(data[[cv]])
      smd <- .ild_smd_weighted(x, A, w)
      wt1 <- w * A
      wt0 <- w * (1 - A)
      sw1 <- sum(wt1[is.finite(wt1)], na.rm = TRUE)
      sw0 <- sum(wt0[is.finite(wt0)], na.rm = TRUE)
      m1 <- if (sw1 > 0) sum(wt1 * x, na.rm = TRUE) / sw1 else NA_real_
      m0 <- if (sw0 > 0) sum(wt0 * x, na.rm = TRUE) / sw0 else NA_real_
      out[[i]] <- tibble::tibble(
        stratum = "pooled",
        covariate = cv,
        smd = smd,
        mean_treated = m1,
        mean_control = m0,
        ess_stratum = ess,
        weight_col = weights_col
      )
    }
    return(dplyr::bind_rows(out))
  }
  if (!time_var %in% names(data)) stop("time_var not found.", call. = FALSE)
  tv <- data[[time_var]]
  u <- sort(unique(as.numeric(tv)))
  blocks <- vector("list", 0L)
  for (t in u) {
    idx <- which(as.numeric(tv) == t)
    if (length(idx) < 2L) next
    At <- A[idx]
    wt <- w[idx]
    ess_t <- .ild_ess_from_w(wt)
    for (cv in covariates) {
      x <- as.numeric(data[[cv]][idx])
      smd <- .ild_smd_weighted(x, At, wt)
      wt1 <- wt * At
      wt0 <- wt * (1 - At)
      sw1 <- sum(wt1[is.finite(wt1)], na.rm = TRUE)
      sw0 <- sum(wt0[is.finite(wt0)], na.rm = TRUE)
      m1 <- if (sw1 > 0) sum(wt1 * x, na.rm = TRUE) / sw1 else NA_real_
      m0 <- if (sw0 > 0) sum(wt0 * x, na.rm = TRUE) / sw0 else NA_real_
      blocks[[length(blocks) + 1L]] <- tibble::tibble(
        stratum = as.character(t),
        covariate = cv,
        smd = smd,
        mean_treated = m1,
        mean_control = m0,
        ess_stratum = ess_t,
        weight_col = weights_col
      )
    }
  }
  if (length(blocks) == 0L) {
    return(tibble::tibble(
      stratum = character(),
      covariate = character(),
      smd = numeric(),
      mean_treated = numeric(),
      mean_control = numeric(),
      ess_stratum = numeric(),
      weight_col = character()
    ))
  }
  dplyr::bind_rows(blocks)
}

#' Propensity overlap plot (pooled or sequential MSM IPTW)
#'
#' For **pooled** [ild_iptw_weights()], uses \code{attr(data, "ild_iptw_fit")} and plots the
#' distribution of fitted \eqn{P(A=1\mid L)} by treatment level.
#'
#' For **sequential** [ild_iptw_msm_weights()], uses \code{attr(data, "ild_iptw_msm_fits")} and
#' facets by occasion: fitted denominator propensity at each \eqn{t}.
#'
#' @param data ILD with treatment column and relevant \code{glm} attributes.
#' @param treatment Binary treatment column name.
#' @param source Character: \code{"auto"} (prefer MSM fits if present), \code{"pooled"}, or
#'   \code{"msm"}.
#' @return A \code{ggplot} object.
#' @export
#' @examples
#' if (requireNamespace("ggplot2", quietly = TRUE)) {
#'   set.seed(3)
#'   d <- ild_simulate(n_id = 15, n_obs_per = 5, seed = 3)
#'   d$stress <- rnorm(nrow(d))
#'   d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
#'   x <- ild_prepare(d, id = "id", time = "time")
#'   x <- ild_iptw_weights(x, treatment = "trt", predictors = "stress")
#'   ild_msm_overlap_plot(x, treatment = "trt", source = "pooled")
#' }
ild_msm_overlap_plot <- function(data, treatment, source = c("auto", "pooled", "msm")) {
  validate_ild(data)
  source <- match.arg(source)
  if (!treatment %in% names(data)) stop("treatment not found.", call. = FALSE)
  Araw <- data[[treatment]]
  if (is.factor(Araw)) {
    if (nlevels(Araw) != 2L) stop("treatment must be binary.", call. = FALSE)
    A <- as.integer(Araw) - 1L
  } else {
    A <- as.numeric(Araw)
  }
  has_msm <- !is.null(attr(data, "ild_iptw_msm_fits", exact = TRUE))
  has_pool <- !is.null(attr(data, "ild_iptw_fit", exact = TRUE))
  if (source == "auto") {
    source <- if (has_msm) "msm" else "pooled"
  }
  if (source == "msm") {
    if (!has_msm) {
      stop("Sequential IPTW fits not found. Run ild_iptw_msm_weights() or use source = \"pooled\".",
           call. = FALSE)
    }
    return(.ild_overlap_plot_msm(data, treatment, A))
  }
  if (!has_pool) {
    stop("Pooled IPTW fit not found. Run ild_iptw_weights() or use source = \"msm\" with MSM weights.",
         call. = FALSE)
  }
  fit <- attr(data, "ild_iptw_fit", exact = TRUE)
  p_hat <- stats::predict(fit, newdata = as.data.frame(data), type = "response")
  df <- tibble::tibble(
    p_hat = p_hat,
    treatment = factor(A, levels = c(0, 1), labels = c("A=0", "A=1"))
  )
  ggplot2::ggplot(df, ggplot2::aes(x = .data$p_hat, color = .data$treatment)) +
    ggplot2::geom_density() +
    ggplot2::labs(
      title = "Propensity overlap (pooled IPTW)",
      x = "Fitted P(A = 1 | L)",
      y = "Density"
    ) +
    ggplot2::theme_minimal()
}

#' @keywords internal
#' @noRd
.ild_overlap_plot_msm <- function(data, treatment, A) {
  fits <- attr(data, "ild_iptw_msm_fits", exact = TRUE)
  seqv <- data[[".ild_seq"]]
  if (is.null(seqv)) stop("data must have .ild_seq for MSM overlap plot.", call. = FALSE)
  rows <- list()
  for (nm in names(fits)) {
    t <- as.numeric(nm)
    fit <- fits[[nm]]
    idx <- which(as.numeric(seqv) == t & !is.na(A))
    if (length(idx) < 2L) next
    p_hat <- stats::predict(fit, newdata = as.data.frame(data[idx, , drop = FALSE]), type = "response")
    rows[[length(rows) + 1L]] <- tibble::tibble(
      occasion = t,
      p_hat = p_hat,
      treatment = factor(A[idx], levels = c(0, 1), labels = c("A=0", "A=1"))
    )
  }
  if (length(rows) == 0L) {
    stop("No rows to plot for MSM overlap.", call. = FALSE)
  }
  df <- dplyr::bind_rows(rows)
  ggplot2::ggplot(df, ggplot2::aes(x = .data$p_hat, color = .data$treatment)) +
    ggplot2::geom_density() +
    ggplot2::facet_wrap(~ occasion, scales = "free_y") +
    ggplot2::labs(
      title = "Propensity overlap (sequential MSM IPTW)",
      x = "Fitted P(A_t = 1 | history)",
      y = "Density"
    ) +
    ggplot2::theme_minimal()
}

# MSM balance, ESS, overlap

test_that("ild_ipw_ess equals nrow for constant unit weights", {
  set.seed(400)
  d <- ild_simulate(n_id = 2, n_obs_per = 2, seed = 400)
  x <- ild_prepare(d, id = "id", time = "time")
  x$.ipw <- rep(1, nrow(x))
  expect_equal(ild_ipw_ess(x, ".ipw"), nrow(x))
})

test_that("ild_ipw_ess pooled and by occasion", {
  set.seed(401)
  d <- ild_simulate(n_id = 4, n_obs_per = 3, seed = 401)
  x <- ild_prepare(d, id = "id", time = "time")
  x$.ipw <- runif(nrow(x), 0.8, 1.2)
  e0 <- ild_ipw_ess(x, ".ipw", by_occasion = FALSE)
  expect_true(is.numeric(e0) && is.finite(e0))
  e1 <- ild_ipw_ess(x, ".ipw", by_occasion = TRUE)
  expect_s3_class(e1, "tbl_df")
  expect_true("ess" %in% names(e1))
})

test_that("ild_msm_balance pooled returns one row per covariate", {
  set.seed(402)
  d <- ild_simulate(n_id = 20, n_obs_per = 6, seed = 402)
  d$stress <- rnorm(nrow(d))
  d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
  x <- ild_prepare(d, id = "id", time = "time")
  x$.ipw_treat <- runif(nrow(x), 0.85, 1.15)
  tab <- ild_msm_balance(x, treatment = "trt", covariates = "stress", weights_col = ".ipw_treat")
  expect_equal(nrow(tab), 1L)
  expect_true(all(c("stratum", "covariate", "smd", "mean_treated", "mean_control") %in% names(tab)))
  expect_equal(tab$stratum[1], "pooled")
})

test_that("ild_msm_overlap_plot pooled uses ild_iptw_fit", {
  skip_if_not_installed("ggplot2")
  set.seed(403)
  d <- ild_simulate(n_id = 15, n_obs_per = 5, seed = 403)
  d$stress <- rnorm(nrow(d))
  d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_iptw_weights(x, treatment = "trt", predictors = "stress")
  g <- ild_msm_overlap_plot(x, treatment = "trt", source = "pooled")
  expect_s3_class(g, "ggplot")
})

test_that("ild_diagnose balance adds causal$balance when requested", {
  set.seed(404)
  d <- ild_simulate(n_id = 12, n_obs_per = 5, seed = 404)
  d$stress <- rnorm(nrow(d))
  d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  x <- ild_iptw_weights(x, treatment = "trt", predictors = "stress")
  fit <- ild_lme(y ~ y_bp + y_wp + stress + (1 | id), data = x,
    ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
  b <- ild_diagnose(fit, data = x, balance = TRUE, balance_treatment = "trt",
    balance_covariates = "stress", balance_weights_col = ".ipw_treat")
  expect_true(!is.null(b$causal$balance))
  expect_s3_class(b$causal$balance$table, "tbl_df")
  expect_true(!is.null(b$causal$balance$ess))
})

test_that("ild_autoplot causal overlap requires treatment", {
  skip_if_not_installed("ggplot2")
  set.seed(405)
  d <- ild_simulate(n_id = 10, n_obs_per = 5, seed = 405)
  d$stress <- rnorm(nrow(d))
  d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  x <- ild_iptw_weights(x, treatment = "trt", predictors = "stress")
  fit <- ild_lme(y ~ y_bp + y_wp + stress + (1 | id), data = x,
    ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
  b <- ild_diagnose(fit, data = x)
  expect_error(
    ild_autoplot(b, section = "causal", type = "overlap"),
    "treatment"
  )
  p <- ild_autoplot(b, section = "causal", type = "overlap", treatment = "trt")
  expect_s3_class(p, "ggplot")
})

#' Example EMA-style intensive longitudinal dataset
#'
#' A small simulated dataset with 10 persons and 14 observations per person,
#' irregular timing, and two variables (mood, stress). For use in examples and
#' vignettes. Use [ild_prepare()] to convert to an ILD object.
#'
#' @format A data frame with 140 rows and 4 columns:
#' \describe{
#'   \item{id}{Person identifier (1--10).}
#'   \item{time}{POSIXct timestamp (irregular within person).}
#'   \item{mood}{Simulated mood score.}
#'   \item{stress}{Simulated stress score.}
#' }
#' @source Simulated with a fixed seed (12345) for reproducibility.
#' @docType data
#' @name ema_example
#' @keywords datasets
NULL

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)

## ----minimal-crosslag---------------------------------------------------------
library(tidyILD)
set.seed(1)
d <- ild_simulate(n_id = 12, n_obs_per = 10, seed = 1)
x <- ild_prepare(d, id = "id", time = "time")
out <- ild_crosslag(x, y, y, lag = 1L, ar1 = FALSE, warn_no_ar1 = FALSE)
out$lag_term[, c("term", "estimate", "std_error")]

## ----minimal-ar1, eval = requireNamespace("nlme", quietly = TRUE)-------------
x2 <- ild_center(x, y)
fit_ar <- tryCatch(
  ild_lme(y ~ y_bp + y_wp, data = x2, ar1 = TRUE, warn_no_ar1 = FALSE, warn_uncentered = FALSE),
  error = function(e) NULL
)
if (!is.null(fit_ar)) {
  print(fit_ar)
} else {
  "nlme fit skipped on this platform"
}

## ----minimal-tvem-------------------------------------------------------------
x3 <- ild_simulate(n_id = 10, n_obs_per = 15, seed = 2)
x3$x <- rnorm(nrow(x3))
x3 <- ild_prepare(x3, id = "id", time = "time")
tv <- ild_tvem(x3, "y", "x", k = 5, re_id = TRUE)
summary(tv)

## ----session-info, echo = FALSE-----------------------------------------------
sessionInfo()

# Umbrella documentation: standalone diagnostics utilities and bundle section providers
# (see ?ild_diagnostics_utilities)

#' Standalone diagnostics utilities and bundle section providers
#'
#' @description
#' These functions are **standalone utilities** you can run on prepared ILD data at any time
#' (typically before or alongside modeling). The same logic also **feeds**
#' \code{\link{ild_diagnostics_bundle}} sections when you call \code{\link{ild_diagnose}} on a
#' fitted model, so the package presents one coherent diagnostics story: a top-level façade
#' (\code{ild_diagnose}) plus reusable building blocks below.
#'
#' @section Relationship to \code{\link{ild_diagnose}}:
#' \describe{
#'   \item{\code{\link{ild_design_check}}}{Populates the bundle \strong{design} slot (embedded as
#'     \code{design$ild_design_check}) and contributes WP/BP and design-time missingness views.}
#'   \item{\code{\link{ild_missing_pattern}}}{Drives \strong{data} (\code{data$missing_pattern},
#'     global columns) and \strong{missingness} (variable-specific summaries when model predictors
#'     are known).}
#'   \item{\code{\link{ild_missing_model}}}{Not called automatically by \code{ild_diagnose}.
#'     Use when you fit a missingness model for IPW or interpretation; it is the usual
#'     prerequisite for \code{\link{ild_ipw_weights}}.}
#'   \item{\code{\link{ild_ipw_weights}}}{Adds \code{.ipw} (and related steps in the IPW workflow).
#'     When those columns are present on the ILD data attached to the fit, \code{ild_diagnose}
#'     fills the bundle \strong{causal} slot with weight summaries.}
#'   \item{\code{\link{ild_iptw_weights}}, \code{\link{ild_iptw_msm_weights}}, \code{\link{ild_ipcw_weights}}, \code{\link{ild_joint_msm_weights}}}{MSM-style
#'     inverse-probability weights for \strong{treatment} (pooled or sequential \eqn{A_t}) and \strong{monotone censoring}, combined into \code{.ipw}.
#'     Distinct from outcome missingness IPW above.}
#'   \item{\code{\link{ild_msm_bootstrap}}, \code{\link{tidy_ild_msm_bootstrap}}}{Person-level bootstrap for weighted \code{lmer} after \code{\link{ild_ipw_refit}}; see \code{\link{ild_msm_inference}}.}
#'   \item{\code{\link{ild_msm_balance}}, \code{\link{ild_ipw_ess}}, \code{\link{ild_msm_overlap_plot}}}{Weighted covariate balance (SMD), effective sample size, and propensity overlap plots for MSM / IPW workflows.}
#'   \item{\code{\link{ild_msm_estimand}}, \code{\link{ild_msm_fit}}, \code{\link{ild_msm_diagnose}}, \code{\link{ild_msm_contrast_over_time}}, \code{\link{ild_msm_history_spec}}, \code{\link{ild_build_msm_history}}, \code{\link{ild_msm_recovery}}}{Estimand-first MSM workflow, one-call diagnostics bridge, time-indexed contrasts, history construction, and simulation-based recovery checks.}
#' }
#'
#' @seealso \code{\link{ild_diagnose}}, \code{\link{ild_diagnostics_bundle}}, \code{\link{ild_diagnostics}}
#' @name ild_diagnostics_utilities
NULL

# Declarative brms formula templates for simple dynamics (no automatic fit)

#' Template brms formula for person-varying lag slope
#'
#' Returns a suggested formula and notes for fitting with [ild_brms()].
#' This does **not** run the model. Identification, priors, and convergence
#' depend on N, T, and scaling; see \code{vignette("brms-dynamics-recipes", package = "tidyILD")}.
#'
#' @param outcome Character name of the response.
#' @param lag_var Character name of an existing lag column (e.g. from [ild_lag()]).
#' @param id_var Character name of the grouping column in \code{data} (default
#'   \code{"id"} as stored after [ild_prepare()] — use the original id column name).
#' @return A list: \code{formula} (\code{formula} object), \code{notes} (character).
#' @seealso [ild_brms()], [ild_panel_lag_prepare()].
#' @export
ild_brms_dynamics_formula <- function(outcome, lag_var, id_var = "id") {
  outcome <- as.character(outcome)[1L]
  lag_var <- as.character(lag_var)[1L]
  id_var <- as.character(id_var)[1L]
  f <- stats::as.formula(paste0(
    outcome, " ~ ", lag_var, " + (", lag_var, " | ", id_var, ") + (1 | ", id_var, ")"
  ))
  notes <- paste(
    "Random slope for the lag column with random intercept; requires scaling and",
    "informative-ish priors for sd. Not a full VAR; one lagged predictor only."
  )
  list(formula = f, notes = notes)
}

# Augment method for ild_kfas

#' @rdname ild_augment
#' @method ild_augment ild_fit_kfas
#' @export
ild_augment.ild_fit_kfas <- function(x, ...) {
  rlang::check_installed("KFAS", reason = "to use ild_augment() for ild_kfas fits")
  data <- attr(x, "ild_data", exact = TRUE)
  if (is.null(data)) {
    stop("ild_augment() for ild_fit_kfas requires attr(fit, \"ild_data\").", call. = FALSE)
  }
  validate_ild(data)
  resp <- attr(x, "ild_response", exact = TRUE)
  if (is.null(resp)) {
    stop("ild_response attribute missing; refit with ild_kfas().", call. = FALSE)
  }
  y <- data[[resp]]
  y <- if (is.numeric(y)) y else suppressWarnings(as.numeric(y))

  ex <- ild_kfas_extract_augment(x$kfs, y, use_smoother = TRUE)
  fit <- ex$fitted
  resid <- ex$resid
  rs <- ex$resid_std
  rs_k <- tryCatch(
    as.numeric(stats::rstandard(x$kfs)),
    error = function(e) NULL
  )
  if (!is.null(rs_k) && length(rs_k) == length(y)) {
    rs <- rs_k
  }

  st <- ild_kfas_extract_level_state(x$kfs)
  opt <- list()
  if (!is.null(st$mean) && length(st$mean) == nrow(data)) {
    opt$.state <- st$mean
    # crude 95% for smoothed state if V available
    V <- x$kfs$V
    if (!is.null(V) && length(dim(V)) >= 3) {
      lo <- hi <- rep(NA_real_, nrow(data))
      for (i in seq_len(nrow(data))) {
        v <- tryCatch(V[1L, 1L, i], error = function(e) NA_real_)
        if (is.finite(v) && v >= 0) {
          se <- sqrt(v)
          lo[i] <- st$mean[i] - 1.96 * se
          hi[i] <- st$mean[i] + 1.96 * se
        }
      }
      opt$.state_lower <- lo
      opt$.state_upper <- hi
    }
  }

  ild_augment_assemble(
    .ild_id = data[[".ild_id"]],
    .ild_time = data[[".ild_time"]],
    .outcome = y,
    .fitted = fit,
    .resid = resid,
    .resid_std = rs,
    engine = "KFAS",
    model_class = "ild_fit_kfas",
    optional = opt
  )
}

#' Per-row latent states (long format) for complex multi-state models
#'
#' For \code{local_level} only, use [ild_augment()] which already provides
#' \code{.state} for the level. This function is reserved for future multi-state
#' models.
#'
#' @param x An object from [ild_kfas()].
#' @param ... Reserved.
#' @return A \code{\link[tibble]{tibble}}.
#' @export
ild_augment_states <- function(x, ...) {
  UseMethod("ild_augment_states")
}

#' @export
ild_augment_states.ild_fit_kfas <- function(x, ...) {
  ild_tidy_states(x)
}

# KFAS visualization: bundle autoplot helpers and user-facing ild_plot_* functions

#' Plot smoothed latent states (first state by default)
#'
#' @param x An object from [ild_kfas()].
#' @param state_index Integer; which column of \code{alphahat} to plot (default \code{1L}).
#' @param ... Passed to [ggplot2::labs()].
#' @return A \code{ggplot} object.
#' @export
ild_plot_states <- function(x, state_index = 1L, ...) {
  rlang::check_installed("KFAS", reason = "to plot ild_kfas states")
  if (!inherits(x, "ild_fit_kfas")) {
    stop("ild_plot_states() expects an ild_fit_kfas object from ild_kfas().", call. = FALSE)
  }
  dat <- attr(x, "ild_data", exact = TRUE)
  if (is.null(dat)) stop("Missing attr(fit, \"ild_data\").", call. = FALSE)
  ah <- x$kfs$alphahat
  if (is.null(ah) || !is.matrix(ah) || ncol(ah) < state_index) {
    stop("Smoothed states (alphahat) not available.", call. = FALSE)
  }
  tidx <- seq_len(nrow(dat))
  df <- data.frame(
    time = tidx,
    state = as.numeric(ah[, state_index]),
    stringsAsFactors = FALSE
  )
  ggplot2::ggplot(df, ggplot2::aes(x = .data$time, y = .data$state)) +
    ggplot2::geom_line(color = "steelblue") +
    ggplot2::labs(
      title = sprintf("Smoothed state (%s)", colnames(ah)[state_index] %||% paste0("state_", state_index)),
      x = "Time index",
      y = "Estimate",
      ...
    ) +
    ggplot2::theme_minimal()
}

#' Plot filtered vs smoothed state (first state)
#'
#' Compares one-step filtered state (\code{att}) with smoothed (\code{alphahat}) when available.
#'
#' @param x An object from [ild_kfas()].
#' @return A \code{ggplot} object.
#' @export
ild_plot_filtered_vs_smoothed <- function(x) {
  rlang::check_installed("KFAS", reason = "to plot ild_kfas states")
  if (!inherits(x, "ild_fit_kfas")) {
    stop("ild_plot_filtered_vs_smoothed() expects ild_fit_kfas.", call. = FALSE)
  }
  dat <- attr(x, "ild_data", exact = TRUE)
  if (is.null(dat)) stop("Missing attr(fit, \"ild_data\").", call. = FALSE)
  kfs <- x$kfs
  ah <- kfs$alphahat
  att <- kfs$att
  if (is.null(ah) || !is.matrix(ah)) {
    stop("Smoothed states not available.", call. = FALSE)
  }
  sm <- as.numeric(ah[, 1L])
  n <- nrow(dat)
  tidx <- seq_len(n)
  if (!is.null(att) && is.matrix(att) && nrow(att) >= n) {
    ft <- as.numeric(att[seq_len(n), 1L])
  } else {
    ft <- rep(NA_real_, n)
  }
  df <- data.frame(
    time = tidx,
    smoothed = sm,
    filtered = ft,
    stringsAsFactors = FALSE
  )
  df_long <- rbind(
    data.frame(time = df$time, value = df$smoothed, series = "smoothed", stringsAsFactors = FALSE),
    data.frame(time = df$time, value = df$filtered, series = "filtered", stringsAsFactors = FALSE)
  )
  ggplot2::ggplot(df_long, ggplot2::aes(x = .data$time, y = .data$value, colour = .data$series)) +
    ggplot2::geom_line(na.rm = TRUE) +
    ggplot2::labs(
      title = "Filtered vs smoothed (first latent state)",
      x = "Time index",
      y = "Estimate",
      colour = NULL
    ) +
    ggplot2::theme_minimal()
}

#' Plot forecast or future simulation (stub when no horizon)
#'
#' When \code{forecast_horizon > 0} was used in [ild_kfas()], attempts a short
#' ahead forecast via \pkg{KFAS}; otherwise returns an informative empty panel.
#'
#' @param x An object from [ild_kfas()].
#' @param ... Passed to \code{\link[stats]{predict}} for the fitted \pkg{KFAS} model when forecasting is available.
#' @return A \code{ggplot} object.
#' @export
ild_plot_forecast <- function(x, ...) {
  rlang::check_installed("KFAS", reason = "to forecast ild_kfas")
  if (!inherits(x, "ild_fit_kfas")) {
    stop("ild_plot_forecast() expects ild_fit_kfas.", call. = FALSE)
  }
  fh <- x$mapping$forecast_horizon
  if (is.null(fh) || fh < 1L) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate(
          "text",
          x = 0.5, y = 0.5,
          label = "No forecast requested (forecast_horizon = 0).\nPass forecast_horizon > 0 to ild_kfas() when implemented.",
          hjust = 0.5, vjust = 0.5
        ) +
        ggplot2::theme_void() +
        ggplot2::labs(title = "Forecast") +
        ggplot2::xlim(0, 1) +
        ggplot2::ylim(0, 1)
    )
  }
  pred <- tryCatch(
    stats::predict(x$kfas_model, n.ahead = as.integer(fh), ...),
    error = function(e) NULL
  )
  if (is.null(pred)) {
    return(
      ggplot2::ggplot() +
        ggplot2::annotate("text", x = 0.5, y = 0.5, label = "Forecast unavailable", hjust = 0.5) +
        ggplot2::theme_void() +
        ggplot2::labs(title = "Forecast") +
        ggplot2::xlim(0, 1) +
        ggplot2::ylim(0, 1)
    )
  }
  p <- as.numeric(pred)
  df <- data.frame(h = seq_along(p), y = p, stringsAsFactors = FALSE)
  ggplot2::ggplot(df, ggplot2::aes(x = .data$h, y = .data$y)) +
    ggplot2::geom_line() +
    ggplot2::labs(x = "Horizon", y = "Forecast", title = "KFAS forecast (mean)") +
    ggplot2::theme_minimal()
}

#' @keywords internal
#' @noRd
plot_bundle_kfas_residual_acf <- function(x) {
  rlang::check_installed("KFAS")
  ac <- x$residual$acf %||% x$residual$acf_innovations
  fit <- attr(x, "ild_fit", exact = TRUE)
  if (!is.null(ac) && is.list(ac) && !is.null(ac$lag) && !is.null(ac$acf)) {
    df <- data.frame(lag = ac$lag, acf = ac$acf, stringsAsFactors = FALSE)
  } else if (inherits(fit, "ild_fit_kfas")) {
    innov <- tryCatch(as.numeric(stats::rstandard(fit$kfs)), error = function(e) NULL)
    if (is.null(innov) || length(innov) < 2L) {
      v <- fit$kfs$v
      Fv <- fit$kfs$F
      if (!is.null(v) && !is.null(Fv)) innov <- as.numeric(v) / sqrt(pmax(as.numeric(Fv), 1e-12))
    }
    if (is.null(innov) || length(innov) < 2L) {
      stop("Could not extract innovations for ACF.", call. = FALSE)
    }
    a <- stats::acf(innov, plot = FALSE, lag.max = min(20L, length(innov) - 1L))
    la <- a$lag
    acf <- a$acf
    if (is.matrix(la)) la <- as.numeric(la[, 1L]) else as.numeric(la)
    if (is.matrix(acf)) acf <- as.numeric(acf[, 1L]) else as.numeric(acf)
    df <- data.frame(lag = la, acf = acf, stringsAsFactors = FALSE)
  } else {
    stop("KFAS bundle residual ACF requires KFAS diagnostics or ild_fit_kfas.", call. = FALSE)
  }
  ggplot2::ggplot(df, ggplot2::aes(x = .data$lag, y = .data$acf)) +
    ggplot2::geom_col(width = 0.02) +
    ggplot2::geom_hline(yintercept = 0, linetype = 2) +
    ggplot2::labs(
      title = "ACF of standardized innovations",
      x = "Lag",
      y = "ACF"
    ) +
    ggplot2::theme_minimal()
}

#' @keywords internal
#' @noRd
plot_bundle_kfas_residual_qq <- function(x) {
  rlang::check_installed("KFAS")
  fit <- attr(x, "ild_fit", exact = TRUE)
  if (!inherits(fit, "ild_fit_kfas")) {
    stop("KFAS Q-Q requires bundle from ild_diagnose(ild_kfas_fit).", call. = FALSE)
  }
  innov <- tryCatch(as.numeric(stats::rstandard(fit$kfs)), error = function(e) NULL)
  if (is.null(innov)) {
    v <- fit$kfs$v
    Fv <- fit$kfs$F
    if (!is.null(v) && !is.null(Fv)) innov <- as.numeric(v) / sqrt(pmax(as.numeric(Fv), 1e-12))
  }
  if (is.null(innov) || length(innov) < 3L) {
    stop("Could not extract standardized innovations for Q-Q.", call. = FALSE)
  }
  qq <- stats::qqnorm(innov, plot = FALSE)
  df <- data.frame(x = qq$x, y = qq$y, stringsAsFactors = FALSE)
  ggplot2::ggplot(df, ggplot2::aes(x = .data$x, y = .data$y)) +
    ggplot2::geom_point(alpha = 0.6) +
    ggplot2::geom_abline(slope = 1, intercept = 0, linetype = 2, colour = "grey50") +
    ggplot2::labs(
      title = "Normal Q-Q of standardized innovations",
      x = "Theoretical quantiles",
      y = "Sample quantiles"
    ) +
    ggplot2::theme_minimal()
}

#' @keywords internal
#' @noRd
plot_bundle_kfas_residual_fitted <- function(x) {
  fit <- attr(x, "ild_fit", exact = TRUE)
  dat <- attr(x, "ild_data", exact = TRUE)
  if (!inherits(fit, "ild_fit_kfas") || is.null(dat)) {
    stop("KFAS fitted vs signal requires ild_fit_kfas and ild_data on bundle.", call. = FALSE)
  }
  resp <- attr(fit, "ild_response", exact = TRUE)
  y <- dat[[resp]]
  y <- if (is.numeric(y)) y else suppressWarnings(as.numeric(y))
  ex <- ild_kfas_extract_augment(fit$kfs, y, TRUE)
  tidx <- seq_len(nrow(dat))
  df <- data.frame(
    time = tidx,
    y = y,
    fitted = ex$fitted,
    stringsAsFactors = FALSE
  )
  ggplot2::ggplot(df, ggplot2::aes(x = .data$time)) +
    ggplot2::geom_line(ggplot2::aes(y = .data$y, colour = "Observed")) +
    ggplot2::geom_line(ggplot2::aes(y = .data$fitted, colour = "Signal")) +
    ggplot2::labs(
      title = "Outcome vs signal expectation",
      x = "Time index",
      y = NULL,
      colour = NULL
    ) +
    ggplot2::theme_minimal()
}

#' @keywords internal
#' @noRd
plot_bundle_kfas_fit_convergence <- function(x) {
  fd <- x$fit
  meta <- x$meta
  ll <- fd$log_likelihood
  if (is.null(ll)) ll <- NA_real_
  opt <- fd$optimizer
  conv <- if (!is.null(fd$convergence)) fd$convergence$converged else NA
  meth <- if (!is.null(opt$method)) opt$method else "?"
  txt <- paste(
    c(
      sprintf("Engine: %s", if (!is.null(fd$engine)) fd$engine else "KFAS"),
      sprintf("State spec: %s", if (!is.null(meta$state_spec)) meta$state_spec else "?"),
      sprintf("Observation family: %s", if (!is.null(meta$observation_family)) meta$observation_family else "gaussian"),
      sprintf("Log-likelihood: %s", if (is.finite(ll)) format(round(ll, 4)) else "NA"),
      sprintf("Optimizer: %s", meth),
      sprintf("Converged: %s", if (isTRUE(conv)) "TRUE" else if (isFALSE(conv)) "FALSE" else "NA"),
      sprintf("Pool mode: %s", if (!is.null(meta$pool_mode)) meta$pool_mode else "none"),
      sprintf("Fit context: %s", if (!is.null(meta$fit_context)) meta$fit_context else "single_series")
    ),
    collapse = "\n"
  )
  ggplot2::ggplot() +
    ggplot2::annotate("text", x = 0, y = 1, label = txt, hjust = 0, vjust = 1, size = 3.2) +
    ggplot2::theme_void() +
    ggplot2::labs(title = "KFAS model fit (MLE)") +
    ggplot2::xlim(0, 1) +
    ggplot2::ylim(0, 1.05)
}

#' @keywords internal
#' @noRd
plot_bundle_kfas_predictive_forecast <- function(x) {
  fit <- attr(x, "ild_fit", exact = TRUE)
  if (!inherits(fit, "ild_fit_kfas")) {
    stop("KFAS forecast plot requires ild_fit_kfas.", call. = FALSE)
  }
  ild_plot_forecast(fit)
}

#' @keywords internal
#' @noRd
plot_bundle_kfas_predictive_errors <- function(x) {
  fit <- attr(x, "ild_fit", exact = TRUE)
  dat <- attr(x, "ild_data", exact = TRUE)
  if (!inherits(fit, "ild_fit_kfas") || is.null(dat)) {
    stop("KFAS errors plot requires bundle from ild_diagnose(ild_kfas_fit).", call. = FALSE)
  }
  rlang::check_installed("KFAS")
  innov <- tryCatch(as.numeric(stats::rstandard(fit$kfs)), error = function(e) NULL)
  if (is.null(innov)) {
    v <- fit$kfs$v
    Fv <- fit$kfs$F
    if (!is.null(v) && !is.null(Fv)) innov <- as.numeric(v) / sqrt(pmax(as.numeric(Fv), 1e-12))
  }
  if (is.null(innov)) {
    stop("Standardized innovations unavailable.", call. = FALSE)
  }
  n <- min(length(innov), nrow(dat))
  df <- data.frame(
    time = seq_len(n),
    abs_innovation = abs(innov[seq_len(n)]),
    stringsAsFactors = FALSE
  )
  ggplot2::ggplot(df, ggplot2::aes(x = .data$time, y = .data$abs_innovation)) +
    ggplot2::geom_line(colour = "coral") +
    ggplot2::geom_hline(yintercept = 0, linetype = 3) +
    ggplot2::labs(
      title = "One-step |standardized innovation|",
      x = "Time index",
      y = "|rstandard|"
    ) +
    ggplot2::theme_minimal()
}

# Internal helpers for ild_tidy() schema conformance (ILD_TIDY_REQUIRED_COLS / optional)

#' Conservative effect_level for fixed-effect coefficient rows (term-name based)
#' @keywords internal
#' @noRd
ild_tidy_effect_level_fixed <- function(term) {
  v <- as.character(term)[1]
  if (is.na(v)) return("unknown")
  if (v %in% c("(Intercept)", "Intercept")) return("population")
  if (grepl("_wp$", v)) return("within")
  if (grepl("_bp$", v)) return("between")
  if (grepl(":", v, fixed = TRUE)) {
    if (grepl("_wp", v) && grepl("_bp", v)) return("cross_level")
    return("unknown")
  }
  "unknown"
}

#' @keywords internal
#' @noRd
ild_tidy_model_class_string <- function(fit) {
  paste(class(fit), collapse = ", ")
}

#' Extract test statistic vector aligned to fixed-effect rows
#' @keywords internal
#' @noRd
ild_tidy_extract_statistic <- function(fit, est, se_vec, se_mode) {
  n <- length(est)
  stat <- rep(NA_real_, n)
  if (identical(se_mode, "robust")) {
    ok <- is.finite(est) & is.finite(se_vec) & se_vec > 0
    stat[ok] <- est[ok] / se_vec[ok]
    return(stat)
  }
  if (inherits(fit, "lmerMod")) {
    cc <- stats::coef(summary(fit))
    if (!is.null(cc) && "t value" %in% colnames(cc) && nrow(cc) == n) {
      stat <- as.vector(cc[, "t value"])
    }
  } else if (inherits(fit, "lme")) {
    tt <- summary(fit)$tTable
    if (!is.null(tt) && "t-value" %in% colnames(tt) && nrow(tt) == n) {
      stat <- as.vector(tt[, "t-value"])
    }
  }
  stat
}

#' Interval label for frequentist ild_lme fits
#' @keywords internal
#' @noRd
ild_tidy_interval_type_frequentist <- function(se_mode) {
  if (identical(se_mode, "robust")) return("Wald")
  "Wald"
}

#' Assemble full tidy tibble with required + optional columns (schema order)
#' @keywords internal
#' @noRd
ild_tidy_assemble <- function(term,
                              estimate,
                              std_error,
                              conf_low,
                              conf_high,
                              p_value,
                              statistic,
                              interval_type,
                              engine,
                              model_class,
                              component = "fixed",
                              effect_level = NULL,
                              optional = NULL) {
  if (is.null(effect_level)) {
    effect_level <- vapply(term, ild_tidy_effect_level_fixed, character(1), USE.NAMES = FALSE)
  }
  n <- length(term)
  stopifnot(
    length(estimate) == n, length(std_error) == n, length(conf_low) == n,
    length(conf_high) == n, length(p_value) == n, length(statistic) == n,
    length(interval_type) == n, length(component) == n, length(effect_level) == n
  )
  out <- tibble::tibble(
    term = term,
    component = component,
    effect_level = effect_level,
    estimate = estimate,
    std_error = std_error,
    conf_low = conf_low,
    conf_high = conf_high,
    statistic = statistic,
    p_value = p_value,
    interval_type = interval_type,
    engine = engine,
    model_class = model_class
  )
  opt <- ILD_TIDY_OPTIONAL_COLS
  if (!is.null(optional) && is.list(optional)) {
    for (nm in opt) {
      out[[nm]] <- if (!is.null(optional[[nm]])) optional[[nm]] else rep(NA_real_, n)
    }
  } else {
    for (nm in opt) {
      out[[nm]] <- rep(NA_real_, n)
    }
  }
  out
}

# MSM post-fit contrasts

#' Compute MSM contrasts over time from a fitted weighted model
#'
#' Computes marginal treatment contrasts by occasion using model fixed effects.
#' This helper is intentionally standalone: it accepts either an \code{ild_msm_fit}
#' object or a weighted \code{lmerMod} with attached \code{ild_data}.
#'
#' @param object Output from [ild_msm_fit()] or a \code{lmerMod}.
#' @param treatment Optional treatment column name. Defaults to
#'   \code{object$estimand$treatment} for \code{ild_msm_fit}, otherwise required.
#' @param time_var Occasion column (default \code{".ild_seq"}).
#' @param target_time Optional target subset: \code{"all"} (default), \code{"final"},
#'   or numeric vector of occasions.
#' @param conf_level Confidence level for Wald intervals (default \code{0.95}).
#' @return Tibble with `ild_tidy_schema` columns plus \code{method},
#'   \code{terms}, and \code{target_time}.
#' @export
ild_msm_contrast_over_time <- function(object,
                                       treatment = NULL,
                                       time_var = ".ild_seq",
                                       target_time = "all",
                                       conf_level = 0.95) {
  fit <- object
  data <- NULL
  inf_method <- "model_wald"
  est_target_time <- NULL
  if (inherits(object, "ild_msm_fit")) {
    fit <- object$fit
    data <- object$weights_data
    if (is.null(treatment)) treatment <- object$estimand$treatment
    est_target_time <- object$estimand$target_time
    if (!is.null(object$inference) && !is.null(object$inference$method)) {
      inf_method <- as.character(object$inference$method)[1L]
    }
  }
  if (!inherits(fit, "lmerMod")) stop("object must contain a lmerMod fit.", call. = FALSE)
  if (is.null(data)) data <- attr(fit, "ild_data", exact = TRUE)
  if (is.null(data)) stop("Unable to find ILD data (attr(fit, \"ild_data\")).", call. = FALSE)
  if (is.null(treatment) || !treatment %in% names(data)) {
    stop("treatment column must be supplied and present in data.", call. = FALSE)
  }
  if (!time_var %in% names(data)) {
    stop("time_var not found in data.", call. = FALSE)
  }
  if (identical(target_time, "all") && !is.null(est_target_time)) {
    target_time <- est_target_time
  }
  beta <- lme4::fixef(fit)
  vc <- tryCatch(stats::vcov(fit), error = function(e) NULL)
  if (is.null(vc)) stop("Could not compute vcov for contrast SEs.", call. = FALSE)
  f_fix <- lme4::nobars(stats::formula(fit))
  z <- stats::qnorm(1 - (1 - conf_level) / 2)

  tt <- sort(unique(data[[time_var]]))
  if (is.character(target_time)) {
    tgt <- tolower(as.character(target_time)[1L])
    if (identical(tgt, "final")) {
      tt <- tt[length(tt)]
    } else if (!identical(tgt, "all")) {
      stop("target_time must be \"all\", \"final\", or numeric vector.", call. = FALSE)
    }
  } else {
    tnum <- as.numeric(target_time)
    tt <- tt[tt %in% tnum]
  }
  if (!length(tt)) {
    stop("No observed occasions matched target_time.", call. = FALSE)
  }
  rows <- vector("list", length(tt))
  for (i in seq_along(tt)) {
    tval <- tt[[i]]
    d_t <- data[data[[time_var]] == tval, , drop = FALSE]
    d1 <- d_t
    d0 <- d_t
    d1[[treatment]] <- 1
    d0[[treatment]] <- 0
    X1 <- stats::model.matrix(f_fix, data = d1)
    X0 <- stats::model.matrix(f_fix, data = d0)
    cm <- intersect(colnames(X1), names(beta))
    if (!length(cm)) {
      stop("Could not align fixed-effect design columns with coefficient names.", call. = FALSE)
    }
    X1 <- X1[, cm, drop = FALSE]
    X0 <- X0[, cm, drop = FALSE]
    b_use <- beta[cm]
    vc_use <- vc[cm, cm, drop = FALSE]
    dbar <- colMeans(X1 - X0)
    est <- as.numeric(sum(dbar * b_use))
    se <- as.numeric(sqrt(drop(dbar %*% vc_use %*% dbar)))
    stat <- est / se
    pval <- 2 * (1 - stats::pnorm(abs(stat)))
    rows[[i]] <- tibble::tibble(
      term = paste0("contrast_t", tval),
      estimate = est,
      std_error = se,
      conf_low = est - z * se,
      conf_high = est + z * se,
      p_value = pval,
      statistic = stat,
      interval_type = "Wald",
      method = inf_method,
      component = "fixed",
      effect_level = "population",
      engine = "lmer",
      model_class = ild_tidy_model_class_string(fit),
      target_time = as.numeric(tval)
    )
  }
  out <- dplyr::bind_rows(rows)
  for (nm in ILD_TIDY_OPTIONAL_COLS) out[[nm]] <- NA_real_
  out$terms <- out$term
  out
}

# KFAS state-space models for single-subject ILD (optional dependency: KFAS)

#' Fit a Gaussian state-space model via KFAS (single-subject ILD)
#'
#' Opinionated entry point for **local level** models in v1. Requires the
#' \pkg{KFAS} package. One distinct \code{.ild_id} per fit; subset your data first.
#'
#' @param data An object that passes [validate_ild()] (e.g. after [ild_prepare()]).
#' @param outcome Character name of the numeric outcome column.
#' @param state_spec Character; use \code{"local_level"} in v1. See
#'   \code{inst/dev/KFAS_V1_BACKEND.md} for the full controlled vocabulary.
#' @param observation_family Currently only \code{"gaussian"}.
#' @param time_units Character describing the time column (e.g. \code{"days"},
#'   \code{"hours"}). Stored in provenance; required for clear reporting.
#' @param irregular_time Logical; if \code{TRUE}, skips the irregular-spacing warning.
#' @param smoother Logical; passed to \code{KFS()} smoothing (default \code{TRUE}).
#' @param forecast_horizon Integer; when \code{> 0}, used by \code{\link{ild_plot_forecast}()}
#'   with \code{\link[stats]{predict}} on the fitted model and stored in \code{mapping$forecast_horizon}.
#' @param fit_context \code{"single_series"} (default) or \code{"independent_series_per_id"}
#'   when you have fit the **same** state-space template separately for multiple persons
#'   and must not interpret results as a pooled multilevel latent model. The latter
#'   triggers guardrail \code{GR_KFAS_UNMODELED_BETWEEN_PERSON_HETEROGENEITY}.
#' @param ... Passed to \code{KFAS::fitSSM()} (e.g. \code{method}, \code{inits}).
#' @return An object of class \code{c("ild_fit_kfas", "ild_fit_model", "ild_analysis")}
#'   with list slots \code{kfas_model}, \code{kfs}, \code{spec}, \code{state_labels},
#'   \code{mapping}, \code{schema_version}, \code{preprocessing}. Attributes \code{ild_data} and
#'   \code{ild_provenance} are set.
#' @seealso \code{inst/dev/KFAS_V1_BACKEND.md}, [ild_tidy()], [ild_augment()], [ild_diagnose()].
#'
#' \strong{Conceptual vignettes:}
#' \itemize{
#'   \item \href{../doc/kfas-state-space-modeling.html}{\code{vignette("kfas-state-space-modeling", package = "tidyILD")}}
#'   \item \href{../doc/kfas-irregular-timing-spacing.html}{\code{vignette("kfas-irregular-timing-spacing", package = "tidyILD")}}
#'   \item \href{../doc/kfas-choosing-backend.html}{\code{vignette("kfas-choosing-backend", package = "tidyILD")}}
#' }
#' @export
ild_kfas <- function(data,
                     outcome,
                     state_spec = "local_level",
                     observation_family = "gaussian",
                     time_units,
                     irregular_time = FALSE,
                     smoother = TRUE,
                     forecast_horizon = 0L,
                     fit_context = c("single_series", "independent_series_per_id"),
                     ...) {
  rlang::check_installed("KFAS", reason = "to use ild_kfas()")
  if (missing(time_units) || !nzchar(as.character(time_units)[1L])) {
    stop("time_units must be a non-empty string (e.g. 'days', 'hours').", call. = FALSE)
  }
  observation_family <- match.arg(observation_family, c("gaussian"))
  fit_context <- match.arg(fit_context)
  state_spec <- ild_kfas_normalize_state_spec(state_spec, implemented_only = TRUE)
  if (!isTRUE(irregular_time)) {
    ild_kfas_warn_irregular(data)
  }

  prep <- ild_kfas_prepare_ts(data, outcome)
  d <- prep$data
  y <- prep$y
  is_na_y <- is.na(y)
  n_na_segments <- ild_kfas_na_run_count(is_na_y)
  pct_outcome_na <- if (length(is_na_y) > 0L) mean(is_na_y) else NA_real_
  n_rows_input <- nrow(d)
  if (any(is_na_y)) {
    warning("Removing rows with NA in outcome for KFAS.", call. = FALSE)
    ok <- !is_na_y
    d <- d[ok, , drop = FALSE]
    y <- y[ok]
  }
  if (length(y) < 5L) {
    stop("Need at least 5 non-missing outcome values for ild_kfas().", call. = FALSE)
  }
  preprocessing <- list(
    n_rows_input = n_rows_input,
    n_outcome_na = sum(is_na_y),
    n_na_segments = n_na_segments,
    pct_outcome_na = pct_outcome_na,
    n_rows_used = nrow(d)
  )

  dots <- list(...)
  if (state_spec == "local_level") {
    # Local level: random walk + noise; variances estimated by ML
    # Formula must see unqualified SSMtrend() for stats::model.frame (KFAS requirement).
    SSMtrend <- KFAS::SSMtrend
    mod <- KFAS::SSModel(
      y ~ SSMtrend(degree = 1L, Q = list(matrix(NA_real_))),
      H = matrix(NA_real_)
    )
    vy <- stats::var(y, na.rm = TRUE)
    vd <- stats::var(diff(as.numeric(y)), na.rm = TRUE)
    if (!is.finite(vy) || vy <= 0) vy <- 1
    if (!is.finite(vd) || vd <= 0) vd <- vy
    inits <- c(log(vy), log(vd))
    fit_ssm <- do.call(
      KFAS::fitSSM,
      c(list(mod, inits = inits, method = "BFGS"), dots)
    )
    sm <- if (isTRUE(smoother)) "state" else "none"
    kfs <- KFAS::KFS(fit_ssm$model, smoothing = sm, filtering = "state")
  } else {
    stop("Unreachable: state_spec not implemented.", call. = FALSE)
  }

  spec <- ild_kfas_internal_spec(
    state_spec = state_spec,
    observation_family = observation_family,
    time_units = as.character(time_units)[1L],
    irregular_time = isTRUE(irregular_time),
    fit_context = fit_context,
    pool_mode = "none"
  )

  m_state <- tryCatch(ncol(kfs$alphahat), error = function(e) NA_integer_)
  diffuse_end <- tryCatch(as.integer(kfs$d), error = function(e) NA_integer_)
  state_labels <- c(level = "local_level")
  mapping <- list(
    outcome = outcome,
    n_time = nrow(d),
    id = prep$id_label,
    time_units = spec$time_units,
    state_dimension = m_state,
    observation_dimension = 1L,
    pool_mode = "none",
    smoother = isTRUE(smoother),
    diffuse_phase_end = diffuse_end,
    forecast_horizon = as.integer(forecast_horizon)
  )

  out <- list(
    kfas_model = fit_ssm$model,
    kfs = kfs,
    fit_ssm = fit_ssm,
    spec = spec,
    state_labels = state_labels,
    mapping = mapping,
    preprocessing = preprocessing,
    schema_version = ILD_KFAS_SCHEMA_VERSION
  )
  class(out) <- c("ild_fit_kfas", "ild_fit_model", "ild_analysis")

  attr(out, "ild_data") <- d
  attr(out, "ild_response") <- outcome
  attr(out, "ild_kfas_schema") <- list(version = ILD_KFAS_SCHEMA_VERSION)
  attr(out, "ild_provenance") <- ild_new_analysis_provenance(
    data,
    "ild_kfas",
    args = list(
      outcome = outcome,
      state_spec = state_spec,
      observation_family = observation_family,
      time_units = spec$time_units,
      irregular_time = irregular_time,
      smoother = smoother,
      kfs_smoothing = sm,
      kfs_filtering = "state",
      forecast_horizon = as.integer(forecast_horizon),
      fit_context = fit_context,
      state_labels = state_labels,
      fit_ssm_dots = dots
    ),
    outputs = list(
      n_obs = nrow(d),
      n_id = 1L,
      logLik = tryCatch(as.numeric(kfs$logLik), error = function(e) NA_real_),
      kfas_version = tryCatch(
        as.character(utils::packageVersion("KFAS")),
        error = function(e) NA_character_
      ),
      diffuse_phase_end = diffuse_end,
      pool_mode = "none"
    )
  )
  out
}

#' Spacing-aware lag within person
#'
#' Computes lagged values within each person. Use this instead of
#' [dplyr::lag()], which assumes equal spacing and no gaps and is unsafe
#' for irregular ILD.
#'
#' @param x An ILD object (see [is_ild()]).
#' @param ... Variables to lag (tidy-select). Unquoted names or selection.
#' @param n Integer. Lag order (default 1 = previous observation).
#' @param mode Character. \code{"index"}: row-based lag. \code{"gap_aware"}: same but NA when interval exceeds \code{max_gap}. \code{"time_window"}: value from (time - window, time] with \code{resolution}.
#' @param max_gap Numeric. For \code{gap_aware} only. Same units as \code{.ild_time_num}.
#'   If \code{NULL}, uses \code{ild_meta(x)$ild_gap_threshold} (metadata-driven default).
#' @param window Numeric or lubridate duration. For \code{time_window} only: time window width.
#'   Numeric is in same units as \code{.ild_time_num} (e.g. seconds for POSIXct). You can pass
#'   a lubridate period/duration (e.g. \code{lubridate::hours(2)}); it is converted to seconds for POSIXct.
#' @param resolution Character. For \code{time_window}: \code{"closest_prior"} (default: most recent observation in window),
#'   \code{"last_in_window"}, or \code{"mean_in_window"}.
#' @return The same ILD tibble with new lag columns. ILD attributes preserved.
#' @importFrom dplyr group_by lag mutate select ungroup
#' @export
ild_lag <- function(x, ..., n = 1L, mode = c("index", "gap_aware", "time_window"),
                   max_gap = NULL, window = NULL,
                   resolution = c("closest_prior", "last_in_window", "mean_in_window")) {
  validate_ild(x)
  mode <- match.arg(mode)
  resolution <- match.arg(resolution)
  if (mode == "gap_aware" && is.null(max_gap)) {
    max_gap <- ild_meta(x)$ild_gap_threshold
    if (!is.finite(max_gap)) max_gap <- Inf
  }
  if (mode != "gap_aware" && is.null(max_gap)) max_gap <- Inf
  n <- as.integer(n)[1]
  if (n < 1 && mode != "time_window") stop("'n' must be >= 1 for index/gap_aware.", call. = FALSE)
  vars <- names(dplyr::select(x, ...))
  if (length(vars) == 0) stop("No variables selected for lag.", call. = FALSE)
  if (mode == "time_window") {
    if (is.null(window)) stop("'window' is required for mode = 'time_window'.", call. = FALSE)
    window <- ild_window_to_num(window, x)
    if (!is.finite(window) || window <= 0) stop("'window' must be a positive number for mode = 'time_window'.", call. = FALSE)
  }
  id_col <- ".ild_id"
  seq_col <- ".ild_seq"
  dt_col <- ".ild_dt"
  time_num_col <- ".ild_time_num"
  out <- x
  created_nms <- character(0)
  for (v in vars) {
    if (!v %in% names(out)) stop("Variable '", v, "' not found in data.", call. = FALSE)
    lag_nm <- if (mode == "time_window") paste0(v, "_lag_window") else paste0(v, "_lag", n)
    created_nms <- c(created_nms, lag_nm)
    out <- dplyr::group_by(out, .data[[id_col]])
    if (mode == "index") {
      out <- dplyr::mutate(out, !!lag_nm := dplyr::lag(.data[[v]], n = n, default = NA))
    } else if (mode == "gap_aware") {
      out <- dplyr::mutate(out,
        !!lag_nm := ild_lag_gap_aware(.data[[v]], .data[[seq_col]], .data[[dt_col]], n, max_gap)
      )
    } else {
      out <- dplyr::mutate(out,
        !!lag_nm := ild_lag_time_window(.data[[v]], .data[[time_num_col]], window, resolution)
      )
    }
    out <- dplyr::ungroup(out)
  }
  out <- restore_ild_attrs(x, out)
  out <- ild_add_step(out, "ild_lag",
    list(vars = vars, n = n, mode = mode, max_gap = max_gap, window = window, resolution = resolution),
    list(created = created_nms)
  )
  out
}

#' For one person's vector: lag by n rows, NA if any dt in the window > max_gap
#' @noRd
ild_lag_gap_aware <- function(val, seq, dt, n, max_gap) {
  len <- length(val)
  out <- rep(NA, len)
  for (i in (n + 1):len) {
    # intervals between row i-n and i: dt[i-n+1], ..., dt[i]
    start <- i - n
    dts <- dt[(start + 1):i]
    if (all(!is.na(dts)) && all(dts <= max_gap)) {
      out[i] <- val[start]
    }
  }
  out
}

#' For one person: time-window lag. For each row, observations in (time - window, time]; apply resolution.
#' @noRd
ild_lag_time_window <- function(val, time_num, window, resolution) {
  len <- length(val)
  out <- rep(NA_real_, len)
  for (i in seq_len(len)) {
    t_i <- time_num[i]
    if (is.na(t_i)) next
    lo <- t_i - window
    in_window <- which(time_num > lo & time_num < t_i)
    if (length(in_window) == 0) next
    if (resolution == "closest_prior" || resolution == "last_in_window") {
      idx <- in_window[which.max(time_num[in_window])]
      out[i] <- val[idx]
    } else {
      out[i] <- mean(val[in_window], na.rm = TRUE)
    }
  }
  out
}

#' Convert window to numeric (seconds for POSIXct). Accepts lubridate Period/Duration.
#' @param window Numeric, or Period, or Duration.
#' @param x ILD object (to infer time type if needed); currently only used if window is not numeric.
#' @noRd
ild_window_to_num <- function(window, x) {
  if (is.numeric(window)) return(as.numeric(window)[1L])
  if (inherits(window, "Period") || inherits(window, "Duration")) {
    return(as.numeric(lubridate::as.duration(window)))
  }
  as.numeric(window)[1]
}

#' Autocorrelation function for ILD variables or model residuals
#'
#' Computes ACF on a variable in ILD data or on residuals from an [ild_lme()] fit.
#' Use this to check whether AR1 is appropriate before fitting models. ACF is
#' computed over the ordered observation sequence (pooled or within person); it
#' does not adjust for irregular time gaps.
#'
#' @param x Either an ILD object (see [is_ild()]) or a fitted model from [ild_lme()].
#' @param ... When \code{x} is ILD data, the variable(s) to compute ACF on (tidy-select; one variable). Ignored when \code{x} is a fit.
#' @param by_id Logical. If \code{TRUE}, also return per-person ACF in \code{acf_by_id} (default \code{FALSE}).
#' @return A list with \code{acf}: a tibble with columns \code{lag} and \code{acf} (pooled). If \code{by_id = TRUE}, \code{acf_by_id} is a named list of tibbles (one per person).
#' @export
#' @examples
#' d <- ild_simulate(n_id = 5, n_obs_per = 10, seed = 1)
#' x <- ild_prepare(d, id = "id", time = "time")
#' ild_acf(x, "y")
#' fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
#' ild_acf(fit)
#' @importFrom dplyr select
ild_acf <- function(x, ..., by_id = FALSE) {
  is_fit <- inherits(x, "lmerMod") || inherits(x, "lme")
  if (is_fit && !is.null(attr(x, "ild_data", exact = TRUE))) {
    data_ref <- augment_ild_model(x)
    id_col <- ".ild_id"
    res <- data_ref$.resid
    ids <- unique(data_ref[[id_col]])
  } else if (is_fit) {
    stop("Fit was not produced by tidyILD; refit using ild_lme() so the fit carries ild_data.",
         call. = FALSE)
  } else {
    validate_ild(x)
    if (length(list(...)) == 0) stop("No variable selected; provide a variable when x is ILD data (e.g. ild_acf(x, mood)).", call. = FALSE)
    data_ref <- x
    id_col <- ".ild_id"
    vars <- names(dplyr::select(x, ...))
    if (length(vars) == 0) stop("No variable selected; provide a variable when x is ILD data.", call. = FALSE)
    if (length(vars) > 1) stop("Select a single variable for ACF.", call. = FALSE)
    res <- x[[vars]]
    ids <- unique(x[[id_col]])
  }
  pooled_tbl <- ild_acf_vec(res[!is.na(res)])
  out <- list(acf = pooled_tbl)
  if (by_id) {
    by_id_tbls <- vector("list", length(ids))
    names(by_id_tbls) <- as.character(ids)
    for (i in seq_along(ids)) {
      r <- res[data_ref[[id_col]] == ids[i]]
      r <- r[!is.na(r)]
      by_id_tbls[[i]] <- ild_acf_vec(r)
    }
    out$acf_by_id <- by_id_tbls
  }
  out
}

#' ACF of a numeric vector as tibble (lag, acf)
#' @noRd
ild_acf_vec <- function(v) {
  if (length(v) <= 2) return(tibble::tibble(lag = numeric(), acf = numeric()))
  acf_obj <- stats::acf(v, plot = FALSE, na.action = stats::na.pass)
  tibble::tibble(lag = as.vector(acf_obj$lag), acf = as.vector(acf_obj$acf))
}

#' ILD-specific plots
#'
#' Produces trajectory (spaghetti), heatmap, gaps, and (if a fitted model
#' is provided) fitted vs actual and residual ACF. Works for both lmerMod and
#' lme (ild_lme with ar1 = TRUE).
#'
#' @param x An ILD tibble or a fitted [ild_lme()] model.
#' @param type Character (or vector). One or more of: `"trajectory"`, `"heatmap"`,
#'   `"gaps"`, `"missingness"`, `"fitted"` or `"fitted_vs_actual"` (requires fitted model),
#'   `"predicted_trajectory"` (observed and fitted lines vs time; requires fitted model),
#'   `"residual_acf"` (requires fitted model; ACF is over observation sequence, not adjusted for
#'   irregular time gaps). If length > 1, returns a named list of ggplots.
#' @param var For `trajectory` or `heatmap`, the variable to plot (optional;
#'   if missing and only one non-.ild_* column exists, it is used).
#' @param id_var For trajectory, variable used for grouping (default `.ild_id`).
#' @param time_var For trajectory/gaps/`predicted_trajectory`, x-axis: `.ild_time_num` or `.ild_seq`.
#' @param max_ids For trajectory and `predicted_trajectory`, max number of persons to plot (sampled if
#'   larger; default 20). Set to `Inf` to plot all.
#' @param seed Integer. Seed for sampling ids when `max_ids` is set (default 42).
#' @param facet_by Optional character: name of a column in the ILD (e.g. cluster or site) to pass to
#'   [ggplot2::facet_wrap()] for `trajectory`, `heatmap`, `gaps`, and `predicted_trajectory`.
#' @param ... Unused.
#' @return A single ggplot when `length(type) == 1`, or a named list of ggplots when `length(type) > 1`.
#' @examples
#' x <- ild_prepare(ild_simulate(n_id = 3, n_obs_per = 6, seed = 1), id = "id", time = "time")
#' fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
#' ild_plot(fit, type = "fitted_vs_actual")
#' ild_plot(fit, type = c("fitted_vs_actual", "residual_acf"))
#' @importFrom ggplot2 aes geom_line geom_point geom_tile geom_abline labs theme_minimal ggplot vars
#' @importFrom ggplot2 facet_wrap
#' @importFrom rlang .data sym
#' @export
ild_plot <- function(x,
                     type = c(
                       "trajectory", "heatmap", "gaps", "missingness", "fitted", "fitted_vs_actual",
                       "predicted_trajectory", "residual_acf"
                     ),
                     var = NULL,
                     id_var = ".ild_id",
                     time_var = c(".ild_time_num", ".ild_seq"),
                     max_ids = 20L,
                     seed = 42L,
                     facet_by = NULL,
                     ...) {
  type <- match.arg(type, several.ok = TRUE)
  time_var <- match.arg(time_var)
  fit_types <- c("fitted", "fitted_vs_actual", "predicted_trajectory", "residual_acf")
  if (any(type %in% fit_types)) {
    data <- if (inherits(x, "ild_lme") || !is.null(attr(x, "ild_data", exact = TRUE))) attr(x, "ild_data", exact = TRUE) else NULL
    if (is.null(data)) stop("Fitted model is missing ild_data attribute.", call. = FALSE)
  } else {
    data <- if (inherits(x, "ild_lme") || !is.null(attr(x, "ild_data", exact = TRUE))) attr(x, "ild_data", exact = TRUE) else x
  }
  if (!is.null(data)) validate_ild(data)
  ild_cols <- c(".ild_id", ".ild_time", ".ild_time_num", ".ild_seq", ".ild_dt", ".ild_gap")
  data_cols <- setdiff(names(data), ild_cols)
  if (any(type %in% c("trajectory", "heatmap")) && is.null(var)) {
    if (length(data_cols) == 0) stop("No data columns to plot.", call. = FALSE)
    if (length(data_cols) > 1) stop("Specify 'var' when data has multiple non-.ild_* columns.", call. = FALSE)
    var <- data_cols[1]
  }
  one_plot <- function(t) {
    if (t == "fitted_vs_actual") t <- "fitted"
    if (t == "missingness") {
      miss_vars <- if (is.null(var)) data_cols else var
      if (length(miss_vars) == 0) stop("No data columns for missingness plot.", call. = FALSE)
      return(ild_plot_missingness(data, miss_vars, id_var, time_var))
    }
    switch(t,
      trajectory = ild_plot_trajectory(data, var, id_var, time_var, max_ids, seed, facet_by),
      heatmap = ild_plot_heatmap(data, var, id_var, time_var, facet_by),
      gaps = ild_plot_gaps(data, id_var, time_var, facet_by),
      fitted = ild_plot_fitted(x, data),
      predicted_trajectory = .ild_plot_predicted_trajectory_impl(x, data, time_var, max_ids, seed, facet_by),
      residual_acf = ild_plot_residual_acf(x, data)
    )
  }
  if (length(type) == 1) {
    if (type == "missingness") {
      miss_vars <- if (is.null(var)) data_cols else var
      if (length(miss_vars) == 0) stop("No data columns for missingness plot.", call. = FALSE)
      return(ild_plot_missingness(data, miss_vars, id_var, time_var))
    }
    plot_type <- if (type == "fitted_vs_actual") "fitted" else type
    return(switch(plot_type,
      trajectory = ild_plot_trajectory(data, var, id_var, time_var, max_ids, seed, facet_by),
      heatmap = ild_plot_heatmap(data, var, id_var, time_var, facet_by),
      gaps = ild_plot_gaps(data, id_var, time_var, facet_by),
      fitted = ild_plot_fitted(x, data),
      predicted_trajectory = .ild_plot_predicted_trajectory_impl(x, data, time_var, max_ids, seed, facet_by),
      residual_acf = ild_plot_residual_acf(x, data)
    ))
  }
  out <- lapply(type, one_plot)
  names(out) <- type
  out
}

.ild_plot_add_facet <- function(p, data, facet_by) {
  if (is.null(facet_by) || !nzchar(as.character(facet_by)[1L])) {
    return(p)
  }
  fb <- as.character(facet_by)[1L]
  if (!fb %in% names(data)) {
    stop("facet_by column '", fb, "' not found in data.", call. = FALSE)
  }
  p + ggplot2::facet_wrap(ggplot2::vars(!!rlang::sym(fb)))
}

ild_plot_trajectory <- function(data, var, id_var, time_var, max_ids, seed, facet_by = NULL) {
  ids <- unique(data[[id_var]])
  if (length(ids) > max_ids && is.finite(max_ids)) {
    set.seed(seed)
    ids <- sample(ids, max_ids)
    data <- data[data[[id_var]] %in% ids, ]
  }
  p <- ggplot2::ggplot(data, ggplot2::aes(x = .data[[time_var]], y = .data[[var]], group = .data[[id_var]])) +
    ggplot2::geom_line(alpha = 0.7) +
    ggplot2::geom_point(alpha = 0.5, size = 1) +
    ggplot2::labs(x = "Time", y = var, title = "Trajectories") +
    ggplot2::theme_minimal()
  .ild_plot_add_facet(p, data, facet_by)
}

ild_plot_heatmap <- function(data, var, id_var, time_var, facet_by = NULL) {
  p <- ggplot2::ggplot(data, ggplot2::aes(x = .data[[time_var]], y = factor(.data[[id_var]]), fill = .data[[var]])) +
    ggplot2::geom_tile() +
    ggplot2::labs(x = "Time", y = "Person", fill = var, title = "Heatmap") +
    ggplot2::theme_minimal()
  .ild_plot_add_facet(p, data, facet_by)
}

ild_plot_gaps <- function(data, id_var, time_var, facet_by = NULL) {
  p <- ggplot2::ggplot(data, ggplot2::aes(x = .data[[time_var]], y = .data[[".ild_dt"]], color = .data[[id_var]])) +
    ggplot2::geom_point(alpha = 0.7) +
    ggplot2::labs(x = "Time", y = "Interval", title = "Intervals (gaps)") +
    ggplot2::theme_minimal()
  .ild_plot_add_facet(p, data, facet_by)
}

.ild_plot_predicted_trajectory_impl <- function(fit, data, time_var, max_ids, seed, facet_by = NULL) {
  aug <- if (inherits(fit, "brmsfit")) ild_augment(fit) else augment_ild_model(fit)
  if (nrow(aug) != nrow(data)) {
    stop(
      "Augmented rows must match ILD rows; refit with ild_lme() / ild_brms() on the same ILD object.",
      call. = FALSE
    )
  }
  if (!time_var %in% names(data)) {
    stop("time_var '", time_var, "' not found in ILD data.", call. = FALSE)
  }
  df <- data.frame(
    .ild_id = aug$.ild_id,
    time = data[[time_var]],
    observed = aug$.outcome,
    fitted = aug$.fitted,
    stringsAsFactors = FALSE
  )
  if (!is.null(facet_by)) {
    fb <- as.character(facet_by)[1L]
    if (!fb %in% names(data)) {
      stop("facet_by column '", fb, "' not found in data.", call. = FALSE)
    }
    df[[fb]] <- data[[fb]]
  }
  ids <- unique(df$.ild_id)
  if (length(ids) > max_ids && is.finite(max_ids)) {
    set.seed(seed)
    keep <- sample(ids, max_ids)
    df <- df[df$.ild_id %in% keep, , drop = FALSE]
  }
  dfl <- rbind(
    data.frame(df, series = "observed", value = df$observed, stringsAsFactors = FALSE),
    data.frame(df, series = "fitted", value = df$fitted, stringsAsFactors = FALSE)
  )
  p <- ggplot2::ggplot(
    dfl,
    ggplot2::aes(
      x = .data$time,
      y = .data$value,
      group = interaction(.data$.ild_id, .data$series),
      color = .data$series
    )
  ) +
    ggplot2::geom_line(alpha = 0.65, linewidth = 0.4) +
    ggplot2::labs(
      x = time_var,
      y = NULL,
      color = NULL,
      title = "Observed and fitted vs time"
    ) +
    ggplot2::theme_minimal()
  if (!is.null(facet_by) && nzchar(as.character(facet_by)[1L])) {
    fb <- as.character(facet_by)[1L]
    p <- p + ggplot2::facet_wrap(ggplot2::vars(!!rlang::sym(fb)))
  }
  p
}

ild_plot_missingness <- function(data, vars, id_var, time_var) {
  vars <- intersect(vars, names(data))
  if (length(vars) == 0) stop("No matching columns for missingness plot.", call. = FALSE)
  parts <- lapply(vars, function(v) {
    data.frame(
      id = data[[id_var]],
      time = data[[time_var]],
      variable = v,
      missing = is.na(data[[v]]),
      stringsAsFactors = FALSE
    )
  })
  long <- do.call(rbind, parts)
  p <- ggplot2::ggplot(long, ggplot2::aes(x = .data$time, y = factor(.data$id), fill = .data$missing)) +
    ggplot2::geom_tile() +
    ggplot2::scale_fill_manual(values = c("FALSE" = "gray90", "TRUE" = "darkred"), labels = c("observed", "missing")) +
    ggplot2::labs(x = "Time", y = "Person", fill = NULL, title = "Missingness") +
    ggplot2::theme_minimal()
  if (length(vars) > 1) p <- p + ggplot2::facet_wrap(ggplot2::vars(.data$variable), ncol = 1)
  p
}

ild_plot_fitted <- function(fit, data) {
  aug <- if (inherits(fit, "brmsfit")) ild_augment(fit) else augment_ild_model(fit)
  ggplot2::ggplot(aug, ggplot2::aes(x = .data$.outcome, y = .data$.fitted, color = .data$.ild_id)) +
    ggplot2::geom_point(alpha = 0.7) +
    ggplot2::geom_abline(slope = 1, intercept = 0, linetype = 2) +
    ggplot2::labs(x = "Observed", y = "Fitted", title = "Fitted vs Observed") +
    ggplot2::theme_minimal()
}

ild_plot_residual_acf <- function(fit, data) {
  diag <- ild_diagnostics(fit, data = data, by_id = FALSE, type = "residual_acf")
  acf_pooled <- diag$stats$acf$pooled
  if (is.null(acf_pooled) || nrow(acf_pooled) == 0) {
    return(ggplot2::ggplot() + ggplot2::theme_minimal() + ggplot2::labs(title = "ACF (insufficient data)"))
  }
  ggplot2::ggplot(acf_pooled, ggplot2::aes(x = .data$lag, y = .data$acf)) +
    ggplot2::geom_col(width = 0.1) +
    ggplot2::geom_hline(yintercept = 0) +
    ggplot2::labs(x = "Lag", y = "ACF", title = "Residual ACF") +
    ggplot2::theme_minimal()
}

#' ILD heatmap (alias for ild_plot with type = "heatmap")
#'
#' Person x time heatmap of a variable. See [ild_plot()].
#' @param x ILD object or fitted model (for heatmap, data are taken from ild_data if model).
#' @param var Variable to plot. If NULL, single data column is used.
#' @param facet_by Optional column in ILD for [ggplot2::facet_wrap()].
#' @param ... Passed to [ild_plot()] (e.g. id_var, time_var).
#' @return A ggplot object.
#' @export
ild_heatmap <- function(x, var = NULL, facet_by = NULL, ...) {
  ild_plot(x, type = "heatmap", var = var, facet_by = facet_by, ...)
}

#' ILD spaghetti / person trajectories (alias for ild_plot with type = "trajectory")
#'
#' Line plot of variable over time, one line per person. See [ild_plot()].
#' @param x ILD object or fitted model.
#' @param var Variable to plot. If NULL, single data column is used.
#' @param facet_by Optional column in ILD for [ggplot2::facet_wrap()] (e.g. cluster).
#' @param ... Passed to [ild_plot()] (e.g. max_ids, seed, id_var, time_var).
#' @return A ggplot object.
#' @export
ild_spaghetti <- function(x, var = NULL, facet_by = NULL, ...) {
  ild_plot(x, type = "trajectory", var = var, facet_by = facet_by, ...)
}

#' Observed and fitted values vs time (trajectory overlay)
#'
#' Uses [augment_ild_model()] or [ild_augment()] on the fitted object and plots
#' two lines per person (observed vs fitted) against `time_var`. For a scatter of
#' observed vs fitted, use [ild_plot()] with `type = "fitted"`.
#'
#' @param fit Model from [ild_lme()] or [ild_brms()] (must carry `ild_data`).
#' @param time_var `.ild_time_num` or `.ild_seq` (default first of these in [match.arg()]).
#' @param max_ids,seed Passed through for subsampling persons.
#' @param facet_by Optional column name in `ild_data` for [ggplot2::facet_wrap()].
#' @return A `ggplot` object.
#' @export
ild_plot_predicted_trajectory <- function(fit,
                                          time_var = c(".ild_time_num", ".ild_seq"),
                                          max_ids = 20L,
                                          seed = 42L,
                                          facet_by = NULL) {
  time_var <- match.arg(time_var)
  data <- attr(fit, "ild_data", exact = TRUE)
  if (is.null(data)) {
    stop("Fit must have ild_data (use ild_lme() or ild_brms()).", call. = FALSE)
  }
  validate_ild(data)
  ild_plot(
    fit,
    type = "predicted_trajectory",
    time_var = time_var,
    max_ids = max_ids,
    seed = seed,
    facet_by = facet_by
  )
}

test_that("ild_diagnostics returns expected structure", {
  d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  diag <- ild_diagnostics(fit, data = x)
  expect_true(all(c("meta", "data", "stats") %in% names(diag)))
  expect_equal(nrow(diag$data$residuals), nrow(x))
  expect_true(all(c(".ild_id", ".ild_time", ".resid", ".fitted") %in% names(diag$data$residuals)))
  expect_true("type" %in% names(diag$meta))
  plots <- plot_ild_diagnostics(diag)
  expect_true(is.list(plots))
  expect_true("qq" %in% names(plots))
})

test_that("ild_plot trajectory runs without error", {
  d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  p <- ild_plot(x, type = "trajectory", var = "y")
  expect_s3_class(p, "ggplot")
})

test_that("ild_plot trajectory facet_by runs when column exists", {
  d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 2)
  d$grp <- rep(c("a", "b"), length.out = nrow(d))
  x <- ild_prepare(d, id = "id", time = "time")
  p <- ild_plot(x, type = "trajectory", var = "y", facet_by = "grp", max_ids = 4L)
  expect_s3_class(p, "ggplot")
})

test_that("ild_plot facet_by errors when column missing", {
  d <- ild_simulate(n_id = 2, n_obs_per = 4, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  expect_error(
    ild_plot(x, type = "trajectory", var = "y", facet_by = "nosuch"),
    "facet_by column"
  )
})

test_that("ild_plot_predicted_trajectory and type predicted_trajectory return ggplot", {
  d <- ild_simulate(n_id = 4, n_obs_per = 6, seed = 3)
  d$grp <- rep(letters[1:2], length.out = nrow(d))
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  p1 <- ild_plot_predicted_trajectory(fit, time_var = ".ild_seq", max_ids = 4L)
  expect_s3_class(p1, "ggplot")
  p2 <- ild_plot(fit, type = "predicted_trajectory", time_var = ".ild_seq", max_ids = 4L, facet_by = "grp")
  expect_s3_class(p2, "ggplot")
})

test_that("ild_plot gaps runs without error", {
  d <- ild_simulate(n_id = 2, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  p <- ild_plot(x, type = "gaps")
  expect_s3_class(p, "ggplot")
})

test_that("ild_plot fitted runs with ild_lme object", {
  d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  p <- ild_plot(fit, type = "fitted")
  expect_s3_class(p, "ggplot")
})

test_that("ild_plot residual_acf runs with ild_lme object", {
  d <- ild_simulate(n_id = 3, n_obs_per = 8, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  p <- ild_plot(fit, type = "residual_acf")
  expect_s3_class(p, "ggplot")
})

test_that("ild_plot missingness runs and returns ggplot", {
  d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 1)
  d$y[2] <- NA
  x <- ild_prepare(d, id = "id", time = "time")
  p <- ild_plot(x, type = "missingness", var = "y")
  expect_s3_class(p, "ggplot")
})

test_that("ild_diagnostics type = residual_acf returns ACF; lme includes ar1_param", {
  d <- ild_simulate(n_id = 3, n_obs_per = 8, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  diag_lmer <- ild_diagnostics(ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE), data = x, type = "residual_acf")
  expect_true(!is.null(diag_lmer$stats$acf$pooled))
  expect_true("residual_acf" %in% names(plot_ild_diagnostics(diag_lmer)))
  fit_lme <- ild_lme(y ~ 1, data = x, ar1 = TRUE)
  diag_lme <- ild_diagnostics(fit_lme, type = "residual_acf")
  expect_true("ar1_param" %in% names(diag_lme$stats))
  expect_true(is.numeric(diag_lme$stats$ar1_param) || is.null(diag_lme$stats$ar1_param))
})

test_that("ild_diagnostics type = residual_time returns residuals vs time and plot", {
  d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  diag <- ild_diagnostics(fit, data = x, type = "residual_time")
  expect_true(".fitted" %in% names(diag$data$residuals))
  plots <- plot_ild_diagnostics(diag)
  expect_true("residuals_vs_time" %in% names(plots))
  expect_true("residuals_vs_fitted" %in% names(plots))
})

test_that("ild_diagnostics type = qq returns Q-Q plot", {
  d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  diag <- ild_diagnostics(fit, data = x, type = "qq")
  expect_true("qq" %in% names(plot_ild_diagnostics(diag)))
  expect_true(is.null(diag$stats$acf))
})

test_that("ild_diagnostics type = c(residual_acf, qq) returns only those plot elements and acf", {
  d <- ild_simulate(n_id = 3, n_obs_per = 6, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  diag <- ild_diagnostics(fit, data = x, type = c("residual_acf", "qq"))
  expect_true(!is.null(diag$stats$acf))
  plots <- plot_ild_diagnostics(diag)
  expect_true("residual_acf" %in% names(plots))
  expect_true("qq" %in% names(plots))
  expect_false("residuals_vs_fitted" %in% names(plots))
  expect_false("residuals_vs_time" %in% names(plots))
})

test_that("ild_plot type = fitted_vs_actual works like fitted", {
  d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  p_fitted <- ild_plot(fit, type = "fitted")
  p_fva <- ild_plot(fit, type = "fitted_vs_actual")
  expect_s3_class(p_fva, "ggplot")
  expect_s3_class(p_fitted, "ggplot")
})

test_that("ild_plot type = c(fitted_vs_actual, residual_acf) returns list of two ggplots", {
  d <- ild_simulate(n_id = 3, n_obs_per = 8, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  plist <- ild_plot(fit, type = c("fitted_vs_actual", "residual_acf"))
  expect_true(is.list(plist))
  expect_equal(names(plist), c("fitted_vs_actual", "residual_acf"))
  expect_s3_class(plist$fitted_vs_actual, "ggplot")
  expect_s3_class(plist$residual_acf, "ggplot")
})

test_that("ild_plot fitted and residual_acf work with lme fit (ar1 = TRUE)", {
  d <- ild_simulate(n_id = 3, n_obs_per = 6, irregular = TRUE, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit_lme <- ild_lme(y ~ 1, data = x, ar1 = TRUE)
  p_fitted <- ild_plot(fit_lme, type = "fitted")
  p_acf <- ild_plot(fit_lme, type = "residual_acf")
  expect_s3_class(p_fitted, "ggplot")
  expect_s3_class(p_acf, "ggplot")
})

#' Residual diagnostics for an ILD model
#'
#' Computes residual ACF (by person and/or pooled), residual vs fitted,
#' residual vs time, and optional Q-Q. Use `type` to request only specific
#' diagnostics. For `ild_lme` models with `ar1 = TRUE`, the estimated AR/CAR
#' parameter is reported when `type` includes `"residual_acf"`.
#'
#' The return value follows a stable schema: `meta` (engine, ar1, id/time columns,
#' n_obs, n_id), `data$residuals` (tibble with `.ild_id`, `.ild_time`, response, `.resid`, `.fitted`),
#' and `stats` (e.g. `acf`, `ar1_param`). Plots are not stored in the object; use
#' [plot_ild_diagnostics()] to generate them from a diagnostics object. The column
#' \code{.resid} is always filled; \code{.fitted} is filled when it can be computed
#' without refitting, otherwise it is \code{NA} (same for both engines and all \code{type} values).
#'
#' Residual ACF is computed over the ordered observation sequence within person;
#' it does not adjust for irregular time gaps.
#'
#' @param object A fitted model from [ild_lme()] (or an object with
#'   `residuals()`, and optional `fitted()`; if not `ild_lme`, pass `data` with
#'   `.ild_id` and `.ild_time_num` or `.ild_seq`).
#' @param data Optional. ILD data (required if `object` is not from [ild_lme()]).
#' @param type Character vector. Which diagnostics to compute: `"residual_acf"`,
#'   `"residual_time"` (residuals vs time and vs fitted), `"qq"`. Default is all three.
#' @param by_id Logical. If `TRUE`, compute ACF within each person (default `TRUE`).
#' @param ... Unused.
#' @return A list of class `ild_diagnostics` with: `meta` (engine, ar1, id_col, time_col,
#'   n_obs, n_id, type, by_id), `data` (list with `residuals` = tibble of .ild_id, .ild_time,
#'   response (name from formula), .resid, .fitted; \code{data$residuals} always exists, \code{.resid} is always filled,
#'   \code{.fitted} is returned when it can be computed without refitting, otherwise NA),
#'   `stats` (list with `acf` = list(pooled = tibble, by_id = list) when requested,
#'   `ar1_param` = numeric or NULL for lme). Use [plot_ild_diagnostics()] for plots.
#' @examples
#' x <- ild_prepare(ild_simulate(n_id = 3, n_obs_per = 6, seed = 1), id = "id", time = "time")
#' fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
#' diag <- ild_diagnostics(fit, type = c("residual_acf", "qq"))
#' plot_ild_diagnostics(diag)
#' @importFrom ggplot2 aes geom_point geom_hline geom_col labs theme_minimal ggplot stat_qq stat_qq_line
#' @importFrom stats coef
#' @export
ild_diagnostics <- function(object, data = NULL, type = c("residual_acf", "residual_time", "qq"), by_id = TRUE, ...) {
  type <- match.arg(type, choices = c("residual_acf", "residual_time", "qq"), several.ok = TRUE)
  if (inherits(object, "ild_lme") || !is.null(attr(object, "ild_data", exact = TRUE))) {
    data <- attr(object, "ild_data", exact = TRUE)
    if (is.null(data)) stop("Model object missing ild_data attribute.", call. = FALSE)
  }
  if (is.null(data)) stop("Provide 'data' (ILD object) for diagnostics.", call. = FALSE)
  validate_ild(data)
  meta_ild <- ild_meta(data)
  id_col <- ".ild_id"
  if (!is.null(attr(object, "ild_data", exact = TRUE))) {
    data_resid <- augment_ild_model(object)
  } else {
    res <- stats::residuals(object)
    if (length(res) != nrow(data)) stop("Residual length does not match data.", call. = FALSE)
    f <- tryCatch(stats::fitted(object), error = function(e) NULL)
    if (is.null(f) || length(f) != nrow(data)) f <- rep(NA_real_, nrow(data))
    mf <- tryCatch(stats::model.frame(object, data = data), error = function(e) stats::model.frame(object))
    y <- tryCatch(stats::model.response(mf), error = function(e) rep(NA_real_, nrow(data)))
    out_name <- ild_response_name(object)
    data_resid <- tibble::tibble(
      .ild_id = data[[id_col]],
      .ild_time = data[[".ild_time"]],
      .fitted = f,
      .resid = res
    )
    data_resid[[out_name]] <- y
    data_resid <- data_resid[c(".ild_id", ".ild_time", out_name, ".fitted", ".resid")]
  }
  res <- data_resid$.resid
  engine <- if (inherits(object, "lme")) "lme" else "lmer"
  ar1_flag <- isTRUE(attr(object, "ild_ar1", exact = TRUE))
  n_obs <- nrow(data)
  n_id <- length(unique(data[[id_col]]))
  meta <- list(
    engine = engine,
    ar1 = ar1_flag,
    id_col = meta_ild$ild_id,
    time_col = meta_ild$ild_time,
    n_obs = n_obs,
    n_id = n_id,
    type = type,
    by_id = by_id
  )
  stats <- list(acf = NULL, ar1_param = NULL)
  ar1_param <- NULL
  if ("residual_acf" %in% type && inherits(object, "ild_lme") && ar1_flag && inherits(object, "lme")) {
    cs <- object$modelStruct$corStruct
    if (!is.null(cs)) {
      phi <- tryCatch(coef(cs, unconstrained = FALSE)[[1]], error = function(e) NULL)
      if (!is.null(phi)) ar1_param <- phi
    }
  }
  stats$ar1_param <- ar1_param
  if ("residual_acf" %in% type) {
    r_pooled <- res[!is.na(res)]
    pooled_tbl <- if (length(r_pooled) > 2) {
      acf_obj <- stats::acf(r_pooled, plot = FALSE, na.action = stats::na.pass)
      tibble::tibble(lag = as.vector(acf_obj$lag), acf = as.vector(acf_obj$acf))
    } else {
      tibble::tibble(lag = numeric(), acf = numeric())
    }
    by_id_tbls <- NULL
    if (by_id) {
      ids <- unique(data[[id_col]])
      by_id_tbls <- vector("list", length(ids))
      names(by_id_tbls) <- as.character(ids)
      for (i in seq_along(ids)) {
        idx <- data[[id_col]] == ids[i]
        r <- res[idx]
        r <- r[!is.na(r)]
        if (length(r) > 2) {
          acf_i <- stats::acf(r, plot = FALSE, na.action = stats::na.pass)
          by_id_tbls[[i]] <- tibble::tibble(lag = as.vector(acf_i$lag), acf = as.vector(acf_i$acf))
        } else {
          by_id_tbls[[i]] <- tibble::tibble(lag = numeric(), acf = numeric())
        }
      }
    }
    stats$acf <- list(pooled = pooled_tbl, by_id = by_id_tbls)
  }
  out <- list(
    meta = meta,
    data = list(residuals = data_resid),
    stats = stats
  )
  class(out) <- "ild_diagnostics"
  attr(out, "ild_provenance") <- ild_new_analysis_provenance(object, "ild_diagnostics", list(
    type = type,
    by_id = by_id,
    engine = meta$engine
  ), list(n_obs = n_obs, n_residuals = nrow(data_resid)))
  out
}

#' Plot diagnostics from an ild_diagnostics object
#'
#' Generates ggplot objects for the requested diagnostic types. Plots are not
#' stored in the diagnostics object; call this function to create them.
#'
#' @param diag An object returned by [ild_diagnostics()].
#' @param type Character vector. Which plots to build (default: the types stored in \code{diag$meta$type}).
#' @return A named list of ggplot objects (e.g. \code{residual_acf}, \code{residuals_vs_fitted}, \code{residuals_vs_time}, \code{qq}).
#' @export
plot_ild_diagnostics <- function(diag, type = NULL) {
  if (!inherits(diag, "ild_diagnostics")) stop("diag must be an object from ild_diagnostics().", call. = FALSE)
  if (is.null(type)) type <- diag$meta$type
  type <- match.arg(type, choices = c("residual_acf", "residual_time", "qq"), several.ok = TRUE)
  dr <- diag$data$residuals
  out_plots <- list()
  if ("residual_acf" %in% type) {
    acf_pooled <- diag$stats$acf$pooled
    if (!is.null(acf_pooled) && nrow(acf_pooled) > 0) {
      p_acf <- ggplot2::ggplot(acf_pooled, ggplot2::aes(x = .data$lag, y = .data$acf)) +
        ggplot2::geom_col(width = 0.1) +
        ggplot2::geom_hline(yintercept = 0) +
        ggplot2::labs(x = "Lag", y = "ACF", title = "Residual ACF") +
        ggplot2::theme_minimal()
      out_plots$residual_acf <- p_acf
    } else {
      out_plots$residual_acf <- ggplot2::ggplot() + ggplot2::theme_minimal() +
        ggplot2::labs(title = "Residual ACF (insufficient data)")
    }
  }
  if ("residual_time" %in% type) {
    has_fitted <- !all(is.na(dr$.fitted))
    if (has_fitted) {
      p_res_fitted <- ggplot2::ggplot(dr, ggplot2::aes(x = .data$.fitted, y = .data$.resid)) +
        ggplot2::geom_point(ggplot2::aes(color = .data$.ild_id), alpha = 0.7) +
        ggplot2::geom_hline(yintercept = 0, linetype = 2) +
        ggplot2::labs(title = "Residuals vs Fitted", x = "Fitted", y = "Residuals") +
        ggplot2::theme_minimal()
      out_plots$residuals_vs_fitted <- p_res_fitted
    }
    p_res_time <- ggplot2::ggplot(dr, ggplot2::aes(x = .data$.ild_time, y = .data$.resid)) +
      ggplot2::geom_point(ggplot2::aes(color = .data$.ild_id), alpha = 0.7) +
      ggplot2::geom_hline(yintercept = 0, linetype = 2) +
      ggplot2::labs(title = "Residuals vs Time", x = "Time", y = "Residuals") +
      ggplot2::theme_minimal()
    out_plots$residuals_vs_time <- p_res_time
  }
  if ("qq" %in% type) {
    r <- dr$.resid[!is.na(dr$.resid)]
    if (length(r) > 0) {
      df_qq <- data.frame(residuals = r)
      p_qq <- ggplot2::ggplot(df_qq, ggplot2::aes(sample = .data$residuals)) +
        ggplot2::stat_qq() +
        ggplot2::stat_qq_line() +
        ggplot2::labs(title = "Q-Q plot (residuals)", x = "Theoretical", y = "Sample") +
        ggplot2::theme_minimal()
      out_plots$qq <- p_qq
    }
  }
  out_plots
}

#' @export
print.ild_diagnostics <- function(x, ...) {
  m <- x$meta
  cat("ILD diagnostics\n")
  cat("  Engine: ", m$engine, "\n", sep = "")
  cat("  AR1/CAR1: ", if (m$ar1) "yes" else "no", "\n", sep = "")
  cat("  Id column: ", m$id_col, "  Time column: ", m$time_col, "\n", sep = "")
  cat("  N persons: ", m$n_id, "  N observations: ", m$n_obs, "\n", sep = "")
  cat("  Types computed: ", paste(m$type, collapse = ", "), "\n", sep = "")
  if (!is.null(x$stats$ar1_param) && is.numeric(x$stats$ar1_param)) {
    cat("  Estimated correlation (phi): ", round(x$stats$ar1_param, 4), "\n", sep = "")
  }
  invisible(x)
}

# MSM / IPW inference: cluster bootstrap for weighted lmer fits

#' @title MSM and IPW inference for weighted mixed models
#' @name ild_msm_inference
#' @description
#' After [ild_ipw_refit()] (or any \code{lmer(..., weights =)} with \code{attr(fit, "ild_data")}),
#' uncertainty for MSM-style **estimands** should not rely on default \code{lmer} standard errors.
#' tidyILD supports:
#' \itemize{
#'   \item \strong{Cluster bootstrap} — [ild_msm_bootstrap()] resamples persons (clusters),
#'     refits the weighted model, and collects fixed effects; use
#'     \code{weight_policy = "reestimate_weights"} and a \code{weights_fn} to re-fit propensity
#'     weights on each replicate when that matches your estimand (computationally heavier;
#'     preferred for many MSM workflows). \code{weight_policy = "fixed_weights"} keeps the
#'     weights from the resampled rows (fast approximation; ignores first-stage uncertainty).
#'   \item \strong{Cluster-robust sandwich Wald} — [ild_robust_se()] / [tidy_ild_model(..., se = "robust")]
#'     apply \pkg{clubSandwich} at the **person** level on the **weighted outcome model only**; they
#'     do **not** account for estimation error in IPW/IPCW weights unless you treat weights as fixed.
#'   \item \strong{Bayesian} — Full joint Bayesian MSM is not built in. See [ild_brms()] for
#'     hierarchical outcome models; combining external IPW row weights with Stan is delicate;
#'     prefer explicit joint/sensitivity designs and specialized causal packages for Bayes MSM.
#' }
#' @seealso [ild_msm_bootstrap()], [tidy_ild_msm_bootstrap()], [ild_ipw_refit()],
#'   [ild_robust_se()], [tidy_ild_model()], [ild_brms()], [ild_msm_estimand()],
#'   [ild_msm_fit()]
#' @keywords internal
NULL

#' Cluster bootstrap inference for weighted \code{lmer} (MSM / IPW sensitivity)
#'
#' Resamples **clusters** (persons) with replacement, subsets \code{attr(fit, "ild_data")}
#' (or \code{data}), optionally re-estimates weights, and refits \code{\link[lme4]{lmer}} with
#' the same formula. Collects fixed effects across replicates for bootstrap standard errors and
#' percentile confidence intervals.
#'
#' @param fit A \code{lmerMod} from [ild_ipw_refit()] or [ild_lme()] with \code{attr(fit, "ild_data")}.
#'   Ignored if \code{formula} and \code{data} are supplied.
#' @param formula Optional. If \code{fit} is \code{NULL}, mixed-model formula for refitting.
#' @param data Optional ILD object when \code{fit} is \code{NULL}; must contain \code{weights_col}.
#' @param weights_col Name of the row-level weight column in each bootstrap dataset (default \code{".ipw"}).
#' @param n_boot Number of bootstrap replicates.
#' @param weight_policy \code{"fixed_weights"} (subset rows; weights unchanged within replicated rows)
#'   or \code{"reestimate_weights"} (apply \code{weights_fn} to each resampled dataset before fitting).
#' @param weights_fn Function \code{function(x)} taking ILD \code{x} and returning ILD with
#'   \code{weights_col} present. Required when \code{weight_policy = "reestimate_weights"}.
#'   Ignored for \code{"fixed_weights"}.
#' @param cluster \code{"id"} (default: cluster by \code{ild_meta(data)$ild_id}) or \code{"data"}.
#' @param cluster_vec When \code{cluster = "data"}, length-\code{nrow(data)} vector of cluster IDs
#'   aligned to \code{ild_data} rows.
#' @param coef_fun \code{function(fit)} returning named numeric vector of estimands (default \code{lme4::fixef}).
#' @param seed Optional integer passed to \code{set.seed()} before resampling.
#' @param verbose Logical. If \code{TRUE}, warn on failed refits.
#' @param ... Passed to \code{lme4::lmer()} on each successful replicate (e.g. \code{control}).
#'   Do not pass \code{weights} here; use \code{weights_col} on the ILD data.
#' @return An object of class \code{ild_msm_bootstrap}: \code{replicates} (matrix \code{n_boot} x \code{p}),
#'   \code{estimate} (point estimates from \code{fit}), \code{bootstrap_se}, \code{conf_low}, \code{conf_high},
#'   \code{term_names}, \code{n_boot}, \code{n_success}, \code{n_cluster}, metadata, and \code{fit} (original).
#'   Use [tidy_ild_msm_bootstrap()] for an \code{\link{ild_tidy_schema}} tibble.
#' @seealso [ild_msm_inference], [ild_ipw_refit()], [tidy_ild_msm_bootstrap()]
#' @export
#' @examples
#' if (requireNamespace("lme4", quietly = TRUE)) {
#'   set.seed(5001)
#'   d <- ild_simulate(n_id = 12, n_obs_per = 6, seed = 5001)
#'   d$stress <- rnorm(nrow(d))
#'   d <- ild_prepare(d, id = "id", time = "time")
#'   d <- ild_center(d, y)
#'   d$.ipw <- runif(nrow(d), 0.8, 1.2)
#'   f0 <- ild_lme(y ~ y_bp + y_wp + stress + (1 | id), data = d, ar1 = FALSE,
#'     warn_no_ar1 = FALSE, warn_uncentered = FALSE)
#'   fw <- ild_ipw_refit(f0, data = d, weights = ".ipw")
#'   b <- ild_msm_bootstrap(fw, n_boot = 15L, weight_policy = "fixed_weights", seed = 2)
#'   print(b)
#'   tidy_ild_msm_bootstrap(b)
#' }
ild_msm_bootstrap <- function(fit = NULL,
                              formula = NULL,
                              data = NULL,
                              weights_col = ".ipw",
                              n_boot = 200L,
                              weight_policy = c("fixed_weights", "reestimate_weights"),
                              weights_fn = NULL,
                              cluster = c("id", "data"),
                              cluster_vec = NULL,
                              coef_fun = function(f) lme4::fixef(f),
                              seed = NULL,
                              verbose = FALSE,
                              ...) {
  weight_policy <- match.arg(weight_policy)
  cluster <- match.arg(cluster)
  if (weight_policy == "reestimate_weights" && is.null(weights_fn)) {
    stop("weights_fn must be supplied when weight_policy = \"reestimate_weights\".", call. = FALSE)
  }
  if (!is.null(seed)) set.seed(seed)

  if (!is.null(fit)) {
    if (!inherits(fit, "lmerMod")) {
      stop("fit must be a lmerMod from ild_ipw_refit() or ild_lme().", call. = FALSE)
    }
    formula <- stats::formula(fit)
    ild_data <- attr(fit, "ild_data", exact = TRUE)
    if (is.null(ild_data)) {
      stop("fit must have attr(fit, \"ild_data\") (use ild_ipw_refit() / ild_lme()).", call. = FALSE)
    }
    validate_ild(ild_data)
  } else {
    if (is.null(formula) || is.null(data)) {
      stop("Supply either fit, or both formula and data.", call. = FALSE)
    }
    validate_ild(data)
    ild_data <- data
    fit <- NULL
  }

  if (!weights_col %in% names(ild_data)) {
    stop("weights_col '", weights_col, "' not found in ILD data.", call. = FALSE)
  }

  meta <- ild_meta(ild_data)
  cl_vec <- if (cluster == "id") {
    if (is.null(meta$ild_id) || !meta$ild_id %in% names(ild_data)) {
      stop("ild_data must contain id column from ild_meta().", call. = FALSE)
    }
    ild_data[[meta$ild_id]]
  } else {
    if (is.null(cluster_vec)) stop("When cluster = \"data\", supply cluster_vec.", call. = FALSE)
    if (length(cluster_vec) != nrow(ild_data)) {
      stop("cluster_vec length must match nrow(ild_data).", call. = FALSE)
    }
    cluster_vec
  }

  ucl <- unique(as.vector(cl_vec))
  n_cl <- length(ucl)
  if (n_cl < 2L && n_boot > 1L) {
    warning("Fewer than 2 clusters; bootstrap is degenerate.", call. = FALSE)
  }

  lmer_dots <- list(...)
  beta_point <- if (!is.null(fit)) {
    coef_fun(fit)
  } else {
    w0 <- ild_data[[weights_col]]
    ftmp <- do.call(
      lme4::lmer,
      c(list(formula = formula, data = ild_data, weights = w0), lmer_dots)
    )
    attr(ftmp, "ild_data") <- ild_data
    fit <- ftmp
    coef_fun(fit)
  }

  p_names <- names(beta_point)
  if (length(p_names) == 0L) stop("Could not determine coefficient names from fit.", call. = FALSE)

  n_boot <- as.integer(n_boot)[1L]
  if (!is.finite(n_boot) || n_boot < 1L) stop("n_boot must be a positive integer.", call. = FALSE)

  reps <- matrix(NA_real_, nrow = n_boot, ncol = length(p_names), dimnames = list(NULL, p_names))
  n_ok <- 0L

  for (b in seq_len(n_boot)) {
    sid <- sample(ucl, size = n_cl, replace = TRUE)
    idx <- unlist(lapply(sid, function(s) which(cl_vec == s)), use.names = FALSE)
    if (length(idx) == 0L) next
    data_b <- ild_data[idx, , drop = FALSE]
    attrs <- attributes(ild_data)
    data_b <- .ild_ipw_restore_attrs(data_b, attrs)

    if (weight_policy == "reestimate_weights") {
      data_b <- weights_fn(data_b)
      validate_ild(data_b)
      if (!weights_col %in% names(data_b)) {
        stop("weights_fn must return data with weights column '", weights_col, "'.", call. = FALSE)
      }
    }

    w_vec <- data_b[[weights_col]]
    fit_b <- tryCatch(
      do.call(
        lme4::lmer,
        c(list(formula = formula, data = data_b, weights = w_vec), lmer_dots)
      ),
      error = function(e) {
        if (verbose) {
          warning(
            "Bootstrap replicate ", b, " failed to fit: ", conditionMessage(e),
            call. = FALSE
          )
        }
        NULL
      }
    )
    if (is.null(fit_b)) next
    cb <- coef_fun(fit_b)
    vec <- rep(NA_real_, length(p_names))
    names(vec) <- p_names
    nm <- names(cb)
    vec[nm] <- as.numeric(unname(cb[nm]))
    reps[b, ] <- vec
    n_ok <- n_ok + 1L
  }

  q <- 0.025
  boot_se <- apply(reps, 2L, stats::sd, na.rm = TRUE)
  c_lo <- apply(reps, 2L, stats::quantile, probs = q, na.rm = TRUE, type = 7)
  c_hi <- apply(reps, 2L, stats::quantile, probs = 1 - q, na.rm = TRUE, type = 7)

  cluster_name <- if (cluster == "id") meta$ild_id else "user"

  out <- list(
    replicates = reps,
    estimate = stats::setNames(as.numeric(beta_point[p_names]), p_names),
    term_names = p_names,
    bootstrap_se = boot_se,
    conf_low = c_lo,
    conf_high = c_hi,
    n_boot = n_boot,
    n_success = n_ok,
    n_cluster = n_cl,
    weight_policy = weight_policy,
    weights_col = weights_col,
    cluster_name = cluster_name,
    formula = formula,
    fit = fit
  )
  class(out) <- c("ild_msm_bootstrap", "list")

  attr(out, "ild_provenance") <- ild_new_analysis_provenance(
    if (!is.null(fit)) fit else ild_data,
    "ild_msm_bootstrap",
    list(
      weights_col = weights_col,
      n_boot = n_boot,
      weight_policy = weight_policy,
      cluster = cluster
    ),
    list(n_success = n_ok, n_cluster = n_cl)
  )
  out
}

#' @export
print.ild_msm_bootstrap <- function(x, ...) {
  cat("tidyILD MSM bootstrap (weighted lmer)\n")
  cat("  replicates: ", x$n_boot, " (successful: ", x$n_success, ")\n", sep = "")
  cat("  clusters: ", x$n_cluster, " (by ", x$cluster_name, ")\n", sep = "")
  cat("  weight_policy: ", x$weight_policy, "\n", sep = "")
  cat("  weights_col: ", x$weights_col, "\n", sep = "")
  invisible(x)
}

#' Tidy fixed effects from \code{ild_msm_bootstrap}
#'
#' Returns a tibble matching \code{\link{ild_tidy_schema}} with \code{interval_type}
#' \code{bootstrap_percentile} (equal-tailed over replicate coefficients).
#'
#' @param x Object from [ild_msm_bootstrap()].
#' @param conf_level Numeric. Used for implied tail quantiles (default 0.95).
#' @param ... Unused.
#' @return A tibble; see \code{\link{ild_tidy_schema}}.
#' @seealso [ild_msm_bootstrap()], [ild_msm_inference]
#' @export
tidy_ild_msm_bootstrap <- function(x, conf_level = 0.95, ...) {
  if (!inherits(x, "ild_msm_bootstrap")) {
    stop("x must be from ild_msm_bootstrap().", call. = FALSE)
  }
  q <- (1 - conf_level) / 2
  term <- x$term_names
  est <- as.numeric(x$estimate)
  se <- as.numeric(x$bootstrap_se)
  # Recompute CIs for conf_level
  reps <- x$replicates
  c_lo <- apply(reps, 2L, stats::quantile, probs = q, na.rm = TRUE, type = 7)
  c_hi <- apply(reps, 2L, stats::quantile, probs = 1 - q, na.rm = TRUE, type = 7)
  pval <- rep(NA_real_, length(term))
  stat <- rep(NA_real_, length(term))
  ok <- is.finite(se) & se > 0
  stat[ok] <- est[ok] / se[ok]
  pval[ok] <- 2 * (1 - stats::pnorm(abs(stat[ok])))
  mc <- if (!is.null(x$fit)) ild_tidy_model_class_string(x$fit) else "lmerMod"
  ild_tidy_assemble(
    term = term,
    estimate = est,
    std_error = se,
    conf_low = c_lo,
    conf_high = c_hi,
    p_value = pval,
    statistic = stat,
    interval_type = rep("bootstrap_percentile", length(term)),
    engine = "lmer",
    model_class = mc,
    component = rep("fixed", length(term)),
    effect_level = NULL,
    optional = NULL
  )
}

msm_boot_data <- function(seed, n_id = 16L, n_obs = 8L) {
  set.seed(seed)
  d <- ild_simulate(n_id = n_id, n_obs_per = n_obs, seed = seed)
  ild_prepare(d, id = "id", time = "time")
}

msm_boot_ctl <- function() {
  lme4::lmerControl(
    optimizer = "bobyqa",
    check.conv.singular = "ignore",
    check.conv.hess = "ignore",
    check.conv.grad = "ignore"
  )
}

msm_boot_fit <- function(x, wcol = ".ipw") {
  x[[wcol]] <- runif(nrow(x), 0.92, 1.08)
  ctl <- msm_boot_ctl()
  f <- ild_lme(
    y ~ (1 | id),
    data = x,
    ar1 = FALSE,
    warn_no_ar1 = FALSE,
    warn_uncentered = FALSE,
    control = ctl
  )
  ild_ipw_refit(f, data = x, weights = wcol, control = ctl)
}

test_that("ild_msm_bootstrap fixed_weights returns expected structure", {
  x <- msm_boot_data(21L)
  fw <- msm_boot_fit(x)
  b <- ild_msm_bootstrap(
    fw,
    n_boot = 20L,
    weight_policy = "fixed_weights",
    seed = 71L,
    control = msm_boot_ctl()
  )
  expect_s3_class(b, "ild_msm_bootstrap")
  expect_equal(dim(b$replicates), c(20L, length(lme4::fixef(fw))))
  expect_equal(b$n_boot, 20L)
  expect_equal(b$n_success, 20L)
  expect_equal(names(b$estimate), b$term_names)
  prov <- attr(b, "ild_provenance", exact = TRUE)
  expect_true(is.list(prov))
  expect_equal(prov$object_type, "ild_msm_bootstrap")
  expect_true(any(vapply(prov$analysis_steps, function(s) identical(s$step, "ild_msm_bootstrap"), logical(1))))
})

test_that("tidy_ild_msm_bootstrap uses bootstrap_percentile interval_type", {
  x <- msm_boot_data(22L)
  fw <- msm_boot_fit(x)
  b <- ild_msm_bootstrap(
    fw,
    n_boot = 15L,
    seed = 32L,
    weight_policy = "fixed_weights",
    control = msm_boot_ctl()
  )
  expect_equal(b$n_success, 15L)
  td <- tidy_ild_msm_bootstrap(b, conf_level = 0.9)
  expect_s3_class(td, "tbl_df")
  expect_true(all(td$interval_type == "bootstrap_percentile"))
  expect_true(all(ild_tidy_schema()$required %in% names(td)))
})

test_that("ild_msm_bootstrap requires weights_fn for reestimate_weights", {
  x <- msm_boot_data(23L, n_id = 10L, n_obs = 6L)
  x$.ipw <- 1
  expect_error(
    ild_msm_bootstrap(
      NULL,
      formula = y ~ (1 | id),
      data = x,
      n_boot = 2L,
      weight_policy = "reestimate_weights",
      control = msm_boot_ctl()
    ),
    "weights_fn"
  )
})

test_that("ild_msm_bootstrap reestimate_weights runs with identity-like weights_fn", {
  x <- msm_boot_data(24L)
  fw <- msm_boot_fit(x)
  b <- ild_msm_bootstrap(
    fw,
    n_boot = 12L,
    seed = 51L,
    weight_policy = "reestimate_weights",
    weights_fn = function(z) z,
    control = msm_boot_ctl()
  )
  expect_equal(b$n_boot, 12L)
  expect_equal(b$n_success, 12L)
})

test_that("tidy_ild_model rejects ild_msm_bootstrap objects", {
  x <- msm_boot_data(25L, n_id = 12L, n_obs = 6L)
  fw <- msm_boot_fit(x)
  b <- ild_msm_bootstrap(
    fw,
    n_boot = 5L,
    seed = 11L,
    weight_policy = "fixed_weights",
    control = msm_boot_ctl()
  )
  expect_error(tidy_ild_model(b), "tidy_ild_msm_bootstrap")
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## ----recipe-random-lag-slope--------------------------------------------------
# library(tidyILD)
# set.seed(1)
# d <- ild_simulate(n_id = 15, n_obs_per = 12, seed = 1)
# d$x <- rnorm(nrow(d))
# x <- ild_prepare(d, id = "id", time = "time")
# x <- ild_lag(x, x, n = 1L, mode = "gap_aware")
# tpl <- ild_brms_dynamics_formula("y", "x_lag1", id_var = "id")
# tpl$formula
# tpl$notes
# # Short chains for illustration only:
# # fit <- ild_brms(tpl$formula, data = x, iter = 500, chains = 2, refresh = 0)
# # ild_tidy(fit)

## ----recipe-two-lags----------------------------------------------------------
# x2 <- ild_panel_lag_prepare(x, c("x", "y"), n = c(1L, 1L), mode = "gap_aware")
# names(x2$data)
# # Example fixed structure only (not run):
# # f2 <- y ~ x_lag1 + y_lag1 + (1 | id)
# # fit2 <- ild_brms(f2, data = x2$data, iter = 500, chains = 2, refresh = 0)

## ----recipe-mvbind------------------------------------------------------------
# # Not run — requires brms; illustration only:
# # library(brms)
# # f_mv <- bf(mvbind(mood, stress) ~ mood_lag1 + stress_lag1 + (1 | id)) + set_rescor(TRUE)
# # fit_mv <- brm(f_mv, data = x2$data, chains = 2, iter = 500, refresh = 0)
# # Build mood_lag1 / stress_lag1 with ild_lag() or ild_panel_lag_prepare() first.

## ----session-info, echo = FALSE-----------------------------------------------
# sessionInfo()

# Person-level compliance, cohort missingness over time, discrete first-event hazard.
# Used by ild_missing_pattern(..., outcome =) and ild_missingness_report().

#' @keywords internal
#' @noRd
.longest_run_true <- function(lgl) {
  if (!length(lgl)) {
    return(0L)
  }
  r <- rle(lgl)
  m <- 0L
  for (i in seq_along(r$values)) {
    if (isTRUE(r$values[i])) {
      m <- max(m, r$lengths[i], na.rm = TRUE)
    }
  }
  as.integer(m)
}

#' @keywords internal
#' @noRd
.monotone_missing_after_first_na <- function(o) {
  na <- is.na(o)
  if (!any(na)) {
    return(NA)
  }
  i <- min(which(na))
  all(na[i:length(na)])
}

#' Person-level missingness compliance for one outcome
#'
#' Per person (after sorting by `.ild_seq`), returns coverage on the outcome,
#' longest streak of observed values, whether missingness is monotone (dropout
#' pattern: once missing, stays missing), and optional comparison to an expected
#' number of occasions.
#'
#' @family ild_diagnostics_utilities
#' @param x An ILD object (see [is_ild()]).
#' @param outcome Character; column with possible `NA` (e.g. EMA outcome).
#' @param expected_occasions Optional positive integer: planned waves per person
#'   for adherence (`pct_of_expected`). If persons have different row counts,
#'   this is only a rough benchmark; see Details.
#' @return A tibble with columns `.ild_id`, `n_rows`, `n_obs_outcome`,
#'   `pct_nonmissing_outcome`, `longest_run_observed`, `monotone_missing`
#'   (`NA` if no missing values on the outcome for that person), and if
#'   `expected_occasions` is set, `pct_of_expected` and `meets_expected_rows`.
#' @details
#' **Monotone missing** is defined only when the person has at least one missing
#' outcome: all observations after the first missing are missing. Intermittent
#' missing yields `FALSE`. **Discrete hazard** for first missing is provided by
#' [ild_missing_hazard_first()] (cohort-level by `.ild_seq`).
#'
#' @seealso [ild_missing_pattern()], [ild_missingness_report()]
#' @importFrom dplyr arrange bind_rows
#' @importFrom rlang .data
#' @importFrom tibble tibble
#' @export
ild_missing_compliance <- function(x, outcome, expected_occasions = NULL) {
  validate_ild(x)
  if (!outcome %in% names(x)) {
    stop("outcome '", outcome, "' not found in data.", call. = FALSE)
  }
  eo <- NULL
  if (!is.null(expected_occasions)) {
    eo <- as.integer(expected_occasions)[1L]
    if (!is.finite(eo) || eo < 1L) {
      stop("expected_occasions must be a positive integer.", call. = FALSE)
    }
  }
  id_col <- ".ild_id"
  x <- dplyr::arrange(x, .data[[id_col]], .data$.ild_seq)
  sp <- split(as.data.frame(x, stringsAsFactors = FALSE), as.character(x[[id_col]]))
  rows <- lapply(sp, function(d) {
    d <- d[order(d$.ild_seq), , drop = FALSE]
    o <- d[[outcome]]
    n <- length(o)
    n_obs <- sum(!is.na(o))
    lr <- .longest_run_true(!is.na(o))
    mono <- .monotone_missing_after_first_na(o)
    row <- tibble::tibble(
      .ild_id = d[[id_col]][1L],
      n_rows = n,
      n_obs_outcome = as.integer(n_obs),
      pct_nonmissing_outcome = if (n > 0L) 100 * n_obs / n else NA_real_,
      longest_run_observed = lr,
      monotone_missing = mono
    )
    if (!is.null(eo)) {
      row$pct_of_expected <- 100 * n_obs / eo
      row$meets_expected_rows <- n >= eo
    }
    row
  })
  dplyr::bind_rows(rows)
}

#' Cohort-level fraction observed by occasion index
#'
#' For each distinct `.ild_seq` value, computes how many rows exist and what
#' fraction has a non-missing outcome. Optional line plot.
#'
#' @family ild_diagnostics_utilities
#' @param x An ILD object.
#' @param outcome Character; outcome column.
#' @param plot Logical; if `TRUE`, return a `ggplot2` object in `$plot`.
#' @return A list with `by_occasion` (tibble: `.ild_seq`, `n_rows`, `n_obs`,
#'   `pct_observed`) and if `plot = TRUE`, `plot`.
#' @details
#' Uses **within-study sequence** (`.ild_seq`), not calendar time. For irregular
#' timing, interpret as ordinal wave position rather than equal time spacing.
#'
#' @seealso [ild_missing_pattern()], [ild_missing_compliance()]
#' @importFrom dplyr group_by summarise
#' @importFrom rlang .data
#' @export
ild_missing_cohort <- function(x, outcome, plot = TRUE) {
  validate_ild(x)
  if (!outcome %in% names(x)) {
    stop("outcome '", outcome, "' not found in data.", call. = FALSE)
  }
  by_occ <- dplyr::group_by(x, .data$.ild_seq)
  by_occ <- dplyr::summarise(
    by_occ,
    n_rows = dplyr::n(),
    n_obs = sum(!is.na(.data[[outcome]])),
    pct_observed = 100 * sum(!is.na(.data[[outcome]])) / dplyr::n(),
    .groups = "drop"
  )
  by_occ <- by_occ[order(by_occ$.ild_seq), , drop = FALSE]
  p <- NULL
  if (isTRUE(plot)) {
    p <- ggplot2::ggplot(by_occ, ggplot2::aes(x = .data$.ild_seq, y = .data$pct_observed)) +
      ggplot2::geom_line(linewidth = 0.6) +
      ggplot2::geom_point(size = 1.5) +
      ggplot2::labs(
        x = "Occasion (.ild_seq)",
        y = "% non-missing outcome",
        title = "Cohort observed fraction by occasion"
      ) +
      ggplot2::ylim(0, 100) +
      ggplot2::theme_minimal()
  }
  list(by_occasion = tibble::as_tibble(by_occ), plot = p)
}

#' Discrete hazard of first missing outcome on an ordinal schedule
#'
#' For each `.ild_seq`, estimates the fraction of person-occasions **at risk**
#' (previous occasion observed, or first occasion) that are **missing** on the
#' outcome. This is a coarse discrete-time hazard for **first** missing spell
#' when missingness is intermittent; under **monotone dropout** it matches the
#' usual discrete hazard of dropout.
#'
#' @family ild_diagnostics_utilities
#' @param x An ILD object.
#' @param outcome Character; outcome column.
#' @return A tibble with `.ild_seq`, `n_at_risk`, `n_missing`, `hazard`.
#' @details
#' **Assumptions:** Rows are ordered within person by `.ild_seq`. The first row
#' per person is always at risk. Later rows are at risk only if the previous
#' row (same person) was non-missing—so a person who is already missing is not
#' counted again. This targets a **first event** narrative; it is not a full
#' repeated-events model.
#'
#' @seealso [ild_missing_cohort()], [ild_missing_compliance()]
#' @importFrom dplyr arrange group_by mutate ungroup summarise
#' @importFrom rlang .data
#' @export
ild_missing_hazard_first <- function(x, outcome) {
  validate_ild(x)
  if (!outcome %in% names(x)) {
    stop("outcome '", outcome, "' not found in data.", call. = FALSE)
  }
  xd <- dplyr::arrange(x, .data$.ild_id, .data$.ild_seq)
  xd <- dplyr::group_by(xd, .data$.ild_id)
  xd <- dplyr::mutate(
    xd,
    prev_obs = dplyr::lag(!is.na(.data[[outcome]]), default = TRUE),
    at_risk = .data$prev_obs,
    event = .data$at_risk & is.na(.data[[outcome]])
  )
  xd <- dplyr::ungroup(xd)
  hz <- dplyr::group_by(xd, .data$.ild_seq)
  hz <- dplyr::summarise(
    hz,
    n_at_risk = sum(.data$at_risk, na.rm = TRUE),
    n_missing = sum(.data$event, na.rm = TRUE),
    hazard = if (sum(.data$at_risk, na.rm = TRUE) > 0) {
      sum(.data$event, na.rm = TRUE) / sum(.data$at_risk, na.rm = TRUE)
    } else {
      NA_real_
    },
    .groups = "drop"
  )
  hz <- hz[order(hz$.ild_seq), , drop = FALSE]
  tibble::as_tibble(hz)
}

test_that("ild_missing_compliance monotone dropout and expected occasions", {
  d <- ild_simulate(n_id = 2, n_obs_per = 5, seed = 1201L)
  x <- ild_prepare(d, id = "id", time = "time")
  w1 <- x$.ild_id == 1L
  x$y[w1] <- NA
  x$y[x$.ild_id == 1L & x$.ild_seq <= 2L] <- 1
  cm <- ild_missing_compliance(x, outcome = "y", expected_occasions = 5L)
  expect_true(all(c("n_rows", "monotone_missing", "pct_of_expected") %in% names(cm)))
  m1 <- cm$monotone_missing[cm$.ild_id == 1L][1L]
  expect_true(isTRUE(m1))
  m2 <- cm$monotone_missing[cm$.ild_id == 2L][1L]
  expect_true(is.na(m2))
})

test_that("ild_missing_hazard_first sums to sensible risk set", {
  d <- ild_simulate(n_id = 20, n_obs_per = 8, seed = 1202L)
  x <- ild_prepare(d, id = "id", time = "time")
  hz <- ild_missing_hazard_first(x, outcome = "y")
  expect_true(all(hz$n_at_risk >= hz$n_missing, na.rm = TRUE))
  expect_true(all(hz$hazard <= 1, na.rm = TRUE))
})

test_that("ild_missing_pattern joins compliance when outcome set", {
  d <- ild_simulate(n_id = 8, n_obs_per = 10, seed = 1203L)
  d$y[sample(nrow(d), 12)] <- NA
  x <- ild_prepare(d, id = "id", time = "time")
  mp <- ild_missing_pattern(x, vars = "y", outcome = "y", expected_occasions = 10L)
  expect_true("pct_nonmissing_outcome" %in% names(mp$by_id))
  expect_true("longest_run_observed" %in% names(mp$by_id))
})

test_that("ild_missingness_report returns expected components", {
  d <- ild_simulate(n_id = 12, n_obs_per = 9, seed = 1204L)
  d$x <- rnorm(nrow(d))
  d$y[sample(nrow(d), 15)] <- NA
  x <- ild_prepare(d, id = "id", time = "time")
  r <- ild_missingness_report(
    x,
    outcome = "y",
    predictors = "x",
    fit_missing_model = TRUE,
    cohort_plot = FALSE
  )
  expect_s3_class(r, "ild_missingness_report")
  expect_true(all(c("compliance", "pattern", "cohort", "hazard", "flags", "snippets") %in% names(r)))
  expect_true(is.logical(r$flags$dropout_late_pooled))
  expect_true(length(r$snippets) >= 1L)
})

# MSM estimand-first API (v1.1 scaffold)

#' Define an MSM estimand specification
#'
#' Creates a lightweight estimand object used by [ild_msm_fit()].
#' v1.1 preserves backward compatibility with v1 static ATE calls while adding
#' explicit slots for regime specification, time targeting, and contrasts.
#'
#' @param type Estimand type. Supports \code{"ate"} and placeholder \code{"att"}.
#' @param regime Regime class. Supports \code{"static"} and scaffold \code{"dynamic"}.
#' @param treatment Binary treatment column name.
#' @param time_var Occasion index column (default \code{".ild_seq"}).
#' @param contrast Optional contrast definition. Character labels are accepted.
#'   Lists may include \code{label}, \code{treated}, and \code{control}.
#' @param target_time Optional target occasion:
#'   \code{"all"} (default), \code{"final"}, or numeric vector.
#' @param regime_value For \code{regime = "static"}, assignment target (default \code{1}).
#' @param dynamic_rule For \code{regime = "dynamic"}, a deterministic rule:
#'   function taking \code{data} and returning binary assignment, or one of
#'   \code{"always_treat"}, \code{"never_treat"}, \code{"as_observed"}.
#' @return Object of class \code{ild_msm_estimand}.
#' @export
#' @examples
#' e <- ild_msm_estimand(treatment = "trt")
#' e
ild_msm_estimand <- function(type = c("ate", "att"),
                             regime = "static",
                             treatment,
                             time_var = ".ild_seq",
                             contrast = NULL,
                             target_time = "all",
                             regime_value = 1,
                             dynamic_rule = NULL) {
  type <- match.arg(type)
  regime_class <- if (is.list(regime) && !is.null(regime$class)) {
    as.character(regime$class)[1L]
  } else {
    match.arg(as.character(regime)[1L], c("static", "dynamic"))
  }
  if (missing(treatment) || !nzchar(as.character(treatment)[1L])) {
    stop("treatment must be provided as a non-empty column name.", call. = FALSE)
  }
  target_time <- .ild_msm_normalize_target_time(target_time)
  regime_spec <- .ild_msm_normalize_regime_spec(
    regime = regime,
    regime_value = regime_value,
    dynamic_rule = dynamic_rule
  )
  contrast_spec <- .ild_msm_normalize_contrast(contrast)
  out <- list(
    type = type,
    regime = regime_class,
    treatment = as.character(treatment)[1L],
    time_var = as.character(time_var)[1L],
    target_time = target_time,
    contrast = contrast_spec,
    regime_spec = regime_spec
  )
  class(out) <- c("ild_msm_estimand", "list")
  out
}

#' @keywords internal
#' @noRd
validate_ild_msm_estimand <- function(x) {
  if (!inherits(x, "ild_msm_estimand")) {
    stop("estimand must be from ild_msm_estimand().", call. = FALSE)
  }
  if (!x$type %in% c("ate", "att")) {
    stop("type must be one of \"ate\" or \"att\".", call. = FALSE)
  }
  if (!x$regime %in% c("static", "dynamic")) {
    stop("regime must be one of \"static\" or \"dynamic\".", call. = FALSE)
  }
  req <- c("treatment", "time_var")
  for (nm in req) {
    if (is.null(x[[nm]]) || !nzchar(as.character(x[[nm]])[1L])) {
      stop("estimand$", nm, " must be non-empty.", call. = FALSE)
    }
  }
  if (!is.null(x$target_time)) {
    if (is.character(x$target_time)) {
      if (!x$target_time %in% c("all", "final")) {
        stop("target_time string must be \"all\" or \"final\".", call. = FALSE)
      }
    } else {
      tt <- as.numeric(x$target_time)
      if (!length(tt) || any(!is.finite(tt))) {
        stop("Numeric target_time must contain finite values.", call. = FALSE)
      }
    }
  }
  if (is.null(x$regime_spec) || !is.list(x$regime_spec) || is.null(x$regime_spec$class)) {
    stop("estimand$regime_spec must be a structured regime specification.", call. = FALSE)
  }
  if (!identical(as.character(x$regime_spec$class)[1L], x$regime)) {
    stop("estimand$regime and estimand$regime_spec$class must agree.", call. = FALSE)
  }
  if (!is.null(x$contrast) && !is.list(x$contrast)) {
    stop("estimand$contrast must be NULL or a contrast list.", call. = FALSE)
  }
  if (is.list(x$contrast)) {
    if (!is.null(x$contrast$treated)) {
      tr <- as.numeric(x$contrast$treated)[1L]
      if (!is.finite(tr) || !tr %in% c(0, 1)) {
        stop("estimand$contrast$treated must be 0 or 1.", call. = FALSE)
      }
    }
    if (!is.null(x$contrast$control)) {
      ct <- as.numeric(x$contrast$control)[1L]
      if (!is.finite(ct) || !ct %in% c(0, 1)) {
        stop("estimand$contrast$control must be 0 or 1.", call. = FALSE)
      }
    }
  }
  invisible(TRUE)
}

#' @export
print.ild_msm_estimand <- function(x, ...) {
  cat("MSM estimand\n")
  cat("  type: ", x$type, "\n", sep = "")
  cat("  regime: ", x$regime, "\n", sep = "")
  cat("  treatment: ", x$treatment, "\n", sep = "")
  cat("  time_var: ", x$time_var, "\n", sep = "")
  if (!is.null(x$target_time)) cat("  target_time: ", paste(x$target_time, collapse = ","), "\n", sep = "")
  if (!is.null(x$contrast) && !is.null(x$contrast$label)) cat("  contrast: ", x$contrast$label, "\n", sep = "")
  invisible(x)
}

#' @keywords internal
#' @noRd
.ild_msm_normalize_target_time <- function(target_time) {
  if (is.null(target_time)) return(NULL)
  if (is.character(target_time)) {
    if (length(target_time) > 1L) {
      target_time <- stats::na.omit(target_time)
      if (length(target_time) == 0L) {
        stop("target_time must be \"all\", \"final\", or numeric vector.", call. = FALSE)
      }
      target_time <- as.character(target_time)[1L]
    }
    val <- tolower(as.character(target_time)[1L])
    if (!val %in% c("all", "final")) {
      stop("target_time must be \"all\", \"final\", or numeric vector.", call. = FALSE)
    }
    return(val)
  }
  tt <- as.numeric(target_time)
  if (!length(tt) || any(!is.finite(tt))) {
    stop("Numeric target_time must contain finite values.", call. = FALSE)
  }
  tt
}

#' @keywords internal
#' @noRd
.ild_msm_normalize_regime_spec <- function(regime, regime_value, dynamic_rule) {
  if (is.list(regime)) {
    cls <- as.character(regime$class)[1L]
    if (!cls %in% c("static", "dynamic")) {
      stop("regime$class must be \"static\" or \"dynamic\".", call. = FALSE)
    }
    if (identical(cls, "static")) {
      rv <- as.numeric(if (is.null(regime$value)) regime_value else regime$value)[1L]
      if (!is.finite(rv) || !rv %in% c(0, 1)) {
        stop("regime$value for static regime must be 0 or 1.", call. = FALSE)
      }
      return(list(class = "static", value = as.integer(rv)))
    }
    rule <- if (is.null(regime$rule)) dynamic_rule else regime$rule
    if (is.null(rule)) {
      stop("regime$rule (or dynamic_rule) must be supplied for dynamic regime.", call. = FALSE)
    }
    if (is.character(rule)) {
      dr <- tolower(as.character(rule)[1L])
      if (!dr %in% c("always_treat", "never_treat", "as_observed")) {
        stop(
          "dynamic rule string must be one of \"always_treat\", \"never_treat\", \"as_observed\".",
          call. = FALSE
        )
      }
      return(list(class = "dynamic", rule = dr, deterministic = TRUE))
    }
    if (!is.function(rule)) {
      stop("dynamic regime rule must be a function or supported character value.", call. = FALSE)
    }
    return(list(class = "dynamic", rule = rule, deterministic = TRUE))
  }
  if (identical(regime, "static")) {
    rv <- as.numeric(regime_value)[1L]
    if (!is.finite(rv) || !rv %in% c(0, 1)) {
      stop("regime_value for static regime must be 0 or 1.", call. = FALSE)
    }
    return(list(class = "static", value = as.integer(rv)))
  }
  if (is.null(dynamic_rule)) {
    stop("dynamic_rule must be supplied when regime = \"dynamic\".", call. = FALSE)
  }
  if (is.character(dynamic_rule)) {
    dr <- tolower(as.character(dynamic_rule)[1L])
    if (!dr %in% c("always_treat", "never_treat", "as_observed")) {
      stop(
        "dynamic_rule string must be one of \"always_treat\", \"never_treat\", \"as_observed\".",
        call. = FALSE
      )
    }
    return(list(class = "dynamic", rule = dr, deterministic = TRUE))
  }
  if (is.function(dynamic_rule)) {
    return(list(class = "dynamic", rule = dynamic_rule, deterministic = TRUE))
  }
  stop("dynamic_rule must be a function or a supported character rule.", call. = FALSE)
}

#' @keywords internal
#' @noRd
.ild_msm_normalize_contrast <- function(contrast) {
  if (is.null(contrast)) {
    return(list(label = "A=1_vs_A=0", treated = 1L, control = 0L))
  }
  if (is.character(contrast)) {
    return(list(label = as.character(contrast)[1L], treated = 1L, control = 0L))
  }
  if (is.list(contrast)) {
    out <- contrast
    if (is.null(out$label)) out$label <- "custom_contrast"
    if (is.null(out$treated)) out$treated <- 1L
    if (is.null(out$control)) out$control <- 0L
    return(out)
  }
  stop("contrast must be NULL, character, or list.", call. = FALSE)
}

msm_fixture_data <- function(seed = 1L, n_id = 12L, n_obs_per = 6L) {
  set.seed(seed)
  d <- ild_simulate(n_id = n_id, n_obs_per = n_obs_per, seed = seed)
  d$stress <- rnorm(nrow(d))
  d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
  d <- ild_prepare(d, id = "id", time = "time")
  d <- ild_center(d, y)
  d <- ild_lag(d, stress, mode = "gap_aware", max_gap = Inf)
  d <- ild_lag(d, trt, mode = "gap_aware", max_gap = Inf)
  d
}

msm_recovery_scenario_grid_fixture <- function() {
  tibble::tibble(
    scenario_id = c("baseline", "positivity_stress", "misspecified_treatment"),
    positivity_stress = c(1, 1.8, 1),
    misspec_treatment_model = c(FALSE, FALSE, TRUE)
  )
}

# Heterogeneity interpretation for mixed models (partial pooling, BLUPs, summaries)
# See vignette("heterogeneity-interpretation", package = "tidyILD").

#' Heterogeneity and person-specific effects from mixed models
#'
#' Summarizes **conditional modes** (empirical Bayes / BLUP deviations) and
#' **person-specific coefficients** (\code{fixef + ranef}) for hierarchical models
#' fit with [ild_lme()] or [ild_brms()]. Outputs are labeled as **partial-pooling**
#' estimates; they are not the same as no-pooling separate regressions (see
#' [ild_person_model()]) or population-average fixed effects alone.
#'
#' @param fit A fitted \code{lmerMod}, \code{lme}, or \code{brmsfit} (from [ild_brms()]).
#' @param term Character vector of random-effect row names to restrict (e.g. \code{"x"} for slope).
#'   Default \code{NULL} includes all random-effect terms except optional filters.
#' @param group Name of the grouping factor in \code{ranef(fit)} to use (default first).
#' @param threshold Optional numeric; with \code{scale}, proportion of person-specific
#'   **total** coefficients exceeding this threshold is reported in \code{summary}.
#' @param scale If \code{threshold} is set: \code{"raw"} (default), \code{"sd_x"} (multiply
#'   threshold by residual SD of a focal predictor from model frame), or \code{"sd_y"}
#'   (multiply by SD of response). Only used when a single numeric predictor is found
#'   for \code{sd_x} / \code{sd_y} heuristics.
#' @param conf_level For \code{lmerMod}, Wald interval level for conditional modes (normal approx.).
#' @param n_draws For \code{brmsfit}, number of posterior draws for summaries (capped).
#' @param ... Reserved.
#'
#' @return An object of class \code{ild_heterogeneity}: a list with \code{meta}, \code{random_effects}
#'   (long tibble), \code{varcorr} (tibble), \code{summary} (per-term metrics), and \code{fit}.
#' @export
#' @seealso [ild_tidy.ild_heterogeneity()], [ild_autoplot.ild_heterogeneity()],
#'   [ild_heterogeneity_stratified()], [ild_person_model()]
ild_heterogeneity <- function(fit, ...) {
  UseMethod("ild_heterogeneity")
}

#' @export
ild_heterogeneity.default <- function(fit, ...) {
  stop(
    "No ild_heterogeneity method for ",
    paste(class(fit), collapse = ", "),
    call. = FALSE
  )
}

#' @rdname ild_heterogeneity
#' @export
ild_heterogeneity.lmerMod <- function(fit,
                                      term = NULL,
                                      group = NULL,
                                      threshold = NULL,
                                      scale = c("raw", "sd_x", "sd_y"),
                                      conf_level = 0.95,
                                      ...) {
  scale <- match.arg(scale)
  nr <- tryCatch(lme4::getME(fit, "n_rtrms"), error = function(e) 0L)
  if (is.null(nr) || nr < 1L) {
    stop("Model has no random effects; nothing to summarize for ild_heterogeneity().", call. = FALSE)
  }
  re <- tryCatch(lme4::ranef(fit, condVar = TRUE), error = function(e) NULL)
  co <- tryCatch(stats::coef(fit), error = function(e) NULL)
  if (is.null(re) || is.null(co)) {
    stop("Could not extract ranef/coef from lmer fit.", call. = FALSE)
  }
  gn <- names(re)
  if (is.null(gn) || length(gn) == 0L) {
    stop("Empty random-effects structure.", call. = FALSE)
  }
  if (!is.null(group)) {
    if (!group %in% gn) {
      stop("group '", group, "' not found in ranef names: ", paste(gn, collapse = ", "), call. = FALSE)
    }
    gn_use <- group
  } else {
    gn_use <- gn[1L]
  }

  mf <- tryCatch(stats::model.frame(fit), error = function(e) NULL)
  sd_y <- NA_real_
  sd_x_map <- list()
  if (!is.null(mf)) {
    resp <- tryCatch(stats::model.response(mf), error = function(e) NULL)
    if (!is.null(resp) && is.numeric(resp)) {
      sd_y <- stats::sd(resp, na.rm = TRUE)
    }
    for (nm in names(mf)) {
      v <- mf[[nm]]
      if (is.numeric(v)) {
        sd_x_map[[nm]] <- stats::sd(v, na.rm = TRUE)
      }
    }
  }

  rows <- list()
  for (g in gn_use) {
    R <- re[[g]]
    C <- co[[g]]
    if (is.null(R) || is.null(C)) next
    levs <- rownames(R)
    cn <- colnames(R)
    if (is.null(cn)) next
    pv <- attr(R, "postVar")
    z <- abs(stats::qnorm((1 - conf_level) / 2))
    for (ti in cn) {
      if (!is.null(term) && length(term) > 0L && !ti %in% term) next
      for (i in seq_along(levs)) {
        r_est <- as.numeric(R[levs[i], ti, drop = TRUE])
        t_est <- as.numeric(C[levs[i], ti, drop = TRUE])
        se <- NA_real_
        if (!is.null(pv) && length(dim(pv)) == 3L) {
          j <- match(ti, cn)
          if (!is.na(j) && j >= 1L && j <= dim(pv)[1L] && i <= dim(pv)[3L]) {
            vv <- pv[j, j, i]
            if (is.finite(vv) && vv >= 0) {
              se <- sqrt(vv)
            }
          }
        }
        lo <- if (is.finite(se)) t_est - z * se else NA_real_
        hi <- if (is.finite(se)) t_est + z * se else NA_real_
        rows[[length(rows) + 1L]] <- tibble::tibble(
          group = g,
          level_id = levs[i],
          term = ti,
          estimate_ranef = r_est,
          estimate_total = t_est,
          std_error = se,
          conf_low = lo,
          conf_high = hi,
          estimand = "conditional_mode_total"
        )
      }
    }
  }
  if (length(rows) == 0L) {
    stop("No random-effect rows extracted (check term filter).", call. = FALSE)
  }
  tbl <- dplyr::bind_rows(rows)

  vc_df <- tryCatch(
    as.data.frame(lme4::VarCorr(fit)),
    error = function(e) NULL
  )
  vc_tbl <- if (!is.null(vc_df)) {
    tibble::as_tibble(vc_df)
  } else {
    tibble::tibble()
  }

  thr_resolved <- .ild_heterogeneity_resolve_threshold(
    threshold, scale, tbl$term, sd_x_map, sd_y
  )

  summ <- .ild_heterogeneity_summary_table(tbl, vc_tbl, thr_resolved)
  meta <- list(
    engine = "lmer",
    class = "lmerMod",
    conf_level = conf_level,
    threshold = thr_resolved$threshold,
    scale = thr_resolved$scale_label,
    note = paste(
      "estimate_ranef: deviation from population; estimate_total: fixef + ranef (partial pooling).",
      "Intervals are conditional (empirical Bayes), not population CIs."
    )
  )
  structure(
    list(
      meta = meta,
      random_effects = tbl,
      varcorr = vc_tbl,
      summary = summ,
      fit = fit
    ),
    class = c("ild_heterogeneity", "list")
  )
}

#' @rdname ild_heterogeneity
#' @export
ild_heterogeneity.lme <- function(fit,
                                  term = NULL,
                                  group = NULL,
                                  threshold = NULL,
                                  scale = c("raw", "sd_x", "sd_y"),
                                  conf_level = 0.95,
                                  ...) {
  scale <- match.arg(scale)
  if (!inherits(fit, "lme")) {
    stop("Expected nlme::lme fit.", call. = FALSE)
  }
  re <- tryCatch(nlme::ranef(fit), error = function(e) NULL)
  co <- tryCatch(stats::coef(fit), error = function(e) NULL)
  if (is.null(re) || is.null(co)) {
    stop("Could not extract ranef/coef from lme fit.", call. = FALSE)
  }
  mf <- tryCatch(stats::model.frame(fit), error = function(e) NULL)
  sd_y <- NA_real_
  sd_x_map <- list()
  if (!is.null(mf)) {
    resp <- tryCatch(stats::model.response(mf), error = function(e) NULL)
    if (!is.null(resp) && is.numeric(resp)) {
      sd_y <- stats::sd(resp, na.rm = TRUE)
    }
    for (nm in names(mf)) {
      v <- mf[[nm]]
      if (is.numeric(v)) {
        sd_x_map[[nm]] <- stats::sd(v, na.rm = TRUE)
      }
    }
  }

  rows <- list()
  if (is.data.frame(re) && is.data.frame(co)) {
    if (!is.null(term) && length(term) > 0L) {
      cn <- intersect(term, colnames(re))
      if (length(cn) == 0L) {
        stop("No matching term columns in ranef() for lme.", call. = FALSE)
      }
    } else {
      cn <- colnames(re)
    }
    levs <- rownames(re)
    if (is.null(levs)) {
      levs <- as.character(seq_len(nrow(re)))
    }
    grp <- if (!is.null(group)) group else "id"
    for (ti in cn) {
      for (i in seq_along(levs)) {
        r_est <- as.numeric(re[levs[i], ti, drop = TRUE])
        t_est <- as.numeric(co[levs[i], ti, drop = TRUE])
        rows[[length(rows) + 1L]] <- tibble::tibble(
          group = grp,
          level_id = levs[i],
          term = ti,
          estimate_ranef = r_est,
          estimate_total = t_est,
          std_error = NA_real_,
          conf_low = NA_real_,
          conf_high = NA_real_,
          estimand = "conditional_mode_total"
        )
      }
    }
  } else if (is.list(co) && !is.data.frame(co)) {
    for (nm in names(co)) {
      if (!is.null(group) && !identical(nm, group)) next
      cm <- co[[nm]]
      if (!is.matrix(cm) && !is.data.frame(cm)) next
      cn <- colnames(cm)
      if (is.null(cn)) next
      levs <- rownames(cm)
      if (is.null(levs)) levs <- as.character(seq_len(nrow(cm)))
      for (ti in cn) {
        if (!is.null(term) && length(term) > 0L && !ti %in% term) next
        for (i in seq_along(levs)) {
          t_est <- as.numeric(cm[levs[i], ti, drop = TRUE])
          r_est <- NA_real_
          if (is.data.frame(re) || is.matrix(re)) {
            if (ti %in% colnames(re) && levs[i] %in% rownames(re)) {
              r_est <- as.numeric(re[levs[i], ti, drop = TRUE])
            }
          }
          rows[[length(rows) + 1L]] <- tibble::tibble(
            group = nm,
            level_id = levs[i],
            term = ti,
            estimate_ranef = r_est,
            estimate_total = t_est,
            std_error = NA_real_,
            conf_low = NA_real_,
            conf_high = NA_real_,
            estimand = "conditional_mode_total"
          )
        }
      }
    }
  }
  if (length(rows) == 0L) {
    stop("Could not reshape lme random effects into long format.", call. = FALSE)
  }
  tbl <- dplyr::bind_rows(rows)
  vc_df <- tryCatch(nlme::VarCorr(fit), error = function(e) NULL)
  vc_tbl <- if (!is.null(vc_df)) {
    tibble::tibble(note = "See nlme::VarCorr(fit) raw table", raw = list(vc_df))
  } else {
    tibble::tibble()
  }
  thr_resolved <- .ild_heterogeneity_resolve_threshold(
    threshold, scale, tbl$term, sd_x_map, sd_y
  )
  summ <- .ild_heterogeneity_summary_table(tbl, tibble::tibble(), thr_resolved)
  meta <- list(
    engine = "lme",
    class = "lme",
    conf_level = conf_level,
    threshold = thr_resolved$threshold,
    scale = thr_resolved$scale_label,
    note = paste(
      "nlme: per-level std_error/confidence for random effects not computed here;",
      "use lme4 for conditional variances or Bayesian brms for posterior intervals."
    )
  )
  structure(
    list(
      meta = meta,
      random_effects = tbl,
      varcorr = vc_tbl,
      summary = summ,
      fit = fit
    ),
    class = c("ild_heterogeneity", "list")
  )
}

#' @keywords internal
#' @noRd
.ild_heterogeneity_resolve_threshold <- function(threshold, scale, term_vec, sd_x_map, sd_y) {
  scale_label <- scale
  if (is.null(threshold)) {
    return(list(threshold = NULL, scale_label = NA_character_, effective = NULL))
  }
  th <- as.numeric(threshold)[1L]
  if (!is.finite(th)) {
    return(list(threshold = NULL, scale_label = NA_character_, effective = NULL))
  }
  eff <- th
  if (identical(scale, "sd_y") && is.finite(sd_y) && sd_y > 0) {
    eff <- th * sd_y
    scale_label <- sprintf("sd_y (x %.4g)", sd_y)
  } else if (identical(scale, "sd_x")) {
    terms_u <- unique(as.character(term_vec))
    terms_u <- terms_u[!terms_u %in% c("(Intercept)", "Intercept")]
    sdx <- NA_real_
    for (t in terms_u) {
      if (!is.null(sd_x_map[[t]])) {
        sdx <- sd_x_map[[t]]
        break
      }
    }
    if (is.finite(sdx) && sdx > 0) {
      eff <- th * sdx
      scale_label <- sprintf("sd_x (x %.4g)", sdx)
    } else {
      scale_label <- "raw (sd_x unavailable)"
    }
  }
  list(threshold = th, scale_label = scale_label, effective = eff)
}

#' @keywords internal
#' @noRd
.ild_heterogeneity_summary_table <- function(tbl, vc_tbl, thr) {
  eff_thr <- thr$effective
  summ <- tbl |>
    dplyr::group_by(.data$group, .data$term) |>
    dplyr::summarise(
      n_levels = dplyr::n(),
      mean_total = mean(.data$estimate_total, na.rm = TRUE),
      sd_total = stats::sd(.data$estimate_total, na.rm = TRUE),
      prop_gt_zero = mean(.data$estimate_total > 0, na.rm = TRUE),
      q25 = stats::quantile(.data$estimate_total, 0.25, na.rm = TRUE, names = FALSE, type = 7),
      q50 = stats::median(.data$estimate_total, na.rm = TRUE),
      q75 = stats::quantile(.data$estimate_total, 0.75, na.rm = TRUE, names = FALSE, type = 7),
      prop_gt_threshold = if (is.null(eff_thr) || !is.finite(eff_thr)) {
        NA_real_
      } else {
        mean(.data$estimate_total > eff_thr, na.rm = TRUE)
      },
      .groups = "drop"
    )
  if (nrow(vc_tbl) > 0L && all(c("grp", "var1", "sdcor") %in% names(vc_tbl))) {
    vc2 <- vc_tbl
    if ("var2" %in% names(vc2)) {
      vc2 <- vc2 |>
        dplyr::filter(is.na(.data$var2) | as.character(.data$var2) == "")
    }
    vc2 <- vc2 |>
      dplyr::select("grp", "var1", "sdcor") |>
      dplyr::rename(varcorr_sdcor = "sdcor", group = "grp", term = "var1")
    summ <- dplyr::left_join(summ, vc2, by = c("group", "term"))
  } else {
    summ$varcorr_sdcor <- NA_real_
  }
  summ
}

#' @rdname ild_heterogeneity
#' @param formula,data Passed to [ild_lme()] within each subgroup.
#' @param subgroup Character name of a column in \code{data} defining subgroups.
#' @param min_n_id Minimum number of distinct persons required in a subgroup.
#' @param ... Passed to [ild_lme()].
#'
#' @return A tibble with one row per successful subgroup fit, including key columns from
#'   \code{ild_heterogeneity()$summary} flattened (\code{subgroup_level}, \code{term}, etc.).
#'   This is a **descriptive** comparison across refits, not a formal test of variance heterogeneity.
#' @export
ild_heterogeneity_stratified <- function(formula,
                                         data,
                                         subgroup,
                                         min_n_id = 5L,
                                         ...) {
  validate_ild(data)
  if (!subgroup %in% names(data)) {
    stop("subgroup column '", subgroup, "' not found in data.", call. = FALSE)
  }
  min_n_id <- as.integer(min_n_id)[1L]
  levs <- unique(data[[subgroup]])
  out_rows <- list()
  for (lev in levs) {
    sub <- data[data[[subgroup]] == lev, , drop = FALSE]
    n_id <- length(unique(sub[[".ild_id"]]))
    if (n_id < min_n_id) {
      next
    }
    fit <- tryCatch(
      ild_lme(formula, data = sub, ...),
      error = function(e) NULL
    )
    if (is.null(fit)) {
      next
    }
    het <- tryCatch(ild_heterogeneity(fit), error = function(e) NULL)
    if (is.null(het)) {
      next
    }
    s <- het$summary
    s$subgroup_level <- lev
    s$n_id_subgroup <- n_id
    out_rows[[length(out_rows) + 1L]] <- s
  }
  if (length(out_rows) == 0L) {
    return(tibble::tibble(
      subgroup_level = character(),
      term = character(),
      note = character()
    ))
  }
  dplyr::bind_rows(out_rows)
}

#' @describeIn ild_heterogeneity Tidy long table of person-level random effects (for \code{ild_heterogeneity} objects).
#' @param x An \code{ild_heterogeneity} object.
#' @export
ild_tidy.ild_heterogeneity <- function(x, ...) {
  if (!inherits(x, "ild_heterogeneity")) {
    stop("Expected ild_heterogeneity object.", call. = FALSE)
  }
  x$random_effects
}

#' @describeIn ild_heterogeneity Caterpillar or histogram of person-specific effects.
#' @param type \code{"caterpillar"} (default) or \code{"histogram"}.
#' @param term Which random-effect term to plot (required if multiple).
#' @export
ild_autoplot.ild_heterogeneity <- function(x, type = c("caterpillar", "histogram"), term = NULL, ...) {
  if (!inherits(x, "ild_heterogeneity")) {
    stop("Expected ild_heterogeneity object.", call. = FALSE)
  }
  type <- match.arg(type)
  d <- x$random_effects
  terms_u <- unique(d$term)
  if (is.null(term)) {
    term <- if (length(terms_u) == 1L) {
      terms_u[1L]
    } else {
      non_int <- terms_u[!terms_u %in% c("(Intercept)", "Intercept")]
      if (length(non_int) > 0L) {
        non_int[1L]
      } else {
        terms_u[1L]
      }
    }
  }
  if (!term %in% terms_u) {
    stop("term '", term, "' not found; available: ", paste(terms_u, collapse = ", "), call. = FALSE)
  }
  sub <- d[d$term == term, , drop = FALSE]
  sub <- sub[order(sub$estimate_total), , drop = FALSE]
  sub$level_ord <- factor(sub$level_id, levels = sub$level_id)
  if (type == "caterpillar") {
    ggplot2::ggplot(sub, ggplot2::aes(x = .data$level_ord, y = .data$estimate_total)) +
      ggplot2::geom_point() +
      ggplot2::geom_errorbar(
        ggplot2::aes(
          ymin = .data$conf_low,
          ymax = .data$conf_high
        ),
        width = 0,
        na.rm = TRUE
      ) +
      ggplot2::geom_hline(yintercept = 0, linetype = 2, alpha = 0.6) +
      ggplot2::coord_flip() +
      ggplot2::labs(
        x = NULL,
        y = "Person-specific coefficient (partial pooling)",
        title = paste0("Caterpillar: ", term)
      ) +
      ggplot2::theme_minimal() +
      ggplot2::theme(axis.text.y = ggplot2::element_text(size = 6))
  } else {
    ggplot2::ggplot(sub, ggplot2::aes(x = .data$estimate_total)) +
      ggplot2::geom_histogram(bins = min(30L, max(5L, nrow(sub) %/% 2)), fill = "gray70", color = "gray40") +
      ggplot2::geom_vline(xintercept = 0, linetype = 2, alpha = 0.6) +
      ggplot2::labs(
        x = "Person-specific coefficient",
        y = "Count",
        title = paste0("Distribution: ", term, " (partial pooling)")
      ) +
      ggplot2::theme_minimal()
  }
}

#' @export
print.ild_heterogeneity <- function(x, ...) {
  cat("ild_heterogeneity:", x$meta$engine, "\n")
  cat(x$meta$note, "\n\n")
  print(x$summary, n = min(20L, nrow(x$summary)))
  invisible(x)
}

#' @rdname ild_heterogeneity
#' @export
ild_heterogeneity.brmsfit <- function(fit,
                                      term = NULL,
                                      group = NULL,
                                      threshold = NULL,
                                      scale = c("raw", "sd_x", "sd_y"),
                                      n_draws = 500L,
                                      ...) {
  rlang::check_installed("brms")
  if (is.null(attr(fit, "ild_data", exact = TRUE))) {
    stop(
      "ild_heterogeneity() for brmsfit requires ild_brms() (ild_data attribute missing).",
      call. = FALSE
    )
  }
  scale <- match.arg(scale)
  n_draws <- as.integer(n_draws)[1L]
  co <- tryCatch(stats::coef(fit, summary = TRUE), error = function(e) NULL)
  if (is.null(co)) {
    stop("Could not extract coef(fit, summary = TRUE) from brms fit.", call. = FALSE)
  }
  gn <- names(co)
  if (is.null(gn) || length(gn) == 0L) {
    stop("Empty coef() list from brms fit.", call. = FALSE)
  }
  if (!is.null(group)) {
    if (!group %in% gn) {
      stop("group '", group, "' not in coef() names: ", paste(gn, collapse = ", "), call. = FALSE)
    }
    gn_use <- group
  } else {
    gn_use <- gn[1L]
  }
  data <- attr(fit, "ild_data", exact = TRUE)
  mf <- tryCatch(stats::model.frame(fit), error = function(e) NULL)
  sd_y <- NA_real_
  sd_x_map <- list()
  if (!is.null(mf)) {
    resp <- tryCatch(stats::model.response(mf), error = function(e) NULL)
    if (!is.null(resp) && is.numeric(resp)) {
      sd_y <- stats::sd(resp, na.rm = TRUE)
    }
    for (nm in names(mf)) {
      v <- mf[[nm]]
      if (is.numeric(v)) {
        sd_x_map[[nm]] <- stats::sd(v, na.rm = TRUE)
      }
    }
  }
  rows <- list()
  for (g in gn_use) {
    arr <- co[[g]]
    if (is.null(arr)) next
    d <- dim(arr)
    # brms::coef(summary=TRUE): dim = c(n_level, n_stat, n_coef) (see dimnames order)
    dn2 <- dimnames(arr)[[2L]]
    dn3 <- dimnames(arr)[[3L]]
    if (length(d) == 3L && !is.null(dn2) && !is.null(dn3)) {
      i_est <- match("Estimate", dn2)
      i_se <- match("Est.Error", dn2)
      if (is.na(i_se)) {
        w <- grep("Est.Error", dn2, fixed = TRUE)
        i_se <- if (length(w)) w[1L] else NA_integer_
      }
      i_lo <- match("Q2.5", dn2)
      if (is.na(i_lo)) {
        w <- grep("2.5%", dn2, fixed = TRUE)
        if (length(w) == 0L) w <- grep("Q2.5", dn2, fixed = TRUE)
        i_lo <- if (length(w)) w[1L] else NA_integer_
      }
      i_hi <- match("Q97.5", dn2)
      if (is.na(i_hi)) {
        w <- grep("97.5%", dn2, fixed = TRUE)
        if (length(w) == 0L) w <- grep("Q97.5", dn2, fixed = TRUE)
        i_hi <- if (length(w)) w[1L] else NA_integer_
      }
      if (is.na(i_est)) next
      levs <- dimnames(arr)[[1L]]
      tns <- dn3
      if (is.null(levs) || is.null(tns)) next
      for (ti in seq_along(tns)) {
        tnm <- tns[ti]
        if (!is.null(term) && length(term) > 0L && !tnm %in% term) next
        for (i in seq_along(levs)) {
          t_est <- as.numeric(arr[i, i_est, ti])
          se <- if (!is.na(i_se)) as.numeric(arr[i, i_se, ti]) else NA_real_
          lo <- if (!is.na(i_lo)) as.numeric(arr[i, i_lo, ti]) else NA_real_
          hi <- if (!is.na(i_hi)) as.numeric(arr[i, i_hi, ti]) else NA_real_
          rows[[length(rows) + 1L]] <- tibble::tibble(
            group = g,
            level_id = levs[i],
            term = tnm,
            estimate_ranef = NA_real_,
            estimate_total = t_est,
            std_error = se,
            conf_low = lo,
            conf_high = hi,
            estimand = "posterior_summary_total"
          )
        }
      }
    }
  }
  if (length(rows) == 0L) {
    stop("Could not parse brms coef(summary) arrays for heterogeneity.", call. = FALSE)
  }
  tbl <- dplyr::bind_rows(rows)
  thr_resolved <- .ild_heterogeneity_resolve_threshold(
    threshold, scale, tbl$term, sd_x_map, sd_y
  )
  summ <- .ild_heterogeneity_summary_table(tbl, tibble::tibble(), thr_resolved)
  meta <- list(
    engine = "brms",
    class = "brmsfit",
    conf_level = NA_real_,
    threshold = thr_resolved$threshold,
    scale = thr_resolved$scale_label,
    n_draws = n_draws,
    note = paste(
      "Person-specific coefficients from brms::coef(summary=TRUE) (posterior means/median-aligned summaries).",
      "Not equivalent to lme4 BLUPs; intervals are posterior quantiles."
    )
  )
  structure(
    list(
      meta = meta,
      random_effects = tbl,
      varcorr = tibble::tibble(),
      summary = summ,
      fit = fit
    ),
    class = c("ild_heterogeneity", "list")
  )
}

#' @keywords internal
#' @noRd
fill_diagnostics_heterogeneity <- function(fit) {
  out <- list(available = FALSE, reason = NA_character_, object = NULL)
  if (!inherits(fit, c("lmerMod", "lme", "brmsfit"))) {
    out$reason <- "unsupported fit class"
    return(out)
  }
  h <- tryCatch(
    ild_heterogeneity(fit),
    error = function(e) {
      out$reason <<- conditionMessage(e)
      NULL
    }
  )
  if (!is.null(h)) {
    out$available <- TRUE
    out$object <- h
    out$reason <- NA_character_
  }
  out
}

# Phase 1 column dictionaries for ild_tidy() and ild_augment() — contract only; outputs migrate later

#' Column dictionary for ild_tidy() outputs
#'
#' Minimum stable contract for parameter-level tables returned by \code{\link{ild_tidy}}.
#' Implementations (\code{\link{tidy_ild_model}}, \code{\link{ild_tidy.brmsfit}}) emit all
#' required columns; \code{conf_low} / \code{conf_high} replace legacy \code{ci_low} / \code{ci_high}.
#'
#' @return A list with \code{required} and \code{optional} character vectors.
#' @details
#' **Required columns:** \code{term}, \code{component}, \code{effect_level}, \code{estimate},
#' \code{std_error}, \code{conf_low}, \code{conf_high}, \code{statistic}, \code{p_value},
#' \code{interval_type}, \code{engine}, \code{model_class}.
#'
#' **Optional:** \code{rhat}, \code{ess_bulk}, \code{ess_tail}, \code{pd}, \code{rope_low}, \code{rope_high}.
#'
#' **\code{interval_type}:** Frequentist rows are typically \code{Wald} (model or robust; see \code{\link{tidy_ild_model}}).
#' Bootstrap output from \code{\link{tidy_ild_msm_bootstrap}} uses \code{bootstrap_percentile} (equal-tailed over replicate coefficients).
#' @seealso \code{\link{ild_augment_schema}}, \code{\link{ild_diagnostics_bundle}}.
#'   Full semantics for \code{component}, \code{effect_level}, and \code{interval_type}:
#'   \code{vignette("developer-contracts", package = "tidyILD")}.
#' @export
ild_tidy_schema <- function() {
  list(
    required = ILD_TIDY_REQUIRED_COLS,
    optional = ILD_TIDY_OPTIONAL_COLS
  )
}

#' @keywords internal
ILD_TIDY_REQUIRED_COLS <- c(
  "term",
  "component",
  "effect_level",
  "estimate",
  "std_error",
  "conf_low",
  "conf_high",
  "statistic",
  "p_value",
  "interval_type",
  "engine",
  "model_class"
)

ILD_TIDY_OPTIONAL_COLS <- c(
  "rhat",
  "ess_bulk",
  "ess_tail",
  "pd",
  "rope_low",
  "rope_high"
)

#' Column dictionary for \code{ild_augment()} outputs
#'
#' Defines the minimum stable contract for augmented observation-level tables.
#' \code{\link{augment_ild_model}} and \code{\link{ild_augment}} methods emit all required
#' columns; see \code{\link{augment_ild_model}} for \code{.resid_std} semantics.
#'
#' @details
#' **Required columns**
#' \tabular{ll}{
#'   \code{.ild_id} \tab Person identifier (from ILD) \cr
#'   \code{.ild_time} \tab Time variable \cr
#'   \code{.outcome} \tab Observed outcome (numeric vector aligned to rows) \cr
#'   \code{.fitted} \tab Fitted / posterior mean fitted value \cr
#'   \code{.resid} \tab Raw residual \cr
#'   \code{.resid_std} \tab Pearson-type residual when \code{residuals(fit, type = "pearson")} is available and length-matched; otherwise \code{NA} (not ad hoc z-scores) \cr
#'   \code{engine} \tab Character engine id \cr
#'   \code{model_class} \tab Model class string \cr
#' }
#'
#' **Optional:** \code{.fitted_lower}, \code{.fitted_upper}, \code{.influence},
#' \code{.state}, \code{.state_lower}, \code{.state_upper}.
#'
#' @return A list with \code{required} and \code{optional} character vectors.
#' @seealso \code{\link{ild_tidy_schema}}, \code{\link{ild_diagnostics_bundle}}.
#'   Reserved prefixes and migration notes: \code{vignette("developer-contracts", package = "tidyILD")}.
#' @export
ild_augment_schema <- function() {
  list(
    required = ILD_AUGMENT_REQUIRED_COLS,
    optional = ILD_AUGMENT_OPTIONAL_COLS
  )
}

#' @keywords internal
ILD_AUGMENT_REQUIRED_COLS <- c(
  ".ild_id",
  ".ild_time",
  ".outcome",
  ".fitted",
  ".resid",
  ".resid_std",
  "engine",
  "model_class"
)

ILD_AUGMENT_OPTIONAL_COLS <- c(
  ".fitted_lower",
  ".fitted_upper",
  ".influence",
  ".state",
  ".state_lower",
  ".state_upper"
)

test_that("augment_ild_model returns ild_augment_schema() columns", {
  req <- ild_augment_schema()$required
  d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  aug <- augment_ild_model(fit)
  expect_s3_class(aug, "tbl_df")
  expect_true(all(req %in% names(aug)))
  expect_equal(nrow(aug), nrow(x))
  expect_equal(aug$.resid, as.numeric(stats::residuals(fit)))
  expect_equal(aug$.fitted, as.numeric(stats::fitted(fit)))
  expect_equal(aug$.outcome, as.numeric(x$y))
  expect_true(is.numeric(aug$.resid_std))
})

test_that("augment_ild_model errors when ild_data missing", {
  fit <- lme4::lmer(y ~ 1 + (1 | id), data = data.frame(id = rep(1:2, each = 3), y = rnorm(6)))
  expect_error(augment_ild_model(fit), "refit using ild_lme")
})

test_that("tidy_ild_model returns ild_tidy_schema() columns for lmer", {
  req <- ild_tidy_schema()$required
  d <- ild_simulate(n_id = 3, n_obs_per = 6, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
  tt <- tidy_ild_model(fit)
  expect_s3_class(tt, "tbl_df")
  expect_true(all(req %in% names(tt)))
  expect_equal(tt$term, "(Intercept)")
  expect_true(all(tt$component == "fixed"))
  expect_equal(tt$effect_level[1], "population")
  expect_equal(tt$interval_type[1], "Wald")
  expect_equal(tt$engine[1], "lmer")
  expect_true(grepl("lmerMod", tt$model_class[1]))
})

test_that("tidy_ild_model effect_level uses _wp/_bp suffixes", {
  d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 2)
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_center(x, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
  tt <- tidy_ild_model(fit)
  expect_equal(tt$effect_level[tt$term == "y_wp"], "within")
  expect_equal(tt$effect_level[tt$term == "y_bp"], "between")
})

test_that("tidy_ild_model returns ild_tidy_schema() columns for lme", {
  req <- ild_tidy_schema()$required
  d <- ild_simulate(n_id = 3, n_obs_per = 6, irregular = TRUE, seed = 1)
  x <- ild_prepare(d, id = "id", time = "time")
  fit <- ild_lme(y ~ 1, data = x, ar1 = TRUE)
  tt <- tidy_ild_model(fit)
  expect_s3_class(tt, "tbl_df")
  expect_true(all(req %in% names(tt)))
  expect_true(is.numeric(tt$p_value))
  expect_equal(tt$engine[1], "lme")
})

# Bayesian ILD models via brms (optional dependency: brms, rstan / cmdstanr)

#' Prior specifications for common ILD mixed models
#'
#' Returns a \code{brmsprior} object suitable for the \code{prior} argument of
#' [ild_brms()] / [brms::brm()]. Templates follow weakly informative defaults:
#' Student-t on intercepts and group SDs, normal on regression coefficients, and
#' exponential on residual SD where applicable.
#'
#' @param template Character selecting the template:
#'   \describe{
#'     \item{\code{default}}{Student-t intercept; Normal(0, 5) fixed effects;
#'       Student-t group SDs; exponential residual SD (Gaussian models).}
#'     \item{\code{weakly_informative}}{Wider scales (more diffuse) on intercept
#'       and \code{b} coefficients.}
#'     \item{\code{minimal_shrinkage}}{Tighter Normal(0, 2.5) on \code{b} (similar
#'       to Gelman et al. scaled logistic guidance; use with scaled predictors).}
#'   }
#' @return A \code{brmsprior} data frame (class \code{brmsprior}).
#' @seealso [ild_brms()]
#' @export
ild_prior_ild <- function(template = c("default", "weakly_informative", "minimal_shrinkage")) {
  rlang::check_installed("brms", reason = "to use ild_prior_ild()")
  template <- match.arg(template)
  switch(template,
    default = c(
      brms::set_prior("student_t(3, 0, 10)", class = "Intercept"),
      brms::set_prior("normal(0, 5)", class = "b"),
      brms::set_prior("student_t(3, 0, 2.5)", class = "sd"),
      brms::set_prior("exponential(1)", class = "sigma")
    ),
    weakly_informative = c(
      brms::set_prior("student_t(3, 0, 10)", class = "Intercept"),
      brms::set_prior("normal(0, 10)", class = "b"),
      brms::set_prior("student_t(3, 0, 5)", class = "sd"),
      brms::set_prior("exponential(1)", class = "sigma")
    ),
    minimal_shrinkage = c(
      brms::set_prior("student_t(3, 0, 5)", class = "Intercept"),
      brms::set_prior("normal(0, 2.5)", class = "b"),
      brms::set_prior("student_t(3, 0, 2.5)", class = "sd"),
      brms::set_prior("exponential(1)", class = "sigma")
    )
  )
}

#' Build standardized posterior / sampler metadata (internal)
#' @param fit A \code{brmsfit} object.
#' @param prior_template Character name of template if used, else \code{NULL}.
#' @noRd
ild_posterior_attr <- function(fit, prior_template = NULL) {
  sm <- summary(fit)
  ca <- tryCatch(fit$fit@stan_args[[1L]], error = function(e) NULL)
  adapt_delta <- if (!is.null(ca$adapt_delta)) ca$adapt_delta else NA_real_
  max_td <- if (!is.null(ca$max_treedepth)) as.integer(ca$max_treedepth) else NA_integer_
  seed <- if (!is.null(ca$seed)) ca$seed else NA_integer_
  ps_df <- tryCatch(
    as.data.frame(brms::prior_summary(fit)),
    error = function(e) NULL
  )
  list(
    engine = "brms",
    backend = "brms",
    brms_version = ild_backend_pkg_version("brms"),
    rstan_version = ild_backend_pkg_version("rstan"),
    prior_template = prior_template,
    prior_summary = ps_df,
    chains = sm$chains,
    iter = sm$iter,
    warmup = sm$warmup,
    thin = sm$thin,
    total_post_warmup_draws = sm$total_ndraws,
    sampler = sm$sampler,
    algorithm = sm$algorithm,
    adapt_delta = adapt_delta,
    max_treedepth = max_td,
    seed = seed,
    n_divergent = ild_brms_n_divergent(fit),
    n_max_treedepth_hits = ild_brms_n_treedepth_hits(fit, max_td)
  )
}

#' @noRd
ild_brms_n_divergent <- function(fit) {
  np <- tryCatch(brms::nuts_params(fit), error = function(e) NULL)
  if (is.null(np) || !all(c("Parameter", "Value") %in% names(np))) return(NA_integer_)
  div <- np[np$Parameter == "divergent__", , drop = FALSE]
  if (nrow(div) == 0L) return(0L)
  sum(as.integer(div$Value))
}

#' Count post-warmup iterations hitting max treedepth (heuristic)
#' @noRd
ild_brms_n_treedepth_hits <- function(fit, max_td = NA_integer_) {
  np <- tryCatch(brms::nuts_params(fit), error = function(e) NULL)
  if (is.null(np) || !all(c("Parameter", "Value") %in% names(np))) return(NA_integer_)
  td <- np[np$Parameter == "treedepth__", , drop = FALSE]
  if (nrow(td) == 0L) return(NA_integer_)
  if (is.na(max_td) || max_td <= 0L) max_td <- 10L
  sum(as.integer(td$Value) >= max_td)
}

#' Fit a Bayesian mixed model to ILD data with \pkg{brms}
#'
#' Wraps [brms::brm()] with ILD validation, default priors for common longitudinal
#' mixed models, standardized posterior metadata in \code{attr(fit, "ild_posterior")},
#' and analysis provenance (priors, chains, iterations). Attach \code{ild_data} for
#' use with [ild_tidy()], [ild_augment()], [ild_diagnose()], and [ild_methods()].
#'
#' Requires packages \pkg{brms} and a Stan backend (\pkg{rstan} or \pkg{cmdstanr}).
#'
#' @param formula A \pkg{brms} / \pkg{lme4}-style formula (e.g. \code{y ~ x + (1|id)}).
#' @param data An ILD object from [ild_prepare()].
#' @param prior Optional \code{brmsprior} object. If \code{NULL}, built from \code{prior_template}.
#' @param prior_template Passed to [ild_prior_ild()] when \code{prior} is \code{NULL}.
#' @param warn_uncentered If \code{TRUE} (default), warn when a numeric predictor varies
#'   within and between persons without \code{_wp}/\code{_bp} (same idea as [ild_lme()]).
#' @param ... Arguments passed to [brms::brm()] (e.g. \code{chains}, \code{iter},
#'   \code{warmup}, \code{control}, \code{backend}, \code{seed}).
#' @return A \code{brmsfit} object with attributes:
#'   \describe{
#'     \item{\code{ild_data}}{The ILD data frame.}
#'     \item{\code{ild_posterior}}{List: sampler settings, prior summary, divergences, etc.}
#'     \item{\code{ild_provenance}}{Standard tidyILD analysis provenance.}
#'   }
#' @seealso [ild_fit()] with \code{backend = "brms"} is equivalent to \code{ild_brms()}.
#'   For IPW/MSM workflows, full Bayesian joint MSM is not in tidyILD; see [ild_msm_inference].
#' @importFrom stats formula
#' @export
ild_brms <- function(formula,
                     data,
                     prior = NULL,
                     prior_template = c("default", "weakly_informative", "minimal_shrinkage"),
                     warn_uncentered = TRUE,
                     ...) {
  rlang::check_installed("brms", reason = "to use ild_brms()")
  validate_ild(data)
  prior_template <- match.arg(prior_template)
  if (is.null(prior)) {
    prior <- ild_prior_ild(prior_template)
    pt_name <- prior_template
  } else {
    pt_name <- NULL
  }
  if (warn_uncentered) {
    f_unc <- tryCatch(stats::formula(formula), error = function(e) formula)
    uncentered <- ild_detect_uncentered_predictors(data, f_unc)
    if (length(uncentered) > 0) {
      warning("Predictor(s) '", paste(uncentered, collapse = "', '"),
              "' vary both within and between persons. Consider using ild_center(",
              paste(uncentered, collapse = ", "), ") for within/between decomposition.",
              call. = FALSE)
    }
  }
  fit <- brms::brm(formula = formula, data = data, prior = prior, ...)
  attr(fit, "ild_data") <- data
  attr(fit, "ild_posterior") <- ild_posterior_attr(fit, prior_template = pt_name)
  ps <- attr(fit, "ild_posterior", exact = TRUE)
  prior_txt <- tryCatch(
    paste(utils::capture.output(print(brms::prior_summary(fit))), collapse = "\n"),
    error = function(e) "prior_summary unavailable"
  )
  attr(fit, "ild_provenance") <- ild_new_analysis_provenance(
    data, "ild_brms",
    list(
      formula = safe_deparse_formula(formula),
      prior_template = pt_name,
      prior_summary_text = prior_txt,
      chains = ps$chains,
      iter = ps$iter,
      warmup = ps$warmup,
      thin = ps$thin,
      adapt_delta = ps$adapt_delta,
      max_treedepth = ps$max_treedepth,
      seed = ps$seed,
      backend = "brms"
    ),
    list(
      n_obs = nrow(data),
      n_id = length(unique(data[[".ild_id"]])),
      brms_version = ps$brms_version,
      n_divergent = ps$n_divergent,
      max_rhat = ild_brms_max_rhat(fit)
    )
  )
  fit
}

#' @noRd
safe_deparse_formula <- function(f) {
  paste(deparse(f), collapse = " ")
}

#' Maximum R-hat across model parameters (brms)
#' @noRd
ild_brms_max_rhat <- function(fit) {
  rh <- tryCatch(brms::rhat(fit), error = function(e) NULL)
  if (is.null(rh) || length(rh) == 0L) return(NA_real_)
  suppressWarnings(max(as.numeric(rh), na.rm = TRUE))
}

#' Augment an ILD model fit with fitted values and residuals
#'
#' Returns a tibble conforming to \code{\link{ild_augment_schema}}: \code{.ild_id},
#' \code{.ild_time}, \code{.outcome}, \code{.fitted}, \code{.resid}, \code{.resid_std},
#' \code{engine}, \code{model_class}, plus optional columns (see schema).
#' \code{.resid_std} is Pearson-type when \code{residuals(fit, type = "pearson")} is
#' available and length-matched; otherwise \code{NA} (principled but sparse).
#' Used internally by [ild_diagnostics()] and [ild_plot()]. Requires
#' \code{attr(fit, "ild_data")}; refit with [ild_lme()] if missing.
#'
#' @param fit A fitted model from [ild_lme()] (must have \code{attr(fit, "ild_data")}).
#' @param ... Unused.
#' @return A tibble; see \code{\link{ild_augment_schema}}.
#' @export
augment_ild_model <- function(fit, ...) {
  data <- attr(fit, "ild_data", exact = TRUE)
  if (is.null(data)) {
    stop("Fit was not produced by tidyILD; refit using ild_lme() so the fit carries ild_data.",
         call. = FALSE)
  }
  validate_ild(data)
  n <- nrow(data)
  res <- stats::residuals(fit)
  f <- tryCatch(stats::fitted(fit), error = function(e) NULL)
  if (is.null(f) || length(f) != n) f <- rep(NA_real_, n)
  mf <- tryCatch(stats::model.frame(fit, data = data), error = function(e) stats::model.frame(fit))
  y <- stats::model.response(mf)
  y_num <- as.numeric(y)
  if (length(y_num) != n) y_num <- rep(NA_real_, n)
  r_std <- ild_augment_pearson_residuals(fit, n)
  engine <- if (inherits(fit, "lme")) "lme" else "lmer"
  mc <- ild_tidy_model_class_string(fit)
  ild_augment_assemble(
    .ild_id = data[[".ild_id"]],
    .ild_time = data[[".ild_time"]],
    .outcome = y_num,
    .fitted = as.numeric(f),
    .resid = as.numeric(res),
    .resid_std = r_std,
    engine = engine,
    model_class = mc,
    optional = NULL
  )
}

#' Get response variable name from an ild_lme fit (lmer or lme)
#' @param fit lmerMod or lme object.
#' @return Character; name of the response (e.g. "y"). Falls back to "outcome" if not determined.
#' @noRd
ild_response_name <- function(fit) {
  if (inherits(fit, "ild_fit_kfas")) {
    r <- attr(fit, "ild_response", exact = TRUE)
    if (!is.null(r)) return(r)
  }
  if (inherits(fit, "brmsfit")) {
    f <- tryCatch(stats::formula(fit), error = function(e) NULL)
    if (!is.null(f)) {
      vars <- all.vars(f)
      if (length(vars) >= 1L) return(vars[1L])
    }
  }
  f <- tryCatch(stats::formula(fit), error = function(e) NULL)
  if (is.null(f)) return("outcome")
  vars <- all.vars(f)
  if (length(vars) >= 1L) return(vars[1L])
  lhs <- tryCatch(f[[2L]], error = function(e) NULL)
  if (!is.null(lhs)) return(deparse(lhs, nlines = 1L))
  "outcome"
}

#' Tidy fixed effects from an ILD model fit
#'
#' Returns a tibble conforming to \code{\link{ild_tidy_schema}}: required columns include
#' \code{term}, \code{component}, \code{effect_level}, \code{estimate}, \code{std_error},
#' \code{conf_low}, \code{conf_high}, \code{statistic}, \code{p_value}, \code{interval_type},
#' \code{engine}, \code{model_class}; optional Bayesian columns are \code{NA} for these engines.
#' \code{component} is \code{"fixed"} for coefficient rows; \code{effect_level} is inferred
#' conservatively from term names (e.g. \verb{_wp}/\verb{_bp} suffixes) or \code{"population"}
#' for the intercept, \code{"unknown"} when ambiguous.
#'
#' With \code{object = TRUE}, returns an object of class \code{tidyild_model} (meta + table) for use with
#' \code{print.tidyild_model}.
#'
#' @param fit A fitted model from [ild_lme()] (lmerMod or lme).
#' @param conf_level Numeric. Confidence level for intervals (default 0.95).
#' @param object Logical. If \code{TRUE}, return a list with \code{meta} and \code{table}
#'   and class \code{tidyild_model} for polished printing (default \code{FALSE}).
#' @param se Character. \code{"model"} (default) uses the model's standard errors;
#'   \code{"robust"} uses cluster-robust SEs from [ild_robust_se()] (requires
#'   \pkg{clubSandwich}).
#' @param robust_type Character. When \code{se = "robust"}, the correction type:
#'   \code{"CR2"} (recommended), \code{"CR3"}, or \code{"CR0"}.
#' @param ... Passed to [ild_robust_se()] when \code{se = "robust"} (e.g. \code{cluster}, \code{cluster_vec}).
#' @return A tibble, or when \code{object = TRUE} a list of class \code{tidyild_model}.
#'
#' @section Model-based vs robust SE:
#' With \code{se = "model"} (default), standard errors and CIs come from the fitted
#' model. With \code{se = "robust"}, cluster-robust (sandwich) SEs are used;
#' CIs and p-values are based on a Wald normal approximation. Install the
#' \pkg{clubSandwich} package to use robust SEs. For **IPW / MSM** weighted \code{lmer}
#' fits ([ild_ipw_refit()]), robust SEs do not account for estimated weights; use
#' [ild_msm_bootstrap()] and [tidy_ild_msm_bootstrap()] when you need bootstrap CIs;
#' see [ild_msm_inference].
#'
#' @export
tidy_ild_model <- function(fit, conf_level = 0.95, object = FALSE,
                          se = c("model", "robust"), robust_type = c("CR2", "CR3", "CR0"), ...) {
  if (inherits(fit, "ild_msm_bootstrap")) {
    stop("Use tidy_ild_msm_bootstrap() for objects from ild_msm_bootstrap(); tidy_ild_model() expects a model fit.",
         call. = FALSE)
  }
  se <- match.arg(se)
  robust_type <- match.arg(robust_type)
  q <- (1 - conf_level) / 2
  engine <- if (inherits(fit, "lme")) "lme" else "lmer"
  ar1 <- isTRUE(attr(fit, "ild_ar1", exact = TRUE))
  if (inherits(fit, "lme")) {
    tt <- summary(fit)$tTable
    if (is.null(tt)) stop("Could not extract fixed-effect table from lme fit.", call. = FALSE)
    est <- as.vector(tt[, "Value"])
    se_vec <- as.vector(tt[, "Std.Error"])
    pval <- as.vector(tt[, "p-value"])
    terms <- rownames(tt)
    df_use <- tt[, "DF"]
  } else if (inherits(fit, "lmerMod")) {
    cc <- summary(fit)$coefficients
    if (is.null(cc)) stop("Could not extract coefficient table from lmer fit.", call. = FALSE)
    est <- as.vector(cc[, "Estimate"])
    se_vec <- as.vector(cc[, "Std. Error"])
    pval <- if ("Pr(>|t|)" %in% colnames(cc)) {
      as.vector(cc[, "Pr(>|t|)"])
    } else {
      rep(NA_real_, length(est))
    }
    terms <- rownames(cc)
    df_use <- NULL
  } else {
    stop("tidy_ild_model() supports only fits from ild_lme() (lme or lmerMod).", call. = FALSE)
  }
  if (se == "robust") {
    rlang::check_installed("clubSandwich", reason = "to use robust standard errors (se = \"robust\")")
    rob <- ild_robust_se(fit, type = robust_type, ...)
    vc <- rob$vcov
    if (nrow(vc) != length(est) || ncol(vc) != length(est)) {
      stop("Robust vcov dimension does not match fixed-effect length.", call. = FALSE)
    }
    se_vec <- sqrt(diag(vc))
    pval <- 2 * (1 - stats::pnorm(abs(est / se_vec)))
  }
  conf_low <- if (se == "robust" || inherits(fit, "lmerMod")) {
    est + stats::qnorm(q) * se_vec
  } else {
    est + stats::qt(q, df_use) * se_vec
  }
  conf_high <- if (se == "robust" || inherits(fit, "lmerMod")) {
    est + stats::qnorm(1 - q) * se_vec
  } else {
    est + stats::qt(1 - q, df_use) * se_vec
  }
  stat_vec <- ild_tidy_extract_statistic(fit, est, se_vec, se_mode = se)
  int_type <- rep(ild_tidy_interval_type_frequentist(se), length(est))
  mc <- ild_tidy_model_class_string(fit)
  tbl <- ild_tidy_assemble(
    term = terms,
    estimate = est,
    std_error = se_vec,
    conf_low = conf_low,
    conf_high = conf_high,
    p_value = pval,
    statistic = stat_vec,
    interval_type = int_type,
    engine = engine,
    model_class = mc,
    component = rep("fixed", length(terms)),
    effect_level = NULL,
    optional = NULL
  )
  if (!object) return(tbl)
  out <- list(
    meta = list(engine = engine, ar1 = ar1, se = se),
    table = tbl
  )
  class(out) <- "tidyild_model"
  out
}

#' @export
print.tidyild_model <- function(x, ...) {
  se_info <- if (identical(x$meta$se, "robust")) " robust SE" else ""
  cat("Fixed effects (", x$meta$engine, if (x$meta$ar1) ", AR1/CAR1" else "", se_info, ")\n", sep = "")
  show <- x$table
  keep <- intersect(
    c("term", "estimate", "std_error", "conf_low", "conf_high", "component", "effect_level", "p_value"),
    names(show)
  )
  print(show[, keep, drop = FALSE])
  invisible(x)
}

#' Compute inverse-probability-of-observation weights from a missingness model
#'
#' Uses the fitted missingness model from [ild_missing_model()] to compute
#' weights: unstabilized \code{w = 1 / p_obs} or stabilized
#' \code{w = mean(p_obs) / p_obs}. Weights are trimmed to avoid extremes.
#' Use with [ild_ipw_refit()] for sensitivity analysis. This is diagnostic
#' and sensitivity tooling, not a full MNAR solution.
#'
#' Also a **section provider** for [ild_diagnose()] when \code{.ipw} is present on the
#' analysis data (see [ild_diagnostics_utilities]).
#'
#' @section Bundle integration:
#' Adds \code{.ipw} to \code{x}. When that column is on the ILD data attached to the fit,
#' [ild_diagnose()] fills \code{causal} (weight summaries) on \code{\link{ild_diagnostics_bundle}}.
#'
#' @family ild_diagnostics_utilities
#' @param x An ILD object (see [is_ild()]); rows must align to the data used to fit \code{miss_fit}.
#' @param miss_fit The return value of [ild_missing_model()] (list with \code{fit}, \code{p_missing}, etc.).
#' @param stabilize Logical. If \code{TRUE} (default), use stabilized weights
#'   \code{mean(p_obs) / p_obs}; otherwise \code{1 / p_obs}.
#' @param trim Numeric of length 2. Quantiles to trim weights (default \code{c(0.01, 0.99)}).
#'   Weights below/above these quantiles are set to the quantile values.
#' @return \code{x} with an added column \code{.ipw} (numeric). ILD attributes are preserved.
#' @seealso [ild_diagnose()], [ild_diagnostics_bundle()], [ild_missing_model()],
#'   [ild_iptw_weights()], [ild_ipcw_weights()], [ild_joint_msm_weights()]
#' @export
#' @examples
#' set.seed(1)
#' d <- ild_simulate(n_id = 15, n_obs_per = 8, seed = 1)
#' d$stress <- rnorm(nrow(d))
#' d$mood <- d$y
#' d$mood[sample(nrow(d), 20)] <- NA
#' x <- ild_prepare(d, id = "id", time = "time")
#' mm <- ild_missing_model(x, "mood", "stress")
#' xw <- ild_ipw_weights(x, mm, stabilize = TRUE)
#' summary(xw$.ipw)
ild_ipw_weights <- function(x, miss_fit, stabilize = TRUE, trim = c(0.01, 0.99)) {
  validate_ild(x)
  if (is.null(miss_fit$fit)) {
    stop("miss_fit has no fit (e.g. no missingness). Use ild_missing_model() first.", call. = FALSE)
  }
  p_missing <- miss_fit$p_missing
  if (is.null(p_missing) || length(p_missing) != nrow(x)) {
    stop("miss_fit p_missing length must match nrow(x). Re-fit ild_missing_model() with this data.", call. = FALSE)
  }
  p_obs <- 1 - p_missing
  p_obs <- pmax(pmin(p_obs, 1), 1e-6)
  w <- if (stabilize) mean(p_obs, na.rm = TRUE) / p_obs else 1 / p_obs
  if (length(trim) >= 2L && is.numeric(trim)) {
    w <- .ild_ipw_trim(w, trim)
  }
  attrs <- attributes(x)
  x$.ipw <- w
  x <- .ild_ipw_restore_attrs(x, attrs)
  x <- ild_add_step(x, "ild_ipw_weights",
    list(stabilize = stabilize, trim = trim),
    list(created = ".ipw")
  )
  x
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
# Optional: set `cache = TRUE` on slow chunks when developing locally (see chunks below).

## ----recovery-run, cache = FALSE----------------------------------------------
library(tidyILD)
library(dplyr)
library(ggplot2)

truth <- 0.35
n_sim <- 200L

set.seed(2026)
res <- ild_power(
  formula = y ~ x + (1 | id),
  n_sim = n_sim,
  n_id = 25L,
  n_obs_per = 12L,
  effect_size = truth,
  seed = 2026L,
  return_sims = TRUE,
  verbose = FALSE
)

sim <- res$sim_results |> dplyr::filter(converged)

## ----recovery-metrics---------------------------------------------------------
ild_recovery_metrics(res$sim_results, truth = truth, level = 0.95)

## ----recovery-plot, fig.alt = "Histogram of simulation estimates of the fixed effect", cache = FALSE----
ggplot2::ggplot(sim, ggplot2::aes(x = estimate)) +
  ggplot2::geom_histogram(fill = "steelblue", color = "white", bins = 20) +
  ggplot2::geom_vline(xintercept = truth, linetype = 2, linewidth = 0.8) +
  ggplot2::labs(
    x = "Estimated coefficient for x",
    y = "Count",
    title = "Sampling distribution of fixed-effect estimates",
    subtitle = sprintf("True beta = %s (vertical line)", truth)
  )

## ----power-link---------------------------------------------------------------
tibble::tibble(
  power = res$power,
  n_reject = res$n_reject,
  n_converged = res$n_converged,
  alpha = res$alpha
)

## ----varcorr-snapshot, eval = requireNamespace("lme4", quietly = TRUE)--------
set.seed(1)
d_one <- ild_simulate(n_id = 30L, n_obs_per = 15L, seed = 99)
d_one$x <- rnorm(nrow(d_one))
d_one$y <- d_one$y + 0.3 * d_one$x
prep_one <- ild_prepare(d_one, id = "id", time = "time")
fit_one <- ild_lme(y ~ x + (1 | id), prep_one, ar1 = FALSE, warn_no_ar1 = FALSE)
lme4::VarCorr(fit_one)

## ----ar1-illustration, eval = requireNamespace("nlme", quietly = TRUE)--------
set.seed(2)
d_ar <- ild_simulate(n_id = 20L, n_obs_per = 14L, ar1 = 0.35, wp_effect = 0.5, seed = 2)
d_ar$x <- rnorm(nrow(d_ar))
d_ar$y <- d_ar$y + 0.25 * d_ar$x
prep_ar <- ild_prepare(d_ar, id = "id", time = "time")
fit_ar <- tryCatch(
  ild_lme(y ~ x, prep_ar, ar1 = TRUE, random = ~ 1 | id, warn_no_ar1 = FALSE),
  error = function(e) NULL
)
if (!is.null(fit_ar)) summary(fit_ar) else "nlme fit failed on this platform (skip)"

## ----session-info, echo = FALSE-----------------------------------------------
sessionInfo()

#' Inverse probability of censoring weights (IPCW) for monotone dropout
#'
#' Builds **discrete-time** IPCW weights for **monotone loss-to-follow-up**:
#' each person contributes a sequence of observed visits; after the last visit
#' the person is censored. This differs from [ild_missing_model()] + [ild_ipw_weights()],
#' which model **sporadic item missingness** on an outcome.
#'
#' Internally, each person-occasion row is labeled with an indicator
#' \code{drop_next}: whether the person drops before the next scheduled occasion
#' (including the last observed row). A pooled logistic regression models
#' \code{drop_next} given covariates; stabilized or unstabilized inverse
#' probabilities are accumulated within person (product over follow-up) to yield
#' \code{.ipw_censor}.
#'
#' **Assumptions:** Monotone censoring, positivity, correct censoring model.
#'
#' @section Bundle integration:
#' Adds \code{.ipw_censor}. Combine with [ild_iptw_weights()] or
#' [ild_iptw_msm_weights()] via [ild_joint_msm_weights()].
#'
#' @family ild_diagnostics_utilities
#' @param x An ILD object (see [is_ild()]).
#' @param predictors Character vector of covariate names (time-varying or baseline
#'   carried forward in \code{x}).
#' @param stabilize Logical. If \code{TRUE} (default), multiply factors
#'   \code{mean(1 - p_drop) / (1 - p_drop)} at each row before cumulating; if
#'   \code{FALSE}, use \code{1 / (1 - p_drop)}.
#' @param trim Numeric of length 2. Quantiles applied to **final** cumulative
#'   \code{.ipw_censor} (default \code{c(0.01, 0.99)}).
#' @param ... Passed to \code{stats::glm()}.
#' @return \code{x} with added column \code{.ipw_censor}. Attribute
#'   \code{ild_ipcw_fit} holds the fitted \code{glm}.
#' @seealso [ild_iptw_weights()], [ild_joint_msm_weights()]
#' @export
#' @examples
#' set.seed(3)
#' d <- ild_simulate(n_id = 18, n_obs_per = 7, seed = 3)
#' d$stress <- rnorm(nrow(d))
#' x <- ild_prepare(d, id = "id", time = "time")
#' x <- ild_ipcw_weights(x, predictors = "stress")
#' summary(x$.ipw_censor)
ild_ipcw_weights <- function(x, predictors, stabilize = TRUE,
                             trim = c(0.01, 0.99), ...) {
  validate_ild(x)
  nms <- names(x)
  for (p in predictors) {
    if (!p %in% nms) stop("predictor '", p, "' not found in data.", call. = FALSE)
  }
  id <- x[[".ild_id"]]
  seqv <- x[[".ild_seq"]]
  n <- nrow(x)
  ord <- order(id, seqv)
  ido <- id[ord]
  seqo <- seqv[ord]
  drop_o <- rep(NA_real_, n)
  i <- 1L
  while (i <= n) {
    g <- ido[i]
    j <- i
    while (j < n && ido[j + 1L] == g) j <- j + 1L
    for (k in i:j) {
      if (k == j) {
        drop_o[k] <- 1
      } else {
        sk <- as.numeric(seqo[k])
        sn <- as.numeric(seqo[k + 1L])
        drop_o[k] <- if (sn == sk + 1L) 0 else 1
      }
    }
    i <- j + 1L
  }
  drop_next <- rep(NA_real_, n)
  drop_next[ord] <- drop_o
  id_col <- ".ild_id"
  pred_cols <- if (length(predictors) > 0L) c(id_col, predictors) else id_col
  df <- as.data.frame(x[, pred_cols, drop = FALSE])
  df$.ild_drop_next <- drop_next
  rhs <- if (length(predictors) > 0L) paste(predictors, collapse = " + ") else "1"
  form <- stats::as.formula(paste(".ild_drop_next ~", rhs))
  fit <- stats::glm(form, data = df, family = stats::binomial(), ...)
  p_drop <- stats::predict(fit, newdata = df, type = "response")
  p_drop <- pmax(pmin(p_drop, 1 - 1e-6), 1e-6)
  p_stay <- 1 - p_drop
  if (stabilize) {
    ms <- mean(p_stay, na.rm = TRUE)
    ms <- max(ms, 1e-6)
    contrib <- ms / p_stay
  } else {
    contrib <- 1 / p_stay
  }
  contrib_ord <- contrib[ord]
  w_ord <- rep(NA_real_, n)
  i <- 1L
  while (i <= n) {
    g <- ido[i]
    j <- i
    while (j < n && ido[j + 1L] == g) j <- j + 1L
    idx <- i:j
    w_ord[idx] <- cumprod(contrib_ord[idx])
    i <- j + 1L
  }
  w <- rep(NA_real_, n)
  w[ord] <- w_ord
  if (length(trim) >= 2L && is.numeric(trim)) {
    w <- .ild_ipw_trim(w, trim)
  }
  attrs <- attributes(x)
  x$.ipw_censor <- w
  x <- .ild_ipw_restore_attrs(x, attrs)
  attr(x, "ild_ipcw_fit") <- fit
  ild_add_step(x, "ild_ipcw_weights",
    list(predictors = predictors, stabilize = stabilize, trim = trim),
    list(created = ".ipw_censor")
  )
}

# Continuous-time state-space models via ctsem (optional dependency)

#' Fit a continuous-time state-space model via ctsem
#'
#' First-class tidyILD wrapper for \pkg{ctsem}. This backend is designed for
#' irregularly spaced ILD where continuous-time dynamics are preferable to
#' discrete-time approximations.
#'
#' v1 focuses on a conservative single-outcome pathway with explicit class,
#' provenance, and diagnostics compatibility. Users may supply a custom ctsem
#' model object; otherwise a simple 1-manifest local-level style template is used.
#'
#' @param data An object that passes [validate_ild()] (typically after [ild_prepare()]).
#' @param outcome Character name of numeric outcome column.
#' @param ct_model Optional ctsem model object. If \code{NULL}, a conservative
#'   default model is created: for \code{model_type = "stanct"}, \code{ctsem::ctModel(type = "ct", ...)}
#'   with one manifest loading on one latent; for \code{"ctfit"}, \code{type = "omx"} with
#'   \code{Tpoints = max(rows per id, 2)} from the prepared long data (custom \code{ct_model}
#'   may still be needed for irregular designs).
#' @param model_type Character. \code{"stanct"} (default) or \code{"ctfit"}.
#' @param id_col Person-id column in \code{data} (default \code{".ild_id"}).
#' @param time_col Continuous-time index column in \code{data}
#'   (default \code{".ild_time_num"}).
#' @param time_scale Optional positive scalar to rescale \code{time_col}
#'   before fitting (default \code{1}).
#' @param ... Passed to \code{ctsem::ctStanFit()} or \code{ctsem::ctFit()}.
#' @return An object of class \code{c("ild_fit_ctsem", "ild_fit_model", "ild_analysis")}.
#'   Attributes \code{ild_data} and \code{ild_provenance} are attached.
#' @seealso [ild_tidy()], [ild_augment()], [ild_diagnose()], [ild_autoplot()].
#' @export
ild_ctsem <- function(data,
                      outcome,
                      ct_model = NULL,
                      model_type = c("stanct", "ctfit"),
                      id_col = ".ild_id",
                      time_col = ".ild_time_num",
                      time_scale = 1,
                      ...) {
  rlang::check_installed("ctsem", reason = "to use ild_ctsem()")
  validate_ild(data)
  model_type <- match.arg(model_type)
  if (missing(outcome) || !nzchar(as.character(outcome)[1L])) {
    stop("outcome must be a non-empty column name.", call. = FALSE)
  }
  outcome <- as.character(outcome)[1L]
  id_col <- as.character(id_col)[1L]
  time_col <- as.character(time_col)[1L]
  if (!id_col %in% names(data)) {
    stop("id_col '", id_col, "' not found in data.", call. = FALSE)
  }
  if (!time_col %in% names(data)) {
    stop("time_col '", time_col, "' not found in data.", call. = FALSE)
  }
  if (!outcome %in% names(data)) {
    stop("outcome column '", outcome, "' not found in data.", call. = FALSE)
  }
  time_scale <- as.numeric(time_scale)[1L]
  if (!is.finite(time_scale) || time_scale <= 0) {
    stop("time_scale must be a positive finite scalar.", call. = FALSE)
  }

  dat <- data[, c(id_col, time_col, outcome), drop = FALSE]
  names(dat) <- c("id", "time", "y")
  dat$y <- suppressWarnings(as.numeric(dat$y))
  dat$time <- suppressWarnings(as.numeric(dat$time)) / time_scale
  if (any(!is.finite(dat$time))) {
    stop("time column contains non-finite values after scaling.", call. = FALSE)
  }
  if (sum(is.finite(dat$y)) < 5L) {
    stop("Need at least 5 finite outcome values to fit ctsem model.", call. = FALSE)
  }
  dat <- dat[is.finite(dat$y), , drop = FALSE]

  if (is.null(ct_model)) {
    max_tp <- max(2L, as.integer(max(table(dat$id))))
    ct_model <- .ild_ctsem_default_model(
      manifest_name = "y",
      model_type = model_type,
      max_Tpoints = max_tp
    )
  }

  fit <- .ild_ctsem_fit_dispatch(
    dat = dat,
    ct_model = ct_model,
    model_type = model_type,
    dots = list(...)
  )

  out <- list(
    ct_fit = fit$fit,
    fit_call = fit$fit_call,
    fit_type = fit$fit_type,
    ct_model = ct_model,
    outcome = outcome,
    id_col = id_col,
    time_col = time_col,
    time_scale = time_scale,
    n_obs = nrow(dat),
    n_id = length(unique(dat$id))
  )
  class(out) <- c("ild_fit_ctsem", "ild_fit_model", "ild_analysis")

  attr(out, "ild_data") <- data
  attr(out, "ild_response") <- outcome
  attr(out, "ild_provenance") <- ild_new_analysis_provenance(
    data,
    "ild_ctsem",
    args = list(
      outcome = outcome,
      model_type = fit$fit_type,
      id_col = id_col,
      time_col = time_col,
      time_scale = time_scale
    ),
    outputs = list(
      n_obs = nrow(dat),
      n_id = length(unique(dat$id)),
      converged = .ild_ctsem_is_converged(out),
      ctsem_version = tryCatch(as.character(utils::packageVersion("ctsem")), error = function(e) NA_character_)
    )
  )
  out
}

#' @export
print.ild_fit_ctsem <- function(x, ...) {
  cat("tidyILD ctsem fit\n")
  cat("  outcome: ", x$outcome, "\n", sep = "")
  cat("  fit_type: ", x$fit_type, "\n", sep = "")
  cat("  n_obs: ", x$n_obs, "\n", sep = "")
  cat("  n_id: ", x$n_id, "\n", sep = "")
  cat("  converged: ", .ild_ctsem_is_converged(x), "\n", sep = "")
  invisible(x)
}

#' @keywords internal
#' @noRd
.ild_ctsem_default_model <- function(manifest_name = "y",
                                      model_type = c("stanct", "ctfit"),
                                      max_Tpoints = 2L) {
  model_type <- match.arg(model_type)
  lam <- matrix(1, nrow = 1L, ncol = 1L)
  if (identical(model_type, "stanct")) {
    ctsem::ctModel(
      type = "ct",
      n.manifest = 1L,
      n.latent = 1L,
      manifestNames = manifest_name,
      latentNames = "eta",
      LAMBDA = lam
    )
  } else {
    tp <- max(2L, as.integer(max_Tpoints)[1L])
    ctsem::ctModel(
      type = "omx",
      n.manifest = 1L,
      n.latent = 1L,
      Tpoints = tp,
      manifestNames = manifest_name,
      latentNames = "eta",
      LAMBDA = lam
    )
  }
}

#' @keywords internal
#' @noRd
.ild_ctsem_fit_dispatch <- function(dat, ct_model, model_type, dots = list()) {
  if (identical(model_type, "stanct")) {
    tries <- list(
      c(list(datalong = dat, ctstanmodel = ct_model), dots),
      c(list(data = dat, ctstanmodel = ct_model), dots)
    )
    for (args in tries) {
      fit <- tryCatch(do.call(ctsem::ctStanFit, args), error = function(e) e)
      if (!inherits(fit, "error")) {
        return(list(fit = fit, fit_type = "stanct", fit_call = "ctStanFit"))
      }
    }
    stop("ctStanFit failed for provided data/model; verify ct_model compatibility.", call. = FALSE)
  }

  tries <- list(
    c(list(data = dat, ctmodelobj = ct_model), dots),
    c(list(datalong = dat, ctmodelobj = ct_model), dots)
  )
  for (args in tries) {
    fit <- tryCatch(do.call(ctsem::ctFit, args), error = function(e) e)
    if (!inherits(fit, "error")) {
      return(list(fit = fit, fit_type = "ctfit", fit_call = "ctFit"))
    }
  }
  stop("ctFit failed for provided data/model; verify ct_model compatibility.", call. = FALSE)
}

#' @keywords internal
#' @noRd
.ild_ctsem_is_converged <- function(object) {
  fit <- if (inherits(object, "ild_fit_ctsem")) object$ct_fit else object
  if (is.null(fit)) return(NA)
  cv <- tryCatch(fit$optimization$convergence, error = function(e) NULL)
  if (is.null(cv)) cv <- tryCatch(fit$opt$convergence, error = function(e) NULL)
  if (is.null(cv)) cv <- tryCatch(fit$stanfit@sim$adaptation_info, error = function(e) NULL)
  if (is.numeric(cv) && length(cv) > 0L) return(isTRUE(cv[1L] == 0))
  if (is.logical(cv) && length(cv) > 0L) return(isTRUE(cv[1L]))
  NA
}

# ild_diagnose() for ild_fit_kfas — diagnostics-centered bundle (KFAS)

#' @rdname ild_diagnose
#' @method ild_diagnose ild_fit_kfas
#' @export
ild_diagnose.ild_fit_kfas <- function(object,
                                      data = NULL,
                                      type = NULL,
                                      by_id = NULL,
                                      ...) {
  rlang::check_installed("KFAS", reason = "to diagnose ild_kfas fits")
  data <- if (is.null(data)) attr(object, "ild_data", exact = TRUE) else data
  if (is.null(data)) {
    stop("ild_diagnose() for ild_fit_kfas requires `data` or attr(fit, \"ild_data\").", call. = FALSE)
  }
  validate_ild(data)
  preds <- NULL
  rn <- attr(object, "ild_response", exact = TRUE)
  outcome_vars <- unique(c(stats::na.omit(c(rn, preds))))
  outcome_vars <- outcome_vars[outcome_vars %in% names(data)]

  kfs <- object$kfs
  y <- data[[rn]]
  y <- if (is.numeric(y)) y else suppressWarnings(as.numeric(y))
  res_diag <- ild_kfas_residual_diagnostics(kfs, y)

  innov <- tryCatch(as.numeric(stats::rstandard(kfs)), error = function(e) NULL)
  ex <- ild_kfas_extract_augment(kfs, y, TRUE)
  mae <- mean(abs(ex$resid), na.rm = TRUE)
  rmse <- sqrt(mean(ex$resid^2, na.rm = TRUE))
  bias <- mean(ex$resid, na.rm = TRUE)

  opt <- object$fit_ssm$optim.out
  loglik <- tryCatch(as.numeric(kfs$logLik), error = function(e) NA_real_)
  diffuse_note <- NULL
  ddiff <- tryCatch(as.integer(kfs$d), error = function(e) NA_integer_)
  if (is.finite(ddiff) && ddiff > 0L) {
    diffuse_note <- sprintf("Diffuse initialization phase ended after %d time index(es); see KFS() output.", ddiff)
  }

  spec <- object$spec
  map <- object$mapping
  pre <- object$preprocessing

  has_wpbp <- any(grepl("_wp$|_bp$", names(data)))

  fd <- list(
    engine = "KFAS",
    state_spec = spec$state_spec,
    observation_family = spec$observation_family,
    convergence = list(
      converged = tryCatch(isTRUE(opt$convergence == 0L), error = function(e) NA),
      note = "ML via KFAS::fitSSM; see fit$fit_ssm$optim.out."
    ),
    rank = list(note = "Not applicable (SSModel)."),
    optimizer = list(
      method = tryCatch(opt$method, error = function(e) NA_character_),
      convergence_code = tryCatch(opt$convergence, error = function(e) NA_integer_),
      value = tryCatch(opt$value, error = function(e) NA_real_)
    ),
    log_likelihood = loglik,
    diffuse_initialization = list(
      d = ddiff,
      note = diffuse_note
    ),
    state_dimension = map$state_dimension,
    observation_dimension = map$observation_dimension,
    residual_correlation = list(
      modeled = TRUE,
      note = paste(
        "Temporal dependence is modeled through the latent state (",
        spec$state_spec,
        "); standardized one-step-ahead innovations are primary residual objects.",
        sep = ""
      )
    ),
    random_effects_structure = list(
      note = "Latent states in fit$kfs$alphahat; see ild_tidy_states()."
    )
  )

  pred <- list(
    obs_metrics = list(
      engine = "KFAS",
      n = nrow(data),
      mae = mae,
      rmse = rmse,
      mean_residual = bias,
      mean_abs_innovation = if (!is.null(innov)) mean(abs(innov), na.rm = TRUE) else NA_real_
    ),
    simulation_checks = list(ppc = NULL),
    one_step_ahead = list(
      standardized_innovations = res_diag$standardized_innovations_summary
    )
  )

  residual_block <- list(
    engine = "KFAS",
    standardized_prediction_errors = res_diag$standardized_innovations_summary,
    acf = res_diag$acf_innovations,
    qq = res_diag$qq,
    outliers = res_diag$outlier_flags,
    ljung_box = res_diag$ljung_box,
    legacy_ild_diagnostics = NULL
  )

  warn_rows <- list()
  if (!is.null(opt) && !is.null(opt$message) && nzchar(paste(opt$message, collapse = ""))) {
    warn_rows[[length(warn_rows) + 1L]] <- list(
      source = "KFAS",
      level = "note",
      message = paste(as.character(opt$message), collapse = " "),
      code = "kfas_optim_message"
    )
  }
  qv <- ild_kfas_qh_scalar(object$kfas_model$Q)
  hv <- ild_kfas_qh_scalar(object$kfas_model$H)
  if (is.finite(qv) && qv < 1e-8) {
    warn_rows[[length(warn_rows) + 1L]] <- list(
      source = "tidyILD",
      level = "warning",
      message = "Estimated state variance Q is effectively zero (boundary).",
      code = "kfas_degenerate_Q"
    )
  }
  if (is.finite(hv) && hv < 1e-8) {
    warn_rows[[length(warn_rows) + 1L]] <- list(
      source = "tidyILD",
      level = "warning",
      message = "Estimated observation variance H is effectively zero (boundary).",
      code = "kfas_degenerate_H"
    )
  }

  warn_df <- if (length(warn_rows) > 0L) {
    tibble::tibble(
      source = vapply(warn_rows, function(z) z$source, character(1L)),
      level = vapply(warn_rows, function(z) z$level, character(1L)),
      message = vapply(warn_rows, function(z) z$message, character(1L)),
      code = vapply(warn_rows, function(z) z$code, character(1L))
    )
  } else {
    tibble::tibble()
  }

  sm_data <- tryCatch(ild_summary(data), error = function(e) NULL)
  time_span <- if (!is.null(sm_data)) {
    list(
      min = sm_data$summary$time_min[1],
      max = sm_data$summary$time_max[1],
      median_dt_sec = sm_data$summary$median_dt_sec[1],
      iqr_dt_sec = sm_data$summary$iqr_dt_sec[1]
    )
  } else {
    NULL
  }

  data_sec <- fill_diagnostics_data(data, outcome_vars = outcome_vars)
  data_sec$intervals_summary <- time_span
  data_sec$n_ids <- dplyr::n_distinct(data[[".ild_id"]])
  data_sec$kfas_preprocessing <- pre

  design_sec <- fill_diagnostics_design(data, vars = NULL)
  design_sec$irregular_spacing_class <- tryCatch(
    ild_spacing_class(data),
    error = function(e) NA_character_
  )
  design_sec$predictors_wp_bp_decomposed <- has_wpbp

  bundle <- ild_diagnostics_bundle(
    meta = list(
      engine = "KFAS",
      model_type = spec$state_spec,
      state_spec = spec$state_spec,
      observation_family = spec$observation_family,
      smoother = isTRUE(map$smoother),
      filter_smoother_settings = list(
        kfs_smoothing = isTRUE(map$smoother),
        kfs_filtering = "state"
      ),
      pool_mode = map$pool_mode,
      fit_context = spec$fit_context,
      n_obs = nrow(data),
      n_id = dplyr::n_distinct(data[[".ild_id"]]),
      time_units = spec$time_units
    ),
    data = data_sec,
    design = design_sec,
    fit = fd,
    residual = residual_block,
    predictive = pred,
    missingness = fill_diagnostics_missingness_section(data, vars = outcome_vars),
    causal = fill_diagnostics_causal(data, causal_detail = FALSE),
    warnings = warn_df,
    guardrails = guardrails_empty_tibble(),
    summary_text = character()
  )

  bundle$guardrails <- evaluate_guardrails_kfas(object, data, bundle)
  bundle$summary_text <- build_diagnostics_bundle_summary(bundle)
  bundle <- enrich_bundle_semantic_sections(bundle)
  attr(bundle, "ild_fit") <- object
  attr(bundle, "ild_data") <- data
  bundle
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)

## ----export-pattern-----------------------------------------------------------
library(tidyILD)
set.seed(1)
d <- ild_simulate(n_id = 8, n_obs_per = 10, seed = 1)
d$x <- rnorm(nrow(d))
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_center(x, y, x)
x <- ild_lag(x, dplyr::all_of(c("y", "x")), n = 1L, mode = "gap_aware")

meta <- ild_meta(x)
meta$ild_id
# Plain data frame for any external package:
dat <- as.data.frame(x)
head(dat[, c(meta$ild_id, ".ild_time_num", "y", "y_wp", "y_bp", "x_lag1")])

## ----stub-dynamite, eval = FALSE----------------------------------------------
# # library(dynamite)
# # After building `dat` from an ILD object (see above):
# # - Map id/time/outcome columns to dynamite's expected data layout.
# # - Specify channels, lags, and priors per package documentation.
# # Example placeholder only (not valid without a real dynamite specification):
# # fit_dyn <- dynamite::dynamite(
# #   dformula = <your dformula>,
# #   data = dat,
# #   ...
# # )

## ----stub-pgee, eval = FALSE--------------------------------------------------
# # library(PGEE)
# # Penalized GEE expects a long data frame with id, repeated outcome, and a matrix
# # or formula interface for high-dimensional covariates — see ?PGEE::PGEE.
# # Use `dat` from tidyILD after centering/lagging so covariates align with your estimand.
# # fit_pgee <- PGEE::PGEE(<formula or design>, data = dat, ...)

## ----stub-lavaan, eval = FALSE------------------------------------------------
# # library(lavaan)
# # Dynamic SEM is model-syntax-specific; export `dat` and define your model
# # in lavaan's longitudinal / DSEM extensions. tidyILD does not generate lavaan syntax.

## ----session-info, echo = FALSE-----------------------------------------------
sessionInfo()

#' One-shot summary of an ILD object
#'
#' Reports number of persons, number of observations, time range,
#' descriptive spacing (median/IQR of intervals, percent gaps), and
#' duplicate info. Uses [ild_meta()] and `.ild_*` columns only.
#' No hard "regular"/"irregular" label; use [ild_spacing_class()] for that.
#'
#' @param x An ILD object (see [is_ild()]).
#' @return A list with elements: \code{summary} (one-row tibble with n_id, n_obs,
#'   time_min, time_max, prop_gap, median_dt_sec, iqr_dt_sec), \code{n_units},
#'   \code{n_obs}, \code{time_range}, \code{spacing}, \code{n_gaps}, \code{pct_gap}.
#'   If \code{ild_prepare()} was run on a \code{tbl_ts}, \code{tsibble} is also
#'   present (same as [ild_tsibble_meta()]). The \code{summary} tibble is the
#'   primary contract for programmatic use.
#' @importFrom tibble tibble
#' @export
ild_summary <- function(x) {
  validate_ild(x)
  meta <- ild_meta(x)
  time_num <- x[[".ild_time_num"]]
  n_gaps <- sum(x[[".ild_gap"]] %in% TRUE, na.rm = TRUE)
  n_units <- meta$ild_n_units
  n_obs <- meta$ild_n_obs
  time_range <- range(time_num, na.rm = TRUE)
  spacing <- meta$ild_spacing
  n_intervals <- if (!is.null(spacing$n_intervals)) spacing$n_intervals else sum(!is.na(x[[".ild_dt"]]))
  prop_gap <- if (n_intervals > 0) n_gaps / n_intervals else NA_real_
  median_dt_sec <- if (!is.null(spacing$median_dt)) spacing$median_dt else stats::median(x[[".ild_dt"]], na.rm = TRUE)
  iqr_dt_sec <- if (!is.null(spacing$iqr_dt)) spacing$iqr_dt else stats::IQR(x[[".ild_dt"]], na.rm = TRUE)
  summary_tbl <- tibble::tibble(
    n_id = n_units,
    n_obs = n_obs,
    time_min = time_range[1],
    time_max = time_range[2],
    prop_gap = prop_gap,
    median_dt_sec = median_dt_sec,
    iqr_dt_sec = iqr_dt_sec
  )
  out <- list(
    summary = summary_tbl,
    n_units = n_units,
    n_obs = n_obs,
    time_range = time_range,
    spacing = spacing,
    n_gaps = n_gaps,
    pct_gap = meta$ild_spacing$pct_gap
  )
  tsb <- ild_tsibble_meta(x)
  if (!is.null(tsb)) {
    out$tsibble <- tsb
  }
  out
}

#' Classify spacing as regular-ish vs irregular-ish
#'
#' Returns a simple classification for use in documentation or when choosing
#' correlation structure (e.g. AR1 vs CAR1 in [ild_lme()]). The rule is
#' documented and overridable via arguments. Does not change core ILD behavior.
#'
#' @param x An ILD object (see [is_ild()]).
#' @param cv_threshold Numeric. Coefficient of variation of within-person
#'   intervals above which spacing is "irregular-ish" (default 0.2).
#' @param pct_gap_threshold Numeric. Percent of intervals flagged as gaps
#'   above which spacing is "irregular-ish" (default 10).
#' @return Character: `"regular-ish"` or `"irregular-ish"`.
#' @seealso
#' \href{../doc/kfas-irregular-timing-spacing.html}{\code{vignette("kfas-irregular-timing-spacing", package = "tidyILD")}}
#' (KFAS and irregular timing). \href{../doc/ild-decomposition-and-spacing.html}{\code{vignette("ild-decomposition-and-spacing", package = "tidyILD")}}
#' (within-between and spacing).
#' @export
ild_spacing_class <- function(x,
                              cv_threshold = 0.2,
                              pct_gap_threshold = 10) {
  validate_ild(x)
  spacing <- attr(x, "ild_spacing", exact = TRUE)
  dt_median <- spacing$median_dt
  dt_iqr <- spacing$iqr_dt
  pct_gap <- spacing$pct_gap
  if (is.na(dt_median) || dt_median <= 0) return("irregular-ish")
  # CV of interval: use IQR/median as a robust proxy for variability
  cv <- if (is.na(dt_iqr) || dt_iqr == 0) 0 else dt_iqr / dt_median
  if (cv > cv_threshold) return("irregular-ish")
  if (!is.na(pct_gap) && pct_gap > pct_gap_threshold) return("irregular-ish")
  "regular-ish"
}

---
title: "State-space modeling in tidyILD with KFAS"
author: "Alex Litovchenko"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{State-space modeling in tidyILD with KFAS}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
has_kfas <- requireNamespace("KFAS", quietly = TRUE)
```

## What is a state-space model?

In a **state-space** (or **dynamic linear**) model, you observe a sequence \(y_1,\ldots,y_T\) and posit **latent** states \(\alpha_t\) that evolve over time and drive the observations. A minimal Gaussian **local level** model is:

- **State:** \(\alpha_t = \alpha_{t-1} + \eta_t\) with \(\eta_t \sim \mathcal{N}(0, Q)\) (random walk).
- **Observation:** \(y_t = \alpha_t + \varepsilon_t\) with \(\varepsilon_t \sim \mathcal{N}(0, H)\).

So the "level" of the process drifts slowly; the data are noisy measurements of that level. In **tidyILD**, `ild_kfas(..., state_spec = "local_level")` fits this structure (via **KFAS**) for a **single** time series per call—one distinct `.ild_id` after `ild_prepare()`.

## When use this instead of mixed-model residual correlation?

**Multilevel models** (`ild_lme()`, `ild_brms()`) are the right tool when you want **population** inference: fixed and random effects across many persons, within/between decomposition, and **residual** dynamics (e.g. AR1 or CAR1 on the **within-person** residuals) as a **nuisance** correlation structure.

**State-space** models in this package focus on **explicit latent dynamics** for **one** series at a time: estimating a **time-varying level** (or, in future specs, trend or AR) in **state space**, with diagnostics built on **one-step-ahead innovations** from the Kalman filter.

Conceptual contrast:

- **AR1 on residuals (nlme / lme):** correlation among **errors** around a smooth mean structure.
- **Local level (KFAS):** the **mean** itself is a **random walk** and is **smoothed**; the "residual" is often summarized as **standardized prediction errors** (innovations).

Neither replaces the other in general—choose based on whether your primary goal is **hierarchical population inference** or **structured univariate latent dynamics** for one person (or one series).

## Filtered vs smoothed states

In **ILD** terms:

- **Filtered** (“online” / **nowcast**): your best estimate of the latent level **at occasion *t* using only measurements up through *t***—what the model would have said about the state **at that moment** as data arrived. Useful when you think about **sequential** self-report or real-time summaries.
- **Smoothed** (“offline” / **full-history**): your best estimate of the level **at each occasion using the entire series**—what you report after seeing **all** waves, including revising earlier time points. This is usually what you want for **scientific summaries** of a completed diary or EMA study.

Formally, after fitting, **KFAS** runs [`KFS()`](https://CRAN.R-project.org/package=KFAS) to obtain:

- **Filtered** state: \(E(\alpha_t \mid y_1,\ldots,y_t)\). In KFAS output this is often `att`.
- **Smoothed** state: \(E(\alpha_t \mid y_1,\ldots,y_T)\). In KFAS output this is often `alphahat`.

In **tidyILD**, `ild_kfas(..., smoother = TRUE)` requests smoothing in `KFS()`; when `FALSE`, smoothed states may be unavailable or less central. Use `ild_plot_filtered_vs_smoothed()` to compare the first latent state over time.

## Minimal example

If the **KFAS** package is installed, you can run:

```{r example, eval = has_kfas}
library(tidyILD)
set.seed(1)
d <- ild_simulate(n_id = 1, n_obs_per = 60, seed = 42)
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_center(x, y)
fit <- suppressWarnings(
  ild_kfas(x, outcome = "y", state_spec = "local_level", time_units = "sim_steps")
)
b <- ild_diagnose(fit)
class(b)
ild_autoplot(b, section = "residual", type = "acf")
```

If **KFAS** is not installed, install it with `install.packages("KFAS")` and load **tidyILD**; the same code then runs end-to-end.

## What the backend does not yet do

Read this section before relying on **KFAS** in a paper or preregistration. The normative scope document `inst/dev/KFAS_V1_BACKEND.md` in the package source has full detail; the points below are the **trust** boundaries for **v1**.

**What `ild_kfas()` is:**

- **Discrete-time state-space** modeling: the latent state advances **one step per observation** (row order after `ild_prepare()` for that series). This is standard dynamic linear modeling on an **index**, not a continuous-time differential equation.

**What it is not (today):**

- **Not** **ctsem**-style (or similar) **continuous-time** latent dynamics with unequal physical intervals baked into the transition model. Those workflows are a **later tier**; this backend does not replace them for that use case.
- **Not** a **multilevel latent-state model**: there is **no** pooled latent trajectory across persons in v1. Fitting one series per call is the supported semantics; **pooling mode** across IDs is **limited** (see the backend doc). Stacking independent per-person fits is explicit via **`fit_context`** and guardrails—not hierarchical partial pooling of a shared state.
- **`local_level` only** in v1; other `state_spec` labels (`local_trend`, `ar1_state`, `regression_local_level`, …) are **reserved** for later releases.
- **Optional short-horizon forecasts** and richer **uncertainty** quantification are **planned**; see `?ild_plot_forecast` and `NEWS.md`.

**Irregular timing:** see `vignette("kfas-irregular-timing-spacing", package = "tidyILD")`—tidyILD **diagnoses** spacing; the KFAS wrapper fits under **discrete-time** choices and does not, by itself, “solve” irregular measurement in a continuous-time sense.

## See also

- `vignette("kfas-irregular-timing-spacing", package = "tidyILD")` — irregular measurement and spacing diagnostics.
- `vignette("kfas-choosing-backend", package = "tidyILD")` — **lme/nlme**, **brms**, and **KFAS**.
- `vignette("ild-decomposition-and-spacing", package = "tidyILD")` — within/between and **spacing** tools.

---
title: "Tsibble interoperability"
author: "tidyILD authors"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Tsibble interoperability}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
has_tsibble <- requireNamespace("tsibble", quietly = TRUE)
```

The [**tsibble**](https://tsibble.tidyverts.org/) package represents tidy temporal data as `tbl_ts` objects with a declared **key** (e.g. subject) and **index** (time). **tidyILD** works on ordinary tibbles with provenance in attributes; when you start from a `tbl_ts`, **tidyILD** records formal tsibble semantics so you can **inspect** them and **round-trip** back to a tsibble where appropriate.

This vignette requires the **tsibble** package (`install.packages("tsibble")`). Chunks below run only when **tsibble** is available.

## Ingesting a tsibble with `ild_prepare()`

Pass a **`tbl_ts`** to **`ild_prepare()`** in either of these ways:

1. **Omit `id` and `time`** — tidyILD infers the subject column from the tsibble **key** (exactly **one** key column) and the time column from the **index**.
2. **Supply `id` and `time`** — names must match the key and index columns; the input must still be a **`tbl_ts`**.

The **`tbl_ts`** class is **not** kept on the result: the prepared object is a tibble with ILD structure (`.ild_*` columns, etc.). Tsibble-specific information is copied into the **tidyILD** attribute **before** coercion.

```{r ingest, eval = has_tsibble}
suppressPackageStartupMessages({
  library(tidyILD)
  library(tsibble)
})
t <- seq.POSIXt(as.POSIXct("2020-01-01", tz = "UTC"), by = "hour", length.out = 4)
d <- tsibble(
  id = rep(1L, 4),
  t = t,
  y = 1:4,
  key = id,
  index = t
)
x <- ild_prepare(d)
class(x)
```

## What provenance is kept

When the source was a **`tbl_ts`**, metadata is stored under **`attr(x, "tidyILD")$tsibble`**. The exported reader is **`ild_tsibble_meta()`**, which returns **`NULL`** if the object was **not** prepared from a tsibble.

Typical fields include:

- **`key_vars`**, **`index_var`** — names of the subject and time columns.
- **`interval_format`** — human-readable summary of the declared **`tsibble::interval()`** (when available).
- **`is_regular`** — tsibble’s regularity flag at **input** time.
- **`index_type`**, **`tsibble_version`**, **`source_class`** — for reporting and reproducibility.

```{r meta, eval = has_tsibble}
str(ild_tsibble_meta(x), max.level = 1)
```

**`ild_spacing()`** may include a **`tsibble`** sub-list that **bridges** the declared tsibble interval/regularity to **empirical** spacing computed from **`.ild_dt`** after **`ild_prepare()`** (sorting, duplicate handling, etc.). Declared and empirical summaries can **diverge** if rows change.

```{r spacing-bridge, eval = has_tsibble}
sp <- ild_spacing(x)
names(sp$tsibble)
```

## How to inspect

- **`ild_tsibble_meta(x)`** — full tsibble provenance list (or **`NULL`**).
- **`ild_summary(x)`** — when tsibble metadata exists, the summary includes a **`tsibble`** element (same content as **`ild_tsibble_meta()`**).
- **`ild_spacing(x)$tsibble`** — declared vs empirical spacing bridge (when present).

```{r summary-ts, eval = has_tsibble}
names(ild_summary(x))
ild_summary(x)$tsibble[c("key_vars", "index_var", "interval_format", "is_regular")]
```

## Round-trip with `ild_as_tsibble()`

**`ild_as_tsibble(x)`** builds a **`tbl_ts`** from the prepared tibble using **`ild_meta()`** for key and index column names. If **`ild_tsibble_meta(x)`** is not **`NULL`**, **`regular`** is set from the stored **`is_regular`** so **`tsibble::interval()`** often **matches the original** for data that was **unchanged** after **`ild_prepare()`**.

```{r roundtrip, eval = has_tsibble}
d2 <- ild_as_tsibble(x)
tsibble::interval(d2)
identical(tsibble::interval(d), tsibble::interval(d2))
```

You can pass extra arguments through to **`tsibble::as_tsibble()`** (e.g. override **`regular`**).

## Limitations and policy

- **One-key policy:** **`ild_prepare()`** accepts a **`tbl_ts`** with **exactly one** key column (the subject id). Multiple keys (e.g. country + person) are not supported; reshape or aggregate first.
- **Class is dropped:** Output is not a **`tbl_ts`**; use **`ild_as_tsibble()`** when you need tsibble semantics again.
- **Pre- vs post-prepare divergence:** Declared interval and regularity describe the **input** **`tbl_ts`**. After **`ild_prepare()`**, empirical spacing (gaps, irregularity) comes from **`.ild_dt`**. Sorting, duplicate removal, or row drops can make **`ild_as_tsibble()`**’s reconstruction **not regular** even when the source was; in that case a **warning** may be emitted when reconstructed regularity disagrees with stored provenance.

For normative developer detail, see **`vignette("developer-contracts", package = "tidyILD")`** and **`inst/dev/DEVELOPER_CONTRACTS.md`** (tsibble section). Related: **`vignette("ild-decomposition-and-spacing", package = "tidyILD")`** for spacing tools.

---
title: "Temporal dynamics: choosing a model for ILD"
author: "tidyILD authors"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Temporal dynamics: choosing a model for ILD}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
```

This article is a **decision guide** for **temporal structure**: lags in the mean, residual autocorrelation, time-varying effects, and latent dynamics. It does not replace statistical theory or domain expertise; it maps **scientific questions** to **tidyILD entry points** you already have.

## Three axes before you fit anything

1. **Estimand** — Associational lag effect? Stability of the outcome (lagged outcome in the mean)? Causal effect under a clear identification strategy (see MSM vignettes)? Latent smooth trajectory?
2. **Spacing** — Occasions are **regular-ish** or **irregular-ish**? See `ild_spacing_class()`, `ild_spacing()`, and `vignette("ild-decomposition-and-spacing", package = "tidyILD")`.
3. **Where dynamics live** — In the **fixed-effects mean** (e.g. `y ~ x_lag1`), in **residual correlation** (AR1/CAR1 on within-person residuals), in a **smooth over calendar/study time** (`ild_tvem()`), or in a **latent state** (`ild_kfas()`, `ild_ctsem()`).

## Decision flow (conceptual)

The following diagram is a navigation aid (rendered on sites that support Mermaid; otherwise read the labels as a checklist). Source in Mermaid syntax:

    flowchart TD
      q1[Estimand clear?]
      spacing[Check ild_spacing_class]
      meanLag[Lags in mean: ild_lag ild_panel_lag_prepare ild_crosslag]
      residAR[Residual AR: ild_lme ar1=TRUE]
      tvem[Effect varies over time: ild_tvem]
      ct[Continuous-time latent: ild_ctsem]
      kfas[Discrete-time latent level: ild_kfas]
      q1 --> spacing
      spacing --> meanLag
      spacing --> residAR
      meanLag --> residAR
      spacing --> tvem
      spacing --> ct
      spacing --> kfas

- **Irregular spacing** makes **index lags** (row order) easy to misinterpret as equal time. Prefer **`gap_aware`** or **`time_window`** in `ild_lag()`, or align to a grid, or use **continuous-time** models when the estimand requires it (`vignette("kfas-irregular-timing-spacing", package = "tidyILD")`).
- **Residual ACF** after a lag-mean model may indicate **unmodeled autocorrelation**; compare to **`ild_lme(..., ar1 = TRUE)`** (nlme path with AR1 or CAR1 chosen from spacing). See `vignette("temporal-dynamics-model-choice", package = "tidyILD")` guardrails in `ild_diagnose()`.

## Feature map

| Question | tidyILD tools | Backend / notes | Not a substitute for |
|----------|---------------|-----------------|----------------------|
| Lagged predictor → outcome | `ild_lag()`, `ild_crosslag()`, `ild_panel_lag_prepare()` | `ild_lme` / `ild_brms` | Full panel VAR / DSEM (multivariate lag system) |
| Lagged outcome (stability) | Same lag helpers; model `y ~ y_lag1 + ...` carefully | Mixed model | Dynamic structural equation modeling software if that is the estimand |
| Residual serial correlation | `ild_lme(..., ar1 = TRUE)` | nlme AR1 or CAR1 | Does not add lagged **mean** structure by itself |
| Effect changes over study time | `ild_tvem()`, `ild_tvem_plot()` | mgcv GAM | Random **slopes over time** per person (consider `ild_brms` recipes) |
| Discrete-time latent level | `ild_kfas()` | KFAS | Pooled multilevel latent model across many IDs |
| Continuous-time latent dynamics | `ild_ctsem()` | ctsem / Stan | Quick lag regression on irregular data without CT assumptions |
| Compare a few fitted models | `ild_compare_fits()` | AIC/BIC where defined | Likelihood-ratio tests unless models are nested and comparable |
| Multivariate lags / feedback (joint system) | Same lag helpers; `ild_crosslag()` is one equation at a time | Export preprocessed data; see `vignette("ild-specialist-backends", package = "tidyILD")` | dynamite, lavaan DSEM, multivariate brms / ctsem |
| High-dimensional time-varying predictors (p >> n) | Unpenalized `ild_lme` / `lme4` is not designed for this | Same vignette: hand off to penalized longitudinal tools | PGEE and related methods |

## Minimal examples

```{r minimal-crosslag}
library(tidyILD)
set.seed(1)
d <- ild_simulate(n_id = 12, n_obs_per = 10, seed = 1)
x <- ild_prepare(d, id = "id", time = "time")
out <- ild_crosslag(x, y, y, lag = 1L, ar1 = FALSE, warn_no_ar1 = FALSE)
out$lag_term[, c("term", "estimate", "std_error")]
```

```{r minimal-ar1, eval = requireNamespace("nlme", quietly = TRUE)}
x2 <- ild_center(x, y)
fit_ar <- tryCatch(
  ild_lme(y ~ y_bp + y_wp, data = x2, ar1 = TRUE, warn_no_ar1 = FALSE, warn_uncentered = FALSE),
  error = function(e) NULL
)
if (!is.null(fit_ar)) {
  print(fit_ar)
} else {
  "nlme fit skipped on this platform"
}
```

```{r minimal-tvem}
x3 <- ild_simulate(n_id = 10, n_obs_per = 15, seed = 2)
x3$x <- rnorm(nrow(x3))
x3 <- ild_prepare(x3, id = "id", time = "time")
tv <- ild_tvem(x3, "y", "x", k = 5, re_id = TRUE)
summary(tv)
```

## Further reading

- `vignette("tidyILD-workflow", package = "tidyILD")`
- `vignette("ild-specialist-backends", package = "tidyILD")` (multivariate / high-p handoffs)
- `vignette("kfas-choosing-backend", package = "tidyILD")`
- `vignette("brms-dynamics-recipes", package = "tidyILD")` (Bayesian templates)
- `vignette("heterogeneity-interpretation", package = "tidyILD")` (person-specific slopes vs dynamics)

```{r session-info, echo = FALSE}
sessionInfo()
```

---
title: "Within-between decomposition and handling irregular spacing"
author: "tidyILD authors"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Within-between decomposition and handling irregular spacing}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
```

## Within-between decomposition

In intensive longitudinal data, predictors often vary both **between persons** (e.g. some people are higher on average) and **within person** (e.g. momentary fluctuations). Using the raw variable in a multilevel model conflates these two sources. tidyILD's `ild_center()` makes the decomposition explicit.

- **Between-person (BP)**: person mean of the variable.
- **Within-person (WP)**: deviation from the person mean (variable minus BP).

Use the WP component at level 1 (within-person effect) and the BP component at level 2 or in cross-level interactions to avoid ecological fallacy and conflation bias.

```{r center}
library(tidyILD)
d <- ild_simulate(n_id = 5, n_obs_per = 8, seed = 1)
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_center(x, y)
# New columns: y_bp (person mean), y_wp (deviation from person mean)
head(x[, c("id", "y", "y_bp", "y_wp")])
```

## Irregular spacing and lags

ILD often has irregular time spacing (e.g. EMA prompts at random times). Using `dplyr::lag()` assumes equal spacing and can produce misaligned lags. Use `ild_lag()` with `mode = "gap_aware"` so that lags are set to NA when the time distance to the previous observation exceeds a threshold.

```{r lag}
d <- ild_simulate(n_id = 3, n_obs_per = 6, irregular = TRUE, seed = 2)
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_lag(x, y, mode = "gap_aware", max_gap = 4000)
# Compare: .ild_dt (interval) and y_lag1 (NA after large gaps)
x[, c(".ild_id", ".ild_dt", "y", "y_lag1")]
```

## Spacing classification

`ild_spacing_class()` returns "regular-ish" or "irregular-ish" based on the variability of intervals and the proportion of large gaps. The rule is overridable. This classification can inform the choice of correlation structure in `ild_lme()` (AR1 for regular-ish, CAR1 for irregular-ish).

```{r spacing}
ild_summary(x)$spacing
ild_spacing_class(x)
```

If your data start as a **tsibble** (`tbl_ts`), see `vignette("tsibble-interoperability", package = "tidyILD")` for how **tidyILD** records key/index and interval provenance and how to round-trip with **`ild_as_tsibble()`**.

---
title: "MSM Identification and Recovery in tidyILD"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{MSM Identification and Recovery in tidyILD}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

## Why this vignette exists

This vignette documents the **identification assumptions** behind the MSM/IPW workflow and
shows how to run the causal recovery harness added for regression testing and simulation-based checks.

## Identification assumptions

In this workflow, interpretation of weighted outcome contrasts depends on:

1. **Sequential exchangeability**: all confounders needed for treatment assignment at each \(t\)
   are captured in the history set used for IPTW.
2. **Positivity / overlap**: treatment probabilities are bounded away from 0 and 1 in relevant strata.
3. **Consistency**: observed outcomes under observed treatment history equal potential outcomes under that same history.
4. **Correct weight models**: treatment and censoring models are correctly specified.

Use diagnostics to stress-test these assumptions:

- `ild_msm_balance()` for weighted SMD checks;
- `ild_ipw_ess()` for effective sample size;
- `ild_msm_overlap_plot()` for propensity overlap;
- `ild_diagnose(..., balance = TRUE, ...)` for integrated causal diagnostics + guardrails.

## Estimand-first + history-builder workflow (v1)

```{r eval = FALSE}
library(tidyILD)

d <- ild_msm_simulate_scenario(n_id = 100, n_obs_per = 12, true_ate = 0.5, seed = 101)
d <- ild_center(d, y)

hist_spec <- ild_msm_history_spec(vars = c("stress", "trt"), lags = 1:2)
d <- ild_build_msm_history(d, hist_spec)

estimand <- ild_msm_estimand(type = "ate", regime = "static", treatment = "trt")

fit_obj <- ild_msm_fit(
  estimand = estimand,
  data = d,
  outcome_formula = y ~ y_bp + y_wp + stress + trt + (1 | id),
  history = ~ stress_lag1 + trt_lag1,
  predictors_censor = "stress",
  inference = "bootstrap",
  n_boot = 200,
  strict_inference = FALSE
)

fit_obj
fit_obj$inference$status
fit_obj$inference$reason
```

## Recovery harness

```{r eval = FALSE}
rec <- ild_msm_recovery(
  n_sim = 100,
  n_id = 120,
  n_obs_per = 12,
  true_ate = 0.5,
  n_boot = 200,
  inference = "bootstrap",
  seed = 1001,
  censoring = TRUE
)

rec$summary
rec$summary_by_scenario
```

Scenario-grid validation (positivity stress and treatment-model misspecification):

```{r eval = FALSE}
grid <- tibble::tibble(
  scenario_id = c("baseline", "positivity_stress", "misspecified_treatment"),
  positivity_stress = c(1, 1.8, 1),
  misspec_treatment_model = c(FALSE, FALSE, TRUE)
)

rec_grid <- ild_msm_recovery(
  n_sim = 50,
  n_id = 120,
  n_obs_per = 12,
  true_ate = 0.5,
  n_boot = 200,
  inference = "bootstrap",
  scenario_grid = grid,
  seed = 1101
)

rec_grid$summary_by_scenario
```

Interpretation:

- `bias` and `rmse` target point-estimate recovery;
- `coverage` targets interval calibration under the chosen inference mode;
- `ess_mean` / `ess_min` and `weight_ratio_median` summarize positivity stress.

## Inference caveats and strict mode

- `inference = "robust"` can degrade on weighted `lmer` paths where robust
  variance is not supported.
- `ild_msm_fit()` records this explicitly in:
  - `fit_obj$inference$status` (`"ok"`, `"degraded"`, `"unsupported"`),
  - `fit_obj$inference$reason` (machine-readable reason code),
  - `fit_obj$inference$message` (user-facing explanation).
- Set `strict_inference = TRUE` to error instead of degrading.
- Use `ild_msm_bootstrap(..., weight_policy = "reestimate_weights")` when you
  want first-stage weight uncertainty represented in intervals.

## Notes on v1 scope

- v1.1 estimand schema accepts static and dynamic regime specs, but dynamic
  weighting is still scaffold-only in `ild_msm_fit` and will report degraded
  status unless strict mode is enabled.
- Joint Bayesian MSM estimation is out of scope in v1 (see `?ild_msm_inference`).

---
title: "Heterogeneity and person-specific effects"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Heterogeneity and person-specific effects}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Estimands: population, partial pooling, and no pooling

Mixed models estimate **population** fixed effects and **variance components** for random effects.
Person-specific quantities are usually summarized by **conditional modes** (empirical Bayes / BLUPs in `lme4`):

- **`ranef(fit)`**: deviations from the population mean.
- **`coef(fit)`**: person-specific coefficients (population + deviation), i.e. partial pooling.

These are **not** the same as fitting a separate regression in each person (**no pooling**), which
`ild_person_model()` implements for teaching and idiographic workflows. No-pooling estimates do not
shrink and can be very unstable with few observations per person.

Bayesian fits from `ild_brms()` report person-specific summaries via posterior distributions;
`ild_heterogeneity()` reads `coef(fit, summary = TRUE)` for posterior means and intervals.

## `ild_heterogeneity()`

After `ild_lme()` or `ild_brms()` with random effects, use:

```{r eval = requireNamespace("lme4", quietly = TRUE)}
library(tidyILD)
d <- ild_simulate(n_id = 20, n_obs_per = 10, seed = 7)
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_center(x, y)
fit <- ild_lme(y ~ y_wp + y_bp + (1 | id), data = x)
h <- ild_heterogeneity(fit)
print(h$summary)
head(ild_tidy(h))
```

The `$summary` table includes the proportion of person-specific **total** coefficients greater than zero,
quantiles, and (for `lmer`) joins `VarCorr` standard deviations when names align.

Optional `threshold` and `scale = c("raw", "sd_x", "sd_y")` define substantively motivated cutoffs for
the proportion exceeding a threshold (e.g. a fraction of the SD of $x$ or $y$).

## Diagnostics bundle and plots

`ild_diagnose()` stores `fit$heterogeneity` when extraction succeeds. Plot with:

```{r eval = FALSE}
ild_autoplot(bundle, section = "fit", type = "heterogeneity", term = "y_wp")
```

Use `heterogeneity_type = "histogram"` in addition if you prefer a histogram (passed as `...`).

Guardrails **`GR_RE_SLOPE_VARIANCE_VERSUS_RESIDUAL_LOW`** and **`GR_PERSON_SPECIFIC_SLOPES_EMPIRICALLY_TIGHT`**
flag cases where estimated slope heterogeneity is small relative to residual noise (heuristic interpretation aids).

## Stratified descriptive comparison

`ild_heterogeneity_stratified()` refits the same formula within levels of a grouping column and binds
per-subgroup summaries. This is a **descriptive** tool, not a formal test of differences in variance
components; use with adequate $N$ per subgroup (`min_n_id`).

```{r eval = FALSE}
ild_heterogeneity_stratified(
  y ~ y_wp + (y_wp | id),
  data = x,
  subgroup = "cohort",
  min_n_id = 8L
)
```

## See also

`vignette("developer-contracts", package = "tidyILD")` documents the optional `fit$heterogeneity` slot
on `ild_diagnostics_bundle`.

---
title: "Short analysis report"
author: "Alex Litovchenko"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Short analysis report}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
fit_ar1 <- NULL
```

This vignette shows a minimal analysis report: fit a model, print fixed effects, plot fitted vs observed, and show residual ACF and Q-Q.

## 1. Fit model(s)

```{r fit}
library(tidyILD)
d <- ild_simulate(n_id = 8, n_obs_per = 10, irregular = TRUE, seed = 101)
x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)
x <- ild_center(x, y)
fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
```

## 2. Tidy fixed-effects table

```{r tidy}
tidy_ild_model(fit)
```

With cluster-robust standard errors (requires the **clubSandwich** package):

```{r tidy_robust, eval = requireNamespace("clubSandwich", quietly = TRUE)}
tidy_ild_model(fit, se = "robust", robust_type = "CR2")
```

## 3. Fitted vs observed

```{r fitted_plot, fig.alt = "Fitted vs observed"}
ild_plot(fit, type = "fitted")
```

## 4. Residual diagnostics: ACF and Q-Q

```{r diag}
diag <- ild_diagnostics(fit, type = c("residual_acf", "qq"))
diag
```

```{r diag_plots, fig.alt = "Residual ACF and Q-Q", fig.show = "hold"}
plots <- plot_ild_diagnostics(diag)
plots$residual_acf
plots$qq
```

## With AR1 (nlme)

Same workflow with residual autocorrelation (nlme path). If the AR1 model fails to converge on this small example, we show the code only; in practice use more data or a different seed.

```{r fit_ar1}
fit_ar1 <- tryCatch(
  ild_lme(y ~ y_bp + y_wp, data = x, random = ~ 1 | id, ar1 = TRUE),
  error = function(e) NULL
)
if (!is.null(fit_ar1)) {
  tidy_ild_model(fit_ar1)
} else {
  message("AR1 fit did not converge on this run; use ild_lme(..., ar1 = TRUE) with your data.")
}
```

```{r diag_ar1, eval = !is.null(fit_ar1)}
if (!is.null(fit_ar1)) {
  diag_ar1 <- ild_diagnostics(fit_ar1, type = c("residual_acf", "qq"))
  diag_ar1
}
```

## Time-varying effects (TVEM)

When the effect of a predictor may change over time (e.g. stronger in the morning), use `ild_tvem()` to fit a GAM with a smooth in time and a time-varying coefficient. The plot shows how the effect of the predictor varies over the study timeline; interpret the curve as the estimated effect at each time point (with uncertainty band).

```{r tvem}
set.seed(101)
d2 <- ild_simulate(n_id = 6, n_obs_per = 12, seed = 101)
d2$x <- rnorm(nrow(d2))
x2 <- ild_prepare(d2, id = "id", time = "time")
tv <- ild_tvem(x2, "y", "x", k = 5, re_id = TRUE)
ild_tvem_plot(tv)
```

---
title: "Bayesian dynamics recipes (ild_brms)"
author: "tidyILD authors"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Bayesian dynamics recipes (ild_brms)}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

This article gives **copy-paste templates** for modest **dynamics** extensions with [`ild_brms()`]. It is **not** a substitute for identification work, prior sensitivity, or specialized **DSEM** software when that is your estimand. See `vignette("temporal-dynamics-model-choice", package = "tidyILD")` for how these pieces fit the bigger picture.

## Prerequisites

- Packages: **brms**, **Stan** backend (**rstan** or **cmdstanr**).
- Data prepared with [`ild_prepare()`]; lag columns from [`ild_lag()`] or [`ild_panel_lag_prepare()`].

## Recipe 1: Random slope for a lagged predictor

```{r recipe-random-lag-slope}
library(tidyILD)
set.seed(1)
d <- ild_simulate(n_id = 15, n_obs_per = 12, seed = 1)
d$x <- rnorm(nrow(d))
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_lag(x, x, n = 1L, mode = "gap_aware")
tpl <- ild_brms_dynamics_formula("y", "x_lag1", id_var = "id")
tpl$formula
tpl$notes
# Short chains for illustration only:
# fit <- ild_brms(tpl$formula, data = x, iter = 500, chains = 2, refresh = 0)
# ild_tidy(fit)
```

`ild_brms_dynamics_formula()` returns a **suggested** formula `(lag | id) + (1|id)`; adjust priors via [`ild_prior_ild()`] and check [`ild_diagnose()`] (R-hat, ESS).

## Recipe 2: Bivariate lag sketch (two variables)

Build two lag columns, then include both fixed effects (and optionally random slopes). A full **panel VAR** requires more structure than this template; this is a **minimal** associational extension.

```{r recipe-two-lags}
x2 <- ild_panel_lag_prepare(x, c("x", "y"), n = c(1L, 1L), mode = "gap_aware")
names(x2$data)
# Example fixed structure only (not run):
# f2 <- y ~ x_lag1 + y_lag1 + (1 | id)
# fit2 <- ild_brms(f2, data = x2$data, iter = 500, chains = 2, refresh = 0)
```

## Recipe 3: Multivariate outcomes with `mvbind` (sketch)

For **two or more outcomes measured on the same occasions**, **brms** can model them **jointly** with `mvbind()` so residuals can correlate across outcomes. This is still **not** a full panel VAR / DSEM: you must choose lags, priors, and residual structure deliberately. When the estimand is a **fully specified dynamic system** with feedback, **dynamite** or **lavaan** DSEM (or similar) is often more appropriate than stacking outcomes in a single `brms` formula.

```{r recipe-mvbind}
# Not run — requires brms; illustration only:
# library(brms)
# f_mv <- bf(mvbind(mood, stress) ~ mood_lag1 + stress_lag1 + (1 | id)) + set_rescor(TRUE)
# fit_mv <- brm(f_mv, data = x2$data, chains = 2, iter = 500, refresh = 0)
# Build mood_lag1 / stress_lag1 with ild_lag() or ild_panel_lag_prepare() first.
```

For routing when joint multivariate dynamics or high-dimensional predictors matter, see `vignette("ild-specialist-backends", package = "tidyILD")`.

## See also

- `?ild_brms`, `?ild_prior_ild`, `?ild_diagnose`
- `vignette("temporal-dynamics-model-choice", package = "tidyILD")`
- `vignette("ild-specialist-backends", package = "tidyILD")`

```{r session-info, echo = FALSE}
sessionInfo()
```

# tidyILD Phase 1 contracts (user-facing summary)

`ild_tidy()` and `ild_augment()` behavior is unchanged until later phases. **`ild_diagnose()`** returns **`ild_diagnostics_bundle`** for supported model classes (see `?ild_diagnose`).
Standalone utilities that also feed bundle sections are documented under **`?ild_diagnostics_utilities`**.
This file mirrors the R-level contracts: see `?ild_diagnostics_bundle`, `?ild_tidy_schema`, `?ild_augment_schema`.

**Developers:** the full normative specification (section fields, tidy/augment semantics, backend checklist) is **`inst/dev/DEVELOPER_CONTRACTS.md`** and the vignette **`vignette("developer-contracts", package = "tidyILD")`**.

## 1. `ild_diagnostics_bundle`

Fixed top-level slots (same names for every engine; content may be `NULL`):

| Slot | Role |
|------|------|
| `meta` | Run metadata, engine, dimensions |
| `data` | Spacing, gaps, compliance, missingness, distributions |
| `design` | Within/between variation, time coverage, occasions, imbalance |
| `fit` | Convergence, singularity, optimizer / MCMC state |
| `residual` | ACF, Q-Q, fitted vs observed, heteroskedasticity |
| `predictive` | CV, PPC, forecast error |
| `missingness` | IPW / imputation summaries when applicable |
| `causal` | Weights, positivity, causal diagnostics when applicable |
| `warnings` | Tibble (0+ rows): `source`, `level`, `message`, `code` |
| `guardrails` | Tibble (0+ rows): `rule_id`, `section`, `severity`, `triggered`, `message`, `recommendation` (see `?guardrail_registry`) |
| `summary_text` | Character vector |

## 2. `ild_tidy()` columns

**Required:** `term`, `component`, `effect_level`, `estimate`, `std_error`, `conf_low`, `conf_high`, `statistic`, `p_value`, `interval_type`, `engine`, `model_class`

**Optional:** `rhat`, `ess_bulk`, `ess_tail`, `pd`, `rope_low`, `rope_high`

## 3. `ild_augment()` columns

**Required:** `.ild_id`, `.ild_time`, `.outcome`, `.fitted`, `.resid`, `.resid_std`, `engine`, `model_class`

**Optional:** `.fitted_lower`, `.fitted_upper`, `.influence`, `.state`, `.state_lower`, `.state_upper`

---
title: "Continuous-Time Dynamics with ctsem in tidyILD"
author: "tidyILD authors"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Continuous-Time Dynamics with ctsem in tidyILD}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

This vignette introduces the `ctsem` backend in tidyILD via `ild_ctsem()`.
Use this path when your scientific target is **continuous-time latent dynamics**
under irregular measurement timing.

## When to use `ild_ctsem()`

- Choose `ild_ctsem()` for continuous-time latent process modeling.
- Choose `ild_kfas()` for discrete-time state-space workflows.
- Choose `ild_lme()` / `ild_brms()` for multilevel regression targets.

## Minimal workflow

```{r eval = FALSE}
library(tidyILD)

d <- ild_simulate(n_id = 1, n_obs_per = 60, seed = 501)
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_center(x, y)

fit_ct <- ild_ctsem(
  data = x,
  outcome = "y",
  model_type = "stanct",
  chains = 1,
  iter = 400
)

fit_ct
td <- ild_tidy(fit_ct)
ag <- ild_augment(fit_ct)
dg <- ild_diagnose(fit_ct)
```

## Diagnostics and plots

```{r eval = FALSE}
ild_autoplot(fit_ct, type = "fitted_vs_actual")
ild_autoplot(fit_ct, type = "residual_time")
ild_autoplot(dg, section = "fit", type = "convergence")
ild_autoplot(dg, section = "residual", type = "acf")
```

## Guardrails and reporting

`ild_diagnose(fit_ct)` may trigger ctsem-focused guardrails such as:

- `GR_CTSEM_NONCONVERGENCE`
- `GR_CTSEM_UNSTABLE_DYNAMICS`
- `GR_CTSEM_SHORT_SERIES_FOR_COMPLEX_DYNAMICS`

These guardrails are surfaced in `print(dg)`, `ild_methods(fit_ct, bundle = dg)`,
and `ild_report(fit_ct)`.

## Notes

- The v1 ctsem wrapper is intentionally conservative.
- You can pass a custom `ct_model` object for advanced specifications.
- Keep time scaling explicit (`time_col`, `time_scale`) for interpretability.

---
title: "Choosing between lme/nlme, brms, KFAS, and ctsem"
author: "Alex Litovchenko"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Choosing between lme/nlme, brms, KFAS, and ctsem}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
```

This vignette is a **decision guide**, not a technical deep dive. Use it to pick an entry point:

- **`ild_lme()`** when the main need is **multilevel regression** with **residual** temporal correlation (AR1/CAR1, random effects).
- **`ild_brms()`** when **uncertainty**, **priors**, and **Bayesian partial pooling** across persons matter most.
- **`ild_kfas()`** when the target is **discrete-time** latent temporal state estimation (single series per fit).
- **`ild_ctsem()`** when the target is **continuous-time latent dynamics** under irregular measurement timing.

## Decision axes

Ask:

1. **Population vs single-series focus** — Do you need **inference for a population** of persons (fixed/random effects, WP/BP) or a **structured model for one** long series (latent level/trend)?
2. **Multilevel vs state-space** — Is the scientific target **between- and within-person** structure with **residual** temporal correlation, or **explicit latent dynamics** (Kalman smoothing/filtering) for one trajectory?
3. **Frequentist vs Bayesian** — `ild_lme()` / **nlme** vs **`ild_brms()`** (Stan); **KFAS** in tidyILD is **likelihood / ML**-oriented for the wrapped fits.
4. **Diagnostics** — Frequentist/brms: `ild_diagnostics()` / bundle from `ild_diagnose()` for standard models; **KFAS**: `ild_diagnose()` emphasizes **innovations**, ACF, and bundle sections tuned for **state space** (see `?ild_diagnose`).

## Comparison table

| | **`ild_lme()`** (lme4 / nlme) | **`ild_brms()`** | **`ild_kfas()`** (KFAS) | **`ild_ctsem()`** (ctsem) |
|---|---|---|---|---|
| **Typical use** | Multilevel regression, AR1/CAR1 on residuals | Hierarchical Bayes, flexible priors | Univariate latent level (v1), single `.ild_id` per fit | Continuous-time latent dynamics with irregular timing |
| **ID structure** | Multiple `.ild_id`, random effects | Multiple `.ild_id`, partial pooling | One person per fit (v1); not pooled latent state across IDs | v1 wrapper is conservative (single-series focus) |
| **Temporal structure** | Residual AR1/CAR1; spacing-informed choice | Flexible time / residual modeling | Discrete-time local level on observation order | Continuous-time drift/diffusion parameterization |
| **Priors** | N/A (frequentist) | Yes (user-specified) | Implicit via likelihood / ML | Depends on ctsem fit mode (`ctStanFit` / `ctFit`) |
| **Predictive checks** | Via bundle / extensions | PPC via **brms** (`pp_check`) | Forecast / innovation plots (not PPC); see `ild_autoplot(bundle)` for KFAS | Bundle diagnostics + ctsem residual/fitted views |
| **Provenance / report** | `ild_provenance`, `ild_report` | Same | Same; KFAS stores state spec, smoothing, `fit_context` | Same; ctsem fit type and scaling recorded |

**Important:** `ild_kfas()` is **not** a drop-in replacement for a **multilevel** model with person-level random effects. If you fit **separate** state-space models per person, use **`fit_context = "independent_series_per_id"`** when that matches your design and read **`GR_KFAS_UNMODELED_BETWEEN_PERSON_HETEROGENEITY`** and related guardrails.

## Short pointers to other vignettes

- **Temporal dynamics and model choice (lags, AR, TVEM vs latent):** `vignette("temporal-dynamics-model-choice", package = "tidyILD")`
- **Overall workflow:** `vignette("tidyILD-workflow", package = "tidyILD")`
- **Report-style analysis:** `vignette("ild-analysis-report", package = "tidyILD")`
- **State-space concepts (KFAS):** `vignette("kfas-state-space-modeling", package = "tidyILD")`
- **Irregular timing + KFAS:** `vignette("kfas-irregular-timing-spacing", package = "tidyILD")`
- **Continuous-time ctsem backend:** `vignette("ctsem-continuous-time-dynamics", package = "tidyILD")`
- **Spacing and WP/BP:** `vignette("ild-decomposition-and-spacing", package = "tidyILD")`

## See also

`inst/dev/KFAS_V1_BACKEND.md` (backend scope), `?ild_kfas`, `?ild_ctsem`, `?ild_lme`, `?ild_brms`.

---
title: "Irregular measurement and latent state tracking"
author: "Alex Litovchenko"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Irregular measurement and latent state tracking}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
has_kfas <- requireNamespace("KFAS", quietly = TRUE)
```

**Important:** This vignette does **not** claim that **KFAS** in tidyILD “fixes” or removes the challenges of irregular measurement. **tidyILD** helps you **measure and report** timing structure (intervals, gaps, spacing class). **`ild_kfas()`** fits a **discrete-time** state-space model along **observation order**; it is **not** a drop-in substitute for **continuous-time** latent models (e.g. ctsem-like workflows). High irregularity can trigger a **guardrail**; continuous-time solutions are a **later tier**.

## Motivation

Intensive longitudinal data (EMA, diaries, wearable prompts) often have **irregular** observation times: gaps, bursts, and varying intervals between measurements. **tidyILD** makes that structure visible and actionable **before** you choose a model:

- **`ild_prepare()`** computes per-row intervals (e.g. `.ild_dt`) and respects **`gap_threshold`** for episode boundaries.
- **`ild_summary()`** and **`ild_spacing()`** / **`ild_spacing_class()`** summarize interval variability and classify spacing as **regular-ish** vs **irregular-ish** (see `vignette("ild-decomposition-and-spacing", package = "tidyILD")`).
- **`ild_design_check()`** aggregates spacing and design summaries for reporting.

Those tools support **mixed models** (e.g. choosing AR1 vs CAR1 for residual correlation in `ild_lme()`) and **transparent reporting** of how timing was handled. **KFAS** in tidyILD adds another branch: **latent state** tracking for a **single** series—but the same spacing diagnostics still describe **whether** the timing pattern matches the **assumptions** of the state-space fit you use.

## What `ild_kfas()` assumes today

The **KFAS** backend in tidyILD fits models on a **discrete time index**: observations are ordered after `ild_prepare()`, and the state evolves **one step per row** for that series. It does **not** implement a **continuous-time** Kalman filter with unequal physical intervals built into the transition matrix (that design space points elsewhere, e.g. specialized continuous-time software—see `inst/dev/KFAS_V1_BACKEND.md`).

So:

- **Irregular calendar timing** is still reflected in **ILD** metadata (intervals, spacing class, gaps).
- The **state-space** fit itself is **discrete-time** along the **observation order** unless and until the backend gains explicit continuous-time support.

Use spacing outputs to **document** irregularity and to **interpret** results cautiously when intervals vary wildly: a local-level model on the observation index is not the same as a model that correctly weights by elapsed time in continuous time.

## `irregular_time` and guardrails

`ild_kfas(..., irregular_time = TRUE)` **suppresses** the default warning when spacing looks highly irregular relative to a discrete-time local-level assumption. It does **not** turn the model into a continuous-time model; it acknowledges that you have read the spacing diagnostics and accept the discrete-time framing (or a deliberate approximation).

After fitting, **`ild_diagnose()`** may attach **KFAS-specific guardrails**, including **`GR_KFAS_HIGH_IRREGULARITY_FOR_DISCRETE_TIME`** when the design is **irregular-ish** but the fit remains discrete-time. That rule is a **design-space bridge**: it connects **spacing classification** to **state-space** workflow without claiming to fix misspecification.

Browse rule definitions with `guardrail_registry()` and inspect triggered rows in `ild_diagnose(fit)$guardrails`.

## Workflow sketch

```{r spacing_demo, eval = TRUE}
library(tidyILD)
set.seed(3)
d <- ild_simulate(n_id = 1, n_obs_per = 50, irregular = TRUE, seed = 11)
x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)
ild_summary(x)$summary
ild_spacing_class(x)
```

If **KFAS** is installed, you can fit with explicit acknowledgment of irregular spacing:

```{r kfas_irregular, eval = has_kfas}
fit <- suppressWarnings(
  ild_kfas(
    x,
    outcome = "y",
    state_spec = "local_level",
    time_units = "seconds",
    irregular_time = TRUE
  )
)
diag <- ild_diagnose(fit)
# When spacing is irregular-ish and IQR/median interval ratio > 0.75, expect
# GR_KFAS_HIGH_IRREGULARITY_FOR_DISCRETE_TIME among triggered guardrails:
diag$guardrails[, c("rule_id", "message")]
```

If **KFAS** is not installed, install the **KFAS** package to run the chunk above.

## Where to go next

- `vignette("kfas-state-space-modeling", package = "tidyILD")` — state-space concepts and **filtered vs smoothed** states.
- `vignette("kfas-choosing-backend", package = "tidyILD")` — when to use **lme/nlme**, **brms**, or **KFAS**.

---
title: "Visualization in tidyILD"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Visualization in tidyILD}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

## Role of visual inspection

Intensive longitudinal data (ILD) benefits from **layered plots**: raw trajectories, spacing and missingness, within/between structure, model-based residuals, and (when fitted) heterogeneity of person-specific parameters. tidyILD spreads these across `ild_plot()`, `ild_autoplot()` on diagnostics bundles, and backend-specific helpers (KFAS, TVEM, circadian). There is no single `ild_visualize()`; use the table below as an index.

## Map: question → function → bundle section (if any)

| Question | Primary API | Notes |
|----------|-------------|--------|
| Spaghetti trajectories | `ild_spaghetti()`, `ild_plot(..., type = "trajectory")` | `max_ids`, `time_var`, optional `facet_by` |
| Person × time heatmap | `ild_heatmap()`, `ild_plot(..., type = "heatmap")` | Optional `facet_by` for clusters |
| Spacing / intervals | `ild_plot(..., type = "gaps")`, `ild_autoplot(bundle, section = "design", type = "coverage")` | `ild_spacing()` is tabular + AR1/CAR1 hint |
| Missingness pattern | `ild_plot(..., type = "missingness")`, `ild_missing_pattern()$plot` | Bundle: `section = "data", type = "missingness"` |
| Within/between densities | `ild_center_plot()`, `ild_decomposition(..., plot = TRUE)` | Not marginal effects of fitted model |
| Residual ACF | `ild_plot(fit, type = "residual_acf")`, `ild_autoplot(bundle, section = "residual", type = "acf")` | Sequence-based lag; see `?ild_acf` |
| Obs vs fitted **scatter** | `ild_plot(fit, type = "fitted")` | |
| Obs vs fitted **over time** | `ild_plot_predicted_trajectory(fit)`, `ild_plot(..., type = "predicted_trajectory")` | Two lines per id |
| Person-specific / partial pooling | `ild_autoplot(h, type = "caterpillar")` on `ild_heterogeneity()`; bundle `section = "fit", type = "heterogeneity"` | |
| Time-varying effect | `ild_tvem_plot()` | After `ild_tvem()` |
| State space / continuous time | `ild_plot_states()`, `ild_plot_forecast()`, `ild_plot_filtered_vs_smoothed()` on `ild_kfas()` fits; `ild_autoplot` for ctsem where supported | Optional **KFAS** / **ctsem** |
| Diurnal pattern | `ild_circadian()` | |

After `ild_diagnose()`, prefer `ild_autoplot(bundle, ...)` for a consistent section layout (`residual`, `fit`, `predictive`, `data`, `design`, `causal`).

## Example: spaghetti, predicted trajectories, facet by cluster

```{r plot-examples}
library(tidyILD)
set.seed(1)
d <- ild_simulate(n_id = 24, n_obs_per = 10, seed = 1)
d$cluster <- rep(LETTERS[1:3], length.out = nrow(d))
x <- ild_prepare(d, id = "id", time = "time")
ild_spaghetti(x, var = "y", facet_by = "cluster", max_ids = 12L)

fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
ild_plot_predicted_trajectory(fit, time_var = ".ild_seq", max_ids = 8L, facet_by = "cluster")
```

## Recipe: facet panels without a dedicated helper

Any ggplot from `ild_*` can be stored and extended (e.g. `p + facet_wrap(vars(week))`) if you add a column to the ILD first. The `facet_by` argument on `ild_plot()` passes through to [`ggplot2::facet_wrap()`](https://ggplot2.tidyverse.org/reference/facet_wrap.html) for trajectory, heatmap, gaps, and predicted trajectories.

## Partial effects for `_wp` and `_bp` (external packages)

tidyILD does **not** reimplement marginal effects. After `ild_center()`, fit with `y ~ x_wp + x_bp + (1|id)` and use **marginaleffects** or **ggeffects** on the fitted object. Keep the estimand explicit: slopes on `x_wp` are **within-person** associations; `x_bp` captures **between-person** differences in level.

```{r partial-effects-template, eval = FALSE}
# install.packages(c("marginaleffects", "ggeffects"))  # if needed
x <- ild_center(x, x)
fit <- ild_lme(y ~ x_wp + x_bp + (1 | id), data = x, warn_uncentered = FALSE)
# Example (syntax may vary by package version):
# marginaleffects::plot_predictions(fit, condition = "x_wp")
# marginaleffects::plot_predictions(fit, condition = "x_bp")
# ggeffects::ggpredict(fit, terms = "x_wp")  # then plot() or ggplot2 layer
```

## See also

`vignette("tidyILD-workflow", package = "tidyILD")`, `vignette("ild-decomposition-and-spacing", package = "tidyILD")`, `?ild_plot`, `?ild_autoplot`.

# Cross-backend validation benchmark contract

This document defines the **scenario matrix**, **metric semantics**, **runtime tiers**, and **artifact schema** for the tidyILD backend validation harness (`scripts/run-backend-validation-benchmarks.R`, `tests/testthat/helper-backend-validation-harness.R`).

## Architecture alignment

- Entry points remain: `ild_fit()` for `lme4` / `nlme` / `brms`; `ild_kfas()` and `ild_ctsem()` are invoked only when the scenario type requires them.
- Outputs are validated against existing contracts where applicable: `ild_tidy()` / `ild_augment()` schemas, `ild_diagnostics_bundle` structure, `guardrail_registry()` rule IDs in diagnostic bundles.
- The harness lives under `tests/` and `scripts/`; it is **not** part of the exported user API unless promoted later.

## Runtime tiers

| Tier     | Intended use              | Typical `n_sim` | Scenarios                          |
|----------|---------------------------|-----------------|------------------------------------|
| `smoke`  | PR / manual quick check   | 2–5             | Fast MLM recovery + KFAS smoke     |
| `nightly`| Scheduled regression      | 10–30           | + nlme, brms, ctsem (optional)     |
| `full`   | Release / deep validation | 50+             | All scenarios, larger samples      |

Tiers are **not** CRAN-facing; they run in CI or locally with optional dependencies installed.

## Scenario IDs (v1)

| `scenario_id`        | `type`      | `tier_min` | Description |
|----------------------|------------|------------|-------------|
| `mlm_power_ri`       | `mlm_power`| `smoke`    | `ild_simulate` + Gaussian `x`, `y += effect * x`; estimand: coefficient of `x` (same DGP for all MLM backends). |
| `mlm_lag_x`         | `mlm_lag`  | `smoke`    | `x` with `ild_lag(..., gap_aware)`; outcome `y = beta * x_lag1 + noise`; estimand: coefficient of `x_lag1`; **lme4** only in the harness matrix. |
| `kfas_local_level` | `kfas`     | `smoke`    | Single-series ILD subset; `ild_kfas()` local level; metrics focus on **fit success** and variance auxiliaries (truth optional / `NA`). |
| `ctsem_univariate`  | `ctsem`    | `nightly`  | Single-ID continuous-time path; short Stan chains in CI; **warn-only** thresholds by default. |

Additional scenarios may append rows to the manifest in `helper-backend-validation-harness.R` and update this table.

## Backend eligibility

- `mlm_power_ri`: `lme4`, `nlme`, `brms` (each skipped if package unavailable or not requested via `--backends`).
- `mlm_lag_x`: `lme4` only.
- `kfas_local_level`: `kfas` only (requires **KFAS**).
- `ctsem_univariate`: `ctsem` only (requires **ctsem**, **rstan**; slow).

## Metric definitions (per replication)

- **truth**: Known estimand when defined (e.g. `effect_size` for `x`); `NA` for auxiliary-only KFAS smoke rows.
- **estimate**, **std_error**, **conf_low**, **conf_high**: From `ild_tidy()` / `tidy_ild_model()` for the target term when available; otherwise `NA`.
- **bias** = `estimate - truth` (if `truth` finite).
- **squared_error** = `(estimate - truth)^2`.
- **covered**: `truth` inside `[conf_low, conf_high]` when intervals exist; else `NA`.
- **calibration_gap** = `nominal_level - mean(covered)` aggregated at summary level (not per row when `covered` is NA).
- **converged**: Backend-specific heuristic (e.g. lme4 optimizer, brms Rhat heuristic, ctsem flag).
- **singular** (lme4): `lme4::isSingular` when applicable.
- **n_guardrails**: Row count in `bundle$guardrails` from `ild_diagnose()`.
- **elapsed_sec**: Wall time for fit + diagnostics slice used in harness.

## Artifact schema

### `benchmark_raw.csv`

Columns (minimum):

`run_id`, `git_sha`, `tier`, `scenario_id`, `backend`, `replicate_id`, `truth`, `estimate`, `bias`, `squared_error`, `std_error`, `ci_low`, `ci_high`, `covered`, `nominal_level`, `elapsed_sec`, `converged`, `singular`, `fit_error`, `n_guardrails`, `guardrail_rule_ids`, `skipped_reason`

### `benchmark_summary.csv`

Grouped by `tier`, `scenario_id`, `backend`:

`n_replicates`, `n_skipped`, `n_converged`, `convergence_rate`, `mean_bias`, `rmse`, `mean_coverage`, `calibration_gap`, `mean_elapsed_sec`, `mean_n_guardrails`

### `benchmark_metadata.json`

`package_version`, `r_version`, `platform`, `git_sha`, `seed_base`, `tier`, `backends_requested`, `backends_skipped`, `scenario_manifest_version`, `timestamp_utc`.

## Threshold checking

Threshold files live in `inst/benchmarks/thresholds-*.json`. The checker script evaluates **aggregates** in `benchmark_summary.csv` against:

- `min_coverage` (for scenarios with defined interval coverage)
- `max_convergence_failure_rate`
- optional `max_rmse`, `min_n_converged`

Optional backends (`brms`, `ctsem`) may use `"warn_only": true` so CI emits warnings without failing.

## Adding a new backend or scenario

1. Extend the scenario manifest and `tier_min` in the harness helper.
2. Implement a branch in `harness_fit_one()` and metric extraction if the estimand differs.
3. Document the scenario here and in `inst/dev/DEVELOPER_CONTRACTS.md`.
4. Add deterministic unit tests for schema and threshold logic in `test-backend-validation-harness.R`.

---
title: "Specialist backends: when to move beyond the default stack"
author: "tidyILD authors"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Specialist backends: when to move beyond the default stack}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
```

This article explains how **tidyILD** fits next to **specialist** R packages for multivariate dynamics, high-dimensional predictors, and full latent-variable / DSEM-style estimands. It is not a tutorial for those packages; it gives a **contract**, a **routing table**, and **export patterns** so you can preprocess in tidyILD and estimate elsewhere without re-deriving time rules or centering.

## Contract: what tidyILD owns vs what it does not

**tidyILD is responsible for** (when you use its pipeline):

- Encoding **person** and **time**, ordering within person, and **gap** metadata (`ild_prepare()`, `ild_meta()`).
- **Within-between decomposition** for interpretable mixed models (`ild_center()`, `ild_decomposition()`).
- **Spacing-aware lags** (`ild_lag()`, `ild_panel_lag_prepare()`, `ild_check_lags()`, `ild_crosslag()` for single-equation shortcuts).
- **Provenance** on analytic objects and **diagnostics** for models that tidyILD fits (`ild_diagnose()`, guardrails).

**Specialist packages are responsible for** (examples below):

- **Joint** multivariate time-series models, **feedback** systems, and **time-varying parameters** at the system level (e.g. **dynamite**, **lavaan** dynamic SEM, multivariate **brms**).
- **Penalized** longitudinal models and **variable selection** when **p >> n** (e.g. **PGEE** and related penalized GEE / regularized approaches).
- **Measurement models** and **latent** structural models beyond the conservative paths wrapped in tidyILD (**ctsem**, **lavaan**, **blavaan**, multivariate **brms**).

tidyILD does **not** aim to reimplement those estimators. The recommended pattern is: **prepare → center → lag → export a plain data frame** → fit with the specialist tool, recording package versions (e.g. `sessionInfo()` or `ild_manifest()` alongside your script).

## Decision table

| Scientific situation | Often sufficient in tidyILD | Bridge (preprocess here, then …) | Primary external tools (examples) |
|---------------------|----------------------------|----------------------------------|-----------------------------------|
| Lag predictor → one outcome; AR1/CAR1 residuals | `ild_lag`, `ild_lme`, `ild_brms`, `ild_tvem` | — | — |
| **Reciprocal** or **multivariate** lags (e.g. stress ↔ mood), joint dynamics | Two separate `ild_lme` / `ild_crosslag` fits are **not** a joint likelihood | Same lags/centering; export data | **dynamite**; **lavaan** DSEM; multivariate **brms**; multivariate **ctsem** |
| **High-dimensional** time-varying predictors (p >> n), selection | `lme4` / unpenalized mixed models: unstable or non-identified | Screen or penalize **outside** default `ild_lme` path | **PGEE**; regularized GLM/GEE; strong priors in **brms** |
| **Latent** constructs + dynamics | Observed-variable pipeline; `ild_ctsem()` v1 is intentionally narrow | Export; specify full SEM in target package | **ctsem**, **lavaan** / **blavaan** DSEM |
| **Nonlinear** or **non-Gaussian** dynamics | `ild_tvem` (GAM); `ild_brms` with appropriate families/splines | Same | **dynamite**; **brms**; state-space with custom observation model |
| **Time-varying coefficients** (effect of X on Y evolves) | `ild_tvem`; interactions with time; random slopes | Full state-space / Bayesian TVP | **dynamite**; **brms** hierarchical structure |
| **Causal** estimands with **many** time-varying confounders | MSM / IPW tools in tidyILD for supported paths | High-dimensional confounding may need dedicated methods | Doubly robust / **DML** literature; penalized exposure/confounder models |
| **Correlated outcomes**, shared residual structure | Separate outcomes ignore cross-outcome correlation | Multivariate formula or dynamic multivariate model | **dynamite**; `mvbind` in **brms** (see `vignette("brms-dynamics-recipes", package = "tidyILD")`) |

For **temporal structure inside** tidyILD (lags vs AR vs TVEM vs KFAS vs ctsem), start with `vignette("temporal-dynamics-model-choice", package = "tidyILD")`.

## Handoff pattern: export after prepare, center, and lag

Use [`ild_meta()`] to recover the **person id column name** stored in metadata (often still `"id"` in your data, but the prepared object also uses internal columns such as `.ild_id`).

```{r export-pattern}
library(tidyILD)
set.seed(1)
d <- ild_simulate(n_id = 8, n_obs_per = 10, seed = 1)
d$x <- rnorm(nrow(d))
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_center(x, y, x)
x <- ild_lag(x, dplyr::all_of(c("y", "x")), n = 1L, mode = "gap_aware")

meta <- ild_meta(x)
meta$ild_id
# Plain data frame for any external package:
dat <- as.data.frame(x)
head(dat[, c(meta$ild_id, ".ild_time_num", "y", "y_wp", "y_bp", "x_lag1")])
```

**Typical columns** to retain for dynamic panel estimators:

- Person identifier: `meta$ild_id` (and/or `.ild_id` if present).
- Time: `.ild_time_num` and/or your original time column, depending on whether the target software expects continuous time, integer occasion, or clock time.
- Outcomes and predictors, including `_wp` / `_bp` columns if you want to preserve within-between interpretation in the exported file.
- Lag columns created by [`ild_lag()`] or [`ild_panel_lag_prepare()`].

[`ild_export_provenance()`] applies to objects that carry provenance (e.g. after modeling in tidyILD). For **external** fits, record inputs and script versions in your project manifest.

## Code stubs (not evaluated)

The chunks below are **illustrative**. They are not run during `R CMD check` because **dynamite** and **PGEE** are not required dependencies of tidyILD. Install those packages from their own instructions, read their vignettes, and adapt column names to match `dat` above.

### dynamite (multivariate dynamic models)

```{r stub-dynamite, eval = FALSE}
# library(dynamite)
# After building `dat` from an ILD object (see above):
# - Map id/time/outcome columns to dynamite's expected data layout.
# - Specify channels, lags, and priors per package documentation.
# Example placeholder only (not valid without a real dynamite specification):
# fit_dyn <- dynamite::dynamite(
#   dformula = <your dformula>,
#   data = dat,
#   ...
# )
```

### PGEE (penalized GEE / high-dimensional longitudinal)

```{r stub-pgee, eval = FALSE}
# library(PGEE)
# Penalized GEE expects a long data frame with id, repeated outcome, and a matrix
# or formula interface for high-dimensional covariates — see ?PGEE::PGEE.
# Use `dat` from tidyILD after centering/lagging so covariates align with your estimand.
# fit_pgee <- PGEE::PGEE(<formula or design>, data = dat, ...)
```

### lavaan / blavaan (DSEM)

```{r stub-lavaan, eval = FALSE}
# library(lavaan)
# Dynamic SEM is model-syntax-specific; export `dat` and define your model
# in lavaan's longitudinal / DSEM extensions. tidyILD does not generate lavaan syntax.
```

## See also

- `vignette("temporal-dynamics-model-choice", package = "tidyILD")`
- `vignette("brms-dynamics-recipes", package = "tidyILD")` (including multivariate `mvbind` sketch)
- `vignette("ctsem-continuous-time-dynamics", package = "tidyILD")`
- `vignette("msm-identification-and-recovery", package = "tidyILD")` (causal MSM path in tidyILD)

```{r session-info, echo = FALSE}
sessionInfo()
```

---
title: "Developer contracts (package standards)"
description: >
  Internal specification for the diagnostics bundle, tidy/augment schemas, and backend adapters.
  Normative text is maintained in `inst/dev/DEVELOPER_CONTRACTS.md`.
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Developer contracts (package standards)}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include=FALSE}
knitr::opts_chunk$set(echo = FALSE)
md_path <- if (file.exists("../inst/dev/DEVELOPER_CONTRACTS.md")) {
  "../inst/dev/DEVELOPER_CONTRACTS.md"
} else {
  system.file("dev", "DEVELOPER_CONTRACTS.md", package = "tidyILD")
}
```

This vignette embeds the **developer contract** shipped with the package.
Use it when adding engines (KFAS, ctsem, …) or evolving `ild_tidy()` / `ild_augment()` / `ild_diagnose()`.
For a **code skeleton** (S3 stubs, bundle assembly, guardrails, provenance, tests), see `inst/dev/backend-adapter-template.R` in the source tree.

```{r contracts, results='asis'}
if (!nzchar(md_path) || !file.exists(md_path)) {
  stop("DEVELOPER_CONTRACTS.md not found; expected inst/dev/DEVELOPER_CONTRACTS.md", call. = FALSE)
}
cat(knitr::knit_child(md_path, quiet = TRUE))
```

## See also

- `help(ild_diagnostics_bundle)`, `help(ild_tidy_schema)`, `help(ild_augment_schema)`, `help(guardrail_registry)`.
- User-facing short mirror: `system.file("CONTRACTS.md", package = "tidyILD")` (if installed from source with `inst/`).

# KFAS backend (tidyILD v1) — design specification

**Status:** Normative target for `ild_kfas()` and related methods. Implementation is incremental; see `NEWS.md` for what is shipped vs planned.

**Related:** `inst/dev/backend-adapter-template.R`, `inst/dev/DEVELOPER_CONTRACTS.md`, `?ild_tidy_schema`, `?ild_augment_schema`.

---

## Scope (v1)

Smaller than the full KFAS surface area. Focus on recurring ILD use cases:

- Latent level tracking over time; latent trend/change; AR(1)-style latent dynamics (staged).
- **Gaussian observation models first**; Poisson/binomial only if the object contract stays stable.
- Single-subject or stacked-by-subject workflows with clear semantics; **no pooled latent state across IDs in v1**.

### In scope for v1 (incremental delivery)

- Gaussian observation models.
- Local level (`SSMtrend(degree = 1)`).
- Local linear trend, AR(1) state, regression + state hybrids — **state_spec vocabulary reserved**; implementation may follow after local level is stable.
- Filtered and smoothed state extraction.
- One-step-ahead fitted values and prediction errors; standardized residuals via KFAS (`rstandard.KFS` when available).
- State-space residual diagnostics (bundle + autoplot hooks).
- Optional short-horizon forecasts (planned).
- Simulation-based state uncertainty where practical (planned).

### Out of scope for v1

- Pooled multi-subject latent states.
- Continuous-time unequal-interval modeling (reserve for ctsem later).

---

## User-facing API

### Primary: `ild_kfas()`

Opinionated entry point. Conceptual arguments:

| Argument | Role |
|----------|------|
| `data` | Prepared ILD (`validate_ild`) or data frame passed through `ild_prepare_state_space()` |
| `outcome` | Outcome column name (string) |
| `state_spec` | Controlled vocabulary (see below) |
| `observation_family` | `"gaussian"` in v1 |
| `time_units` | **Required explicit** declaration (e.g. `"hours"`, `"days"`) for provenance and spacing summaries |
| `irregular_time` | Logical / detection; warn when the state model assumes near-equal spacing but data are highly irregular |
| `smoother` | Pass through to `KFS()` smoothing |
| `forecast_horizon` | Optional short horizon (planned) |
| `...` | KFAS control (`fitSSM`, `KFS` args) |

### State spec vocabulary (v1)

Controlled labels (not free-form formulas initially) so tidyILD can generate guardrails, diagnostics, methods text, state labels, and tidy outputs:

- `local_level`
- `local_trend`
- `ar1_state`
- `regression_local_level`
- `regression_ar1_state`

**v1 implementation:** `local_level` only unless noted in `NEWS.md`.

### Pooling modes

- **`pool = "none"` (default):** one time series per fit — **one subject per call** in v1 (single distinct `.ild_id`).
- **Mode B (stacked per-ID fits):** wrapper loops over IDs; combined tidyILD object + per-ID subfits — **planned**; not required for first merge.

---

## Return object class and contents

Suggested class vector:

```r
c("ild_fit_kfas", "ild_fit_model", "ild_analysis")
```

**Contents (list or similar):**

| Slot | Description |
|------|-------------|
| `kfas_model` | Fitted `SSModel` after `fitSSM` |
| `kfs` | `KFS` output |
| `spec` | Standardized internal spec (from `ild_kfas_spec.R`) |
| `state_labels` | Names for states (for tidiers and reports) |
| `mapping` | Data mapping metadata (outcome column, time order, ID) |
| `schema_version` | Character, e.g. `"0.1"` |

**Attributes:**

- `attr(., "ild_data")` — ILD data used (single-subject slice)
- `attr(., "ild_provenance")` — via `ild_new_analysis_provenance(..., step = "ild_kfas", ...)`

---

## Data contract

KFAS requires a stricter temporal contract than mixed-model wrappers.

1. Input must be **`validate_ild` ILD** with valid `id` / `time` structure, **or** a raw data frame processed by internal `ild_prepare_state_space()`.
2. **Explicit `time_units`** stored in provenance and echoed in bundle `meta` when diagnosing.
3. Interval summaries and interval distributions belong in bundle `data` / `design` when `ild_diagnose()` is implemented for this engine.

---

## File layout (source)

| File | Role |
|------|------|
| `R/ild_kfas_spec.R` | Vocabulary + internal spec object |
| `R/ild_kfas_helpers.R` | `ild_prepare_state_space()`, spacing warnings, KFS extraction |
| `R/ild_kfas.R` | `ild_kfas()` — fit layer |
| `R/ild_kfas_tidiers.R` | `ild_tidy()`, `ild_tidy_states()` |
| `R/ild_kfas_augment.R` | `ild_augment()` |
| `R/ild_kfas_diagnose.R` | `ild_diagnose()` — bundle contract adapter |
| `R/ild_kfas_autoplot.R` | `ild_autoplot()` |

Internal layers: **specification → model build (`SSModel`) → `fitSSM` → `KFS` → extraction → contract adapter.**

---

## `ild_tidy()` (parameters)

Minimum: one row per estimated variance / hyperparameter where available.

**Required schema columns:** per `ild_tidy_schema()`.

**`component` values (KFAS-specific):**

- `observation` — regression coefficients (when regression spec is implemented)
- `state` — transition / state dynamics parameters
- `variance` — observation and state innovation variances
- `initial` — initial state parameters if exposed
- `auxiliary` — other quantities

**`effect_level`:** conservative defaults — variances and latent dynamics often `auxiliary`; `_wp` / `_bp` terms follow existing tidyILD rules when present.

**`interval_type`:** `"Wald"` if asymptotic intervals are returned; `"none"` if unavailable (document in `?ild_tidy` method).

---

## `ild_tidy_states()`

One row per time point and state summary (backend-specific). Does **not** replace `ild_tidy()` for parameters.

---

## `ild_augment()`

Must satisfy `ild_augment_schema()` required columns.

**KFAS-specific optional columns:** `.state`, `.state_lower`, `.state_upper` — use meaningfully; for multi-state models prefer **`ild_augment_states()`** (long format) instead of overloading columns.

**v1 preference:** single interpretable latent level → populate `.state` with the smoothed level when `state_spec = "local_level"`.

**`.resid_std`:** standardized **one-step-ahead prediction errors** from the state-space model (e.g. `rstandard(KFS)`), **not** arbitrary z-scores of raw residuals.

---

## Tests

- `skip_if_not_installed("KFAS")` for integration tests.
- Single-subject simulated series; assert `ild_tidy` / `ild_augment` column names, `validate_ild` on `ild_data`, and object class.

## Guardrails (KFAS-specific)

Registered in `guardrail_registry()` / `ILD_GUARDRAIL_REGISTRY` with `rule_id` prefix `GR_KFAS_*`:

| rule_id | Intent |
|---------|--------|
| `GR_KFAS_HIGH_IRREGULARITY_FOR_DISCRETE_TIME` | Irregular spacing vs discrete-time indexing |
| `GR_KFAS_SHORT_SERIES_FOR_STATE_SPACE` | Too few time points for the chosen state complexity |
| `GR_KFAS_STATE_DIMENSION_HIGH_FOR_N` | Latent dimension large relative to `n` |
| `GR_KFAS_DEGENERATE_VARIANCE_ESTIMATE` | `Q` or `H` near zero |
| `GR_KFAS_NONCONVERGENCE` | `fitSSM` optimization did not converge |
| `GR_KFAS_MANY_MISSING_OUTCOME_SEGMENTS` | Many NA runs in the outcome before row omission |
| `GR_KFAS_UNMODELED_BETWEEN_PERSON_HETEROGENEITY` | Set `fit_context = "independent_series_per_id"` when fitting the same template per person; not a pooled latent model |

Evaluation: `evaluate_guardrails_kfas()` in `R/ild_kfas_guardrails.R`, called from `ild_diagnose.ild_fit_kfas`.

---

## Changelog

Track implementation status in `NEWS.md` and bump `schema_version` when the stored object shape changes.

---
title: "Missingness in ILD: diagnostics and sensitivity routes"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Missingness in ILD: diagnostics and sensitivity routes}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Why missingness matters in intensive longitudinal data

In EMA and diary studies, missing responses are often **non-ignorable in substance** even when analysts assume **missing at random (MAR)** for estimation: burden, symptom severity, context, or device issues can co-determine both **whether** a prompt is answered and **the outcome**. tidyILD does not replace dedicated missing-data software; it gives **structured diagnostics**, **person-level adherence views**, **time-oriented summaries**, and **hooks** to IPW-based sensitivity workflows already in the package.

**MNAR** (missing not at random) means missingness depends on unobserved values or latent states. No routine plot proves MAR vs MNAR. Use multiple **sensitivity routes** and transparent reporting.

## Types of missingness (useful labels)

- **Unit non-response / attrition:** a person stops contributing (often modeled as **monotone** dropout on the outcome).
- **Intermittent missingness:** gaps with later observed values again; common in EMA.
- **Item missingness:** some variables missing while others are observed on the same prompt.

The **ordinal occasion index** `.ild_seq` (from `ild_prepare()`) is the default backbone for “wave” summaries; it is **not** the same as equal calendar spacing—see `vignette("ild-decomposition-and-spacing", package = "tidyILD")` when timing is irregular.

## Descriptive profiling: `ild_missing_pattern()` and heatmaps

`ild_missing_pattern()` tabulates NA rates by variable and by person, and builds a **person × occasion** heatmap (sequence index on the x-axis). Pass `outcome` to **enrich** `by_id` with compliance metrics from `ild_missing_compliance()` (see below).

```{r pattern}
library(tidyILD)
set.seed(11)
d <- ild_simulate(n_id = 25, n_obs_per = 12, seed = 11)
d$stress <- rnorm(nrow(d))
d$mood <- d$y
miss_i <- sample(nrow(d), 45)
d$mood[miss_i] <- NA
x <- ild_prepare(d, id = "id", time = "time")
mp <- ild_missing_pattern(x, vars = c("mood", "stress"), outcome = "mood")
mp$summary
head(mp$by_id, 3)
```

Plot the same view with `ild_plot(x, type = "missingness", var = "mood")` (see `?ild_plot`).

## Person-level compliance: `ild_missing_compliance()`

`tidyILD::ild_missing_compliance()` returns, per `.ild_id`:

- **`pct_nonmissing_outcome`**, **`longest_run_observed`** (longest streak of observed values in time order),
- **`monotone_missing`**: `TRUE` if, after the first missing outcome, all later values are missing (`NA` if there is no missingness for that person),
- optional **`expected_occasions`** for rough **adherence vs planned N** (`pct_of_expected`, `meets_expected_rows`).

```{r compliance}
cm <- ild_missing_compliance(x, outcome = "mood", expected_occasions = 12L)
summary(cm$pct_nonmissing_outcome)
```

## When to use `ild_missing_model()` and `ild_missing_bias()`

- **`ild_missing_model()`** fits a **logistic** model for `is.na(outcome)` ~ predictors (pooling `glm` or `glmer` with `random = TRUE`). Use it as a **diagnostic** for whether observed covariates predict missingness, not as proof of MAR.
- **`ild_missing_bias()`** is a shortcut for **one numeric predictor** vs missingness (teaching / quick screening).

If predictors are associated with missingness, **complete-case summaries** of the outcome can be **biased** even when a mixed model uses all rows—because the **composition** of who contributes at each occasion may shift. Compare descriptive means by missingness pattern only as exploratory, not causal.

## Complete-case vs mixed models (careful wording)

A **linear mixed model** fitted to all available rows uses the **likelihood contribution from observed outcomes** conditional on random effects. Under **MAR** *and* correct **mean and covariance** specification, inference for the **outcome model** can be appropriate while **ignoring** the missingness mechanism (likelihood-based inference). That statement has **scope limits**:

- It concerns **outcome missingness** in the modeled response, not necessarily intermittent **predictor** missingness handled ad hoc.
- **MNAR** breaks the interpretation; **informative dropout** requires sensitivity analysis.
- **Complete-case analysis** (drop any person-occasion with missing outcome) changes the estimand when missingness is not MCAR.

tidyILD encourages comparing **descriptives** and **fits** on full vs complete-case data as a **coarse** sensitivity check, not a formal test.

## Cohort-level and hazard summaries

- **`ild_missing_cohort()`**: fraction of non-missing outcomes at each `.ild_seq` plus an optional line plot.
- **`ild_missing_hazard_first()`**: discrete **hazard** of being missing on the current row among rows **at risk** (previous occasion observed, or first occasion). Under **intermittent** missingness this is a rough **first-event** summary; under **monotone dropout** it aligns better with a discrete-time dropout hazard.

```{r cohort-hazard}
coh <- ild_missing_cohort(x, outcome = "mood", plot = FALSE)
head(coh$by_occasion)
head(ild_missing_hazard_first(x, outcome = "mood"))
```

## One entry point: `ild_missingness_report()`

`ild_missingness_report()` bundles compliance, `ild_missing_pattern()` (with `outcome` enrichment), cohort and hazard tables, optional `ild_missing_model()`, the same **late-dropout** heuristic used in guardrails (`GR_DROPOUT_LATE_CONCENTRATION`), and short **`snippets`** for methods text.

```{r report}
rpt <- ild_missingness_report(
  x,
  outcome = "mood",
  predictors = "stress",
  fit_missing_model = TRUE,
  random = FALSE,
  cohort_plot = FALSE
)
names(rpt)
rpt$snippets["overview"]
```

## MNAR as sensitivity (no single fix)

tidyILD does **not** fit selection models, pattern-mixture models, or joint models for MNAR. Consider external packages and pre-specified sensitivity analyses. The snippets in `ild_missingness_report()` remind readers that logistic missingness models are **diagnostic / sensitivity**, not proof of MAR.

## IPW and causal tools as **one** sensitivity route

If you fit `ild_missing_model()`, you can feed predicted probabilities into **`ild_ipw_weights()`** and **`ild_ipw_refit()`** for inverse-probability weighting (see `?ild_ipw_weights` and causal vignettes). This addresses **observed** confounding of missingness under a **MAR-like** weighting story; it is **not** a blanket MNAR solution.

```{r ipw-template, eval = FALSE}
mm <- ild_missing_model(x, outcome = "mood", predictors = c("stress"), random = TRUE)
x_w <- ild_ipw_weights(x, mm, stabilize = TRUE)
fit_w <- ild_ipw_refit(mood ~ stress + (1 | id), data = x_w, weights = ".ipw")
```

## Other templates (not evaluated here)

**Compare complete-case vs full mixed model** (same formula):

```{r cc-vs-full, eval = FALSE}
x_cc <- dplyr::filter(x, !is.na(mood))
fit_full <- ild_lme(mood ~ stress + (1 | id), data = x, warn_uncentered = FALSE)
fit_cc <- ild_lme(mood ~ stress + (1 | id), data = x_cc, warn_uncentered = FALSE)
```

**Multiple imputation outside tidyILD**, then `ild_prepare()` per imputed dataset and pool with **mice** / **mitools** / **brms**—keep the imputation model and substantive model aligned with your estimand.

## What tidyILD does **not** do (and where to look)

- **Full MI pipelines** — **mice**, **Amelia**, **jomo**, etc.
- **MNAR selection / pattern-mixture** — specialized books and packages; consult a statistician.
- **Continuous-time event models** for dropout — **survival** / **joint longitudinal-survival** packages.
- **Replacing** domain knowledge about **why** prompts are missed.

## See also

`vignette("tidyILD-workflow", package = "tidyILD")`, `vignette("msm-identification-and-recovery", package = "tidyILD")`, `?ild_diagnose`, `?ild_missing_pattern`, `?ild_missingness_report`.

---
title: "Simulation benchmarks: recovery and power"
author: "tidyILD authors"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Simulation benchmarks: recovery and power}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
# Optional: set `cache = TRUE` on slow chunks when developing locally (see chunks below).
```

This article shows how to use **`ild_simulate()`** and **`ild_power(..., return_sims = TRUE)`** to assess **parameter recovery** for a fixed effect under the package’s bundled Gaussian mixed-model simulation setup, and how that connects to **empirical power**. It is an **illustrative benchmark** for that DGP—not a proof that every tidyILD method is validated for all data.

**Power** answers: “How often do we reject the null when the effect is present?” **Recovery** answers: “How close are point estimates (and intervals) to the known coefficient across repeated simulations?” Both use the same replication loop inside **`ild_power()`**.

## Simulation size and precision

Recovery summaries (**bias**, **RMSE**, **coverage**) are **Monte Carlo estimates**: their sampling variability shrinks as **`n_sim`** increases (roughly \(\propto 1/\sqrt{n_\text{sim}}\) for means). The worked example below uses **`n_sim = 200`** as a balance between stability and build time on **CRAN**; for publication-style tables, consider larger **`n_sim`** and reporting uncertainty (e.g. bootstrap across replications or analytical approximations where available). Chunks that re-run **`ild_power()`** are marked with optional **`cache = TRUE`** so local rebuilds can skip re-simulation when the code is unchanged (not used on CRAN by default).

## What `ild_simulate()` encodes

The base outcome **`y`** (before `ild_power()` adds a predictor) is built as follows:

- **Between-person**: independent random intercepts per `id` (scale **`bp_effect`**).
- **Within-person**: optional AR(1) structure on de-trended within-person series when **`ar1`** is set (numeric \(\phi\)); otherwise i.i.d. Gaussian noise (scale **`wp_effect`**).
- **Time**: a mild linear-in-time component plus **`time`** as `POSIXct` from a regular grid (with optional jitter if **`irregular = TRUE`**).

The formula you fit should match this structure (e.g. random intercept for `id` when using **`ild_lme(..., ar1 = FALSE)`**). Recovery in this vignette focuses on a **single fixed effect** added by **`ild_power()`** on top of that DGP (see below).

## Fixed-effect recovery with `ild_power(..., return_sims = TRUE)`

For each replication, **`ild_power()`**:

1. Simulates data with **`ild_simulate()`**.
2. Adds a standard normal predictor **`x`** and sets **`y <- y + effect_size * x`** so the **true** coefficient of **`x`** is **`effect_size`** (call it β).
3. Runs **`ild_prepare()`** and **`ild_lme()`** with your formula.
4. Records the Wald-based estimate and standard error for the test term.

With **`return_sims = TRUE`**, you get a tibble of per-replication **`estimate`**, **`std_error`**, **`p_value`**, and **`converged`**. Use **`ild_recovery_metrics()`** for **bias**, **RMSE**, and **nominal Wald interval coverage** (normal approximation \(\hat\beta \pm z_{1-\alpha/2} \cdot \mathrm{SE}\)), consistent with the manual calculation shown in earlier package versions.

```{r recovery-run, cache = FALSE}
library(tidyILD)
library(dplyr)
library(ggplot2)

truth <- 0.35
n_sim <- 200L

set.seed(2026)
res <- ild_power(
  formula = y ~ x + (1 | id),
  n_sim = n_sim,
  n_id = 25L,
  n_obs_per = 12L,
  effect_size = truth,
  seed = 2026L,
  return_sims = TRUE,
  verbose = FALSE
)

sim <- res$sim_results |> dplyr::filter(converged)
```

```{r recovery-metrics}
ild_recovery_metrics(res$sim_results, truth = truth, level = 0.95)
```

Across replications, the mean estimate should be near **β**, bias near 0, RMSE positive but modest if the model is well specified, and **coverage** near the nominal level (e.g. 0.95) for large **`n_sim`**—subject to Monte Carlo noise when **`n_sim`** is moderate.

```{r recovery-plot, fig.alt = "Histogram of simulation estimates of the fixed effect", cache = FALSE}
ggplot2::ggplot(sim, ggplot2::aes(x = estimate)) +
  ggplot2::geom_histogram(fill = "steelblue", color = "white", bins = 20) +
  ggplot2::geom_vline(xintercept = truth, linetype = 2, linewidth = 0.8) +
  ggplot2::labs(
    x = "Estimated coefficient for x",
    y = "Count",
    title = "Sampling distribution of fixed-effect estimates",
    subtitle = sprintf("True beta = %s (vertical line)", truth)
  )
```

## From recovery to power

The same object **`res`** already contains **empirical power**: the fraction of converged runs with **`p_value < alpha`**.

```{r power-link}
tibble::tibble(
  power = res$power,
  n_reject = res$n_reject,
  n_converged = res$n_converged,
  alpha = res$alpha
)
```

**Interpretation:** **Recovery** summarizes the **sampling distribution** of \(\hat\beta\). **Power** summarizes how often that distribution leads to rejection at **`alpha`**. A large effect with low variance yields good recovery *and* high power; a small effect may show acceptable recovery (unbiased estimates) but low power at a fixed sample size.

## Variance components (illustrative snapshot)

**`ild_power()`** is built around a **fixed-effect** estimand. Recovery of **variance components** (random-intercept SD, residual variance) is a separate question. As a **single-run sanity check**, you can inspect **`VarCorr()`** on one simulated dataset ( **`lme4`** or **`nlme`** output depending on **`ar1`**):

```{r varcorr-snapshot, eval = requireNamespace("lme4", quietly = TRUE)}
set.seed(1)
d_one <- ild_simulate(n_id = 30L, n_obs_per = 15L, seed = 99)
d_one$x <- rnorm(nrow(d_one))
d_one$y <- d_one$y + 0.3 * d_one$x
prep_one <- ild_prepare(d_one, id = "id", time = "time")
fit_one <- ild_lme(y ~ x + (1 | id), prep_one, ar1 = FALSE, warn_no_ar1 = FALSE)
lme4::VarCorr(fit_one)
```

This is **not** a simulation-based assessment of variance recovery—only a template for where to look when extending benchmarks.

## AR(1) in the DGP vs residual correlation in the fit

- **`ild_simulate(..., ar1 = phi)`** (numeric \(\phi\)) injects **AR(1) within-person** correlation in the **outcome** before **`ild_power()`** adds **`x`**.
- **`ild_lme(..., ar1 = TRUE)`** fits a **residual** AR1/CAR1 structure in **`nlme`** (different object from the DGP’s \(\phi\)); when **`ar1 = TRUE`**, the formula must be **fixed-effects-only** with **`random = ~ 1 | id`** (see **`?ild_lme`**).

**`ild_power()`** exposes **`ar1`** as a **logical** flag for the **fit**, not the DGP’s \(\phi\). To study AR(1) DGPs with **`ild_power()`**, you would need a **custom loop** calling **`ild_simulate(..., ar1 = ...)`** yourself, or prioritize alignment by matching structures in both steps. Minimal illustration of **DGP AR + nlme residual correlation** on one draw:

```{r ar1-illustration, eval = requireNamespace("nlme", quietly = TRUE)}
set.seed(2)
d_ar <- ild_simulate(n_id = 20L, n_obs_per = 14L, ar1 = 0.35, wp_effect = 0.5, seed = 2)
d_ar$x <- rnorm(nrow(d_ar))
d_ar$y <- d_ar$y + 0.25 * d_ar$x
prep_ar <- ild_prepare(d_ar, id = "id", time = "time")
fit_ar <- tryCatch(
  ild_lme(y ~ x, prep_ar, ar1 = TRUE, random = ~ 1 | id, warn_no_ar1 = FALSE),
  error = function(e) NULL
)
if (!is.null(fit_ar)) summary(fit_ar) else "nlme fit failed on this platform (skip)"
```

## Bayesian and state-space extensions

- **`ild_power()`** currently drives **`ild_lme()`** only. For **brms** / Stan, use **`ild_simulate()`** + **`ild_fit(..., backend = "brms")`** (or **`ild_brms()`**) in your own replication loop, then summarize posterior intervals against **`truth`** with the same **coverage** idea (credible vs Wald nominal coverage is **not** interchangeable—document which you use).
- **KFAS** targets **state-space** models; see **`vignette("kfas-state-space-modeling", package = "tidyILD")`** and **`?ild_kfas`**. Cross-checking mixed-model simulation benchmarks against KFAS is out of scope here but natural for method comparison papers.

## Cross-backend validation harness (optional)

For **repeatable CI-style checks** across several engines (`lme4`, optional `nlme` / `brms` / **KFAS** / **ctsem**), the package ships a **non-exported** harness documented in **`inst/dev/BACKEND_VALIDATION_BENCHMARK_CONTRACT.md`**. From a **source tree**, run `scripts/run-backend-validation-benchmarks.R` with a **`--tier`** (`smoke`, `nightly`, `full`) and compare outputs to JSON thresholds via `scripts/check-backend-validation-thresholds.R`. This complements the **`ild_power()`** / **`ild_recovery_metrics()`** workflow above by standardizing **raw and summary tables** for trend comparison; it is **not** a substitute for study-specific simulation design.

## Limitations and scope

- **One estimand in `ild_power()`**: this path targets a **single fixed effect** in a linear mixed model fit with **`ild_lme()`**, matching the current **`ild_power()`** implementation.
- **Inference**: for **`lme4`** fits, **`ild_power()`** uses a **Wald z-approximation** when **p-values** are unavailable from the engine; that is appropriate for this illustration but not identical to likelihood-ratio tests.
- **Variance / AR / other backends**: recovery of **variance components** and **AR** parameters, **brms**, and **KFAS** are discussed **above**; full simulation suites for those paths are left to application-specific code.

## See also

- **`vignette("tidyILD-workflow", package = "tidyILD")`** — end-to-end pipeline.
- **`?ild_power`**, **`?ild_recovery_metrics`** — simulation-based power and recovery summaries.
- **`?ild_simulate`** — base data generator.

```{r session-info, echo = FALSE}
sessionInfo()
```

# tidyILD developer contracts (package standards)

**Status:** Normative for new backends and API evolution. User-facing summaries live in `?ild_diagnostics_bundle`, `?ild_tidy_schema`, and `?ild_augment_schema`; this document is the **operational** specification for implementers.

**Version:** Align with **tidyILD** package version; schema constants live in `R/ild_diagnostics_bundle.R`, `R/ild_schema_tidy_augment.R`, and `R/ild_guardrail_registry.R`.

---

## 1. Diagnostics bundle standard (`ild_diagnostics_bundle`)

### 1.1 Object-level contract

- **Class:** `c("ild_diagnostics_bundle", "list")`.
- **Top-level names (fixed order, non-optional):**  
  `meta`, `data`, `design`, `fit`, `residual`, `predictive`, `missingness`, `causal`, `warnings`, `guardrails`, `summary_text`  
  (constant `ILD_DIAGNOSTICS_BUNDLE_SLOTS` in source).
- **Validation:** `validate_ild_diagnostics_bundle()` — list names must match exactly; `warnings` and `guardrails` must be tibbles; `guardrails` must include the columns in §1.10; `summary_text` must be character.
- **NULL sections:** Any of `meta` … `causal` may be `NULL` if not applicable for an engine or run. **`warnings`** and **`guardrails`** are never `NULL` (use empty tibbles).

### 1.2 Section: `meta`

| Field | Required | Type | Semantics |
|-------|----------|------|-----------|
| `engine` | Recommended | character | Backend id: `"lmer"`, `"lme"`, `"brms"`, or future ids (e.g. `"kfas"`). |
| `n_obs`, `n_id` | Recommended | integer | Rows and distinct persons on the ILD used for diagnostics. |
| `ar1` | Optional | logical | Whether residual AR1/CAR1 was used (frequentist path). |
| `brms_types` | Optional | character | Subset of diagnostics requested for brms (`convergence`, `sampler`, `ppc`). |

Additional named entries are allowed if documented per engine.

### 1.3 Section: `data`

**Purpose:** Observation-level and cohort-level **data** quality (spacing, gaps, global missingness pattern).

| Field | Required | Semantics |
|-------|----------|------------|
| `summary` | Recommended | One-row tibble from `ild_summary()$summary` (e.g. `n_id`, `n_obs`, `prop_gap`, `median_dt_sec`, `iqr_dt_sec`). |
| `spacing_class` | Recommended | `"regular-ish"` or `"irregular-ish"` (`ild_spacing_class()`). |
| `n_gaps`, `pct_gap` | Recommended | Gap counts / percent of intervals flagged as gaps. |
| `median_dt_sec`, `iqr_dt_sec` | Optional | Timing dispersion. |
| `time_range` | Optional | Numeric range of `.ild_time_num` or equivalent. |
| `cohort` | Recommended | `list(n_id, n_obs, n_intervals)` aligned with `ild_summary`. |
| `obs_per_id` | Recommended | `list(min, median, max, sd)` observation counts per person (occasion imbalance). |
| `outcome_summaries` | Optional | Tibble per outcome column: `mean`, `sd`, `min`, `max`, `pct_na`, `n` (when model response/predictors passed to `ild_diagnose`). |
| `missingness_rates` | Optional | Tibble `variable`, `pct_na` for those columns. |
| `missing_pattern` | Optional | List: `summary`, `overall`, `n_complete` from `ild_missing_pattern()` (global columns). |

### 1.4 Section: `design`

**Purpose:** **Design** structure (WP/BP, imbalance, design-time missingness).

| Field | Required | Semantics |
|-------|----------|------------|
| `ild_design_check` | Recommended | Full return value of `ild_design_check()` (embedded list object). |
| `spacing_class`, `recommendation` | Optional | Redundant copies acceptable for fast reporting. |
| `wp_bp` | Optional | Decomposition tibble or `NULL` if `vars` omitted. |
| `design_missingness` | Optional | Summaries bundled inside `ild_design_check` / design slice. |
| `flags` | Recommended | `list(has_wp_bp, spacing_class, irregular)` for fast checks. |
| `time_coverage` | Recommended | `list(min, max, span_sec)` from pooled timeline. |
| `occasion_imbalance` | Recommended | Same object as `data$obs_per_id` (mirrored for visibility). |

### 1.5 Section: `fit`

**Purpose:** **Estimation** diagnostics (convergence, singularity, MCMC).

**Frequentist `lmerMod`:** `engine`, `singular`, `converged`, `optimizer_messages`, `optinfo` (lmer); **`reml`**, **`optimizer`**, **`theta_summary`** (`length`/`min`/`max` of variance-component vector), **`X_rank`** (fixed-effects design rank from `qr(X)`), **`residual_correlation`** (`list(modeled, structure, note)` — lmer has no AR1/CAR1; `modeled` reflects `attr(ild_ar1)`).

**Frequentist `lme`:** above where applicable; **`apVar_ok`**; **`residual_correlation`** with **`class`** (`ild_correlation_class`) and **`coef_corStruct`** when present.

**Bayesian (`brmsfit`):** `engine`, `ild_posterior`, `convergence_table`, `max_rhat`, `n_divergent`, `n_max_treedepth_hits`; **`residual_correlation`** (`note` only; residual structure is model-specific).

**Optional `heterogeneity` (`lmerMod`, `lme`, `brmsfit`):** `list(available, reason, object)`.
When `available` is `TRUE`, `object` is an `ild_heterogeneity` result from [ild_heterogeneity()] (person-specific partial-pooling summaries).
When extraction fails (e.g. intercept-only fixed model without random effects, or parsing error), `available` is `FALSE` and `reason` is a short message.
`ild_autoplot(bundle, section = "fit", type = "heterogeneity", ...)` forwards `term` / `heterogeneity_type` to [ild_autoplot.ild_heterogeneity()].

### 1.6 Section: `residual`

**Purpose:** **Residual** behavior (ACF, Q-Q, vs time/fitted).

| Field | Required | Semantics |
|-------|----------|------------|
| `legacy_ild_diagnostics` | Optional | Full `ild_diagnostics()` object for frequentist engines (enables `plot_ild_diagnostics()` / `ild_autoplot(bundle, section = "residual", type = NULL)` multi-plot or short `type` names `acf` / `qq` / `fitted`). |
| `residual_sd`, `cor_observed_fitted` | Optional | Scalar summaries. |
| `engine` | Recommended | Engine id. |

Bayesian fits may omit `legacy_ild_diagnostics` and keep lighter summaries only.

### 1.7 Section: `predictive`

**Purpose:** **Predictive** checks (obs vs fitted, PPC).

| Content | Semantics |
|---------|-----------|
| Frequentist | `engine`, `n`, `mean_abs_error`, `rmse`, `mean_residual` (bias), `max_abs_error`, `cor_observed_fitted`. |
| brms | `ppc` from `ild_brms_ppc_summary()` when requested; same observation-level scalars as frequentist when `ild_augment` succeeds (`n`, MAE, RMSE, etc.). |

### 1.8 Section: `missingness`

**Purpose:** **Variable-level** missingness aligned to model terms (not duplicate of global `data$missing_pattern` when both exist).

| Field | Semantics |
|-------|------------|
| `note` | Set when no model variables; else `NULL`. |
| `summary`, `overall` | From `ild_missing_pattern(data, vars = predictors)` when predictors exist. |
| `pct_na_by_var` | Subset of `summary` (`variable`, `pct_na`) when available. |
| `missing_model_diagnostic` | Optional: `tidy`, `message`, `outcome`, `predictors` when `ild_diagnose(..., missing_model = TRUE)` runs `ild_missing_model()`. |

### 1.9 Section: `causal`

**Purpose:** **Causal / weighting** diagnostics (IPW, positivity).

Typical content: `columns_found`, `weight_summary` (`min`, `max`, `mean`) when `.ipw` or related columns exist on `ild_data`. Optional **`weight_detail`** (`quantiles`, `sum_w`) when `ild_diagnose(..., causal_detail = TRUE)`.

### 1.10 Tibbles: `warnings` and `guardrails`

**`warnings` (event log from software / sampler):**

| Column | Semantics |
|--------|-----------|
| `source` | e.g. `"lme4"`, `"stan"`, `"tidyILD"`. |
| `level` | e.g. `"warning"`, `"note"`. |
| `message` | Human-readable text. |
| `code` | Short machine id: e.g. `lmer_warning`, `divergent_transitions`. |

**`guardrails` (methodological rules; see `?guardrail_registry`):**

| Column | Semantics |
|--------|-----------|
| `rule_id` | Stable id (e.g. `GR_SINGULAR_RANDOM_EFFECTS`). |
| `section` | Bundle section the rule relates to (`fit`, `design`, `data`, …). |
| `severity` | e.g. `info`, `warning`. |
| `triggered` | Always `TRUE` for rows that appear (only triggered rules are stored). |
| `message`, `recommendation` | Run-specific or default text. |

**Surfacing guardrails (package identity):**

- **`print()` on `ild_diagnostics_bundle`:** After the slot listing, prints a **Summary** block: warning row count, guardrail row count, and (when `nrow(guardrails) > 0`) **highest severity** (`info` < `warning`) and up to **five** unique `rule_id` values. Helpers: `guardrail_severity_rank()` / `guardrail_max_severity()` in `R/ild_guardrail_registry.R`.
- **`ild_report()`:** On successful `ild_diagnose()`, `diagnostics_summary` includes **`guardrails_narrative`** (short string) and **`guardrails`** (`n`, `max_severity`, `rule_ids`). If any guardrails fired, **`methods_with_guardrails`** repeats the methods paragraph with the same guardrail sentence appended (equivalent to `ild_methods(fit, bundle = diagnose_result)`).
- **`ild_methods(..., bundle = NULL)`:** Optional `bundle` from `ild_diagnose(fit)`; when `nrow(bundle$guardrails) > 0`, appends one sentence after provenance: *Methodological cautions (tidyILD guardrails): …*
- **New rules:** Each new `rule_id` added to `ILD_GUARDRAIL_REGISTRY` must have a **deterministic regression test** that asserts the rule can still be triggered (see `tests/testthat/test-guardrails-triggers.R` and fit-level tests via `evaluate_guardrails_fit()` with a constructed `fit_diag` when full `ild_diagnose()` is impractical).

**Contract regression (semantics, not only shape):** `tests/testthat/helper-contract-fixtures.R` builds stable seeded scenarios (regular vs irregular spacing, IPW instability, late dropout with merged contextual guardrails, merged singular / brms fit-level guardrails). `tests/testthat/test-contract-regression.R` checks dense bundle sections, expected `rule_id`s, `ild_tidy` / `ild_augment` required columns, core `ild_autoplot()` routes, and guardrail-aware `ild_methods(..., bundle)` text. The brms scenario uses `skip_on_cran()` because it fits a small `brms` model.

### 1.11 `summary_text`

Character vector of short narrative lines for reporting (`ild_report`, printing). Not a substitute for structured sections.

---

## 2. Tidy output standard (`ild_tidy()`)

**Schema function:** `ild_tidy_schema()` — constants `ILD_TIDY_REQUIRED_COLS`, `ILD_TIDY_OPTIONAL_COLS`.

### 2.1 Required columns (implemented)

`term`, `component`, `effect_level`, `estimate`, `std_error`, `conf_low`, `conf_high`, `statistic`, `p_value`, `interval_type`, `engine`, `model_class`

**Frequentist (`tidy_ild_model`):** `interval_type` is `"Wald"`; `statistic` is the model-reported **t** (or **z**-ratio when `se = "robust"`). **brms:** `interval_type` is `"quantile"` (equal-tailed default intervals); `p_value` is `NA`; optional posterior columns filled when `intervals = TRUE`.

### 2.2 Optional columns

`rhat`, `ess_bulk`, `ess_tail`, `pd`, `rope_low`, `rope_high` (Bayesian / posterior summaries).

### 2.3 Semantics: `component`

Conservative vocabulary (extend with engine-specific values only if documented):

| Value | Meaning |
|-------|---------|
| `fixed` | Standard fixed-effect regression coefficients (current default for all exposed FE rows). |
| `random` | Random-effect variance / covariance summaries when exposed as rows. |
| `auxiliary` | Dispersion, residual variance, or residual-correlation parameters (not yet exposed for lme4/nlme in `ild_tidy`). |
| `scale` | Dispersion parameters (legacy alias in docs; prefer `auxiliary` where appropriate). |
| `correlation` | Correlation structure parameters (e.g. AR1). |
| `nonlinear` | Smooth / spline terms (e.g. mgcv-style). |

### 2.4 Semantics: `effect_level`

| Value | Meaning |
|-------|---------|
| `population` | Intercept / cohort-mean type estimands (`(Intercept)` rows). |
| `within` | Term name ends with `_wp` (within-person component). |
| `between` | Term name ends with `_bp` (between-person component). |
| `cross_level` | Interaction term involving both `_wp` and `_bp` substrings. |
| `unknown` | Coefficient rows that are not clearly classified (default for generic predictors). |
| `auxiliary` | Variance / sampler / state parameters (when such rows exist). |

### 2.5 Semantics: `interval_type`

| Value | Meaning |
|-------|---------|
| `Wald` | Symmetric interval from SE (normal or t as implemented). |
| `quantile` | Equal-tailed posterior quantiles. |
| `HPD` | Highest posterior density interval. |
| `normal` | Alias for large-sample normal approximation when distinct from `Wald`. |

---

## 3. Augment output standard (`ild_augment()`)

**Schema function:** `ild_augment_schema()` — `ILD_AUGMENT_REQUIRED_COLS`, `ILD_AUGMENT_OPTIONAL_COLS`.

### 3.1 Required columns (implemented)

`.ild_id`, `.ild_time`, `.outcome`, `.fitted`, `.resid`, `.resid_std`, `engine`, `model_class`

**`.resid_std` semantics (principled but sparse):** set from `residuals(fit, type = "pearson")` when the engine returns a numeric vector of the correct length; otherwise all `NA`. Do **not** populate with arbitrary z-scores of `.resid` in the same column (avoids mixing definitions). For **brms**, Pearson residuals are used when available; otherwise `NA`.

### 3.2 Optional columns

`.fitted_lower`, `.fitted_upper` (e.g. equal-tailed 95% for **brms** `ild_augment`), `.influence`, `.state`, `.state_lower`, `.state_upper` (e.g. latent states; often `NA` placeholders).

### 3.3 Reserved prefixes and names

| Pattern | Reserved for |
|---------|----------------|
| `.ild_*` | ILD system columns from `ild_prepare()` (do not overwrite for non-ILD semantics). |
| `.fitted*` | Fitted / predicted mean response. |
| `.resid*` | Residuals (raw or transformed). |
| `.state*` | Latent trajectory / random-effect line values when exposed per row. |

The observed response is always **`.outcome`** (no duplicate formula-named column in the augmented tibble).

---

## 4. Backend adapter checklist

**Skeleton:** `inst/dev/backend-adapter-template.R` — commented R patterns for `ild_tidy`, `ild_augment`, `ild_diagnose`, guardrails, provenance, tests, and docs (not sourced by the package; copy into `R/` when implementing).

Each new estimation backend (e.g. KFAS, ctsem) should ship:

### 4.1 S3 methods (operational generics)

| Generic | Requirement |
|---------|-------------|
| `ild_tidy()` | Return a tibble **aligned with `ild_tidy_schema()`** (required columns when feasible). |
| `ild_augment()` | Return a tibble **aligned with `ild_augment_schema()`**; attach `attr(..., "ild_data")` on the **model object** used by diagnostics. |
| `ild_diagnose()` | Return **`ild_diagnostics_bundle`** only; populate all relevant sections; use shared fillers/helpers where possible (`fill_diagnostics_*`, guardrail evaluation). |
| `ild_autoplot()` | For bundle: **section-first** routing (`section` + `type`); implement plotters that read bundle slots; call engine-specific code only inside plotters. Attach `attr(bundle, "ild_fit")` and `attr(bundle, "ild_data")` from `ild_diagnose()` so PPC, fitted, and missingness plots work without a second user argument. |

### 4.2 Data attachment

- Fitted object must carry **`attr(fit, "ild_data")`** (validated ILD) for `ild_augment` / `ild_diagnose`.

### 4.3 Provenance

- Attach **`attr(fit, "ild_provenance")`** via `ild_new_analysis_provenance()` (or equivalent) with `step` set to the fitting function name (e.g. `ild_kfas`), serializable `args`, and `outputs` summarizing key choices.

### 4.4 Tests (minimum)

- **Schema:** `expect_named()` / `expect_true(all(ild_tidy_schema()$required %in% names(...)))` for tidy output (or explicit waiver for transitional columns).
- **Bundle:** `expect_s3_class(..., "ild_diagnostics_bundle")`; `expect_no_error(validate_ild_diagnostics_bundle(...))` when constructed via `ild_diagnose()`.
- **Smoke:** one small simulated ILD dataset, one fitted model, `ild_tidy`, `ild_augment`, `ild_diagnose` run without error (use `skip_if_not_installed` for heavy deps).

### 4.5 Documentation

- Export methods documented on the same `@rdname` as existing engines where possible; describe engine-specific **fit** / **residual** slots in **Details**.

---

## 5. Tsibble interoperability

**Status:** Phase 1 — provenance on input; best-effort round-trip via `ild_as_tsibble()`.

- **Conceptual choice (keys):** ILD uses **one** subject identifier column. `tbl_ts` inputs must have **exactly one** key variable. Multiple keys (e.g. crossed factors) are **not** supported; reshape or paste into a single id before `ild_prepare()`.
- **Provenance:** When `data` is a `tbl_ts`, `ild_prepare()` stores metadata in `attr(x, "tidyILD")$tsibble` (key/index names, interval summary string, `is_regular`, tsibble version). The `tbl_ts` class is still dropped on output; this records formal time-series semantics for reporting and for `ild_as_tsibble()`.
- **Spacing:** `ild_spacing$tsibble` (when present) links declared tsibble regularity/interval text to empirical `.ild_dt` summaries; they may diverge after sorting, duplicate handling, or row drops.
- **Round-trip:** `ild_as_tsibble()` passes `regular =` from stored `is_regular` when available so `interval()` often matches the original for **unchanged** data.

---

## 6. Cross-backend validation benchmark harness

**Normative spec:** `inst/dev/BACKEND_VALIDATION_BENCHMARK_CONTRACT.md` (scenario IDs, tiers, metric columns, artifact layout).

**Purpose:** Run shared simulation scenarios across `lme4` / `nlme` / `brms` / KFAS / ctsem entry points (where installed), write `benchmark_raw.csv`, `benchmark_summary.csv`, and `benchmark_metadata.json`, and optionally gate regressions with JSON thresholds in `inst/benchmarks/thresholds-*.json`.

**Implementation (not exported API):**

| Location | Role |
|----------|------|
| `tests/testthat/helper-backend-validation-harness.R` | `harness_run_benchmark()`, metric extraction, summarization, threshold evaluation. |
| `scripts/run-backend-validation-benchmarks.R` | CLI runner (`--tier`, `--backends`, `--n-sim`, `--seed`, `--out-dir`). Expects package root + `pkgload::load_all()` or `devtools::load_all()`. |
| `scripts/check-backend-validation-thresholds.R` | Reads `benchmark_summary.csv` + thresholds JSON; writes `benchmark_checks.csv`; exit code 1 on hard failures. |
| `.github/workflows/backend-validation-benchmarks.yml` | Scheduled / manual CI; uploads artifacts. |

**When adding a new backend:** extend the scenario manifest and `harness_fit_one()` in the harness helper; document the scenario here and in the benchmark contract; add or adjust `skip_if_not_installed()` tests in `tests/testthat/test-backend-validation-harness.R`. Prefer **warn-only** thresholds for slow or fragile optional engines until metrics stabilize.

---

## 7. Temporal dynamics helpers and guardrails

**User-facing functions** (see `vignette("temporal-dynamics-model-choice", package = "tidyILD")`):

| Function | Role |
|----------|------|
| `ild_panel_lag_prepare()` | Multi-variable `ild_lag()` + single `ild_check_lags()`; provenance step `ild_panel_lag_prepare`. |
| `ild_compare_fits()` | Named list of fits → tibble (`aic`, `bic`, `n_obs`, `converged`, optional `n_guardrails`); **not** an automatic nested-model test. |
| `ild_brms_dynamics_formula()` | Returns a suggested `formula` + `notes` for `ild_brms()`; does not fit. |

**Guardrails** (registry + `evaluate_guardrails_contextual()`):

- `GR_LAG_MEAN_STRONG_RESIDUAL_ACF_NO_AR` — lag terms in mean, no residual AR, pooled residual ACF at lag 1 above a fixed threshold (heuristic).
- `GR_INDEX_LAG_IRREGULAR_SPACING` — `ild_spacing_class` irregular-ish and data provenance records `ild_lag(..., mode = "index")`.

---

## 8. Related files in the source tree

| Location | Role |
|----------|------|
| `inst/CONTRACTS.md` | Short user-facing mirror of slot/column names. |
| `inst/dev/backend-adapter-template.R` | Skeleton methods and checklist for new engines (copy-paste starting point). |
| `inst/dev/KFAS_V1_BACKEND.md` | KFAS / `ild_kfas()` v1 design (Gaussian local level first; full vocabulary staged). |
| `inst/dev/BACKEND_VALIDATION_BENCHMARK_CONTRACT.md` | Benchmark scenario matrix and artifact schema. |
| `R/ild_diagnostics_bundle.R` | Bundle constructor and validation. |
| `R/ild_schema_tidy_augment.R` | Tidy/augment column lists. |
| `R/ild_guardrail_registry.R` | Guardrail rule ids and evaluation helpers. |
| `R/ild_diagnose_fillers.R` | Reference implementations for section contents. |

This file is the **single extended normative spec** for developers; keep it updated when slots or schemas change.

---
title: "Reproducible ILD workflows with tidyILD provenance"
author: "Alex Litovchenko"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Reproducible ILD workflows with tidyILD provenance}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
```

tidyILD records **provenance** at each step: data preparation, centering, lags, alignment, weighting, and model fitting. This vignette shows how to inspect history, generate methods text, build a report, export provenance, and compare two pipelines.

## 1. Prepare data

```{r prepare}
library(tidyILD)
set.seed(1)
d <- ild_simulate(n_id = 8, n_obs_per = 10, irregular = TRUE, seed = 42)
x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)
```

## 2. Center and lag

```{r center_lag}
x <- ild_center(x, y)
x <- ild_lag(x, y, n = 1, mode = "gap_aware", max_gap = 7200)
```

## 3. Fit model

```{r fit}
fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
```

## 4. Run diagnostics

```{r diag}
diag <- ild_diagnostics(fit, type = c("residual_acf", "qq"))
```

## 5. Inspect ild_history()

`ild_history()` prints a human-readable log of all preprocessing and analysis steps recorded on the object.

```{r history}
ild_history(x)
```

For a model fit, provenance includes the source data steps plus the analysis step:

```{r history_fit}
ild_history(fit)
```

## 6. Generate ild_methods()

`ild_methods()` turns provenance into a single methods-style paragraph suitable for a manuscript.

```{r methods}
ild_methods(fit)
```

If you reported fixed effects with cluster-robust SEs (e.g. via `tidy_ild_model(fit, se = "robust", robust_type = "CR2")`), pass that so the methods text can mention it:

```{r methods_robust, eval = requireNamespace("clubSandwich", quietly = TRUE)}
ild_methods(fit, robust_se = "CR2")
```

## 7. Run ild_report()

`ild_report()` assembles a standardized list: meta (n_obs, n_id, engine), methods text, the fixed-effects table, a diagnostics summary, provenance, and an optional export path.

```{r report}
r <- ild_report(fit)
names(r)
r$meta
r$methods
r$model_table
```

The return schema is stable: `meta`, `methods`, `model_table`, `diagnostics_summary`, `provenance`, `provenance_export_path`.

## 8. Export provenance

Export the full provenance (data + analysis steps) to JSON or YAML for reproducibility supplements or archiving.

```{r export}
tmp <- tempfile(fileext = ".json")
ild_export_provenance(fit, tmp, format = "json")
readLines(tmp, n = 20)
```

With `ild_report()`, you can export in one call:

```{r report_export}
tmp2 <- tempfile(fileext = ".yaml")
r2 <- ild_report(fit, export_provenance_path = tmp2)
r2$provenance_export_path
```

## 9. Compare two pipelines

Use `ild_compare_pipelines()` to see how two objects differ (e.g. different gap thresholds, lag modes, or model formula).

```{r compare_setup}
x2 <- ild_prepare(d, id = "id", time = "time", gap_threshold = 3600)
x2 <- ild_center(x2, y)
x2 <- ild_lag(x2, y, n = 1, mode = "index")
fit2 <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = x2, ar1 = FALSE, warn_no_ar1 = FALSE)
```

```{r compare}
cmp <- ild_compare_pipelines(fit, fit2)
cmp
```

This makes it easy to document sensitivity analyses or to check what changed between two analyses.

---
title: "From raw data to model with tidyILD"
author: "tidyILD authors"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{From raw data to model with tidyILD}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
```

This vignette walks through the full tidyILD pipeline: prepare data, inspect structure, apply within-between decomposition and lags, fit a mixed-effects model, and run diagnostics and plots.

For **which temporal assumptions** (lags in the mean, residual AR, time-varying effects, state-space) fit your question, see `vignette("temporal-dynamics-model-choice", package = "tidyILD")`.

## Simulate and prepare

```{r prepare}
library(tidyILD)
# Simulate simple ILD
d <- ild_simulate(n_id = 10, n_obs_per = 12, irregular = TRUE, seed = 42)
# Prepare: encode time structure and add .ild_* columns
x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)
```

## Inspect

```{r summary}
ild_summary(x)
ild_spacing_class(x)
```

## Within-person centering and lags

```{r center_lag}
x <- ild_center(x, y)
x <- ild_lag(x, y, mode = "gap_aware", max_gap = 7200)
```

## Fit a model

Without residual autocorrelation (lmer):

```{r lme_no_ar1}
fit0 <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
```

With AR1 residual correlation (nlme):

```{r lme_ar1}
fit1 <- ild_lme(y ~ 1, data = x, ar1 = TRUE, correlation_class = "CAR1")
```

## Diagnostics and plots

```{r diagnostics}
diag <- ild_diagnostics(fit1, data = x)
names(diag)  # meta, data, stats
names(plot_ild_diagnostics(diag))  # plot names for requested types
# Pooled residual ACF (tibble)
head(diag$stats$acf$pooled)
# By-id ACF when by_id = TRUE: one tibble per person
names(diag$stats$acf$by_id)
head(diag$stats$acf$by_id[[1]])
```

```{r plot_trajectory, fig.alt = "Trajectory plot"}
ild_plot(x, type = "trajectory", var = "y", max_ids = 5)
```

```{r plot_fitted, fig.alt = "Fitted vs observed"}
ild_plot(fit1, type = "fitted")
```

## MSM-style weights (IPTW and IPCW) — optional

For a **marginal structural model**–style sensitivity analysis, you can build **inverse probability of treatment** weights: either **pooled** (`ild_iptw_weights()`) or **sequential MSM** for time-varying `A_t` (`ild_iptw_msm_weights()` after building history with `ild_lag()`), then **inverse probability of censoring** weights for **monotone dropout** (`ild_ipcw_weights()`), multiply into a joint analysis weight with `ild_joint_msm_weights()`, and refit with `ild_ipw_refit()`. This path targets **treatment assignment** and **loss-to-follow-up**; it is distinct from **outcome missingness** IPW (`ild_missing_model()` + `ild_ipw_weights()`). Assumptions (positivity, correct models) are still yours; for **uncertainty** with estimated weights, use **`ild_msm_bootstrap()`** (see `?ild_msm_inference`) instead of trusting default `lmer` SEs alone—**`weight_policy = "reestimate_weights"`** with a **`weights_fn`** that rebuilds IPW on each bootstrap resample is often the appropriate choice when you want first-stage uncertainty reflected; **`fixed_weights`** resamples clusters but keeps the weights attached to resampled rows (faster, approximate).

```{r msm_weights, eval = FALSE}
# Example skeleton (not run in the vignette build)
x2 <- ild_simulate(n_id = 12, n_obs_per = 10, seed = 1)
x2$stress <- rnorm(nrow(x2))
x2$trt <- rbinom(nrow(x2), 1L, 0.45)
x2 <- ild_prepare(x2, id = "id", time = "time")
x2 <- ild_center(x2, y)
x2 <- ild_iptw_weights(x2, treatment = "trt", predictors = "stress")
# Sequential A_t: x2 <- ild_lag(x2, stress); x2 <- ild_lag(x2, trt); ...
# x2 <- ild_iptw_msm_weights(x2, treatment = "trt", history = ~ stress_lag1 + trt_lag1)
x2 <- ild_ipcw_weights(x2, predictors = "stress")
x2 <- ild_joint_msm_weights(x2)
fit_msm <- ild_lme(y ~ y_bp + y_wp + stress + (1 | id), data = x2,
  ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
fit_msm_w <- ild_ipw_refit(fit_msm, data = x2)
```

**Balance and overlap:** After weights are attached, use `ild_msm_balance()` (weighted SMDs), `ild_ipw_ess()` (Kish effective sample size), and `ild_msm_overlap_plot()` (propensity densities by treatment—pooled IPTW or sequential MSM via `attr(x, "ild_iptw_msm_fits")`). Optional: `ild_diagnose(..., balance = TRUE, balance_treatment = "trt", balance_covariates = c("stress"))` fills `causal$balance` and may trigger guardrails `GR_MSM_BALANCE_SMD_HIGH` / `GR_MSM_ESS_LOW`. `ild_autoplot(bundle, section = "causal", type = "overlap", treatment = "trt")` draws overlap when `ggplot2` is available.

For a full assumptions-oriented walkthrough (exchangeability, positivity/overlap, consistency, and weight-model correctness) and simulation-based recovery checks, see `vignette("msm-identification-and-recovery", package = "tidyILD")`. When using `ild_msm_fit()`, check `fit_obj$inference$status`/`reason` for degraded paths, or set `strict_inference = TRUE` to fail fast.

A minimal **bootstrap** illustration (small `n_boot` for speed; increase in real analyses):

```{r msm_bootstrap, eval = requireNamespace("lme4", quietly = TRUE)}
set.seed(3)
xb <- ild_simulate(n_id = 10, n_obs_per = 5, seed = 3)
xb$stress <- rnorm(nrow(xb))
xb <- ild_prepare(xb, id = "id", time = "time")
xb <- ild_center(xb, y)
xb$.ipw <- runif(nrow(xb), 0.85, 1.15)
fb <- ild_lme(y ~ y_bp + y_wp + stress + (1 | id), data = xb,
  ar1 = FALSE, warn_no_ar1 = FALSE, warn_uncentered = FALSE)
fwb <- ild_ipw_refit(fb, data = xb, weights = ".ipw")
bs_fixed <- ild_msm_bootstrap(fwb, n_boot = 20L, weight_policy = "fixed_weights", seed = 3)
tidy_ild_msm_bootstrap(bs_fixed)
# reestimate_weights: weights_fn must return ILD with the weight column, e.g. re-run IPTW pipeline:
bs_re <- ild_msm_bootstrap(fwb, n_boot = 12L, weight_policy = "reestimate_weights",
  seed = 4, weights_fn = function(d) { d$.ipw <- runif(nrow(d), 0.85, 1.15); d })
```

## Reproducibility

Use a fixed seed when simulating or fitting models so results can be recreated. The pipeline is deterministic for a given seed and data. When saving results (e.g. after [ild_lme()] or [ild_diagnostics()]), you can attach a reproducibility manifest and save a single bundle with [ild_manifest()] and [ild_bundle()]:

```{r reproducibility}
# Optional: build a manifest with scenario and seed, then bundle the fit for saving
manifest <- ild_manifest(seed = 42, scenario = ild_summary(x), include_session = FALSE)
bundle <- ild_bundle(fit1, manifest = manifest, label = "model_ar1")
# saveRDS(bundle, "run.rds")  # one file with result + manifest + label
names(bundle)
```

For **simulation-based recovery and power** under the bundled `ild_simulate()` DGP, see `vignette("benchmark-simulation-recovery", package = "tidyILD")`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_kfas_augment.R
\name{ild_augment_states}
\alias{ild_augment_states}
\title{Per-row latent states (long format) for complex multi-state models}
\usage{
ild_augment_states(x, ...)
}
\arguments{
\item{x}{An object from [ild_kfas()].}

\item{...}{Reserved.}
}
\value{
A \code{\link[tibble]{tibble}}.
}
\description{
For \code{local_level} only, use [ild_augment()] which already provides
\code{.state} for the level. This function is reserved for future multi-state
models.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild-class.R
\name{ild_meta}
\alias{ild_meta}
\title{Get ILD metadata attributes}
\usage{
ild_meta(x)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}
}
\value{
A named list of metadata (ild_id, ild_time, ild_gap_threshold,
  ild_n_units, ild_n_obs, ild_spacing). \code{ild_spacing} includes overall
  stats and may contain \code{by_id}, a tibble of per-person spacing stats.
}
\description{
Returns the metadata attributes set by [ild_prepare()]: user-facing
id/time column names, gap threshold, n_units, n_obs, and spacing
(descriptive stats only).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_person_model.R
\name{ild_person_model}
\alias{ild_person_model}
\title{Fit a model separately per person (N-of-1 / idiographic)}
\usage{
ild_person_model(formula, data, method = c("lm"), min_obs = 2L)
}
\arguments{
\item{formula}{A formula (e.g. \code{y ~ x}). Used for each person's \code{lm}.}

\item{data}{An ILD object (see [is_ild()]).}

\item{method}{Character. Currently only \code{"lm"} (default).}

\item{min_obs}{Integer. Minimum observations per person to fit (default 2).
Persons with fewer are omitted or get NA rows.}
}
\value{
A tibble with columns \code{.ild_id} (or the id column name from metadata),
  \code{term}, \code{estimate}, \code{std_error}, \code{p_value}, and optionally
  \code{sigma}, \code{n_obs}. One row per person per term (long format).
}
\description{
Splits the ILD by person and fits the same formula (e.g. \code{lm}) within each.
Returns a tibble of person-level estimates for teaching, N-of-1 analysis, or
inspecting heterogeneity. Use [ild_person_distribution()] to visualize the
distribution of estimates across persons.
}
\examples{
d <- ild_simulate(n_id = 5, n_obs_per = 8, seed = 1)
x <- ild_prepare(d, id = "id", time = "time")
pm <- ild_person_model(y ~ 1, x)
ild_person_distribution(pm, term = "(Intercept)")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_bootstrap.R
\name{tidy_ild_msm_bootstrap}
\alias{tidy_ild_msm_bootstrap}
\title{Tidy fixed effects from \code{ild_msm_bootstrap}}
\usage{
tidy_ild_msm_bootstrap(x, conf_level = 0.95, ...)
}
\arguments{
\item{x}{Object from [ild_msm_bootstrap()].}

\item{conf_level}{Numeric. Used for implied tail quantiles (default 0.95).}

\item{...}{Unused.}
}
\value{
A tibble; see \code{\link{ild_tidy_schema}}.
}
\description{
Returns a tibble matching \code{\link{ild_tidy_schema}} with \code{interval_type}
\code{bootstrap_percentile} (equal-tailed over replicate coefficients).
}
\seealso{
[ild_msm_bootstrap()], [ild_msm_inference]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_kfas_plot.R
\name{ild_plot_forecast}
\alias{ild_plot_forecast}
\title{Plot forecast or future simulation (stub when no horizon)}
\usage{
ild_plot_forecast(x, ...)
}
\arguments{
\item{x}{An object from [ild_kfas()].}

\item{...}{Passed to \code{\link[stats]{predict}} for the fitted \pkg{KFAS} model when forecasting is available.}
}
\value{
A \code{ggplot} object.
}
\description{
When \code{forecast_horizon > 0} was used in [ild_kfas()], attempts a short
ahead forecast via \pkg{KFAS}; otherwise returns an informative empty panel.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_power.R
\name{ild_power}
\alias{ild_power}
\title{Simulation-based power analysis for a fixed effect in ILD models}
\usage{
ild_power(
  formula,
  n_sim = 500L,
  n_id,
  n_obs_per,
  effect_size,
  test_term = NULL,
  alpha = 0.05,
  ar1 = FALSE,
  seed = 42L,
  return_sims = FALSE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{formula}{Fixed-effects formula including the predictor to power for
and random effects, e.g. \code{y ~ x + (1 | id)}. For \code{ar1 = TRUE},
use a fixed-only formula (random passed internally).}

\item{n_sim}{Integer. Number of simulation replications (default 500).}

\item{n_id}{Integer. Number of persons per replication.}

\item{n_obs_per}{Integer. Observations per person per replication.}

\item{effect_size}{Numeric. True coefficient for \code{test_term} in the DGP.}

\item{test_term}{Character or \code{NULL}. Which fixed-effect term to test.
If \code{NULL}, taken as the first non-intercept fixed-effect term from
the model (inferred from the formula).}

\item{alpha}{Numeric. Significance level for rejection (default 0.05).}

\item{ar1}{Logical. If \code{TRUE}, fit with nlme and residual AR1/CAR1
(default \code{FALSE}).}

\item{seed}{Integer. Base random seed; replication \code{i} uses \code{seed + i}.}

\item{return_sims}{Logical. If \code{TRUE}, include a tibble of per-run
estimate, std_error, p_value, rejected in the result (default \code{FALSE}).}

\item{verbose}{Logical. If \code{TRUE}, message progress (default \code{TRUE}).}

\item{...}{Passed to [ild_simulate()] (e.g. \code{irregular}, \code{wp_effect},
\code{bp_effect}) and to [ild_lme()].}
}
\value{
A list: \code{power} (proportion of converged runs with p < alpha),
  \code{n_sim}, \code{n_reject}, \code{n_converged}, \code{n_failed},
  \code{alpha}, \code{test_term}. If \code{return_sims = TRUE}, also
  \code{sim_results} (tibble of per-run results).
}
\description{
Estimates empirical power by repeatedly simulating data with a known effect
(via [ild_simulate()] plus one added predictor), fitting with [ild_lme()],
and counting the proportion of runs where the target term is significant
(Wald p < alpha). The workflow (simulate, fit, reject/retain) mirrors
simulation-based power in packages like mixpower; \code{ild_power()} is
focused on ILD and \code{ild_lme()}. For multi-parameter grids, LRT, or
general LMMs, consider mixpower.
}
\details{
The data-generating process adds one predictor (name from \code{test_term})
as standard normal and adds \code{effect_size * predictor} to the outcome
on top of the base [ild_simulate()] DGP (id, time, y). No change to
\code{ild_simulate()} is required.

For \code{ar1 = FALSE} (lmer), the lme4 backend does not report p-values;
inference for the test term uses a Wald z-approximation (estimate / SE)
so that power is still computed. For \code{ar1 = TRUE} (nlme), p-values
come from the model summary.
}
\examples{
set.seed(42)
res <- ild_power(
  formula = y ~ x + (1 | id),
  n_sim = 25L,
  n_id = 15L,
  n_obs_per = 10L,
  effect_size = 0.3,
  seed = 42L,
  verbose = FALSE
)
res$power
res$n_reject
}
\seealso{
[ild_recovery_metrics()] to summarize bias, RMSE, and coverage from
  \code{return_sims}. \href{../doc/benchmark-simulation-recovery.html}{\code{vignette("benchmark-simulation-recovery", package = "tidyILD")}}
  (worked example).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_lme.R
\name{ild_lme}
\alias{ild_lme}
\title{Fit a linear mixed-effects model to ILD}
\usage{
ild_lme(
  formula,
  data,
  ar1 = FALSE,
  correlation_class = c("auto", "AR1", "CAR1"),
  random = ~1 | .ild_id,
  warn_no_ar1 = TRUE,
  warn_uncentered = TRUE,
  ...
)
}
\arguments{
\item{formula}{Fixed-effects formula. For `ar1 = TRUE`, must be fixed-only
(e.g. `y ~ x`); random structure is set to `~ 1 | .ild_id` internally.
For `ar1 = FALSE`, formula may include random effects (e.g. `y ~ x + (1|id)`).}

\item{data}{An ILD object (see [is_ild()]).}

\item{ar1}{Logical. If `TRUE`, fit with nlme and residual AR1/CAR1
correlation; if `FALSE`, fit with lme4 (no residual correlation).}

\item{correlation_class}{Character. `"auto"` (default) uses [ild_spacing_class()]
to choose CAR1 (irregular-ish) or AR1 (regular-ish). Use `"CAR1"` or
`"AR1"` to override.}

\item{random}{For `ar1 = TRUE`, the random effects formula (default
`~ 1 | .ild_id`). Must use `.ild_id` as grouping for correlation to match.}

\item{warn_no_ar1}{If `TRUE` (default), warn when `ar1 = FALSE` that
temporal autocorrelation is not modeled.}

\item{warn_uncentered}{If `TRUE` (default), warn when a predictor in the
formula varies both within and between persons but is not decomposed
(no \code{_wp}/\code{_bp}); suggests using [ild_center()].}

\item{...}{Passed to [lme4::lmer()] or [nlme::lme()].}
}
\value{
A fitted model object (class `lmerMod` or `lme`) with attribute
  `ild_data` (the ILD data) and `ild_ar1` (logical). When `ar1 = TRUE`,
  the returned object has class `ild_lme` prepended and attribute
  `ild_random_resolved` (the formula actually passed to nlme, e.g. `~ 1 | M2ID`).
  See [ild_diagnostics()] and [ild_plot()].
}
\description{
When `ar1 = FALSE`, fits with [lme4::lmer()] (no residual correlation).
When `ar1 = TRUE`, fits with [nlme::lme()] using a residual correlation
structure: CAR1 (continuous-time) by default for irregular spacing,
or AR1 when spacing is regular-ish. Use [ild_spacing_class()] to inform
the choice; override with `correlation_class`.
}
\examples{
# lme4 path: formula includes random effects
set.seed(1)
dat <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 1)
dat <- ild_prepare(dat, id = "id", time = "time")
dat <- ild_center(dat, y)
fit_lmer <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = dat,
                    ar1 = FALSE, warn_no_ar1 = FALSE)
# nlme path (may not converge on all platforms; see ?nlme::lme)
\dontrun{
fit_lme <- ild_lme(y ~ y_bp + y_wp, data = dat,
                   random = ~ 1 | id, ar1 = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_balance.R
\name{ild_msm_overlap_plot}
\alias{ild_msm_overlap_plot}
\title{Propensity overlap plot (pooled or sequential MSM IPTW)}
\usage{
ild_msm_overlap_plot(data, treatment, source = c("auto", "pooled", "msm"))
}
\arguments{
\item{data}{ILD with treatment column and relevant \code{glm} attributes.}

\item{treatment}{Binary treatment column name.}

\item{source}{Character: \code{"auto"} (prefer MSM fits if present), \code{"pooled"}, or
\code{"msm"}.}
}
\value{
A \code{ggplot} object.
}
\description{
For **pooled** [ild_iptw_weights()], uses \code{attr(data, "ild_iptw_fit")} and plots the
distribution of fitted \eqn{P(A=1\mid L)} by treatment level.
}
\details{
For **sequential** [ild_iptw_msm_weights()], uses \code{attr(data, "ild_iptw_msm_fits")} and
facets by occasion: fitted denominator propensity at each \eqn{t}.
}
\examples{
if (requireNamespace("ggplot2", quietly = TRUE)) {
  set.seed(3)
  d <- ild_simulate(n_id = 15, n_obs_per = 5, seed = 3)
  d$stress <- rnorm(nrow(d))
  d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
  x <- ild_prepare(d, id = "id", time = "time")
  x <- ild_iptw_weights(x, treatment = "trt", predictors = "stress")
  ild_msm_overlap_plot(x, treatment = "trt", source = "pooled")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_brms.R
\name{ild_brms}
\alias{ild_brms}
\title{Fit a Bayesian mixed model to ILD data with \pkg{brms}}
\usage{
ild_brms(
  formula,
  data,
  prior = NULL,
  prior_template = c("default", "weakly_informative", "minimal_shrinkage"),
  warn_uncentered = TRUE,
  ...
)
}
\arguments{
\item{formula}{A \pkg{brms} / \pkg{lme4}-style formula (e.g. \code{y ~ x + (1|id)}).}

\item{data}{An ILD object from [ild_prepare()].}

\item{prior}{Optional \code{brmsprior} object. If \code{NULL}, built from \code{prior_template}.}

\item{prior_template}{Passed to [ild_prior_ild()] when \code{prior} is \code{NULL}.}

\item{warn_uncentered}{If \code{TRUE} (default), warn when a numeric predictor varies
within and between persons without \code{_wp}/\code{_bp} (same idea as [ild_lme()]).}

\item{...}{Arguments passed to [brms::brm()] (e.g. \code{chains}, \code{iter},
\code{warmup}, \code{control}, \code{backend}, \code{seed}).}
}
\value{
A \code{brmsfit} object with attributes:
  \describe{
    \item{\code{ild_data}}{The ILD data frame.}
    \item{\code{ild_posterior}}{List: sampler settings, prior summary, divergences, etc.}
    \item{\code{ild_provenance}}{Standard tidyILD analysis provenance.}
  }
}
\description{
Wraps [brms::brm()] with ILD validation, default priors for common longitudinal
mixed models, standardized posterior metadata in \code{attr(fit, "ild_posterior")},
and analysis provenance (priors, chains, iterations). Attach \code{ild_data} for
use with [ild_tidy()], [ild_augment()], [ild_diagnose()], and [ild_methods()].
}
\details{
Requires packages \pkg{brms} and a Stan backend (\pkg{rstan} or \pkg{cmdstanr}).
}
\seealso{
[ild_fit()] with \code{backend = "brms"} is equivalent to \code{ild_brms()}.
  For IPW/MSM workflows, full Bayesian joint MSM is not in tidyILD; see [ild_msm_inference].
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_missing_pattern.R
\name{ild_missing_pattern}
\alias{ild_missing_pattern}
\title{Summarize missingness pattern in ILD}
\usage{
ild_missing_pattern(
  x,
  vars = NULL,
  max_ids = NULL,
  seed = NULL,
  outcome = NULL,
  expected_occasions = NULL
)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{vars}{Optional character vector of variable names to summarize.
If \code{NULL}, all non-.ild_* data columns are used.}

\item{max_ids}{Optional integer. If set, subset to this many persons (sampled)
before computing \code{by_id}, \code{summary}, and \code{plot} to handle large N.}

\item{seed}{Optional integer. Seed for sampling when \code{max_ids} is set.}

\item{outcome}{Optional character; if set, \code{by_id} is left-joined with
[ild_missing_compliance()] for that column (interpretable adherence metrics).}

\item{expected_occasions}{Passed to [ild_missing_compliance()] when \code{outcome} is set.}
}
\value{
A list with: \code{summary} (tibble: one row per var, columns var, n_obs, n_na, pct_na),
  \code{plot} (ggplot2 object for missingness heatmap), \code{by_id}, \code{overall},
  \code{n_complete}, \code{vars}. When \code{outcome} is set, \code{by_id} includes
  compliance columns (see [ild_missing_compliance()]).
}
\description{
Returns a tabular summary of missingness by person and/or by variable,
plus an optional heatmap plot. Complements [ild_summary()] and supports
checking data before modeling. When \code{vars = NULL}, all non-internal
data columns are used (observation presence across variables).
}
\details{
Also a **section provider** for [ild_diagnose()] (see [ild_diagnostics_utilities]).
}
\section{Bundle integration}{

[ild_diagnose()] passes summaries into \code{data$missing_pattern} (global) and
\code{missingness} (model variables) on the \code{\link{ild_diagnostics_bundle}}.
}

\seealso{
[ild_diagnose()], [ild_diagnostics_bundle()]

Other ild_diagnostics_utilities: 
\code{\link{ild_design_check}()},
\code{\link{ild_ipcw_weights}()},
\code{\link{ild_iptw_msm_weights}()},
\code{\link{ild_iptw_weights}()},
\code{\link{ild_ipw_weights}()},
\code{\link{ild_joint_msm_weights}()},
\code{\link{ild_missing_cohort}()},
\code{\link{ild_missing_compliance}()},
\code{\link{ild_missing_hazard_first}()},
\code{\link{ild_missing_model}()},
\code{\link{ild_missingness_report}()}
}
\concept{ild_diagnostics_utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_tvem.R
\name{ild_tvem}
\alias{ild_tvem}
\title{Fit a time-varying effects model (TVEM) for ILD}
\usage{
ild_tvem(
  data,
  outcome,
  predictor,
  time_var = ".ild_time_num",
  k = 10L,
  re_id = TRUE,
  ...
)
}
\arguments{
\item{data}{An ILD object (see [is_ild()]).}

\item{outcome}{Character. Name of the outcome variable.}

\item{predictor}{Character. Name of the predictor with a time-varying effect.}

\item{time_var}{Character. Name of the time variable (default \code{".ild_time_num"}).}

\item{k}{Integer. Basis dimension for smooth terms (default 10).}

\item{re_id}{Logical. If \code{TRUE} (default), include a random intercept by person (\code{s(.ild_id, bs="re")}).}

\item{...}{Passed to [mgcv::gam()].}
}
\value{
A fitted \code{gam} object with class \code{c("tidyild_tvem", "gam", ...)}
  and attribute \code{ild_tvem_meta} (list with \code{outcome}, \code{predictor},
  \code{time_var}, \code{k}, \code{re_id}).
}
\description{
Fits a GAM with a smooth in time and a time-varying coefficient for a
predictor using [mgcv::gam()]. Use [ild_tvem_plot()] to plot the
time-varying effect. Requires \code{.ild_time_num} (or a numeric time column).
}
\examples{
set.seed(1)
d <- ild_simulate(n_id = 10, n_obs_per = 15, seed = 1)
d$x <- rnorm(nrow(d))
x <- ild_prepare(d, id = "id", time = "time")
tv <- ild_tvem(x, "y", "x", k = 5, re_id = TRUE)
ild_tvem_plot(tv)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_missingness_report.R
\name{ild_missingness_report}
\alias{ild_missingness_report}
\title{Missingness workflow report (orchestration)}
\usage{
ild_missingness_report(
  x,
  outcome,
  predictors = NULL,
  fit_missing_model = TRUE,
  random = FALSE,
  expected_occasions = NULL,
  max_ids = NULL,
  seed = NULL,
  cohort_plot = TRUE
)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{outcome}{Character; outcome column with possible `NA`.}

\item{predictors}{Optional character vector passed to [ild_missing_model()]
when `fit_missing_model` is `TRUE`. If empty, no missingness model is fit.}

\item{fit_missing_model}{Logical; fit [ild_missing_model()] when predictors
are non-empty.}

\item{random}{Logical; passed to [ild_missing_model()] (`glmer` vs `glm`).}

\item{expected_occasions}{Optional integer for [ild_missing_compliance()].}

\item{max_ids, seed}{Passed to [ild_missing_pattern()] for large cohorts.}

\item{cohort_plot}{Logical; include ggplot for [ild_missing_cohort()].}
}
\value{
A list (class `ild_missingness_report`) with:
  \describe{
    \item{compliance}{Tibble from [ild_missing_compliance()].}
    \item{pattern}{Output of [ild_missing_pattern()] (includes compliance
      columns on `by_id` if `outcome` was passed through).}
    \item{cohort}{List from [ild_missing_cohort()].}
    \item{hazard}{Tibble from [ild_missing_hazard_first()].}
    \item{flags}{Named list: `dropout_late_pooled` (logical from same
      heuristic as guardrail `GR_DROPOUT_LATE_CONCENTRATION`).}
    \item{missing_model}{Result of [ild_missing_model()] or `NULL`.}
    \item{snippets}{Character vector of short paragraphs for copy-paste.}
  }
}
\description{
Bundles person-level compliance, pattern summaries, cohort and hazard tables,
optional [ild_missing_model()], heuristic flags, and short text snippets for
methods sections. This is **diagnostic and reporting** tooling, not a substitute
for formal sensitivity analysis or MNAR models.
}
\seealso{
[ild_missing_pattern()], [ild_missing_bias()], [ild_ipw_weights()],
  `vignette("ild-missingness-workflow", package = "tidyILD")`

Other ild_diagnostics_utilities: 
\code{\link{ild_design_check}()},
\code{\link{ild_ipcw_weights}()},
\code{\link{ild_iptw_msm_weights}()},
\code{\link{ild_iptw_weights}()},
\code{\link{ild_ipw_weights}()},
\code{\link{ild_joint_msm_weights}()},
\code{\link{ild_missing_cohort}()},
\code{\link{ild_missing_compliance}()},
\code{\link{ild_missing_hazard_first}()},
\code{\link{ild_missing_model}()},
\code{\link{ild_missing_pattern}()}
}
\concept{ild_diagnostics_utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_manifest.R
\name{ild_manifest}
\alias{ild_manifest}
\title{Create a reproducibility manifest}
\usage{
ild_manifest(
  seed = NULL,
  scenario = NULL,
  include_session = TRUE,
  include_git = FALSE,
  git_path = "."
)
}
\arguments{
\item{seed}{Optional integer. Seed used for the run (e.g. from
[ild_simulate()] or set before fitting). Not captured automatically;
pass explicitly if you want it in the manifest.}

\item{scenario}{Optional. Named list or character string describing the
run (e.g. formula, n_obs, n_id, ar1). Build from [ild_summary()] or
a short list when calling after [ild_lme()] / [ild_diagnostics()].}

\item{include_session}{Logical. If `TRUE` (default), include
[utils::sessionInfo()] in the manifest. Set to `FALSE` to reduce size.}

\item{include_git}{Logical. If `TRUE`, attempt to record the current
git commit SHA from \code{git_path}. Default `FALSE`.}

\item{git_path}{Character. Path to the repository root (default
\code{"."}). Used only when \code{include_git = TRUE}.}
}
\value{
A list with elements \code{timestamp} (POSIXct), \code{seed}
  (integer or NULL), \code{scenario} (as provided or NULL),
  \code{session_info} (list from sessionInfo() or NULL),
  \code{git_sha} (length-1 character or NA). All elements are
  serializable.
}
\description{
Captures timestamp, optional seed, optional scenario fingerprint,
session info, and optional git SHA for use when saving or serializing
results (e.g. after [ild_lme()] or [ild_diagnostics()]). The return
value is a serializable list suitable for [saveRDS()] or [ild_bundle()].
}
\examples{
m <- ild_manifest()
names(m)
m <- ild_manifest(seed = 42, scenario = list(n_obs = 100, formula = "y ~ x"))
m$seed
m$scenario
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_simulation.R
\name{ild_msm_simulate_scenario}
\alias{ild_msm_simulate_scenario}
\title{Simulate a simple longitudinal MSM scenario}
\usage{
ild_msm_simulate_scenario(
  n_id = 60L,
  n_obs_per = 10L,
  true_ate = 0.5,
  censoring = TRUE,
  positivity_stress = 1,
  treatment_strength = 0.8,
  censoring_strength = 0.6,
  seed = 42L
)
}
\arguments{
\item{n_id}{Number of persons.}

\item{n_obs_per}{Planned observations per person.}

\item{true_ate}{True additive treatment effect on outcome.}

\item{censoring}{If \code{TRUE}, apply monotone dropout and return observed rows only.}

\item{positivity_stress}{Multiplier for treatment/censoring logits
(>1 worsens overlap; <1 improves overlap).}

\item{treatment_strength}{Base strength of treatment dependence on confounders/history.}

\item{censoring_strength}{Base strength of dropout dependence on stress/treatment.}

\item{seed}{Integer seed.}
}
\value{
ILD object with columns \code{stress} and binary \code{trt}; attributes include
  \code{true_ate}.
}
\description{
Generates ILD with time-varying treatment, confounding, and optional monotone
censoring/dropout under a known treatment effect (\code{true_ate}).
}
\examples{
d <- ild_msm_simulate_scenario(n_id = 20, n_obs_per = 8, true_ate = 0.5, seed = 12)
ild_summary(d)$summary
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_missing_model.R
\name{ild_missing_model}
\alias{ild_missing_model}
\title{Fit a model for missingness (diagnostic / sensitivity)}
\usage{
ild_missing_model(
  x,
  outcome,
  predictors,
  random = FALSE,
  family = stats::binomial(),
  ...
)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{outcome}{Character. Name of the variable with missingness (e.g. \code{"mood"}).}

\item{predictors}{Character vector. Names of covariates to predict missingness.}

\item{random}{Logical. If \code{TRUE}, fit \code{glmer(is_missing ~ ... + (1|id))};
if \code{FALSE} (default), fit \code{glm(is_missing ~ ., family = binomial)}.}

\item{family}{Passed to \code{glm} / \code{glmer} (default \code{binomial()}).}

\item{...}{Passed to \code{glm()} or \code{lme4::glmer()}.}
}
\value{
A list with \code{fit} (the fitted model or \code{NULL} if no/all missing),
  \code{tidy} (tibble: term, estimate, std_error, p_value), \code{outcome},
  \code{predictors}, and \code{message}. If \code{fit} is not \code{NULL},
  \code{p_missing} is a numeric vector of predicted P(missing) per row (aligned to \code{x}).
}
\description{
Fits a logistic model predicting whether the outcome is missing from
covariates. Use as a diagnostic; then [ild_ipw_weights()] to compute
inverse-probability weights and [ild_ipw_refit()] for a sensitivity analysis.
This is not a full MNAR solution—treat as diagnostic and sensitivity tooling.
}
\details{
**Not** run automatically by [ild_diagnose()]; it is the usual prerequisite for
[ild_ipw_weights()] when you want IPW summaries in the bundle (see [ild_diagnostics_utilities]).
}
\section{Bundle integration}{

There is no separate \dQuote{missingness model} slot on the bundle. This function
supports the IPW workflow; after [ild_ipw_weights()], [ild_diagnose()] can summarize
weights in \code{causal} on \code{\link{ild_diagnostics_bundle}}.
}

\examples{
set.seed(1)
d <- ild_simulate(n_id = 15, n_obs_per = 8, seed = 1)
d$stress <- rnorm(nrow(d))
d$mood <- d$y
d$mood[sample(nrow(d), 20)] <- NA
x <- ild_prepare(d, id = "id", time = "time")
mm <- ild_missing_model(x, "mood", "stress")
mm$tidy
}
\seealso{
[ild_diagnose()], [ild_diagnostics_bundle()], [ild_ipw_weights()]

Other ild_diagnostics_utilities: 
\code{\link{ild_design_check}()},
\code{\link{ild_ipcw_weights}()},
\code{\link{ild_iptw_msm_weights}()},
\code{\link{ild_iptw_weights}()},
\code{\link{ild_ipw_weights}()},
\code{\link{ild_joint_msm_weights}()},
\code{\link{ild_missing_cohort}()},
\code{\link{ild_missing_compliance}()},
\code{\link{ild_missing_hazard_first}()},
\code{\link{ild_missing_pattern}()},
\code{\link{ild_missingness_report}()}
}
\concept{ild_diagnostics_utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild-class.R
\name{validate_ild}
\alias{validate_ild}
\title{Validate an ILD object and error if invalid}
\usage{
validate_ild(x)
}
\arguments{
\item{x}{Object to validate (expected to be an ILD tibble).}
}
\value{
Invisibly returns \code{x} if valid.
}
\description{
Checks presence and types of `.ild_*` columns and `ild_*` attributes.
Errors with a clear message if anything is missing or invalid.
Calls \code{ild_normalize_internal()} so legacy objects get \code{attr(x, "tidyILD")} and class \code{tidyild_df}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_recovery_metrics.R
\name{ild_recovery_metrics}
\alias{ild_recovery_metrics}
\title{Recovery metrics from simulation replications}
\usage{
ild_recovery_metrics(sim_results, truth, level = 0.95, converged_only = TRUE)
}
\arguments{
\item{sim_results}{A data frame or tibble with columns \code{estimate},
\code{std_error}, and \code{converged} (as in \code{ild_power(..., return_sims = TRUE)$sim_results}).}

\item{truth}{Known value of the estimand (e.g. \code{effect_size} in [ild_power()]).}

\item{level}{Nominal coverage level for Wald intervals (default \code{0.95}).}

\item{converged_only}{If \code{TRUE} (default), only rows with \code{converged == TRUE}
are used.}
}
\value{
A one-row \code{\link[tibble]{tibble}} with \code{truth}, \code{n} (rows used),
  \code{n_total} (rows in \code{sim_results} if \code{converged_only}), \code{mean_estimate},
  \code{bias}, \code{rmse}, \code{coverage}, \code{level}.
}
\description{
Summarizes **bias**, **RMSE**, and **nominal Wald interval coverage** for a known
coefficient using the per-replication table from [ild_power()] when
\code{return_sims = TRUE}. Purely descriptive; does not run new simulations.
}
\seealso{
[ild_power()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_diagnostics.R
\name{plot_ild_diagnostics}
\alias{plot_ild_diagnostics}
\title{Plot diagnostics from an ild_diagnostics object}
\usage{
plot_ild_diagnostics(diag, type = NULL)
}
\arguments{
\item{diag}{An object returned by [ild_diagnostics()].}

\item{type}{Character vector. Which plots to build (default: the types stored in \code{diag$meta$type}).}
}
\value{
A named list of ggplot objects (e.g. \code{residual_acf}, \code{residuals_vs_fitted}, \code{residuals_vs_time}, \code{qq}).
}
\description{
Generates ggplot objects for the requested diagnostic types. Plots are not
stored in the diagnostics object; call this function to create them.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_simulate.R
\name{ild_simulate}
\alias{ild_simulate}
\title{Simulate simple ILD for examples, tests, and power analysis}
\usage{
ild_simulate(
  n_id = 5L,
  n_obs_per = 10L,
  n_time = NULL,
  irregular = FALSE,
  ar1 = NULL,
  wp_effect = 0.5,
  bp_effect = 1,
  seed = 42L
)
}
\arguments{
\item{n_id}{Integer. Number of persons (default 5).}

\item{n_obs_per}{Integer. Observations per person (default 10).}

\item{n_time}{Integer. Alias for \code{n_obs_per} (observations per person). If provided, overrides \code{n_obs_per}.}

\item{irregular}{Logical. If \code{TRUE}, add random jitter to time (default \code{FALSE}).}

\item{ar1}{Numeric or \code{NULL}. If numeric, within-person AR(1) correlation (e.g. 0.4). If \code{NULL} or 0, no AR (default \code{NULL}).}

\item{wp_effect}{Numeric. Scale (SD) of within-person innovation (default 0.5).}

\item{bp_effect}{Numeric. Scale (SD) of between-person random intercept (default 1).}

\item{seed}{Integer. Random seed for reproducibility (default 42).}
}
\value{
A data frame with columns \code{id}, \code{time} (POSIXct), and \code{y}.
}
\description{
Generates a tibble with id, time, and outcome \code{y}. Optionally uses
AR(1) within-person correlation and configurable WP/BP variance. Use
[ild_prepare()] after to get a proper ILD object.
}
\examples{
d <- ild_simulate(n_id = 3, n_obs_per = 5, seed = 1)
x <- ild_prepare(d, id = "id", time = "time")
d2 <- ild_simulate(n_id = 100, n_time = 50, ar1 = 0.4, wp_effect = 0.6,
  bp_effect = 0.3, irregular = TRUE, seed = 1)
}
\seealso{
\href{../doc/benchmark-simulation-recovery.html}{\code{vignette("benchmark-simulation-recovery", package = "tidyILD")}}
  (simulation benchmarks and DGP description).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_prepare.R
\name{ild_prepare}
\alias{ild_prepare}
\title{Prepare a data frame as an ILD (intensive longitudinal data) object}
\usage{
ild_prepare(
  data,
  id = NULL,
  time = NULL,
  gap_threshold = Inf,
  duplicate_handling = c("first", "last", "error", "collapse"),
  collapse_fn = NULL
)
}
\arguments{
\item{data}{A data frame or tibble with at least an id and a time column.
  If \code{id} and \code{time} are both omitted, \code{data} must be a
  \code{tbl_ts} from \pkg{tsibble} (requires a single key column and an index);
  key and index names are inferred (see [ild_as_tsibble()] for the reverse).

  \strong{Tsibble interoperability:} Accepting a \code{tbl_ts} is an input
  convenience; the result is always a plain ILD tibble (the \code{tbl_ts} class
  is dropped). Metadata from the source tsibble (key, index, interval summary,
  regularity) is stored in \code{attr(x, "tidyILD")$tsibble} when available;
  use [ild_tsibble_meta()] to read it. Use [ild_as_tsibble()] for a
  best-effort round-trip. \strong{Conceptual choice (keys):} ILD uses one
  subject identifier per row; \code{tbl_ts} must have exactly one key
  column. Compound keys are not supported—combine levels into a single id
  column before calling \code{ild_prepare()}.}

\item{id}{Character. Name of the subject/unit identifier column. Omit both
\code{id} and \code{time} when \code{data} is a tsibble (see above).}

\item{time}{Character. Name of the time column (Date, POSIXct, or numeric).}

\item{gap_threshold}{Numeric. Time distance above which an interval is
flagged as a gap (`.ild_gap` TRUE). Same units as the numeric time
(e.g. seconds if time is POSIXct). Use `Inf` to disable gap flagging.}

\item{duplicate_handling}{Character. How to handle duplicate timestamps
within the same id: `"first"` (keep first), `"last"` (keep last),
`"error"` (stop with an error), `"collapse"` (aggregate with \code{collapse_fn}).}

\item{collapse_fn}{Named list of functions, one per variable to collapse.
Used only when \code{duplicate_handling = "collapse"}. E.g. \code{list(x = mean, y = function(z) z[1])}.
Variables not in \code{collapse_fn} keep their first value within the duplicate group.}
}
\value{
An ILD tibble with `.ild_*` columns and metadata attributes.
  Spacing metadata (see [ild_meta()]) includes overall stats and a
  \code{by_id} tibble of per-person spacing stats (median_dt, iqr_dt,
  n_intervals, pct_gap). Use [ild_summary()] to inspect and check gap
  flags before modeling.
}
\description{
Validates and encodes longitudinal structure: parses time, sorts by id and
time, handles duplicate timestamps, and adds internal columns (`.ild_*`)
and metadata. All downstream functions assume the result of `ild_prepare()`.
}
\seealso{
\href{../doc/tsibble-interoperability.html}{\code{vignette("tsibble-interoperability", package = "tidyILD")}}
  (tsibble ingestion, provenance, \code{ild_as_tsibble()}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_heterogeneity.R
\name{ild_heterogeneity}
\alias{ild_heterogeneity}
\alias{ild_heterogeneity.lmerMod}
\alias{ild_heterogeneity.lme}
\alias{ild_heterogeneity_stratified}
\alias{ild_tidy.ild_heterogeneity}
\alias{ild_autoplot.ild_heterogeneity}
\alias{ild_heterogeneity.brmsfit}
\title{Heterogeneity and person-specific effects from mixed models}
\usage{
ild_heterogeneity(fit, ...)

\method{ild_heterogeneity}{lmerMod}(
  fit,
  term = NULL,
  group = NULL,
  threshold = NULL,
  scale = c("raw", "sd_x", "sd_y"),
  conf_level = 0.95,
  ...
)

\method{ild_heterogeneity}{lme}(
  fit,
  term = NULL,
  group = NULL,
  threshold = NULL,
  scale = c("raw", "sd_x", "sd_y"),
  conf_level = 0.95,
  ...
)

ild_heterogeneity_stratified(formula, data, subgroup, min_n_id = 5L, ...)

\method{ild_tidy}{ild_heterogeneity}(x, ...)

\method{ild_autoplot}{ild_heterogeneity}(x, type = c("caterpillar", "histogram"), term = NULL, ...)

\method{ild_heterogeneity}{brmsfit}(
  fit,
  term = NULL,
  group = NULL,
  threshold = NULL,
  scale = c("raw", "sd_x", "sd_y"),
  n_draws = 500L,
  ...
)
}
\arguments{
\item{fit}{A fitted \code{lmerMod}, \code{lme}, or \code{brmsfit} (from [ild_brms()]).}

\item{...}{Passed to [ild_lme()].}

\item{term}{Which random-effect term to plot (required if multiple).}

\item{group}{Name of the grouping factor in \code{ranef(fit)} to use (default first).}

\item{threshold}{Optional numeric; with \code{scale}, proportion of person-specific
**total** coefficients exceeding this threshold is reported in \code{summary}.}

\item{scale}{If \code{threshold} is set: \code{"raw"} (default), \code{"sd_x"} (multiply
threshold by residual SD of a focal predictor from model frame), or \code{"sd_y"}
(multiply by SD of response). Only used when a single numeric predictor is found
for \code{sd_x} / \code{sd_y} heuristics.}

\item{conf_level}{For \code{lmerMod}, Wald interval level for conditional modes (normal approx.).}

\item{formula, data}{Passed to [ild_lme()] within each subgroup.}

\item{subgroup}{Character name of a column in \code{data} defining subgroups.}

\item{min_n_id}{Minimum number of distinct persons required in a subgroup.}

\item{x}{An \code{ild_heterogeneity} object.}

\item{type}{\code{"caterpillar"} (default) or \code{"histogram"}.}

\item{n_draws}{For \code{brmsfit}, number of posterior draws for summaries (capped).}
}
\value{
An object of class \code{ild_heterogeneity}: a list with \code{meta}, \code{random_effects}
  (long tibble), \code{varcorr} (tibble), \code{summary} (per-term metrics), and \code{fit}.

A tibble with one row per successful subgroup fit, including key columns from
  \code{ild_heterogeneity()$summary} flattened (\code{subgroup_level}, \code{term}, etc.).
  This is a **descriptive** comparison across refits, not a formal test of variance heterogeneity.
}
\description{
Summarizes **conditional modes** (empirical Bayes / BLUP deviations) and
**person-specific coefficients** (\code{fixef + ranef}) for hierarchical models
fit with [ild_lme()] or [ild_brms()]. Outputs are labeled as **partial-pooling**
estimates; they are not the same as no-pooling separate regressions (see
[ild_person_model()]) or population-average fixed effects alone.
}
\section{Methods (by generic)}{
\itemize{
\item \code{ild_tidy(ild_heterogeneity)}: Tidy long table of person-level random effects (for \code{ild_heterogeneity} objects).

\item \code{ild_autoplot(ild_heterogeneity)}: Caterpillar or histogram of person-specific effects.

}}
\seealso{
[ild_tidy.ild_heterogeneity()], [ild_autoplot.ild_heterogeneity()],
  [ild_heterogeneity_stratified()], [ild_person_model()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.R
\docType{package}
\name{tidyILD-package}
\alias{tidyILD}
\alias{tidyILD-package}
\title{tidyILD: Tidy Intensive Longitudinal Data Analysis}
\description{
tidyILD is a reproducible, tidyverse-style framework for intensive longitudinal
data (ILD) analysis in R, with built-in methodological safeguards, provenance
tracking, and reporting tools. It supports ecological momentary assessment
(EMA) and diary studies with a tidy pipeline from raw data to mixed-effects
models: explicit time structure, within-between decomposition, spacing-aware
lags, and diagnostics. Use it when you have repeated measures per person over
time and want consistent handling of time, gaps, centering, and residual
correlation (AR1/CAR1).
}
\details{
All ILD structure (`.ild_*` columns and `ild_*` metadata) is created only
by \code{\link{ild_prepare}} (via the internal constructor). Downstream
functions expect data prepared with \code{ild_prepare()}. For the full
workflow and applications, see the vignettes. Estimands that require
\strong{joint multivariate dynamics}, \strong{penalized high-dimensional}
longitudinal models, or full \strong{latent-variable DSEM} are better handled
by specialist packages after preprocessing; see
\code{vignette("ild-specialist-backends", package = "tidyILD")}.
}
\section{Getting started}{

A minimal workflow: simulate or load data, prepare with
\code{\link{ild_prepare}}, inspect with \code{\link{ild_summary}}, apply
\code{\link{ild_center}} and \code{\link{ild_lag}}, fit with
\code{\link{ild_lme}} or \code{\link{ild_fit}} (same engines plus optional
\code{backend = "brms"}), then \code{\link{ild_diagnostics}} or
\code{\link{ild_plot}}. State-space: \code{\link{ild_kfas}}. See the examples below.
}

\section{Function index by topic}{

\describe{
  \item{Contracts (schemas)}{\code{\link{ild_diagnostics_bundle}}, \code{\link{guardrail_registry}},
    \code{\link{ild_tidy_schema}}, \code{\link{ild_augment_schema}}}
  \item{Setup and validation}{\code{\link{ild_prepare}}, \code{\link{ild_as_tsibble}},
    \code{\link{as_ild}},
    \code{\link{is_ild}}, \code{\link{validate_ild}}, \code{\link{ild_meta}}}
  \item{Summaries and inspection}{\code{\link{ild_summary}},
    \code{\link{ild_spacing_class}}, \code{\link{ild_spacing}},
    \code{\link{ild_design_check}}, \code{\link{ild_missing_pattern}},
    \code{\link{ild_missing_compliance}}, \code{\link{ild_missing_cohort}},
    \code{\link{ild_missing_hazard_first}}, \code{\link{ild_missingness_report}},
    \code{\link{ild_missing_bias}}, \code{\link{ild_missing_model}},
    \code{\link{ild_ipw_weights}}, \code{\link{ild_ipw_refit}},
    \code{\link{ild_plot}} (types: trajectory, gaps, missingness, predicted_trajectory).
    These overlap with \code{\link{ild_diagnostics_utilities}} (bundle section providers).}
  \item{Within-person and lags}{\code{\link{ild_center}}, \code{\link{ild_center_plot}},
    \code{\link{ild_decomposition}}, \code{\link{ild_lag}}, \code{\link{ild_check_lags}},
    \code{\link{ild_panel_lag_prepare}}, \code{\link{ild_crosslag}}, \code{\link{ild_align}}}
  \item{Modeling}{\code{\link{ild_fit}} (unified: lme4, nlme, brms), \code{\link{ild_lme}},
    \code{\link{ild_brms}} (Bayesian), \code{\link{ild_kfas}} (KFAS state-space; not via \code{ild_fit}),
    \code{\link{ild_person_model}}, \code{\link{ild_heterogeneity}}, \code{\link{ild_heterogeneity_stratified}},
    \code{\link{ild_tvem}} (time-varying effects)}
  \item{Diagnostics and visualization}{\code{\link{ild_diagnose}} (\code{\link{ild_diagnostics_bundle}}),
    \code{\link{ild_diagnostics_utilities}}, \code{\link{ild_acf}}, \code{\link{ild_diagnostics}},
    \code{\link{ild_plot}} (types: fitted, predicted_trajectory, residual_acf; optional \code{facet_by}),
    \code{\link{ild_plot_predicted_trajectory}}, \code{\link{ild_heatmap}},
    \code{\link{ild_spaghetti}}, \code{\link{ild_circadian}}, \code{\link{ild_tvem_plot}}}
  \item{Provenance and methods}{\code{\link{ild_provenance}}, \code{\link{ild_history}},
    \code{\link{ild_methods}}, \code{\link{ild_report}}, \code{\link{ild_compare_pipelines}},
    \code{\link{ild_compare_fits}}, \code{\link{ild_export_provenance}}}
  \item{Reproducibility}{\code{\link{ild_manifest}}, \code{\link{ild_bundle}}}
  \item{Package standards (developers)}{\code{vignette("developer-contracts", package = "tidyILD")};
    normative spec: \code{system.file("dev", "DEVELOPER_CONTRACTS.md", package = "tidyILD")}}
  \item{Utilities and data}{\code{\link{ild_simulate}}, \code{\link{ild_power}}, \code{\link{ema_example}}}
  \item{Person-level}{\code{\link{ild_person_model}}, \code{\link{ild_person_distribution}}}
  \item{Model tidiers}{\code{\link{ild_prior_ild}}, \code{\link{ild_tidy}}, \code{\link{ild_augment}},
    \code{\link{ild_diagnose}}, \code{\link{ild_autoplot}},
    \code{\link{augment_ild_model}}, \code{\link{tidy_ild_model}}
    (model or robust SE via \code{se = "robust"}), \code{\link{ild_robust_se}};
    \code{\link{tidy.ild_lme}}, \code{\link{augment.ild_lme}} (broom.mixed, see \code{\link{broom_ild_lme}})}
}
}

\section{Visualization}{

\code{\link{ild_plot}}, \code{\link{ild_spaghetti}}, and \code{\link{ild_heatmap}} cover trajectories,
heatmaps, gaps, and missingness; optional \code{facet_by} adds panels (e.g. by cluster).
For observed and fitted lines over time, use \code{\link{ild_plot_predicted_trajectory}} or
\code{ild_plot(..., type = "predicted_trajectory")}. After \code{\link{ild_diagnose}}, use
\code{\link{ild_autoplot}} on the bundle. Backend-specific plots include \code{\link{ild_tvem_plot}},
\code{\link{ild_plot_states}}, \code{\link{ild_plot_forecast}}, and \code{\link{ild_plot_filtered_vs_smoothed}}
(KFAS). See \code{vignette("visualization-in-tidyILD", package = "tidyILD")} for a question-to-function map
and recipes (including partial effects via \strong{marginaleffects} / \strong{ggeffects} on \code{_wp} / \code{_bp} columns).
}

\section{Vignettes}{

\code{browseVignettes("tidyILD")} lists all vignettes. Key entries:
\itemize{
  \item \emph{From raw data to model with tidyILD} --- Full pipeline:
    prepare, inspect, center, lag, fit, diagnose.
  \item \emph{Visualization in tidyILD} --- Index of plots, bundle sections, \code{facet_by},
    predicted trajectories, and partial-effects templates.
  \item \emph{Short analysis report} --- Fit, tidy fixed effects,
    fitted vs observed, residual ACF and Q-Q.
  \item \emph{Within-between decomposition and irregular spacing} ---
    Centering (BP/WP), gap-aware lags, spacing classification.
  \item \emph{Temporal dynamics: choosing a model for ILD} ---
    Maps estimands to lags, residual AR (\code{ild_lme}), time-varying effects (\code{ild_tvem}),
    and state-space backends (\code{ild_kfas}, \code{ild_ctsem}).
  \item \emph{Specialist backends: when to move beyond the default stack} ---
    Handoffs to dynamite, PGEE, DSEM, and multivariate workflows; export patterns after prepare/center/lag.
  \item \emph{Reproducible ILD workflows with tidyILD provenance} ---
    Inspect history, generate methods text, ild_report(), export and compare pipelines.
  \item \emph{Glossary and quick-start checklist} --- Table of main
    functions and a short checklist.
  \item \emph{Heterogeneity and person-specific effects} --- BLUPs vs \code{ild_person_model},
    \code{ild_heterogeneity()}, bundle plots.
  \item \emph{Missingness in ILD: diagnostics and sensitivity routes} ---
    \code{ild_missingness_report()}, MAR/MNAR context, IPW templates.
}
}

\section{Key concepts}{

\itemize{
  \item \strong{ILD:} Intensive longitudinal data; many repeated
    measurements per person over time (e.g. EMA).
  \item \strong{Within-between decomposition:} \code{\link{ild_center}}
    adds \code{_bp} (person mean) and \code{_wp} (within-person deviation);
    use WP for within-person effects and BP for between-person or
    cross-level terms.
  \item \strong{Spacing-aware lags:} \code{\link{ild_lag}} supports
    \code{index}, \code{gap_aware} (NA when \code{gap > max_gap}), and
    \code{time_window}; avoids misalignment from assuming equal spacing.
  \item \strong{Residual correlation:} \code{\link{ild_lme}} can fit nlme
    with AR1 or CAR1 for residual autocorrelation;
    \code{\link{ild_spacing_class}} helps choose \code{regular-ish} vs
    \code{irregular-ish} spacing.
  \item \strong{Person-level:} \code{\link{ild_person_model}} fits models separately
    per participant; \code{\link{ild_person_distribution}} plots the distribution of
    estimates across persons (N-of-1 / idiographic). \code{\link{ild_heterogeneity}} summarizes
    partial-pooling person-specific effects from mixed models (contrasts with no pooling).
}
}

\examples{
library(tidyILD)
d <- ild_simulate(n_id = 10, n_obs_per = 12, irregular = TRUE, seed = 42)
x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)
ild_summary(x)
x <- ild_center(x, y)
x <- ild_lag(x, y, mode = "gap_aware", max_gap = 7200)
fit <- ild_lme(y ~ 1, data = x, ar1 = TRUE, correlation_class = "CAR1")
ild_diagnostics(fit, data = x)
ild_plot(fit, type = "fitted")

}
\seealso{
\code{\link{browseVignettes}} and \code{vignette(package = "tidyILD")} for
vignettes. Core entry points: \code{\link{ild_prepare}}, \code{\link{ild_lme}}.
\code{vignette("ild-specialist-backends", package = "tidyILD")} for using
tidyILD as a preprocessing layer with external multivariate or high-dimensional estimators.
Related packages: \pkg{lme4}, \pkg{nlme} (model backends),
\pkg{broom.mixed} (tidiers).
}
\author{
Alex Litovchenko \email{al4877@columbia.edu}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_generics.R, R/ild_kfas_diagnose.R,
%   R/ild_diagnose_methods.R
\name{ild_diagnose}
\alias{ild_diagnose}
\alias{ild_diagnose.ild_fit_kfas}
\alias{ild_diagnose.lmerMod}
\alias{ild_diagnose.lme}
\alias{ild_diagnose.brmsfit}
\alias{ild_diagnose.ild_fit_ctsem}
\alias{ild_diagnose.default}
\title{Engine-agnostic diagnostics façade}
\usage{
ild_diagnose(object, ...)

\method{ild_diagnose}{ild_fit_kfas}(object, data = NULL, type = NULL, by_id = NULL, ...)

\method{ild_diagnose}{lmerMod}(
  object,
  data = NULL,
  type = c("residual_acf", "residual_time", "qq"),
  by_id = TRUE,
  missing_model = FALSE,
  missing_model_predictors = NULL,
  causal_detail = FALSE,
  balance = FALSE,
  balance_treatment = NULL,
  balance_covariates = NULL,
  balance_weights_col = ".ipw_treat",
  balance_by_occasion = FALSE,
  ...
)

\method{ild_diagnose}{lme}(
  object,
  data = NULL,
  type = c("residual_acf", "residual_time", "qq"),
  by_id = TRUE,
  missing_model = FALSE,
  missing_model_predictors = NULL,
  causal_detail = FALSE,
  balance = FALSE,
  balance_treatment = NULL,
  balance_covariates = NULL,
  balance_weights_col = ".ipw_treat",
  balance_by_occasion = FALSE,
  ...
)

\method{ild_diagnose}{brmsfit}(
  object,
  data = NULL,
  type = c("all", "convergence", "sampler", "ppc"),
  by_id = TRUE,
  ppc_ndraws = 500L,
  missing_model = FALSE,
  missing_model_predictors = NULL,
  causal_detail = FALSE,
  balance = FALSE,
  balance_treatment = NULL,
  balance_covariates = NULL,
  balance_weights_col = ".ipw_treat",
  balance_by_occasion = FALSE,
  ...
)

\method{ild_diagnose}{ild_fit_ctsem}(
  object,
  data = NULL,
  type = NULL,
  by_id = NULL,
  missing_model = FALSE,
  missing_model_predictors = NULL,
  causal_detail = FALSE,
  balance = FALSE,
  balance_treatment = NULL,
  balance_covariates = NULL,
  balance_weights_col = ".ipw_treat",
  balance_by_occasion = FALSE,
  ...
)

\method{ild_diagnose}{default}(object, ...)
}
\arguments{
\item{object}{A fitted model.}

\item{...}{Reserved.}

\item{data}{Optional ILD data (defaults to \code{attr(object, "ild_data")}).}

\item{type}{Residual diagnostic types for \code{lmerMod}/\code{lme} (passed to [ild_diagnostics()]).
For \code{brmsfit}: \code{"all"} or any of \code{"convergence"}, \code{"sampler"}, \code{"ppc"}.}

\item{by_id}{For \code{lmerMod}/\code{lme}: ACF by person.}

\item{missing_model}{If \code{TRUE}, fit \code{\link{ild_missing_model}} for the response (optional; can be slow). Requires at least one predictor.}

\item{missing_model_predictors}{Character vector for \code{ild_missing_model}; default \code{NULL} uses formula predictors.}

\item{causal_detail}{If \code{TRUE}, add extra quantile summaries for \code{.ipw} when present.}

\item{balance}{If \code{TRUE}, add weighted covariate balance (SMD) and ESS to \code{causal$balance}
when \code{balance_covariates} and \code{balance_treatment} are set.}

\item{balance_treatment}{Character. Binary treatment column for balance (required if \code{balance = TRUE}).}

\item{balance_covariates}{Character vector of covariate columns for SMDs (required if \code{balance = TRUE}).}

\item{balance_weights_col}{Weights column for balance (default \code{".ipw_treat"}; use \code{".ipw"} for joint MSM).}

\item{balance_by_occasion}{If \code{TRUE}, balance and ESS are stratified by \code{.ild_seq}.}

\item{ppc_ndraws}{For \code{brmsfit}: posterior predictive draws.}
}
\value{
An \code{\link{ild_diagnostics_bundle}}.
}
\description{
Returns a single \code{\link{ild_diagnostics_bundle}} for \code{lmerMod}, \code{lme},
\code{brmsfit}, \code{ild_fit_kfas}, and \code{ild_fit_ctsem}.
Sections are parallel across engines;
see \code{?ild_diagnostics_bundle}. Residual ACF/Q-Q for frequentist models are stored in
\code{residual$legacy_ild_diagnostics} for [plot_ild_diagnostics()]. For raw residual-only
computation without the bundle, call [ild_diagnostics()] directly.
}
\details{
**Building blocks:** the same analyses are available as standalone functions
([ild_design_check()], [ild_missing_pattern()], etc.); see [ild_diagnostics_utilities].
}
\seealso{
\code{\link{ild_autoplot}}, \code{\link{ild_diagnostics_bundle}},
  \code{\link{guardrail_registry}}, \code{\link{ild_diagnostics_utilities}},
  \code{\link{ild_diagnostics}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_missing_extensions.R
\name{ild_missing_hazard_first}
\alias{ild_missing_hazard_first}
\title{Discrete hazard of first missing outcome on an ordinal schedule}
\usage{
ild_missing_hazard_first(x, outcome)
}
\arguments{
\item{x}{An ILD object.}

\item{outcome}{Character; outcome column.}
}
\value{
A tibble with `.ild_seq`, `n_at_risk`, `n_missing`, `hazard`.
}
\description{
For each `.ild_seq`, estimates the fraction of person-occasions **at risk**
(previous occasion observed, or first occasion) that are **missing** on the
outcome. This is a coarse discrete-time hazard for **first** missing spell
when missingness is intermittent; under **monotone dropout** it matches the
usual discrete hazard of dropout.
}
\details{
**Assumptions:** Rows are ordered within person by `.ild_seq`. The first row
per person is always at risk. Later rows are at risk only if the previous
row (same person) was non-missing—so a person who is already missing is not
counted again. This targets a **first event** narrative; it is not a full
repeated-events model.
}
\seealso{
[ild_missing_cohort()], [ild_missing_compliance()]

Other ild_diagnostics_utilities: 
\code{\link{ild_design_check}()},
\code{\link{ild_ipcw_weights}()},
\code{\link{ild_iptw_msm_weights}()},
\code{\link{ild_iptw_weights}()},
\code{\link{ild_ipw_weights}()},
\code{\link{ild_joint_msm_weights}()},
\code{\link{ild_missing_cohort}()},
\code{\link{ild_missing_compliance}()},
\code{\link{ild_missing_model}()},
\code{\link{ild_missing_pattern}()},
\code{\link{ild_missingness_report}()}
}
\concept{ild_diagnostics_utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{ema_example}
\alias{ema_example}
\title{Example EMA-style intensive longitudinal dataset}
\format{
A data frame with 140 rows and 4 columns:
\describe{
  \item{id}{Person identifier (1--10).}
  \item{time}{POSIXct timestamp (irregular within person).}
  \item{mood}{Simulated mood score.}
  \item{stress}{Simulated stress score.}
}
}
\source{
Simulated with a fixed seed (12345) for reproducibility.
}
\description{
A small simulated dataset with 10 persons and 14 observations per person,
irregular timing, and two variables (mood, stress). For use in examples and
vignettes. Use [ild_prepare()] to convert to an ILD object.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_missing_bias.R
\name{ild_missing_bias}
\alias{ild_missing_bias}
\title{Test whether missingness is associated with a predictor (informative missingness)}
\usage{
ild_missing_bias(x, outcome_var, predictor_var, random = FALSE)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{outcome_var}{Character. Name of the variable with missingness (e.g. \code{"mood"}).}

\item{predictor_var}{Character. Name of the suspected predictor of missingness (e.g. \code{"stress"}).}

\item{random}{Logical. If \code{TRUE}, fit a mixed-effects logistic model
\code{is_missing ~ predictor + (1 | id)} via \code{lme4::glmer}; if \code{FALSE}
(default), fit \code{glm(is_missing ~ predictor, family = binomial)}.}
}
\value{
A list with \code{predictor} (name), \code{estimate}, \code{std_error},
  \code{p_value}, and \code{message} (short note about informative missingness).
}
\description{
Fits a logistic model of missingness (binary: is the outcome NA?) on a
predictor variable. Use as a diagnostic: if the predictor is significant,
missingness may be informative and results could be biased. This function
does not correct for missingness; it flags the assumption for sensitivity analyses.
}
\examples{
set.seed(1)
d <- ild_simulate(n_id = 20, n_obs_per = 10, seed = 1)
d$stress <- rnorm(nrow(d))
d$mood <- d$y
d$mood[sample(nrow(d), 30)] <- NA  # some missing
x <- ild_prepare(d, id = "id", time = "time")
ild_missing_bias(x, "mood", "stress")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_fit.R
\name{ild_msm_fit}
\alias{ild_msm_fit}
\title{Fit an MSM-style weighted mixed model from an estimand spec}
\usage{
ild_msm_fit(
  estimand,
  data,
  outcome_formula,
  history,
  history_spec = NULL,
  treatment_engine = c("sequential_msm", "pooled"),
  predictors_censor = NULL,
  weights_col = ".ipw",
  stabilize_treat = c("marginal", "prior_treatment", "none"),
  prior_treatment = NULL,
  stabilize_joint = c("mean1", "none"),
  trim = c(0.01, 0.99),
  inference = c("none", "robust", "bootstrap"),
  robust_type = c("CR2", "CR3", "CR0"),
  n_boot = 200L,
  weight_policy = c("fixed_weights", "reestimate_weights"),
  weights_fn = NULL,
  seed = NULL,
  strict_inference = FALSE,
  ...
)
}
\arguments{
\item{estimand}{Object from [ild_msm_estimand()].}

\item{data}{ILD data.}

\item{outcome_formula}{Outcome model formula for [ild_lme()] (typically includes treatment term).}

\item{history}{Predictors for treatment process weights:
one-sided formula (e.g. \code{~ stress_lag1 + trt_lag1}) or character vector.}

\item{history_spec}{Optional [ild_msm_history_spec()] to build lagged history before weighting.}

\item{treatment_engine}{\code{"sequential_msm"} (default) or \code{"pooled"}.}

\item{predictors_censor}{Optional character vector for IPCW model. If supplied, joint weights
are built with [ild_joint_msm_weights()] and stored in \code{weights_col}.}

\item{weights_col}{Column used by [ild_ipw_refit()] (default \code{".ipw"}).}

\item{stabilize_treat}{Treatment-weight stabilization mode. For sequential MSM:
\code{"marginal"}, \code{"prior_treatment"}, \code{"none"}. For pooled IPTW:
mapped to logical \code{stabilize}.}

\item{prior_treatment}{Required when \code{stabilize_treat = "prior_treatment"} in sequential MSM.}

\item{stabilize_joint}{Joint-weight stabilization for [ild_joint_msm_weights()].}

\item{trim}{Quantile trimming applied during weight construction.}

\item{inference}{\code{"none"} (default), \code{"robust"}, or \code{"bootstrap"}.}

\item{robust_type}{Passed to [ild_robust_se()] and [tidy_ild_model()] when \code{inference = "robust"}.}

\item{n_boot}{Passed to [ild_msm_bootstrap()] when \code{inference = "bootstrap"}.}

\item{weight_policy}{Passed to [ild_msm_bootstrap()] when \code{inference = "bootstrap"}.}

\item{weights_fn}{Passed to [ild_msm_bootstrap()] when \code{inference = "bootstrap"} and
\code{weight_policy = "reestimate_weights"}.}

\item{seed}{Optional seed for bootstrap inference.}

\item{strict_inference}{Logical. If \code{TRUE}, stop when requested inference or
regime dispatch cannot be delivered as requested. If \code{FALSE} (default),
degrade with warning and record machine-readable status.}

\item{...}{Passed to [ild_lme()].}
}
\value{
Object of class \code{ild_msm_fit} with elements:
  \code{estimand}, \code{history_spec}, \code{weights_data}, \code{fit}, \code{inference},
  \code{treatment_engine}, \code{weights_col}. Analysis provenance is attached.
}
\description{
High-level runner that takes an [ild_msm_estimand()] plus ILD data, builds
treatment (and optional censoring) weights, refits a weighted \code{lmer},
and optionally computes inference via robust SEs or cluster bootstrap.
}
\details{
v1.1 keeps the v1 workflow and adds explicit capability signaling for
inference/regime paths that are degraded or unsupported.
}
\examples{
set.seed(31)
d <- ild_simulate(n_id = 12, n_obs_per = 6, seed = 31)
d$stress <- rnorm(nrow(d))
d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
d <- ild_prepare(d, id = "id", time = "time")
d <- ild_center(d, y)
d <- ild_lag(d, stress, max_gap = Inf)
d <- ild_lag(d, trt, max_gap = Inf)
est <- ild_msm_estimand(treatment = "trt")
res <- ild_msm_fit(
  estimand = est,
  data = d,
  outcome_formula = y ~ y_bp + y_wp + stress + trt + (1 | id),
  history = ~ stress_lag1 + trt_lag1,
  predictors_censor = "stress",
  inference = "none",
  warn_no_ar1 = FALSE,
  warn_uncentered = FALSE
)
res$fit
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_robust_se.R
\name{ild_robust_se}
\alias{ild_robust_se}
\title{Cluster-robust variance-covariance matrix for ILD model fits}
\usage{
ild_robust_se(
  fit,
  type = c("CR2", "CR3", "CR0"),
  cluster = c("id", "data"),
  cluster_vec = NULL,
  ...
)
}
\arguments{
\item{fit}{A fitted model from [ild_lme()] (lmerMod or lme).}

\item{type}{Character. Correction type: \code{"CR2"} (recommended), \code{"CR3"}, or \code{"CR0"}.}

\item{cluster}{Either \code{"id"} (default; cluster by the ILD id column from
\code{ild_data}) or \code{"data"} to use a user-supplied vector via \code{cluster_vec}.}

\item{cluster_vec}{When \code{cluster = "data"}, a vector of cluster IDs aligned to
the model rows (same length and order as \code{attr(fit, "ild_data")}).}

\item{...}{Passed to \code{clubSandwich::vcovCR()}.}
}
\value{
A list with \code{vcov} (matrix), \code{type}, \code{cluster_name},
  \code{engine} (\code{"lmer"} or \code{"lme"}), and optionally \code{message}
  if a fallback was used (e.g. lme not fully supported on this build).
}
\description{
Computes cluster-robust (sandwich) variance estimators with small-sample
corrections via the \pkg{clubSandwich} package. Use with [tidy_ild_model()]
via \code{se = "robust"} for fixed-effect inference. Requires
\code{attr(fit, "ild_data")}; refit with [ild_lme()] if missing.
}
\details{
For fits from [ild_ipw_refit()] with **estimated** IPW/IPCW weights, these SEs
apply to the **weighted outcome (mixed) model** only; they do **not** incorporate
first-stage uncertainty from the propensity/censoring models unless you treat
weights as fixed. For MSM-style inference, prefer [ild_msm_bootstrap()] with an
appropriate \code{weight_policy}; see [ild_msm_inference].
}
\examples{
if (requireNamespace("clubSandwich", quietly = TRUE)) {
  set.seed(1)
  dat <- ild_simulate(n_id = 8, n_obs_per = 6, seed = 1)
  dat <- ild_prepare(dat, id = "id", time = "time")
  dat <- ild_center(dat, y)
  fit <- ild_lme(y ~ y_bp + y_wp + (1 | id), data = dat, ar1 = FALSE, warn_no_ar1 = FALSE)
  rv <- ild_robust_se(fit, type = "CR2")
  rv$engine
  dim(rv$vcov)
}
}
\seealso{
[tidy_ild_model()] with \code{se = "robust"}, \code{clubSandwich::vcovCR},
  [ild_msm_bootstrap()], [ild_msm_inference].
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/broom.R
\name{broom_ild_lme}
\alias{broom_ild_lme}
\alias{tidy.ild_lme}
\alias{augment.ild_lme}
\title{Tidy and augment ild_lme fits with broom.mixed}
\usage{
tidy.ild_lme(x, ...)

augment.ild_lme(x, ...)
}
\arguments{
\item{x}{A fitted model from [ild_lme()].}

\item{...}{Passed to \code{broom.mixed::tidy()} or \code{broom.mixed::augment()}.}
}
\value{
Same as the corresponding broom.mixed method.
}
\description{
These S3 methods delegate to [broom.mixed::tidy()] and [broom.mixed::augment()]
on the underlying model object so that ild_lme fits work in tidy workflows.
Package \pkg{broom.mixed} must be attached (e.g. \code{library(broom.mixed)}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_lag.R
\name{ild_lag}
\alias{ild_lag}
\title{Spacing-aware lag within person}
\usage{
ild_lag(
  x,
  ...,
  n = 1L,
  mode = c("index", "gap_aware", "time_window"),
  max_gap = NULL,
  window = NULL,
  resolution = c("closest_prior", "last_in_window", "mean_in_window")
)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{...}{Variables to lag (tidy-select). Unquoted names or selection.}

\item{n}{Integer. Lag order (default 1 = previous observation).}

\item{mode}{Character. \code{"index"}: row-based lag. \code{"gap_aware"}: same but NA when interval exceeds \code{max_gap}. \code{"time_window"}: value from (time - window, time] with \code{resolution}.}

\item{max_gap}{Numeric. For \code{gap_aware} only. Same units as \code{.ild_time_num}.
If \code{NULL}, uses \code{ild_meta(x)$ild_gap_threshold} (metadata-driven default).}

\item{window}{Numeric or lubridate duration. For \code{time_window} only: time window width.
Numeric is in same units as \code{.ild_time_num} (e.g. seconds for POSIXct). You can pass
a lubridate period/duration (e.g. \code{lubridate::hours(2)}); it is converted to seconds for POSIXct.}

\item{resolution}{Character. For \code{time_window}: \code{"closest_prior"} (default: most recent observation in window),
\code{"last_in_window"}, or \code{"mean_in_window"}.}
}
\value{
The same ILD tibble with new lag columns. ILD attributes preserved.
}
\description{
Computes lagged values within each person. Use this instead of
[dplyr::lag()], which assumes equal spacing and no gaps and is unsafe
for irregular ILD.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_bootstrap.R
\name{ild_msm_inference}
\alias{ild_msm_inference}
\title{MSM and IPW inference for weighted mixed models}
\description{
After [ild_ipw_refit()] (or any \code{lmer(..., weights =)} with \code{attr(fit, "ild_data")}),
uncertainty for MSM-style **estimands** should not rely on default \code{lmer} standard errors.
tidyILD supports:
\itemize{
  \item \strong{Cluster bootstrap} — [ild_msm_bootstrap()] resamples persons (clusters),
    refits the weighted model, and collects fixed effects; use
    \code{weight_policy = "reestimate_weights"} and a \code{weights_fn} to re-fit propensity
    weights on each replicate when that matches your estimand (computationally heavier;
    preferred for many MSM workflows). \code{weight_policy = "fixed_weights"} keeps the
    weights from the resampled rows (fast approximation; ignores first-stage uncertainty).
  \item \strong{Cluster-robust sandwich Wald} — [ild_robust_se()] / [tidy_ild_model(..., se = "robust")]
    apply \pkg{clubSandwich} at the **person** level on the **weighted outcome model only**; they
    do **not** account for estimation error in IPW/IPCW weights unless you treat weights as fixed.
  \item \strong{Bayesian} — Full joint Bayesian MSM is not built in. See [ild_brms()] for
    hierarchical outcome models; combining external IPW row weights with Stan is delicate;
    prefer explicit joint/sensitivity designs and specialized causal packages for Bayes MSM.
}
}
\seealso{
[ild_msm_bootstrap()], [tidy_ild_msm_bootstrap()], [ild_ipw_refit()],
  [ild_robust_se()], [tidy_ild_model()], [ild_brms()], [ild_msm_estimand()],
  [ild_msm_fit()]
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_ctsem.R
\name{ild_ctsem}
\alias{ild_ctsem}
\title{Fit a continuous-time state-space model via ctsem}
\usage{
ild_ctsem(
  data,
  outcome,
  ct_model = NULL,
  model_type = c("stanct", "ctfit"),
  id_col = ".ild_id",
  time_col = ".ild_time_num",
  time_scale = 1,
  ...
)
}
\arguments{
\item{data}{An object that passes [validate_ild()] (typically after [ild_prepare()]).}

\item{outcome}{Character name of numeric outcome column.}

\item{ct_model}{Optional ctsem model object. If \code{NULL}, a conservative
default model is created: for \code{model_type = "stanct"}, \code{ctsem::ctModel(type = "ct", ...)}
with one manifest loading on one latent; for \code{"ctfit"}, \code{type = "omx"} with
\code{Tpoints = max(rows per id, 2)} from the prepared long data (custom \code{ct_model}
may still be needed for irregular designs).}

\item{model_type}{Character. \code{"stanct"} (default) or \code{"ctfit"}.}

\item{id_col}{Person-id column in \code{data} (default \code{".ild_id"}).}

\item{time_col}{Continuous-time index column in \code{data}
(default \code{".ild_time_num"}).}

\item{time_scale}{Optional positive scalar to rescale \code{time_col}
before fitting (default \code{1}).}

\item{...}{Passed to \code{ctsem::ctStanFit()} or \code{ctsem::ctFit()}.}
}
\value{
An object of class \code{c("ild_fit_ctsem", "ild_fit_model", "ild_analysis")}.
  Attributes \code{ild_data} and \code{ild_provenance} are attached.
}
\description{
First-class tidyILD wrapper for \pkg{ctsem}. This backend is designed for
irregularly spaced ILD where continuous-time dynamics are preferable to
discrete-time approximations.
}
\details{
v1 focuses on a conservative single-outcome pathway with explicit class,
provenance, and diagnostics compatibility. Users may supply a custom ctsem
model object; otherwise a simple 1-manifest local-level style template is used.
}
\seealso{
[ild_tidy()], [ild_augment()], [ild_diagnose()], [ild_autoplot()].
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_balance.R
\name{ild_ipw_ess}
\alias{ild_ipw_ess}
\title{Effective sample size from weights}
\usage{
ild_ipw_ess(
  data,
  weights_col = ".ipw",
  by_occasion = FALSE,
  time_var = ".ild_seq"
)
}
\arguments{
\item{data}{An ILD object.}

\item{weights_col}{Name of the nonnegative weight column (e.g. \code{".ipw"} or \code{".ipw_treat"}).}

\item{by_occasion}{If \code{TRUE}, one ESS per level of \code{time_var}.}

\item{time_var}{Column for occasion index (default \code{".ild_seq"}).}
}
\value{
A single numeric (pooled) or a \code{tibble} with \code{occasion} and \code{ess} when
  \code{by_occasion = TRUE}.
}
\description{
For nonnegative weights \eqn{w}, \eqn{\mathrm{ESS} = (\sum w)^2 / \sum w^2}.
}
\examples{
set.seed(1)
d <- ild_simulate(n_id = 10, n_obs_per = 5, seed = 1)
x <- ild_prepare(d, id = "id", time = "time")
x$.ipw <- runif(nrow(x), 0.5, 1.5)
ild_ipw_ess(x, ".ipw")
ild_ipw_ess(x, ".ipw", by_occasion = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_generics.R
\name{ild_fit}
\alias{ild_fit}
\title{Fit a mixed model to ILD data (unified entry point)}
\usage{
ild_fit(
  formula,
  data,
  backend = c("lme4", "nlme", "brms"),
  correlation_class = c("auto", "AR1", "CAR1"),
  random = ~1 | .ild_id,
  warn_no_ar1 = TRUE,
  warn_uncentered = TRUE,
  prior = NULL,
  prior_template = c("default", "weakly_informative", "minimal_shrinkage"),
  ...
)
}
\arguments{
\item{formula}{Passed to [ild_lme()] or [ild_brms()].}

\item{data}{An ILD object (see [is_ild()]).}

\item{backend}{Character. \code{"lme4"} (default, [lme4::lmer()]), \code{"nlme"}
([nlme::lme()] with AR1/CAR1 residual structure), or \code{"brms"} ([brms::brm()]).}

\item{correlation_class, random, warn_no_ar1}{Passed to [ild_lme()] when
\code{backend} is \code{"lme4"} or \code{"nlme"}. Ignored when \code{backend = "brms"},
except \code{correlation_class} must be \code{"auto"} for \code{"brms"} (non-\code{"auto"}
values are for residual correlation under \code{nlme} only and will error).}

\item{warn_uncentered}{Passed to [ild_lme()] or [ild_brms()].}

\item{prior, prior_template}{Passed to [ild_brms()] when \code{backend = "brms"}.
For \code{"lme4"} / \code{"nlme"}, \code{prior} must be \code{NULL} and
\code{prior_template} must remain the default (\code{"default"}); otherwise an error
is raised so mistaken cross-backend arguments are not silently ignored.}

\item{...}{Passed to [ild_lme()] / [lme4::lmer()] / [nlme::lme()] or [ild_brms()].}
}
\value{
A fitted model: \code{lmerMod} / \code{lme} from [ild_lme()], or \code{brmsfit}
  from [ild_brms()], each with the same attributes as those functions document.
}
\description{
Façade that selects the estimation backend explicitly. Equivalent to calling
[ild_lme()] for \code{"lme4"} / \code{"nlme"}, or [ild_brms()] for \code{"brms"}.
Named functions remain the primary APIs for full documentation and examples.
}
\details{
**State-space / latent-dynamics models** are not mixed-model formulas:
use [ild_kfas()] or [ild_ctsem()] directly;
see \code{vignette("kfas-choosing-backend", package = "tidyILD")}.

To pass \code{backend} to \code{brms::brm()} (e.g. \code{"rstan"} vs \code{"cmdstanr"}), use
[ild_brms()] directly: \code{ild_fit()} reserves \code{backend} for the tidyILD engine
(\code{"lme4"}, \code{"nlme"}, \code{"brms"}).
}
\seealso{
[ild_kfas()] and [ild_ctsem()] for state-space / latent-dynamics
  backends (not available via \code{ild_fit()}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_generics.R, R/ild_brms_s3.R,
%   R/ild_ctsem_s3.R, R/ild_kfas_autoplot.R, R/ild_autoplot_bundle.R
\name{ild_autoplot}
\alias{ild_autoplot}
\alias{ild_autoplot.ild_diagnostics}
\alias{ild_autoplot.lmerMod}
\alias{ild_autoplot.lme}
\alias{ild_autoplot.default}
\alias{ild_autoplot.brmsfit}
\alias{ild_autoplot.ild_fit_ctsem}
\alias{ild_autoplot.ild_fit_kfas}
\alias{ild_autoplot.ild_diagnostics_bundle}
\title{Autoplot for ILD model fits or diagnostics (S3 generic)}
\usage{
ild_autoplot(x, ...)

\method{ild_autoplot}{ild_diagnostics}(x, ...)

\method{ild_autoplot}{lmerMod}(x, type = "fitted_vs_actual", ...)

\method{ild_autoplot}{lme}(x, type = "fitted_vs_actual", ...)

\method{ild_autoplot}{default}(x, ...)

\method{ild_autoplot}{brmsfit}(x, type = "pp_check", ...)

\method{ild_autoplot}{ild_fit_ctsem}(x, type = c("fitted_vs_actual", "residual_time", "qq"), ...)

\method{ild_autoplot}{ild_fit_kfas}(x, type = c("states", "innovations", "signal"), ...)

\method{ild_autoplot}{ild_diagnostics_bundle}(x, section = "residual", type = NULL, ...)
}
\arguments{
\item{x}{A bundle, \code{ild_diagnostics} object, or a fitted model with \code{ild_data}.}

\item{...}{Passed to [plot_ild_diagnostics()] (\code{ild_diagnostics} objects, or bundle
\code{section = "residual"} with \code{type = NULL} and legacy diagnostics), to
[ild_plot()] (\code{lmerMod}/\code{lme}), or to \code{brms::pp_check()} (bundle
\code{section = "predictive"}, \code{type = "ppc"}).}

\item{type}{Character: \code{"states"} (smoothed level vs time), \code{"innovations"}
(standardized), or \code{"signal"} (observation expectation vs outcome).}

\item{section}{For \code{\link{ild_diagnostics_bundle}} only: \code{"residual"},
\code{"fit"}, \code{"predictive"}, \code{"data"}, \code{"design"}, or \code{"causal"}.}
}
\description{
For \code{ild_diagnostics_bundle} from [ild_diagnose()], dispatches by \code{section}
and \code{type} (section-first; see \code{?ild_autoplot} method for
\code{ild_diagnostics_bundle}). For [ild_diagnostics()], calls [plot_ild_diagnostics()]
directly. For \code{lmerMod} or \code{lme} fits from [ild_lme()], calls [ild_plot()]
(default \code{type = "fitted_vs_actual"}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_missing_extensions.R
\name{ild_missing_cohort}
\alias{ild_missing_cohort}
\title{Cohort-level fraction observed by occasion index}
\usage{
ild_missing_cohort(x, outcome, plot = TRUE)
}
\arguments{
\item{x}{An ILD object.}

\item{outcome}{Character; outcome column.}

\item{plot}{Logical; if `TRUE`, return a `ggplot2` object in `$plot`.}
}
\value{
A list with `by_occasion` (tibble: `.ild_seq`, `n_rows`, `n_obs`,
  `pct_observed`) and if `plot = TRUE`, `plot`.
}
\description{
For each distinct `.ild_seq` value, computes how many rows exist and what
fraction has a non-missing outcome. Optional line plot.
}
\details{
Uses **within-study sequence** (`.ild_seq`), not calendar time. For irregular
timing, interpret as ordinal wave position rather than equal time spacing.
}
\seealso{
[ild_missing_pattern()], [ild_missing_compliance()]

Other ild_diagnostics_utilities: 
\code{\link{ild_design_check}()},
\code{\link{ild_ipcw_weights}()},
\code{\link{ild_iptw_msm_weights}()},
\code{\link{ild_iptw_weights}()},
\code{\link{ild_ipw_weights}()},
\code{\link{ild_joint_msm_weights}()},
\code{\link{ild_missing_compliance}()},
\code{\link{ild_missing_hazard_first}()},
\code{\link{ild_missing_model}()},
\code{\link{ild_missing_pattern}()},
\code{\link{ild_missingness_report}()}
}
\concept{ild_diagnostics_utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_crosslag.R
\name{ild_crosslag}
\alias{ild_crosslag}
\title{Cross-lag model: lag predictor then fit outcome ~ lag}
\usage{
ild_crosslag(
  data,
  outcome,
  predictor,
  lag = 1L,
  mode = c("gap_aware", "index", "time_window"),
  max_gap = NULL,
  ar1 = FALSE,
  include_diagnostics = FALSE,
  ...
)
}
\arguments{
\item{data}{An ILD object (see [is_ild()]).}

\item{outcome}{Character or symbol. Name of the outcome variable (e.g. \code{"mood"}).}

\item{predictor}{Character or symbol. Name of the predictor to lag (e.g. \code{"stress"}).}

\item{lag}{Integer. Lag order (default 1).}

\item{mode}{Character. Passed to [ild_lag()]: \code{"index"}, \code{"gap_aware"} (default), or \code{"time_window"}.}

\item{max_gap}{Numeric or NULL. Passed to [ild_lag()] when \code{mode = "gap_aware"}.}

\item{ar1}{Logical. If \code{TRUE}, fit with [ild_lme()] using AR1/CAR1 (default \code{FALSE}).}

\item{include_diagnostics}{Logical. If \code{TRUE}, attach [ild_diagnostics()] to the return (default \code{FALSE}).}

\item{...}{Passed to [ild_lme()].}
}
\value{
A list: \code{fit} (fitted model), \code{lag_term} (one-row tibble from [tidy_ild_model()] for the lag variable),
  \code{lag_check} (tibble from [ild_check_lags()]), \code{data} (ILD with lag column), and optionally \code{diagnostics}.
}
\description{
One-call pipeline: [ild_lag()] the predictor, [ild_check_lags()] to audit,
then [ild_lme()] to fit outcome on the lagged predictor. Returns the fit,
the lag-term coefficient (estimate, CI, p), and lag validity check.
}
\examples{
d <- ild_simulate(n_id = 10, n_obs_per = 8, seed = 1)
x <- ild_prepare(d, id = "id", time = "time")
out <- ild_crosslag(x, "y", "y", lag = 1, ar1 = FALSE, warn_no_ar1 = FALSE)
out$lag_term
out$lag_check
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_simulation.R
\name{ild_msm_recovery}
\alias{ild_msm_recovery}
\title{MSM recovery simulation harness}
\usage{
ild_msm_recovery(
  n_sim = 100L,
  n_id = 80L,
  n_obs_per = 10L,
  true_ate = 0.5,
  n_boot = 200L,
  inference = c("bootstrap", "robust", "none"),
  seed = 1001L,
  censoring = TRUE,
  scenario_grid = NULL
)
}
\arguments{
\item{n_sim}{Number of simulations.}

\item{n_id}{Number of persons per simulation.}

\item{n_obs_per}{Planned observations per person.}

\item{true_ate}{True additive treatment effect.}

\item{n_boot}{Bootstrap replicates when \code{inference = "bootstrap"}.}

\item{inference}{\code{"bootstrap"} (default), \code{"robust"}, or \code{"none"}.}

\item{seed}{Base seed.}

\item{censoring}{If \code{TRUE}, include monotone dropout in DGP and fit IPCW+joint pipeline.}

\item{scenario_grid}{Optional data frame/list of scenario settings. Supported columns:
\code{scenario_id, n_id, n_obs_per, true_ate, censoring, positivity_stress,
treatment_strength, censoring_strength, misspec_treatment_model}.}
}
\value{
List with \code{summary}, \code{summary_by_scenario}, and \code{sim_results}.
}
\description{
Repeats an MSM analysis over simulated datasets with known \code{true_ate} and
summarizes estimation bias, RMSE, CI coverage, and positivity stress summaries.
}
\examples{
\dontrun{
out <- ild_msm_recovery(n_sim = 10, n_id = 40, n_obs_per = 10, true_ate = 0.5, n_boot = 50)
out$summary
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_generics.R, R/ild_brms_s3.R,
%   R/ild_ctsem_s3.R, R/ild_kfas_augment.R
\name{ild_augment}
\alias{ild_augment}
\alias{ild_augment.lmerMod}
\alias{ild_augment.lme}
\alias{ild_augment.default}
\alias{ild_augment.brmsfit}
\alias{ild_augment.ild_fit_ctsem}
\alias{ild_augment.ild_fit_kfas}
\title{Augment observations with fitted values and residuals (S3 generic)}
\usage{
ild_augment(x, ...)

\method{ild_augment}{lmerMod}(x, ...)

\method{ild_augment}{lme}(x, ...)

\method{ild_augment}{default}(x, ...)

\method{ild_augment}{brmsfit}(x, summary = TRUE, ...)

\method{ild_augment}{ild_fit_ctsem}(x, ...)

\method{ild_augment}{ild_fit_kfas}(x, ...)
}
\arguments{
\item{x}{A fitted model from [ild_lme()].}

\item{...}{Passed to [augment_ild_model()].}

\item{summary}{Logical. For \code{brmsfit}, use posterior predictive summary
for fitted values (default \code{TRUE}).}
}
\description{
Dispatches to [augment_ild_model()] for \code{lmerMod} and \code{lme} objects from [ild_lme()],
or the \code{brmsfit} method for [ild_brms()]. All methods return tables conforming to
\code{\link{ild_augment_schema}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_diagnostics_utilities.R
\name{ild_diagnostics_utilities}
\alias{ild_diagnostics_utilities}
\title{Standalone diagnostics utilities and bundle section providers}
\description{
These functions are **standalone utilities** you can run on prepared ILD data at any time
(typically before or alongside modeling). The same logic also **feeds**
\code{\link{ild_diagnostics_bundle}} sections when you call \code{\link{ild_diagnose}} on a
fitted model, so the package presents one coherent diagnostics story: a top-level façade
(\code{ild_diagnose}) plus reusable building blocks below.
}
\section{Relationship to \code{\link{ild_diagnose}}}{

\describe{
  \item{\code{\link{ild_design_check}}}{Populates the bundle \strong{design} slot (embedded as
    \code{design$ild_design_check}) and contributes WP/BP and design-time missingness views.}
  \item{\code{\link{ild_missing_pattern}}}{Drives \strong{data} (\code{data$missing_pattern},
    global columns) and \strong{missingness} (variable-specific summaries when model predictors
    are known).}
  \item{\code{\link{ild_missing_model}}}{Not called automatically by \code{ild_diagnose}.
    Use when you fit a missingness model for IPW or interpretation; it is the usual
    prerequisite for \code{\link{ild_ipw_weights}}.}
  \item{\code{\link{ild_ipw_weights}}}{Adds \code{.ipw} (and related steps in the IPW workflow).
    When those columns are present on the ILD data attached to the fit, \code{ild_diagnose}
    fills the bundle \strong{causal} slot with weight summaries.}
  \item{\code{\link{ild_iptw_weights}}, \code{\link{ild_iptw_msm_weights}}, \code{\link{ild_ipcw_weights}}, \code{\link{ild_joint_msm_weights}}}{MSM-style
    inverse-probability weights for \strong{treatment} (pooled or sequential \eqn{A_t}) and \strong{monotone censoring}, combined into \code{.ipw}.
    Distinct from outcome missingness IPW above.}
  \item{\code{\link{ild_msm_bootstrap}}, \code{\link{tidy_ild_msm_bootstrap}}}{Person-level bootstrap for weighted \code{lmer} after \code{\link{ild_ipw_refit}}; see \code{\link{ild_msm_inference}}.}
  \item{\code{\link{ild_msm_balance}}, \code{\link{ild_ipw_ess}}, \code{\link{ild_msm_overlap_plot}}}{Weighted covariate balance (SMD), effective sample size, and propensity overlap plots for MSM / IPW workflows.}
  \item{\code{\link{ild_msm_estimand}}, \code{\link{ild_msm_fit}}, \code{\link{ild_msm_diagnose}}, \code{\link{ild_msm_contrast_over_time}}, \code{\link{ild_msm_history_spec}}, \code{\link{ild_build_msm_history}}, \code{\link{ild_msm_recovery}}}{Estimand-first MSM workflow, one-call diagnostics bridge, time-indexed contrasts, history construction, and simulation-based recovery checks.}
}
}

\seealso{
\code{\link{ild_diagnose}}, \code{\link{ild_diagnostics_bundle}}, \code{\link{ild_diagnostics}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_joint_msm_weights.R
\name{ild_joint_msm_weights}
\alias{ild_joint_msm_weights}
\title{Joint MSM weights from IPTW and IPCW components}
\usage{
ild_joint_msm_weights(
  x,
  stabilize = c("mean1", "none"),
  trim = c(0.01, 0.99),
  joint_name = ".ipw"
)
}
\arguments{
\item{x}{An ILD object containing \code{.ipw_treat} and \code{.ipw_censor}
(from [ild_iptw_weights()] or [ild_iptw_msm_weights()], and [ild_ipcw_weights()]).}

\item{stabilize}{Character. \code{"none"}: \code{w_treat * w_censor} only.
\code{"mean1"} (default): divide joint weights by their mean so the mean is 1
(helps comparability with stabilized components).}

\item{trim}{Numeric of length 2. Quantiles for trimming the **joint** vector
(default \code{c(0.01, 0.99)}).}

\item{joint_name}{Character. Name of the combined weight column (default \code{".ipw"}).}
}
\value{
\code{x} with \code{joint_name} set; component columns unchanged.
}
\description{
Multiplies \code{.ipw_treat} and \code{.ipw_censor} into a single analysis
weight column (default \code{.ipw}) for use with [ild_ipw_refit()] and
[ild_diagnose()]. Optional scaling sets the mean joint weight to 1.
}
\section{Bundle integration}{

Sets \code{.ipw} (or \code{joint_name}) so existing causal diagnostics and
guardrails apply to the **joint** weight.
}

\examples{
set.seed(4)
d <- ild_simulate(n_id = 15, n_obs_per = 8, seed = 4)
d$stress <- rnorm(nrow(d))
d$trt <- as.integer(d$stress > 0)
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_iptw_weights(x, "trt", "stress")
x <- ild_ipcw_weights(x, "stress")
x <- ild_joint_msm_weights(x)
summary(x$.ipw)
}
\seealso{
[ild_iptw_weights()], [ild_iptw_msm_weights()], [ild_ipcw_weights()], [ild_ipw_refit()]

Other ild_diagnostics_utilities: 
\code{\link{ild_design_check}()},
\code{\link{ild_ipcw_weights}()},
\code{\link{ild_iptw_msm_weights}()},
\code{\link{ild_iptw_weights}()},
\code{\link{ild_ipw_weights}()},
\code{\link{ild_missing_cohort}()},
\code{\link{ild_missing_compliance}()},
\code{\link{ild_missing_hazard_first}()},
\code{\link{ild_missing_model}()},
\code{\link{ild_missing_pattern}()},
\code{\link{ild_missingness_report}()}
}
\concept{ild_diagnostics_utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_kfas_plot.R
\name{ild_plot_filtered_vs_smoothed}
\alias{ild_plot_filtered_vs_smoothed}
\title{Plot filtered vs smoothed state (first state)}
\usage{
ild_plot_filtered_vs_smoothed(x)
}
\arguments{
\item{x}{An object from [ild_kfas()].}
}
\value{
A \code{ggplot} object.
}
\description{
Compares one-step filtered state (\code{att}) with smoothed (\code{alphahat}) when available.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_schema_tidy_augment.R
\name{ild_augment_schema}
\alias{ild_augment_schema}
\title{Column dictionary for \code{ild_augment()} outputs}
\usage{
ild_augment_schema()
}
\value{
A list with \code{required} and \code{optional} character vectors.
}
\description{
Defines the minimum stable contract for augmented observation-level tables.
\code{\link{augment_ild_model}} and \code{\link{ild_augment}} methods emit all required
columns; see \code{\link{augment_ild_model}} for \code{.resid_std} semantics.
}
\details{
**Required columns**
\tabular{ll}{
  \code{.ild_id} \tab Person identifier (from ILD) \cr
  \code{.ild_time} \tab Time variable \cr
  \code{.outcome} \tab Observed outcome (numeric vector aligned to rows) \cr
  \code{.fitted} \tab Fitted / posterior mean fitted value \cr
  \code{.resid} \tab Raw residual \cr
  \code{.resid_std} \tab Pearson-type residual when \code{residuals(fit, type = "pearson")} is available and length-matched; otherwise \code{NA} (not ad hoc z-scores) \cr
  \code{engine} \tab Character engine id \cr
  \code{model_class} \tab Model class string \cr
}

**Optional:** \code{.fitted_lower}, \code{.fitted_upper}, \code{.influence},
\code{.state}, \code{.state_lower}, \code{.state_upper}.
}
\seealso{
\code{\link{ild_tidy_schema}}, \code{\link{ild_diagnostics_bundle}}.
  Reserved prefixes and migration notes: \code{vignette("developer-contracts", package = "tidyILD")}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_plot.R
\name{ild_heatmap}
\alias{ild_heatmap}
\title{ILD heatmap (alias for ild_plot with type = "heatmap")}
\usage{
ild_heatmap(x, var = NULL, facet_by = NULL, ...)
}
\arguments{
\item{x}{ILD object or fitted model (for heatmap, data are taken from ild_data if model).}

\item{var}{Variable to plot. If NULL, single data column is used.}

\item{facet_by}{Optional column in ILD for [ggplot2::facet_wrap()].}

\item{...}{Passed to [ild_plot()] (e.g. id_var, time_var).}
}
\value{
A ggplot object.
}
\description{
Person x time heatmap of a variable. See [ild_plot()].
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_guardrail_registry.R
\name{guardrail_registry}
\alias{guardrail_registry}
\title{Guardrail rule registry (analysis safety layer)}
\usage{
guardrail_registry()
}
\value{
A tibble with columns \code{rule_id}, \code{section}, \code{severity},
  \code{default_message}, \code{default_recommendation}.
}
\description{
Returns the canonical catalog of \strong{guardrail} rules tidyILD may use. Each rule has a
stable \code{rule_id}, \code{section} (bundle slot), \code{severity}, and default
\code{default_message} / \code{default_recommendation} text. When you call
\code{\link{ild_diagnose}}, \strong{triggered} rules appear as rows in
\code{\link{ild_diagnostics_bundle}$guardrails} with \code{message} and
\code{recommendation} possibly customized for the run.
}
\seealso{
\code{\link{ild_diagnose}}, \code{\link{ild_diagnostics_bundle}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_tvem_plot.R
\name{ild_tvem_plot}
\alias{ild_tvem_plot}
\title{Plot the time-varying coefficient from a TVEM fit}
\usage{
ild_tvem_plot(tvem_fit, n_grid = 100L, level = 0.95)
}
\arguments{
\item{tvem_fit}{A fitted object from [ild_tvem()] (class \code{tidyild_tvem}).}

\item{n_grid}{Integer. Number of points over the time range for the curve (default 100).}

\item{level}{Numeric. Confidence level for the band (default 0.95).}
}
\value{
A \code{ggplot} object (time on x-axis, estimated effect on y-axis).
}
\description{
Builds a line plot of the smooth term for the time-varying effect of the
predictor (with optional confidence band). Uses a grid over the time variable
and [mgcv::predict.gam()] with \code{type = "terms"}.
}
\examples{
set.seed(1)
d <- ild_simulate(n_id = 10, n_obs_per = 15, seed = 1)
d$x <- rnorm(nrow(d))
x <- ild_prepare(d, id = "id", time = "time")
tv <- ild_tvem(x, "y", "x", k = 5, re_id = TRUE)
ild_tvem_plot(tv)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_decomposition.R
\name{ild_center_plot}
\alias{ild_center_plot}
\title{Standalone WP/BP centering plot}
\usage{
ild_center_plot(x, ...)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{...}{Variables to plot (tidy-select). Must be numeric. Only the first is plotted.}
}
\value{
A ggplot object (WP vs BP density overlay for the first selected variable).
}
\description{
Shows within-person deviation and between-person (person mean) distribution
for selected variable(s). Uses the same plot as [ild_decomposition(..., plot = TRUE)].
Useful when you only want the visualization without the variance table.
}
\examples{
d <- ild_simulate(n_id = 10, n_obs_per = 8, seed = 1)
x <- ild_prepare(d, id = "id", time = "time")
ild_center_plot(x, y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_iptw_msm_weights.R
\name{ild_iptw_msm_weights}
\alias{ild_iptw_msm_weights}
\title{Sequential MSM inverse probability of treatment weights (IPTW for A_t)}
\usage{
ild_iptw_msm_weights(
  x,
  treatment,
  history,
  time_var = ".ild_seq",
  stabilize = c("marginal", "prior_treatment", "none"),
  prior_treatment = NULL,
  trim = c(0.01, 0.99),
  ...
)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{treatment}{Character. Binary \eqn{A_t} column (\code{0}/\code{1} or two-level factor).}

\item{history}{Predictors \eqn{\bar{L}_t} for the **denominator** model: a
\code{character} vector of column names, or a one-sided \code{formula}
like \code{~ stress + trt_lag1}. Empty history fits an intercept-only model.}

\item{time_var}{Character. Occasion index column (default \code{".ild_seq"}).}

\item{stabilize}{Character: \code{"marginal"}, \code{"prior_treatment"}, or \code{"none"}.}

\item{prior_treatment}{Character. Required when \code{stabilize = "prior_treatment"}:
name of the \eqn{A_{t-1}} (or prior treatment) column for the numerator model.}

\item{trim}{Numeric of length 2. Quantiles applied to the **final** cumulative
\code{.ipw_treat} (default \code{c(0.01, 0.99)}). Use length-0 numeric to skip.}

\item{...}{Passed to \code{stats::glm()} for each occasion (denominator model;
numerator model when \code{prior_treatment} is used also uses \code{...}).}
}
\value{
\code{x} with \code{.ipw_treat} set to the cumulative sequential weight.
  Attributes: \code{ild_iptw_msm_fits} (named list of denominator \code{glm} fits
  by occasion), \code{ild_iptw_msm_numerator_fits} (same for numerator when
  \code{prior_treatment}).
}
\description{
For **time-varying binary treatment** \eqn{A_t}, fits a separate treatment
propensity model at each occasion \eqn{t} (pooling persons at that time):
\eqn{P(A_t \mid \bar{L}_t)} where \eqn{\bar{L}_t} is supplied by the user
(e.g. lags from [ild_lag()], prior treatment, time-varying confounders). The
stabilized factor at each person-occasion is multiplied **within person**
over time to yield a **cumulative** MSM IPTW in \code{.ipw_treat}.
}
\details{
This differs from [ild_iptw_weights()], which fits **one pooled** logistic
over all rows (appropriate for a single treatment decision or descriptive
pooling, not the standard sequential MSM estimand for \eqn{\bar{A}_K}).

**Assumptions:** Sequential exchangeability given \eqn{\bar{L}_t}, positivity
at each \eqn{t}, and correct models. Variance (e.g. bootstrap by person) is
not computed here.
}
\section{Stabilization}{

\describe{
  \item{\code{marginal}}{Numerator at each occasion uses the sample marginal
    \eqn{P(A_t = 1)} among rows at that time (same pattern as pooled stabilized
    IPTW).}
  \item{\code{prior_treatment}}{Numerator uses \eqn{P(A_t \mid A_{t-1})} from
    a logistic model \code{A_t ~ prior_treatment} fit at each occasion (pass
    \code{prior_treatment} column name). Denominator remains \eqn{P(A_t \mid \bar{L}_t)}.}
  \item{\code{none}}{Unstabilized \eqn{1/P(A_t \mid \bar{L}_t)} for the observed \eqn{A_t}.}
}
}

\section{Bundle integration}{

Sets \code{.ipw_treat} for use with [ild_ipcw_weights()] and
[ild_joint_msm_weights()].
}

\examples{
set.seed(11)
d <- ild_simulate(n_id = 20, n_obs_per = 6, seed = 11)
d$stress <- rnorm(nrow(d))
d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.4))
d <- ild_prepare(d, id = "id", time = "time")
d <- ild_lag(d, stress, mode = "gap_aware", max_gap = Inf)
d <- ild_lag(d, trt, mode = "gap_aware", max_gap = Inf)
x <- ild_iptw_msm_weights(d, treatment = "trt", history = ~ stress_lag1 + trt_lag1)
summary(x$.ipw_treat)
}
\seealso{
[ild_iptw_weights()], [ild_ipcw_weights()], [ild_joint_msm_weights()]

Other ild_diagnostics_utilities: 
\code{\link{ild_design_check}()},
\code{\link{ild_ipcw_weights}()},
\code{\link{ild_iptw_weights}()},
\code{\link{ild_ipw_weights}()},
\code{\link{ild_joint_msm_weights}()},
\code{\link{ild_missing_cohort}()},
\code{\link{ild_missing_compliance}()},
\code{\link{ild_missing_hazard_first}()},
\code{\link{ild_missing_model}()},
\code{\link{ild_missing_pattern}()},
\code{\link{ild_missingness_report}()}
}
\concept{ild_diagnostics_utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_ipcw_weights.R
\name{ild_ipcw_weights}
\alias{ild_ipcw_weights}
\title{Inverse probability of censoring weights (IPCW) for monotone dropout}
\usage{
ild_ipcw_weights(x, predictors, stabilize = TRUE, trim = c(0.01, 0.99), ...)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{predictors}{Character vector of covariate names (time-varying or baseline
carried forward in \code{x}).}

\item{stabilize}{Logical. If \code{TRUE} (default), multiply factors
\code{mean(1 - p_drop) / (1 - p_drop)} at each row before cumulating; if
\code{FALSE}, use \code{1 / (1 - p_drop)}.}

\item{trim}{Numeric of length 2. Quantiles applied to **final** cumulative
\code{.ipw_censor} (default \code{c(0.01, 0.99)}).}

\item{...}{Passed to \code{stats::glm()}.}
}
\value{
\code{x} with added column \code{.ipw_censor}. Attribute
  \code{ild_ipcw_fit} holds the fitted \code{glm}.
}
\description{
Builds **discrete-time** IPCW weights for **monotone loss-to-follow-up**:
each person contributes a sequence of observed visits; after the last visit
the person is censored. This differs from [ild_missing_model()] + [ild_ipw_weights()],
which model **sporadic item missingness** on an outcome.
}
\details{
Internally, each person-occasion row is labeled with an indicator
\code{drop_next}: whether the person drops before the next scheduled occasion
(including the last observed row). A pooled logistic regression models
\code{drop_next} given covariates; stabilized or unstabilized inverse
probabilities are accumulated within person (product over follow-up) to yield
\code{.ipw_censor}.

**Assumptions:** Monotone censoring, positivity, correct censoring model.
}
\section{Bundle integration}{

Adds \code{.ipw_censor}. Combine with [ild_iptw_weights()] or
[ild_iptw_msm_weights()] via [ild_joint_msm_weights()].
}

\examples{
set.seed(3)
d <- ild_simulate(n_id = 18, n_obs_per = 7, seed = 3)
d$stress <- rnorm(nrow(d))
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_ipcw_weights(x, predictors = "stress")
summary(x$.ipw_censor)
}
\seealso{
[ild_iptw_weights()], [ild_joint_msm_weights()]

Other ild_diagnostics_utilities: 
\code{\link{ild_design_check}()},
\code{\link{ild_iptw_msm_weights}()},
\code{\link{ild_iptw_weights}()},
\code{\link{ild_ipw_weights}()},
\code{\link{ild_joint_msm_weights}()},
\code{\link{ild_missing_cohort}()},
\code{\link{ild_missing_compliance}()},
\code{\link{ild_missing_hazard_first}()},
\code{\link{ild_missing_model}()},
\code{\link{ild_missing_pattern}()},
\code{\link{ild_missingness_report}()}
}
\concept{ild_diagnostics_utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_diagnostics.R
\name{ild_diagnostics}
\alias{ild_diagnostics}
\title{Residual diagnostics for an ILD model}
\usage{
ild_diagnostics(
  object,
  data = NULL,
  type = c("residual_acf", "residual_time", "qq"),
  by_id = TRUE,
  ...
)
}
\arguments{
\item{object}{A fitted model from [ild_lme()] (or an object with
`residuals()`, and optional `fitted()`; if not `ild_lme`, pass `data` with
`.ild_id` and `.ild_time_num` or `.ild_seq`).}

\item{data}{Optional. ILD data (required if `object` is not from [ild_lme()]).}

\item{type}{Character vector. Which diagnostics to compute: `"residual_acf"`,
`"residual_time"` (residuals vs time and vs fitted), `"qq"`. Default is all three.}

\item{by_id}{Logical. If `TRUE`, compute ACF within each person (default `TRUE`).}

\item{...}{Unused.}
}
\value{
A list of class `ild_diagnostics` with: `meta` (engine, ar1, id_col, time_col,
  n_obs, n_id, type, by_id), `data` (list with `residuals` = tibble of .ild_id, .ild_time,
  response (name from formula), .resid, .fitted; \code{data$residuals} always exists, \code{.resid} is always filled,
  \code{.fitted} is returned when it can be computed without refitting, otherwise NA),
  `stats` (list with `acf` = list(pooled = tibble, by_id = list) when requested,
  `ar1_param` = numeric or NULL for lme). Use [plot_ild_diagnostics()] for plots.
}
\description{
Computes residual ACF (by person and/or pooled), residual vs fitted,
residual vs time, and optional Q-Q. Use `type` to request only specific
diagnostics. For `ild_lme` models with `ar1 = TRUE`, the estimated AR/CAR
parameter is reported when `type` includes `"residual_acf"`.
}
\details{
The return value follows a stable schema: `meta` (engine, ar1, id/time columns,
n_obs, n_id), `data$residuals` (tibble with `.ild_id`, `.ild_time`, response, `.resid`, `.fitted`),
and `stats` (e.g. `acf`, `ar1_param`). Plots are not stored in the object; use
[plot_ild_diagnostics()] to generate them from a diagnostics object. The column
\code{.resid} is always filled; \code{.fitted} is filled when it can be computed
without refitting, otherwise it is \code{NA} (same for both engines and all \code{type} values).

Residual ACF is computed over the ordered observation sequence within person;
it does not adjust for irregular time gaps.
}
\examples{
x <- ild_prepare(ild_simulate(n_id = 3, n_obs_per = 6, seed = 1), id = "id", time = "time")
fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
diag <- ild_diagnostics(fit, type = c("residual_acf", "qq"))
plot_ild_diagnostics(diag)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_summary.R
\name{ild_summary}
\alias{ild_summary}
\title{One-shot summary of an ILD object}
\usage{
ild_summary(x)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}
}
\value{
A list with elements: \code{summary} (one-row tibble with n_id, n_obs,
  time_min, time_max, prop_gap, median_dt_sec, iqr_dt_sec), \code{n_units},
  \code{n_obs}, \code{time_range}, \code{spacing}, \code{n_gaps}, \code{pct_gap}.
  If \code{ild_prepare()} was run on a \code{tbl_ts}, \code{tsibble} is also
  present (same as [ild_tsibble_meta()]). The \code{summary} tibble is the
  primary contract for programmatic use.
}
\description{
Reports number of persons, number of observations, time range,
descriptive spacing (median/IQR of intervals, percent gaps), and
duplicate info. Uses [ild_meta()] and `.ild_*` columns only.
No hard "regular"/"irregular" label; use [ild_spacing_class()] for that.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_person_model.R
\name{ild_person_distribution}
\alias{ild_person_distribution}
\title{Plot distribution of person-level estimates from ild_person_model}
\usage{
ild_person_distribution(
  person_fit,
  term = NULL,
  type = c("histogram", "density")
)
}
\arguments{
\item{person_fit}{Tibble returned by [ild_person_model()] (columns \code{term}, \code{estimate}, etc.).}

\item{term}{Character. Which term to plot (e.g. \code{"(Intercept)"} or a covariate name).
If \code{NULL}, the first term in the table is used.}

\item{type}{Character. \code{"histogram"} (default) or \code{"density"}.}
}
\value{
A ggplot object.
}
\description{
Draws a histogram or density of the selected term's estimates across persons.
Useful to visualize heterogeneity (e.g. distribution of slopes or intercepts).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_spacing.R
\name{ild_spacing}
\alias{ild_spacing}
\title{Spacing diagnostics and correlation-structure recommendation}
\usage{
ild_spacing(x, gap_large_hours = 12)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{gap_large_hours}{Numeric. Intervals (in hours) above which to count
as "large gaps" for \code{large_gaps_pct} (default 12). Ignored if time
is not in seconds (e.g. numeric day indices).}
}
\value{
A list with \code{median_interval} (hours), \code{iqr} (hours),
  \code{large_gaps_pct} (percent of intervals \code{> gap_large_hours}),
  \code{coefficient_of_variation}, \code{recommendation} (character: use CAR1 or AR1),
  and \code{spacing_class} (\code{regular-ish} or \code{irregular-ish}).
}
\description{
Reports observation intervals in human-friendly units (e.g. hours) and
recommends AR1 vs CAR1 for use in [ild_lme()]. Surfaces the same logic
that \code{ild_lme(..., ar1 = TRUE)} uses internally so users can see
why a correlation structure was chosen.
}
\examples{
d <- ild_simulate(n_id = 5, n_obs_per = 10, irregular = TRUE, seed = 1)
x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)
ild_spacing(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_circadian.R
\name{ild_circadian}
\alias{ild_circadian}
\title{Time-of-day pattern plot for ILD (circadian-style)}
\usage{
ild_circadian(x, var, type = c("boxplot", "line"))
}
\arguments{
\item{x}{An ILD object (see [is_ild()]) with \code{.ild_time} as POSIXct.}

\item{var}{Character or symbol. Variable to plot (e.g. mood, activity).}

\item{type}{Character. \code{"boxplot"} (default) or \code{"line"} (mean per hour with optional SE).}
}
\value{
A ggplot object (variable by hour of day).
}
\description{
Plots a variable by hour of day (or time-of-day) when \code{.ild_time} is
POSIXct. Useful for EMA (e.g. mood or activity by hour). Does not add
columns to the ILD object; hour is derived internally for plotting.
}
\examples{
d <- ild_simulate(n_id = 5, n_obs_per = 12, seed = 1)
x <- ild_prepare(d, id = "id", time = "time")
ild_circadian(x, y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_center.R
\name{ild_center}
\alias{ild_center}
\title{Within-person and between-person decomposition (centering)}
\usage{
ild_center(
  x,
  ...,
  type = c("person_mean", "grand_mean"),
  naming = c("suffix", "prefix")
)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{...}{Variables to center (tidy-select). Unquoted names or a single
character vector of column names. Must be numeric.}

\item{type}{Character. `"person_mean"` (default) for person-mean centering
(x_bp, x_wp); `"grand_mean"` for grand-mean centering (x_gm, x_wp_gm).}

\item{naming}{Character. `"suffix"` (default): new columns \code{var_bp}, \code{var_wp};
\code{"prefix"}: new columns \code{BP_var}, \code{WP_var}.}
}
\value{
The same ILD tibble with additional columns. ILD attributes are preserved.
}
\description{
For each selected variable, computes the person mean (between-person
component) and the within-person deviation (variable minus person mean).
Use `*_wp` at level-1 and `*_bp` at level-2 or in cross-level interactions
to avoid ecological fallacy and conflation bias. Selected variables must be numeric.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_spacing_class.R
\name{ild_spacing_class}
\alias{ild_spacing_class}
\title{Classify spacing as regular-ish vs irregular-ish}
\usage{
ild_spacing_class(x, cv_threshold = 0.2, pct_gap_threshold = 10)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{cv_threshold}{Numeric. Coefficient of variation of within-person
intervals above which spacing is "irregular-ish" (default 0.2).}

\item{pct_gap_threshold}{Numeric. Percent of intervals flagged as gaps
above which spacing is "irregular-ish" (default 10).}
}
\value{
Character: `"regular-ish"` or `"irregular-ish"`.
}
\description{
Returns a simple classification for use in documentation or when choosing
correlation structure (e.g. AR1 vs CAR1 in [ild_lme()]). The rule is
documented and overridable via arguments. Does not change core ILD behavior.
}
\seealso{
\href{../doc/kfas-irregular-timing-spacing.html}{\code{vignette("kfas-irregular-timing-spacing", package = "tidyILD")}}
(KFAS and irregular timing). \href{../doc/ild-decomposition-and-spacing.html}{\code{vignette("ild-decomposition-and-spacing", package = "tidyILD")}}
(within-between and spacing).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_fit.R
\name{ild_msm_diagnose}
\alias{ild_msm_diagnose}
\title{Diagnose an \code{ild_msm_fit} result in one call}
\usage{
ild_msm_diagnose(object, ...)
}
\arguments{
\item{object}{Output from [ild_msm_fit()].}

\item{...}{Passed to [ild_diagnose()].}
}
\value{
An \code{\link{ild_diagnostics_bundle}}.
}
\description{
Convenience bridge to [ild_diagnose()] using \code{object$fit} and
\code{object$weights_data}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_compare_pipelines.R
\name{ild_compare_pipelines}
\alias{ild_compare_pipelines}
\title{Compare provenance of two objects and report differences}
\usage{
ild_compare_pipelines(x1, x2)
}
\arguments{
\item{x1}{First object (ILD data, model fit, or diagnostics with provenance).}

\item{x2}{Second object (same types as x1).}
}
\value{
A list of class \code{ild_compare_pipelines} with \code{only_in_first}
  (step names only in x1), \code{only_in_second}, \code{differing} (list of
  per-step arg differences), and \code{summary} (character vector of
  human-readable differences). If either object has no provenance, returns
  a list with \code{message} and empty comparison components.
}
\description{
Flattens provenance for each object (data steps plus analysis steps), then
compares step sequences and key arguments. Useful to compare preprocessing
pipelines, model settings, or to see what changed between two analyses.
}
\examples{
set.seed(1)
d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 1)
x1 <- ild_prepare(d, id = "id", time = "time")
x1 <- ild_center(x1, y)
x2 <- ild_prepare(d, id = "id", time = "time")
x2 <- ild_center(x2, y)
ild_compare_pipelines(x1, x2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_kfas_tidiers.R
\name{ild_tidy_states}
\alias{ild_tidy_states}
\title{Tidy per-time state summaries for ild_kfas}
\usage{
ild_tidy_states(x, ...)
}
\arguments{
\item{x}{An object from [ild_kfas()].}

\item{...}{Reserved.}
}
\value{
A \code{\link[tibble]{tibble}}.
}
\description{
Returns a tibble with one row per time index and columns for the smoothed
level (and optional variance). Separate from [ild_tidy()] which tidies
**parameters**.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_panel_lag_prepare.R
\name{ild_panel_lag_prepare}
\alias{ild_panel_lag_prepare}
\title{Prepare several lag columns and audit them}
\usage{
ild_panel_lag_prepare(
  data,
  variables,
  n = 1L,
  mode = c("gap_aware", "index", "time_window"),
  max_gap = NULL,
  window = NULL,
  resolution = c("closest_prior", "last_in_window", "mean_in_window")
)
}
\arguments{
\item{data}{An ILD object (see [is_ild()]).}

\item{variables}{Character vector of column names to lag.}

\item{n}{Integer vector of lag orders, recycled to \code{length(variables)}.
Ignored for \code{mode = "time_window"} (see [ild_lag()]).}

\item{mode, max_gap, window, resolution}{Passed to each [ild_lag()] call.}
}
\value{
A list with \code{data} (ILD with lag columns), \code{lag_vars}
  (names of created columns), \code{check} (tibble from [ild_check_lags()]),
  and \code{spec} (list of arguments for provenance).
}
\description{
Sequentially applies [ild_lag()] to each named variable with the same
\code{mode}, \code{max_gap}, \code{window}, and \code{resolution}, then runs
[ild_check_lags()] once on all created lag columns. Use this when building
multivariate lag specifications (e.g. several predictors or lag orders) with
consistent gap or time-window semantics.
}
\seealso{
\code{vignette("temporal-dynamics-model-choice", package = "tidyILD")},
  [ild_crosslag()], [ild_lag()].
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_align.R
\name{ild_align}
\alias{ild_align}
\title{Align a secondary stream to primary ILD within a time window}
\usage{
ild_align(
  primary,
  secondary,
  value_var,
  window,
  time_secondary = "time",
  fun = c("mean", "median", "closest")
)
}
\arguments{
\item{primary}{An ILD object (see [is_ild()]); the stream to keep as rows.}

\item{secondary}{A data frame with id and time columns and the value variable(s) to align.}

\item{value_var}{Character. Name of the column in \code{secondary} to align (e.g. \code{"heart_rate"}).}

\item{window}{Numeric or lubridate duration. Time window (same units as \code{.ild_time_num},
e.g. seconds for POSIXct). Only secondary observations with time in
\code{(primary_time - window, primary_time]} are used.}

\item{time_secondary}{Character. Name of the time column in \code{secondary} (default \code{"time"}).}

\item{fun}{Character. Aggregation for values in window: \code{"mean"}, \code{"median"}, or \code{"closest"} (most recent in window).}
}
\value{
The primary data with a new column \code{<value_var>_aligned} (numeric; NA where no secondary obs in window).
}
\description{
For each row in the primary ILD, finds observations in the secondary
data set (same id, time within \code{window} before the primary time)
and attaches an aggregated value (e.g. mean, median, or closest).
Use when combining self-report with wearables or other streams that
have different timestamps.
}
\examples{
prim <- ild_prepare(
  data.frame(
    id = rep(1:2, each = 3),
    time = as.POSIXct(rep(c(0, 3600, 7200), 2), origin = "1970-01-01"),
    y = rnorm(6)
  ),
  id = "id", time = "time"
)
sec <- data.frame(
  id = rep(1:2, each = 4),
  time = as.POSIXct(rep(c(0, 1800, 3600, 5400), 2), origin = "1970-01-01"),
  heart_rate = 60 + rnorm(8, 0, 5)
)
ild_align(prim, sec, "heart_rate", window = 3600, fun = "mean")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_plot.R
\name{ild_spaghetti}
\alias{ild_spaghetti}
\title{ILD spaghetti / person trajectories (alias for ild_plot with type = "trajectory")}
\usage{
ild_spaghetti(x, var = NULL, facet_by = NULL, ...)
}
\arguments{
\item{x}{ILD object or fitted model.}

\item{var}{Variable to plot. If NULL, single data column is used.}

\item{facet_by}{Optional column in ILD for [ggplot2::facet_wrap()] (e.g. cluster).}

\item{...}{Passed to [ild_plot()] (e.g. max_ids, seed, id_var, time_var).}
}
\value{
A ggplot object.
}
\description{
Line plot of variable over time, one line per person. See [ild_plot()].
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_estimand.R
\name{ild_msm_estimand}
\alias{ild_msm_estimand}
\title{Define an MSM estimand specification}
\usage{
ild_msm_estimand(
  type = c("ate", "att"),
  regime = "static",
  treatment,
  time_var = ".ild_seq",
  contrast = NULL,
  target_time = "all",
  regime_value = 1,
  dynamic_rule = NULL
)
}
\arguments{
\item{type}{Estimand type. Supports \code{"ate"} and placeholder \code{"att"}.}

\item{regime}{Regime class. Supports \code{"static"} and scaffold \code{"dynamic"}.}

\item{treatment}{Binary treatment column name.}

\item{time_var}{Occasion index column (default \code{".ild_seq"}).}

\item{contrast}{Optional contrast definition. Character labels are accepted.
Lists may include \code{label}, \code{treated}, and \code{control}.}

\item{target_time}{Optional target occasion:
\code{"all"} (default), \code{"final"}, or numeric vector.}

\item{regime_value}{For \code{regime = "static"}, assignment target (default \code{1}).}

\item{dynamic_rule}{For \code{regime = "dynamic"}, a deterministic rule:
function taking \code{data} and returning binary assignment, or one of
\code{"always_treat"}, \code{"never_treat"}, \code{"as_observed"}.}
}
\value{
Object of class \code{ild_msm_estimand}.
}
\description{
Creates a lightweight estimand object used by [ild_msm_fit()].
v1.1 preserves backward compatibility with v1 static ATE calls while adding
explicit slots for regime specification, time targeting, and contrasts.
}
\examples{
e <- ild_msm_estimand(treatment = "trt")
e
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_generics.R, R/ild_brms_s3.R,
%   R/ild_ctsem_s3.R, R/ild_kfas_tidiers.R
\name{ild_tidy}
\alias{ild_tidy}
\alias{ild_tidy.lmerMod}
\alias{ild_tidy.lme}
\alias{ild_tidy.default}
\alias{ild_tidy.brmsfit}
\alias{ild_tidy.ild_fit_ctsem}
\alias{ild_tidy.ild_fit_kfas}
\title{Tidy fixed effects from an ILD model fit (S3 generic)}
\usage{
ild_tidy(x, ...)

\method{ild_tidy}{lmerMod}(x, ...)

\method{ild_tidy}{lme}(x, ...)

\method{ild_tidy}{default}(x, ...)

\method{ild_tidy}{brmsfit}(x, intervals = TRUE, object = FALSE, ...)

\method{ild_tidy}{ild_fit_ctsem}(x, ...)

\method{ild_tidy}{ild_fit_kfas}(x, ...)
}
\arguments{
\item{x}{A fitted model from [ild_lme()].}

\item{...}{Passed to [tidy_ild_model()].}

\item{intervals}{Logical. For \code{brmsfit} from [ild_brms()], include
posterior \code{rhat}, \code{ess_bulk}, \code{ess_tail} (default \code{TRUE}).}

\item{object}{Logical. Passed through for non-brms fits; ignored for \code{brmsfit}.}
}
\description{
Dispatches to [tidy_ild_model()] for \code{lmerMod} and \code{lme} objects from [ild_lme()],
or the \code{brmsfit} method for [ild_brms()]. All methods return tables conforming to
\code{\link{ild_tidy_schema}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_iptw_weights.R
\name{ild_iptw_weights}
\alias{ild_iptw_weights}
\title{Inverse probability of treatment weights (IPTW)}
\usage{
ild_iptw_weights(
  x,
  treatment,
  predictors,
  stabilize = TRUE,
  trim = c(0.01, 0.99),
  ...
)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{treatment}{Character. Name of a **binary** treatment column (numeric
\code{0}/\code{1} or two-level factor).}

\item{predictors}{Character vector of covariate names in \code{x} (main terms only).}

\item{stabilize}{Logical. If \code{TRUE} (default), stabilized weights
\eqn{P(A)/P(A|X)} for the observed \eqn{A}; otherwise \eqn{1/P(A|X)}.}

\item{trim}{Numeric of length 2. Quantiles for trimming (default \code{c(0.01, 0.99)}).}

\item{...}{Passed to \code{stats::glm()}.}
}
\value{
\code{x} with added column \code{.ipw_treat}. Attributes include
  \code{ild_iptw_fit} (the fitted \code{glm} object).
}
\description{
Fits a **single pooled** treatment propensity model (logistic regression)
over all person-occasions and computes stabilized or unstabilized IPTW weights
(one factor per row). Use this for a **time-invariant** treatment or descriptive
pooling. For **time-varying** \eqn{A_t} with a sequential MSM estimand, use
[ild_iptw_msm_weights()] instead. Outcome missingness IPW is
[ild_missing_model()] + [ild_ipw_weights()].
}
\details{
**Assumptions:** Positivity (\eqn{0 < P(A|X) < 1}), correct specification of
the treatment model. This is sensitivity / MSM-style tooling, not a substitute
for careful causal design.
}
\section{Bundle integration}{

Adds \code{.ipw_treat}. Use [ild_joint_msm_weights()] to combine with
\code{.ipw_censor} into \code{.ipw} for [ild_ipw_refit()] and [ild_diagnose()].
}

\examples{
set.seed(2)
d <- ild_simulate(n_id = 20, n_obs_per = 6, seed = 2)
d$stress <- rnorm(nrow(d))
d$trt <- as.integer(d$stress > 0)
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_iptw_weights(x, treatment = "trt", predictors = "stress")
summary(x$.ipw_treat)
}
\seealso{
[ild_iptw_msm_weights()], [ild_ipcw_weights()], [ild_joint_msm_weights()],
  [ild_ipw_refit()]

Other ild_diagnostics_utilities: 
\code{\link{ild_design_check}()},
\code{\link{ild_ipcw_weights}()},
\code{\link{ild_iptw_msm_weights}()},
\code{\link{ild_ipw_weights}()},
\code{\link{ild_joint_msm_weights}()},
\code{\link{ild_missing_cohort}()},
\code{\link{ild_missing_compliance}()},
\code{\link{ild_missing_hazard_first}()},
\code{\link{ild_missing_model}()},
\code{\link{ild_missing_pattern}()},
\code{\link{ild_missingness_report}()}
}
\concept{ild_diagnostics_utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_kfas.R
\name{ild_kfas}
\alias{ild_kfas}
\title{Fit a Gaussian state-space model via KFAS (single-subject ILD)}
\usage{
ild_kfas(
  data,
  outcome,
  state_spec = "local_level",
  observation_family = "gaussian",
  time_units,
  irregular_time = FALSE,
  smoother = TRUE,
  forecast_horizon = 0L,
  fit_context = c("single_series", "independent_series_per_id"),
  ...
)
}
\arguments{
\item{data}{An object that passes [validate_ild()] (e.g. after [ild_prepare()]).}

\item{outcome}{Character name of the numeric outcome column.}

\item{state_spec}{Character; use \code{"local_level"} in v1. See
\code{inst/dev/KFAS_V1_BACKEND.md} for the full controlled vocabulary.}

\item{observation_family}{Currently only \code{"gaussian"}.}

\item{time_units}{Character describing the time column (e.g. \code{"days"},
\code{"hours"}). Stored in provenance; required for clear reporting.}

\item{irregular_time}{Logical; if \code{TRUE}, skips the irregular-spacing warning.}

\item{smoother}{Logical; passed to \code{KFS()} smoothing (default \code{TRUE}).}

\item{forecast_horizon}{Integer; when \code{> 0}, used by \code{\link{ild_plot_forecast}()}
with \code{\link[stats]{predict}} on the fitted model and stored in \code{mapping$forecast_horizon}.}

\item{fit_context}{\code{"single_series"} (default) or \code{"independent_series_per_id"}
when you have fit the **same** state-space template separately for multiple persons
and must not interpret results as a pooled multilevel latent model. The latter
triggers guardrail \code{GR_KFAS_UNMODELED_BETWEEN_PERSON_HETEROGENEITY}.}

\item{...}{Passed to \code{KFAS::fitSSM()} (e.g. \code{method}, \code{inits}).}
}
\value{
An object of class \code{c("ild_fit_kfas", "ild_fit_model", "ild_analysis")}
  with list slots \code{kfas_model}, \code{kfs}, \code{spec}, \code{state_labels},
  \code{mapping}, \code{schema_version}, \code{preprocessing}. Attributes \code{ild_data} and
  \code{ild_provenance} are set.
}
\description{
Opinionated entry point for **local level** models in v1. Requires the
\pkg{KFAS} package. One distinct \code{.ild_id} per fit; subset your data first.
}
\seealso{
\code{inst/dev/KFAS_V1_BACKEND.md}, [ild_tidy()], [ild_augment()], [ild_diagnose()].

\strong{Conceptual vignettes:}
\itemize{
  \item \href{../doc/kfas-state-space-modeling.html}{\code{vignette("kfas-state-space-modeling", package = "tidyILD")}}
  \item \href{../doc/kfas-irregular-timing-spacing.html}{\code{vignette("kfas-irregular-timing-spacing", package = "tidyILD")}}
  \item \href{../doc/kfas-choosing-backend.html}{\code{vignette("kfas-choosing-backend", package = "tidyILD")}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_diagnostics_bundle.R
\name{ild_diagnostics_bundle}
\alias{ild_diagnostics_bundle}
\alias{new_ild_diagnostics_bundle}
\alias{is_ild_diagnostics_bundle}
\title{Canonical ILD diagnostics bundle}
\usage{
ild_diagnostics_bundle(
  meta = NULL,
  data = NULL,
  design = NULL,
  fit = NULL,
  residual = NULL,
  predictive = NULL,
  missingness = NULL,
  causal = NULL,
  warnings = NULL,
  guardrails = NULL,
  summary_text = character()
)

new_ild_diagnostics_bundle(
  meta = NULL,
  data = NULL,
  design = NULL,
  fit = NULL,
  residual = NULL,
  predictive = NULL,
  missingness = NULL,
  causal = NULL,
  warnings = NULL,
  guardrails = NULL,
  summary_text = character()
)

is_ild_diagnostics_bundle(x)
}
\arguments{
\item{meta, data, design, fit, residual, predictive, missingness, causal}{Optional sections
(list or structured object, or \code{NULL}).}

\item{warnings, guardrails}{Tibbles (default empty).}

\item{summary_text}{Character vector (default empty).}

\item{x}{Object to test.}
}
\value{
An object of class \code{ild_diagnostics_bundle}.

\code{is_ild_diagnostics_bundle}: logical.
}
\description{
The bundle has identical top-level \strong{slot names} for every estimation backend.
Content differs by engine; unavailable sections are \code{NULL}.
\code{warnings} and \code{guardrails} are always tibbles (possibly 0 rows).
Guardrails use the canonical schema documented in \code{\link{guardrail_registry}}:
\code{rule_id}, \code{section}, \code{severity}, \code{triggered}, \code{message},
\code{recommendation} (only triggered rules are stored as rows).
}
\details{
Several standalone functions **feed** these slots when you use \code{\link{ild_diagnose}};
see \code{\link{ild_diagnostics_utilities}} (\code{\link{ild_design_check}},
\code{\link{ild_missing_pattern}}, \code{\link{ild_missing_model}}, \code{\link{ild_ipw_weights}}).

Bundles returned by \code{\link{ild_diagnose}} also set \code{attr(bundle, "ild_fit")} and
\code{attr(bundle, "ild_data")} for \code{\link{ild_autoplot}}; re-run \code{ild_diagnose} if
these are missing (e.g. after loading a saved bundle without the fitted object).

The \code{\link{print}} method lists each slot and prints a short \strong{summary} line:
number of warning rows, number of guardrail rows, highest guardrail severity
(\code{info} < \code{warning}), and up to five triggered \code{rule_id} values.
}
\section{Slots}{

\describe{
  \item{\code{meta}}{List or \code{NULL}: run ids, engine, dimensions.}
  \item{\code{data}}{List or \code{NULL}: spacing, gaps, compliance, missingness, distributions.}
  \item{\code{design}}{List or \code{NULL}: within/between variation, time coverage, occasions, imbalance.}
  \item{\code{fit}}{List or \code{NULL}: convergence, singularity, optimizer / MCMC diagnostics.}
  \item{\code{residual}}{List or \code{NULL}: ACF, Q-Q, fitted vs observed, heteroskedasticity.}
  \item{\code{predictive}}{List or \code{NULL}: CV, PPC, forecast error.}
  \item{\code{missingness}}{List or \code{NULL}: IPW / imputation summaries when applicable.}
  \item{\code{causal}}{List or \code{NULL}: weights, positivity, causal diagnostics when applicable.}
  \item{\code{warnings}}{Tibble: structured warnings.}
  \item{\code{guardrails}}{Tibble: methodological guardrails (see \code{\link{guardrail_registry}}).}
  \item{\code{summary_text}}{Character: short narrative summary.}
}
}

\seealso{
\code{\link{ild_diagnose}}, \code{\link{guardrail_registry}},
  \code{\link{ild_diagnostics_utilities}},
  \code{\link{ild_tidy_schema}}, \code{\link{ild_augment_schema}}.
  Per-section field semantics: \code{vignette("developer-contracts", package = "tidyILD")}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_decomposition.R
\name{ild_decomposition}
\alias{ild_decomposition}
\title{Within-person and between-person variance decomposition}
\usage{
ild_decomposition(x, ..., plot = FALSE)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{...}{Variables to decompose (tidy-select). Must be numeric.}

\item{plot}{Logical. If \code{TRUE}, return a list with \code{table} and \code{plot} (WP vs BP density overlay). Default \code{FALSE}.}
}
\value{
A tibble with columns \code{variable}, \code{wp_var}, \code{bp_var}, \code{ratio} (wp_var / bp_var; \code{Inf} if bp_var is 0). If \code{plot = TRUE}, a list with \code{table} and \code{plot} (ggplot).
}
\description{
Reports WP and BP variance and their ratio for selected variables. Use as a
diagnostic and teaching tool: large ratio suggests within-person variance
dominates; small ratio suggests between-person differences dominate. Helps
avoid conflating WP and BP effects in modeling.
}
\examples{
d <- ild_simulate(n_id = 10, n_obs_per = 8, seed = 1)
x <- ild_prepare(d, id = "id", time = "time")
ild_decomposition(x, y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_missing_extensions.R
\name{ild_missing_compliance}
\alias{ild_missing_compliance}
\title{Person-level missingness compliance for one outcome}
\usage{
ild_missing_compliance(x, outcome, expected_occasions = NULL)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{outcome}{Character; column with possible `NA` (e.g. EMA outcome).}

\item{expected_occasions}{Optional positive integer: planned waves per person
for adherence (`pct_of_expected`). If persons have different row counts,
this is only a rough benchmark; see Details.}
}
\value{
A tibble with columns `.ild_id`, `n_rows`, `n_obs_outcome`,
  `pct_nonmissing_outcome`, `longest_run_observed`, `monotone_missing`
  (`NA` if no missing values on the outcome for that person), and if
  `expected_occasions` is set, `pct_of_expected` and `meets_expected_rows`.
}
\description{
Per person (after sorting by `.ild_seq`), returns coverage on the outcome,
longest streak of observed values, whether missingness is monotone (dropout
pattern: once missing, stays missing), and optional comparison to an expected
number of occasions.
}
\details{
**Monotone missing** is defined only when the person has at least one missing
outcome: all observations after the first missing are missing. Intermittent
missing yields `FALSE`. **Discrete hazard** for first missing is provided by
[ild_missing_hazard_first()] (cohort-level by `.ild_seq`).
}
\seealso{
[ild_missing_pattern()], [ild_missingness_report()]

Other ild_diagnostics_utilities: 
\code{\link{ild_design_check}()},
\code{\link{ild_ipcw_weights}()},
\code{\link{ild_iptw_msm_weights}()},
\code{\link{ild_iptw_weights}()},
\code{\link{ild_ipw_weights}()},
\code{\link{ild_joint_msm_weights}()},
\code{\link{ild_missing_cohort}()},
\code{\link{ild_missing_hazard_first}()},
\code{\link{ild_missing_model}()},
\code{\link{ild_missing_pattern}()},
\code{\link{ild_missingness_report}()}
}
\concept{ild_diagnostics_utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_model_tidiers.R
\name{tidy_ild_model}
\alias{tidy_ild_model}
\title{Tidy fixed effects from an ILD model fit}
\usage{
tidy_ild_model(
  fit,
  conf_level = 0.95,
  object = FALSE,
  se = c("model", "robust"),
  robust_type = c("CR2", "CR3", "CR0"),
  ...
)
}
\arguments{
\item{fit}{A fitted model from [ild_lme()] (lmerMod or lme).}

\item{conf_level}{Numeric. Confidence level for intervals (default 0.95).}

\item{object}{Logical. If \code{TRUE}, return a list with \code{meta} and \code{table}
and class \code{tidyild_model} for polished printing (default \code{FALSE}).}

\item{se}{Character. \code{"model"} (default) uses the model's standard errors;
\code{"robust"} uses cluster-robust SEs from [ild_robust_se()] (requires
\pkg{clubSandwich}).}

\item{robust_type}{Character. When \code{se = "robust"}, the correction type:
\code{"CR2"} (recommended), \code{"CR3"}, or \code{"CR0"}.}

\item{...}{Passed to [ild_robust_se()] when \code{se = "robust"} (e.g. \code{cluster}, \code{cluster_vec}).}
}
\value{
A tibble, or when \code{object = TRUE} a list of class \code{tidyild_model}.
}
\description{
Returns a tibble conforming to \code{\link{ild_tidy_schema}}: required columns include
\code{term}, \code{component}, \code{effect_level}, \code{estimate}, \code{std_error},
\code{conf_low}, \code{conf_high}, \code{statistic}, \code{p_value}, \code{interval_type},
\code{engine}, \code{model_class}; optional Bayesian columns are \code{NA} for these engines.
\code{component} is \code{"fixed"} for coefficient rows; \code{effect_level} is inferred
conservatively from term names (e.g. \verb{_wp}/\verb{_bp} suffixes) or \code{"population"}
for the intercept, \code{"unknown"} when ambiguous.
}
\details{
With \code{object = TRUE}, returns an object of class \code{tidyild_model} (meta + table) for use with
\code{print.tidyild_model}.
}
\section{Model-based vs robust SE}{

With \code{se = "model"} (default), standard errors and CIs come from the fitted
model. With \code{se = "robust"}, cluster-robust (sandwich) SEs are used;
CIs and p-values are based on a Wald normal approximation. Install the
\pkg{clubSandwich} package to use robust SEs. For **IPW / MSM** weighted \code{lmer}
fits ([ild_ipw_refit()]), robust SEs do not account for estimated weights; use
[ild_msm_bootstrap()] and [tidy_ild_msm_bootstrap()] when you need bootstrap CIs;
see [ild_msm_inference].
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_plot.R
\name{ild_plot_predicted_trajectory}
\alias{ild_plot_predicted_trajectory}
\title{Observed and fitted values vs time (trajectory overlay)}
\usage{
ild_plot_predicted_trajectory(
  fit,
  time_var = c(".ild_time_num", ".ild_seq"),
  max_ids = 20L,
  seed = 42L,
  facet_by = NULL
)
}
\arguments{
\item{fit}{Model from [ild_lme()] or [ild_brms()] (must carry `ild_data`).}

\item{time_var}{`.ild_time_num` or `.ild_seq` (default first of these in [match.arg()]).}

\item{max_ids, seed}{Passed through for subsampling persons.}

\item{facet_by}{Optional column name in `ild_data` for [ggplot2::facet_wrap()].}
}
\value{
A `ggplot` object.
}
\description{
Uses [augment_ild_model()] or [ild_augment()] on the fitted object and plots
two lines per person (observed vs fitted) against `time_var`. For a scatter of
observed vs fitted, use [ild_plot()] with `type = "fitted"`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_provenance.R
\name{ild_provenance}
\alias{ild_provenance}
\title{Return the raw provenance object}
\usage{
ild_provenance(x)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]) or an analysis object with \code{ild_provenance} attribute.}
}
\value{
For data: list with \code{version} and \code{steps}. For analysis: list with
  \code{version}, \code{source_data_provenance}, \code{analysis_steps}. \code{NULL} if none.
}
\description{
For ILD data: returns \code{attr(x, "tidyILD")$provenance} (version + steps from
preprocessing). For analysis objects (e.g. fits from [ild_lme()], [ild_diagnostics()],
[ild_tvem()], [ild_power()], [ild_missing_model()]): returns \code{attr(x, "ild_provenance")},
which has \code{source_data_provenance} (snapshot of data provenance) and
\code{analysis_steps} (list of analysis step records).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild-class.R
\name{is_ild}
\alias{is_ild}
\title{Check if an object is a valid ILD tibble}
\usage{
is_ild(x)
}
\arguments{
\item{x}{Any object.}
}
\value{
Logical.
}
\description{
Returns TRUE if the object has all required `.ild_*` columns and
`ild_*` metadata attributes (as set by [ild_prepare()]).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_history.R
\name{ild_msm_history_spec}
\alias{ild_msm_history_spec}
\title{Build a declarative MSM history specification}
\usage{
ild_msm_history_spec(
  vars,
  lags = 1L,
  mode = c("gap_aware", "index", "time_window"),
  max_gap = Inf
)
}
\arguments{
\item{vars}{Character vector of variable names to lag.}

\item{lags}{Integer vector of lag orders. Recycled across \code{vars}. Defaults to \code{1L}.}

\item{mode}{Passed to [ild_lag()] (default \code{"gap_aware"}).}

\item{max_gap}{Passed to [ild_lag()] for \code{mode = "gap_aware"}.}
}
\value{
Object of class \code{ild_msm_history_spec}.
}
\description{
Creates a lightweight spec object used by [ild_build_msm_history()] to generate
lagged confounder/treatment history columns with deterministic names (e.g.
\code{stress_lag1}, \code{trt_lag2}).
}
\examples{
s <- ild_msm_history_spec(vars = c("stress", "trt"), lags = 1:2)
s
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_design_check.R
\name{ild_design_check}
\alias{ild_design_check}
\title{ILD design diagnostics: spacing, WP/BP, missingness, and recommendations}
\usage{
ild_design_check(x, vars = NULL)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]).}

\item{vars}{Optional character vector of variable names for decomposition
and missingness. If \code{NULL}, only spacing and summary are computed;
\code{wp_bp} and \code{missingness} will be \code{NULL}.}
}
\value{
A list of class \code{ild_design_check}: \code{summary} (from ild_summary),
  \code{spacing_class} (regular-ish / irregular-ish), \code{spacing} (from ild_spacing),
  \code{recommendation} (AR1/CAR1 text), \code{wp_bp} (decomposition tibble or NULL),
  \code{missingness} (list with \code{summary} tibble and \code{pct_na} overall, or NULL).
  Use \code{print()} for a human-readable summary.
}
\description{
Aggregates [ild_summary()], [ild_spacing()], [ild_spacing_class()], and
optionally [ild_decomposition()] and [ild_missing_pattern()] into one
design summary. Use before modeling to see spacing class, correlation
recommendation, within- vs between-person variance, and missingness.
}
\details{
Also a **section provider** for [ild_diagnose()] (see [ild_diagnostics_utilities]).
}
\section{Bundle integration}{

[ild_diagnose()] embeds this object in \code{design$ild_design_check} on the
\code{\link{ild_diagnostics_bundle}}.
}

\examples{
d <- ild_simulate(n_id = 10, n_obs_per = 8, irregular = TRUE, seed = 1)
x <- ild_prepare(d, id = "id", time = "time", gap_threshold = 7200)
ild_design_check(x, vars = "y")
}
\seealso{
[ild_diagnose()], [ild_diagnostics_bundle()]

Other ild_diagnostics_utilities: 
\code{\link{ild_ipcw_weights}()},
\code{\link{ild_iptw_msm_weights}()},
\code{\link{ild_iptw_weights}()},
\code{\link{ild_ipw_weights}()},
\code{\link{ild_joint_msm_weights}()},
\code{\link{ild_missing_cohort}()},
\code{\link{ild_missing_compliance}()},
\code{\link{ild_missing_hazard_first}()},
\code{\link{ild_missing_model}()},
\code{\link{ild_missing_pattern}()},
\code{\link{ild_missingness_report}()}
}
\concept{ild_diagnostics_utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_bootstrap.R
\name{ild_msm_bootstrap}
\alias{ild_msm_bootstrap}
\title{Cluster bootstrap inference for weighted \code{lmer} (MSM / IPW sensitivity)}
\usage{
ild_msm_bootstrap(
  fit = NULL,
  formula = NULL,
  data = NULL,
  weights_col = ".ipw",
  n_boot = 200L,
  weight_policy = c("fixed_weights", "reestimate_weights"),
  weights_fn = NULL,
  cluster = c("id", "data"),
  cluster_vec = NULL,
  coef_fun = function(f) lme4::fixef(f),
  seed = NULL,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{fit}{A \code{lmerMod} from [ild_ipw_refit()] or [ild_lme()] with \code{attr(fit, "ild_data")}.
Ignored if \code{formula} and \code{data} are supplied.}

\item{formula}{Optional. If \code{fit} is \code{NULL}, mixed-model formula for refitting.}

\item{data}{Optional ILD object when \code{fit} is \code{NULL}; must contain \code{weights_col}.}

\item{weights_col}{Name of the row-level weight column in each bootstrap dataset (default \code{".ipw"}).}

\item{n_boot}{Number of bootstrap replicates.}

\item{weight_policy}{\code{"fixed_weights"} (subset rows; weights unchanged within replicated rows)
or \code{"reestimate_weights"} (apply \code{weights_fn} to each resampled dataset before fitting).}

\item{weights_fn}{Function \code{function(x)} taking ILD \code{x} and returning ILD with
\code{weights_col} present. Required when \code{weight_policy = "reestimate_weights"}.
Ignored for \code{"fixed_weights"}.}

\item{cluster}{\code{"id"} (default: cluster by \code{ild_meta(data)$ild_id}) or \code{"data"}.}

\item{cluster_vec}{When \code{cluster = "data"}, length-\code{nrow(data)} vector of cluster IDs
aligned to \code{ild_data} rows.}

\item{coef_fun}{\code{function(fit)} returning named numeric vector of estimands (default \code{lme4::fixef}).}

\item{seed}{Optional integer passed to \code{set.seed()} before resampling.}

\item{verbose}{Logical. If \code{TRUE}, warn on failed refits.}

\item{...}{Passed to \code{lme4::lmer()} on each successful replicate (e.g. \code{control}).
Do not pass \code{weights} here; use \code{weights_col} on the ILD data.}
}
\value{
An object of class \code{ild_msm_bootstrap}: \code{replicates} (matrix \code{n_boot} x \code{p}),
  \code{estimate} (point estimates from \code{fit}), \code{bootstrap_se}, \code{conf_low}, \code{conf_high},
  \code{term_names}, \code{n_boot}, \code{n_success}, \code{n_cluster}, metadata, and \code{fit} (original).
  Use [tidy_ild_msm_bootstrap()] for an \code{\link{ild_tidy_schema}} tibble.
}
\description{
Resamples **clusters** (persons) with replacement, subsets \code{attr(fit, "ild_data")}
(or \code{data}), optionally re-estimates weights, and refits \code{\link[lme4]{lmer}} with
the same formula. Collects fixed effects across replicates for bootstrap standard errors and
percentile confidence intervals.
}
\examples{
if (requireNamespace("lme4", quietly = TRUE)) {
  set.seed(5001)
  d <- ild_simulate(n_id = 12, n_obs_per = 6, seed = 5001)
  d$stress <- rnorm(nrow(d))
  d <- ild_prepare(d, id = "id", time = "time")
  d <- ild_center(d, y)
  d$.ipw <- runif(nrow(d), 0.8, 1.2)
  f0 <- ild_lme(y ~ y_bp + y_wp + stress + (1 | id), data = d, ar1 = FALSE,
    warn_no_ar1 = FALSE, warn_uncentered = FALSE)
  fw <- ild_ipw_refit(f0, data = d, weights = ".ipw")
  b <- ild_msm_bootstrap(fw, n_boot = 15L, weight_policy = "fixed_weights", seed = 2)
  print(b)
  tidy_ild_msm_bootstrap(b)
}
}
\seealso{
[ild_msm_inference], [ild_ipw_refit()], [tidy_ild_msm_bootstrap()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_methods.R
\name{ild_report}
\alias{ild_report}
\title{Assemble a light report from a model fit}
\usage{
ild_report(fit, export_provenance_path = NULL, robust_se = NULL, ...)
}
\arguments{
\item{fit}{A fitted model from [ild_lme()] (or a list with a \code{fit} component, e.g. from [ild_crosslag()]).}

\item{export_provenance_path}{Optional. If provided, [ild_export_provenance()] is called to write provenance to this path; the path is included in the returned list.}

\item{robust_se}{Optional. Passed to [ild_methods()] when building the methods text (e.g. \code{"CR2"} if you used [tidy_ild_model()] with \code{se = "robust"}).}

\item{...}{Unused.}
}
\value{
A list with a stable schema: \code{meta} (list with \code{n_obs}, \code{n_id}, \code{engine} when available), \code{methods}, \code{model_table}, \code{diagnostics_summary} (includes \code{guardrails_narrative} and \code{guardrails} when [ild_diagnose()] succeeds), \code{provenance}, \code{provenance_export_path} (character or NULL). When guardrails are triggered, \code{methods_with_guardrails} repeats the methods paragraph with guardrails appended (same pattern as \code{\link{ild_methods}(fit, bundle = d)}).
}
\description{
Builds a structured list with methods narrative (from [ild_methods()]),
fixed-effects table (from [tidy_ild_model()]), a short diagnostics summary,
and the raw provenance. Optionally exports provenance to a file.
}
\examples{
set.seed(1)
x <- ild_prepare(ild_simulate(n_id = 5, n_obs_per = 6, seed = 1), id = "id", time = "time")
fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
r <- ild_report(fit)
r$methods
r$model_table
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_balance.R
\name{ild_msm_balance_topic}
\alias{ild_msm_balance_topic}
\title{MSM and IPW balance diagnostics}
\description{
After building weights ([ild_iptw_weights()], [ild_iptw_msm_weights()], [ild_joint_msm_weights()]),
assess **covariate balance** (weighted SMD), **effective sample size** (ESS), and **overlap**
of fitted propensities. These complement weight summaries in [fill_diagnostics_causal()] and
guardrails on weight ranges.

**Weights:** Use \code{.ipw_treat} to check balance for the **treatment** mechanism;
use \code{.ipw} for balance in the **joint** MSM pseudo-population used in the weighted outcome
model—they need not coincide.
}
\seealso{
[ild_msm_balance()], [ild_ipw_ess()], [ild_msm_overlap_plot()],
  [ild_iptw_weights()], [ild_iptw_msm_weights()], [ild_joint_msm_weights()],
  [ild_msm_estimand()], [ild_msm_fit()], [ild_msm_recovery()]
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_schema_tidy_augment.R
\name{ild_tidy_schema}
\alias{ild_tidy_schema}
\title{Column dictionary for ild_tidy() outputs}
\usage{
ild_tidy_schema()
}
\value{
A list with \code{required} and \code{optional} character vectors.
}
\description{
Minimum stable contract for parameter-level tables returned by \code{\link{ild_tidy}}.
Implementations (\code{\link{tidy_ild_model}}, \code{\link{ild_tidy.brmsfit}}) emit all
required columns; \code{conf_low} / \code{conf_high} replace legacy \code{ci_low} / \code{ci_high}.
}
\details{
**Required columns:** \code{term}, \code{component}, \code{effect_level}, \code{estimate},
\code{std_error}, \code{conf_low}, \code{conf_high}, \code{statistic}, \code{p_value},
\code{interval_type}, \code{engine}, \code{model_class}.

**Optional:** \code{rhat}, \code{ess_bulk}, \code{ess_tail}, \code{pd}, \code{rope_low}, \code{rope_high}.

**\code{interval_type}:** Frequentist rows are typically \code{Wald} (model or robust; see \code{\link{tidy_ild_model}}).
Bootstrap output from \code{\link{tidy_ild_msm_bootstrap}} uses \code{bootstrap_percentile} (equal-tailed over replicate coefficients).
}
\seealso{
\code{\link{ild_augment_schema}}, \code{\link{ild_diagnostics_bundle}}.
  Full semantics for \code{component}, \code{effect_level}, and \code{interval_type}:
  \code{vignette("developer-contracts", package = "tidyILD")}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_methods.R
\name{ild_methods}
\alias{ild_methods}
\title{Generate methods-style narrative from provenance}
\usage{
ild_methods(x, robust_se = NULL, bundle = NULL, ...)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]), a model from [ild_lme()] or [ild_tvem()],
a diagnostics object from [ild_diagnostics()], or another object with
[ild_provenance()] (e.g. [ild_power()] result, [ild_missing_model()] result).}

\item{robust_se}{Optional. If you reported fixed effects with cluster-robust SEs
via [tidy_ild_model()] with \code{se = "robust"}, pass the type here (e.g. \code{"CR2"})
so the methods text can mention it.}

\item{bundle}{Optional. An object from [ild_diagnose()] (class
\code{\link{ild_diagnostics_bundle}}). When provided and
\code{nrow(bundle$guardrails) > 0}, a short methodological-cautions sentence
is appended after the provenance paragraph.}

\item{...}{Unused.}
}
\value{
A single character string (one or more sentences) suitable for a
  methods section. Use \code{cat()} or \code{print()} to display.
}
\description{
Takes an ILD data object, a model fit, or a diagnostics object and produces
a concise methods-style paragraph based on the recorded provenance (data
preparation, centering, lagging, modeling, etc.).
}
\details{
After fitting, call \code{diag <- ild_diagnose(fit)} and pass
\code{bundle = diag} to surface triggered guardrails in the methods narrative
without duplicating [ild_report()]'s structured \code{diagnostics_summary}.
}
\examples{
set.seed(1)
d <- ild_simulate(n_id = 5, n_obs_per = 6, seed = 1)
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_center(x, y)
ild_methods(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_provenance.R
\name{ild_export_provenance}
\alias{ild_export_provenance}
\title{Export provenance to a JSON or YAML file}
\usage{
ild_export_provenance(x, path, format = c("auto", "json", "yaml"))
}
\arguments{
\item{x}{An ILD object or an analysis object with [ild_provenance()].}

\item{path}{Character. File path to write (e.g. \code{"analysis_provenance.json"}).}

\item{format}{Character. \code{"auto"} (default) infers from file extension
(\code{.json} -> JSON, \code{.yaml} or \code{.yml} -> YAML). Use \code{"json"}
or \code{"yaml"} to force a format.}
}
\value{
The \code{path} invisibly, after writing the file.
}
\description{
Writes the full provenance structure (data or analysis) to a file for
reproducibility supplements, preregistration appendices, or lab archiving.
Requires the \pkg{jsonlite} package for JSON or \pkg{yaml} for YAML.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_manifest.R
\name{ild_bundle}
\alias{ild_bundle}
\title{Bundle a result with a reproducibility manifest}
\usage{
ild_bundle(result, manifest = NULL, label = NULL)
}
\arguments{
\item{result}{Any object (e.g. fitted model, diagnostics list).}

\item{manifest}{List. Reproducibility manifest from [ild_manifest()].
If \code{NULL}, [ild_manifest()] is called with default arguments.}

\item{label}{Optional character. Short label for the run (e.g.
\code{"model_ar1"} or \code{"diagnostics"}).}
}
\value{
A list with elements \code{result}, \code{manifest}, \code{label},
  suitable for [saveRDS()].
}
\description{
Combines a result (e.g. a fit from [ild_lme()] or output from
[ild_diagnostics()]) with a manifest and optional label for one-shot
saving. Typical use: \code{saveRDS(ild_bundle(fit, label = "model_ar1"), "run.rds")}.
You can build a manifest with [ild_manifest()] and pass \code{scenario}
(e.g. from [ild_summary()]) and \code{seed} before bundling.
}
\examples{
dat <- ild_prepare(ild_simulate(seed = 1), "id", "time")
fit <- ild_lme(y ~ 1 + (1 | id), dat, ar1 = FALSE, warn_no_ar1 = FALSE)
b <- ild_bundle(fit, label = "ar1")
names(b)
b <- ild_bundle(fit, manifest = ild_manifest(seed = 1, scenario = list(n_obs = 50)), label = "run1")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_brms.R
\name{ild_prior_ild}
\alias{ild_prior_ild}
\title{Prior specifications for common ILD mixed models}
\usage{
ild_prior_ild(
  template = c("default", "weakly_informative", "minimal_shrinkage")
)
}
\arguments{
\item{template}{Character selecting the template:
\describe{
  \item{\code{default}}{Student-t intercept; Normal(0, 5) fixed effects;
    Student-t group SDs; exponential residual SD (Gaussian models).}
  \item{\code{weakly_informative}}{Wider scales (more diffuse) on intercept
    and \code{b} coefficients.}
  \item{\code{minimal_shrinkage}}{Tighter Normal(0, 2.5) on \code{b} (similar
    to Gelman et al. scaled logistic guidance; use with scaled predictors).}
}}
}
\value{
A \code{brmsprior} data frame (class \code{brmsprior}).
}
\description{
Returns a \code{brmsprior} object suitable for the \code{prior} argument of
[ild_brms()] / [brms::brm()]. Templates follow weakly informative defaults:
Student-t on intercepts and group SDs, normal on regression coefficients, and
exponential on residual SD where applicable.
}
\seealso{
[ild_brms()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_acf.R
\name{ild_acf}
\alias{ild_acf}
\title{Autocorrelation function for ILD variables or model residuals}
\usage{
ild_acf(x, ..., by_id = FALSE)
}
\arguments{
\item{x}{Either an ILD object (see [is_ild()]) or a fitted model from [ild_lme()].}

\item{...}{When \code{x} is ILD data, the variable(s) to compute ACF on (tidy-select; one variable). Ignored when \code{x} is a fit.}

\item{by_id}{Logical. If \code{TRUE}, also return per-person ACF in \code{acf_by_id} (default \code{FALSE}).}
}
\value{
A list with \code{acf}: a tibble with columns \code{lag} and \code{acf} (pooled). If \code{by_id = TRUE}, \code{acf_by_id} is a named list of tibbles (one per person).
}
\description{
Computes ACF on a variable in ILD data or on residuals from an [ild_lme()] fit.
Use this to check whether AR1 is appropriate before fitting models. ACF is
computed over the ordered observation sequence (pooled or within person); it
does not adjust for irregular time gaps.
}
\examples{
d <- ild_simulate(n_id = 5, n_obs_per = 10, seed = 1)
x <- ild_prepare(d, id = "id", time = "time")
ild_acf(x, "y")
fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
ild_acf(fit)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_brms_dynamics_formula.R
\name{ild_brms_dynamics_formula}
\alias{ild_brms_dynamics_formula}
\title{Template brms formula for person-varying lag slope}
\usage{
ild_brms_dynamics_formula(outcome, lag_var, id_var = "id")
}
\arguments{
\item{outcome}{Character name of the response.}

\item{lag_var}{Character name of an existing lag column (e.g. from [ild_lag()]).}

\item{id_var}{Character name of the grouping column in \code{data} (default
\code{"id"} as stored after [ild_prepare()] — use the original id column name).}
}
\value{
A list: \code{formula} (\code{formula} object), \code{notes} (character).
}
\description{
Returns a suggested formula and notes for fitting with [ild_brms()].
This does **not** run the model. Identification, priors, and convergence
depend on N, T, and scaling; see \code{vignette("brms-dynamics-recipes", package = "tidyILD")}.
}
\seealso{
[ild_brms()], [ild_panel_lag_prepare()].
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_check_lags.R
\name{ild_check_lags}
\alias{ild_check_lags}
\title{Check lag variable validity (gap-aware)}
\usage{
ild_check_lags(x, lag_vars = NULL, max_gap = NULL)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]) that contains lag columns
(e.g. from [ild_lag()] with \code{mode = "gap_aware"}).}

\item{lag_vars}{Character vector of lag column names (e.g. \code{"y_lag1"}).
If \code{NULL}, attempts to detect columns ending in \code{_lag{n}} or \code{_lag_window}.}

\item{max_gap}{Numeric. Threshold used to define invalid (same units as
\code{.ild_time_num}). If \code{NULL}, uses \code{ild_meta(x)$ild_gap_threshold}.}
}
\value{
A data frame with one row per lag variable: \code{var}, \code{lag} (parsed lag order or \code{"window"}),
  \code{n_valid}, \code{n_invalid}, \code{n_first}, \code{n_total}, \code{pct_valid}, \code{pct_invalid}.
}
\description{
Given an ILD object and lag variable names, reports how many lagged
values are valid vs invalid (NA because the time distance to the
lagged row exceeded a threshold). Useful to audit lag columns before
modeling without re-specifying \code{max_gap}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_plot.R
\name{ild_plot}
\alias{ild_plot}
\title{ILD-specific plots}
\usage{
ild_plot(
  x,
  type = c("trajectory", "heatmap", "gaps", "missingness", "fitted", "fitted_vs_actual",
    "predicted_trajectory", "residual_acf"),
  var = NULL,
  id_var = ".ild_id",
  time_var = c(".ild_time_num", ".ild_seq"),
  max_ids = 20L,
  seed = 42L,
  facet_by = NULL,
  ...
)
}
\arguments{
\item{x}{An ILD tibble or a fitted [ild_lme()] model.}

\item{type}{Character (or vector). One or more of: `"trajectory"`, `"heatmap"`,
`"gaps"`, `"missingness"`, `"fitted"` or `"fitted_vs_actual"` (requires fitted model),
`"predicted_trajectory"` (observed and fitted lines vs time; requires fitted model),
`"residual_acf"` (requires fitted model; ACF is over observation sequence, not adjusted for
irregular time gaps). If length > 1, returns a named list of ggplots.}

\item{var}{For `trajectory` or `heatmap`, the variable to plot (optional;
if missing and only one non-.ild_* column exists, it is used).}

\item{id_var}{For trajectory, variable used for grouping (default `.ild_id`).}

\item{time_var}{For trajectory/gaps/`predicted_trajectory`, x-axis: `.ild_time_num` or `.ild_seq`.}

\item{max_ids}{For trajectory and `predicted_trajectory`, max number of persons to plot (sampled if
larger; default 20). Set to `Inf` to plot all.}

\item{seed}{Integer. Seed for sampling ids when `max_ids` is set (default 42).}

\item{facet_by}{Optional character: name of a column in the ILD (e.g. cluster or site) to pass to
[ggplot2::facet_wrap()] for `trajectory`, `heatmap`, `gaps`, and `predicted_trajectory`.}

\item{...}{Unused.}
}
\value{
A single ggplot when `length(type) == 1`, or a named list of ggplots when `length(type) > 1`.
}
\description{
Produces trajectory (spaghetti), heatmap, gaps, and (if a fitted model
is provided) fitted vs actual and residual ACF. Works for both lmerMod and
lme (ild_lme with ar1 = TRUE).
}
\examples{
x <- ild_prepare(ild_simulate(n_id = 3, n_obs_per = 6, seed = 1), id = "id", time = "time")
fit <- ild_lme(y ~ 1 + (1 | id), data = x, ar1 = FALSE, warn_no_ar1 = FALSE)
ild_plot(fit, type = "fitted_vs_actual")
ild_plot(fit, type = c("fitted_vs_actual", "residual_acf"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_contrast.R
\name{ild_msm_contrast_over_time}
\alias{ild_msm_contrast_over_time}
\title{Compute MSM contrasts over time from a fitted weighted model}
\usage{
ild_msm_contrast_over_time(
  object,
  treatment = NULL,
  time_var = ".ild_seq",
  target_time = "all",
  conf_level = 0.95
)
}
\arguments{
\item{object}{Output from [ild_msm_fit()] or a \code{lmerMod}.}

\item{treatment}{Optional treatment column name. Defaults to
\code{object$estimand$treatment} for \code{ild_msm_fit}, otherwise required.}

\item{time_var}{Occasion column (default \code{".ild_seq"}).}

\item{target_time}{Optional target subset: \code{"all"} (default), \code{"final"},
or numeric vector of occasions.}

\item{conf_level}{Confidence level for Wald intervals (default \code{0.95}).}
}
\value{
Tibble with `ild_tidy_schema` columns plus \code{method},
  \code{terms}, and \code{target_time}.
}
\description{
Computes marginal treatment contrasts by occasion using model fixed effects.
This helper is intentionally standalone: it accepts either an \code{ild_msm_fit}
object or a weighted \code{lmerMod} with attached \code{ild_data}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_ipw_refit.R
\name{ild_ipw_refit}
\alias{ild_ipw_refit}
\title{Refit an ILD model with inverse-probability weights (sensitivity analysis)}
\usage{
ild_ipw_refit(fit_or_formula, data, weights = ".ipw", ar1 = FALSE, ...)
}
\arguments{
\item{fit_or_formula}{Either a fitted model from [ild_lme()] (lmerMod), or a
formula. If a formula, \code{data} and \code{weights} are required.}

\item{data}{An ILD object (see [is_ild()]) containing a weight column (default \code{.ipw})
from [ild_ipw_weights()] or joint MSM weights from [ild_joint_msm_weights()], when refitting
from a fit (or the data to fit when \code{fit_or_formula} is a formula).}

\item{weights}{Character. Name of the weight column in \code{data} (default \code{".ipw"}).}

\item{ar1}{Logical. Must be \code{FALSE} for IPW refit (nlme weighted refit not supported).}

\item{...}{Passed to [ild_lme()] or \code{lme4::lmer()}.}
}
\value{
A fitted model (lmerMod) with \code{attr(..., "ild_data")} set so that
  [tidy_ild_model()] and [ild_diagnostics()] work. For inference under estimated
  weights, see [ild_msm_bootstrap()] and [ild_msm_inference].
}
\description{
Takes a fit from [ild_lme()] (or a formula and data) and refits the model
using observation weights from [ild_ipw_weights()]. Only the lme4 (lmer)
path is supported; nlme (ar1 = TRUE) is not supported for weighted refits.
This is a sensitivity tool, not a full MNAR solution.
}
\examples{
set.seed(1)
d <- ild_simulate(n_id = 12, n_obs_per = 10, seed = 1)
d$stress <- rnorm(nrow(d))
d$mood <- d$y
d$mood[sample(nrow(d), 25)] <- NA
x <- ild_prepare(d, id = "id", time = "time")
x <- ild_center(x, mood)
mm <- ild_missing_model(x, "mood", "stress")
xw <- ild_ipw_weights(x, mm, stabilize = TRUE)
fit0 <- ild_lme(mood ~ mood_bp + mood_wp + stress + (1 | id), data = xw,
  ar1 = FALSE, warn_no_ar1 = FALSE)
fitw <- ild_ipw_refit(fit0, data = xw)
tidy_ild_model(fitw)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_tsibble.R
\name{ild_as_tsibble}
\alias{ild_as_tsibble}
\title{Convert an ILD object to a tsibble}
\usage{
ild_as_tsibble(x, ...)
}
\arguments{
\item{x}{An object that passes [validate_ild()].}

\item{...}{Optional arguments passed to [tsibble::as_tsibble()] (e.g.
override \code{regular}).}
}
\value{
A \code{tbl_ts} object (see \pkg{tsibble}).
}
\description{
Wraps [tsibble::as_tsibble()] using the subject and time column names from
[ild_meta()] as key and index. When [ild_prepare()] recorded tsibble
provenance (\code{ild_tsibble_meta()} is non-\code{NULL}), the \code{regular}
argument is set from the stored \code{is_regular} flag so the reconstructed
\code{interval} often matches the original for unchanged data. Otherwise
defaults to \code{regular = TRUE}.
}
\examples{
\donttest{
d <- ild_simulate(n_id = 4, n_obs_per = 5, seed = 1)
x <- ild_prepare(d, id = "id", time = "time")
if (requireNamespace("tsibble", quietly = TRUE)) {
  ts <- ild_as_tsibble(x)
  class(ts)
}
}
}
\seealso{
\href{../doc/tsibble-interoperability.html}{\code{vignette("tsibble-interoperability", package = "tidyILD")}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_compare_fits.R
\name{ild_compare_fits}
\alias{ild_compare_fits}
\title{Compare fitted models (AIC, BIC, nobs)}
\usage{
ild_compare_fits(fits, guardrail_bundles = NULL)
}
\arguments{
\item{fits}{A named (or unnamed) list of fitted model objects.}

\item{guardrail_bundles}{Optional named list of [ild_diagnostics_bundle] objects
aligned with \code{fits} (same length and order). When provided, adds
\code{n_guardrails} per row.}
}
\value{
A tibble with columns \code{model}, \code{engine} (from [ild_methods()]
  truncated to one line when long), \code{aic}, \code{bic}, \code{n_obs},
  \code{converged} (best-effort), \code{n_guardrails} (if bundles given),
  and \code{notes} (fit-specific caveats).
}
\description{
Builds a compact comparison table for a named list of fitted objects.
Uses [stats::AIC()], [stats::BIC()], and [stats::nobs()] where those methods
exist. This is **not** a likelihood-ratio test and does not establish that
models are nested; interpret only when model comparison is scientifically
appropriate for your estimand.
}
\seealso{
\code{vignette("temporal-dynamics-model-choice", package = "tidyILD")}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_ipw_weights.R
\name{ild_ipw_weights}
\alias{ild_ipw_weights}
\title{Compute inverse-probability-of-observation weights from a missingness model}
\usage{
ild_ipw_weights(x, miss_fit, stabilize = TRUE, trim = c(0.01, 0.99))
}
\arguments{
\item{x}{An ILD object (see [is_ild()]); rows must align to the data used to fit \code{miss_fit}.}

\item{miss_fit}{The return value of [ild_missing_model()] (list with \code{fit}, \code{p_missing}, etc.).}

\item{stabilize}{Logical. If \code{TRUE} (default), use stabilized weights
\code{mean(p_obs) / p_obs}; otherwise \code{1 / p_obs}.}

\item{trim}{Numeric of length 2. Quantiles to trim weights (default \code{c(0.01, 0.99)}).
Weights below/above these quantiles are set to the quantile values.}
}
\value{
\code{x} with an added column \code{.ipw} (numeric). ILD attributes are preserved.
}
\description{
Uses the fitted missingness model from [ild_missing_model()] to compute
weights: unstabilized \code{w = 1 / p_obs} or stabilized
\code{w = mean(p_obs) / p_obs}. Weights are trimmed to avoid extremes.
Use with [ild_ipw_refit()] for sensitivity analysis. This is diagnostic
and sensitivity tooling, not a full MNAR solution.
}
\details{
Also a **section provider** for [ild_diagnose()] when \code{.ipw} is present on the
analysis data (see [ild_diagnostics_utilities]).
}
\section{Bundle integration}{

Adds \code{.ipw} to \code{x}. When that column is on the ILD data attached to the fit,
[ild_diagnose()] fills \code{causal} (weight summaries) on \code{\link{ild_diagnostics_bundle}}.
}

\examples{
set.seed(1)
d <- ild_simulate(n_id = 15, n_obs_per = 8, seed = 1)
d$stress <- rnorm(nrow(d))
d$mood <- d$y
d$mood[sample(nrow(d), 20)] <- NA
x <- ild_prepare(d, id = "id", time = "time")
mm <- ild_missing_model(x, "mood", "stress")
xw <- ild_ipw_weights(x, mm, stabilize = TRUE)
summary(xw$.ipw)
}
\seealso{
[ild_diagnose()], [ild_diagnostics_bundle()], [ild_missing_model()],
  [ild_iptw_weights()], [ild_ipcw_weights()], [ild_joint_msm_weights()]

Other ild_diagnostics_utilities: 
\code{\link{ild_design_check}()},
\code{\link{ild_ipcw_weights}()},
\code{\link{ild_iptw_msm_weights}()},
\code{\link{ild_iptw_weights}()},
\code{\link{ild_joint_msm_weights}()},
\code{\link{ild_missing_cohort}()},
\code{\link{ild_missing_compliance}()},
\code{\link{ild_missing_hazard_first}()},
\code{\link{ild_missing_model}()},
\code{\link{ild_missing_pattern}()},
\code{\link{ild_missingness_report}()}
}
\concept{ild_diagnostics_utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_kfas_plot.R
\name{ild_plot_states}
\alias{ild_plot_states}
\title{Plot smoothed latent states (first state by default)}
\usage{
ild_plot_states(x, state_index = 1L, ...)
}
\arguments{
\item{x}{An object from [ild_kfas()].}

\item{state_index}{Integer; which column of \code{alphahat} to plot (default \code{1L}).}

\item{...}{Passed to [ggplot2::labs()].}
}
\value{
A \code{ggplot} object.
}
\description{
Plot smoothed latent states (first state by default)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_balance.R
\name{ild_msm_balance}
\alias{ild_msm_balance}
\title{Covariate balance (weighted SMD) for MSM / IPW}
\usage{
ild_msm_balance(
  data,
  treatment,
  covariates,
  weights_col = ".ipw_treat",
  by_occasion = FALSE,
  time_var = ".ild_seq"
)
}
\arguments{
\item{data}{An ILD object with \code{treatment}, \code{covariates}, and \code{weights_col}.}

\item{treatment}{Character. Binary treatment column (\code{0}/\code{1} or two-level factor).}

\item{covariates}{Character vector of numeric or binary covariate columns.}

\item{weights_col}{Analysis weights (e.g. \code{".ipw_treat"} or \code{".ipw"}).}

\item{by_occasion}{If \code{TRUE}, stratify by \code{time_var}.}

\item{time_var}{Occasion index column (default \code{".ild_seq"}).}
}
\value{
A \code{tibble} with columns \code{stratum}, \code{covariate}, \code{smd},
  \code{mean_treated}, \code{mean_control}, \code{ess_stratum}, \code{weight_col}.
}
\description{
Computes **standardized mean differences** between weighted treated and control groups for
each named covariate. When \code{by_occasion = TRUE}, balance is computed **within** each
occasion (stratum); when \code{FALSE}, all rows are pooled into one pseudo-population.
}
\examples{
set.seed(2)
d <- ild_simulate(n_id = 12, n_obs_per = 6, seed = 2)
d$stress <- rnorm(nrow(d))
d$trt <- as.integer(stats::rbinom(nrow(d), 1L, 0.45))
x <- ild_prepare(d, id = "id", time = "time")
x$.ipw_treat <- runif(nrow(x), 0.8, 1.2)
ild_msm_balance(x, treatment = "trt", covariates = "stress", weights_col = ".ipw_treat")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_model_tidiers.R
\name{augment_ild_model}
\alias{augment_ild_model}
\title{Augment an ILD model fit with fitted values and residuals}
\usage{
augment_ild_model(fit, ...)
}
\arguments{
\item{fit}{A fitted model from [ild_lme()] (must have \code{attr(fit, "ild_data")}).}

\item{...}{Unused.}
}
\value{
A tibble; see \code{\link{ild_augment_schema}}.
}
\description{
Returns a tibble conforming to \code{\link{ild_augment_schema}}: \code{.ild_id},
\code{.ild_time}, \code{.outcome}, \code{.fitted}, \code{.resid}, \code{.resid_std},
\code{engine}, \code{model_class}, plus optional columns (see schema).
\code{.resid_std} is Pearson-type when \code{residuals(fit, type = "pearson")} is
available and length-matched; otherwise \code{NA} (principled but sparse).
Used internally by [ild_diagnostics()] and [ild_plot()]. Requires
\code{attr(fit, "ild_data")}; refit with [ild_lme()] if missing.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_provenance.R
\name{ild_history}
\alias{ild_history}
\title{Print a human-readable log of preprocessing or analysis steps}
\usage{
ild_history(x)
}
\arguments{
\item{x}{An ILD object (see [is_ild()]) or an analysis object with \code{ild_provenance}.}
}
\value{
The provenance object (from [ild_provenance()]) invisibly, or a message if none.
}
\description{
For ILD data, displays data provenance steps. For analysis objects (e.g. from
[ild_lme()], [ild_diagnostics()]), displays source data provenance (if any)
and analysis steps. Use [ild_provenance()] to get the raw structured object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_msm_history.R
\name{ild_build_msm_history}
\alias{ild_build_msm_history}
\title{Build MSM lagged history columns from a spec}
\usage{
ild_build_msm_history(data, spec, mode = NULL, max_gap = NULL)
}
\arguments{
\item{data}{An ILD object.}

\item{spec}{Object from [ild_msm_history_spec()].}

\item{mode}{Optional override for \code{spec$mode}.}

\item{max_gap}{Optional override for \code{spec$max_gap}.}
}
\value{
ILD object with lagged columns added; attributes include
  \code{ild_msm_history_manifest} (tibble with \code{variable}, \code{lag},
  \code{column}).
}
\description{
Applies [ild_lag()] repeatedly according to an [ild_msm_history_spec()] and
records a manifest under \code{attr(x, "ild_msm_history_manifest")}.
}
\examples{
d <- ild_simulate(n_id = 6, n_obs_per = 5, seed = 1)
d$stress <- rnorm(nrow(d))
d$trt <- rbinom(nrow(d), 1, 0.4)
d <- ild_prepare(d, id = "id", time = "time")
hs <- ild_msm_history_spec(c("stress", "trt"), lags = 1:2)
d2 <- ild_build_msm_history(d, hs)
attr(d2, "ild_msm_history_manifest")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild-class.R
\name{as_ild}
\alias{as_ild}
\title{Coerce to ILD object}
\usage{
as_ild(x)
}
\arguments{
\item{x}{A data frame or tibble that may already be ILD-shaped.}
}
\value{
An ILD tibble with class \code{tidyild_df} and \code{ild_tbl}.
}
\description{
If the object already has the required `.ild_*` columns and
attributes, validates and returns it (with \code{tidyild_df} and \code{ild_tbl} class if missing).
Otherwise errors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ild_tsibble.R
\name{ild_tsibble_meta}
\alias{ild_tsibble_meta}
\title{Tsibble provenance from an ILD object}
\usage{
ild_tsibble_meta(x)
}
\arguments{
\item{x}{An object that passes [validate_ild()].}
}
\value{
A named list or \code{NULL}.
}
\description{
If [ild_prepare()] was called on a \code{tbl_ts}, metadata from the source
tsibble (key, index, interval summary, regularity) is stored in
\code{attr(x, "tidyILD")$tsibble}. Returns \code{NULL} if the object was
not prepared from a tsibble.
}
\seealso{
[ild_prepare()], [ild_as_tsibble()],
  \href{../doc/tsibble-interoperability.html}{\code{vignette("tsibble-interoperability", package = "tidyILD")}}
}
